(* A formalisation of the spi calculus in Coq *)
(* Copyright (C) 2005-2007 Sebastien Briais *)
(* http://lamp.epfl.ch/~sbriais/ *)

(* This library is free software; you can redistribute it and/or modify *)
(* it under the terms of the GNU Lesser General Public License as *)
(* published by the Free Software Foundation; either version 2.1 of the *)
(* License, or (at your option) any later version. *)

(* This library is distributed in the hope that it will be useful, but *)
(* WITHOUT ANY WARRANTY; without even the implied warranty of *)
(* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU *)
(* Lesser General Public License for more details. *)

(* You should have received a copy of the GNU Lesser General Public *)
(* License along with this library; if not, write to the Free Software *)
(* Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA *)
(* 02110-1301 USA *)

Require Import deBruijnType.
Require Import name.
Require Import messageSpi.
Require Import expressionSpi.
Require Import missing.
Require Import ordmissing.
Require Import tactics.
Require Import FSets.
Require Import hedges.
Require Import calculi.
Require Import reductionSpi.
Require Import actionSpi.
Require Import evalSpi.
Require Import hedgedRelations.
Require Import permutation.

Module EarlyHedgedBisimulation(H:Hedge).
  Module HR := HedgedRelations(H).
  Export HR.

  Definition early_hedged_bisimulation_condition (R:hedged_relation_type) :=
    forall (h:H.t)(p q:AgentSpi.P.t),(R (h,p,q))->
  (** silent transitions *)
      (forall (p':AgentSpi.P.t),
        (RedSpiCong.R.reduction p ActionSpi.Tau (AgentSpi.Pr p'))->
        (exists q':AgentSpi.P.t,RedSpiCong.R.reduction q ActionSpi.Tau (AgentSpi.Pr q') /\ exists h':H.t,(H.Equal h h')/\(R (h',p',q')))) /\
  (** input transitions *)
      (forall (a:Name.t)(p':AgentSpi.P.t),
        (RedSpiCong.R.reduction p (ActionSpi.Inp a) (AgentSpi.F (AgentSpi.Abs p')))->
        forall (b:Name.t),(H.In (MessageSpi.MNam a,MessageSpi.MNam b) h)->
          forall (B:H.t)(m n:MessageSpi.message),(possible_inputs h B m n)->
            (exists q':AgentSpi.P.t,RedSpiCong.R.reduction q (ActionSpi.Inp b) (AgentSpi.F (AgentSpi.Abs q')) /\ 
              exists h':H.t,(H.Equal h' (H.union h B))/\(R (h',subst_input m p',subst_input n q')))) /\
  (** output transitions *)
      (forall (a:Name.t)(c:nat)(m:MessageSpi.message)(p':AgentSpi.P.t),
        (RedSpiCong.R.reduction p (ActionSpi.Outp a) (AgentSpi.C (AgentSpi.Conc c (EvalSpi.expr_of_mesg m) p')))->
        forall (tildec:list Name.t),(HOp.n_fresh_left h c tildec)->
        forall (b:Name.t),(H.In (MessageSpi.MNam a,MessageSpi.MNam b) h)->
          (exists d:nat,exists n:MessageSpi.message,exists q':AgentSpi.P.t,RedSpiCong.R.reduction q (ActionSpi.Outp b) (AgentSpi.C (AgentSpi.Conc d (EvalSpi.expr_of_mesg n) q')) /\
            exists tilded:list Name.t,(HOp.n_fresh_right h d tilded)/\
              exists h':H.t,(H.Equal h' (HOp.hedge_irreducible (H.add (instantiate_message tildec m,instantiate_message tilded n) h)))/\
                (R (h',(instantiate_process tildec p'),(instantiate_process tilded q'))))).

  (** R is a early hedged bisimulation iff it is a symmetric left
     consistent hedged relation that satisfies the bisimulation clause *)
  Definition early_hedged_bisimulation (R:hedged_relation_type) := 
    (symmetric_left_consistent_hedged_relation R)/\(early_hedged_bisimulation_condition R).

  (** early hedged bisimilarity *)
  Definition early_hedged_bisimilar (h:H.t)(p q:AgentSpi.P.t) := 
    exists R:hedged_relation_type,(R (h,p,q))/\(early_hedged_bisimulation R).

  (** Definition of progress (early hedged bisimulation clause) *)
  Definition early_hedged_bisimulation_condition_progress (R S:hedged_relation_type) :=
    forall (h:H.t)(p q:AgentSpi.P.t),(R (h,p,q))->
  (** silent transitions *)
      (forall (p':AgentSpi.P.t),
        (RedSpiCong.R.reduction p ActionSpi.Tau (AgentSpi.Pr p'))->
        (exists q':AgentSpi.P.t,RedSpiCong.R.reduction q ActionSpi.Tau (AgentSpi.Pr q') /\ exists h':H.t,(H.Equal h h')/\(S (h',p',q')))) /\
  (** input transitions *)
      (forall (a:Name.t)(p':AgentSpi.P.t),
        (RedSpiCong.R.reduction p (ActionSpi.Inp a) (AgentSpi.F (AgentSpi.Abs p')))->
        forall (b:Name.t),(H.In (MessageSpi.MNam a,MessageSpi.MNam b) h)->
          forall (B:H.t)(m n:MessageSpi.message),(possible_inputs h B m n)->
            (exists q':AgentSpi.P.t,RedSpiCong.R.reduction q (ActionSpi.Inp b) (AgentSpi.F (AgentSpi.Abs q')) /\ 
              exists h':H.t,(H.Equal h' (H.union h B))/\(S (h',subst_input m p',subst_input n q')))) /\
  (** output transitions *)
      (forall (a:Name.t)(c:nat)(m:MessageSpi.message)(p':AgentSpi.P.t),
        (RedSpiCong.R.reduction p (ActionSpi.Outp a) (AgentSpi.C (AgentSpi.Conc c (EvalSpi.expr_of_mesg m) p')))->
        forall (tildec:list Name.t),(HOp.n_fresh_left h c tildec)->
        forall (b:Name.t),(H.In (MessageSpi.MNam a,MessageSpi.MNam b) h)->
          (exists d:nat,exists n:MessageSpi.message,exists q':AgentSpi.P.t,RedSpiCong.R.reduction q (ActionSpi.Outp b) (AgentSpi.C (AgentSpi.Conc d (EvalSpi.expr_of_mesg n) q')) /\
            exists tilded:list Name.t,(HOp.n_fresh_right h d tilded)/\
              exists h':H.t,(H.Equal h' (HOp.hedge_irreducible (H.add (instantiate_message tildec m,instantiate_message tilded n) h)))/\
                (S (h',(instantiate_process tildec p'),(instantiate_process tilded q'))))).

  Lemma early_hedged_bisimulation_condition_iff_progress_itself : forall (R:hedged_relation_type),(early_hedged_bisimulation_condition R)<->(early_hedged_bisimulation_condition_progress R R).
    unfold early_hedged_bisimulation_condition_progress, early_hedged_bisimulation_condition.
    tauto.
  Qed.

  Lemma early_hedged_bisimulation_condition_progress_empty_hedged_relation : forall (R:hedged_relation_type),(early_hedged_bisimulation_condition_progress empty_hedged_relation R).
    unfold early_hedged_bisimulation_condition_progress.
    intros.
    elim H.
  Qed.

  Lemma early_hedged_bisimulation_condition_progress_inclusion :
    forall (R R' S S':hedged_relation_type),
      (inclusion_hedged_relation R R')->
      (inclusion_hedged_relation S' S)->
      (early_hedged_bisimulation_condition_progress R' S')->
      (early_hedged_bisimulation_condition_progress R S).
    intros.
    unfold early_hedged_bisimulation_condition_progress.
    intros.
    elim (H _ _ _ H2).
    intro h0;intros;and_elim.
    elim (H1 _ _ _ H4).
    intros;and_elim.
    repeat (split;intros).
    (* Tau *)
    clear H6;clear H7.
    elim (H5 _ H8).
    intro q';intros;and_elim.
    exists q'.
    split;auto.
    ex_elim H7 h1.
    elim (H0 _ _ _ H9).
    intro h';intros;and_elim.
    exists h'.
    split;auto.
    eapply HOp.HedgeProp.equal_trans.
    apply H3.
    eapply HOp.HedgeProp.equal_trans;eauto.
    (* Inp *)
    clear H5;clear H7.
    elim (H3 (MessageSpi.MNam a,MessageSpi.MNam b));intros.
    generalize (HOp.HedgeProp.subset_equal (HOp.HedgeProp.equal_sym H3));intro.
    generalize (HOp.HedgeProp.subset_equal (HOp.HedgeProp.equal_refl B));intro.
    assert (possible_inputs h0 B m n).
    eapply possible_inputs_Subset.
    apply HOp.HedgeProp.subset_equal.
    apply H3.
    eapply HOp.are_fresh_names_Subset.
    apply H11.
    apply H12.
    elim H10;intros;auto.
    apply HOp.HedgeProp.equal_refl.
    apply H10.
    elim (H6 _ _ H8 _ (H5 H9) _ _ _ H13).
    intro q';intros.
    clear H5;clear H7.
    and_elim.
    exists q'.
    split;auto.
    ex_elim H5 h1.
    elim (H0 _ _ _ H7).
    intro h';intros;and_elim.
    exists h'.
    split;auto.
    eapply HOp.HedgeProp.equal_trans.
    apply HOp.HedgeProp.equal_sym.
    apply H15.
    eapply HOp.HedgeProp.equal_trans.
    apply H5.
    apply HOp.HedgeProp.union_equal_1.
    apply HOp.HedgeProp.equal_sym;auto.
    (* Outp *)
    clear H5;clear H6.
    elim (H3 (MessageSpi.MNam a,MessageSpi.MNam b)).
    intros.
    generalize (HOp.n_fresh_left_Subset _ _ _ _ (HOp.HedgeProp.subset_equal (HOp.HedgeProp.equal_sym H3)) H9).
    intro.
    elim (H7 _ _ _ _ H8 _ H11 _ (H5 H10)).
    clear H5;clear H6.
    intro d;intros.
    ex_elim H5 n.
    ex_elim H5 q'.
    exists d.
    exists n.
    exists q'.
    split;auto.
    ex_elim H6 tilded.
    generalize (HOp.n_fresh_right_Subset _ _ _ _ (HOp.HedgeProp.subset_equal H3) H6).
    intro.
    exists tilded.
    split;auto.
    ex_elim H12 h1.
    clear H7.
    elim (H0 _ _ _ H14).
    intro h';intros;and_elim.
    exists h'.
    split;auto.
    eapply HOp.HedgeProp.equal_trans.
    apply HOp.HedgeProp.equal_sym;eauto.
    eapply HOp.HedgeProp.equal_trans.
    apply H12.
    apply HOp.hedge_irreducible_morphism.
    apply HOp.HedgeOtherProp.compat_op_add;auto.
    apply HOp.HedgeProp.equal_sym;auto.
  Qed.
    
  Lemma same_hedged_relation_early_hedged_bisimulation_condition : 
    forall (R R':hedged_relation_type),
      (same_hedged_relation R R')->
      (early_hedged_bisimulation_condition R)->
      (early_hedged_bisimulation_condition R').
    intros.
    elim H;intros.
    elim (early_hedged_bisimulation_condition_iff_progress_itself R);intros.
    elim (early_hedged_bisimulation_condition_iff_progress_itself R');intros.
    apply H6.
    eapply early_hedged_bisimulation_condition_progress_inclusion;eauto.
  Qed.

  Lemma same_hedged_relation_early_hedged_bisimulation : 
    forall (R R':hedged_relation_type),
      (same_hedged_relation R R')->
      (early_hedged_bisimulation R)->
      (early_hedged_bisimulation R').
    unfold early_hedged_bisimulation;intros;and_elim.
    split.
    eapply same_hedged_relation_symmetric_left_consistent_hedged_relation;eauto.
    eapply same_hedged_relation_early_hedged_bisimulation_condition;eauto.
  Qed.

  Lemma early_hedged_bisimulation_condition_progress_families_aux : 
    forall (R S:hedged_relation_family_type),
      (forall (i:nat),exists j:nat,early_hedged_bisimulation_condition_progress (R i) (S j))->
      early_hedged_bisimulation_condition_progress (union_hedged_relation_family R) (union_hedged_relation_family S).
    intros.
    unfold early_hedged_bisimulation_condition_progress.
    intros.
    elim H0.
    intro i;intros.
    elim (H i);intro j;intros.
    generalize (H2 _ _ _ H1).
    intros.
    and_elim.
    split;intros.
    (* Tau *)
    clear H4;clear H5.
    elim (H3 _ H6).
    intro q';intros;and_elim.
    exists q'.
    split;auto.
    ex_elim H5 h'.
    exists h'.
    split;auto.
    exists j.
    auto.
    (* Inp *)
    clear H3.
    split;intros.
    clear H5.
    elim (H4 _ _ H3 _ H6 _ _ _ H7).
    intro q';intros;and_elim.
    exists q'.
    split;auto.
    ex_elim H8 h'.
    exists h'.
    split;auto.
    exists j;auto.
    (* Outp *)
    clear H4.
    elim (H5 _ _ _ _ H3 _ H6 _ H7).
    intro d;intros.
    ex_elim H4 n.
    ex_elim H4 q'.
    exists d.
    exists n.
    exists q'.
    split;auto.
    ex_elim H8 tilded.
    exists tilded.
    split;auto.
    ex_elim H9 h'.
    exists h'.
    split;auto.
    exists j.
    auto.
  Qed.

  Lemma early_hedged_bisimulation_condition_progress_families : 
    forall (R:hedged_relation_family_type),
      (forall (i:nat),exists j:nat,early_hedged_bisimulation_condition_progress (R i) (R j))->
      early_hedged_bisimulation_condition (union_hedged_relation_family R).
    intros.
    elim (early_hedged_bisimulation_condition_iff_progress_itself (union_hedged_relation_family R)).
    intros.
    apply H1.
    apply early_hedged_bisimulation_condition_progress_families_aux.
    intro.
    elim (H i).
    intro j;intro.
    exists j.
    auto.
  Qed.

  Lemma early_hedged_bisimulation_condition_progress_union : 
    forall (R R' S S':hedged_relation_type),
      (early_hedged_bisimulation_condition_progress R S)->
      (early_hedged_bisimulation_condition_progress R' S')->
      early_hedged_bisimulation_condition_progress (union_hedged_relation R R') (union_hedged_relation S S').
    intros.
    elim (same_hedged_relation_union_hedged_relation_family R R');intros.
    elim (same_hedged_relation_union_hedged_relation_family S S');intros.
    eapply early_hedged_bisimulation_condition_progress_inclusion;eauto.
    clear H1;clear H2;clear H3;clear H4.
    apply early_hedged_bisimulation_condition_progress_families_aux.
    destruct i.
    exists 0.
    auto.
    destruct i.
    exists 1.
    auto.
    exists 0.
    apply early_hedged_bisimulation_condition_progress_empty_hedged_relation.
  Qed.


  Definition hedged_relation_sound (R:hedged_relation_type) := exists S:hedged_relation_type,(inclusion_hedged_relation R S)/\(early_hedged_bisimulation S).

  Definition early_hedged_bisimulation_progress (R S:hedged_relation_type) := 
    (symmetric_left_consistent_hedged_relation R)/\
    (symmetric_left_consistent_hedged_relation S)/\
    (early_hedged_bisimulation_condition_progress R S).

  Lemma early_hedged_bisimulation_iff_progress_itself :
    forall (R:hedged_relation_type),
      (early_hedged_bisimulation R)<->(early_hedged_bisimulation_progress R R).
    unfold early_hedged_bisimulation_progress, early_hedged_bisimulation.
    split;intros;and_elim.
    repeat (split;auto).
    repeat (split;auto).
  Qed.

  Definition early_hedged_bisimulation_upto (R:hedged_relation_type)(f:hedged_relation_operator) := early_hedged_bisimulation_progress R (f R).

  Definition hedged_relation_operator_safe (f:hedged_relation_operator) := forall (R:hedged_relation_type),(early_hedged_bisimulation_upto R f)->(hedged_relation_sound R).

  Definition hedged_relation_operator_respectful (f:hedged_relation_operator) := 
    forall (R R':hedged_relation_type),
      (inclusion_hedged_relation R R')->
      (early_hedged_bisimulation_condition_progress R R')->
      (inclusion_hedged_relation (f R) (f R'))/\(early_hedged_bisimulation_condition_progress (f R) (f R')).

  Lemma progress_hedged_relation_operator_iter : 
    forall (f:hedged_relation_operator),
      (hedged_relation_operator_respectful f)->
      forall (R:hedged_relation_type),(early_hedged_bisimulation_condition_progress R (f R))->
        forall (n:nat),
          (early_hedged_bisimulation_condition_progress (hedged_relation_operator_iter f R n) (hedged_relation_operator_iter f R (S n))).
    intros.
    induction n.
    simpl.
    eapply early_hedged_bisimulation_condition_progress_inclusion.
    apply inclusion_hedged_relation_refl.
    apply inclusion_union_2.
    apply H0.
    simpl.
    apply early_hedged_bisimulation_condition_progress_union.
    auto.
    unfold hedged_relation_operator_respectful in H.
    elim (H _ _ (inclusion_hedged_relation_operator_iter f R n) IHn).
    intros.
    apply H2.
  Qed.

  Lemma progress_hedged_relation_operator_iter_limit : 
    forall (f:hedged_relation_operator),
      (hedged_relation_operator_respectful f)->
      forall (R:hedged_relation_type),(early_hedged_bisimulation_condition_progress R (f R))->
        (early_hedged_bisimulation_condition (hedged_relation_operator_limit f R)).
    intros.
    unfold hedged_relation_operator_limit.
    apply early_hedged_bisimulation_condition_progress_families.    
    intro i.
    exists (S i).
    apply progress_hedged_relation_operator_iter.   
    auto.
    auto.
  Qed.

  Lemma hedged_relation_operator_respectful_is_safe : 
    forall (f:hedged_relation_operator),
      (hedged_relation_operator_correct f)->
      (hedged_relation_operator_respectful f)->
      (hedged_relation_operator_safe f).
    intros.
    unfold hedged_relation_operator_safe.
    intros.
    unfold hedged_relation_sound.
    exists (hedged_relation_operator_limit f R).
    split.
    apply inclusion_hedged_relation_trans with (hedged_relation_operator_iter f R 1).
    simpl.
    apply inclusion_union_1.
    unfold hedged_relation_operator_limit.
    apply inclusion_union_hedged_relation_family.
    split.
    apply hedged_relation_operator_correct_limit;auto.
    elim H1;intros.
    auto.
    apply progress_hedged_relation_operator_iter_limit.  
    auto.
    elim H1;intros.
    elim H3;auto.
  Qed.

  Lemma f_structural_congruence_respectful : hedged_relation_operator_respectful f_structural_congruence.
    unfold hedged_relation_operator_respectful.
    intros.
    split.
    apply f_structural_congruence_inclusion.
    auto.
    unfold early_hedged_bisimulation_condition_progress.
    intros.
    split.
    intros.
    ex_elim H1 p0.
    ex_elim H1 q0.
    elim (reduction_spi_congruence_silent _ _ H2 _ H3).
    intro p'0;intros;and_elim.
    elim (H0 _ _ _ H1).
    intros.
    clear H8.
    elim (H7 _ H5).
    intro q'0;intros;and_elim.
    elim (reduction_spi_congruence_silent _ _ H8 _ (CongruenceSpi.StructuralCongruence_symmetric _ _ H4)).
    intro q';intros.
    and_elim.
    exists q'.
    split.
    auto.
    ex_elim H9 h'.
    exists h'.
    split.
    auto.
    exists p'0.
    exists q'0.
    split;auto.
    split;auto.
    apply CongruenceSpi.StructuralCongruence_symmetric;auto.
    split.
    intros.
    ex_elim H1 p0.
    ex_elim H1 q0.
    elim (reduction_spi_congruence_input _ _ _ H2 _ H5).
    intro p'0;intros;and_elim.
    elim (H0 _ _ _ H1).
    intros.
    clear H9.
    elim H10;clear H10;intros.
    clear H10.
    elim (H9 _ _ H7 _ H3 _ _ _ H4).
    intro q'0;intros;and_elim.
    elim (reduction_spi_congruence_input _ _ _ H10 _ (CongruenceSpi.StructuralCongruence_symmetric _ _ H6)).
    intro q';intros;and_elim.
    exists q'.
    split;auto.
    ex_elim H11 h'.
    exists h'.
    split;auto.
    exists (subst_input m p'0).
    exists (subst_input n q'0).
    split.
    auto.
    unfold subst_input.
    split.
    apply CongruenceSpi.Rcongruence_low_subst.
    apply CongruenceSpi.StructuralCongruenceAxioms_preserves_low_subst.
    auto.
    apply CongruenceSpi.Rcongruence_low_subst.
    apply CongruenceSpi.StructuralCongruenceAxioms_preserves_low_subst.
    apply CongruenceSpi.StructuralCongruence_symmetric;auto.
    intros.
    ex_elim H1 p0.
    ex_elim H1 q0.
    elim (reduction_spi_congruence_output _ _ _ _ _ H2 _ H5).
    intro p'0;intros.
    ex_elim H7 e.
    elim (reduction_spi_conc_message _ _ _ H7 _ _ _ (refl_equal _)).
    intro m';intros.
    subst e.
    elim (H0 _ _ _ H1).
    intros.
    clear H9.
    elim H10;clear H10;intros.
    clear H9.
    generalize (H10 _ _ _ _ H7).
    clear H10.
    intros.
    elim (H8 (map ExpressionSpi.j tildec)).
    intro tildecexpr;intros;and_elim.
    elim (is_permutation_map _ _ _ H10 _ _ _ (refl_equal _)).
    intro tildec';intros;and_elim.
    subst tildecexpr.
    elim (H9 tildec') with b;auto.
    intro d;intros.
    ex_elim H14 n'.
    ex_elim H14 q'0.
    clear H8.
    clear H9.
    elim (reduction_spi_congruence_output _ _ _ _ _ H14 _ (CongruenceSpi.StructuralCongruence_symmetric _ _ H6)).
    intro q';intros.
    ex_elim H8 f.
    elim (reduction_spi_conc_message _ _ _ H8 _ _ _ (refl_equal _)).
    intro n;intros.
    subst f.
    ex_elim H15 tilded'.
    elim (H9 (map ExpressionSpi.j tilded')).
    intro tildedexpr;intros;and_elim.
    clear H9.
    elim (is_permutation_map _ _ _ H17 _ _ _ (refl_equal _)).
    intro tilded;intros;and_elim.
    subst tildedexpr.
    exists d.
    exists n.
    exists q'.
    split;auto.
    exists tilded.
    split.
    eapply HOp.n_fresh_right_permutation;eauto.
    ex_elim H16 h'.
    exists h'.
    split.
    assert (instantiate_message tildec m = instantiate_message tildec' m').
    apply EvalSpi.expr_of_mesg_injective.
    repeat (rewrite instantiate_message_expr).
    auto.
    assert (instantiate_message tilded n = instantiate_message tilded' n').
    apply EvalSpi.expr_of_mesg_injective.
    repeat (rewrite instantiate_message_expr).
    auto.
    rewrite H21.
    rewrite H22.
    auto.
    exists (instantiate_process tildec' p'0).
    exists (instantiate_process tilded' q'0).
    split;auto.
    split.
    unfold instantiate_process.
    auto.
    apply CongruenceSpi.StructuralCongruence_symmetric.
    auto.
    rewrite List_length_map.
    elim H15;auto.
    intros.
    elim H17;auto.
    eapply HOp.n_fresh_left_permutation;eauto.
    rewrite List_length_map.
    elim H3;auto.
    intros.
    elim H10;auto.
  Qed.
End EarlyHedgedBisimulation.


