(* A formalisation of the spi calculus in Coq *)
(* Copyright (C) 2005-2007 Sebastien Briais *)
(* http://lamp.epfl.ch/~sbriais/ *)

(* This library is free software; you can redistribute it and/or modify *)
(* it under the terms of the GNU Lesser General Public License as *)
(* published by the Free Software Foundation; either version 2.1 of the *)
(* License, or (at your option) any later version. *)

(* This library is distributed in the hope that it will be useful, but *)
(* WITHOUT ANY WARRANTY; without even the implied warranty of *)
(* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU *)
(* Lesser General Public License for more details. *)

(* You should have received a copy of the GNU Lesser General Public *)
(* License along with this library; if not, write to the Free Software *)
(* Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA *)
(* 02110-1301 USA *)

Require Import name.
Require Import expressionSpi.
Require Import messageSpi.
Require Import formulaeSpi.
Require Import evalSpi.
Require Import process.
Require Import actionType.
Require Import agent.
Require Import Arith.
Require Import deBruijnType.
Require Import tactics.
Require Import missing.

Require Import calculi.

Module ActionSpi <: ActionType with Module A := AgentSpi.  Module A := AgentSpi.

  Inductive kind : Set :=
    | In : kind
    | Out : kind
    | Silent : kind.

  Inductive action : Set :=
    | Tau : action
    | Inp : Name.t -> action
    | Outp : Name.t -> action.

  Definition t := action.

  Definition equal (x y:t) := x = y.

  Lemma equal_refl : forall (x:t),(equal x x).
    unfold equal;auto.
  Qed.

  Lemma equal_sym : forall (x y:t),(equal x y)->(equal y x).
    unfold equal;auto.
  Qed.

  Lemma equal_trans : forall (x y z:t),(equal x y)->(equal y z)->(equal x z).
    unfold equal;intros;congruence.
  Qed.

  Module Ord <: OrderedType with Definition t := t with Definition eq := @eq t.
    Definition t := t.
    Definition eq := @eq t.
    Definition lt (alpha beta:t) :=
      match (alpha,beta) with
	| (Tau,Tau) => False
	| (Tau,_) => True
	| (_,Tau) => False
	| (Inp a,Inp b) => Name.Ord.lt a b
	| (Inp _,_) => True
	| (_,Inp _) => False
	| (Outp a,Outp b) => Name.Ord.lt a b
      end.

    Lemma eq_refl : forall (x:t),(eq x x).
      unfold eq;reflexivity.
    Qed.

    Lemma eq_sym : forall (x y:t),(eq x y)->(eq y x).
      unfold eq;auto.
    Qed.

    Lemma eq_trans : forall (x y z:t),(eq x y)->(eq y z)->(eq x z).
      unfold eq;intros;congruence.
    Qed.

    Lemma lt_trans : forall (x y z:t),(lt x y)->(lt y z)->(lt x z).
      unfold lt.
      destruct x;destruct y;destruct z;try tauto;eauto.
    Qed.

    Lemma lt_not_eq : forall (x y:t),(lt x y)->~(eq x y).
      unfold lt, eq.
      destruct x;destruct y;try tauto;intros;red;intro;try (discriminate H0);injection H0;intros.
      elim (Name.Ord.lt_not_eq t0 t1);auto.
      elim (Name.Ord.lt_not_eq t0 t1);auto.
    Qed.

    Definition compare : forall (x y:t),(Compare lt eq x y).
      unfold lt, eq.
      destruct x;destruct y;try (apply EQ;trivial;fail);try (apply LT;trivial;fail);try (apply GT;trivial;fail).
      case (Name.Ord.compare t0 t1);intro;[apply LT | apply EQ | apply GT];auto.
      rewrite e;trivial.
      case (Name.Ord.compare t0 t1);intro;[apply LT | apply EQ | apply GT];auto.
      rewrite e;auto.
    Defined.

    Hint Immediate eq_sym.
    Hint Resolve eq_refl eq_trans lt_not_eq lt_trans.
  End Ord.

  Definition t_map (f:nat->Name.t->Name.t) (d:nat)(x:t) :=
    match x with
      | Tau => Tau
      | (Inp a) => (Inp (f d a))
      | (Outp a) => (Outp (f d a))
    end.

  Lemma t_map_struct : deBruijn.map_struct Name.t t t_map.
    split;unfold t_map;try (destruct x;repeat (rewrite H);trivial).
    unfold injective;destruct x;destruct y;intros;try (discriminate H0);auto;injection H0;intros.
    rewrite (H d t0 t1);auto.
    rewrite (H d t0 t1);auto.
  Qed.

  Definition deBruijn_base := deBruijn.deBruijn_map_struct Name.t Name.deBruijn_base t t_map t_map_struct.

  Definition t_fold (A:Set)(f:nat->Name.t->A->A)(d:nat)(x:t)(e:A) : A :=
    match x with
      | Tau => e
      | Inp a => (f d a e)
      | Outp a => (f d a e)
    end.
  
  Lemma t_fold_struct : deBruijn.fold_struct Name.t t t_map t_map_struct t_fold.
    split;intros.
    (* fold_orb_eq *)
    destruct x;simpl;repeat (rewrite H);intros;auto;simpl in H0;auto.
    (* fold_exists *)
    destruct x;simpl;simpl in H0;intros;auto.
    (* fold_map_simpl *)
    destruct x;simpl;intros;auto.
    (* fold_equal_cond_invariant *)
    destruct x;simpl;intros;auto.
    (* fold_invariant *)
    destruct x;simpl;intros;auto.
    (* fold_orb_eq_map_inversion *)
    destruct x;destruct y;try (discriminate H2);auto;simpl in H0;simpl in H1;simpl in H2;injection H2;clear H2;intros;rewrite (H d t0 t1);auto.
  Qed.
  
  Definition deBruijn_fn_base := deBruijn.deBruijn_fn_map_struct Name.t t t_map t_map_struct t_fold Name.deBruijn_base Name.deBruijn_fn_base t_fold_struct.

  Definition get_kind (a:t) := 
    match a with
      | Tau => Silent
      | Inp _ => In
      | _ => Out
    end.

  Lemma lift_kind : forall (d k:nat)(x:t),(get_kind (deBruijn.lift deBruijn_base d k x))=(get_kind x).
    destruct x;try reflexivity.
  Qed.

  Lemma swapR_kind : forall (d k:nat)(x:t),(get_kind (deBruijn.swapR deBruijn_base d k x))=(get_kind x).
    destruct x;try reflexivity.
  Qed.

  Definition mk_input (e:A.P.E.t) : option t :=
    match (EvalSpi.evalc e) with
      | (Some (MessageSpi.MNam a)) => Some (Inp a)
      | _ => None
    end.

  Definition mk_output (e:A.P.E.t) (f:A.P.E.t) : option (t * A.P.E.t) :=
    match (EvalSpi.evalc e,EvalSpi.evalc f) with
      | (Some (MessageSpi.MNam a),Some m) => Some ((Outp a),EvalSpi.expr_of_mesg m)
      | _ => None
    end.

  Definition mk_silent : t := Tau.

  Lemma mk_input_kind : forall (e:A.P.E.t)(a:t),(mk_input e = Some a)->(get_kind a = In).
    unfold mk_input.
    intros.
    destruct (EvalSpi.evalc e).
    destruct t0;try (discriminate H).
    injection H;intros;subst a;reflexivity.
    discriminate H.
  Qed.

  Lemma mk_output_kind : forall (e f:A.P.E.t)(a:t)(m:A.P.E.t),(mk_output e f = Some (a,m))->(get_kind a = Out).
    unfold mk_output.
    intros.
    destruct (EvalSpi.evalc e);try (discriminate H).
    destruct t0;try (discriminate H).
    destruct (EvalSpi.evalc f);try (discriminate H).
    injection H;intros;subst a.
    reflexivity.
  Qed.

  Lemma mk_silent_kind : get_kind mk_silent = Silent.
    reflexivity.
  Qed.

  Definition mk_if (phi:A.P.F.t) (alpha:t) : option t :=
    if EvalSpi.eval phi then Some alpha else None.

  Lemma mk_if_kind : forall (phi:A.P.F.t)(a b:t),(mk_if phi a = Some b)->(get_kind b = get_kind a).
    unfold mk_if;intros.
    destruct (EvalSpi.eval phi);try (discriminate H).
    injection H;intros;subst b;auto.
  Qed.

  Definition mk_res (a:t) : option t := if deBruijn.mem_fn deBruijn_fn_base 0 0 a then None else Some(deBruijn.low1 deBruijn_base 0 a).
	
  Lemma mk_res_kind : forall (a b:t),(mk_res a = Some b)->(get_kind b = get_kind a).
    destruct a;unfold mk_res;repeat (case deBruijn.mem_fn;intro);try (intros;injection H;intros;subst b;reflexivity);try (intros;discriminate H).
  Qed.

  Definition mk_interact (alpha:t) (beta:t) : option t :=
    match (alpha,beta) with
      | (Inp a,Outp b) =>
	(match Name.Ord.compare a b with
	   | EQ _ => Some Tau
	   | _ => None
	 end)
      | (Outp b,Inp a) =>
	(match Name.Ord.compare a b with
	   | EQ _ => Some Tau
	   | _ => None
	 end)
      | _ => None
    end.
  
  Lemma mk_interact_kind : forall (a b c:t),(mk_interact a b = Some c)->(get_kind c = Silent).
    destruct a;destruct b;unfold mk_interact;try (intros;discriminate H);repeat (case Name.Ord.compare;intro);try (intros;discriminate H);intros;injection H;intros;subst c;repeat (split;try reflexivity).
  Qed.
  
  Lemma mk_interact_exg : forall (a b c:t),(mk_interact a b = Some c)->exists c':t,equal c c' /\ (mk_interact b a = Some c').
    destruct a;destruct b;unfold mk_interact;intros;try (discriminate H).
    destruct (Name.Ord.compare t0 t1);try (discriminate H).
    exists Tau.
    split;auto.
    injection H;intros.
    subst c.
    apply equal_refl.
    destruct (Name.Ord.compare t1 t0);try (discriminate H).
    exists Tau.
    split;auto.
    injection H;intros.
    subst c.
    apply equal_refl.
  Qed.

  (* equal properties *)
  Lemma mk_if_equal : forall (phi:A.P.F.t)(a b:t),(mk_if phi a = Some b)->forall (a':t),(equal a a')->(exists b':t,equal b b' /\ (mk_if phi a' = Some b')).
    intros.
    rewrite H0 in H.
    exists b.
    split;auto.
    apply equal_refl.
  Qed.

  Lemma mk_res_equal : forall (a b:t),(mk_res a = Some b)->forall (a':t),(equal a a')->(exists b':t,equal b b' /\ (mk_res a' = Some b')).
    intros.
    rewrite H0 in H.
    exists b.
    split;auto.
    apply equal_refl.
  Qed.

  Lemma mk_interact_equal : forall (a b c:t),(mk_interact a b = Some c)->forall (a' b':t),(equal a a')->(equal b b')->(exists c':t,equal c c' /\ (mk_interact a' b' = Some c')).
    intros.
    rewrite H0 in H.
    rewrite H1 in H.
    exists c.
    split;auto.
    apply equal_refl.
  Qed.

  Lemma mk_res_lift_1 : forall (a:t),exists b:t,(equal b a)/\(mk_res (deBruijn.lift deBruijn_base 0 1 a) = Some b).
    unfold mk_res;intros.
    case (Bool_elim_bool (deBruijn.mem_fn  deBruijn_fn_base 0 0 (deBruijn.lift deBruijn_base 0 1 a)));intro H;rewrite H.
    rewrite deBruijn.lift_not_fn in H;try (discriminate H).
    omega.
    rewrite (deBruijn.low1_lift_simpl deBruijn_base 0 0).
    rewrite deBruijn.lift_0.
    exists a.
    split;auto.
    apply equal_refl.
  Qed.
  
  Lemma mk_res_swapR_aux : forall (a b:t),(mk_res a = Some b)->forall (c:t),(mk_res b = Some c)->(equal a (deBruijn.swapR deBruijn_base 0 1 a)).
    intros a b.
    unfold mk_res.
    case (Bool_elim_bool (deBruijn.mem_fn deBruijn_fn_base 0 0 a));intro H;rewrite H;intro H0;try (discriminate H0).
    intro c.
    case (Bool_elim_bool (deBruijn.mem_fn deBruijn_fn_base 0 0 b));intro H1;rewrite H1;intro H2;try (discriminate H2).
    injection H0;clear H0;intros;subst b.
    injection H2;clear H2;intros;subst c.
    rewrite (deBruijn.swapR_eq_cond deBruijn_fn_base);trivial.
    apply equal_refl.
    intros.
    inversion H0.
    generalize (deBruijn.lift_low1_fn deBruijn_fn_base 0 a H 0);simpl;intros.
    unfold DeBruijnNat.lift in H3.
    simpl in H3.
    rewrite <- H3.
    simpl in H1;auto.
    inversion H3.
    auto.
  Qed.

  Lemma mk_res_swapR : forall (a b:t),(mk_res a = Some b)->forall (c:t),(mk_res b = Some c)->(exists b':t,exists c':t,(equal c' c)/\(mk_res (deBruijn.swapR deBruijn_base 0 1 a) = Some b')/\(mk_res b' = Some c')).
    intros.
    generalize (mk_res_swapR_aux a b H c H0);intro.
    elim (mk_res_equal a b H (deBruijn.swapR deBruijn_base 0 1 a));auto.
    intro b';intros;and_elim.
    exists b'.
    elim (mk_res_equal b c H0 b');auto.
    intro c';intros;and_elim.
    exists c'.
    split.
    apply equal_sym;auto.
    split;auto.
  Qed.
  
  Lemma mk_interact_res_lift : forall (a a':t),(mk_res a = Some a')->forall (b c'':t),(mk_interact a' b = Some c'')->(exists c:t,exists c':t,(equal c' c'')/\(mk_interact a (deBruijn.lift deBruijn_base 0 1 b) = Some c)/\(mk_res c = Some c')).
    unfold mk_res.
    intros a a'.
    case (Bool_elim_bool (deBruijn.mem_fn deBruijn_fn_base 0 0 a));intro H;rewrite H.
    intro H0;discriminate H0.
    intro H0;injection H0;clear H0;intro H0;subst a'.
    intros b c'.
    destruct a;destruct b;simpl;try (intro H0;discriminate H0);simpl in H;unfold mk_interact;case Name.Ord.compare;intros;try (discriminate H0);injection H0;clear H0;intros;subst c'.
    rewrite <- e;generalize (deBruijn.lift_1_low1_simpl Name.deBruijn_fn_base 0 0 t0);simpl;intros;rewrite H0;auto;rewrite (deBruijn.swapR_0 Name.deBruijn_base).
    case Name.Ord.compare;intro;try (elim (Name.Ord.lt_not_eq t0 t0);auto;fail).
    exists Tau.
    exists Tau.
    split.
    apply equal_refl.
    split;auto.
    rewrite e;generalize (deBruijn.lift_1_low1_simpl Name.deBruijn_fn_base 0 0 t0);simpl;intros;rewrite H0;auto;rewrite (deBruijn.swapR_0 Name.deBruijn_base).
    case Name.Ord.compare;intro;try (elim (Name.Ord.lt_not_eq t0 t0);auto;fail).
    exists Tau.
    exists Tau.
    split.
    apply equal_refl.
    split;auto.
  Qed.

  Lemma mk_interact_res_lift_inv : forall (a b c:t),(mk_interact a (deBruijn.lift deBruijn_base 0 1 b) = Some c)->forall (c'':t),(mk_res c = Some c'')->exists a':t,(mk_res a = Some a')/\exists c':t,(mk_interact a' b = Some c').
    intros a b c.
    unfold mk_interact.
    destruct a;destruct b;try (intro H;discriminate H);simpl;case Name.Ord.compare;intro H;try (intro H0;discriminate H0);intro H0;injection H0;clear H0;intros;subst c;(rewrite H || rewrite <- H);unfold mk_res;simpl;rewrite (deBruijn.lift_not_fn Name.deBruijn_fn_base 0 1 0);try omega.
    exists (Inp (Name.low1 0 (Name.lift 0 1 t1)));split;trivial;rewrite (deBruijn.low1_lift_simpl Name.deBruijn_base);rewrite deBruijn.lift_0;exists Tau.
    case Name.Ord.compare;intro;try (elim (Name.Ord.lt_not_eq t1 t1);auto;fail).
    split;auto.
    injection H1.
    intro;subst c''.
    exists (Outp (Name.low1 0 (Name.lift 0 1 t1)));split;trivial;rewrite (deBruijn.low1_lift_simpl Name.deBruijn_base);rewrite deBruijn.lift_0;exists Tau.
    case Name.Ord.compare;intro;try (elim (Name.Ord.lt_not_eq t1 t1);auto;fail).
    auto.
  Qed.

  (* lift and actions *)
  Lemma lift_mk_silent : forall (d k:nat),(equal mk_silent (deBruijn.lift deBruijn_base d k mk_silent)).
    intros.
    unfold mk_silent, equal.
    reflexivity.
  Qed.

  Lemma lift_mk_input : forall (e:A.P.E.t)(inp:t),(mk_input e = Some inp)->forall (d k:nat),(exists inp':t,mk_input (deBruijn.lift A.P.E.deBruijn_base d k e) = Some inp' /\ (equal inp' (deBruijn.lift deBruijn_base d k inp))).
    intros e inp.
    unfold mk_input.
    intros.
    rewrite EvalSpi.evalc_lift.
    destruct (EvalSpi.evalc e);try (discriminate H).
    unfold option_map.
    destruct t0;simpl;try (discriminate H).
    exists (Inp (Name.lift d k t0)).
    injection H;intros;subst inp.
    split;auto.
    unfold equal;auto.
  Qed.

  Lemma lift_mk_output : forall (e f:A.P.E.t)(out:t)(m:A.P.E.t),(mk_output e f = Some (out,m))->forall (d k:nat),(exists out':t,mk_output (deBruijn.lift A.P.E.deBruijn_base d k e) (deBruijn.lift A.P.E.deBruijn_base d k f) = Some (out',deBruijn.lift A.P.E.deBruijn_base d k m) /\ (equal out' (deBruijn.lift deBruijn_base d k out))).
    intros e f out m.
    unfold mk_output.
    intros.
    rewrite EvalSpi.evalc_lift.
    rewrite EvalSpi.evalc_lift.
    destruct (EvalSpi.evalc e);simpl;try (discriminate H).
    destruct t0;try (discriminate H).
    simpl.
    destruct (EvalSpi.evalc f);simpl;try (discriminate H).
    injection H;intros.
    subst out.
    subst m.
    rewrite EvalSpi.map_expr_of_msg.
    exists (Outp (Name.lift d k t0)).
    split;auto.
    unfold equal.
    auto.
  Qed.

  Lemma lift_mk_if : forall (phi:A.P.F.t)(mu mu':t),(mk_if phi mu = Some mu')->forall (d k:nat),(exists mu'':t,(equal mu'' (deBruijn.lift deBruijn_base d k mu')) /\ mk_if (deBruijn.lift A.P.F.deBruijn_base d k phi) (deBruijn.lift deBruijn_base d k mu) = Some mu'').
    intros phi mu mu';unfold mk_if;intros.
    rewrite EvalSpi.eval_lift.
    destruct (EvalSpi.eval phi);try (discriminate H).
    injection H;intros;subst mu'.
    exists (deBruijn.lift deBruijn_base d k mu).
    unfold equal;auto.
  Qed.
    
  Lemma lift_mk_res : forall (mu mu':t),(mk_res mu = Some mu')->forall (d k:nat),(exists mu'':t,(equal mu'' (deBruijn.lift deBruijn_base d k mu'))/\(mk_res (deBruijn.lift deBruijn_base (S d) k mu))=(Some mu'')).
    unfold mk_res.
    intros a b.
    case (Bool_elim_bool (deBruijn.mem_fn deBruijn_fn_base 0 0 a));intro H;rewrite H;try (intros;discriminate H0).
    intros.
    injection H0;intros;subst b.
    generalize (deBruijn.lift_In_fn deBruijn_fn_base d 0 (le_O_n d) 0 k a);simpl;intros.
    rewrite <- H1.
    simpl in H;rewrite H.
    generalize (deBruijn.lift_low1_comm deBruijn_fn_base 0 d k a);simpl;intros.
    rewrite H2;auto.
    exists (deBruijn.lift deBruijn_base d k (deBruijn.low1 deBruijn_base 0 a)).
    unfold equal.
    split;reflexivity.
  Qed.

  Lemma lift_mk_interact : forall (a b c:t),(mk_interact a b = Some c)->forall (d k:nat),(exists c':t,(equal c' (deBruijn.lift deBruijn_base d k c))/\(mk_interact (deBruijn.lift deBruijn_base d k a) (deBruijn.lift deBruijn_base d k b) = Some c')).
    destruct a;destruct b;unfold mk_interact;try (intros;discriminate H).
    case (Name.Ord.compare t0 t1);try (intros;discriminate H);intros;injection H;intros;subst c;simpl;rewrite e;case Name.Ord.compare;intro;try (elim (Name.Ord.lt_not_eq (Name.lift d k t1) (Name.lift d k t1));auto;fail).
    exists Tau;unfold equal;auto.
    case (Name.Ord.compare t1 t0);try (intros;discriminate H);intros;injection H;intros;subst c;simpl;rewrite <- e;case Name.Ord.compare;intro;try (elim (Name.Ord.lt_not_eq (Name.lift d k t1) (Name.lift d k t1));auto;fail).
    exists Tau;unfold equal;auto.
  Qed.

  Lemma lift_mk_input_inv : forall (d k:nat)(e:A.P.E.t)(inp':t),(mk_input (deBruijn.lift A.P.E.deBruijn_base d k e) = Some inp')->(exists inp:t,mk_input e= Some inp).
    intros d k e inp'.
    unfold mk_input.
    rewrite EvalSpi.evalc_lift.
    intro.
    destruct (EvalSpi.evalc e);simpl;try (discriminate H).
    simpl in H.
    destruct t0;simpl in H;try (discriminate H).
    injection H;intros;subst inp'.
    exists (Inp t0).
    auto.
  Qed.
    
  Lemma lift_mk_output_inv : forall (d k:nat)(e f:A.P.E.t)(out':t)(m':A.P.E.t),(mk_output (deBruijn.lift A.P.E.deBruijn_base d k e) (deBruijn.lift A.P.E.deBruijn_base d k f) = Some (out',m'))->(exists out:t,exists m:A.P.E.t,mk_output e f = Some (out,m) /\ m' = deBruijn.lift A.P.E.deBruijn_base d k m).
    intros d k e f out' m'.
    unfold mk_output.
    repeat (rewrite EvalSpi.evalc_lift).
    intro.
    destruct (EvalSpi.evalc e);try (discriminate H).
    simpl in H.
    destruct t0;simpl in H;try (discriminate H).
    destruct (EvalSpi.evalc f);try (discriminate H).
    simpl in H.
    injection H;intros.
    subst out'.
    exists (Outp t0).
    exists (EvalSpi.expr_of_mesg t1).
    split;auto.
    subst m'.
    simpl.
    rewrite EvalSpi.map_expr_of_msg.
    auto.
  Qed.
    
  Lemma lift_mk_if_inv : forall (d k:nat)(phi:A.P.F.t)(mu mu'':t),(mk_if (deBruijn.lift A.P.F.deBruijn_base d k phi) (deBruijn.lift deBruijn_base d k mu) = Some mu'')->(exists mu':t,(mk_if phi mu = Some mu')).
    intros d k phi mu mu'';unfold mk_if.
    rewrite EvalSpi.eval_lift.
    intro.
    destruct (EvalSpi.eval phi);try (discriminate H).
    injection H;intros;subst mu''.
    exists mu.
    auto.
  Qed.

  Lemma lift_mk_res_inv : forall (d k:nat)(mu mu'':t),(mk_res (deBruijn.lift deBruijn_base (S d) k mu) = Some mu'')->(exists mu':t,mk_res mu = Some mu').
    intros d k a c.
    unfold mk_res.
    case (Bool_elim_bool (deBruijn.mem_fn deBruijn_fn_base 0 0 (deBruijn.lift deBruijn_base (S d) k a)));intro H;rewrite H.
    intro H0;discriminate H0.
    intro.
    generalize (deBruijn.lift_In_fn deBruijn_fn_base d 0 (le_O_n d) 0 k a);simpl;intros.
    rewrite H1.
    simpl in H;rewrite H.
    injection H0;intros;subst c.
    exists (deBruijn.low1 deBruijn_base 0 a).
    reflexivity.
  Qed.

  Lemma lift_mk_interact_inv : forall (d k:nat)(a b c':t),(mk_interact (deBruijn.lift deBruijn_base d k a) (deBruijn.lift deBruijn_base d k b) = Some c')->(exists c:t,mk_interact a b = Some c).
    intros d k.
    destruct a;destruct b;unfold mk_interact;simpl;try (intros;discriminate H);intros c'.
    case (Name.Ord.compare (Name.lift d k t0) (Name.lift d k t1));intros;try (discriminate H).
    rewrite (deBruijn.lift_injective Name.deBruijn_base d k t0 t1 e);case Name.Ord.compare;intro;try (elim (Name.Ord.lt_not_eq t1 t1);auto;fail);exists Tau;trivial.
    case (Name.Ord.compare (Name.lift d k t1) (Name.lift d k t0));intros;try (discriminate H).
    rewrite (deBruijn.lift_injective Name.deBruijn_base d k t1 t0 e);case Name.Ord.compare;intro;try (elim (Name.Ord.lt_not_eq t0 t0);auto;fail);exists Tau;trivial.
  Qed.

  (* swapR and actions *)
  Lemma swapR_mk_silent : forall (d k:nat),(equal mk_silent (deBruijn.swapR deBruijn_base d k mk_silent)).
    intros.
    unfold mk_silent, equal.
    reflexivity.
  Qed.

  Lemma swapR_mk_input : forall (e:A.P.E.t)(inp:t),(mk_input e = Some inp)->forall (d k:nat),(exists inp':t,mk_input (deBruijn.swapR A.P.E.deBruijn_base d k e) = Some inp' /\ (equal inp' (deBruijn.swapR deBruijn_base d k inp))).
    intros e inp.
    unfold mk_input.
    intros.
    rewrite EvalSpi.evalc_swapR.
    destruct (EvalSpi.evalc e);simpl;try (discriminate H).
    destruct t0;simpl;try (discriminate H).
    exists (Inp (Name.swapR d k t0)).
    injection H;intros;subst inp.
    split;auto.
    unfold equal;auto.
  Qed.

  Lemma swapR_mk_output : forall (e f:A.P.E.t)(out:t)(m:A.P.E.t),(mk_output e f = Some (out,m))->forall (d k:nat),(exists out':t,mk_output (deBruijn.swapR A.P.E.deBruijn_base d k e) (deBruijn.swapR A.P.E.deBruijn_base d k f) = Some (out',deBruijn.swapR A.P.E.deBruijn_base d k m) /\ (equal out' (deBruijn.swapR deBruijn_base d k out))).
    intros e f out m.
    unfold mk_output.
    intros.
    repeat (rewrite EvalSpi.evalc_swapR).
    destruct (EvalSpi.evalc e);simpl;try (discriminate H).
    destruct t0;try (discriminate H).
    simpl.
    destruct (EvalSpi.evalc f);simpl;try (discriminate H).
    injection H;intros.
    subst out.
    subst m.
    rewrite EvalSpi.map_expr_of_msg.
    exists (Outp (Name.swapR d k t0)).
    split;auto.
    unfold equal.
    auto.
  Qed.

  Lemma swapR_mk_if : forall (phi:A.P.F.t)(mu mu':t),(mk_if phi mu = Some mu')->forall (d k:nat),(exists mu'':t,(equal mu'' (deBruijn.swapR deBruijn_base d k mu')) /\ mk_if (deBruijn.swapR A.P.F.deBruijn_base d k phi) (deBruijn.swapR deBruijn_base d k mu) = Some mu'').
    intros phi mu mu';unfold mk_if;intros.
    rewrite EvalSpi.eval_swapR.
    destruct (EvalSpi.eval phi);try (discriminate H).
    injection H;intros;subst mu'.
    exists (deBruijn.swapR deBruijn_base d k mu).
    unfold equal;auto.
  Qed.
    
  Lemma swapR_mk_res : forall (mu mu':t),(mk_res mu = Some mu')->forall (d k:nat),(exists mu'':t,(equal mu'' (deBruijn.swapR deBruijn_base d k mu'))/\(mk_res (deBruijn.swapR deBruijn_base (S d) k mu))=(Some mu'')).
    unfold mk_res.
    intros a b.
    case (Bool_elim_bool (deBruijn.mem_fn deBruijn_fn_base 0 0 a));intro H;rewrite H;try (intros;discriminate H0).
    intros.
    injection H0;intros;subst b.
    generalize (deBruijn.swapR_In_fn deBruijn_fn_base d 0 (le_O_n d) 0 k a);simpl;intros.
    rewrite <- H1.
    simpl in H;rewrite H.
    generalize (deBruijn.swapR_low1_comm deBruijn_fn_base 0 d k a);simpl;intros.
    rewrite H2;auto.
    exists (deBruijn.swapR deBruijn_base d k (deBruijn.low1 deBruijn_base 0 a)).
    unfold equal.
    split;reflexivity.
  Qed.

  Lemma swapR_mk_interact : forall (a b c:t),(mk_interact a b = Some c)->forall (d k:nat),(exists c':t,(equal c' (deBruijn.swapR deBruijn_base d k c))/\(mk_interact (deBruijn.swapR deBruijn_base d k a) (deBruijn.swapR deBruijn_base d k b) = Some c')).
    destruct a;destruct b;unfold mk_interact;try (intros;discriminate H).
    case (Name.Ord.compare t0 t1);try (intros;discriminate H);intros;injection H;intros;subst c;simpl;rewrite e;case Name.Ord.compare;intro;try (elim (Name.Ord.lt_not_eq (Name.swapR d k t1) (Name.swapR d k t1));auto;fail).
    exists Tau;unfold equal;auto.
    case (Name.Ord.compare t1 t0);try (intros;discriminate H);intros;injection H;intros;subst c;simpl;rewrite <- e;case Name.Ord.compare;intro;try (elim (Name.Ord.lt_not_eq (Name.swapR d k t1) (Name.swapR  d k t1));auto;fail).
    exists Tau;unfold equal;auto.
  Qed.
End ActionSpi.


