(* A formalisation of the spi calculus in Coq *)
(* Copyright (C) 2005-2007 Sebastien Briais *)
(* http://lamp.epfl.ch/~sbriais/ *)

(* This library is free software; you can redistribute it and/or modify *)
(* it under the terms of the GNU Lesser General Public License as *)
(* published by the Free Software Foundation; either version 2.1 of the *)
(* License, or (at your option) any later version. *)

(* This library is distributed in the hope that it will be useful, but *)
(* WITHOUT ANY WARRANTY; without even the implied warranty of *)
(* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU *)
(* Lesser General Public License for more details. *)

(* You should have received a copy of the GNU Lesser General Public *)
(* License along with this library; if not, write to the Free Software *)
(* Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA *)
(* 02110-1301 USA *)

Require Import name.
Require Import expressionSpi.
Require Import messageSpi.
Require Import formulaeSpi.
Require Import evalSpi.
Require Import process.
Require Import actionType.
Require Import Arith.
Require Import deBruijnType.
Require Import tactics.
Require Import missing.
Require Import ordmissing.

Require Import calculi.

Module ActionSpiSymb <: ActionTypeLight.
  Module A := AgentSpi.

  Definition constraint := (nat * list FormulaeSpi.t)%type.

  Module OrdListForm := OrderedList FormulaeSpi.Ord.
  Module OrdListFormNew := OrderedPair OrderedNat OrdListForm.

  Lemma OrdListFormNewEq_equal : forall (x y:constraint),(OrdListFormNew.eq x y)<->x=y.
    split.
    destruct x;destruct y.
    simpl;intros;and_elim.
    rewrite H.
    clear H.
    generalize H0;clear H0.
    generalize l0;clear l0.
    induction l.
    simpl.
    destruct l0;try tauto.
    simpl.
    destruct l0;try tauto.
    intros;and_elim.
    rewrite H0.
    generalize (IHl l0 H).
    intros.
    injection H1;intros.
    subst l0;auto.
    intro.
    subst y;auto.
  Qed.

  Definition t_map_constraint (f:nat->list FormulaeSpi.t->list FormulaeSpi.t) (d:nat) (x:constraint) := 
    let (n,lphi) := x in (n,(f (d+n) lphi)).

  Lemma t_map_struct_constraint : deBruijn.map_struct (list FormulaeSpi.t) constraint t_map_constraint.
    split;intros.
    destruct x;reflexivity.
    unfold injective.
    intros.
    destruct x;destruct y;simpl in H0.
    injection H0;intros.
    subst n0.
    rewrite (H (d+n) l l0);auto.
    destruct x;simpl.
    repeat (rewrite <- plus_assoc).
    rewrite H;auto.
    destruct x;simpl.
    repeat (rewrite <- plus_assoc).
    rewrite H;auto.
  Qed.

  Definition t_fold_constraint (A:Set)(f:nat->list FormulaeSpi.t->A->A)(d:nat)(x:constraint)(e:A) : A :=
    let (n,lphi) := x in (f (d+n) lphi e).

  Lemma t_fold_struct_constraint : deBruijn.fold_struct (list FormulaeSpi.t) constraint t_map_constraint t_map_struct_constraint t_fold_constraint.
    split;intros.
    destruct x;simpl.
    simpl in H0.
    repeat (rewrite <- plus_assoc).
    rewrite H;auto.
    rewrite plus_assoc;auto.
    destruct x;simpl.
    simpl in H0.
    rewrite <- plus_assoc.
    apply H.
    rewrite plus_assoc;auto.
    destruct x;simpl.
    repeat (rewrite plus_assoc).
    auto.
    destruct x;simpl.
    simpl in H1.
    repeat (rewrite <- plus_assoc).
    apply H;auto.
    rewrite plus_assoc;auto.
    destruct x;simpl.
    apply H.
    auto.
    (* fold_orb_eq_map_inversion *)
    destruct x;destruct y;simpl in H2;injection H2;clear H2;intros;subst n0.
    rewrite (H (d+n) l l0);repeat (rewrite plus_assoc);auto.
  Qed.

  Definition deBruijn_base_list := deBruijn.deBruijn_base_list FormulaeSpi.t FormulaeSpi.deBruijn_base.
  Definition deBruijn_base_constraint := deBruijn.deBruijn_map_struct (list FormulaeSpi.t) deBruijn_base_list constraint t_map_constraint t_map_struct_constraint.
  Definition deBruijn_subst_list := deBruijn.deBruijn_subst_list FormulaeSpi.t FormulaeSpi.deBruijn_base ExpressionSpi.t ExpressionSpi.deBruijn_base FormulaeSpi.deBruijn_subst.

  Definition deBruijn_fn_base_list :=  deBruijn.deBruijn_fn_base_list FormulaeSpi.t FormulaeSpi.deBruijn_base FormulaeSpi.deBruijn_fn_base.
  Definition deBruijn_fn_base_constraint :=  deBruijn.deBruijn_fn_map_struct (list FormulaeSpi.t) constraint t_map_constraint t_map_struct_constraint t_fold_constraint deBruijn_base_list deBruijn_fn_base_list t_fold_struct_constraint.
  Definition deBruijn_fn_subst_list := deBruijn.deBruijn_fn_subst_list FormulaeSpi.t FormulaeSpi.deBruijn_base FormulaeSpi.deBruijn_fn_base ExpressionSpi.t ExpressionSpi.deBruijn_base ExpressionSpi.deBruijn_fn_base FormulaeSpi.deBruijn_subst FormulaeSpi.deBruijn_fn_subst.

  Definition union_c (x y:constraint) :=
    match (x,y) with
      | ((n,lphi),(m,lpsi)) => (n+m,app (deBruijn.lift deBruijn_base_list 0 m lphi) (deBruijn.lift deBruijn_base_list m n lpsi))
    end.

  Definition new_c (x:constraint) := 
    match x with
      | (n,lphi) => 
	if deBruijn.mem_fn deBruijn_fn_base_list 0 n lphi 
	  then (S n,lphi) 
	  else (n,deBruijn.low1 deBruijn_base_list n lphi)
    end.

  Definition equal_c (x y:constraint) := x = y.

  Lemma equal_c_refl : forall (x:constraint),(equal_c x x).
    unfold equal_c;auto.
  Qed.

  Lemma equal_c_sym : forall (x y:constraint),(equal_c x y)->(equal_c y x).
    unfold equal_c;auto.
  Qed.

  Lemma equal_c_trans : forall (x y z:constraint),(equal_c x y)->(equal_c y z)->(equal_c x z).
    unfold equal_c;congruence.
  Qed.

  Lemma lift_constraint : forall (d k:nat)(n:nat)(lphi:list FormulaeSpi.t),(deBruijn.lift deBruijn_base_constraint d k (n,lphi))=(n,deBruijn.lift deBruijn_base_list (d+n) k lphi).
    intros;reflexivity.
  Qed.

  Lemma lift_app : forall (d k:nat)(lphi lpsi:list FormulaeSpi.t),(deBruijn.lift deBruijn_base_list d k (app lphi lpsi))=(app (deBruijn.lift deBruijn_base_list d k lphi) (deBruijn.lift deBruijn_base_list d k lpsi)).
    intros.
    simpl.
    repeat (rewrite deBruijn.map_list_app).    
    auto.
  Qed.
      
  Lemma lift_union_c : forall (d k:nat)(x y:constraint),(deBruijn.lift deBruijn_base_constraint d k (union_c x y))=(union_c (deBruijn.lift deBruijn_base_constraint d k x) (deBruijn.lift deBruijn_base_constraint d k y)).
    destruct x as [n lphi].
    destruct y as [m lpsi].
    unfold union_c.
    repeat (rewrite lift_constraint).
    rewrite lift_app.
    generalize (deBruijn.lift_comm deBruijn_base_list 0 (d+n) m k);intros.
    rewrite plus_0_l in H.
    rewrite plus_assoc.
    rewrite H.
    generalize (deBruijn.lift_comm deBruijn_base_list m d n k);intros.
    replace (d+n+m) with (m+d+n);try omega.
    rewrite H0.
    rewrite (plus_comm m d);auto.
  Qed.

  Lemma swapR_constraint : forall (d k:nat)(n:nat)(lphi:list FormulaeSpi.t),(deBruijn.swapR deBruijn_base_constraint d k (n,lphi))=(n,deBruijn.swapR deBruijn_base_list (d+n) k lphi).
    intros;reflexivity.
  Qed.

  Lemma swapR_app : forall (d k:nat)(lphi lpsi:list FormulaeSpi.t),(deBruijn.swapR deBruijn_base_list d k (app lphi lpsi))=(app (deBruijn.swapR deBruijn_base_list d k lphi) (deBruijn.swapR deBruijn_base_list d k lpsi)).
    intros.
    simpl.
    repeat (rewrite deBruijn.map_list_app).    
    auto.
  Qed.

  Lemma swapR_union_c : forall (d k:nat)(x y:constraint),(deBruijn.swapR deBruijn_base_constraint d k (union_c x y))=(union_c (deBruijn.swapR deBruijn_base_constraint d k x) (deBruijn.swapR deBruijn_base_constraint d k y)).
    destruct x as [n lphi].
    destruct y as [m lpsi].
    unfold union_c.
    repeat (rewrite swapR_constraint).
    rewrite swapR_app.
    generalize (deBruijn.swapR_lift_comm deBruijn_base_list 0 (d+n) m k);intros.
    repeat (rewrite plus_0_l in H).
    replace (d+(n+m)) with (m+(d+n));try omega.
    rewrite H.
    generalize (deBruijn.swapR_lift_comm deBruijn_base_list m d n k);intros.
    replace (m+(d+n)) with (m+n+d);try omega.
    rewrite H0.
    rewrite (plus_comm m d);auto.
  Qed.

  Lemma low1_constraint : forall (d:nat)(n:nat)(lphi:list FormulaeSpi.t),(deBruijn.low1 deBruijn_base_constraint d (n,lphi))=(n,deBruijn.low1 deBruijn_base_list (d+n) lphi).
    intros;reflexivity.
  Qed.

  Lemma low1_app : forall (d:nat)(lphi lpsi:list FormulaeSpi.t),(deBruijn.low1 deBruijn_base_list d (app lphi lpsi))=(app (deBruijn.low1 deBruijn_base_list d lphi) (deBruijn.low1 deBruijn_base_list d lpsi)).
    intros.
    simpl.
    repeat (rewrite deBruijn.map_list_app).    
    auto.
  Qed.

  Lemma mem_fn_app : forall (n d:nat)(lphi lpsi:list FormulaeSpi.t),(deBruijn.mem_fn deBruijn_fn_base_list n d (app lphi lpsi))=(orb (deBruijn.mem_fn deBruijn_fn_base_list n d lphi) (deBruijn.mem_fn deBruijn_fn_base_list n d lpsi)).
    intros.
    unfold deBruijn_fn_base_list.
    rewrite deBruijn.mem_fn_app.
    auto.
  Qed.

  Lemma mem_fn_constraint : forall  (n d:nat)(nphi:nat)(lphi:list FormulaeSpi.t),(deBruijn.mem_fn deBruijn_fn_base_constraint n d (nphi,lphi))=(deBruijn.mem_fn deBruijn_fn_base_list n (d+nphi) lphi).
    intros.
    reflexivity.
  Qed.

  Lemma mem_fn_union_c : forall (n d:nat)(x y:constraint),(deBruijn.mem_fn deBruijn_fn_base_constraint n d (union_c x y))=(orb (deBruijn.mem_fn deBruijn_fn_base_constraint n d x) (deBruijn.mem_fn deBruijn_fn_base_constraint n d y)).
    unfold union_c.
    destruct x as [nphi lphi].
    destruct y as [npsi lpsi].
    repeat (rewrite mem_fn_constraint).
    rewrite mem_fn_app.
    generalize (deBruijn.fn_plus_lift deBruijn_fn_base_list 0 (d+(nphi+npsi)));rewrite plus_0_l;intros.
    rewrite <- H.
    replace (DeBruijnNat.lift 0 (d+(nphi+npsi)) n) with (n+(d+(nphi+npsi)));try reflexivity.
    generalize (deBruijn.lift_fn deBruijn_fn_base_list 0 npsi lphi (n+d+nphi)).
    replace (DeBruijnNat.lift 0 npsi (n+d+nphi)) with (n+d+nphi+npsi);try reflexivity.
    repeat (rewrite plus_assoc).
    fold deBruijn_base_list.
    intro.
    rewrite H0.
    generalize (deBruijn.fn_plus_lift deBruijn_fn_base_list npsi (d+nphi));intros.
    replace (npsi+(d+nphi)) with (d+nphi+npsi) in H1;try omega.
    rewrite <- H1.
    replace (DeBruijnNat.lift 0 (d+nphi) n) with (n+(d+nphi));try reflexivity.
    generalize (deBruijn.lift_fn deBruijn_fn_base_list npsi nphi lpsi (n+d)).
    replace (DeBruijnNat.lift 0 nphi (n+d)) with (n+d+nphi);try reflexivity.
    fold deBruijn_base_list.
    intro.
    rewrite plus_assoc.
    rewrite H2.
    generalize (deBruijn.fn_plus_lift deBruijn_fn_base_list 0 (d+nphi));rewrite plus_0_l;intros.
    rewrite <- H3.
    replace (DeBruijnNat.lift 0 (d+nphi) n) with (n+(d+nphi));try reflexivity.
    generalize (deBruijn.fn_plus_lift deBruijn_fn_base_list npsi d);intros.
    rewrite plus_comm in H4.
    rewrite <- H4.
    replace (DeBruijnNat.lift 0 d n) with (n+d);try reflexivity.
    rewrite plus_assoc;auto.
  Qed.

  Lemma lift_new_c : forall (d k:nat)(c:constraint),(deBruijn.lift deBruijn_base_constraint d k (new_c c))=(new_c (deBruijn.lift deBruijn_base_constraint (S d) k c)).
    intros.
    unfold new_c.
    destruct c.
    repeat (rewrite lift_constraint).
    case (Bool_elim_bool (deBruijn.mem_fn deBruijn_fn_base_list 0 n l));intro H;rewrite H.
    rewrite lift_constraint.
    generalize H;clear H.
    generalize (deBruijn.lift_In_fn deBruijn_fn_base_list d 0 (le_O_n d) n k l).
    fold deBruijn_base_list.
    intros.
    rewrite H0 in H.
    replace (S d + n) with (n + S d);try omega.
    rewrite <- H.
    replace (n+S d) with (d+S n);try omega.
    auto.
    rewrite lift_constraint.
    generalize H;clear H.
    generalize (deBruijn.lift_In_fn deBruijn_fn_base_list d 0 (le_O_n d) n k l).
    fold deBruijn_base_list.
    intros.
    rewrite H0 in H.
    replace (S d + n) with (n + S d);try omega.
    rewrite <- H.
    replace (n+S d) with (d+S n);try omega.
    generalize (deBruijn.lift_low1_comm deBruijn_fn_base_list n d k l H0).
    fold deBruijn_base_list.
    intros.
    replace (d+S n) with (n+S d);try omega.
    rewrite H1.
    rewrite plus_comm;auto.
  Qed.

  Lemma swapR_new_c : forall (d k:nat)(c:constraint),(deBruijn.swapR deBruijn_base_constraint d k (new_c c))=(new_c (deBruijn.swapR deBruijn_base_constraint (S d) k c)).
    intros.
    unfold new_c.
    destruct c.
    repeat (rewrite swapR_constraint).
    case (Bool_elim_bool (deBruijn.mem_fn deBruijn_fn_base_list 0 n l));intro H;rewrite H.
    rewrite swapR_constraint.
    generalize H;clear H.
    generalize (deBruijn.swapR_In_fn deBruijn_fn_base_list d 0 (le_O_n d) n k l).
    fold deBruijn_base_list.
    intros.
    rewrite H0 in H.
    replace (S d + n) with (n + S d);try omega.
    rewrite <- H.
    replace (n+S d) with (d+S n);try omega.
    auto.
    rewrite swapR_constraint.
    generalize H;clear H.
    generalize (deBruijn.swapR_In_fn deBruijn_fn_base_list d 0 (le_O_n d) n k l).
    fold deBruijn_base_list.
    intros.
    rewrite H0 in H.
    replace (S d + n) with (n + S d);try omega.
    rewrite <- H.
    replace (n+S d) with (d+S n);try omega.
    generalize (deBruijn.swapR_low1_comm deBruijn_fn_base_list n d k l H0).
    fold deBruijn_base_list.
    intros.
    replace (d+S n) with (n+S d);try omega.
    rewrite H1.
    rewrite plus_comm;auto.
  Qed.

  Inductive kind : Set :=
    | In : kind
    | Out : kind
    | Silent : kind.

  Inductive action : Set :=
    | Tau : action
    | Inp : ExpressionSpi.t -> action
    | Outp : ExpressionSpi.t -> action.

  Definition t := (action * constraint)%type.

  Definition equal (x y:t) := 
    match (x,y) with
      | ((alpha,phi),(beta,psi)) => alpha = beta /\ equal_c phi psi
    end.

  Lemma equal_refl : forall (x:t),(equal x x).
    unfold equal.
    destruct x.
    split;auto.
    apply equal_c_refl.
  Qed.

  Lemma equal_sym : forall (x y:t),(equal x y)->(equal y x).
    unfold equal.
    destruct x;destruct y.
    intros;and_elim.
    split.
    congruence.
    apply equal_c_sym.
    auto.
  Qed.

  Lemma equal_trans : forall (x y z:t),(equal x y)->(equal y z)->(equal x z).
    unfold equal;intros.
    destruct x;destruct y;destruct z;and_elim.
    split.
    congruence.
    eapply equal_c_trans;eauto.
  Qed.

  Module OrdAct <: OrderedType with Definition t := action with Definition eq := @eq action.
    Definition t := action.
    Definition eq := @eq action.
    Definition lt (alpha beta:t) :=
      match (alpha,beta) with
	| (Tau,Tau) => False
	| (Tau,_) => True
	| (_,Tau) => False
	| (Inp a,Inp b) => ExpressionSpi.Ord.lt a b
	| (Inp _,_) => True
	| (_,Inp _) => False
	| (Outp a,Outp b) => ExpressionSpi.Ord.lt a b
      end.

    Lemma eq_refl : forall (x:t),(eq x x).
      unfold eq;reflexivity.
    Qed.

    Lemma eq_sym : forall (x y:t),(eq x y)->(eq y x).
      unfold eq;auto.
    Qed.

    Lemma eq_trans : forall (x y z:t),(eq x y)->(eq y z)->(eq x z).
      unfold eq;intros;congruence.
    Qed.

    Lemma lt_trans : forall (x y z:t),(lt x y)->(lt y z)->(lt x z).
      unfold lt.
      destruct x;destruct y;destruct z;try tauto;eauto.
    Qed.

    Lemma lt_not_eq : forall (x y:t),(lt x y)->~(eq x y).
      unfold lt, eq.
      destruct x;destruct y;try tauto;intros;red;intro;try (discriminate H0);injection H0;intros.
      elim (ExpressionSpi.Ord.lt_not_eq t0 t1);auto.
      elim (ExpressionSpi.Ord.lt_not_eq t0 t1);auto.
    Qed.

    Definition compare : forall (x y:t),(Compare lt eq x y).
      unfold lt, eq.
      destruct x;destruct y;try (apply EQ;trivial;fail);try (apply LT;trivial;fail);try (apply GT;trivial;fail).
      case (ExpressionSpi.Ord.compare t0 t1);intro;[apply LT | apply EQ | apply GT];auto.
      rewrite e;trivial.
      case (ExpressionSpi.Ord.compare t0 t1);intro;[apply LT | apply EQ | apply GT];auto.
      rewrite e;auto.
    Defined.

    Hint Immediate eq_sym.
    Hint Resolve eq_refl eq_trans lt_not_eq lt_trans.
  End OrdAct.

  Module OrdActSymb := OrderedPair OrdAct OrdListFormNew.

  Module Ord <: OrderedType with Definition t := t with Definition eq := @eq t.
    Definition t := t.
    Definition eq := @eq t.
    Definition lt (a b:t) := OrdActSymb.lt a b.

    Lemma OrdActSymb_eq : forall (a b:t),(eq a b)<->(OrdActSymb.eq a b).
      split;intros.
      rewrite H.
      auto.
      destruct a;destruct b;simpl in H.
      and_elim.
      rewrite H.
      elim (OrdListFormNewEq_equal c c0);intros.
      rewrite H1;auto.
      unfold eq.
      auto.
    Qed.

    Lemma eq_refl : forall (x:t),(eq x x).
      unfold eq;reflexivity.
    Qed.

    Lemma eq_sym : forall (x y:t),(eq x y)->(eq y x).
      unfold eq;auto.
    Qed.

    Lemma eq_trans : forall (x y z:t),(eq x y)->(eq y z)->(eq x z).
      unfold eq;intros;congruence.
    Qed.

    Lemma lt_trans : forall (x y z:t),(lt x y)->(lt y z)->(lt x z).
      apply OrdActSymb.lt_trans.
    Qed.

    Lemma lt_not_eq : forall (x y:t),(lt x y)->~(eq x y).
      intros.
      red;intro.
      elim (OrdActSymb.lt_not_eq x y);auto.
      elim (OrdActSymb_eq x y);auto.
    Qed.

    Definition compare : forall (x y:t),(Compare lt eq x y).
      intros.
      case (OrdActSymb.compare x y);intro.
      apply LT;auto.
      apply EQ.
      elim (OrdActSymb_eq x y);auto.
      apply GT;auto.
    Defined.

    Hint Immediate eq_sym.
    Hint Resolve eq_refl eq_trans lt_not_eq lt_trans.
  End Ord.

  Definition t_map_action (f:nat->ExpressionSpi.t->ExpressionSpi.t) (d:nat)(x:action) :=
    match x with
      | Tau => Tau
      | (Inp a) => (Inp (f d a))
      | (Outp a) => (Outp (f d a))
    end.

  Lemma t_map_struct_action : deBruijn.map_struct ExpressionSpi.t action t_map_action.
    split;unfold t_map_action;try (destruct x;repeat (rewrite H);trivial).
    unfold injective;destruct x;destruct y;intros;try (discriminate H0);auto;injection H0;intros.
    rewrite (H d t0 t1);auto.
    rewrite (H d t0 t1);auto.
  Qed.

  Definition deBruijn_base_action := deBruijn.deBruijn_map_struct ExpressionSpi.t ExpressionSpi.deBruijn_base action t_map_action t_map_struct_action.

  Lemma lift_Inp : forall (d k:nat)(e:ExpressionSpi.t),(deBruijn.lift deBruijn_base_action d k (Inp e))=(Inp (deBruijn.lift ExpressionSpi.deBruijn_base d k e)).
    intros;reflexivity.
  Qed.

  Lemma lift_Outp : forall (d k:nat)(e:ExpressionSpi.t),(deBruijn.lift deBruijn_base_action d k (Outp e))=(Outp (deBruijn.lift ExpressionSpi.deBruijn_base d k e)).
    intros;reflexivity.
  Qed.

  Lemma lift_Tau : forall (d k:nat),(deBruijn.lift deBruijn_base_action d k Tau)=Tau.
    intros;reflexivity.
  Qed.

  Lemma swapR_Inp : forall (d k:nat)(e:ExpressionSpi.t),(deBruijn.swapR deBruijn_base_action d k (Inp e))=(Inp (deBruijn.swapR ExpressionSpi.deBruijn_base d k e)).
    intros;reflexivity.
  Qed.

  Lemma swapR_Outp : forall (d k:nat)(e:ExpressionSpi.t),(deBruijn.swapR deBruijn_base_action d k (Outp e))=(Outp (deBruijn.swapR ExpressionSpi.deBruijn_base d k e)).
    intros;reflexivity.
  Qed.

  Lemma swapR_Tau : forall (d k:nat),(deBruijn.swapR deBruijn_base_action d k Tau)=Tau.
    intros;reflexivity.
  Qed.

  Lemma low1_Inp : forall (d:nat)(e:ExpressionSpi.t),(deBruijn.low1 deBruijn_base_action d (Inp e))=(Inp (deBruijn.low1 ExpressionSpi.deBruijn_base d e)).
    intros;reflexivity.
  Qed.

  Lemma low1_Outp : forall (d:nat)(e:ExpressionSpi.t),(deBruijn.low1 deBruijn_base_action d (Outp e))=(Outp (deBruijn.low1 ExpressionSpi.deBruijn_base d e)).
    intros;reflexivity.
  Qed.

  Lemma low1_Tau : forall (d:nat),(deBruijn.low1 deBruijn_base_action d Tau)=Tau.
    intros;reflexivity.
  Qed.

  Definition t_fold_action (A:Set)(f:nat->ExpressionSpi.t->A->A)(d:nat)(x:action)(e:A) : A :=
    match x with
      | Tau => e
      | Inp a => (f d a e)
      | Outp a => (f d a e)
    end.
  
  Lemma t_fold_struct_action : deBruijn.fold_struct ExpressionSpi.t action t_map_action t_map_struct_action t_fold_action.
    split;intros.
    (* fold_orb_eq *)
    destruct x;simpl;repeat (rewrite H);intros;auto;simpl in H0;auto.
    (* fold_exists *)
    destruct x;simpl;simpl in H0;intros;auto.
    (* fold_map_simpl *)
    destruct x;simpl;intros;auto.
    (* fold_equal_cond_invariant *)
    destruct x;simpl;intros;auto.
    (* fold_invariant *)
    destruct x;simpl;intros;auto.
    (* fold_orb_eq_map_inversion *)
    destruct x;destruct y;try (discriminate H2);auto;simpl in H0;simpl in H1;simpl in H2;injection H2;clear H2;intros;rewrite (H d t0 t1);auto.
  Qed.

  Definition deBruijn_fn_base_action := deBruijn.deBruijn_fn_map_struct ExpressionSpi.t action t_map_action t_map_struct_action t_fold_action ExpressionSpi.deBruijn_base ExpressionSpi.deBruijn_fn_base t_fold_struct_action.

  Lemma mem_fn_Inp : forall (d k:nat)(e:ExpressionSpi.t),(deBruijn.mem_fn deBruijn_fn_base_action k d (Inp e))=(deBruijn.mem_fn ExpressionSpi.deBruijn_fn_base k d e).
    intros;reflexivity.
  Qed.

  Lemma mem_fn_Outp : forall (d k:nat)(e:ExpressionSpi.t),(deBruijn.mem_fn deBruijn_fn_base_action k d (Outp e))=(deBruijn.mem_fn ExpressionSpi.deBruijn_fn_base k d e).
    intros;reflexivity.
  Qed.

  Lemma mem_fn_Tau : forall (d k:nat)(e:ExpressionSpi.t),(deBruijn.mem_fn deBruijn_fn_base_action k d Tau)=false.
    intros;reflexivity.
  Qed.

  Definition deBruijn_base := deBruijn.deBruijn_base_pair action deBruijn_base_action constraint deBruijn_base_constraint.
  Definition deBruijn_fn_base := deBruijn.deBruijn_fn_base_pair action deBruijn_base_action deBruijn_fn_base_action constraint deBruijn_base_constraint deBruijn_fn_base_constraint.

  Lemma lift_action : forall (d k:nat)(a:action)(c:constraint),(deBruijn.lift deBruijn_base d k (a,c))=(deBruijn.lift deBruijn_base_action d k a,deBruijn.lift deBruijn_base_constraint d k c).
    intros;reflexivity.
  Qed.

  Lemma swapR_action : forall (d k:nat)(a:action)(c:constraint),(deBruijn.swapR deBruijn_base d k (a,c))=(deBruijn.swapR deBruijn_base_action d k a,deBruijn.swapR deBruijn_base_constraint d k c).
    intros;reflexivity.
  Qed.

  Lemma low1_action : forall (d:nat)(a:action)(c:constraint),(deBruijn.low1 deBruijn_base d (a,c))=(deBruijn.low1 deBruijn_base_action d a,deBruijn.low1 deBruijn_base_constraint d c).
    intros;reflexivity.
  Qed.

  Definition get_kind (a:t) := 
    let (a,_) := a in 
      match a with
	| Inp _ => In
	| Outp _ => Out
	| Tau => Silent
      end.

  Lemma lift_kind : forall (d k:nat)(x:t),(get_kind (deBruijn.lift deBruijn_base d k x))=(get_kind x).
    intros.
    unfold get_kind.
    destruct x.
    rewrite lift_action.
    destruct a;reflexivity.
  Qed.

  Lemma swapR_kind : forall (d k:nat)(x:t),(get_kind (deBruijn.swapR deBruijn_base d k x))=(get_kind x).
    intros.
    unfold get_kind.
    destruct x.
    rewrite swapR_action.
    destruct a;reflexivity.
  Qed.

  Definition mk_input (e:A.P.E.t) : option t := Some (Inp (EvalSpi.evala e),(0%nat,cons (FormulaeSpi.FName e) nil)).

  Definition mk_output (e:A.P.E.t) (f:A.P.E.t) : option (t * A.P.E.t) := Some ((Outp (EvalSpi.evala e),(0%nat,cons (FormulaeSpi.FName e) (cons (FormulaeSpi.Wff f) nil))),EvalSpi.evala f).

  Definition mk_silent : t := (Tau,(0%nat,nil)).

  Lemma mk_input_kind : forall (e:A.P.E.t)(a:t),(mk_input e = Some a)->(get_kind a = In).
    unfold mk_input.
    intros.
    injection H;intros.
    subst a;reflexivity.
  Qed.

  Lemma mk_output_kind : forall (e f:A.P.E.t)(a:t)(m:A.P.E.t),(mk_output e f = Some (a,m))->(get_kind a = Out).
    unfold mk_output.
    intros.
    injection H;intros.
    subst a;reflexivity.
  Qed.

  Lemma mk_silent_kind : get_kind mk_silent = Silent.
    reflexivity.
  Qed.

  Definition mk_if (phi:A.P.F.t) (a:t) : option t :=
    let (alpha,c) := a in Some (alpha,union_c c (0%nat,cons phi nil)).

  Lemma mk_if_kind : forall (phi:A.P.F.t)(a b:t),(mk_if phi a = Some b)->(get_kind b = get_kind a).
    unfold mk_if;intros.
    destruct a.
    injection H;intros;subst b.
    reflexivity.
  Qed.

  Definition mk_res (a:t) : option t :=
    let (alpha,c) := a in
      if deBruijn.mem_fn deBruijn_fn_base_action 0 0 alpha then None else Some (deBruijn.low1 deBruijn_base_action 0 alpha,new_c c).

  Lemma mk_res_kind : forall (a b:t),(mk_res a = Some b)->(get_kind b = get_kind a).
    unfold mk_res;intros.
    destruct a.
    destruct (deBruijn.mem_fn deBruijn_fn_base_action 0 0 a);try (discriminate H).
    injection H;intros.
    subst b.
    destruct a;reflexivity.
  Qed.

  Definition mk_interact (a:t) (b:t) : option t :=
    match (a,b) with
      | ((Inp e,cinp),(Outp f,cout)) =>
	Some (Tau,union_c (0%nat,cons (FormulaeSpi.FEqual e f) nil) (union_c cinp cout))
      | ((Outp f,cout),(Inp e,cinp)) =>
	Some (Tau,union_c (0%nat,cons (FormulaeSpi.FEqual e f) nil) (union_c cinp cout))
      | _ => None
    end.

  Lemma mk_interact_kind : forall (a b c:t),(mk_interact a b = Some c)->(get_kind c = Silent).
    destruct a as [alpha ca];destruct b as [beta cb];unfold mk_interact.
    intros.
    destruct alpha;destruct beta;try (discriminate H);injection H;intros;subst c;reflexivity.
  Qed.

  Lemma mk_interact_exg : forall (a b c:t),(mk_interact a b = Some c)->exists c':t,equal c c' /\ (mk_interact b a = Some c').
    destruct a as [alpha ca];destruct b as [beta cb];unfold mk_interact;intros.
    destruct alpha;destruct beta;try (discriminate H);injection H;clear H;intros;subst c.
    exists (Tau, union_c (0%nat, cons (FormulaeSpi.FEqual t0 t1) nil) (union_c ca cb)).
    split;auto.
    apply equal_refl.
    exists (Tau, union_c (0%nat, cons (FormulaeSpi.FEqual t1 t0) nil) (union_c cb ca)).
    split;auto.
    apply equal_refl.
  Qed.

  (* equal properties *)
  Lemma mk_if_equal : forall (phi:A.P.F.t)(a b:t),(mk_if phi a = Some b)->forall (a':t),(equal a a')->(exists b':t,equal b b' /\ (mk_if phi a' = Some b')).
    intros.
    unfold equal in H0.
    destruct a.
    destruct a'.
    and_elim.
    subst a0.
    rewrite <- H1.
    exists b.
    split;auto.
    apply equal_refl.
  Qed.

  Lemma mk_res_equal : forall (a b:t),(mk_res a = Some b)->forall (a':t),(equal a a')->(exists b':t,equal b b' /\ (mk_res a' = Some b')).
    intros.
    unfold equal in H0.
    destruct a;destruct a'.
    and_elim.
    subst a0.
    rewrite <- H1.
    exists b.
    split;auto.
    apply equal_refl.
  Qed.

  Lemma mk_interact_equal : forall (a b c:t),(mk_interact a b = Some c)->forall (a' b':t),(equal a a')->(equal b b')->(exists c':t,equal c c' /\ (mk_interact a' b' = Some c')).
    intros.
    unfold equal in H0.
    unfold equal in H1.
    destruct a;destruct a'.
    destruct b;destruct b'.
    and_elim.
    subst a0;subst a2.
    rewrite <- H2.
    rewrite <- H3.
    exists c.
    split;auto.
    apply equal_refl.
  Qed.

  (* lift *)
  Lemma lift_mk_silent : forall (d k:nat),(equal mk_silent (deBruijn.lift deBruijn_base d k mk_silent)).
    intros.
    unfold mk_silent.
    simpl.
    apply equal_refl.
  Qed.
    
  Lemma lift_mk_input : forall (e:A.P.E.t)(inp:t),(mk_input e = Some inp)->forall (d k:nat),(exists inp':t,mk_input (deBruijn.lift A.P.E.deBruijn_base d k e) = Some inp' /\ (equal inp' (deBruijn.lift deBruijn_base d k inp))).
    unfold mk_input;intros.
    injection H;intros.
    subst inp.
    rewrite lift_action.
    rewrite lift_Inp.
    rewrite lift_constraint.
    rewrite plus_0_r.
    exists (Inp (EvalSpi.evala (deBruijn.lift A.P.E.deBruijn_base d k e)),(0%nat,cons (FormulaeSpi.FName (deBruijn.lift A.P.E.deBruijn_base d k e)) nil)).
    split;auto.
    simpl.
    rewrite EvalSpi.evala_map.
    apply equal_refl.
  Qed.

  Lemma lift_mk_output : forall (e f:A.P.E.t)(out:t)(m:A.P.E.t),(mk_output e f = Some (out,m))->forall (d k:nat),(exists out':t,mk_output (deBruijn.lift A.P.E.deBruijn_base d k e) (deBruijn.lift A.P.E.deBruijn_base d k f) = Some (out',deBruijn.lift A.P.E.deBruijn_base d k m) /\ (equal out' (deBruijn.lift deBruijn_base d k out))).
    unfold mk_output;intros.
    injection H;intros.
    subst out.
    rewrite lift_action.
    rewrite lift_Outp.
    rewrite lift_constraint.
    rewrite plus_0_r.
    exists (Outp (EvalSpi.evala (deBruijn.lift A.P.E.deBruijn_base d k e)),(0%nat,cons (FormulaeSpi.FName (deBruijn.lift A.P.E.deBruijn_base d k e)) (cons (FormulaeSpi.Wff (deBruijn.lift A.P.E.deBruijn_base d k f)) nil))).
    split;auto.
    simpl.
    repeat (rewrite EvalSpi.evala_map);auto.
    subst m;auto.
    simpl.
    repeat (rewrite EvalSpi.evala_map).
    apply equal_refl.
  Qed.

  Lemma lift_mk_if : forall (phi:A.P.F.t)(mu mu':t),(mk_if phi mu = Some mu')->forall (d k:nat),(exists mu'':t,(equal mu'' (deBruijn.lift deBruijn_base d k mu')) /\ mk_if (deBruijn.lift A.P.F.deBruijn_base d k phi) (deBruijn.lift deBruijn_base d k mu) = Some mu'').
    unfold mk_if;intros.
    destruct mu.
    injection H;intros.
    subst mu'.
    repeat (rewrite lift_action).
    exists (deBruijn.lift deBruijn_base_action d k a,deBruijn.lift deBruijn_base_constraint d k (union_c c (0%nat, cons phi nil))).
    split;auto.
    apply equal_refl.
    rewrite lift_union_c.
    rewrite lift_constraint.
    rewrite plus_0_r.
    auto.
  Qed.

  Lemma lift_mk_res : forall (mu mu':t),(mk_res mu = Some mu')->forall (d k:nat),(exists mu'':t,(equal mu'' (deBruijn.lift deBruijn_base d k mu'))/\(mk_res (deBruijn.lift deBruijn_base (S d) k mu))=(Some mu'')).
    unfold mk_res;intros.
    destruct mu as [a ca].
    rewrite lift_action.
    case (Bool_elim_bool (deBruijn.mem_fn deBruijn_fn_base_action 0 0 a));intro H0;rewrite H0 in H;try (discriminate H).
    generalize (deBruijn.lift_In_fn deBruijn_fn_base_action d 0 (le_O_n d) 0 k a).
    rewrite H0.
    fold deBruijn_base_action.
    rewrite plus_0_l.
    intro.
    rewrite <- H1.
    generalize (deBruijn.lift_low1_comm deBruijn_fn_base_action 0 d k a H0).
    fold deBruijn_base_action.
    repeat (rewrite plus_0_l).
    intro.
    rewrite H2.
    exists (deBruijn.lift deBruijn_base d k mu').
    split.
    apply equal_refl.
    injection H;intros;subst mu'.
    rewrite <- lift_new_c.
    auto.
  Qed.

  Lemma lift_mk_interact : forall (a b c:t),(mk_interact a b = Some c)->forall (d k:nat),(exists c':t,(equal c' (deBruijn.lift deBruijn_base d k c))/\(mk_interact (deBruijn.lift deBruijn_base d k a) (deBruijn.lift deBruijn_base d k b) = Some c')).
    unfold mk_interact.
    destruct a as [a ca].
    destruct b as [b cb].
    intros.
    destruct a;destruct b;try (discriminate H);repeat (rewrite lift_action || rewrite lift_Inp || rewrite lift_Outp).
    exists (deBruijn.lift deBruijn_base d k c). 
    split.
    apply equal_refl.
    injection H;clear H;intros.
    subst c.
    rewrite lift_action.
    repeat (rewrite lift_union_c).
    rewrite lift_constraint.
    rewrite plus_0_r.
    rewrite lift_Tau.
    reflexivity.
    exists (deBruijn.lift deBruijn_base d k c). 
    split.
    apply equal_refl.
    injection H;clear H;intros.
    subst c.
    rewrite lift_action.
    repeat (rewrite lift_union_c).
    rewrite lift_constraint.
    rewrite plus_0_r.
    rewrite lift_Tau.
    reflexivity.
  Qed.

  Lemma lift_mk_input_inv : forall (d k:nat)(e:A.P.E.t)(inp':t),(mk_input (deBruijn.lift A.P.E.deBruijn_base d k e) = Some inp')->(exists inp:t,mk_input e = Some inp).
    unfold mk_input;intros.
    exists (Inp (EvalSpi.evala e), (0%nat, cons (FormulaeSpi.FName e) nil)).
    auto.
  Qed.

  Lemma lift_mk_output_inv : forall (d k:nat)(e f:A.P.E.t)(out':t)(m':A.P.E.t),(mk_output (deBruijn.lift A.P.E.deBruijn_base d k e) (deBruijn.lift A.P.E.deBruijn_base d k f) = Some (out',m'))->(exists out:t,exists m:A.P.E.t,mk_output e f = Some (out,m) /\ m' = deBruijn.lift A.P.E.deBruijn_base d k m).
    intros.
    unfold mk_output.
    exists (Outp (EvalSpi.evala e),(0%nat, cons (FormulaeSpi.FName e) (cons (FormulaeSpi.Wff f) nil))).
    exists (EvalSpi.evala f).
    split;auto.
    simpl.
    injection H;intros.
    subst m'.
    rewrite EvalSpi.evala_map.
    auto.
  Qed.

  Lemma lift_mk_if_inv : forall (d k:nat)(phi:A.P.F.t)(mu mu'':t),(mk_if (deBruijn.lift A.P.F.deBruijn_base d k phi) (deBruijn.lift deBruijn_base d k mu) = Some mu'')->(exists mu':t,(mk_if phi mu = Some mu')).
    intros.
    unfold mk_if.
    destruct mu.
    exists (a,union_c c (0%nat,cons phi nil)).
    auto.
  Qed.

  Lemma lift_mk_res_inv : forall (d k:nat)(mu mu'':t),(mk_res (deBruijn.lift deBruijn_base (S d) k mu) = Some mu'')->(exists mu':t,mk_res mu = Some mu').
    intros.
    unfold mk_res.
    unfold mk_res in H.
    destruct mu.
    rewrite lift_action in H.
    case (Bool_elim_bool (deBruijn.mem_fn deBruijn_fn_base_action 0 0 (deBruijn.lift deBruijn_base_action (S d) k a)));intro H0;rewrite H0 in H;try (discriminate H).
    generalize (deBruijn.lift_In_fn deBruijn_fn_base_action d 0 (le_O_n d) 0 k).
    generalize H0.
    fold deBruijn_base_action.
    intros.
    rewrite plus_0_l in H2.
    rewrite <- H2 in H1.
    rewrite H1.
    exists (deBruijn.low1 deBruijn_base_action 0 a, new_c c).
    auto.
  Qed.

  Lemma lift_mk_interact_inv : forall (d k:nat)(a b c':t),(mk_interact (deBruijn.lift deBruijn_base d k a) (deBruijn.lift deBruijn_base d k b) = Some c')->(exists c:t,mk_interact a b = Some c).
    intros.
    unfold mk_interact in H.
    unfold mk_interact.
    destruct a as [a ca].
    destruct b as [b cb].
    repeat (rewrite lift_action in H).
    destruct a;destruct b;try (discriminate H);repeat (rewrite lift_Inp in H || rewrite lift_Outp in H).
    exists (Tau, union_c (0%nat, cons (FormulaeSpi.FEqual t0 t1) nil) (union_c ca cb)).
    auto.
    exists (Tau, union_c (0%nat, cons (FormulaeSpi.FEqual t1 t0) nil) (union_c cb ca)).
    auto.
  Qed.

  (* swapR *)
  Lemma swapR_mk_silent : forall (d k:nat),(equal mk_silent (deBruijn.swapR deBruijn_base d k mk_silent)).
    intros.
    unfold mk_silent.
    simpl.
    apply equal_refl.
  Qed.
    
  Lemma swapR_mk_input : forall (e:A.P.E.t)(inp:t),(mk_input e = Some inp)->forall (d k:nat),(exists inp':t,mk_input (deBruijn.swapR A.P.E.deBruijn_base d k e) = Some inp' /\ (equal inp' (deBruijn.swapR deBruijn_base d k inp))).
    unfold mk_input;intros.
    injection H;intros.
    subst inp.
    rewrite swapR_action.
    rewrite swapR_Inp.
    rewrite swapR_constraint.
    rewrite plus_0_r.
    exists (Inp (EvalSpi.evala (deBruijn.swapR A.P.E.deBruijn_base d k e)),(0%nat,cons (FormulaeSpi.FName (deBruijn.swapR A.P.E.deBruijn_base d k e)) nil)).
    split;auto.
    simpl.
    rewrite EvalSpi.evala_map.
    apply equal_refl.
  Qed.

  Lemma swapR_mk_output : forall (e f:A.P.E.t)(out:t)(m:A.P.E.t),(mk_output e f = Some (out,m))->forall (d k:nat),(exists out':t,mk_output (deBruijn.swapR A.P.E.deBruijn_base d k e) (deBruijn.swapR A.P.E.deBruijn_base d k f) = Some (out',deBruijn.swapR A.P.E.deBruijn_base d k m) /\ (equal out' (deBruijn.swapR deBruijn_base d k out))).
    unfold mk_output;intros.
    injection H;intros.
    subst out.
    rewrite swapR_action.
    rewrite swapR_Outp.
    rewrite swapR_constraint.
    rewrite plus_0_r.
    exists (Outp (EvalSpi.evala (deBruijn.swapR A.P.E.deBruijn_base d k e)),(0%nat,cons (FormulaeSpi.FName (deBruijn.swapR A.P.E.deBruijn_base d k e)) (cons (FormulaeSpi.Wff (deBruijn.swapR A.P.E.deBruijn_base d k f)) nil))).
    split;auto.
    simpl.
    repeat (rewrite EvalSpi.evala_map);auto.
    subst m;auto.
    simpl.
    repeat (rewrite EvalSpi.evala_map).
    apply equal_refl.
  Qed.

  Lemma swapR_mk_if : forall (phi:A.P.F.t)(mu mu':t),(mk_if phi mu = Some mu')->forall (d k:nat),(exists mu'':t,(equal mu'' (deBruijn.swapR deBruijn_base d k mu')) /\ mk_if (deBruijn.swapR A.P.F.deBruijn_base d k phi) (deBruijn.swapR deBruijn_base d k mu) = Some mu'').
    unfold mk_if;intros.
    destruct mu.
    injection H;intros.
    subst mu'.
    repeat (rewrite swapR_action).
    exists (deBruijn.swapR deBruijn_base_action d k a,deBruijn.swapR deBruijn_base_constraint d k (union_c c (0%nat, cons phi nil))).
    split;auto.
    apply equal_refl.
    rewrite swapR_union_c.
    rewrite swapR_constraint.
    rewrite plus_0_r.
    auto.
  Qed.

  Lemma swapR_mk_res : forall (mu mu':t),(mk_res mu = Some mu')->forall (d k:nat),(exists mu'':t,(equal mu'' (deBruijn.swapR deBruijn_base d k mu'))/\(mk_res (deBruijn.swapR deBruijn_base (S d) k mu))=(Some mu'')).
    unfold mk_res;intros.
    destruct mu as [a ca].
    rewrite swapR_action.
    case (Bool_elim_bool (deBruijn.mem_fn deBruijn_fn_base_action 0 0 a));intro H0;rewrite H0 in H;try (discriminate H).
    generalize (deBruijn.swapR_In_fn deBruijn_fn_base_action d 0 (le_O_n d) 0 k a).
    rewrite H0.
    fold deBruijn_base_action.
    rewrite plus_0_l.
    intro.
    rewrite <- H1.
    generalize (deBruijn.swapR_low1_comm deBruijn_fn_base_action 0 d k a H0).
    fold deBruijn_base_action.
    repeat (rewrite plus_0_l).
    intro.
    rewrite H2.
    exists (deBruijn.swapR deBruijn_base d k mu').
    split.
    apply equal_refl.
    injection H;intros;subst mu'.
    rewrite <- swapR_new_c.
    auto.
  Qed.

  Lemma swapR_mk_interact : forall (a b c:t),(mk_interact a b = Some c)->forall (d k:nat),(exists c':t,(equal c' (deBruijn.swapR deBruijn_base d k c))/\(mk_interact (deBruijn.swapR deBruijn_base d k a) (deBruijn.swapR deBruijn_base d k b) = Some c')).
    unfold mk_interact.
    destruct a as [a ca].
    destruct b as [b cb].
    intros.
    destruct a;destruct b;try (discriminate H);repeat (rewrite swapR_action || rewrite swapR_Inp || rewrite swapR_Outp).
    exists (deBruijn.swapR deBruijn_base d k c). 
    split.
    apply equal_refl.
    injection H;clear H;intros.
    subst c.
    rewrite swapR_action.
    repeat (rewrite swapR_union_c).
    rewrite swapR_constraint.
    rewrite plus_0_r.
    rewrite swapR_Tau.
    reflexivity.
    exists (deBruijn.swapR deBruijn_base d k c). 
    split.
    apply equal_refl.
    injection H;clear H;intros.
    subst c.
    rewrite swapR_action.
    repeat (rewrite swapR_union_c).
    rewrite swapR_constraint.
    rewrite plus_0_r.
    rewrite swapR_Tau.
    reflexivity.
  Qed.
End ActionSpiSymb.

(*   Inductive equal_c : constraint -> constraint -> Prop := *)
(*     | EqualRefl : forall (x:constraint),(equal_c x x) *)
(*     | EqualTrans : forall (x y z:constraint),(equal_c x y)->(equal_c y z)->(equal_c x z) *)
(*     | EqualNew : forall (x y:constraint),(equal_c x y)->(equal_c (new_c x) (new_c y)) *)
(*     | EqualNewSwap : forall (x:constraint),(equal_c (new_c (new_c x)) (new_c (new_c (deBruijn.swapR deBruijn_base_constraint 0 1 x)))) *)
(*     | EqualScopeL : forall (x y:constraint),(equal_c (union_c (new_c x) y) (new_c (union_c x (deBruijn.lift deBruijn_base_constraint 0 1 y)))) *)
(*     | EqualScopeLsym : forall (x y:constraint),(equal_c (new_c (union_c x (deBruijn.lift deBruijn_base_constraint 0 1 y))) (union_c (new_c x) y)) *)
(*     | EqualScopeR : forall (x y:constraint),(equal_c (union_c x (new_c y)) (new_c (union_c (deBruijn.lift deBruijn_base_constraint 0 1 x) y))) *)
(*     | EqualScopeRsym : forall (x y:constraint),(equal_c (new_c (union_c (deBruijn.lift deBruijn_base_constraint 0 1 x) y)) (union_c x (new_c y))) *)
(*     | EqualUnionL : forall (x y:constraint),(equal_c x y)->forall (z:constraint),(equal_c (union_c z x) (union_c z y)) *)
(*     | EqualUnionR : forall (x y:constraint),(equal_c x y)->forall (z:constraint),(equal_c (union_c x z) (union_c y z)). *)

(*   Lemma equal_c_refl : forall (x:constraint),(equal_c x x). *)
(*     intros. *)
(*     apply EqualRefl. *)
(*   Qed. *)

(*   Lemma equal_c_sym : forall (x y:constraint),(equal_c x y)->(equal_c y x). *)
(*     induction 1. *)
(*     apply EqualRefl. *)
(*     eapply EqualTrans;eauto. *)
(*     apply EqualNew;auto. *)
(*     replace (new_c (new_c x)) with (new_c (new_c (deBruijn.swapR deBruijn_base_constraint 0 1 (deBruijn.swapR deBruijn_base_constraint 0 1 x)))). *)
(*     apply EqualNewSwap. *)
(*     rewrite deBruijn.swapR_1_idem;auto. *)
(*     apply EqualScopeLsym. *)
(*     apply EqualScopeL. *)
(*     apply EqualScopeRsym. *)
(*     apply EqualScopeR. *)
(*     apply EqualUnionL;auto. *)
(*     apply EqualUnionR;auto. *)
(*   Qed. *)

(*   Lemma equal_c_trans : forall (x y z:constraint),(equal_c x y)->(equal_c y z)->(equal_c x z). *)
(*     intros. *)
(*     eapply EqualTrans;eauto. *)
(*   Qed. *)

(*   Lemma new_c_scope_1 : forall (x y:constraint),(union_c (new_c x) y)=(new_c (union_c x (deBruijn.lift deBruijn_base_constraint 0 1 y))). *)
(*     intros. *)
(*     unfold union_c, new_c. *)
(*     destruct x as [n lphi]. *)
(*     destruct y as [m lpsi]. *)
(*     repeat (rewrite lift_constraint). *)
(*     repeat (rewrite plus_0_l). *)
(*     rewrite mem_fn_app. *)
(*     generalize (mem_fn_union_c 0 0 (n,lphi) (m,(deBruijn.lift deBruijn_base_list m 1 lpsi))). *)
(*     unfold union_c. *)
(*     repeat (rewrite mem_fn_constraint). *)
(*     repeat (rewrite plus_0_l). *)
(*     rewrite mem_fn_app. *)
(*     intro. *)
(*     rewrite H. *)
(*     case (Bool_elim_bool (deBruijn.mem_fn deBruijn_fn_base_list 0 n lphi));intro. *)
(*     rewrite H0. *)
(*     rewrite orb_true_b. *)
(*     rewrite deBruijn.lift_plus. *)
(*     rewrite <- plus_n_Sm. *)
(*     rewrite plus_0_r. *)
(*     reflexivity. *)
(*     rewrite H0. *)
(*     rewrite orb_false_b. *)
(*     generalize (deBruijn.lift_not_fn deBruijn_fn_base_list m 1 0).     *)
(*     fold deBruijn_base_list. *)
(*     intros. *)
(*     rewrite H1;auto. *)
(*     rewrite low1_app. *)
(*     rewrite deBruijn.lift_plus. *)
(*     generalize (deBruijn.lift_compose deBruijn_base_list m n 1 0). *)
(*     repeat (rewrite plus_0_r). *)
(*     intros. *)
(*     rewrite <- H2. *)
(*     rewrite (plus_comm m n). *)
(*     rewrite (deBruijn.low1_lift_simpl deBruijn_base_list (n+m) 0). *)
(*     rewrite deBruijn.lift_0. *)
(*   Qed. *)

(*   Lemma new_c_scope_2 : forall (l l':constraint),(app l (new_c l'))=(new_c (app (deBruijn.lift deBruijn_base_list 0 1 l) l')). *)
(*     intros. *)
(*     unfold new_c. *)
(*     rewrite remove_app. *)
(*     rewrite low1_app. *)
(*     rewrite (remove_nothing 0 (deBruijn.lift deBruijn_base_list 0 1 l)). *)
(*     rewrite (deBruijn.low1_lift_simpl deBruijn_base_list 0 0 l). *)
(*     rewrite deBruijn.lift_0;auto. *)
(*     apply (deBruijn.lift_not_fn deBruijn_fn_base_list 0 1 0). *)
(*     auto. *)
(*   Qed. *)
 
(*   Lemma new_c_swap : forall (l:constraint),new_c (new_c (deBruijn.swapR deBruijn_base_list 0 1 l)) = (new_c (new_c l)). *)
(*     induction l;try reflexivity. *)
(*     replace (a::l) with ((a::nil) ++ l);try reflexivity. *)
(*     rewrite swapR_app. *)
(*     unfold new_c. *)
(*     repeat (rewrite remove_app || rewrite low1_app). *)
(*     fold (new_c (deBruijn.swapR deBruijn_base_list 0 1 l)). *)
(*     fold (new_c (new_c (deBruijn.swapR deBruijn_base_list 0 1 l))). *)
(*     rewrite IHl. *)
(*     fold (new_c l). *)
(*     fold (new_c (new_c l)). *)
(*     destruct a. *)
(*     destruct n;try reflexivity. *)
(*     destruct n;try reflexivity. *)
(*   Qed. *)

  (* equal properties *)
(*   Lemma mk_res_lift_1 : forall (a:t),exists b:t,(equal b a)/\(mk_res (deBruijn.lift deBruijn_base 0 1 a) = Some b). *)
(*     unfold mk_res;intros. *)
(*     destruct a as [alpha ca]. *)
(*     rewrite lift_action. *)
(*     case (Bool_elim_bool (deBruijn.mem_fn deBruijn_fn_base_action 0 0 (deBruijn.lift deBruijn_base_action 0 1 alpha)));intro H;rewrite H. *)
(*     rewrite deBruijn.lift_not_fn in H;try (discriminate H). *)
(*     omega. *)
(*     rewrite (deBruijn.low1_lift_simpl deBruijn_base_action 0 0). *)
(*     rewrite deBruijn.lift_0. *)
(*     unfold new_c. *)
(*     destruct ca. *)
(*     rewrite lift_constraint. *)
(*     rewrite plus_0_l. *)
(*     rewrite deBruijn.lift_not_fn;try omega. *)
(*     rewrite (deBruijn.low1_lift_simpl deBruijn_base_list). *)
(*     rewrite deBruijn.lift_0. *)
(*     exists (alpha,(n,l)). *)
(*     split;auto. *)
(*     apply equal_refl. *)
(*   Qed. *)

(*   Lemma mk_interact_res_lift : forall (a a':t),(mk_res a = Some a')->forall (b c'':t),(mk_interact a' b = Some c'')->(exists c:t,exists c':t,(equal c' c'')/\(mk_interact a (deBruijn.lift deBruijn_base 0 1 b) = Some c)/\(mk_res c = Some c')). *)
(*     unfold mk_res. *)
(*     destruct a as [alpha ca];intro a'. *)
(*     case (Bool_elim_bool (deBruijn.mem_fn deBruijn_fn_base_action 0 0 alpha));intro H;rewrite H. *)
(*     intro H0;discriminate H0. *)
(*     intro. *)
(*     replace a' with (deBruijn.low1 deBruijn_base_action 0 alpha, new_c ca);try (injection H0;intros;subst a';auto;fail). *)
(*     destruct b as [beta cb];intro c''. *)
(*     unfold mk_interact. *)
(*     intros. *)
(*     destruct alpha;destruct beta;try (discriminate H1);repeat (rewrite lift_action);[rewrite low1_Inp in H1;rewrite lift_Outp | rewrite low1_Outp in H1;rewrite lift_Inp]. *)
(*     exists (Tau,union_c (0%nat,cons (FormulaeSpi.FEqual t0 (deBruijn.lift ExpressionSpi.deBruijn_base 0 1 t1)) nil) (union_c ca (deBruijn.lift deBruijn_base_constraint 0 1 cb))). *)
(*     rewrite low1_Tau. *)
(*     exists (Tau,new_c (union_c (0%nat,cons (FormulaeSpi.FEqual t0 (deBruijn.lift ExpressionSpi.deBruijn_base 0 1 t1)) nil) (union_c ca (deBruijn.lift deBruijn_base_constraint 0 1 cb)))). *)
(*     split;auto. *)
(*     generalize (deBruijn.lift_1_low1_simpl ExpressionSpi.deBruijn_fn_base 0 0 t0 H). *)
(*     rewrite deBruijn.swapR_0. *)
(*     rewrite plus_0_l. *)
(*     fold ExpressionSpi.deBruijn_base. *)
(*     intro. *)
(*     rewrite <- H2. *)
(*     rewrite <- FormulaeSpi.lift_FEqual. *)
(*     injection H1;intros;subst c''. *)
(*     unfold equal. *)
(*     split;auto. *)
(*     eapply equal_c_trans. *)
(*     apply (EqualScopeRsym (0%nat,(cons (FormulaeSpi.FEqual (deBruijn.low1 ExpressionSpi.deBruijn_base 0 t0) t1) nil))). *)
(*     apply EqualUnionL. *)
(*     apply EqualScopeLsym. *)
(*     exists (Tau,union_c (0%nat,cons (FormulaeSpi.FEqual (deBruijn.lift ExpressionSpi.deBruijn_base 0 1 t1) t0) nil) (union_c (deBruijn.lift deBruijn_base_constraint 0 1 cb) ca)). *)
(*     rewrite low1_Tau. *)
(*     exists (Tau,new_c (union_c (0%nat,cons (FormulaeSpi.FEqual (deBruijn.lift ExpressionSpi.deBruijn_base 0 1 t1) t0) nil) (union_c (deBruijn.lift deBruijn_base_constraint 0 1 cb) ca))). *)
(*     split;auto. *)
(*     generalize (deBruijn.lift_1_low1_simpl ExpressionSpi.deBruijn_fn_base 0 0 t0 H). *)
(*     rewrite deBruijn.swapR_0. *)
(*     rewrite plus_0_l. *)
(*     fold ExpressionSpi.deBruijn_base. *)
(*     intro. *)
(*     rewrite <- H2. *)
(*     rewrite <- FormulaeSpi.lift_FEqual. *)
(*     injection H1;intros;subst c''. *)
(*     unfold equal. *)
(*     split;auto. *)
(*     eapply equal_c_trans. *)
(*     apply (EqualScopeRsym (0%nat,(cons (FormulaeSpi.FEqual t1 (deBruijn.low1 ExpressionSpi.deBruijn_base 0 t0)) nil))). *)
(*     apply EqualUnionL. *)
(*     apply EqualScopeRsym. *)
(*   Qed. *)

(*   Lemma mk_interact_res_lift_inv : forall (a b c:t),(mk_interact a (deBruijn.lift deBruijn_base 0 1 b) = Some c)->forall (c'':t),(mk_res c = Some c'')->exists a':t,(mk_res a = Some a')/\exists c':t,(mk_interact a' b = Some c'). *)
(*     intros a b c. *)
(*     unfold mk_interact. *)
(*     destruct a as [alpha la]. *)
(*     destruct b as [beta lb]. *)
(*     rewrite lift_action. *)
(*     unfold mk_res. *)
(*     intros. *)
(*     destruct alpha;destruct beta;try (discriminate H). *)
(*     rewrite lift_Outp in H. *)
(*     rewrite mem_fn_Inp. *)
(*     rewrite low1_Inp. *)
(*   Qed. *)
    
(*   Lemma mk_res_swapR : forall (a b:t),(mk_res a = Some b)->forall (c:t),(mk_res b = Some c)->(exists b':t,exists c':t,(equal c' c)/\(mk_res (deBruijn.swapR deBruijn_base 0 1 a) = Some b')/\(mk_res b' = Some c')). *)
(*     destruct a as [alpha la]. *)
(*     destruct b as [beta lb]. *)
(*     unfold mk_res. *)
(*     generalize (ActionSpi.mk_res_swapR alpha). *)
(*     destruct (ActionSpi.mk_res alpha). *)
(*     intros. *)
(*     generalize (H t0 (refl_equal (Some t0)));clear H;intro. *)
(*     injection H0;clear H0;intros;subst t0. *)
(*     destruct (ActionSpi.mk_res beta);try (discriminate H1). *)
(*     elim (H t0 (refl_equal (Some t0)));clear H. *)
(*     intro beta';intros. *)
(*     ex_elim H gamma'. *)
(*     rewrite swapR_action. *)
(*     rewrite H2. *)
(*     exists (beta',new_c (deBruijn.swapR deBruijn_base_list 0 1 la)). *)
(*     rewrite H3. *)
(*     exists (gamma',new_c (new_c (deBruijn.swapR deBruijn_base_list 0 1 la))). *)
(*     unfold equal. *)
(*     injection H1;clear H1;intros;subst c. *)
(*     split;auto. *)
(*     rewrite H. *)
(*     rewrite <- H0. *)
(*     rewrite new_c_swap. *)
(*     auto. *)
(*     intros. *)
(*     discriminate H0. *)
(*   Qed. *)
