(* A formalisation of the spi calculus in Coq *)
(* Copyright (C) 2005-2007 Sebastien Briais *)
(* http://lamp.epfl.ch/~sbriais/ *)

(* This library is free software; you can redistribute it and/or modify *)
(* it under the terms of the GNU Lesser General Public License as *)
(* published by the Free Software Foundation; either version 2.1 of the *)
(* License, or (at your option) any later version. *)

(* This library is distributed in the hope that it will be useful, but *)
(* WITHOUT ANY WARRANTY; without even the implied warranty of *)
(* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU *)
(* Lesser General Public License for more details. *)

(* You should have received a copy of the GNU Lesser General Public *)
(* License along with this library; if not, write to the Free Software *)
(* Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA *)
(* 02110-1301 USA *)

Require Export expressionType.
Require Export formulaeType.
Require Export processType.
Require Export agentType.
Require Import deBruijnType.

Module Type ActionTypeLight.
  Declare Module A : AgentType.

  Inductive kind : Set :=
    | In : kind
    | Out : kind
    | Silent : kind.

  (* type of action *)
  (* they can be guarded by condition but the condition should have its own binder if needed *)
  Parameter t : Set.

  Parameter equal : t -> t -> Prop.
  Parameter equal_refl : forall (x:t),(equal x x).
  Parameter equal_sym : forall (x y:t),(equal x y)->(equal y x).
  Parameter equal_trans : forall (x y z:t),(equal x y)->(equal y z)->(equal x z).

  Declare Module Ord : OrderedType with Definition t := t with Definition eq := @eq t.

  Parameter deBruijn_base : deBruijn.deBruijn_base t.
  Parameter deBruijn_fn_base : deBruijn.deBruijn_fn t deBruijn_base.

  Parameter get_kind : t -> kind.

  Parameter lift_kind : forall (d k:nat)(x:t),(get_kind (deBruijn.lift deBruijn_base d k x))=(get_kind x).
  Parameter swapR_kind : forall (d k:nat)(x:t),(get_kind (deBruijn.swapR deBruijn_base d k x))=(get_kind x).

  (* basic transitions *)
  Parameter mk_input : A.P.E.t -> option t.
  Parameter mk_output : A.P.E.t -> A.P.E.t -> option (t * A.P.E.t).
  Parameter mk_silent : t.

  Parameter mk_input_kind : forall (e:A.P.E.t)(a:t),(mk_input e = Some a)->(get_kind a = In).
  Parameter mk_output_kind : forall (e f:A.P.E.t)(a:t)(m:A.P.E.t),(mk_output e f = Some (a,m))->(get_kind a = Out).
  Parameter mk_silent_kind : get_kind mk_silent = Silent.

  Parameter mk_if : A.P.F.t -> t -> option t.
  Parameter mk_if_kind : forall (phi:A.P.F.t)(a b:t),(mk_if phi a = Some b)->(get_kind b = get_kind a).

  (* mk_res succeeds if the name 0 is not present in the transition *)
  Parameter mk_res : t -> option t.
  Parameter mk_res_kind : forall (a b:t),(mk_res a = Some b)->(get_kind b = get_kind a).

  (* interaction *)
  Parameter mk_interact : t -> t -> option t.
  Parameter mk_interact_kind : forall (a b c:t),(mk_interact a b = Some c)->(get_kind c = Silent).

  Parameter mk_interact_exg : forall (a b c:t),(mk_interact a b = Some c)->exists c':t,equal c c' /\ (mk_interact b a = Some c').

  (* equal properties *)
  Parameter mk_if_equal : forall (phi:A.P.F.t)(a b:t),(mk_if phi a = Some b)->forall (a':t),(equal a a')->(exists b':t,equal b b' /\ (mk_if phi a' = Some b')).
  Parameter mk_res_equal : forall (a b:t),(mk_res a = Some b)->forall (a':t),(equal a a')->(exists b':t,equal b b' /\ (mk_res a' = Some b')).
  Parameter mk_interact_equal : forall (a b c:t),(mk_interact a b = Some c)->forall (a' b':t),(equal a a')->(equal b b')->(exists c':t,equal c c' /\ (mk_interact a' b' = Some c')).
  
  (* lift *)
  Parameter lift_mk_silent : forall (d k:nat),(equal mk_silent (deBruijn.lift deBruijn_base d k mk_silent)).
  Parameter lift_mk_input : forall (e:A.P.E.t)(inp:t),(mk_input e = Some inp)->forall (d k:nat),(exists inp':t,mk_input (deBruijn.lift A.P.E.deBruijn_base d k e) = Some inp' /\ (equal inp' (deBruijn.lift deBruijn_base d k inp))).
  Parameter lift_mk_output : forall (e f:A.P.E.t)(out:t)(m:A.P.E.t),(mk_output e f = Some (out,m))->forall (d k:nat),(exists out':t,mk_output (deBruijn.lift A.P.E.deBruijn_base d k e) (deBruijn.lift A.P.E.deBruijn_base d k f) = Some (out',deBruijn.lift A.P.E.deBruijn_base d k m) /\ (equal out' (deBruijn.lift deBruijn_base d k out))).
  Parameter lift_mk_if : forall (phi:A.P.F.t)(mu mu':t),(mk_if phi mu = Some mu')->forall (d k:nat),(exists mu'':t,(equal mu'' (deBruijn.lift deBruijn_base d k mu')) /\ mk_if (deBruijn.lift A.P.F.deBruijn_base d k phi) (deBruijn.lift deBruijn_base d k mu) = Some mu'').
  Parameter lift_mk_res : forall (mu mu':t),(mk_res mu = Some mu')->forall (d k:nat),(exists mu'':t,(equal mu'' (deBruijn.lift deBruijn_base d k mu'))/\(mk_res (deBruijn.lift deBruijn_base (S d) k mu))=(Some mu'')).
  Parameter lift_mk_interact : forall (a b c:t),(mk_interact a b = Some c)->forall (d k:nat),(exists c':t,(equal c' (deBruijn.lift deBruijn_base d k c))/\(mk_interact (deBruijn.lift deBruijn_base d k a) (deBruijn.lift deBruijn_base d k b) = Some c')).

  Parameter lift_mk_input_inv : forall (d k:nat)(e:A.P.E.t)(inp':t),(mk_input (deBruijn.lift A.P.E.deBruijn_base d k e) = Some inp')->(exists inp:t,mk_input e = Some inp).
  Parameter lift_mk_output_inv : forall (d k:nat)(e f:A.P.E.t)(out':t)(m':A.P.E.t),(mk_output (deBruijn.lift A.P.E.deBruijn_base d k e) (deBruijn.lift A.P.E.deBruijn_base d k f) = Some (out',m'))->(exists out:t,exists m:A.P.E.t,mk_output e f = Some (out,m) /\ m' = deBruijn.lift A.P.E.deBruijn_base d k m).
  Parameter lift_mk_if_inv : forall (d k:nat)(phi:A.P.F.t)(mu mu'':t),(mk_if (deBruijn.lift A.P.F.deBruijn_base d k phi) (deBruijn.lift deBruijn_base d k mu) = Some mu'')->(exists mu':t,(mk_if phi mu = Some mu')).
  Parameter lift_mk_res_inv : forall (d k:nat)(mu mu'':t),(mk_res (deBruijn.lift deBruijn_base (S d) k mu) = Some mu'')->(exists mu':t,mk_res mu = Some mu').
  Parameter lift_mk_interact_inv : forall (d k:nat)(a b c':t),(mk_interact (deBruijn.lift deBruijn_base d k a) (deBruijn.lift deBruijn_base d k b) = Some c')->(exists c:t,mk_interact a b = Some c).

  (* swapR *)
  Parameter swapR_mk_silent : forall (d k:nat),(equal mk_silent (deBruijn.swapR deBruijn_base d k mk_silent)).
  Parameter swapR_mk_input : forall (e:A.P.E.t)(inp:t),(mk_input e = Some inp)->forall (d k:nat),(exists inp':t,mk_input (deBruijn.swapR A.P.E.deBruijn_base d k e) = Some inp' /\ (equal inp' (deBruijn.swapR deBruijn_base d k inp))).
  Parameter swapR_mk_output : forall (e f:A.P.E.t)(out:t)(m:A.P.E.t),(mk_output e f = Some (out,m))->forall (d k:nat),(exists out':t,mk_output (deBruijn.swapR A.P.E.deBruijn_base d k e) (deBruijn.swapR A.P.E.deBruijn_base d k f) = Some (out',deBruijn.swapR A.P.E.deBruijn_base d k m) /\ (equal out' (deBruijn.swapR deBruijn_base d k out))).
  Parameter swapR_mk_if : forall (phi:A.P.F.t)(mu mu':t),(mk_if phi mu = Some mu')->forall (d k:nat),(exists mu'':t,(equal mu'' (deBruijn.swapR deBruijn_base d k mu')) /\ mk_if (deBruijn.swapR A.P.F.deBruijn_base d k phi) (deBruijn.swapR deBruijn_base d k mu) = Some mu'').
  Parameter swapR_mk_res : forall (mu mu':t),(mk_res mu = Some mu')->forall (d k:nat),(exists mu'':t,(equal mu'' (deBruijn.swapR deBruijn_base d k mu'))/\(mk_res (deBruijn.swapR deBruijn_base (S d) k mu))=(Some mu'')).
  Parameter swapR_mk_interact : forall (a b c:t),(mk_interact a b = Some c)->forall (d k:nat),(exists c':t,(equal c' (deBruijn.swapR deBruijn_base d k c))/\(mk_interact (deBruijn.swapR deBruijn_base d k a) (deBruijn.swapR deBruijn_base d k b) = Some c')).
End ActionTypeLight.

Module Type ActionType.
  Declare Module A : AgentType.

  Inductive kind : Set :=
    | In : kind
    | Out : kind
    | Silent : kind.

  (* type of action *)
  (* they can be guarded by condition but the condition should have its own binder if needed *)
  Parameter t : Set.

  Parameter equal : t -> t -> Prop.
  Parameter equal_refl : forall (x:t),(equal x x).
  Parameter equal_sym : forall (x y:t),(equal x y)->(equal y x).
  Parameter equal_trans : forall (x y z:t),(equal x y)->(equal y z)->(equal x z).

  Declare Module Ord : OrderedType with Definition t := t with Definition eq := @eq t.

  Parameter deBruijn_base : deBruijn.deBruijn_base t.

  Parameter deBruijn_fn_base : deBruijn.deBruijn_fn t deBruijn_base.

  Parameter get_kind : t -> kind.

  Parameter lift_kind : forall (d k:nat)(x:t),(get_kind (deBruijn.lift deBruijn_base d k x))=(get_kind x).
  Parameter swapR_kind : forall (d k:nat)(x:t),(get_kind (deBruijn.swapR deBruijn_base d k x))=(get_kind x).

  (* basic transitions *)
  Parameter mk_input : A.P.E.t -> option t.
  Parameter mk_output : A.P.E.t -> A.P.E.t -> option (t * A.P.E.t).
  Parameter mk_silent : t.

  Parameter mk_input_kind : forall (e:A.P.E.t)(a:t),(mk_input e = Some a)->(get_kind a = In).
  Parameter mk_output_kind : forall (e f:A.P.E.t)(a:t)(m:A.P.E.t),(mk_output e f = Some (a,m))->(get_kind a = Out).
  Parameter mk_silent_kind : get_kind mk_silent = Silent.

  Parameter mk_if : A.P.F.t -> t -> option t.
  Parameter mk_if_kind : forall (phi:A.P.F.t)(a b:t),(mk_if phi a = Some b)->(get_kind b = get_kind a).

  (* mk_res succeeds if the name 0 is not present in the transition *)
  Parameter mk_res : t -> option t.
  Parameter mk_res_kind : forall (a b:t),(mk_res a = Some b)->(get_kind b = get_kind a).

  Parameter mk_res_lift_1 : forall (a:t),exists b:t,(equal b a)/\(mk_res (deBruijn.lift deBruijn_base 0 1 a) = Some b).
  Parameter mk_res_swapR : forall (a b:t),(mk_res a = Some b)->forall (c:t),(mk_res b = Some c)->(exists b':t,exists c':t,(equal c' c)/\(mk_res (deBruijn.swapR deBruijn_base 0 1 a) = Some b')/\(mk_res b' = Some c')).

  (* interaction *)
  Parameter mk_interact : t -> t -> option t.
  Parameter mk_interact_kind : forall (a b c:t),(mk_interact a b = Some c)->(get_kind c = Silent).

  Parameter mk_interact_exg : forall (a b c:t),(mk_interact a b = Some c)->exists c':t,equal c c' /\ (mk_interact b a = Some c').
  Parameter mk_interact_res_lift : forall (a a':t),(mk_res a = Some a')->forall (b c'':t),(mk_interact a' b = Some c'')->(exists c:t,exists c':t,(equal c' c'')/\(mk_interact a (deBruijn.lift deBruijn_base 0 1 b) = Some c)/\(mk_res c = Some c')).
  Parameter mk_interact_res_lift_inv : forall (a b c:t),(mk_interact a (deBruijn.lift deBruijn_base 0 1 b) = Some c)->forall (c'':t),(mk_res c = Some c'')->exists a':t,(mk_res a = Some a')/\exists c':t,(mk_interact a' b = Some c').

  (* equal properties *)
  Parameter mk_if_equal : forall (phi:A.P.F.t)(a b:t),(mk_if phi a = Some b)->forall (a':t),(equal a a')->(exists b':t,equal b b' /\ (mk_if phi a' = Some b')).
  Parameter mk_res_equal : forall (a b:t),(mk_res a = Some b)->forall (a':t),(equal a a')->(exists b':t,equal b b' /\ (mk_res a' = Some b')).
  Parameter mk_interact_equal : forall (a b c:t),(mk_interact a b = Some c)->forall (a' b':t),(equal a a')->(equal b b')->(exists c':t,equal c c' /\ (mk_interact a' b' = Some c')).
  
  (* lift *)
  Parameter lift_mk_silent : forall (d k:nat),(equal mk_silent (deBruijn.lift deBruijn_base d k mk_silent)).
  Parameter lift_mk_input : forall (e:A.P.E.t)(inp:t),(mk_input e = Some inp)->forall (d k:nat),(exists inp':t,mk_input (deBruijn.lift A.P.E.deBruijn_base d k e) = Some inp' /\ (equal inp' (deBruijn.lift deBruijn_base d k inp))).
  Parameter lift_mk_output : forall (e f:A.P.E.t)(out:t)(m:A.P.E.t),(mk_output e f = Some (out,m))->forall (d k:nat),(exists out':t,mk_output (deBruijn.lift A.P.E.deBruijn_base d k e) (deBruijn.lift A.P.E.deBruijn_base d k f) = Some (out',deBruijn.lift A.P.E.deBruijn_base d k m) /\ (equal out' (deBruijn.lift deBruijn_base d k out))).
  Parameter lift_mk_if : forall (phi:A.P.F.t)(mu mu':t),(mk_if phi mu = Some mu')->forall (d k:nat),(exists mu'':t,(equal mu'' (deBruijn.lift deBruijn_base d k mu')) /\ mk_if (deBruijn.lift A.P.F.deBruijn_base d k phi) (deBruijn.lift deBruijn_base d k mu) = Some mu'').
  Parameter lift_mk_res : forall (mu mu':t),(mk_res mu = Some mu')->forall (d k:nat),(exists mu'':t,(equal mu'' (deBruijn.lift deBruijn_base d k mu'))/\(mk_res (deBruijn.lift deBruijn_base (S d) k mu))=(Some mu'')).
  Parameter lift_mk_interact : forall (a b c:t),(mk_interact a b = Some c)->forall (d k:nat),(exists c':t,(equal c' (deBruijn.lift deBruijn_base d k c))/\(mk_interact (deBruijn.lift deBruijn_base d k a) (deBruijn.lift deBruijn_base d k b) = Some c')).

  Parameter lift_mk_input_inv : forall (d k:nat)(e:A.P.E.t)(inp':t),(mk_input (deBruijn.lift A.P.E.deBruijn_base d k e) = Some inp')->(exists inp:t,mk_input e = Some inp).
  Parameter lift_mk_output_inv : forall (d k:nat)(e f:A.P.E.t)(out':t)(m':A.P.E.t),(mk_output (deBruijn.lift A.P.E.deBruijn_base d k e) (deBruijn.lift A.P.E.deBruijn_base d k f) = Some (out',m'))->(exists out:t,exists m:A.P.E.t,mk_output e f = Some (out,m) /\ m' = deBruijn.lift A.P.E.deBruijn_base d k m).
  Parameter lift_mk_if_inv : forall (d k:nat)(phi:A.P.F.t)(mu mu'':t),(mk_if (deBruijn.lift A.P.F.deBruijn_base d k phi) (deBruijn.lift deBruijn_base d k mu) = Some mu'')->(exists mu':t,(mk_if phi mu = Some mu')).
  Parameter lift_mk_res_inv : forall (d k:nat)(mu mu'':t),(mk_res (deBruijn.lift deBruijn_base (S d) k mu) = Some mu'')->(exists mu':t,mk_res mu = Some mu').
  Parameter lift_mk_interact_inv : forall (d k:nat)(a b c':t),(mk_interact (deBruijn.lift deBruijn_base d k a) (deBruijn.lift deBruijn_base d k b) = Some c')->(exists c:t,mk_interact a b = Some c).

  (* swapR *)
  Parameter swapR_mk_silent : forall (d k:nat),(equal mk_silent (deBruijn.swapR deBruijn_base d k mk_silent)).
  Parameter swapR_mk_input : forall (e:A.P.E.t)(inp:t),(mk_input e = Some inp)->forall (d k:nat),(exists inp':t,mk_input (deBruijn.swapR A.P.E.deBruijn_base d k e) = Some inp' /\ (equal inp' (deBruijn.swapR deBruijn_base d k inp))).
  Parameter swapR_mk_output : forall (e f:A.P.E.t)(out:t)(m:A.P.E.t),(mk_output e f = Some (out,m))->forall (d k:nat),(exists out':t,mk_output (deBruijn.swapR A.P.E.deBruijn_base d k e) (deBruijn.swapR A.P.E.deBruijn_base d k f) = Some (out',deBruijn.swapR A.P.E.deBruijn_base d k m) /\ (equal out' (deBruijn.swapR deBruijn_base d k out))).
  Parameter swapR_mk_if : forall (phi:A.P.F.t)(mu mu':t),(mk_if phi mu = Some mu')->forall (d k:nat),(exists mu'':t,(equal mu'' (deBruijn.swapR deBruijn_base d k mu')) /\ mk_if (deBruijn.swapR A.P.F.deBruijn_base d k phi) (deBruijn.swapR deBruijn_base d k mu) = Some mu'').
  Parameter swapR_mk_res : forall (mu mu':t),(mk_res mu = Some mu')->forall (d k:nat),(exists mu'':t,(equal mu'' (deBruijn.swapR deBruijn_base d k mu'))/\(mk_res (deBruijn.swapR deBruijn_base (S d) k mu))=(Some mu'')).
  Parameter swapR_mk_interact : forall (a b c:t),(mk_interact a b = Some c)->forall (d k:nat),(exists c':t,(equal c' (deBruijn.swapR deBruijn_base d k c))/\(mk_interact (deBruijn.swapR deBruijn_base d k a) (deBruijn.swapR deBruijn_base d k b) = Some c')).
End ActionType.


