(* A formalisation of the spi calculus in Coq *)
(* Copyright (C) 2005-2007 Sebastien Briais *)
(* http://lamp.epfl.ch/~sbriais/ *)

(* This library is free software; you can redistribute it and/or modify *)
(* it under the terms of the GNU Lesser General Public License as *)
(* published by the Free Software Foundation; either version 2.1 of the *)
(* License, or (at your option) any later version. *)

(* This library is distributed in the hope that it will be useful, but *)
(* WITHOUT ANY WARRANTY; without even the implied warranty of *)
(* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU *)
(* Lesser General Public License for more details. *)

(* You should have received a copy of the GNU Lesser General Public *)
(* License along with this library; if not, write to the Free Software *)
(* Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA *)
(* 02110-1301 USA *)

Require Export processType.
Require Import deBruijnType.
Require Import tactics.
Require Import Arith.
Require Import missing.
Require Export agentType.

Module Agent(Pr:ProcessType) <: AgentType with Module P := Pr.
  Module P := Pr.

  Inductive abstraction : Set := 
    | Abs : P.t -> abstraction.
  
  Inductive concretion : Set :=
    | Conc : nat -> P.E.t -> P.t -> concretion.
  
  Inductive t : Set :=
    | Pr : P.t -> t
    | F : abstraction -> t
    | C : concretion -> t.

  Module Ord <: OrderedType with Definition t := t with Definition eq := @eq t.
    Definition t := t.
    Definition eq := @eq t.
    
    Definition lt (a b:t) : Prop :=
      match (a,b) with
	| (Pr p, Pr q) => P.Ord.lt p q
	| (Pr _,_) => True
	| (_,Pr _) => False
	| (F (Abs p),F (Abs q)) => P.Ord.lt p q
	| (F _,_) => True
	| (_,F _) => False
	| (C (Conc n e p),C (Conc m f q)) => n<m \/ (n=m /\ (P.E.Ord.lt e f)) \/ (n = m /\ e = f /\ (P.Ord.lt p q))
      end.
    
    Lemma eq_refl : forall (x:t),(eq x x).
      unfold eq;trivial.
    Qed.
    
    Lemma eq_sym : forall (x y:t),(eq x y)->(eq y x).
      unfold eq;auto.
    Qed.
    
    Lemma eq_trans : forall (x y z:t),(eq x y)->(eq y z)->(eq x z).
      unfold eq;intros;congruence.
    Qed.
    
    Lemma lt_not_eq : forall (x y:t),(lt x y)->~(eq x y).
      unfold eq.
      destruct x;destruct y;intro;red;intro;try (discriminate H0).
      unfold lt in H.
      injection H0;intros.
      elim (P.Ord.lt_not_eq H);auto.
      destruct a;destruct a0.
      unfold lt in H.
      injection H0;intros.
      elim (P.Ord.lt_not_eq H);auto.
      destruct c;destruct c0.
      unfold lt in H.
      injection H0;intros.
      case H;intro.
      try_absurd_arith.
      case H4;intro;and_elim.
      elim (P.E.Ord.lt_not_eq H6);auto.
      elim (P.Ord.lt_not_eq H7);auto.
    Qed.
    
    Lemma lt_eq_trans : forall (x y z:t),(lt x y)->(eq y z)->(lt x z).
      intros.
      rewrite H0 in H;auto.
    Qed.
    
    Lemma eq_lt_trans : forall (x y z:t),(eq x y)->(lt y z)->(lt x z).
      intros.
      rewrite H;auto.
    Qed.
    
    Lemma lt_trans : forall (x y z:t),(lt x y)->(lt y z)->(lt x z).
      destruct x;destruct y;destruct z;unfold lt;intros;eauto;try (destruct a;tauto);try (destruct c;tauto);try (destruct a0;tauto);try (destruct c0;tauto).
      destruct a;destruct a0;destruct a1;eauto.
      destruct c;destruct c0;destruct c1.
      case H;case H0;intros;and_elim.
      left;omega.
      case H1;intros;and_elim.
      left;omega.
      left;omega.
      case H2;intros;and_elim.
      left;omega.
      left;omega.
      case H1;case H2;intros;and_elim;right.
      left;split;eauto.
      congruence.
      rewrite H6.
      left;split;eauto.
      congruence.
      rewrite <- H5.
      left;split;eauto.
      congruence.
      right.
      repeat split;eauto;congruence.
    Qed.
    
    Definition compare : forall (x y:t),(Compare lt eq x y).
      unfold eq.
      destruct x;destruct y;try (destruct a);try (destruct a0);try (destruct c);try (destruct c0);try (apply LT;unfold lt;trivial;fail);try (apply GT;unfold lt;trivial;fail).
      case (P.Ord.compare t0 t1);intro;[apply LT | apply EQ | apply GT];unfold lt;auto.
      rewrite e;auto.
      case (P.Ord.compare t0 t1);intro;[apply LT | apply EQ | apply GT];unfold lt;auto.
      rewrite e;auto.
      case (lt_eq_lt_dec n n0);intro;try (case s;clear s;intro);[apply LT;unfold lt;auto | idtac | apply GT;unfold lt;auto].
      subst n0.
      case (P.E.Ord.compare t0 t2);intro;[apply LT;unfold lt;right;left;split;auto | idtac | apply GT;unfold lt;right;left;auto].
      rewrite e;clear e.
      case (P.Ord.compare t1 t3);intro;[apply LT;unfold lt | apply EQ;rewrite e;auto | apply GT;unfold lt];right;right;repeat split;auto.
    Defined.
    
    Hint Immediate eq_sym.
    Hint Resolve eq_refl eq_trans lt_not_eq lt_trans.
  End Ord.
  
  Definition t_map (fe:nat->P.E.t->P.E.t) (fp:nat->P.t->P.t) (d:nat) (a:t) : t :=
    match a with
      | Pr p => Pr (fp d p)
      | F (Abs p) => F (Abs (fp (S d) p))
      | C (Conc n e p) => C (Conc n (fe (d+n) e) (fp (d+n) p))
    end.
  
  Lemma t_map_struct : deBruijn.map2_struct P.E.t P.t t t_map.
    split;intros.
    destruct x;simpl;auto.
    destruct a;auto.
    destruct c;auto.
    unfold injective;intros.
    destruct x;destruct y;try (destruct a);try (destruct a0);try (destruct c);try (destruct c0);simpl in H1;try (discriminate H1).
    injection H1;intros.
    rewrite (H0 d t0 t1);auto.
    injection H1;intros.
    rewrite (H0 (S d) t0 t1);auto.
    injection H1;intros.
    subst n0.
    rewrite (H0 (d+n) t1 t3);auto.
    rewrite (H (d+n) t0 t2);auto.
    destruct x;simpl;auto.
    rewrite H0;auto.
    destruct a;simpl;auto.
    repeat (rewrite plus_n_Sm).
    rewrite H0;auto.
    destruct c;simpl;auto.
    repeat (rewrite <- plus_assoc).
    rewrite H0.
    rewrite H;auto.
    destruct x;simpl;auto.
    rewrite H0;auto.
    destruct a.
    repeat (rewrite plus_n_Sm).
    rewrite H0;auto.
    destruct c.
    repeat (rewrite <- plus_assoc).
    rewrite H0.
    rewrite H;auto.
  Qed. 
  
  Definition deBruijn_base := deBruijn.deBruijn_map2_struct P.E.t P.E.deBruijn_base P.t P.deBruijn_base t t_map t_map_struct.

  Definition deBruijn_subst := deBruijn.deBruijn_s_map2_struct P.E.t P.E.deBruijn_base P.t P.deBruijn_base P.E.t P.E.deBruijn_base P.E.deBruijn_subst P.deBruijn_subst t t_map t_map_struct.

  Lemma low_subst_Pr : forall (d k:nat)(es:list P.E.t)(p:P.t),deBruijn.low_subst deBruijn_subst d k es (Pr p) = Pr (deBruijn.low_subst P.deBruijn_subst d k es p).
    intros;reflexivity.
  Qed.

  Lemma low_subst_F : forall (d k:nat)(es:list P.E.t)(p:P.t),deBruijn.low_subst deBruijn_subst d k es (F (Abs p)) = F (Abs (deBruijn.low_subst P.deBruijn_subst (S d) k es p)).
    intros;reflexivity.
  Qed.

  Lemma low_subst_C : forall (d k:nat)(es:list P.E.t)(n:nat)(m:P.E.t)(p:P.t),deBruijn.low_subst deBruijn_subst d k es (C (Conc n m p)) = C (Conc n (deBruijn.low_subst P.E.deBruijn_subst (d+n) k es m) (deBruijn.low_subst P.deBruijn_subst (d+n) k es p)).
    intros.
    reflexivity.
  Qed.

(*   Definition t_fold (A:Set)(fe:nat->P.E.t->A->A)(fp:nat->P.t->A->A)(d:nat)(x:t)(a:A) : A := *)
(*     match x with *)
(*       | Pr p => fp d p a *)
(*       | F (Abs p) => fp (S d) p a *)
(*       | C (Conc n m p) => fp (d+n) p (fe (d+n) m a) *)
(*     end. *)

  Definition new_abs (f:abstraction) : abstraction :=
    match f with
      | (Abs p) => Abs (P.New (deBruijn.swapR P.deBruijn_base 0 1 p))
    end.

  Definition new_conc (c:concretion) : concretion :=
    match c with
      | (Conc n e p) => 
	if deBruijn.mem_fn P.E.deBruijn_fn_base 0 n e 
	  then (Conc (S n) e p)
	  else (Conc n (deBruijn.low1 P.E.deBruijn_base n e) (P.New (deBruijn.swapR P.deBruijn_base 0 n p)))
    end.

  Definition new (a:t) : t :=
    match a with
      | Pr p => Pr (P.New p)
      | F f => F (new_abs f)
      | C c => C (new_conc c) 
    end.

  Lemma new_abs_def : forall (p:P.t),(new_abs (Abs p))=(Abs (P.New (deBruijn.swapR P.deBruijn_base 0 1 p))).
    intros;reflexivity.
  Qed.

  Lemma new_conc_def : forall (n:nat)(m:P.E.t)(p:P.t),(new_conc (Conc n m p))=(if deBruijn.mem_fn P.E.deBruijn_fn_base 0 n m then (Conc (S n) m p) else (Conc n (deBruijn.low1 P.E.deBruijn_base n m)  (P.New (deBruijn.swapR P.deBruijn_base 0 n p)))).
    intros;reflexivity.
  Qed.
    
  Lemma new_Pr : forall (p:P.t),(new (Pr p))=(Pr (P.New p)).
    intros;reflexivity.
  Qed.

  Lemma new_F : forall (f:abstraction),(new (F f))=(F (new_abs f)).
    intros;reflexivity.
  Qed.

  Lemma new_C : forall (c:concretion),(new (C c))=(C (new_conc c)).
    intros;reflexivity.
  Qed.

  Definition abs_par_l (p:P.t)(f:abstraction) : abstraction :=
    match f with
      | (Abs q) => (Abs (P.Par (deBruijn.lift P.deBruijn_base 0 1 p) q))
    end.

  Definition conc_par_l (p:P.t)(c:concretion) : concretion :=
    match c with
      | (Conc n e q) => (Conc n e (P.Par (deBruijn.lift P.deBruijn_base 0 n p) q))
    end.

  Definition par_l (p:P.t) (a:t) : t :=
    match a with
      | Pr q => Pr (P.Par p q)
      | F f => F (abs_par_l p f)
      | C c => C (conc_par_l p c)
    end.

  Lemma abs_par_l_def : forall (p:P.t)(q:P.t),(abs_par_l p (Abs q))=(Abs (P.Par (deBruijn.lift P.deBruijn_base 0 1 p) q)).
    intros;reflexivity.
  Qed.

  Lemma conc_par_l_def : forall (p:P.t)(n:nat)(m:P.E.t)(q:P.t),(conc_par_l p (Conc n m q))=(Conc n m (P.Par (deBruijn.lift P.deBruijn_base 0 n p) q)).
    intros;reflexivity.
  Qed.

  Lemma par_l_Pr : forall (p q:P.t),(par_l p (Pr q))=(Pr (P.Par p q)).
    intros;reflexivity.
  Qed.

  Lemma par_l_F : forall (p:P.t)(f:abstraction),(par_l p (F f))=(F (abs_par_l p f)).
    intros;reflexivity.
  Qed.

  Lemma par_l_C : forall (p:P.t)(c:concretion),(par_l p (C c))=(C (conc_par_l p c)).
    intros;reflexivity.
  Qed.

  Definition abs_par_r (f:abstraction)(q:P.t) : abstraction :=
    match f with
      | Abs p => (Abs (P.Par p (deBruijn.lift P.deBruijn_base 0 1 q)))
    end.

  Definition conc_par_r (c:concretion)(q:P.t) : concretion :=
    match c with
      | Conc n e p => (Conc n e (P.Par p (deBruijn.lift P.deBruijn_base 0 n q)))
    end.

  Definition par_r (a:t) (q:P.t) : t :=
    match a with
      | Pr p => Pr (P.Par p q)
      | F f => F (abs_par_r f q)
      | C c => C (conc_par_r c q)
    end.

  Lemma abs_par_r_def : forall (p:P.t)(q:P.t),(abs_par_r (Abs p) q)=(Abs (P.Par p (deBruijn.lift P.deBruijn_base 0 1 q))).
    intros;reflexivity.
  Qed.

  Lemma conc_par_r_def : forall (n:nat)(m:P.E.t)(p:P.t)(q:P.t),(conc_par_r (Conc n m p) q)=(Conc n m (P.Par p (deBruijn.lift P.deBruijn_base 0 n q))).
    intros;reflexivity.
  Qed.

  Lemma par_r_Pr : forall (p q:P.t),(par_r (Pr p) q)=(Pr (P.Par p q)).
    intros;reflexivity.
  Qed.

  Lemma par_r_F : forall (f:abstraction)(q:P.t),(par_r (F f) q)=(F (abs_par_r f q)).
    intros;reflexivity.
  Qed.

  Lemma par_r_C : forall (c:concretion)(q:P.t),(par_r (C c) q)=(C (conc_par_r c q)).
    intros;reflexivity.
  Qed.

  Definition pseudo_appl_l (f:abstraction) (c:concretion) : t :=
    match (f,c) with
      | (Abs p,Conc n e q) => Pr (P.mk_nu n (P.Par (deBruijn.low_subst P.deBruijn_subst 0 n (cons e nil) p) q))
    end.

  Definition pseudo_appl_r (c:concretion) (f:abstraction) : t :=
    match (c,f) with
      | (Conc n e p,Abs q) => Pr (P.mk_nu n (P.Par p (deBruijn.low_subst P.deBruijn_subst 0 n (cons e nil) q)))
    end.

  Lemma pseudo_appl_l_def : forall (p:P.t)(n:nat)(m:P.E.t)(q:P.t),(pseudo_appl_l (Abs p) (Conc n m q))=(Pr (P.mk_nu n (P.Par (deBruijn.low_subst P.deBruijn_subst 0 n (cons m nil) p) q))).
    intros;reflexivity.
  Qed.
  
  Lemma pseudo_appl_r_def : forall (n:nat)(m:P.E.t)(p:P.t)(q:P.t),(pseudo_appl_r (Conc n m p) (Abs q))=(Pr (P.mk_nu n (P.Par p (deBruijn.low_subst P.deBruijn_subst 0 n (cons m nil) q)))).
    intros;reflexivity.
  Qed.

  Lemma lift_Pr : forall (d k:nat)(p:P.t),(deBruijn.lift deBruijn_base d k (Pr p))=(Pr (deBruijn.lift P.deBruijn_base d k p)).
    intros;reflexivity.
  Qed.

  Lemma lift_F : forall (d k:nat)(p:P.t),(deBruijn.lift deBruijn_base d k (F (Abs p)))=(F (Abs (deBruijn.lift P.deBruijn_base (S d) k p))).
    intros;reflexivity.
  Qed.

  Lemma lift_C : forall (d k:nat)(n:nat)(e:P.E.t)(p:P.t),(deBruijn.lift deBruijn_base d k (C (Conc n e p)))=(C (Conc n (deBruijn.lift P.E.deBruijn_base (d+n) k e) (deBruijn.lift P.deBruijn_base (d+n) k p))).
    intros;reflexivity.
  Qed.

  Lemma lift_new : forall (d k:nat)(a:t),(deBruijn.lift deBruijn_base d k (new a))=(new (deBruijn.lift deBruijn_base (S d) k a)).
    intros.
    destruct a.
    rewrite lift_Pr.
    simpl.
    rewrite P.lift_New;auto.
    destruct a.
    rewrite lift_F.
    simpl.
    rewrite P.lift_New.
    generalize (deBruijn.lift_swapR_comm P.deBruijn_base 0 d 1 k t0);simpl;intros.
    rewrite H.
    auto.
    destruct c.
    rewrite lift_C.
    unfold new.
    generalize (deBruijn.lift_In_fn P.E.deBruijn_fn_base d 0 (le_O_n d) n k t0);intro.
    simpl.
    case (Bool_elim_bool (deBruijn.mem_fn P.E.deBruijn_fn_base 0 n t0));intro.
    rewrite H0.
    rewrite H0 in H.
    replace (S (d+n)) with (n+S d);try omega.
    rewrite <- H.
    replace (n+S d) with (d+S n);try omega;auto.
    rewrite H0.
    rewrite H0 in H.
    replace (S (d+n)) with (n+S d);try omega.
    rewrite <- H.
    rewrite P.lift_New.
    generalize (deBruijn.lift_swapR_comm P.deBruijn_base 0 d n k t1);simpl;intro.
    replace (S (d+n)) with (n+S d);try omega.
    rewrite H1.
    rewrite (deBruijn.lift_low1_comm P.E.deBruijn_fn_base n d k t0 H0).
    rewrite plus_comm;trivial.
  Qed.

  Lemma lift_par_l : forall (p:P.t)(a:t)(d k:nat),(deBruijn.lift deBruijn_base d k (par_l p a))=(par_l (deBruijn.lift P.deBruijn_base d k p) (deBruijn.lift deBruijn_base d k a)).
    intros.
    destruct a.
    simpl.
    rewrite P.lift_Par.
    auto.
    destruct a.
    simpl.
    rewrite P.lift_Par.
    generalize (deBruijn.lift_comm P.deBruijn_base 0 d 1 k p);simpl.
    rewrite plus_comm;simpl;intros.
    rewrite H.
    auto.
    destruct c.
    simpl.
    rewrite P.lift_Par.
    generalize (deBruijn.lift_comm P.deBruijn_base 0 d n k p);simpl;intro.
    rewrite H.
    auto.
  Qed.
  
  Lemma lift_par_r : forall (a:t)(q:P.t)(d k:nat),(deBruijn.lift deBruijn_base d k (par_r a q))=(par_r (deBruijn.lift deBruijn_base d k a) (deBruijn.lift P.deBruijn_base d k q)).
    intros.
    destruct a.
    simpl.
    rewrite P.lift_Par.
    auto.
    destruct a.
    simpl.
    rewrite P.lift_Par.
    generalize (deBruijn.lift_comm P.deBruijn_base 0 d 1 k q);simpl.
    rewrite plus_comm;simpl;intros.
    rewrite H.
    auto.
    destruct c.
    simpl.
    rewrite P.lift_Par.
    generalize (deBruijn.lift_comm P.deBruijn_base 0 d n k q);simpl;intro.
    rewrite H.
    auto.
  Qed.

  Lemma lift_pseudo_appl_l : forall (p:P.t)(n:nat)(m:P.E.t)(q:P.t)(d k:nat),(deBruijn.lift deBruijn_base d k (pseudo_appl_l (Abs p) (Conc n m q)))=(pseudo_appl_l (Abs (deBruijn.lift P.deBruijn_base (S d) k p)) (Conc n (deBruijn.lift P.E.deBruijn_base (d+n) k m) (deBruijn.lift P.deBruijn_base (d+n) k q))).
    intros.
    unfold pseudo_appl_l.
    rewrite lift_Pr.
    rewrite P.lift_mk_nu.
    rewrite P.lift_Par.
    generalize (deBruijn.lift_low_subst_comm P.deBruijn_subst d 0 n k);rewrite plus_0_r;intro.
    rewrite H.
    simpl.
    replace (d+1) with (S d);try omega.
    auto.
  Qed.

  Lemma lift_pseudo_appl_r : forall (n:nat)(m:P.E.t)(p q:P.t)(d k:nat),(deBruijn.lift deBruijn_base d k (pseudo_appl_r (Conc n m p) (Abs q)))=(pseudo_appl_r (Conc n (deBruijn.lift P.E.deBruijn_base (d+n) k m) (deBruijn.lift P.deBruijn_base (d+n) k p)) (Abs (deBruijn.lift P.deBruijn_base (S d) k q))).
    intros.
    unfold pseudo_appl_r.
    rewrite lift_Pr.
    rewrite P.lift_mk_nu.
    rewrite P.lift_Par.
    generalize (deBruijn.lift_low_subst_comm P.deBruijn_subst d 0 n k);rewrite plus_0_r;intro.
    rewrite H.
    simpl.
    replace (d+1) with (S d);try omega.
    auto.
  Qed.

  Lemma swapR_Pr : forall (d k:nat)(p:P.t),(deBruijn.swapR deBruijn_base d k (Pr p))=(Pr (deBruijn.swapR P.deBruijn_base d k p)).
    intros;reflexivity.
  Qed.

  Lemma swapR_F : forall (d k:nat)(p:P.t),(deBruijn.swapR deBruijn_base d k (F (Abs p)))=(F (Abs (deBruijn.swapR P.deBruijn_base (S d) k p))).
    intros;reflexivity.
  Qed.

  Lemma swapR_C : forall (d k:nat)(n:nat)(e:P.E.t)(p:P.t),(deBruijn.swapR deBruijn_base d k (C (Conc n e p)))=(C (Conc n (deBruijn.swapR P.E.deBruijn_base (d+n) k e) (deBruijn.swapR P.deBruijn_base (d+n) k p))).
    intros;reflexivity.
  Qed.

  Lemma swapR_new : forall (d k:nat)(a:t),(deBruijn.swapR deBruijn_base d k (new a))=(new (deBruijn.swapR deBruijn_base (S d) k a)).
    intros.
    destruct a.
    rewrite swapR_Pr.
    simpl.
    rewrite P.swapR_New;auto.
    destruct a.
    rewrite swapR_F.
    simpl.
    rewrite P.swapR_New.
    generalize (deBruijn.swapR_comm P.deBruijn_base 0 d 1 k t0);simpl;intros.
    rewrite H.
    auto.
    destruct c.
    rewrite swapR_C.
    unfold new.
    generalize (deBruijn.swapR_In_fn P.E.deBruijn_fn_base d 0 (le_O_n d) n k t0);intro.
    simpl.
    case (Bool_elim_bool (deBruijn.mem_fn P.E.deBruijn_fn_base 0 n t0));intro.
    rewrite H0.
    rewrite H0 in H.
    replace (S (d+n)) with (n+S d);try omega.
    rewrite <- H.
    replace (n+S d) with (d+S n);try omega;auto.
    rewrite H0.
    rewrite H0 in H.
    replace (S (d+n)) with (n+S d);try omega.
    rewrite <- H.
    rewrite P.swapR_New.
    generalize (deBruijn.swapR_comm P.deBruijn_base 0 d n k t1);simpl;intro.
    replace (S (d+n)) with (n+S d);try omega.
    rewrite H1.
    rewrite (deBruijn.swapR_low1_comm P.E.deBruijn_fn_base n d k t0 H0).
    rewrite plus_comm;trivial.
  Qed.

  Lemma swapR_par_l : forall (p:P.t)(a:t)(d k:nat),(deBruijn.swapR deBruijn_base d k (par_l p a))=(par_l (deBruijn.swapR P.deBruijn_base d k p) (deBruijn.swapR deBruijn_base d k a)).
    intros.
    destruct a.
    simpl.
    rewrite P.swapR_Par.
    auto.
    destruct a.
    simpl.
    rewrite P.swapR_Par.
    generalize (deBruijn.swapR_lift_comm P.deBruijn_base 0 d 1 k);simpl;intros.
    rewrite H.
    auto.
    destruct c.
    simpl.
    rewrite P.swapR_Par.
    generalize (deBruijn.swapR_lift_comm P.deBruijn_base 0 d n k p);simpl;intro.
    rewrite plus_comm in H.
    rewrite H.
    auto.
  Qed.
  
  Lemma swapR_par_r : forall (a:t)(q:P.t)(d k:nat),(deBruijn.swapR deBruijn_base d k (par_r a q))=(par_r (deBruijn.swapR deBruijn_base d k a) (deBruijn.swapR P.deBruijn_base d k q)).
    intros.
    destruct a.
    simpl.
    rewrite P.swapR_Par.
    auto.
    destruct a.
    simpl.
    rewrite P.swapR_Par.
    generalize (deBruijn.swapR_lift_comm P.deBruijn_base 0 d 1 k);simpl;intros.
    rewrite H.
    auto.
    destruct c.
    simpl.
    rewrite P.swapR_Par.
    generalize (deBruijn.swapR_lift_comm P.deBruijn_base 0 d n k q);simpl;intro.
    rewrite plus_comm in H.
    rewrite H.
    auto.
  Qed.

  Lemma swapR_pseudo_appl_l : forall (p:P.t)(n:nat)(m:P.E.t)(q:P.t)(d k:nat),(deBruijn.swapR deBruijn_base d k (pseudo_appl_l (Abs p) (Conc n m q)))=(pseudo_appl_l (Abs (deBruijn.swapR P.deBruijn_base (S d) k p)) (Conc n (deBruijn.swapR P.E.deBruijn_base (d+n) k m) (deBruijn.swapR P.deBruijn_base (d+n) k q))).
    intros.
    unfold pseudo_appl_l.
    rewrite swapR_Pr.
    rewrite P.swapR_mk_nu.
    rewrite P.swapR_Par.
    generalize (deBruijn.swapR_low_subst_comm P.deBruijn_subst d 0 n k);rewrite plus_0_r;intro.
    rewrite H.
    simpl.
    replace (d+1) with (S d);try omega.
    auto.
  Qed.

  Lemma swapR_pseudo_appl_r : forall (n:nat)(m:P.E.t)(p q:P.t)(d k:nat),(deBruijn.swapR deBruijn_base d k (pseudo_appl_r (Conc n m p) (Abs q)))=(pseudo_appl_r (Conc n (deBruijn.swapR P.E.deBruijn_base (d+n) k m) (deBruijn.swapR P.deBruijn_base (d+n) k p)) (Abs (deBruijn.swapR P.deBruijn_base (S d) k q))).
    intros.
    unfold pseudo_appl_r.
    rewrite swapR_Pr.
    rewrite P.swapR_mk_nu.
    rewrite P.swapR_Par.
    generalize (deBruijn.swapR_low_subst_comm P.deBruijn_subst d 0 n k);rewrite plus_0_r;intro.
    rewrite H.
    simpl.
    replace (d+1) with (S d);try omega.
    auto.
  Qed.

  Lemma low_subst_new : forall (d k:nat)(es:list P.E.t)(a:t),(deBruijn.low_subst deBruijn_subst d k es (new a))=(new (deBruijn.low_subst deBruijn_subst (S d) k es a)).
    destruct a;intros.
    simpl.
    rewrite P.low_subst_New;auto.
    simpl.
    destruct a.
    simpl.
    rewrite P.low_subst_New.
    generalize (deBruijn.low_subst_swapR_comm P.deBruijn_subst 0 d 1 k es t0).
    replace (0+1+S d) with (S (S d));try reflexivity.
    intros.
    rewrite H;auto.
    destruct c;simpl.
    generalize (deBruijn.low_subst_fn_not_supp P.E.deBruijn_fn_subst 0 n d k es t0 (le_O_n d));intro.
    replace (S (d+n)) with (n+S d);try omega.
    rewrite <- H.
    case (Bool_elim_bool (deBruijn.mem_fn P.E.deBruijn_fn_base 0 n t0));intro H0;rewrite H0.
    replace (n+S d) with (d+S n);try omega;auto.
    rewrite P.low_subst_New.
    generalize (deBruijn.low_subst_swapR_comm P.deBruijn_subst 0 d n k es).
    simpl.
    intro.
    rewrite <- H1.
    rewrite <- (deBruijn.low_subst_low1_comm P.E.deBruijn_fn_subst n d k);auto.
    rewrite plus_comm;auto.
    rewrite plus_n_Sm.
    auto.
  Qed.

  Lemma low_subst_par_l : forall (d k:nat)(es:list P.E.t)(p:P.t)(a:t),(deBruijn.low_subst deBruijn_subst d k es (par_l p a))=(par_l (deBruijn.low_subst P.deBruijn_subst d k es p) (deBruijn.low_subst deBruijn_subst d k es a)).
    destruct a;intros.
    simpl.
    rewrite P.low_subst_Par.
    auto.
    destruct a as [q];simpl.
    rewrite P.low_subst_Par.
    generalize (deBruijn.low_subst_lift_comm P.deBruijn_subst 0 d 1 k);simpl;intro.
    rewrite <- H.
    rewrite plus_comm;reflexivity.
    destruct c as [n m q];simpl.
    rewrite P.low_subst_Par.
    generalize (deBruijn.low_subst_lift_comm P.deBruijn_subst 0 d n k);simpl;intro.
    rewrite H;auto.
  Qed.

  Lemma low_subst_par_r : forall (d k:nat)(es:list P.E.t)(a:t)(p:P.t),(deBruijn.low_subst deBruijn_subst d k es (par_r a p))=(par_r (deBruijn.low_subst deBruijn_subst d k es a) (deBruijn.low_subst P.deBruijn_subst d k es p)).
    destruct a;intros.
    simpl.
    rewrite P.low_subst_Par.
    auto.
    destruct a as [q];simpl.
    rewrite P.low_subst_Par.
    generalize (deBruijn.low_subst_lift_comm P.deBruijn_subst 0 d 1 k);simpl;intro.
    rewrite <- H.
    rewrite plus_comm;reflexivity.
    destruct c as [n m q];simpl.
    rewrite P.low_subst_Par.
    generalize (deBruijn.low_subst_lift_comm P.deBruijn_subst 0 d n k);simpl;intro.
    rewrite H;auto.
  Qed.

  Lemma low_subst_pseudo_appl_l : forall (d k:nat)(es:list P.E.t)(p:P.t)(n:nat)(m:P.E.t)(q:P.t),deBruijn.low_subst deBruijn_subst d k es (pseudo_appl_l (Abs p) (Conc n m q)) = pseudo_appl_l (Abs (deBruijn.low_subst P.deBruijn_subst (S d) k es p)) (Conc n (deBruijn.low_subst P.E.deBruijn_subst (d+n) k es m) (deBruijn.low_subst P.deBruijn_subst (d+n) k es q)).
    intros.
    unfold pseudo_appl_l.
    repeat (rewrite low_subst_Pr).
    rewrite P.low_subst_mk_nu.
    rewrite P.low_subst_Par.
    generalize (P.low_subst_lemma p 0 d n k es (cons m nil)).
    simpl.
    intros.
    rewrite H.
    rewrite (plus_comm d 1).
    simpl.
    auto.
  Qed.

  Lemma low_subst_pseudo_appl_r : forall (d k:nat)(es:list P.E.t)(n:nat)(m:P.E.t)(p:P.t)(q:P.t),deBruijn.low_subst deBruijn_subst d k es (pseudo_appl_r (Conc n m p) (Abs q)) = pseudo_appl_r (Conc n (deBruijn.low_subst P.E.deBruijn_subst (d+n) k es m) (deBruijn.low_subst P.deBruijn_subst (d+n) k es p)) (Abs (deBruijn.low_subst P.deBruijn_subst (S d) k es q)).
    intros.
    unfold pseudo_appl_r.
    repeat (rewrite low_subst_Pr).
    rewrite P.low_subst_mk_nu.
    rewrite P.low_subst_Par.
    generalize (P.low_subst_lemma q 0 d n k es (cons m nil)).
    simpl.
    intros.
    rewrite H.
    rewrite (plus_comm d 1).
    simpl.
    auto.
  Qed.

  Inductive well_formed : t -> Prop :=
    | WFProc : forall (p:P.t),(well_formed (Pr p))
    | WFAbs : forall (p:P.t),(well_formed (F (Abs p)))
    | WFConc : forall (n:nat)(m:P.E.t)(p:P.t),(forall (i:nat),i < n->(deBruijn.mem_fn P.E.deBruijn_fn_base i 0 m = true))->(well_formed (C (Conc n m p))).

  Lemma well_formed_new : forall (a:t),(well_formed a)->(well_formed (new a)).
    induction 1;simpl.
    apply WFProc.
    apply WFAbs.
    generalize (deBruijn.fn_plus_lift P.E.deBruijn_fn_base 0 n m 0);simpl.
    unfold DeBruijnNat.lift;simpl;intro.
    rewrite <- H0.
    case (Bool_elim_bool (deBruijn.mem_fn P.E.deBruijn_fn_base n 0 m));intro.
    rewrite H1.
    apply WFConc.
    intros.
    inversion H2.
    auto.
    assert (i < n);try omega.
    auto.
    rewrite H1.
    apply WFConc.
    intros.
    destruct n.
    inversion H2.
    assert (i <= n);try omega.
    generalize (deBruijn.low1_In_fn P.E.deBruijn_fn_base n i H3 0 m);simpl.
    rewrite <- H0.
    intro.
    generalize (H4 H1);clear H4;intro.
    rewrite <- H4.
    apply H.
    auto.
  Qed.

  Lemma well_formed_par_r : forall (a:t)(q:P.t),(well_formed a)->(well_formed (par_r a q)).
    induction 1;simpl.
    apply WFProc.
    apply WFAbs.
    apply WFConc.
    auto.
  Qed.

  Lemma well_formed_par_l : forall (p:P.t)(a:t),(well_formed a)->(well_formed (par_l p a)).
    induction 1;simpl.
    apply WFProc.
    apply WFAbs.
    apply WFConc.
    auto.
  Qed.

  Lemma well_formed_pseudo_appl_l : forall (f:abstraction)(c:concretion),(well_formed (pseudo_appl_l f c)).
    destruct f;destruct c.
    unfold pseudo_appl_l.
    apply WFProc.
  Qed.
  
  Lemma well_formed_pseudo_appl_r : forall (f:abstraction)(c:concretion),(well_formed (pseudo_appl_r c f)).
    destruct f;destruct c.
    unfold pseudo_appl_r.
    apply WFProc.
  Qed.

  Lemma well_formed_concretion_inv : forall (n:nat)(m:P.E.t)(p:P.t),(well_formed (C (Conc (S n) m p)))->(well_formed (C (Conc n m p)))/\(C (Conc (S n) m p))=new (C (Conc n m p)).
    intros.
    inversion H.
    subst n0;subst m0;subst p0.
    split.
    apply WFConc.
    auto.
    simpl.
    generalize (deBruijn.fn_plus_lift P.E.deBruijn_fn_base 0 n m 0);simpl.
    unfold DeBruijnNat.lift;simpl;intro.
    rewrite <- H0.
    rewrite (H1 n);auto.
  Qed.
  
  Lemma well_formed_lift : forall (a:t),(well_formed a)->forall (d k:nat),(well_formed (deBruijn.lift deBruijn_base d k a)).
    induction 1;intros.
    rewrite lift_Pr.
    apply WFProc.
    rewrite lift_F.
    apply WFAbs.
    rewrite lift_C.
    apply WFConc.
    intros.
    destruct n.
    inversion H0.
    assert (i <= d+n);try omega.
    generalize (deBruijn.lift_In_fn P.E.deBruijn_fn_base (d+n) i H1 0 k m).
    rewrite H;auto.
    rewrite plus_n_Sm.
    simpl.
    intro.
    congruence.
  Qed.

  Lemma well_formed_swapR : forall (a:t),(well_formed a)->forall (d k:nat),(well_formed (deBruijn.swapR deBruijn_base d k a)).
    induction 1;intros.
    rewrite swapR_Pr.
    apply WFProc.
    rewrite swapR_F.
    apply WFAbs.
    rewrite swapR_C.
    apply WFConc.
    intros.
    destruct n.
    inversion H0.
    assert (i <= d+n);try omega.
    generalize (deBruijn.swapR_In_fn P.E.deBruijn_fn_base (d+n) i H1 0 k m).
    rewrite H;auto.
    rewrite plus_n_Sm.
    simpl.
    intro.
    congruence.
  Qed.
End Agent.


