(* A formalisation of the spi calculus in Coq *)
(* Copyright (C) 2005-2007 Sebastien Briais *)
(* http://lamp.epfl.ch/~sbriais/ *)

(* This library is free software; you can redistribute it and/or modify *)
(* it under the terms of the GNU Lesser General Public License as *)
(* published by the Free Software Foundation; either version 2.1 of the *)
(* License, or (at your option) any later version. *)

(* This library is distributed in the hope that it will be useful, but *)
(* WITHOUT ANY WARRANTY; without even the implied warranty of *)
(* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU *)
(* Lesser General Public License for more details. *)

(* You should have received a copy of the GNU Lesser General Public *)
(* License along with this library; if not, write to the Free Software *)
(* Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA *)
(* 02110-1301 USA *)

Require Export processType.
Require Export agentType.
Require Import deBruijnType.
Require Import congruenceType.

Module Congruence(Ag:AgentType) <: CongruenceType with Module A := Ag.
  Require Export Relations.
  Require Import Arith.
  Require Import tactics.
  Require Import missing.
  Module A := Ag.

  Lemma mk_nu_plus : forall (n m:nat)(x:A.P.t),(A.P.mk_nu n (A.P.mk_nu m x))=(A.P.mk_nu (n+m) x).
    induction n;intros.
    rewrite A.P.mk_nu_0.
    reflexivity.
    rewrite A.P.mk_nu_S_1.
    rewrite IHn.
    rewrite <- A.P.mk_nu_S_1.
    reflexivity.
  Qed.

  Inductive Rcongruence (R:relation A.P.t) : relation A.P.t :=
    | CongAxiom : forall (p q:A.P.t),(R p q)->(Rcongruence R p q)
    | CongRefl : forall (p:A.P.t),(Rcongruence R p p)
    | CongTrans : forall (p q:A.P.t),(Rcongruence R p q)->forall (r:A.P.t),(Rcongruence R q r)->(Rcongruence R p r)
    | CongInput : forall (p q:A.P.t),(Rcongruence R p q)->forall (e:A.P.E.t),(Rcongruence R (A.P.Input e p) (A.P.Input e q))
    | CongOutput : forall (p q:A.P.t),(Rcongruence R p q)->forall (e f:A.P.E.t),(Rcongruence R (A.P.Output e f p) (A.P.Output e f q))
    | CongIfThen : forall (p q:A.P.t),(Rcongruence R p q)->forall (phi:A.P.F.t),(Rcongruence R (A.P.IfThen phi p) (A.P.IfThen phi q))
    | CongNew : forall (p q:A.P.t),(Rcongruence R p q)->(Rcongruence R (A.P.New p) (A.P.New q))
    | CongSumL : forall (p q:A.P.t),(Rcongruence R p q)->forall (r:A.P.t),(Rcongruence R (A.P.Sum p r) (A.P.Sum q r))
    | CongParL : forall (p q:A.P.t),(Rcongruence R p q)->forall (r:A.P.t),(Rcongruence R (A.P.Par p r) (A.P.Par q r)).

  Hint Resolve CongAxiom CongRefl CongTrans CongInput CongOutput CongIfThen CongNew CongSumL CongParL : Rcongruence_laws.

  Lemma Rcongruence_sym : forall (R:relation A.P.t),(symmetric A.P.t R)->(symmetric A.P.t (Rcongruence R)).
    unfold symmetric.
    intros.
    induction H0;auto with Rcongruence_laws.
    eapply CongTrans;eauto.
  Qed.

  Lemma Rcongruence_inclusion : forall (R:relation A.P.t),(inclusion A.P.t R (Rcongruence R)).
    unfold inclusion;intros.
    apply CongAxiom;auto.
  Qed.

  Lemma Rcongruence_increase : forall (R R':relation A.P.t),(inclusion A.P.t R R')->(inclusion A.P.t (Rcongruence R) (Rcongruence R')).
    unfold inclusion;intros.
    induction H0;auto with Rcongruence_laws.
    eapply CongTrans;eauto.
  Qed.

  Lemma Rcongruence_same_relation : forall (R R':relation A.P.t),(same_relation A.P.t R R')->(same_relation A.P.t (Rcongruence R) (Rcongruence R')).
    unfold same_relation.
    intros;and_elim.
    split;apply Rcongruence_increase;auto.
  Qed.

  Lemma Rcongruence_union : forall (R R':relation A.P.t),(same_relation A.P.t (Rcongruence (union A.P.t (Rcongruence R) R')) (Rcongruence (union A.P.t R R'))).
    unfold same_relation;unfold inclusion.
    split;intros.
    induction H;auto with Rcongruence_laws.
    case H;intro.
    generalize Rcongruence_increase;unfold inclusion;intros.
    apply H1 with R;trivial.
    intros;left;trivial.
    generalize Rcongruence_increase;unfold inclusion;intros.
    apply H1 with R';trivial.
    intros;right;trivial.
    apply CongAxiom;trivial.
    eapply CongTrans;eauto.
    generalize Rcongruence_increase;unfold inclusion;intros.
    apply H0 with (union A.P.t R R').
    intros.
    case H1;intro;[left | right];auto.
    apply Rcongruence_inclusion;trivial.
    auto.
  Qed.

  Lemma Rcongruence_mk_nu : forall (R:relation A.P.t)(p q:A.P.t),(Rcongruence R p q)->forall (n:nat),(Rcongruence R (A.P.mk_nu n p) (A.P.mk_nu n q)).
    induction n;intros.
    repeat (rewrite A.P.mk_nu_0);trivial.
    repeat (rewrite A.P.mk_nu_S_1);auto with Rcongruence_laws.
  Qed.

  Inductive SumMonoid : relation A.P.t :=
    | SumZeroL : forall (p:A.P.t),(SumMonoid (A.P.Sum A.P.Zero p) p)
    | SumZeroR : forall (p:A.P.t),(SumMonoid p (A.P.Sum A.P.Zero p))
    | SumComm : forall (p q:A.P.t),(SumMonoid (A.P.Sum p q) (A.P.Sum q p))
    | SumAssocL : forall (p q r:A.P.t),(SumMonoid (A.P.Sum (A.P.Sum p q) r) (A.P.Sum p (A.P.Sum q r))).

  Hint Resolve SumZeroL SumZeroR SumComm SumAssocL : SumMonoid_laws.

  Lemma Rcongruence_SumMonoid_sym : (symmetric A.P.t (Rcongruence SumMonoid)).
    unfold symmetric.
    induction 1;auto with Rcongruence_laws.
    induction H;try (apply CongAxiom;auto with SumMonoid_laws;fail).
    eapply CongTrans.
    apply CongAxiom;apply SumComm.
    eapply CongTrans.
    apply CongAxiom;apply SumAssocL.
    eapply CongTrans.
    apply CongAxiom;apply SumComm.
    eapply CongTrans.
    apply CongAxiom;apply SumAssocL.
    apply CongAxiom;apply SumComm.
    eapply CongTrans;eauto.
  Qed.

  Inductive ParMonoid : relation A.P.t :=
    | ParZeroL : forall (p:A.P.t),(ParMonoid (A.P.Par A.P.Zero p) p)
    | ParZeroR : forall (p:A.P.t),(ParMonoid p (A.P.Par A.P.Zero p))
    | ParComm : forall (p q:A.P.t),(ParMonoid (A.P.Par p q) (A.P.Par q p))
    | ParAssocL : forall (p q r:A.P.t),(ParMonoid (A.P.Par (A.P.Par p q) r) (A.P.Par p (A.P.Par q r))).

  Hint Resolve ParZeroL ParZeroR ParComm ParAssocL : ParMonoid_laws.

  Lemma Rcongruence_ParMonoid_sym : (symmetric A.P.t (Rcongruence ParMonoid)).
    unfold symmetric.
    induction 1;auto with Rcongruence_laws.
    induction H;try (apply CongAxiom;auto with ParMonoid_laws;fail).
    eapply CongTrans.
    apply CongAxiom;apply ParComm.
    eapply CongTrans.
    apply CongAxiom;apply ParAssocL.
    eapply CongTrans.
    apply CongAxiom;apply ParComm.
    eapply CongTrans.
    apply CongAxiom;apply ParAssocL.
    apply CongAxiom;apply ParComm.
    eapply CongTrans;eauto.
  Qed.

  Inductive ScopeExtrusion : relation A.P.t :=
    | ScopeExtrusionL : forall (p q:A.P.t),(ScopeExtrusion (A.P.Par (A.P.New p) q) (A.P.New (A.P.Par p ((deBruijn.lift A.P.deBruijn_base) 0%nat 1%nat q))))
    | ScopeExtrusionR : forall (p q:A.P.t),(ScopeExtrusion (A.P.New (A.P.Par p ((deBruijn.lift A.P.deBruijn_base) 0%nat 1%nat q))) (A.P.Par (A.P.New p) q)).

  Hint Resolve ScopeExtrusionL ScopeExtrusionR : ScopeExtrusion_laws.

  Lemma ScopeExtrusion_sym : (symmetric A.P.t ScopeExtrusion).
    unfold symmetric;intros.
    induction H;auto with ScopeExtrusion_laws.
  Qed.

  Lemma ScopeExtrusion_scopeL : forall (k:nat)(p q:A.P.t),(Rcongruence ScopeExtrusion (A.P.Par (A.P.mk_nu k p) q) (A.P.mk_nu k (A.P.Par p ((deBruijn.lift A.P.deBruijn_base) 0%nat k q)))).
    induction k;intros;[repeat (rewrite A.P.mk_nu_0) | repeat (rewrite A.P.mk_nu_S_2)].
    rewrite deBruijn.lift_0.
    apply CongRefl.
    eapply CongTrans.
    apply IHk.
    apply Rcongruence_mk_nu.
    replace (S k) with (1+k);try reflexivity.
    rewrite <- deBruijn.lift_plus.
    apply CongAxiom.
    apply ScopeExtrusionL.
  Qed.

  Lemma ParMonoidScopeExtrusion_scopeR : forall (k:nat)(p q:A.P.t),(Rcongruence (union A.P.t ParMonoid ScopeExtrusion) (A.P.Par p (A.P.mk_nu k q)) (A.P.mk_nu k (A.P.Par (deBruijn.lift A.P.deBruijn_base 0%nat k p) q))).
    intros.
    eapply CongTrans.
    apply CongAxiom.
    left;apply ParComm.
    eapply CongTrans.
    apply (Rcongruence_increase ScopeExtrusion (union A.P.t ParMonoid ScopeExtrusion)).
    unfold inclusion;unfold union;tauto.
    apply ScopeExtrusion_scopeL.
    apply Rcongruence_mk_nu.
    apply CongAxiom.
    left;apply ParComm.
  Qed.

  Inductive NewSwap : relation A.P.t :=
    | NewSwapL : forall (p:A.P.t),(NewSwap (A.P.New (A.P.New p)) (A.P.New (A.P.New ((deBruijn.swapR A.P.deBruijn_base) 0%nat 1%nat p)))).

  Hint Resolve NewSwapL : NewSwap_laws.

  Lemma NewSwap_sym : symmetric A.P.t NewSwap.
    unfold symmetric.
    induction 1.
    replace (A.P.New (A.P.New p)) with (A.P.New (A.P.New ((deBruijn.swapR A.P.deBruijn_base) 0%nat 1%nat ((deBruijn.swapR A.P.deBruijn_base) 0%nat 1%nat p)))).
    auto with NewSwap_laws.
    rewrite deBruijn.swapR_1_idem;trivial.
  Qed.

  Lemma NewSwap_mk_nu_swapR : forall (k:nat)(p:A.P.t),(Rcongruence NewSwap (A.P.New (A.P.mk_nu k p)) (A.P.mk_nu k (A.P.New ((deBruijn.swapR A.P.deBruijn_base) 0%nat k p)))).
    induction k;intros.
    repeat (rewrite A.P.mk_nu_0).
    rewrite deBruijn.swapR_0.
    apply CongRefl.
    repeat (rewrite A.P.mk_nu_S_1).
    eapply CongTrans.
    apply CongAxiom.
    apply NewSwapL.
    rewrite A.P.swapR_mk_nu.
    simpl.
    apply CongNew.
    eapply CongTrans.
    apply IHk.
    repeat (rewrite <- A.P.mk_nu_S_2).
    repeat (rewrite A.P.mk_nu_S_1).
    generalize (deBruijn.swapR_simpl A.P.deBruijn_base 0 k 1);simpl;rewrite plus_comm;simpl;intro.
    rewrite H.
    apply CongRefl.
  Qed.

  Lemma NewSwap_mk_nu_swapR_plus : forall (k:nat)(i:nat),(i<=k)->forall (p:A.P.t),(Rcongruence NewSwap (A.P.mk_nu (S k) p) (A.P.mk_nu (S k) (deBruijn.swapR A.P.deBruijn_base i (k-i) p))).
    intros.
    assert (k = (k-i) + i);try omega.
    rewrite H0.
    replace (S (k-i+i)) with ((S (k-i))+i);try omega.
    repeat (rewrite <- mk_nu_plus).
    rewrite <- H0.
    generalize (A.P.swapR_mk_nu 0 (k-i) i);simpl;intro.
    rewrite <- H1.
    rewrite A.P.mk_nu_S_1.
    rewrite A.P.mk_nu_S_2.
    apply NewSwap_mk_nu_swapR.
  Qed.

  Inductive NewAxioms : relation A.P.t :=
  | NewZeroL : NewAxioms A.P.Zero (A.P.New (A.P.Zero))
  | NewZeroR : NewAxioms (A.P.New (A.P.Zero)) A.P.Zero
  | NewSumL : forall (p q:A.P.t),(NewAxioms (A.P.Sum (A.P.New p) q) (A.P.New (A.P.Sum p ((deBruijn.lift A.P.deBruijn_base) 0%nat 1%nat q))))
  | NewSumR : forall (p q:A.P.t),(NewAxioms (A.P.New (A.P.Sum p ((deBruijn.lift A.P.deBruijn_base) 0%nat 1%nat q))) (A.P.Sum (A.P.New p) q)).

  Hint Resolve NewZeroL NewZeroR NewSumL NewSumR : New_laws.

  Lemma NewAxioms_sym : (symmetric A.P.t NewAxioms).
    unfold symmetric;intros.
    induction H;auto with New_laws.
  Qed.

  Inductive BangAxioms : relation A.P.t :=
  | BangExpL : forall (p:A.P.t),(BangAxioms (A.P.Bang p) (A.P.Par p (A.P.Bang p)))
  | BangExpR : forall (p:A.P.t),(BangAxioms (A.P.Par p (A.P.Bang p)) (A.P.Bang p)).

  Hint Resolve BangExpL BangExpR : Bang_laws.

  Lemma BangAxioms_sym : (symmetric A.P.t BangAxioms).
    unfold symmetric;intros.
    induction H;auto with Bang_laws.
  Qed.

  Definition R_preserves_lift (R:relation A.P.t) := forall (d k:nat),(R_preserves_f R ((deBruijn.lift A.P.deBruijn_base) d k)).
  
  Lemma Rcongruence_lift : forall (R:relation A.P.t),(R_preserves_lift R)->(R_preserves_lift (Rcongruence R)).
    unfold R_preserves_lift;unfold R_preserves_f;intros R H d k p q H0.
    generalize k;clear k.
    generalize d;clear d.
    induction H0;intros;repeat (rewrite A.P.lift_Input || rewrite A.P.lift_Output || rewrite A.P.lift_IfThen || rewrite A.P.lift_New || rewrite A.P.lift_Sum || rewrite A.P.lift_Par);auto with Rcongruence_laws.
    eapply CongTrans;eauto.
  Qed.
  
  Lemma SumMonoid_preserves_lift : (R_preserves_lift SumMonoid).
    unfold R_preserves_lift;unfold R_preserves_f;intros.
    induction H;repeat (rewrite A.P.lift_Sum || rewrite A.P.lift_Zero);auto with SumMonoid_laws.
  Qed.

  Lemma ParMonoid_preserves_lift : (R_preserves_lift ParMonoid).
    unfold R_preserves_lift;unfold R_preserves_f;intros.
    induction H;repeat (rewrite A.P.lift_Par || rewrite A.P.lift_Zero);auto with ParMonoid_laws.
  Qed.

  Lemma ScopeExtrusion_preserves_lift : (R_preserves_lift ScopeExtrusion).
    unfold R_preserves_lift;unfold R_preserves_f;intros.
    induction H;repeat (rewrite A.P.lift_Par || rewrite A.P.lift_New);generalize (deBruijn.lift_comm A.P.deBruijn_base 0 d 1 k);rewrite plus_comm;simpl;intro H;rewrite H;auto with ScopeExtrusion_laws.
  Qed.

  Lemma NewSwap_preserves_lift : (R_preserves_lift NewSwap).
    unfold R_preserves_lift;unfold R_preserves_f;intros.
    induction H;repeat (rewrite A.P.lift_Par || rewrite A.P.lift_New);generalize (deBruijn.lift_swapR_comm A.P.deBruijn_base 0 d 1 k);simpl;intro H;rewrite H;auto with NewSwap_laws.
  Qed.

  Lemma NewAxioms_preserves_lift : (R_preserves_lift NewAxioms).
    unfold R_preserves_lift;unfold R_preserves_f;intros.
    induction H;repeat (rewrite A.P.lift_Sum || rewrite A.P.lift_New || rewrite A.P.lift_Zero);auto with New_laws;generalize (deBruijn.lift_comm A.P.deBruijn_base 0 d 1 k);rewrite plus_comm;simpl;intro H;rewrite H;auto with New_laws.
  Qed.

  Lemma BangAxioms_preserves_lift : (R_preserves_lift BangAxioms).
    unfold R_preserves_lift;unfold R_preserves_f;intros.
    induction H;repeat (rewrite A.P.lift_Par || rewrite A.P.lift_Zero || rewrite A.P.lift_Bang);auto with Bang_laws.
  Qed.

  Definition R_preserves_swapR (R:relation A.P.t) := forall (d k:nat),(R_preserves_f R ((deBruijn.swapR A.P.deBruijn_base) d k)).

  Lemma Rcongruence_swapR :  forall (R:relation A.P.t),(R_preserves_swapR R)->(R_preserves_swapR (Rcongruence R)).
    unfold R_preserves_swapR;unfold R_preserves_f;intros R H d k p q H0.
    generalize k;clear k.
    generalize d;clear d.
    induction H0;intros;repeat (rewrite A.P.swapR_Input || rewrite A.P.swapR_Output || rewrite A.P.swapR_IfThen || rewrite A.P.swapR_New || rewrite A.P.swapR_Sum || rewrite A.P.swapR_Par);auto with Rcongruence_laws.
    eapply CongTrans;eauto.
  Qed.
  
  Lemma SumMonoid_preserves_swapR : (R_preserves_swapR SumMonoid).
    unfold R_preserves_swapR;unfold R_preserves_f;intros.
    induction H;repeat (rewrite A.P.swapR_Sum || rewrite A.P.swapR_Zero);auto with SumMonoid_laws.
  Qed.

  Lemma ParMonoid_preserves_swapR : (R_preserves_swapR ParMonoid).
    unfold R_preserves_swapR;unfold R_preserves_f;intros.
    induction H;repeat (rewrite A.P.swapR_Par || rewrite A.P.swapR_Zero);auto with ParMonoid_laws.
  Qed.

  Lemma ScopeExtrusion_preserves_swapR : (R_preserves_swapR ScopeExtrusion).
    unfold R_preserves_swapR;unfold R_preserves_f;intros.
    induction H;repeat (rewrite A.P.swapR_Par || rewrite A.P.swapR_New);generalize (deBruijn.swapR_lift_comm A.P.deBruijn_base 0 d 1 k);simpl;intro H;rewrite H;auto with ScopeExtrusion_laws.
  Qed.

  Lemma NewSwap_preserves_swapR : (R_preserves_swapR NewSwap).
    unfold R_preserves_swapR;unfold R_preserves_f;intros.
    induction H;repeat (rewrite A.P.swapR_Par || rewrite A.P.swapR_New);generalize (deBruijn.swapR_comm A.P.deBruijn_base 0 d 1 k);simpl;intro H;rewrite H;auto with NewSwap_laws.
  Qed.

  Lemma NewAxioms_preserves_swapR : (R_preserves_swapR NewAxioms).
    unfold R_preserves_swapR;unfold R_preserves_f;intros.
    induction H;repeat (rewrite A.P.swapR_Sum || rewrite A.P.swapR_New || rewrite A.P.swapR_Zero);auto with New_laws;generalize (deBruijn.swapR_lift_comm A.P.deBruijn_base 0 d 1 k);simpl;intro H;rewrite H;auto with New_laws.
  Qed.

  Lemma BangAxioms_preserves_swapR : (R_preserves_swapR BangAxioms).
    unfold R_preserves_swapR;unfold R_preserves_f;intros.
    induction H;repeat (rewrite A.P.swapR_Par || rewrite A.P.swapR_Zero || rewrite A.P.swapR_Bang);auto with Bang_laws.
  Qed.

  Definition R_preserves_low_subst (R:relation A.P.t) := forall (d k:nat)(es:list A.P.E.t),(R_preserves_f R (deBruijn.low_subst A.P.deBruijn_subst d k es)).

  Lemma Rcongruence_low_subst :  forall (R:relation A.P.t),(R_preserves_low_subst R)->(R_preserves_low_subst (Rcongruence R)).
    unfold R_preserves_low_subst;unfold R_preserves_f;intros R H d k es p q H0.
    generalize es;clear es.
    generalize k;clear k.
    generalize d;clear d.
    induction H0;intros;repeat (rewrite A.P.low_subst_Input || rewrite A.P.low_subst_Output || rewrite A.P.low_subst_IfThen || rewrite A.P.low_subst_New || rewrite A.P.low_subst_Sum || rewrite A.P.low_subst_Par);auto with Rcongruence_laws.
    eapply CongTrans;eauto.
  Qed.

  Lemma SumMonoid_preserves_low_subst : (R_preserves_low_subst SumMonoid).
    unfold R_preserves_low_subst;unfold R_preserves_f;intros.
    induction H;repeat (rewrite A.P.low_subst_Sum || rewrite A.P.low_subst_Zero);auto with SumMonoid_laws.
  Qed.

  Lemma ParMonoid_preserves_low_subst : (R_preserves_low_subst ParMonoid).
    unfold R_preserves_low_subst;unfold R_preserves_f;intros.
    induction H;repeat (rewrite A.P.low_subst_Par || rewrite A.P.low_subst_Zero);auto with ParMonoid_laws.
  Qed.
  
  Lemma ScopeExtrusion_preserves_low_subst : (R_preserves_low_subst ScopeExtrusion).
    unfold R_preserves_low_subst;unfold R_preserves_f;intros.
    induction H;repeat (rewrite A.P.low_subst_Par || rewrite A.P.low_subst_New);generalize (deBruijn.low_subst_lift_comm A.P.deBruijn_subst 0 d 1 k);rewrite plus_comm;simpl;intro H;rewrite H;auto with ScopeExtrusion_laws.
  Qed.

  Lemma NewSwap_preserves_low_subst : (R_preserves_low_subst NewSwap).
    unfold R_preserves_low_subst;unfold R_preserves_f;intros.
    induction H;repeat (rewrite A.P.low_subst_Par || rewrite A.P.low_subst_New);generalize (deBruijn.low_subst_swapR_comm A.P.deBruijn_subst 0 d 1 k);simpl;intro H;rewrite H;auto with NewSwap_laws.
  Qed.

  Lemma NewAxioms_preserves_low_subst : (R_preserves_low_subst NewAxioms).
    unfold R_preserves_low_subst;unfold R_preserves_f;intros.
    induction H;repeat (rewrite A.P.low_subst_Sum || rewrite A.P.low_subst_New || rewrite A.P.low_subst_Zero);auto with New_laws;generalize (deBruijn.low_subst_lift_comm A.P.deBruijn_subst 0 d 1 k);rewrite plus_comm;simpl;intro H;rewrite H;auto with New_laws.
  Qed.

  Lemma BangAxioms_preserves_low_subst : (R_preserves_low_subst BangAxioms).
    unfold R_preserves_low_subst;unfold R_preserves_f;intros.
    induction H;repeat (rewrite A.P.low_subst_Par || rewrite A.P.low_subst_Zero || rewrite A.P.low_subst_Bang);auto with Bang_laws.
  Qed.

  Lemma ParMonoidScopeExtrusionNewSwap_scopeL_swapL : forall (k:nat)(p q:A.P.t),(Rcongruence (union A.P.t (union A.P.t ParMonoid ScopeExtrusion) NewSwap) (A.P.mk_nu k (A.P.Par (A.P.New p) q)) (A.P.mk_nu k (A.P.New (A.P.Par (deBruijn.swapL A.P.deBruijn_base 0%nat k p) (deBruijn.lift A.P.deBruijn_base k 1 q))))).
    induction k;intros;simpl.
    apply CongAxiom.
    left;right.
    repeat (rewrite A.P.mk_nu_0).
    apply ScopeExtrusionL.
    repeat (rewrite A.P.mk_nu_S_1).
    eapply CongTrans.
    apply CongNew.
    apply IHk.
    repeat (rewrite <- A.P.mk_nu_S_2).
    repeat (rewrite A.P.mk_nu_S_1).
    eapply CongTrans.
    apply CongAxiom.
    right.
    apply NewSwapL.
    rewrite A.P.swapR_mk_nu.
    simpl.
    rewrite A.P.swapR_Par.
    rewrite deBruijn.swapR_lift_simpl.
    generalize (deBruijn.swapR_swapL_simpl A.P.deBruijn_base 0 k).
    rewrite plus_0_l;intro.
    rewrite H.
    apply CongRefl.
  Qed.

  Lemma ParMonoidScopeExtrusion_sym : symmetric A.P.t (Rcongruence (union A.P.t ParMonoid ScopeExtrusion)).
    eapply same_relation_symmetric.
    apply Rcongruence_union.
    apply Rcongruence_sym.
    apply union_symmetric.
    apply Rcongruence_ParMonoid_sym.
    apply ScopeExtrusion_sym.
  Qed.

  Lemma ParMonoidScopeExtrusionNewSwap_sym : symmetric A.P.t (Rcongruence (union A.P.t (union A.P.t ParMonoid ScopeExtrusion) NewSwap)).
    eapply same_relation_symmetric.
    apply Rcongruence_union.
    apply Rcongruence_sym.
    apply union_symmetric.
    apply ParMonoidScopeExtrusion_sym.
    apply NewSwap_sym.
  Qed.

  Lemma NewAxioms_NewGarbageL : forall (p:A.P.t),(Rcongruence (union _ (union A.P.t ParMonoid ScopeExtrusion) NewAxioms) p (A.P.New (deBruijn.lift A.P.deBruijn_base 0 1 p))).
    intros.
    eapply CongTrans.
    apply Rcongruence_increase with ParMonoid.
    unfold inclusion, union;intros;auto.
    apply CongAxiom.
    eapply ParZeroR.
    eapply CongTrans.
    apply CongParL.
    apply Rcongruence_increase with NewAxioms.
    unfold inclusion, union;intros;auto.
    apply CongAxiom.
    apply NewZeroL.
    eapply CongTrans.
    apply Rcongruence_increase with ScopeExtrusion.
    unfold inclusion, union;intros;auto.
    apply CongAxiom.
    apply ScopeExtrusionL.
    apply CongNew.
    apply Rcongruence_increase with ParMonoid.
    unfold inclusion, union;intros;auto.
    apply CongAxiom.
    eapply ParZeroL.
  Qed.

  Lemma NewAxioms_NewGarbageR : forall (p:A.P.t),(Rcongruence (union A.P.t (union _ ParMonoid ScopeExtrusion) NewAxioms) (A.P.New (deBruijn.lift A.P.deBruijn_base 0 1 p)) p).
    intros.
    elim (Rcongruence_union (union _ ParMonoid ScopeExtrusion) NewAxioms).
    intros.
    apply H.
    apply Rcongruence_sym.
    apply union_symmetric.
    generalize (Rcongruence_union ParMonoid ScopeExtrusion).
    intro.
    eapply same_relation_symmetric.
    apply H1.
    apply Rcongruence_sym.
    apply union_symmetric.
    apply Rcongruence_ParMonoid_sym.
    apply ScopeExtrusion_sym.
    apply NewAxioms_sym.
    apply H0.
    apply NewAxioms_NewGarbageL.
  Qed.

  Inductive StructuralCongruenceAxioms : relation A.P.t :=
  | SCSumMonoid : forall (p q:A.P.t),(SumMonoid p q)->(StructuralCongruenceAxioms p q)
  | SCParMonoid : forall (p q:A.P.t),(ParMonoid p q)->(StructuralCongruenceAxioms p q)
  | SCScopeExtrusion : forall (p q:A.P.t),(ScopeExtrusion p q)->(StructuralCongruenceAxioms p q)
  | SCNewSwap : forall (p q:A.P.t),(NewSwap p q)->(StructuralCongruenceAxioms p q)
  | SCNewAxioms : forall (p q:A.P.t),(NewAxioms p q)->(StructuralCongruenceAxioms p q)
  | SCBangAxioms : forall (p q:A.P.t),(BangAxioms p q)->(StructuralCongruenceAxioms p q).

  Lemma StructuralCongruenceAxioms_preserves_lift : (R_preserves_lift StructuralCongruenceAxioms).
    unfold R_preserves_lift;intros.
    unfold R_preserves_f.
    induction 1.
    apply SCSumMonoid.
    apply SumMonoid_preserves_lift;auto.
    apply SCParMonoid.
    apply ParMonoid_preserves_lift;auto.
    apply SCScopeExtrusion.
    apply ScopeExtrusion_preserves_lift;auto.
    apply SCNewSwap.
    apply NewSwap_preserves_lift;auto.
    apply SCNewAxioms.
    apply NewAxioms_preserves_lift;auto.
    apply SCBangAxioms.
    apply BangAxioms_preserves_lift;auto.
  Qed.

  Lemma StructuralCongruenceAxioms_preserves_swapR : (R_preserves_swapR StructuralCongruenceAxioms).
    unfold R_preserves_swapR;intros.
    unfold R_preserves_f.
    induction 1.
    apply SCSumMonoid.
    apply SumMonoid_preserves_swapR;auto.
    apply SCParMonoid.
    apply ParMonoid_preserves_swapR;auto.
    apply SCScopeExtrusion.
    apply ScopeExtrusion_preserves_swapR;auto.
    apply SCNewSwap.
    apply NewSwap_preserves_swapR;auto.
    apply SCNewAxioms.
    apply NewAxioms_preserves_swapR;auto.
    apply SCBangAxioms.
    apply BangAxioms_preserves_swapR;auto.
  Qed.

  Lemma StructuralCongruenceAxioms_preserves_low_subst : (R_preserves_low_subst StructuralCongruenceAxioms).
    unfold R_preserves_low_subst;intros.
    unfold R_preserves_f.
    induction 1.
    apply SCSumMonoid.
    apply SumMonoid_preserves_low_subst;auto.
    apply SCParMonoid.
    apply ParMonoid_preserves_low_subst;auto.
    apply SCScopeExtrusion.
    apply ScopeExtrusion_preserves_low_subst;auto.
    apply SCNewSwap.
    apply NewSwap_preserves_low_subst;auto.
    apply SCNewAxioms.
    apply NewAxioms_preserves_low_subst;auto.
    apply SCBangAxioms.
    apply BangAxioms_preserves_low_subst;auto.
  Qed.

  Lemma StructuralCongruence_symmetric : symmetric _ (Rcongruence StructuralCongruenceAxioms).
    eapply same_relation_symmetric with (Rcongruence (union _ (union _ (union _ (union _ (union _ SumMonoid ParMonoid) ScopeExtrusion) NewSwap) NewAxioms) BangAxioms)).
    apply Rcongruence_same_relation.
    split;unfold inclusion;intros.
    case H;clear H;intro.
    case H;clear H;intro.
    case H;clear H;intro.
    case H;clear H;intro.
    case H;clear H;intro.
    apply SCSumMonoid;auto.
    apply SCParMonoid;auto.
    apply SCScopeExtrusion;auto.
    apply SCNewSwap;auto.
    apply SCNewAxioms;auto.
    apply SCBangAxioms;auto.
    unfold union.
    induction H;try tauto.
    left;left;right;auto.
    eapply same_relation_symmetric.
    apply Rcongruence_union.
    apply Rcongruence_sym.
    apply union_symmetric.
    eapply same_relation_symmetric.
    apply Rcongruence_union.
    apply Rcongruence_sym.
    apply union_symmetric.
    eapply same_relation_symmetric.
    apply Rcongruence_union.
    apply Rcongruence_sym.
    apply union_symmetric.
    eapply same_relation_symmetric.
    apply Rcongruence_union.
    apply Rcongruence_sym.
    apply union_symmetric.
    eapply same_relation_symmetric with (Rcongruence (union _ (Rcongruence SumMonoid) (Rcongruence ParMonoid))).
    eapply same_relation_trans.
    apply Rcongruence_union.
    eapply same_relation_trans.
    apply Rcongruence_same_relation.
    apply union_comm.
    eapply same_relation_trans.
    apply Rcongruence_union.
    apply Rcongruence_same_relation.
    apply union_comm.
    apply Rcongruence_sym.
    apply union_symmetric.
    apply Rcongruence_SumMonoid_sym.
    apply Rcongruence_ParMonoid_sym.
    apply ScopeExtrusion_sym.
    apply NewSwap_sym.
    apply NewAxioms_sym.
    apply BangAxioms_sym.
  Qed.

  Definition preserves_mem_fn (R:relation A.P.t) := forall (p q:A.P.t),(R p q)->forall (i d:nat),(deBruijn.mem_fn A.P.deBruijn_fn_base i d p)=(deBruijn.mem_fn A.P.deBruijn_fn_base i d q).

  Lemma Rcongruence_mem_fn : forall (R:relation A.P.t),(preserves_mem_fn R)->(preserves_mem_fn (Rcongruence R)).
    unfold preserves_mem_fn;intros.
    generalize i;clear i.
    generalize d;clear d.
    induction H0;intros;auto.
    rewrite IHRcongruence1;auto.
    repeat (rewrite A.P.fn_Input).
    rewrite IHRcongruence;auto.
    repeat (rewrite A.P.fn_Output).
    rewrite IHRcongruence;auto.
    repeat (rewrite A.P.fn_IfThen).
    rewrite IHRcongruence;auto.
    repeat (rewrite A.P.fn_New).
    rewrite IHRcongruence;auto.
    repeat (rewrite A.P.fn_Sum).
    rewrite IHRcongruence;auto.
    repeat (rewrite A.P.fn_Par).
    rewrite IHRcongruence;auto.
  Qed.

  Lemma SumMonoid_preserves_mem_fn : preserves_mem_fn SumMonoid.
    unfold preserves_mem_fn.
    induction 1;intros.
    rewrite A.P.fn_Sum.
    rewrite A.P.fn_Zero.
    reflexivity.
    rewrite A.P.fn_Sum.
    rewrite A.P.fn_Zero.
    reflexivity.
    repeat (rewrite A.P.fn_Sum).
    rewrite orb_comm;auto.
    repeat (rewrite A.P.fn_Sum).
    rewrite orb_assoc.
    auto.
  Qed.

  Lemma ParMonoid_preserves_mem_fn : preserves_mem_fn ParMonoid.
    unfold preserves_mem_fn.
    induction 1;intros.
    rewrite A.P.fn_Par.
    rewrite A.P.fn_Zero.
    reflexivity.
    rewrite A.P.fn_Par.
    rewrite A.P.fn_Zero.
    reflexivity.
    repeat (rewrite A.P.fn_Par).
    rewrite orb_comm;auto.
    repeat (rewrite A.P.fn_Par).
    rewrite orb_assoc.
    auto.
  Qed.

  Lemma ScopeExtrusion_preserves_mem_fn : preserves_mem_fn ScopeExtrusion.
    unfold preserves_mem_fn.
    induction 1;intros.
    repeat (rewrite A.P.fn_Par || rewrite A.P.fn_New).
    generalize (deBruijn.fn_plus_lift A.P.deBruijn_fn_base 0 (S d)).
    simpl;intro.
    rewrite <- H.
    rewrite <- H.
    generalize (deBruijn.lift_fn A.P.deBruijn_fn_base 0 1 q (i+d)).
    replace (DeBruijnNat.lift 0 (S d) i) with (i+S d);try reflexivity.
    replace (DeBruijnNat.lift 0 1 (i+d)) with (i+d+1);try reflexivity.
    rewrite <- plus_assoc.
    rewrite (plus_comm d 1);simpl.
    intros.
    rewrite H0.
    generalize (deBruijn.fn_plus_lift A.P.deBruijn_fn_base 0 d q i).
    simpl.
    intro.
    rewrite <- H1.
    reflexivity.
    repeat (rewrite A.P.fn_Par || rewrite A.P.fn_New).
    generalize (deBruijn.fn_plus_lift A.P.deBruijn_fn_base 0 (S d)).
    simpl;intro.
    rewrite <- H.
    rewrite <- H.
    generalize (deBruijn.lift_fn A.P.deBruijn_fn_base 0 1 q (i+d)).
    replace (DeBruijnNat.lift 0 (S d) i) with (i+S d);try reflexivity.
    replace (DeBruijnNat.lift 0 1 (i+d)) with (i+d+1);try reflexivity.
    rewrite <- plus_assoc.
    rewrite (plus_comm d 1);simpl.
    intros.
    rewrite H0.
    generalize (deBruijn.fn_plus_lift A.P.deBruijn_fn_base 0 d q i).
    simpl.
    intro.
    rewrite <- H1.
    reflexivity.
  Qed.

  Lemma NewSwap_preserves_mem_fn : preserves_mem_fn NewSwap.
    unfold preserves_mem_fn.
    induction 1.
    intros.
    repeat (rewrite A.P.fn_New).
    generalize (deBruijn.swapR_fn_same A.P.deBruijn_fn_base 0 d 1 p).
    simpl.
    intros.
    rewrite H.
    auto.
  Qed.

  Lemma NewAxioms_preserves_mem_fn : preserves_mem_fn NewAxioms.
    unfold preserves_mem_fn.
    induction 1;intros.
    repeat (rewrite A.P.fn_Zero || rewrite A.P.fn_New);auto.
    repeat (rewrite A.P.fn_Zero || rewrite A.P.fn_New);auto.
    repeat (rewrite A.P.fn_Sum || rewrite A.P.fn_New).
    generalize (deBruijn.fn_plus_lift A.P.deBruijn_fn_base 0 (S d)).
    simpl;intro.
    rewrite <- H.
    rewrite <- H.
    generalize (deBruijn.lift_fn A.P.deBruijn_fn_base 0 1 q (i+d)).
    replace (DeBruijnNat.lift 0 (S d) i) with (i+S d);try reflexivity.
    replace (DeBruijnNat.lift 0 1 (i+d)) with (i+d+1);try reflexivity.
    rewrite <- plus_assoc.
    rewrite (plus_comm d 1);simpl.
    intros.
    rewrite H0.
    generalize (deBruijn.fn_plus_lift A.P.deBruijn_fn_base 0 d q i).
    simpl.
    intro.
    rewrite <- H1.
    reflexivity.
    repeat (rewrite A.P.fn_Sum || rewrite A.P.fn_New).
    generalize (deBruijn.fn_plus_lift A.P.deBruijn_fn_base 0 (S d)).
    simpl;intro.
    rewrite <- H.
    rewrite <- H.
    generalize (deBruijn.lift_fn A.P.deBruijn_fn_base 0 1 q (i+d)).
    replace (DeBruijnNat.lift 0 (S d) i) with (i+S d);try reflexivity.
    replace (DeBruijnNat.lift 0 1 (i+d)) with (i+d+1);try reflexivity.
    rewrite <- plus_assoc.
    rewrite (plus_comm d 1);simpl.
    intros.
    rewrite H0.
    generalize (deBruijn.fn_plus_lift A.P.deBruijn_fn_base 0 d q i).
    simpl.
    intro.
    rewrite <- H1.
    reflexivity.
  Qed.

  Lemma BangAxioms_preserves_mem_fn : preserves_mem_fn BangAxioms.
    unfold preserves_mem_fn.
    induction 1;intros.
    repeat (rewrite A.P.fn_Par || rewrite A.P.fn_Bang).
    destruct (deBruijn.mem_fn A.P.deBruijn_fn_base i d p);reflexivity.
    repeat (rewrite A.P.fn_Par || rewrite A.P.fn_Bang).
    destruct (deBruijn.mem_fn A.P.deBruijn_fn_base i d p);reflexivity.
  Qed.

  Lemma StructuralCongruenceAxioms_preserves_mem_fn : preserves_mem_fn StructuralCongruenceAxioms.
    unfold preserves_mem_fn.
    induction 1;intros.
    apply SumMonoid_preserves_mem_fn;auto.
    apply ParMonoid_preserves_mem_fn;auto.
    apply ScopeExtrusion_preserves_mem_fn;auto.
    apply NewSwap_preserves_mem_fn;auto.
    apply NewAxioms_preserves_mem_fn;auto.
    apply BangAxioms_preserves_mem_fn;auto.    
  Qed.

  Inductive Acongruence (R:relation A.P.t) : relation A.t :=
    | ACong : forall (p q:A.P.t),(Rcongruence R p q)->(Acongruence R (A.Pr p) (A.Pr q))
    | ACongTrans : forall (a b:A.t),(Acongruence R a b)->forall (c:A.t),(Acongruence R b c)->(Acongruence R a c)
    | AAbs : forall (p q:A.P.t),(Rcongruence R p q)->(Acongruence R (A.F (A.Abs p)) (A.F (A.Abs q)))
    | AConc : forall (p q:A.P.t),(Rcongruence R p q)->forall (e:A.P.E.t),(Acongruence R (A.C (A.Conc 0 e p)) (A.C (A.Conc 0 e q)))
    | ACongNew : forall (a b:A.t),(Acongruence R a b)->(Acongruence R (A.new a) (A.new b))
    | ANewSwapL : forall (a:A.t),(A.well_formed a)->(Acongruence R (A.new (A.new a)) (A.new (A.new (deBruijn.swapR A.deBruijn_base 0 1 a)))).

  Lemma Acongruence_increase : forall (R R':relation A.P.t),(inclusion A.P.t R R')->(inclusion A.t (Acongruence R) (Acongruence R')).
    unfold inclusion;intros.
    induction H0.
    apply ACong.
    apply (Rcongruence_increase R R' H);auto.
    eapply ACongTrans;eauto.
    apply AAbs.
    apply (Rcongruence_increase R R' H);auto.
    apply AConc.
    apply (Rcongruence_increase R R' H);auto.
    apply ACongNew;auto.
    apply ANewSwapL;auto.
  Qed.

  Lemma Acongruence_Pr : forall (R:relation A.P.t),(inclusion A.P.t NewSwap R)->forall (a b:A.t),(Acongruence R a b)->forall (p:A.P.t),(a = A.Pr p)->(exists q:A.P.t,(b = A.Pr q)/\(Rcongruence R p q)).
    intros R RNewSwap.
    induction 1;intros;try (discriminate H0).
    injection H0;intros;subst p0.
    exists q.
    split;auto.
    elim (IHAcongruence1 p H1);intro q;intros;and_elim.
    elim (IHAcongruence2 q H2);intro r;intros;and_elim.
    exists r.
    split;auto.
    eapply CongTrans;eauto.
    destruct a;try (discriminate H0).
    simpl in H0.
    elim (IHAcongruence t (refl_equal (A.Pr t))).
    intro t1;intros;and_elim.
    rewrite A.new_Pr in H0.
    injection H0;intros.
    exists (A.P.New t1).
    rewrite H1.
    rewrite A.new_Pr.
    split;auto.
    rewrite <- H3.
    apply CongNew;auto.
    rewrite A.new_F in H0;discriminate H0.
    rewrite A.new_C in H0;discriminate H0.
    destruct a.
    rewrite A.swapR_Pr.
    repeat (rewrite A.new_Pr).
    exists (A.P.New (A.P.New (deBruijn.swapR A.P.deBruijn_base 0 1 t))).
    split;auto.
    apply CongAxiom.
    repeat (rewrite A.new_Pr in H0).
    injection H0;intros;subst p.
    apply RNewSwap.
    apply NewSwapL.
    repeat (rewrite A.new_F in H0);discriminate H0.
    repeat (rewrite A.new_C in H0);discriminate H0.
  Qed.

  Implicit Arguments Acongruence_Pr [R a b].

  Lemma Acongruence_F : forall (R:relation A.P.t),(inclusion A.P.t NewSwap R)->(R_preserves_swapR (Rcongruence R))->forall (a b:A.t),(Acongruence R a b)->forall (p:A.P.t),(a = A.F (A.Abs p))->(exists q:A.P.t,(b = A.F (A.Abs q))/\(Rcongruence R p q)).
    intros R RNewSwap RswapR.
    induction 1;intros;try (discriminate H0);auto.
    elim (IHAcongruence1 p H1);intro q;intros;and_elim.
    elim (IHAcongruence2 q H2);intro r;intros;and_elim.
    exists r.
    split;auto.
    eapply CongTrans;eauto.
    exists q.
    split;auto.
    injection H0;intros;subst p0;auto.
    destruct a.
    rewrite A.new_Pr in H0;discriminate H0.
    rewrite A.new_F in H0.
    destruct a.
    rewrite A.new_abs_def in H0.
    injection H0;intros.
    subst p.
    elim (IHAcongruence t (refl_equal (A.F (A.Abs t)))).
    intro t1;intros;and_elim.
    exists (A.P.New (deBruijn.swapR A.P.deBruijn_base 0 1 t1)).
    rewrite H1.
    rewrite A.new_F.
    rewrite A.new_abs_def.
    split;auto.
    apply CongNew.
    apply RswapR;auto.
    rewrite A.new_C in H0;discriminate H0.
    destruct a.
    repeat (rewrite A.new_Pr in H0);discriminate H0.
    destruct a.
    repeat (rewrite A.new_F in H0).
    repeat (rewrite A.new_abs_def in H0).
    rewrite A.P.swapR_New in H0.
    injection H0;intros.
    rewrite A.swapR_F.
    repeat (rewrite A.new_F).
    repeat (rewrite A.new_abs_def).
    rewrite A.P.swapR_New.
    exists (A.P.New (A.P.New (deBruijn.swapR A.P.deBruijn_base 1 1 (deBruijn.swapR A.P.deBruijn_base 0 1 (deBruijn.swapR A.P.deBruijn_base 1 1 t))))).
    split;auto.
    subst p.
    generalize (deBruijn.swapR_simpl A.P.deBruijn_base 0 1 1);simpl;intro.
    rewrite H1.
    generalize (deBruijn.swapR_compose A.P.deBruijn_base 0 1 0);simpl;intro.
    rewrite <- H2.
    rewrite <- H1.
    apply CongAxiom.
    apply RNewSwap.
    apply NewSwapL.
    repeat (rewrite A.new_C in H0);discriminate H0.
  Qed.

  Implicit Arguments Acongruence_F [R a b].

  Lemma Acongruence_C : forall (R:relation A.P.t),(inclusion A.P.t NewSwap R)->(inclusion A.P.t ScopeExtrusion R)->forall (a b:A.t),(Acongruence R a b)->forall (n:nat)(m:A.P.E.t)(p:A.P.t),(a = A.C (A.Conc n m p))->(exists m':A.P.E.t,exists q:A.P.t,b = A.C (A.Conc n m' q) /\ (forall (i:nat),(deBruijn.mem_fn A.P.E.deBruijn_fn_base i n m)=(deBruijn.mem_fn A.P.E.deBruijn_fn_base i n m')) /\ forall (k:nat)(f:A.abstraction),(Acongruence R (A.pseudo_appl_r (A.Conc (k+n) m p) f) (A.pseudo_appl_r (A.Conc (k+n) m' q) f))).
    intros R RNewSwap RScopeExtrusion.
    induction 1;intros;try (discriminate H0);auto.
    (* ACongTrans *)
    elim (IHAcongruence1 n m p H1);intro m';intro.
    ex_elim H2 q.
    elim (IHAcongruence2 n m' q H2);intro m'';intro.
    ex_elim H5 r.
    exists m''.
    exists r.
    split;auto.
    split;auto.
    intros.
    rewrite H3.
    auto.
    intros.
    generalize (H4 k f);generalize (H7 k f).
    destruct f;rewrite A.pseudo_appl_r_def.
    intros.
    elim (Acongruence_Pr RNewSwap H9 (A.P.mk_nu (k+n) (A.P.Par p (deBruijn.low_subst A.P.deBruijn_subst 0 (k+n) (cons m nil) t))));auto.
    intro x;intros;and_elim.
    injection H10;clear H10;intros.
    subst x.
    clear H9.
    elim (Acongruence_Pr RNewSwap H8 (A.P.mk_nu (k+n) (A.P.Par q (deBruijn.low_subst A.P.deBruijn_subst 0 (k+n) (cons m' nil) t))));auto.
    intro x;intros;and_elim.
    rewrite A.pseudo_appl_r_def in H9.
    injection H9;clear H9;intros.
    subst x.
    clear H8.
    repeat (rewrite A.pseudo_appl_r_def).
    apply ACong.
    eapply CongTrans;eauto.
    rewrite A.pseudo_appl_r_def;auto.
    (* AConc *)
    injection H0;intros.
    subst p0.
    subst n.
    subst e.
    exists m.
    exists q.
    split;auto.
    split;auto.
    intros.
    destruct f;simpl.
    repeat (rewrite A.pseudo_appl_r_def).
    rewrite plus_0_r.
    apply ACong.
    apply Rcongruence_mk_nu.
    apply CongParL;auto.
    (* ACongNew *)
    destruct a.
    rewrite A.new_Pr in H0;discriminate H0.
    rewrite A.new_F in H0;discriminate H0.
    destruct c;rewrite A.new_C in H0.
    rewrite A.new_conc_def in H0.
    case (Bool_elim_bool (deBruijn.mem_fn A.P.E.deBruijn_fn_base 0 n0 t));intro.
    rewrite H1 in H0.
    elim (IHAcongruence n0 t t0 (refl_equal (A.C (A.Conc n0 t t0))));intro t';intros.
    ex_elim H2 t0'.
    generalize (H3 0%nat).
    rewrite H1;intro.
    symmetry in H5.
    injection H0;intros.
    subst n.
    subst m.
    subst p.
    clear H0.
    exists t'.
    exists t0'.
    subst b.
    rewrite A.new_C.
    rewrite A.new_conc_def.
    rewrite H5.
    split;auto.
    split.
    intro.
    generalize (deBruijn.fn_plus_lift A.P.E.deBruijn_fn_base n0 1).
    unfold DeBruijnNat.lift.
    simpl.
    rewrite plus_comm;simpl.
    intro.
    rewrite <- H0.
    rewrite <- H0.
    auto.
    intros.
    clear IHAcongruence.
    generalize (H4 (S k) f).
    destruct f.
    repeat (rewrite A.pseudo_appl_r_def).
    intro.
    elim (Acongruence_Pr RNewSwap H0 (A.P.mk_nu (S k + n0) (A.P.Par t0 (deBruijn.low_subst A.P.deBruijn_subst 0 (S k + n0) (cons t nil) t1))));auto.
    intro x;intros;and_elim.
    clear H0.
    injection H2;clear H2;intros;subst x.
    apply ACong.
    replace (k+S n0) with (S k+n0);try omega.
    auto.
    rewrite H1 in H0.
    injection H0;clear H0;intros.
    subst n;subst m;subst p.
    elim (IHAcongruence n0 t t0);auto.
    intro t';intros.
    ex_elim H0 t0'.
    generalize (H2 0%nat).
    rewrite H1;intro.
    symmetry in H4.
    subst b.
    rewrite A.new_C.
    rewrite A.new_conc_def.
    rewrite H4.
    exists (deBruijn.low1 A.P.E.deBruijn_base n0 t').
    exists (A.P.New (deBruijn.swapR A.P.deBruijn_base 0 n0 t0')).
    split;auto.
    split.
    intro.
    rewrite (deBruijn.lift_low1_fn A.P.E.deBruijn_fn_base n0 t H1).
    rewrite (deBruijn.lift_low1_fn A.P.E.deBruijn_fn_base n0 t' H4).
    auto.
    clear IHAcongruence.
    intros.
    destruct f.
    repeat (rewrite A.pseudo_appl_r_def).
    apply ACong.
    eapply CongTrans.
    apply Rcongruence_mk_nu.
    apply Rcongruence_increase with ScopeExtrusion;auto.
    apply CongAxiom.
    apply ScopeExtrusionL.
    generalize (deBruijn.lift_low_subst_succ A.P.deBruijn_subst 0 0 (k+n0) 1);simpl;intro.
    rewrite H0.
    simpl.
    rewrite deBruijn.lift_0.
    generalize (deBruijn.lift_1_low1_simpl A.P.E.deBruijn_fn_base 0 n0);simpl;intro.
    rewrite H5;auto.
    replace (k+n0+1) with (n0+S k);try omega.
    generalize (deBruijn.swapR_low_subst_idem A.P.deBruijn_subst 0 n0 k (cons t nil));simpl;intro.
    rewrite <- H6.
    rewrite <- A.P.swapR_Par.
    rewrite <- A.P.mk_nu_S_2.
    replace (S (k+n0)) with (S n0+k);try omega.
    eapply CongTrans.
    apply Rcongruence_increase with NewSwap;auto.
    rewrite plus_comm.
    rewrite <- mk_nu_plus.
    rewrite A.P.mk_nu_S_2.
    apply Rcongruence_sym.
    apply NewSwap_sym.
    apply Rcongruence_mk_nu.
    apply NewSwap_mk_nu_swapR.
    rewrite <- A.P.mk_nu_S_1.
    rewrite mk_nu_plus.
    replace (k+S n0) with (n0+S k);try omega.
    generalize (H3 (S k) (A.Abs t1)).
    repeat (rewrite A.pseudo_appl_r_def).
    rewrite plus_comm;intro.
    elim (Acongruence_Pr RNewSwap H7 (A.P.mk_nu (n0 + S k) (A.P.Par t0 (deBruijn.low_subst A.P.deBruijn_subst 0 (n0 + S k) (cons t nil) t1))));auto.
    intro x;intros;and_elim.
    injection H8;clear H8;intros;subst x.
    clear H7.
    eapply CongTrans.
    apply H9.
    replace (n0+S k) with (k+S n0);try omega.
    rewrite <- mk_nu_plus.
    rewrite A.P.mk_nu_S_1.
    eapply CongTrans.
    apply Rcongruence_mk_nu.
    apply Rcongruence_increase with NewSwap;auto.
    apply NewSwap_mk_nu_swapR.
    rewrite mk_nu_plus.
    clear H9.
    rewrite A.P.swapR_Par.
    clear H6.
    generalize (deBruijn.swapR_low_subst_idem A.P.deBruijn_subst 0 n0 k (cons t' nil));simpl;intro.
    replace (k+S n0) with (n0+S k);try omega.
    rewrite H6.
    rewrite <- H5;auto.
    generalize (H0 (cons (deBruijn.low1 A.P.E.deBruijn_base n0 t') nil) t1);simpl.
    replace (k+n0+1) with (n0+S k);try omega.
    rewrite deBruijn.lift_0.
    intro.
    rewrite <- H7.
    apply Rcongruence_mk_nu.
    apply Rcongruence_increase with ScopeExtrusion;auto.
    eapply CongTrans.
    apply CongAxiom.
    apply ScopeExtrusionR.
    apply CongRefl.
    (* ANewSwapL *)
    destruct a.
    repeat (rewrite A.new_Pr in H0);discriminate H0.
    repeat (rewrite A.new_F in H0);discriminate H0.
    destruct c.
    rewrite A.swapR_C.
    rewrite plus_0_l.
    repeat (rewrite A.new_C in H0).
    repeat (rewrite A.new_C).
    generalize (deBruijn.swapR_fn A.P.E.deBruijn_fn_base n0 1 t 1).
    unfold DeBruijnNat.swapR;simpl;intro.
    rewrite A.new_conc_def.
    rewrite H1;clear H1.
    case (Bool_elim_bool (deBruijn.mem_fn A.P.E.deBruijn_fn_base 0 n0 t));intro.
    rewrite A.new_conc_def in H0.
    rewrite H1 in H0.
    rewrite A.new_conc_def in H0.
    case (Bool_elim_bool (deBruijn.mem_fn A.P.E.deBruijn_fn_base 0 (S n0) t));intro.
    rewrite H2 in H0.
    injection H0;clear H0;intros;subst n;subst m;subst p.
    generalize (deBruijn.fn_plus_lift A.P.E.deBruijn_fn_base n0 1 t 0).
    unfold DeBruijnNat.lift;simpl;intro.
    rewrite H0.
    rewrite plus_comm;simpl.
    rewrite H2.
    rewrite A.new_conc_def.
    generalize (deBruijn.fn_plus_lift A.P.E.deBruijn_fn_base n0 1 (deBruijn.swapR A.P.E.deBruijn_base n0 1 t) 0).
    unfold DeBruijnNat.lift;simpl;intro.
    rewrite plus_comm in H3;simpl in H3.
    rewrite <- H3.
    generalize (deBruijn.swapR_fn A.P.E.deBruijn_fn_base n0 1 t 0).
    unfold DeBruijnNat.swapR;simpl;intro.
    rewrite H4.
    rewrite H1.
    exists (deBruijn.swapR A.P.E.deBruijn_base n0 1 t).
    exists (deBruijn.swapR A.P.deBruijn_base n0 1 t0).
    split;auto.
    split;intros.
    generalize (deBruijn.swapR_fn_same A.P.E.deBruijn_fn_base n0 0 1);simpl;intros.
    replace (n0+1+1) with (S (S n0)) in H5;try omega.
    rewrite H5;auto.
    destruct f.
    repeat (rewrite A.pseudo_appl_r_def).
    apply ACong.
    repeat (rewrite <- mk_nu_plus).
    repeat (rewrite A.P.mk_nu_S_1).
    apply Rcongruence_mk_nu.
    eapply CongTrans.
    apply Rcongruence_increase with NewSwap;auto.
    apply CongAxiom.
    apply NewSwapL.
    rewrite A.P.swapR_mk_nu.
    rewrite A.P.swapR_Par.
    simpl.
    generalize (deBruijn.swapR_low_subst_comm_idem A.P.deBruijn_subst 0 n0 k 1).
    replace (n0+k+2) with (k+S (S n0));try omega.
    simpl;intro.
    rewrite H5.
    simpl.
    apply CongRefl.
    rewrite H2 in H0.
    injection H0;clear H0;intros.
    subst n;subst m;subst p.
    generalize (deBruijn.fn_plus_lift A.P.E.deBruijn_fn_base n0 1 t 0).
    unfold DeBruijnNat.lift;simpl;intros.
    rewrite plus_comm in H0;simpl in H0.
    rewrite H0.
    rewrite H2.
    rewrite A.new_conc_def.
    generalize (deBruijn.low1_swapR_simpl A.P.E.deBruijn_fn_base n0 1).
    rewrite plus_comm;simpl.
    intros.
    rewrite H3;auto.
    generalize (deBruijn.low1_In_fn A.P.E.deBruijn_fn_base 0 0 (le_refl 0) n0);rewrite plus_comm;simpl;intros.
    rewrite <- H4;auto.
    rewrite H1.
    exists (deBruijn.low1 A.P.E.deBruijn_base (S n0) t).
    exists (A.P.New (deBruijn.swapR A.P.deBruijn_base 0 n0 (deBruijn.swapR A.P.deBruijn_base n0 1 t0))).
    split;auto.
    split;auto.
    intros.
    destruct f.
    repeat (rewrite A.pseudo_appl_r_def).
    apply ACong.
    generalize (deBruijn.swapR_simpl A.P.deBruijn_base 0 n0 1).
    simpl.
    rewrite plus_comm;simpl;intros.
    rewrite H5.
    apply CongRefl.
    rewrite A.new_conc_def in H0.
    rewrite H1 in H0.
    rewrite A.new_conc_def in H0.
    generalize (deBruijn.lift_low1_fn A.P.E.deBruijn_fn_base n0 t H1 0).
    unfold DeBruijnNat.lift;simpl;intros.
    rewrite <- H2.
    case (Bool_elim_bool (deBruijn.mem_fn A.P.E.deBruijn_fn_base 0 n0 (deBruijn.low1 A.P.E.deBruijn_base n0 t)));intro;rewrite H3;simpl;rewrite H3 in H2;symmetry in H2.
    generalize (deBruijn.fn_plus_lift A.P.E.deBruijn_fn_base n0 1).
    rewrite plus_comm;unfold DeBruijnNat.lift;simpl.
    intro.
    rewrite A.new_conc_def.
    rewrite <- H4.
    simpl.
    generalize (deBruijn.swapR_fn A.P.E.deBruijn_fn_base n0 1 t 0).
    unfold DeBruijnNat.swapR;simpl.
    intro.
    rewrite H5.
    rewrite H1.
    rewrite H3 in H0.
    injection H0;clear H0;intros;subst n;subst m;subst p. 
    generalize (deBruijn.fn_plus_lift A.P.E.deBruijn_fn_base n0 1 (deBruijn.swapR A.P.E.deBruijn_base n0 1 t) 0);rewrite plus_comm;simpl.
    unfold DeBruijnNat.lift;simpl.
    rewrite H5.
    rewrite H1;intro.
    symmetry in H0.
    generalize (deBruijn.low1_swapR_simpl A.P.E.deBruijn_fn_base n0 1).
    rewrite plus_comm;simpl.
    intros.
    rewrite <- H6;auto.
    rewrite deBruijn.swapR_1_idem.
    exists (deBruijn.low1 A.P.E.deBruijn_base n0 t).
    exists (A.P.New (deBruijn.swapR A.P.deBruijn_base 0 (S n0) (deBruijn.swapR A.P.deBruijn_base n0 1 t0))).
    split;auto.
    split;auto.
    intros.
    generalize (deBruijn.swapR_simpl A.P.deBruijn_base 0 n0 1);simpl;rewrite plus_comm;simpl;intros.
    rewrite <- H7.
    rewrite deBruijn.swapR_1_idem.
    destruct f.
    repeat (rewrite (A.pseudo_appl_r_def)).
    apply ACong.
    apply CongRefl.
    rewrite H3 in H0.
    injection H0;clear H0;intros;subst n;subst m;subst p.
    generalize (deBruijn.swapR_fn A.P.E.deBruijn_fn_base n0 1 t 1);unfold DeBruijnNat.swapR;simpl;intro.
    rewrite H2 in H0.
    generalize (deBruijn.lift_low1_fn A.P.E.deBruijn_fn_base n0).
    unfold DeBruijnNat.lift;simpl.
    intros.
    rewrite A.new_conc_def.
    rewrite H4;auto.
    simpl.
    generalize (deBruijn.swapR_fn A.P.E.deBruijn_fn_base n0 1 t 0);unfold DeBruijnNat.swapR;simpl;intro.
    rewrite H1 in H5.
    rewrite H5.
    rewrite A.P.swapR_New.
    rewrite (deBruijn.swapR_eq_cond A.P.E.deBruijn_fn_base n0 1).
    exists (deBruijn.low1 A.P.E.deBruijn_base n0 (deBruijn.low1 A.P.E.deBruijn_base n0 t)).
    exists (A.P.New (A.P.New (deBruijn.swapR A.P.deBruijn_base 1 n0 (deBruijn.swapR A.P.deBruijn_base 0 n0 (deBruijn.swapR A.P.deBruijn_base n0 1 t0))))).
    split;auto.
    split;auto.
    intros.
    rewrite A.P.swapR_New.
    generalize (deBruijn.swapR_simpl A.P.deBruijn_base 0 n0 1);simpl;intro.
    rewrite H6.
    generalize (deBruijn.swapR_compose A.P.deBruijn_base 0 n0 0);intro.
    rewrite <- H7.
    generalize (deBruijn.swapR_simpl A.P.deBruijn_base 0 1 n0);simpl;intro.
    replace (n0+1) with (S n0);try omega.
    rewrite <- H8.
    destruct f.
    repeat (rewrite A.pseudo_appl_r_def).
    apply ACong.
    apply Rcongruence_mk_nu.
    apply CongParL.
    apply Rcongruence_increase with NewSwap;auto.
    apply CongAxiom.
    apply NewSwapL.
    intros.
    inversion H6;auto.
    inversion H8;auto.
  Qed.

  Implicit Arguments Acongruence_C [R a b].

  Lemma AConcR : forall (R:relation A.P.t)(p q:A.P.t),(Rcongruence R p q)->forall (m:A.P.E.t),forall (n:nat),(forall (i:nat),(i<n)->deBruijn.mem_fn A.P.E.deBruijn_fn_base i 0 m = true)->(Acongruence R (A.C (A.Conc n m p)) (A.C (A.Conc n m q))).
    intros.
    induction n.
    apply AConc.
    auto.
    assert (Acongruence R (A.C (A.Conc n m p)) (A.C (A.Conc n m q))).
    apply IHn.
    intros.
    auto.
    generalize (ACongNew R (A.C (A.Conc n m p)) (A.C (A.Conc n m q)) H1).
    assert (n < S n);try omega.
    generalize (H0 n H2).
    intro.
    generalize (deBruijn.fn_plus_lift A.P.E.deBruijn_fn_base 0 n m 0).
    unfold DeBruijnNat.lift;simpl.
    rewrite H3;intro.
    symmetry in H4.
    repeat (rewrite A.new_C).
    repeat (rewrite A.new_conc_def).
    rewrite H4.
    auto.
  Qed.

  Lemma ACongRefl : forall (R:relation A.P.t)(a:A.t),(A.well_formed a)->(Acongruence R a a).
    induction 1.
    apply ACong.
    apply CongRefl.
    apply AAbs.
    apply CongRefl.
    assert (A.well_formed (A.C (A.Conc n m p))).
    apply A.WFConc.
    auto.
    induction n.
    apply AConc.
    apply CongRefl.
    elim (A.well_formed_concretion_inv n m p H0);intros.
    rewrite H2.
    apply ACongNew.
    apply IHn;auto.
  Qed.

  Lemma ACongSym : forall (R:relation A.P.t),(symmetric A.P.t (Rcongruence R))->(symmetric A.t (Acongruence R)).
    intros.
    unfold symmetric.
    induction 1.
    apply ACong.
    apply H.
    auto.
    eapply ACongTrans;eauto.
    apply AAbs.
    apply H;auto.
    apply AConc.
    apply H;auto.
    apply ACongNew.
    auto.
    generalize (deBruijn.swapR_1_idem A.deBruijn_base 0 a).
    intro.
    replace (A.new (A.new a)) with (A.new (A.new (deBruijn.swapR A.deBruijn_base 0 1 (deBruijn.swapR A.deBruijn_base 0 1 a)))).
    apply ANewSwapL.
    apply A.well_formed_swapR;auto.
    rewrite H1.
    auto.
  Qed.

  Lemma AScopeExtrusionL_r : forall (R:relation A.P.t),(inclusion A.P.t ScopeExtrusion R)->forall (a:A.t)(q:A.P.t),(A.well_formed a)->(Acongruence R (A.par_r (A.new a) q) (A.new (A.par_r a (deBruijn.lift A.P.deBruijn_base 0 1 q)))).
    intros R RScopeExtrusion.
    destruct a;intros.
    repeat (rewrite A.new_Pr || rewrite A.par_r_Pr).
    apply ACong.
    apply Rcongruence_increase with ScopeExtrusion;auto.
    apply CongAxiom.
    apply ScopeExtrusionL.
    destruct a.
    repeat (rewrite A.new_F || rewrite A.par_r_F || rewrite A.new_abs_def || rewrite A.abs_par_r_def).
    apply AAbs.
    rewrite A.P.swapR_Par.
    rewrite (deBruijn.lift_plus A.P.deBruijn_base 0 1 1);simpl.
    generalize (deBruijn.swapR_lift_idem A.P.deBruijn_base 0 0 1 0);simpl;intros.
    rewrite H0.
    replace 2%nat with (1+1)%nat;try reflexivity.
    rewrite <- (deBruijn.lift_plus A.P.deBruijn_base 0 1 1);simpl.
    apply Rcongruence_increase with ScopeExtrusion;auto.
    apply CongAxiom.
    apply ScopeExtrusionL.
    destruct c.
    repeat (rewrite A.new_C || rewrite A.par_r_C || rewrite A.new_conc_def || rewrite A.conc_par_r_def).
    case (Bool_elim_bool (deBruijn.mem_fn A.P.E.deBruijn_fn_base 0 n t));intro H0;rewrite H0;simpl. 
    rewrite A.conc_par_r_def.
    rewrite deBruijn.lift_plus;rewrite plus_comm;simpl.
    apply AConcR.
    apply CongRefl.
    intros.
    inversion H1.
    generalize (deBruijn.fn_plus_lift A.P.E.deBruijn_fn_base 0 n t 0).
    unfold DeBruijnNat.lift;simpl.
    intro.
    rewrite H2.
    auto.
    assert (i < n).
    omega.
    inversion H.
    auto.
    rewrite A.P.swapR_Par.
    rewrite A.conc_par_r_def.
    apply AConcR.
    rewrite deBruijn.lift_plus.
    generalize (deBruijn.swapR_lift_idem A.P.deBruijn_base 0 0 n 0);simpl;intros.
    rewrite H1.
    rewrite plus_comm.
    rewrite <- deBruijn.lift_plus.
    apply Rcongruence_increase with ScopeExtrusion;auto.
    apply CongAxiom.
    apply ScopeExtrusionL.
    intros.
    inversion H.
    destruct n.
    inversion H1.
    assert (i <= n);try omega.
    generalize (deBruijn.low1_In_fn A.P.E.deBruijn_fn_base n i H6 0 t);simpl;intros.
    rewrite <- H7;auto.
  Qed.

  Lemma AScopeExtrusionL_l : forall (R:relation A.P.t),(inclusion A.P.t ScopeExtrusion R)->forall (p:A.P.t)(a:A.t),(A.well_formed a)->(Acongruence R (A.par_l (A.P.New p) a) (A.new (A.par_l p (deBruijn.lift A.deBruijn_base 0 1 a)))).
    intros R RScopeExtrusion.
    destruct a;intros.
    repeat (rewrite A.new_Pr || rewrite A.par_l_Pr || rewrite A.lift_Pr).
    apply ACong.
    apply Rcongruence_increase with ScopeExtrusion;auto.
    apply CongAxiom.
    apply ScopeExtrusionL.
    destruct a.
    repeat (rewrite A.new_F || rewrite A.par_l_F || rewrite A.lift_F || rewrite A.abs_par_l_def || rewrite A.new_abs_def).
    apply AAbs.
    rewrite A.P.swapR_Par.
    rewrite A.P.lift_New.
    rewrite (deBruijn.swapR_lift_simpl A.P.deBruijn_base 0 1).
    rewrite (deBruijn.swapR_lift_1 A.P.deBruijn_base 0 1).
    apply Rcongruence_increase with ScopeExtrusion;auto.
    apply CongAxiom.
    apply ScopeExtrusionL.    
    destruct c.
    repeat (rewrite A.new_C || rewrite A.par_l_C || rewrite A.lift_C || rewrite A.conc_par_l_def || rewrite A.new_conc_def).
    rewrite (deBruijn.lift_not_fn A.P.E.deBruijn_fn_base n 1 0);try omega.
    rewrite (deBruijn.low1_lift_simpl A.P.E.deBruijn_base n 0).
    rewrite deBruijn.lift_0.
    apply AConcR.
    rewrite A.P.swapR_Par.
    rewrite A.P.lift_New.
    rewrite (deBruijn.swapR_lift_simpl A.P.deBruijn_base 0 n).
    rewrite (deBruijn.swapR_lift_1 A.P.deBruijn_base 0 n).
    apply Rcongruence_increase with ScopeExtrusion;auto.
    apply CongAxiom.
    apply ScopeExtrusionL.    
    inversion H;auto.
  Qed.

  Lemma AScopeExtrusionL_pal : forall (R:relation A.P.t),(inclusion A.P.t ScopeExtrusion R)->(inclusion A.P.t NewSwap R)->forall (f:A.abstraction)(n:nat)(m:A.P.E.t)(q:A.P.t),(Acongruence R (A.pseudo_appl_l (A.new_abs f) (A.Conc n m q)) (A.new (A.pseudo_appl_l f (A.Conc n (deBruijn.lift A.P.E.deBruijn_base n 1 m) (deBruijn.lift A.P.deBruijn_base n 1 q))))).
    intros R RScopeExtrusion RNewSwap.
    destruct f;intros.
    repeat (rewrite A.pseudo_appl_l_def || rewrite A.new_abs_def || rewrite A.new_Pr).
    apply ACong.
    rewrite A.P.low_subst_New.
    rewrite (deBruijn.low_subst_swapR_simpl A.P.deBruijn_subst 0 n (cons m nil)).
    simpl.
    eapply CongTrans.
    apply Rcongruence_mk_nu.
    apply Rcongruence_increase with ScopeExtrusion.
    unfold inclusion, union;intros;auto.
    apply CongAxiom.
    apply ScopeExtrusionL.
    apply Rcongruence_increase with NewSwap.
    unfold inclusion, union;intros;auto.
    apply Rcongruence_sym.
    apply NewSwap_sym.
    eapply CongTrans.
    apply NewSwap_mk_nu_swapR.
    rewrite A.P.swapR_Par.
    rewrite (deBruijn.swapR_lift_1 A.P.deBruijn_base 0 n).
    apply CongRefl.
  Qed.

  Lemma AScopeExtrusionL_par : forall (R:relation A.P.t),(inclusion A.P.t ScopeExtrusion R)->(inclusion A.P.t NewSwap R)->(inclusion A.P.t ParMonoid R)->forall (c:A.concretion)(q:A.P.t),(Acongruence R (A.pseudo_appl_r (A.new_conc c) (A.Abs q)) (A.new (A.pseudo_appl_r c (A.Abs (deBruijn.lift A.P.deBruijn_base 1 1 q))))).
    intros R RScopeExtrusion RNewSwap RParMonoid.
    destruct c;intros.
    rewrite A.pseudo_appl_r_def.
    rewrite A.new_conc_def.
    case (Bool_elim_bool (deBruijn.mem_fn A.P.E.deBruijn_fn_base 0 n t));intro H;rewrite H;rewrite A.pseudo_appl_r_def;rewrite A.new_Pr;intros;simpl.
    apply ACong.
    generalize (deBruijn.low_subst_lift_succ_simpl A.P.deBruijn_subst 0 n 1 (cons t nil));simpl;intro.
    rewrite plus_comm in H0;simpl in H0.
    rewrite <- H0.
    rewrite A.P.mk_nu_S_1.
    apply CongRefl.
    apply ACong.
    apply Rcongruence_increase with (union A.P.t (union A.P.t ParMonoid ScopeExtrusion) NewSwap).
    unfold inclusion, union;intros;auto.
    case H0;auto.
    intro.
    case H1;auto.
    apply ParMonoidScopeExtrusionNewSwap_sym.
    eapply CongTrans.
    apply Rcongruence_increase with NewSwap.
    unfold inclusion, union;intros;auto.
    apply NewSwap_mk_nu_swapR.
    rewrite A.P.swapR_Par.
    apply Rcongruence_mk_nu.
    apply ParMonoidScopeExtrusionNewSwap_sym.
    eapply CongTrans.
    apply CongAxiom.
    left;right.
    apply ScopeExtrusionL.
    rewrite (deBruijn.low_subst_lift_succ_simpl A.P.deBruijn_subst 0 n 1 (cons t nil)).
    rewrite (deBruijn.lift_low_subst_succ A.P.deBruijn_subst 0 0 n 1).
    simpl.
    rewrite (deBruijn.swapR_low_subst_idem A.P.deBruijn_subst 0 n 0).
    simpl.
    generalize (deBruijn.lift_1_low1_simpl A.P.E.deBruijn_fn_base 0 n);simpl;intro.
    rewrite H0;auto.
    rewrite deBruijn.lift_0.
    apply CongRefl.
  Qed.

  Lemma AScopeExtrusionR_r : forall (R:relation A.P.t),(inclusion A.P.t ScopeExtrusion R)->forall (a:A.t)(q:A.P.t),(A.well_formed a)->(Acongruence R (A.new (A.par_r a (deBruijn.lift A.P.deBruijn_base 0 1 q))) (A.par_r (A.new a) q)).
    intros R RScopeExtrusion;intros.
    apply Acongruence_increase with ScopeExtrusion;auto.
    apply ACongSym.
    apply Rcongruence_sym.
    apply ScopeExtrusion_sym.
    apply AScopeExtrusionL_r;auto.
    unfold inclusion;auto.
  Qed.

  Lemma AScopeExtrusionR_l : forall (R:relation A.P.t),(inclusion A.P.t ScopeExtrusion R)->forall (p:A.P.t)(a:A.t),(A.well_formed a)->(Acongruence R (A.new (A.par_l p (deBruijn.lift A.deBruijn_base 0 1 a))) (A.par_l (A.P.New p) a)).
    intros R RScopeExtrusion;intros.
    apply Acongruence_increase with ScopeExtrusion;auto.
    apply ACongSym.
    apply Rcongruence_sym.
    apply ScopeExtrusion_sym.
    apply AScopeExtrusionL_l;auto.
    unfold inclusion;auto.
  Qed.

  Lemma AScopeExtrusionR_pal : forall (R:relation A.P.t),(inclusion A.P.t ScopeExtrusion R)->(inclusion A.P.t NewSwap R)->forall (f:A.abstraction)(n:nat)(m:A.P.E.t)(q:A.P.t),(Acongruence R (A.new (A.pseudo_appl_l f (A.Conc n (deBruijn.lift A.P.E.deBruijn_base n 1 m) (deBruijn.lift A.P.deBruijn_base n 1 q)))) (A.pseudo_appl_l (A.new_abs f) (A.Conc n m q))).
    intros R RScopeExtrusion RNewSwap.
    intros.
    apply Acongruence_increase with (union A.P.t ScopeExtrusion NewSwap).
    unfold inclusion, union;intros;auto.
    case H;auto.
    apply ACongSym.
    apply Rcongruence_sym.
    apply union_symmetric.
    apply ScopeExtrusion_sym.
    apply NewSwap_sym.
    apply AScopeExtrusionL_pal.
    unfold inclusion, union;intros;auto.
    unfold inclusion, union;intros;auto.
  Qed.

  Lemma AScopeExtrusionR_par : forall (R:relation A.P.t),(inclusion A.P.t ScopeExtrusion R)->(inclusion A.P.t NewSwap R)->(inclusion A.P.t ParMonoid R)->forall (c:A.concretion)(q:A.P.t),(Acongruence R (A.new (A.pseudo_appl_r c (A.Abs (deBruijn.lift A.P.deBruijn_base 1 1 q)))) (A.pseudo_appl_r (A.new_conc c) (A.Abs q))).
    intros R RScopeExtrusion RNewSwap RParMonoid.
    intros.
    apply Acongruence_increase with (union A.P.t (union A.P.t ParMonoid ScopeExtrusion) NewSwap).
    unfold inclusion, union;intros;auto.
    case H;auto.
    intro.
    case H0;auto.
    apply ACongSym.
    apply ParMonoidScopeExtrusionNewSwap_sym.
    apply AScopeExtrusionL_par.
    unfold inclusion, union;intros;auto.
    unfold inclusion, union;intros;auto.
    unfold inclusion, union;intros;auto.
  Qed.

  Lemma AParZeroL : forall (R:relation A.P.t),(inclusion A.P.t ParMonoid R)->forall (a:A.t),(A.well_formed a)->(Acongruence R (A.par_l A.P.Zero a) a).
    intros R RParMonoid.
    destruct a;intros.
    rewrite A.par_l_Pr.
    apply ACong.
    apply Rcongruence_increase with ParMonoid;auto.
    apply CongAxiom.
    apply ParZeroL.
    destruct a.
    rewrite A.par_l_F;rewrite A.abs_par_l_def.
    rewrite A.P.lift_Zero.
    apply AAbs.
    apply Rcongruence_increase with ParMonoid;auto.
    apply CongAxiom.
    apply ParZeroL.
    destruct c.
    rewrite A.par_l_C;rewrite A.conc_par_l_def.
    rewrite A.P.lift_Zero.
    apply AConcR.
    apply Rcongruence_increase with ParMonoid;auto.
    apply CongAxiom.
    apply ParZeroL.
    inversion H;auto.
  Qed.

  Lemma AParZeroR : forall (R:relation A.P.t),(inclusion A.P.t ParMonoid R)->forall (a:A.t),(A.well_formed a)->(Acongruence R a (A.par_l A.P.Zero a)).
    intros R RParMonoid.
    destruct a;intros.
    rewrite A.par_l_Pr.
    apply ACong.
    apply Rcongruence_increase with ParMonoid;auto.
    apply CongAxiom.
    apply ParZeroR.
    destruct a.
    rewrite A.par_l_F;rewrite A.abs_par_l_def.
    rewrite A.P.lift_Zero.
    apply AAbs.
    apply Rcongruence_increase with ParMonoid;auto.
    apply CongAxiom.
    apply ParZeroR.
    destruct c.
    rewrite A.par_l_C;rewrite A.conc_par_l_def.
    rewrite A.P.lift_Zero.
    apply AConcR.
    apply Rcongruence_increase with ParMonoid;auto.
    apply CongAxiom.
    apply ParZeroR.
    inversion H;auto.
  Qed.

  Lemma AParComm_r_l : forall (R:relation A.P.t),(inclusion A.P.t ParMonoid R)->forall (a:A.t)(q:A.P.t),(A.well_formed a)->(Acongruence R (A.par_r a q) (A.par_l q a)).
    intros R RParMonoid.
    destruct a;intros.
    repeat (rewrite A.par_l_Pr || rewrite A.par_r_Pr).
    apply ACong.
    apply Rcongruence_increase with ParMonoid;auto.
    apply CongAxiom.
    apply ParComm.
    destruct a.
    repeat (rewrite A.par_l_F || rewrite A.abs_par_l_def || rewrite A.par_r_F || rewrite A.abs_par_r_def).
    apply AAbs.
    apply Rcongruence_increase with ParMonoid;auto.
    apply CongAxiom.
    apply ParComm.
    destruct c.
    repeat (rewrite A.par_l_C || rewrite A.conc_par_l_def || rewrite A.par_r_C || rewrite A.conc_par_r_def).
    apply AConcR.
    apply Rcongruence_increase with ParMonoid;auto.
    apply CongAxiom.
    apply ParComm.
    inversion H;auto.
  Qed.

  Lemma AParComm_l_r : forall (R:relation A.P.t),(inclusion A.P.t ParMonoid R)->forall (p:A.P.t)(a:A.t),(A.well_formed a)->(Acongruence R (A.par_l p a) (A.par_r a p)).
    intros R RParMonoid.
    destruct a;intros.
    repeat (rewrite A.par_l_Pr || rewrite A.par_r_Pr).
    apply ACong.
    apply Rcongruence_increase with ParMonoid;auto.
    apply CongAxiom.
    apply ParComm.
    destruct a.
    repeat (rewrite A.par_l_F || rewrite A.abs_par_l_def || rewrite A.par_r_F || rewrite A.abs_par_r_def).
    apply AAbs.
    apply Rcongruence_increase with ParMonoid;auto.
    apply CongAxiom.
    apply ParComm.
    destruct c.
    repeat (rewrite A.par_l_C || rewrite A.conc_par_l_def || rewrite A.par_r_C || rewrite A.conc_par_r_def).
    apply AConcR.
    apply Rcongruence_increase with ParMonoid;auto.
    apply CongAxiom.
    apply ParComm.
    inversion H;auto.
  Qed.

  Lemma AParAssocL_r_r : forall (R:relation A.P.t),(inclusion A.P.t ParMonoid R)->forall (a:A.t)(q r:A.P.t),(A.well_formed a)->(Acongruence R (A.par_r (A.par_r a q) r) (A.par_r a (A.P.Par q r))).
    intros R RParMonoid.
    destruct a;intros.
    repeat (rewrite A.par_l_Pr || rewrite A.par_r_Pr).
    apply ACong.
    apply Rcongruence_increase with ParMonoid;auto.
    apply CongAxiom.
    apply ParAssocL.
    destruct a.
    repeat (rewrite A.par_l_F || rewrite A.abs_par_l_def || rewrite A.par_r_F || rewrite A.abs_par_r_def).
    rewrite A.P.lift_Par.
    apply AAbs.
    apply Rcongruence_increase with ParMonoid;auto.
    apply CongAxiom.
    apply ParAssocL.
    destruct c.
    repeat (rewrite A.par_l_C || rewrite A.conc_par_l_def || rewrite A.par_r_C || rewrite A.conc_par_r_def).
    rewrite A.P.lift_Par.
    apply AConcR.
    apply Rcongruence_increase with ParMonoid;auto.
    apply CongAxiom.
    apply ParAssocL.
    inversion H;auto.
  Qed.

  Lemma AParAssocL_r_l : forall (R:relation A.P.t),(inclusion A.P.t ParMonoid R)->forall (p:A.P.t)(a:A.t)(q:A.P.t),(A.well_formed a)->(Acongruence R (A.par_r (A.par_l p a) q) (A.par_l p (A.par_r a q))).
    intros R RParMonoid.
    destruct a;intros.
    repeat (rewrite A.par_l_Pr || rewrite A.par_r_Pr).
    apply ACong.
    apply Rcongruence_increase with ParMonoid;auto.
    apply CongAxiom.
    apply ParAssocL.
    destruct a.
    repeat (rewrite A.par_l_F || rewrite A.abs_par_l_def || rewrite A.par_r_F || rewrite A.abs_par_r_def).
    apply AAbs.
    apply Rcongruence_increase with ParMonoid;auto.
    apply CongAxiom.
    apply ParAssocL.
    destruct c.
    repeat (rewrite A.par_l_C || rewrite A.conc_par_l_def || rewrite A.par_r_C || rewrite A.conc_par_r_def).
    apply AConcR.
    apply Rcongruence_increase with ParMonoid;auto.
    apply CongAxiom.
    apply ParAssocL.
    inversion H;auto.
  Qed.

  Lemma AParAssocL_l : forall (R:relation A.P.t),(inclusion A.P.t ParMonoid R)->forall (p q:A.P.t)(a:A.t),(A.well_formed a)->(Acongruence R (A.par_l (A.P.Par p q) a) (A.par_l p (A.par_l q a))).
    intros R RParMonoid.
    destruct a;intros.
    repeat (rewrite A.par_l_Pr || rewrite A.par_r_Pr).
    apply ACong.
    apply Rcongruence_increase with ParMonoid;auto.
    apply CongAxiom.
    apply ParAssocL.
    destruct a.
    repeat (rewrite A.par_l_F || rewrite A.abs_par_l_def || rewrite A.par_r_F || rewrite A.abs_par_r_def).
    apply AAbs.
    apply Rcongruence_increase with ParMonoid;auto.
    apply CongAxiom.
    rewrite A.P.lift_Par.
    apply ParAssocL.
    destruct c.
    repeat (rewrite A.par_l_C || rewrite A.conc_par_l_def || rewrite A.par_r_C || rewrite A.conc_par_r_def).
    apply AConcR.
    apply Rcongruence_increase with ParMonoid;auto.
    rewrite A.P.lift_Par.
    apply CongAxiom.
    apply ParAssocL.
    inversion H;auto.
  Qed.

  Lemma APseudoApplComm_l : forall (R:relation A.P.t),(inclusion A.P.t ParMonoid R)->forall (f:A.abstraction)(c:A.concretion),(Acongruence R (A.pseudo_appl_l f c) (A.pseudo_appl_r c f)).
    intros R RParMonoid.
    intros.
    destruct f;destruct c.
    repeat (rewrite A.pseudo_appl_l_def || rewrite A.pseudo_appl_r_def).
    apply ACong.
    apply Rcongruence_mk_nu.
    apply Rcongruence_increase with ParMonoid;auto.
    apply CongAxiom.
    apply ParComm.
  Qed.

  Lemma APseudoApplComm_r : forall (R:relation A.P.t),(inclusion A.P.t ParMonoid R)->forall (f:A.abstraction)(c:A.concretion),(Acongruence R (A.pseudo_appl_r c f) (A.pseudo_appl_l f c)).
    intros R RParMonoid.
    intros.
    destruct f;destruct c.
    repeat (rewrite A.pseudo_appl_l_def || rewrite A.pseudo_appl_r_def).
    apply ACong.
    apply Rcongruence_mk_nu.
    apply Rcongruence_increase with ParMonoid;auto.
    apply CongAxiom.
    apply ParComm.
  Qed.

  Lemma AParAssocL_r_pal : forall (R:relation A.P.t),(inclusion A.P.t ParMonoid R)->(inclusion A.P.t ScopeExtrusion R)->forall (f:A.abstraction)(n:nat)(m:A.P.E.t)(q r:A.P.t),(Acongruence R (A.par_r (A.pseudo_appl_l f (A.Conc n m q)) r) (A.pseudo_appl_l f (A.Conc n m (A.P.Par q (deBruijn.lift A.P.deBruijn_base 0 n r))))).
    intros R RParMonoid RScopeExtrusion.
    destruct f as [p];intros.
    repeat (rewrite A.pseudo_appl_l_def || rewrite A.pseudo_appl_r_def).
    repeat (rewrite A.par_l_Pr || rewrite A.par_r_Pr).
    apply ACong.
    eapply CongTrans.
    apply Rcongruence_increase with ScopeExtrusion;auto.
    apply ScopeExtrusion_scopeL.
    apply Rcongruence_mk_nu.
    apply Rcongruence_increase with ParMonoid;auto.
    apply CongAxiom.
    apply ParAssocL.
  Qed.

  Lemma AParAssocL_r_par : forall (R:relation A.P.t),(inclusion A.P.t ParMonoid R)->(inclusion A.P.t ScopeExtrusion R)->forall (c:A.concretion)(f:A.abstraction)(r:A.P.t),(Acongruence R (A.par_r (A.pseudo_appl_r c f) r) (A.pseudo_appl_r c (A.abs_par_r f r))).
    intros R RParMonoid RScopeExtrusion.
    destruct c as [n m p].
    destruct f as [q].
    intros.
    repeat (rewrite A.abs_par_r_def).
    repeat (rewrite A.pseudo_appl_l_def || rewrite A.pseudo_appl_r_def).
    repeat (rewrite A.par_l_Pr || rewrite A.par_r_Pr).
    apply ACong.
    eapply CongTrans.
    apply Rcongruence_increase with ScopeExtrusion;auto.
    apply ScopeExtrusion_scopeL.
    apply Rcongruence_mk_nu.
    rewrite A.P.low_subst_Par.
    generalize (deBruijn.low_subst_lift_nil A.P.deBruijn_subst 0 n 0 (cons m nil));simpl;intro.
    rewrite H.
    rewrite plus_0_r.
    apply Rcongruence_increase with ParMonoid;auto.
    apply CongAxiom.
    apply ParAssocL.
  Qed.

  Lemma AParAssocL_pal_abs_r : forall (R:relation A.P.t),(inclusion A.P.t ParMonoid R)->forall (f:A.abstraction)(q:A.P.t)(c:A.concretion),(Acongruence R (A.pseudo_appl_l (A.abs_par_r f q) c) (A.pseudo_appl_l f (A.conc_par_l q c))).
    intros R RParMonoid.
    destruct f as [p].
    destruct c as [n m r].
    repeat (rewrite A.abs_par_r_def || rewrite A.conc_par_l_def).
    repeat (rewrite A.pseudo_appl_l_def || rewrite A.pseudo_appl_r_def).
    repeat (rewrite A.par_l_Pr || rewrite A.par_r_Pr).
    repeat (rewrite A.P.low_subst_Par).
    generalize (deBruijn.low_subst_lift_nil A.P.deBruijn_subst 0 n 0 (cons m nil));simpl;intros.
    rewrite H.
    rewrite plus_0_r.
    apply ACong.
    apply Rcongruence_mk_nu.
    apply Rcongruence_increase with ParMonoid;auto.
    apply CongAxiom.
    apply ParAssocL.
  Qed.

  Lemma AParAssocL_pal_abs_l : forall (R:relation A.P.t),(inclusion A.P.t ParMonoid R)->(inclusion A.P.t ScopeExtrusion R)->forall (p:A.P.t)(f:A.abstraction)(c:A.concretion),(Acongruence R (A.pseudo_appl_l (A.abs_par_l p f) c) (A.par_l p (A.pseudo_appl_l f c))).
    intros R RParMonoid RScopeExtrusion.
    destruct f as [q].
    destruct c as [n m r].
    intros.
    repeat (rewrite A.abs_par_l_def).
    repeat (rewrite A.pseudo_appl_l_def || rewrite A.pseudo_appl_r_def).
    repeat (rewrite A.par_l_Pr || rewrite A.par_r_Pr).
    repeat (rewrite A.P.low_subst_Par).
    generalize (deBruijn.low_subst_lift_nil A.P.deBruijn_subst 0 n 0 (cons m nil));simpl;intros.
    rewrite H.
    rewrite plus_0_r.
    apply ACong.
    apply Rcongruence_increase with (union A.P.t ParMonoid ScopeExtrusion).
    unfold inclusion, union;intros.
    case H0;auto.
    elim (Rcongruence_union ParMonoid ScopeExtrusion);intros.
    apply H0.
    assert (symmetric A.P.t (union A.P.t (Rcongruence ParMonoid) ScopeExtrusion)).
    apply union_symmetric.
    apply Rcongruence_ParMonoid_sym.
    apply ScopeExtrusion_sym.
    apply Rcongruence_sym;auto.
    apply H1.
    eapply CongTrans.
    apply ParMonoidScopeExtrusion_scopeR.
    apply Rcongruence_mk_nu.
    apply H0.
    apply Rcongruence_sym;auto.
    apply H1.
    apply CongAxiom.
    left.
    apply ParAssocL.
  Qed.

  Lemma AParAssocL_pal_conc_r : forall (R:relation A.P.t),(inclusion A.P.t ParMonoid R)->forall (c:A.concretion)(q:A.P.t)(f:A.abstraction),(Acongruence R (A.pseudo_appl_r (A.conc_par_r c q) f) (A.pseudo_appl_r c (A.abs_par_l q f))).
    intros R RParMonoid.
    destruct c as [n m p].
    destruct f as [r].
    rewrite A.conc_par_r_def.
    rewrite A.abs_par_l_def.
    repeat (rewrite A.pseudo_appl_l_def || rewrite A.pseudo_appl_r_def).
    repeat (rewrite A.par_l_Pr || rewrite A.par_r_Pr).
    apply ACong.
    rewrite A.P.low_subst_Par.
    generalize (deBruijn.low_subst_lift_nil A.P.deBruijn_subst 0 n 0 (cons m nil));simpl;intros.
    rewrite H.
    rewrite plus_0_r.
    apply Rcongruence_mk_nu.
    apply Rcongruence_increase with ParMonoid;auto.
    apply CongAxiom.
    apply ParAssocL.
  Qed.

  Lemma AParAssocL_pal_conc_l : forall (R:relation A.P.t),(inclusion A.P.t ParMonoid R)->(inclusion A.P.t ScopeExtrusion R)->forall (p:A.P.t)(c:A.concretion)(f:A.abstraction),(Acongruence R (A.pseudo_appl_r (A.conc_par_l p c) f) (A.par_l p (A.pseudo_appl_r c f))).
    intros R RParMonoid RScopeExtrusion.
    destruct c as [n m q].
    destruct f as [r];intros.
    rewrite A.conc_par_l_def.
    repeat (rewrite A.pseudo_appl_l_def || rewrite A.pseudo_appl_r_def).
    repeat (rewrite A.par_l_Pr || rewrite A.par_r_Pr).
    apply ACong.
    apply Rcongruence_increase with (union A.P.t ParMonoid ScopeExtrusion).
    unfold inclusion, union;intros.
    case H;auto.
    elim (Rcongruence_union ParMonoid ScopeExtrusion);intros.
    apply H.
    assert (symmetric A.P.t (union A.P.t (Rcongruence ParMonoid) ScopeExtrusion)).
    apply union_symmetric.
    apply Rcongruence_ParMonoid_sym.
    apply ScopeExtrusion_sym.
    apply Rcongruence_sym;auto.
    apply H0.
    eapply CongTrans.
    apply ParMonoidScopeExtrusion_scopeR.
    apply Rcongruence_mk_nu.
    apply H.
    apply Rcongruence_sym;auto.
    apply H0.
    apply CongAxiom.
    left.
    apply ParAssocL.
  Qed.

  Lemma Acongruence_well_formed : forall (R:relation A.P.t)(a b:A.t),(Acongruence R a b)->(A.well_formed a)/\(A.well_formed b).
    induction 1;split;and_elim;auto.
    apply A.WFProc.
    apply A.WFProc.
    apply A.WFAbs.
    apply A.WFAbs.
    apply A.WFConc.
    intros.
    inversion H0.
    apply A.WFConc.
    intros.
    inversion H0.
    apply A.well_formed_new;auto.
    apply A.well_formed_new;auto.
    repeat (apply A.well_formed_new;auto).
    repeat (apply A.well_formed_new;auto).
    apply A.well_formed_swapR;auto.
  Qed.

  Lemma ACongParL_par_r : forall (R:relation A.P.t),(inclusion A.P.t NewSwap R)->(inclusion A.P.t ScopeExtrusion R)->(R_preserves_swapR (Rcongruence R))->forall (a b:A.t),(Acongruence R a b)->forall (q:A.P.t),(Acongruence R (A.par_r a q) (A.par_r b q)).
    intros R RNewSwap RScopeExtrusion RswapR.
    induction 1;intros.
    repeat (rewrite A.par_l_Pr || rewrite A.par_r_Pr).
    apply ACong.
    apply CongParL;auto.
    eapply ACongTrans;eauto.
    repeat (rewrite A.par_l_F || rewrite A.abs_par_l_def || rewrite A.par_r_F || rewrite A.abs_par_r_def).
    apply AAbs.
    apply CongParL;auto.
    repeat (rewrite A.par_l_C || rewrite A.conc_par_l_def || rewrite A.par_r_C || rewrite A.conc_par_r_def).
    apply AConc.
    apply CongParL;auto.
    eapply ACongTrans.
    apply AScopeExtrusionL_r;auto.
    elim (Acongruence_well_formed R a b H);auto.
    eapply ACongTrans.
    apply ACongNew.
    apply IHAcongruence.
    apply AScopeExtrusionR_r;auto.
    elim (Acongruence_well_formed R a b H);auto.
    eapply ACongTrans.
    apply AScopeExtrusionL_r;auto.
    apply A.well_formed_new;auto.
    eapply ACongTrans.
    apply ACongNew.
    apply AScopeExtrusionL_r;auto.
    eapply ACongTrans.
    apply ANewSwapL.
    apply A.well_formed_par_r;auto.
    rewrite A.swapR_par_r.
    rewrite deBruijn.lift_plus.
    generalize (deBruijn.swapR_lift_idem A.P.deBruijn_base 0 0 1 0 q);intro.
    simpl in H0.
    replace (1+1) with 2%nat;try reflexivity.
    rewrite H0.
    replace 2%nat with (1+1);try reflexivity.
    rewrite <- deBruijn.lift_plus.
    eapply ACongTrans.
    apply ACongNew.
    apply AScopeExtrusionR_r;auto.
    apply A.well_formed_swapR;auto.
    apply AScopeExtrusionR_r;auto.
    apply A.well_formed_new.
    apply A.well_formed_swapR;auto.
  Qed.

  Lemma ACongParL_par_l : forall (R:relation A.P.t),(R_preserves_lift (Rcongruence R))->forall (p q:A.P.t),(Rcongruence R p q)->forall (a:A.t),(A.well_formed a)->(Acongruence R (A.par_l p a) (A.par_l q a)).
    intros R Rlift.
    intros.
    destruct a.
    repeat (rewrite A.par_l_Pr || rewrite A.par_r_Pr).
    apply ACong.
    apply CongParL.
    auto.
    destruct a.
    repeat (rewrite A.par_l_F || rewrite A.abs_par_l_def || rewrite A.par_r_F || rewrite A.abs_par_r_def).
    apply AAbs.
    apply CongParL.
    apply Rlift;auto.
    destruct c;simpl.
    repeat (rewrite A.par_l_C || rewrite A.conc_par_l_def || rewrite A.par_r_C || rewrite A.conc_par_r_def).
    apply AConcR.
    apply CongParL.
    apply Rlift;auto.
    inversion H0;auto.
  Qed.

  Lemma ACongParR_par_r : forall (R:relation A.P.t),(inclusion A.P.t NewSwap R)->(inclusion A.P.t ScopeExtrusion R)->(inclusion _ ParMonoid R)->(R_preserves_lift (Rcongruence R))->(R_preserves_swapR (Rcongruence R))->forall (p q:A.P.t),(Rcongruence R p q)->forall (a:A.t),(A.well_formed a)->(Acongruence R (A.par_r a p) (A.par_r a q)).
    intros R RNewSwap RScopeExtrusion RParMonoid RswapR Rlift.
    intros.
    eapply ACongTrans.
    apply AParComm_r_l;auto.
    eapply ACongTrans.
    apply ACongParL_par_l;eauto.
    apply AParComm_l_r;auto.
  Qed.
    
  Lemma ACongParL_pseudo_appl_l : forall (R:relation A.P.t),(R_preserves_low_subst (Rcongruence R))->forall (p q:A.P.t),(Rcongruence R p q)->forall (c:A.concretion),(Acongruence R (A.pseudo_appl_l (A.Abs p) c) (A.pseudo_appl_l (A.Abs q) c)).
    intros R Rlow_subst.
    intros.
    destruct c.
    repeat (rewrite A.pseudo_appl_l_def).
    apply ACong.
    apply Rcongruence_mk_nu.
    apply CongParL.
    apply Rlow_subst.
    auto.
  Qed.

  Lemma ANewGarbageL : forall (R:relation A.P.t),(inclusion _ (union _ (union _ ParMonoid ScopeExtrusion) NewAxioms) R)->forall (a:A.t),(A.well_formed a)->(Acongruence R a (A.new (deBruijn.lift A.deBruijn_base 0 1 a))).
    intros.
    destruct a.
    rewrite A.lift_Pr.
    rewrite A.new_Pr.
    apply ACong.
    apply Rcongruence_increase with (union _ (union _ ParMonoid ScopeExtrusion) NewAxioms);auto.
    apply NewAxioms_NewGarbageL.
    destruct a as [p].
    rewrite A.lift_F.
    rewrite A.new_F.
    rewrite A.new_abs_def.
    apply AAbs.
    generalize (deBruijn.swapR_lift_1 A.P.deBruijn_base 0 1).
    simpl;intros.
    rewrite H1.
    apply Rcongruence_increase with (union _ (union _ ParMonoid ScopeExtrusion) NewAxioms);auto.
    apply NewAxioms_NewGarbageL.
    destruct c as [n m p]. 
    rewrite A.lift_C.
    rewrite A.new_C.
    rewrite plus_0_l.
    rewrite A.new_conc_def.
    rewrite (deBruijn.lift_not_fn A.P.E.deBruijn_fn_base n 1 0);try omega.
    rewrite (deBruijn.low1_lift_simpl A.P.E.deBruijn_base n 0).
    rewrite deBruijn.lift_0.
    apply AConcR.
    generalize (deBruijn.swapR_lift_1 A.P.deBruijn_base 0 n).
    simpl;intros.
    rewrite H1.
    apply Rcongruence_increase with (union _ (union _ ParMonoid ScopeExtrusion) NewAxioms);auto.
    apply NewAxioms_NewGarbageL.
    inversion H0.
    auto.
  Qed.

  Lemma ANewGarbageR : forall (R:relation A.P.t),(inclusion _ (union _ (union _ ParMonoid ScopeExtrusion) NewAxioms) R)->forall (a:A.t),(A.well_formed a)->(Acongruence R (A.new (deBruijn.lift A.deBruijn_base 0 1 a)) a).
    intros.
    apply Acongruence_increase with (union _ (union _ ParMonoid ScopeExtrusion) NewAxioms).
    auto.
    apply ACongSym.
    eapply same_relation_symmetric.
    apply Rcongruence_union.
    apply Rcongruence_sym.
    apply union_symmetric.
    eapply same_relation_symmetric.
    apply Rcongruence_union.
    apply Rcongruence_sym.
    apply union_symmetric.
    apply Rcongruence_ParMonoid_sym.
    apply ScopeExtrusion_sym.
    apply NewAxioms_sym.
    apply ANewGarbageL;auto.
    unfold inclusion;auto.
  Qed.

  Require Import permutation.

  Lemma Acongruence_C_permutation : forall (R:relation A.P.t),(inclusion A.P.t NewSwap R)->(R_preserves_swapR (Rcongruence R))->(R_preserves_low_subst (Rcongruence R))->forall (a b:A.t),(Acongruence R a b)->forall (n:nat)(m:A.P.E.t)(p:A.P.t),(a = A.C (A.Conc n m p))->(exists m':A.P.E.t,exists q:A.P.t,b = A.C (A.Conc n m' q) /\ (forall (i:nat),(deBruijn.mem_fn A.P.E.deBruijn_fn_base i n m)=(deBruijn.mem_fn A.P.E.deBruijn_fn_base i n m')) /\ (forall (es:list A.P.E.t),(length es = n)->exists fs:list A.P.E.t,(is_permutation _ es fs) /\ (deBruijn.low_subst A.P.E.deBruijn_subst 0 0 es m = deBruijn.low_subst A.P.E.deBruijn_subst 0 0 fs m') /\ (Rcongruence R (deBruijn.low_subst A.P.deBruijn_subst 0 0 es p) (deBruijn.low_subst A.P.deBruijn_subst 0 0 fs q)))).
    intros R RNewSwap R_swapR R_low_subst.
    induction 1;intros;try (discriminate H0);auto.
    (* ACongTrans *)
    elim (IHAcongruence1 n m p H1);intro m';intro.
    ex_elim H2 q.
    elim (IHAcongruence2 n m' q H2);intro m'';intro.
    ex_elim H5 r.
    exists m''.
    exists r.
    split;auto.
    split;intros.
    rewrite H3.
    auto.
    elim (H4 es H8).
    intro fs;intros;and_elim.
    generalize (permutation_length _ _ _ H9).
    intro.
    rewrite H8 in H12.
    symmetry in H12.
    elim (H7 _ H12).
    intro gs;intros;and_elim.
    exists gs.
    split.
    eapply is_permutation_trans;eauto.
    split;try congruence.
    eapply CongTrans;eauto.
    (* AConc *)
    injection H0;intros.
    subst p0.
    subst n.
    subst e.
    exists m.
    exists q.
    split;auto.
    split;intros.
    auto.
    exists es.
    split.
    apply is_permutation_refl.
    split;trivial.
    apply R_low_subst;auto.
    (* ACongNew *)
    destruct a.
    rewrite A.new_Pr in H0;discriminate H0.
    rewrite A.new_F in H0;discriminate H0.
    destruct c;rewrite A.new_C in H0.
    rewrite A.new_conc_def in H0.
    case (Bool_elim_bool (deBruijn.mem_fn A.P.E.deBruijn_fn_base 0 n0 t));intro.
    rewrite H1 in H0.
    elim (IHAcongruence n0 t t0 (refl_equal (A.C (A.Conc n0 t t0))));intro t';intros.
    ex_elim H2 t0'.
    generalize (H3 0%nat).
    rewrite H1;intro.
    symmetry in H5.
    injection H0;intros.
    subst n.
    subst m.
    subst p.
    clear H0.
    exists t'.
    exists t0'.
    subst b.
    rewrite A.new_C.
    rewrite A.new_conc_def.
    rewrite H5.
    split;auto.
    split;intros.
    generalize (deBruijn.fn_plus_lift A.P.E.deBruijn_fn_base n0 1).
    unfold DeBruijnNat.lift.
    simpl.
    rewrite plus_comm;simpl.
    intro.
    rewrite <- H0.
    rewrite <- H0.
    auto.
    elim (List_length_S_inversion _ _ _ H0).
    intro es0;intro.
    ex_elim H2 e.
    subst es.
    rewrite List_length_append in H0.
    rewrite plus_comm in H0.
    simpl in H0.
    injection H0;intros.
    rewrite (A.P.E.low_subst_snoc_lemma t 0 0 e es0).
    generalize (List_length_map _ _ (deBruijn.lift A.P.E.deBruijn_base 0 1) es0).
    rewrite H2.
    intro.
    elim (H4 _ H6).
    intro x;intros;and_elim.
    elim (is_permutation_map _ _ _ H7 _ _ es0 (refl_equal _)).
    intro fs0;intros;and_elim.
    exists (app fs0 (cons e nil)).
    split.
    apply is_permutation_reverse_impl_is_permutation.
    repeat (rewrite distr_rev).
    simpl.
    eapply cons_is_permutation.
    apply is_permutation_impl_is_permutation_reverse.
    apply H10.
    apply head_insertion.
    split.
    rewrite H8.
    subst x.
    rewrite <- (A.P.E.low_subst_snoc_lemma t' 0 0 e fs0).
    auto.
    rewrite A.P.low_subst_snoc_lemma.
    rewrite A.P.low_subst_snoc_lemma.
    apply R_low_subst;auto.
    apply R_swapR;auto.
    subst x.
    auto.
    (* *)
    rewrite H1 in H0.
    injection H0;clear H0;intros.
    subst n;subst m;subst p.
    elim (IHAcongruence n0 t t0);auto.
    intro t';intros.
    ex_elim H0 t0'.
    generalize (H2 0%nat).
    rewrite H1;intro.
    symmetry in H4.
    subst b.
    rewrite A.new_C.
    rewrite A.new_conc_def.
    rewrite H4.
    exists (deBruijn.low1 A.P.E.deBruijn_base n0 t').
    exists (A.P.New (deBruijn.swapR A.P.deBruijn_base 0 n0 t0')).
    split;auto.
    split;intros.
    rewrite (deBruijn.lift_low1_fn A.P.E.deBruijn_fn_base n0 t H1).
    rewrite (deBruijn.lift_low1_fn A.P.E.deBruijn_fn_base n0 t' H4).
    auto.
    generalize (List_length_map _ _ (deBruijn.lift A.P.E.deBruijn_base 0 1) es).
    intro.
    rewrite H0 in H5.
    elim (H3 _ H5).
    intro x;intros;and_elim.
    elim (is_permutation_map _ _ _ H6 _ _ es (refl_equal _)).
    intro fs;intros;and_elim.
    subst x.
    exists fs.
    split;auto.
    split.
    assert (A.P.E.t).
    apply t.
    rewrite <- (deBruijn.low_subst_low1_same A.P.E.deBruijn_fn_subst _ H10 _ H1).
    generalize (H2 0).
    rewrite H1;intro.
    symmetry in H11.
    rewrite <- (deBruijn.low_subst_low1_same A.P.E.deBruijn_fn_subst _ H10 _ H11).
    generalize (A.P.E.low_subst_split_lemma t 0 0 es (cons H10 nil)).
    rewrite H0.
    simpl.
    intro.
    rewrite H12.
    generalize (A.P.E.low_subst_split_lemma t' 0 0 fs (cons H10 nil)).
    simpl.
    rewrite <- (permutation_length _ _ _ H9).
    rewrite H0.
    intro.
    rewrite H13.
    rewrite (A.P.E.low_subst_snoc_lemma t 0 0 H10 es).
    rewrite (A.P.E.low_subst_snoc_lemma t' 0 0 H10 fs).
    rewrite H7.
    auto.
    repeat (rewrite A.P.low_subst_New).
    generalize (deBruijn.low_subst_swapR_simpl A.P.deBruijn_subst 0 0 es).
    rewrite H0.
    intros.
    rewrite H10.
    generalize (deBruijn.low_subst_swapR_simpl A.P.deBruijn_subst 0 0 fs).
    rewrite <- (permutation_length _ _ _ H9).
    rewrite H0.
    intro.
    rewrite H11.
    apply CongNew.
    apply R_swapR;auto.
    (* ANewSwapL *)
    destruct a.
    repeat (rewrite A.new_Pr in H0);discriminate H0.
    repeat (rewrite A.new_F in H0);discriminate H0.
    destruct c.
    rewrite A.swapR_C.
    rewrite plus_0_l.
    repeat (rewrite A.new_C in H0).
    repeat (rewrite A.new_C).
    generalize (deBruijn.swapR_fn A.P.E.deBruijn_fn_base n0 1 t 1).
    unfold DeBruijnNat.swapR;simpl;intro.
    rewrite A.new_conc_def.
    rewrite H1;clear H1.
    case (Bool_elim_bool (deBruijn.mem_fn A.P.E.deBruijn_fn_base 0 n0 t));intro.
    rewrite A.new_conc_def in H0.
    rewrite H1 in H0.
    rewrite A.new_conc_def in H0.
    case (Bool_elim_bool (deBruijn.mem_fn A.P.E.deBruijn_fn_base 0 (S n0) t));intro.
    rewrite H2 in H0.
    injection H0;clear H0;intros;subst n;subst m;subst p.
    generalize (deBruijn.fn_plus_lift A.P.E.deBruijn_fn_base n0 1 t 0).
    unfold DeBruijnNat.lift;simpl;intro.
    rewrite H0.
    rewrite plus_comm;simpl.
    rewrite H2.
    rewrite A.new_conc_def.
    generalize (deBruijn.fn_plus_lift A.P.E.deBruijn_fn_base n0 1 (deBruijn.swapR A.P.E.deBruijn_base n0 1 t) 0).
    unfold DeBruijnNat.lift;simpl;intro.
    rewrite plus_comm in H3;simpl in H3.
    rewrite <- H3.
    generalize (deBruijn.swapR_fn A.P.E.deBruijn_fn_base n0 1 t 0).
    unfold DeBruijnNat.swapR;simpl;intro.
    rewrite H4.
    rewrite H1.
    exists (deBruijn.swapR A.P.E.deBruijn_base n0 1 t).
    exists (deBruijn.swapR A.P.deBruijn_base n0 1 t0).
    split;auto.
    split;intros.
    generalize (deBruijn.swapR_fn_same A.P.E.deBruijn_fn_base n0 0 1);simpl;intros.
    replace (n0+1+1) with (S (S n0)) in H5;try omega.
    rewrite H5;auto.
    elim (List_length_S_inversion _ _ _ H5).
    intro es';intro.
    ex_elim H6 e0.
    subst es.
    rewrite List_length_append in H5.
    rewrite plus_comm in H5.
    injection H5.
    intro.
    elim (List_length_S_inversion _ _ _ H6).
    intro es0;intro.
    ex_elim H7 e1.
    subst es'.
    rewrite List_length_append in H5.
    simpl in H5.
    rewrite plus_comm in H5.
    injection H5.
    clear H5;clear H6.
    intro.
    exists (app es0 (cons e0 (cons e1 nil))).
    split.
    rewrite app_ass.
    simpl.
    apply is_permutation_append.
    apply is_permutation_refl.
    eapply is_permutation_trans.
    apply is_permutation_right_rotate.
    simpl.
    apply is_permutation_refl.
    split.
    generalize (A.P.E.low_subst_split_lemma t 0 0 es0 (cons e1 (cons e0 nil))).
    simpl.
    intro.
    generalize (A.P.E.low_subst_split_lemma (deBruijn.swapR A.P.E.deBruijn_base n0 1 t) 0 0 es0 (cons e0 (cons e1 nil))).
    simpl.
    intro.
    rewrite <- H7.
    rewrite H5.
    generalize (deBruijn.low_subst_swapR_rotate A.P.E.deBruijn_subst n0 0 e1 (cons e0 nil) t).
    simpl.
    intro.
    rewrite H8.
    rewrite <- H5.
    rewrite H6.
    rewrite app_ass.
    simpl.
    auto.
    generalize (A.P.low_subst_split_lemma t0 0 0 es0 (cons e1 (cons e0 nil))).
    simpl.
    intro.
    generalize (A.P.low_subst_split_lemma (deBruijn.swapR A.P.deBruijn_base n0 1 t0) 0 0 es0 (cons e0 (cons e1 nil))).
    simpl.
    intro.
    rewrite <- H7.
    rewrite H5.
    generalize (deBruijn.low_subst_swapR_rotate A.P.deBruijn_subst n0 0 e1 (cons e0 nil) t0).
    simpl.
    intro.
    rewrite H8.
    rewrite <- H5.
    rewrite H6.
    rewrite app_ass.
    simpl.
    apply CongRefl.
    (* *)
    rewrite H2 in H0.
    injection H0;clear H0;intros.
    subst n;subst m;subst p.
    generalize (deBruijn.fn_plus_lift A.P.E.deBruijn_fn_base n0 1 t 0).
    unfold DeBruijnNat.lift;simpl;intros.
    rewrite plus_comm in H0;simpl in H0.
    rewrite H0.
    rewrite H2.
    rewrite A.new_conc_def.
    generalize (deBruijn.low1_swapR_simpl A.P.E.deBruijn_fn_base n0 1).
    rewrite plus_comm;simpl.
    intros.
    rewrite H3;auto.
    generalize (deBruijn.low1_In_fn A.P.E.deBruijn_fn_base 0 0 (le_refl 0) n0);rewrite plus_comm;simpl;intros.
    rewrite <- H4;auto.
    rewrite H1.
    exists (deBruijn.low1 A.P.E.deBruijn_base (S n0) t).
    exists (A.P.New (deBruijn.swapR A.P.deBruijn_base 0 n0 (deBruijn.swapR A.P.deBruijn_base n0 1 t0))).
    split;auto.
    split;intros.
    auto.
    exists es.
    split.
    apply is_permutation_refl.
    split.
    auto.
    repeat (rewrite A.P.low_subst_New).
    generalize (deBruijn.swapR_simpl A.P.deBruijn_base 0 n0 1).
    simpl;rewrite plus_comm;simpl;intros.
    rewrite H6.
    apply CongRefl.
    (* *)
    rewrite A.new_conc_def in H0.
    rewrite H1 in H0.
    rewrite A.new_conc_def in H0.
    generalize (deBruijn.lift_low1_fn A.P.E.deBruijn_fn_base n0 t H1 0).
    unfold DeBruijnNat.lift;simpl;intros.
    rewrite <- H2.
    case (Bool_elim_bool (deBruijn.mem_fn A.P.E.deBruijn_fn_base 0 n0 (deBruijn.low1 A.P.E.deBruijn_base n0 t)));intro;rewrite H3;simpl;rewrite H3 in H2;symmetry in H2.
    generalize (deBruijn.fn_plus_lift A.P.E.deBruijn_fn_base n0 1).
    rewrite plus_comm;unfold DeBruijnNat.lift;simpl.
    intro.
    rewrite A.new_conc_def.
    rewrite <- H4.
    simpl.
    generalize (deBruijn.swapR_fn A.P.E.deBruijn_fn_base n0 1 t 0).
    unfold DeBruijnNat.swapR;simpl.
    intro.
    rewrite H5.
    rewrite H1.
    rewrite H3 in H0.
    injection H0;clear H0;intros;subst n;subst m;subst p. 
    generalize (deBruijn.fn_plus_lift A.P.E.deBruijn_fn_base n0 1 (deBruijn.swapR A.P.E.deBruijn_base n0 1 t) 0);rewrite plus_comm;simpl.
    unfold DeBruijnNat.lift;simpl.
    rewrite H5.
    rewrite H1;intro.
    symmetry in H0.
    generalize (deBruijn.low1_swapR_simpl A.P.E.deBruijn_fn_base n0 1).
    rewrite plus_comm;simpl.
    intros.
    rewrite <- H6;auto.
    rewrite deBruijn.swapR_1_idem.
    exists (deBruijn.low1 A.P.E.deBruijn_base n0 t).
    exists (A.P.New (deBruijn.swapR A.P.deBruijn_base 0 (S n0) (deBruijn.swapR A.P.deBruijn_base n0 1 t0))).
    split;auto.
    split;auto.
    intros.
    exists es.
    split.
    apply is_permutation_refl.
    split.
    auto.
    generalize (deBruijn.swapR_simpl A.P.deBruijn_base 0 n0 1).
    simpl;rewrite plus_comm;simpl;intros.
    rewrite <- H8.
    rewrite deBruijn.swapR_1_idem.
    apply CongRefl.
    (* *)
    rewrite H3 in H0.
    injection H0;clear H0;intros;subst n;subst m;subst p.
    generalize (deBruijn.swapR_fn A.P.E.deBruijn_fn_base n0 1 t 1);unfold DeBruijnNat.swapR;simpl;intro.
    rewrite H2 in H0.
    generalize (deBruijn.lift_low1_fn A.P.E.deBruijn_fn_base n0).
    unfold DeBruijnNat.lift;simpl.
    intros.
    rewrite A.new_conc_def.
    rewrite H4;auto.
    simpl.
    generalize (deBruijn.swapR_fn A.P.E.deBruijn_fn_base n0 1 t 0);unfold DeBruijnNat.swapR;simpl;intro.
    rewrite H1 in H5.
    rewrite H5.
    rewrite A.P.swapR_New.
    rewrite (deBruijn.swapR_eq_cond A.P.E.deBruijn_fn_base n0 1).
    exists (deBruijn.low1 A.P.E.deBruijn_base n0 (deBruijn.low1 A.P.E.deBruijn_base n0 t)).
    exists (A.P.New (A.P.New (deBruijn.swapR A.P.deBruijn_base 1 n0 (deBruijn.swapR A.P.deBruijn_base 0 n0 (deBruijn.swapR A.P.deBruijn_base n0 1 t0))))).
    split;auto.
    split;auto.
    intros.
    exists es.
    split.
    apply is_permutation_refl.
    split.
    auto.
    rewrite A.P.swapR_New.
    generalize (deBruijn.swapR_simpl A.P.deBruijn_base 0 n0 1).
    simpl.
    rewrite plus_comm;simpl;intros.
    rewrite H7.
    generalize (deBruijn.swapR_compose A.P.deBruijn_base 0 n0 0).
    rewrite plus_comm;simpl;intros.
    rewrite <- H8.
    generalize (deBruijn.swapR_simpl A.P.deBruijn_base 0 1 n0).
    simpl.
    intros.
    rewrite <- H9.
    apply R_low_subst;auto.
    apply CongAxiom.
    apply RNewSwap.
    apply NewSwapL.
    intros.
    inversion H6;auto.
    inversion H8;auto.
  Qed.

  Implicit Arguments Acongruence_C_permutation [R a b].
End Congruence.
