(* A formalisation of the spi calculus in Coq *)
(* Copyright (C) 2005-2007 Sebastien Briais *)
(* http://lamp.epfl.ch/~sbriais/ *)

(* This library is free software; you can redistribute it and/or modify *)
(* it under the terms of the GNU Lesser General Public License as *)
(* published by the Free Software Foundation; either version 2.1 of the *)
(* License, or (at your option) any later version. *)

(* This library is distributed in the hope that it will be useful, but *)
(* WITHOUT ANY WARRANTY; without even the implied warranty of *)
(* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU *)
(* Lesser General Public License for more details. *)

(* You should have received a copy of the GNU Lesser General Public *)
(* License along with this library; if not, write to the Free Software *)
(* Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA *)
(* 02110-1301 USA *)

Require Export processType.
Require Export agentType.
Require Export Relations.

Require Import deBruijnType.
Require Import missing.
Require Import permutation.

Module Type CongruenceType.
  Declare Module A : AgentType.

  Parameter mk_nu_plus : forall (n m:nat)(x:A.P.t),(A.P.mk_nu n (A.P.mk_nu m x))=(A.P.mk_nu (n+m) x).

  Inductive Rcongruence (R:relation A.P.t) : relation A.P.t :=
    | CongAxiom : forall (p q:A.P.t),(R p q)->(Rcongruence R p q)
    | CongRefl : forall (p:A.P.t),(Rcongruence R p p)
    | CongTrans : forall (p q:A.P.t),(Rcongruence R p q)->forall (r:A.P.t),(Rcongruence R q r)->(Rcongruence R p r)
    | CongInput : forall (p q:A.P.t),(Rcongruence R p q)->forall (e:A.P.E.t),(Rcongruence R (A.P.Input e p) (A.P.Input e q))
    | CongOutput : forall (p q:A.P.t),(Rcongruence R p q)->forall (e f:A.P.E.t),(Rcongruence R (A.P.Output e f p) (A.P.Output e f q))
    | CongIfThen : forall (p q:A.P.t),(Rcongruence R p q)->forall (phi:A.P.F.t),(Rcongruence R (A.P.IfThen phi p) (A.P.IfThen phi q))
    | CongNew : forall (p q:A.P.t),(Rcongruence R p q)->(Rcongruence R (A.P.New p) (A.P.New q))
    | CongSumL : forall (p q:A.P.t),(Rcongruence R p q)->forall (r:A.P.t),(Rcongruence R (A.P.Sum p r) (A.P.Sum q r))
    | CongParL : forall (p q:A.P.t),(Rcongruence R p q)->forall (r:A.P.t),(Rcongruence R (A.P.Par p r) (A.P.Par q r)).

  Parameter Rcongruence_sym : forall (R:relation A.P.t),(symmetric A.P.t R)->(symmetric A.P.t (Rcongruence R)).

  Parameter Rcongruence_inclusion : forall (R:relation A.P.t),(inclusion A.P.t R (Rcongruence R)).

  Parameter Rcongruence_increase : forall (R R':relation A.P.t),(inclusion A.P.t R R')->(inclusion A.P.t (Rcongruence R) (Rcongruence R')).

  Parameter Rcongruence_same_relation : forall (R R':relation A.P.t),(same_relation A.P.t R R')->(same_relation A.P.t (Rcongruence R) (Rcongruence R')).

  Parameter Rcongruence_union : forall (R R':relation A.P.t),(same_relation A.P.t (Rcongruence (union A.P.t (Rcongruence R) R')) (Rcongruence (union A.P.t R R'))).

  Parameter Rcongruence_mk_nu : forall (R:relation A.P.t)(p q:A.P.t),(Rcongruence R p q)->forall (n:nat),(Rcongruence R (A.P.mk_nu n p) (A.P.mk_nu n q)).

  Inductive SumMonoid : relation A.P.t :=
    | SumZeroL : forall (p:A.P.t),(SumMonoid (A.P.Sum A.P.Zero p) p)
    | SumZeroR : forall (p:A.P.t),(SumMonoid p (A.P.Sum A.P.Zero p))
    | SumComm : forall (p q:A.P.t),(SumMonoid (A.P.Sum p q) (A.P.Sum q p))
    | SumAssocL : forall (p q r:A.P.t),(SumMonoid (A.P.Sum (A.P.Sum p q) r) (A.P.Sum p (A.P.Sum q r))).

  Parameter Rcongruence_SumMonoid_sym : (symmetric A.P.t (Rcongruence SumMonoid)).

  Inductive ParMonoid : relation A.P.t :=
    | ParZeroL : forall (p:A.P.t),(ParMonoid (A.P.Par A.P.Zero p) p)
    | ParZeroR : forall (p:A.P.t),(ParMonoid p (A.P.Par A.P.Zero p))
    | ParComm : forall (p q:A.P.t),(ParMonoid (A.P.Par p q) (A.P.Par q p))
    | ParAssocL : forall (p q r:A.P.t),(ParMonoid (A.P.Par (A.P.Par p q) r) (A.P.Par p (A.P.Par q r))).

  Parameter Rcongruence_ParMonoid_sym : (symmetric A.P.t (Rcongruence ParMonoid)).

  Inductive ScopeExtrusion : relation A.P.t :=
    | ScopeExtrusionL : forall (p q:A.P.t),(ScopeExtrusion (A.P.Par (A.P.New p) q) (A.P.New (A.P.Par p ((deBruijn.lift A.P.deBruijn_base) 0%nat 1%nat q))))
    | ScopeExtrusionR : forall (p q:A.P.t),(ScopeExtrusion (A.P.New (A.P.Par p ((deBruijn.lift A.P.deBruijn_base) 0%nat 1%nat q))) (A.P.Par (A.P.New p) q)).

  Parameter ScopeExtrusion_sym : (symmetric A.P.t ScopeExtrusion).

  Parameter ScopeExtrusion_scopeL : forall (k:nat)(p q:A.P.t),(Rcongruence ScopeExtrusion (A.P.Par (A.P.mk_nu k p) q) (A.P.mk_nu k (A.P.Par p ((deBruijn.lift A.P.deBruijn_base) 0%nat k q)))).

  Parameter ParMonoidScopeExtrusion_scopeR : forall (k:nat)(p q:A.P.t),(Rcongruence (union A.P.t ParMonoid ScopeExtrusion) (A.P.Par p (A.P.mk_nu k q)) (A.P.mk_nu k (A.P.Par (deBruijn.lift A.P.deBruijn_base 0%nat k p) q))).

  Inductive NewSwap : relation A.P.t :=
    | NewSwapL : forall (p:A.P.t),(NewSwap (A.P.New (A.P.New p)) (A.P.New (A.P.New ((deBruijn.swapR A.P.deBruijn_base) 0%nat 1%nat p)))).

  Parameter NewSwap_sym : symmetric A.P.t NewSwap.

  Parameter NewSwap_mk_nu_swapR : forall (k:nat)(p:A.P.t),(Rcongruence NewSwap (A.P.New (A.P.mk_nu k p)) (A.P.mk_nu k (A.P.New ((deBruijn.swapR A.P.deBruijn_base) 0%nat k p)))).

  Parameter NewSwap_mk_nu_swapR_plus : forall (k:nat)(i:nat),(i<=k)->forall (p:A.P.t),(Rcongruence NewSwap (A.P.mk_nu (S k) p) (A.P.mk_nu (S k) (deBruijn.swapR A.P.deBruijn_base i (k-i) p))).

  Inductive NewAxioms : relation A.P.t :=
  | NewZeroL : NewAxioms A.P.Zero (A.P.New (A.P.Zero))
  | NewZeroR : NewAxioms (A.P.New (A.P.Zero)) A.P.Zero
  | NewSumL : forall (p q:A.P.t),(NewAxioms (A.P.Sum (A.P.New p) q) (A.P.New (A.P.Sum p ((deBruijn.lift A.P.deBruijn_base) 0%nat 1%nat q))))
  | NewSumR : forall (p q:A.P.t),(NewAxioms (A.P.New (A.P.Sum p ((deBruijn.lift A.P.deBruijn_base) 0%nat 1%nat q))) (A.P.Sum (A.P.New p) q)).

  Parameter NewAxioms_sym : (symmetric A.P.t NewAxioms).

  Inductive BangAxioms : relation A.P.t :=
  | BangExpL : forall (p:A.P.t),(BangAxioms (A.P.Bang p) (A.P.Par p (A.P.Bang p)))
  | BangExpR : forall (p:A.P.t),(BangAxioms (A.P.Par p (A.P.Bang p)) (A.P.Bang p)).

  Parameter BangAxioms_sym : (symmetric A.P.t BangAxioms).

  Definition R_preserves_lift (R:relation A.P.t) := forall (d k:nat),(R_preserves_f R ((deBruijn.lift A.P.deBruijn_base) d k)).
  
  Parameter Rcongruence_lift : forall (R:relation A.P.t),(R_preserves_lift R)->(R_preserves_lift (Rcongruence R)).
  Parameter SumMonoid_preserves_lift : (R_preserves_lift SumMonoid).
  Parameter ParMonoid_preserves_lift : (R_preserves_lift ParMonoid).
  Parameter ScopeExtrusion_preserves_lift : (R_preserves_lift ScopeExtrusion).
  Parameter NewSwap_preserves_lift : (R_preserves_lift NewSwap).
  Parameter NewAxioms_preserves_lift : (R_preserves_lift NewAxioms).
  Parameter BangAxioms_preserves_lift : (R_preserves_lift BangAxioms).

  Definition R_preserves_swapR (R:relation A.P.t) := forall (d k:nat),(R_preserves_f R ((deBruijn.swapR A.P.deBruijn_base) d k)).

  Parameter Rcongruence_swapR :  forall (R:relation A.P.t),(R_preserves_swapR R)->(R_preserves_swapR (Rcongruence R)).
  Parameter SumMonoid_preserves_swapR : (R_preserves_swapR SumMonoid).
  Parameter ParMonoid_preserves_swapR : (R_preserves_swapR ParMonoid).
  Parameter ScopeExtrusion_preserves_swapR : (R_preserves_swapR ScopeExtrusion).
  Parameter NewSwap_preserves_swapR : (R_preserves_swapR NewSwap).
  Parameter NewAxioms_preserves_swapR : (R_preserves_swapR NewAxioms).
  Parameter BangAxioms_preserves_swapR : (R_preserves_swapR BangAxioms).

  Definition R_preserves_low_subst (R:relation A.P.t) := forall (d k:nat)(es:list A.P.E.t),(R_preserves_f R (deBruijn.low_subst A.P.deBruijn_subst d k es)).

  Parameter Rcongruence_low_subst :  forall (R:relation A.P.t),(R_preserves_low_subst R)->(R_preserves_low_subst (Rcongruence R)).
  Parameter SumMonoid_preserves_low_subst : (R_preserves_low_subst SumMonoid).
  Parameter ParMonoid_preserves_low_subst : (R_preserves_low_subst ParMonoid).
  Parameter ScopeExtrusion_preserves_low_subst : (R_preserves_low_subst ScopeExtrusion).
  Parameter NewSwap_preserves_low_subst : (R_preserves_low_subst NewSwap).
  Parameter NewAxioms_preserves_low_subst : (R_preserves_low_subst NewAxioms).
  Parameter BangAxioms_preserves_low_subst : (R_preserves_low_subst BangAxioms).

  Parameter ParMonoidScopeExtrusionNewSwap_scopeL_swapL : forall (k:nat)(p q:A.P.t),(Rcongruence (union A.P.t (union A.P.t ParMonoid ScopeExtrusion) NewSwap) (A.P.mk_nu k (A.P.Par (A.P.New p) q)) (A.P.mk_nu k (A.P.New (A.P.Par (deBruijn.swapL A.P.deBruijn_base 0%nat k p) (deBruijn.lift A.P.deBruijn_base k 1 q))))).

  Parameter ParMonoidScopeExtrusion_sym : symmetric A.P.t (Rcongruence (union A.P.t ParMonoid ScopeExtrusion)).
  Parameter ParMonoidScopeExtrusionNewSwap_sym : symmetric A.P.t (Rcongruence (union A.P.t (union A.P.t ParMonoid ScopeExtrusion) NewSwap)).

  Parameter NewAxioms_NewGarbageL : forall (p:A.P.t),(Rcongruence (union _ (union A.P.t ParMonoid ScopeExtrusion) NewAxioms) p (A.P.New (deBruijn.lift A.P.deBruijn_base 0 1 p))).
  Parameter NewAxioms_NewGarbageR : forall (p:A.P.t),(Rcongruence (union A.P.t (union _ ParMonoid ScopeExtrusion) NewAxioms) (A.P.New (deBruijn.lift A.P.deBruijn_base 0 1 p)) p).

  Inductive StructuralCongruenceAxioms : relation A.P.t :=
  | SCSumMonoid : forall (p q:A.P.t),(SumMonoid p q)->(StructuralCongruenceAxioms p q)
  | SCParMonoid : forall (p q:A.P.t),(ParMonoid p q)->(StructuralCongruenceAxioms p q)
  | SCScopeExtrusion : forall (p q:A.P.t),(ScopeExtrusion p q)->(StructuralCongruenceAxioms p q)
  | SCNewSwap : forall (p q:A.P.t),(NewSwap p q)->(StructuralCongruenceAxioms p q)
  | SCNewAxioms : forall (p q:A.P.t),(NewAxioms p q)->(StructuralCongruenceAxioms p q)
  | SCBangAxioms : forall (p q:A.P.t),(BangAxioms p q)->(StructuralCongruenceAxioms p q).

  Parameter StructuralCongruenceAxioms_preserves_lift : (R_preserves_lift StructuralCongruenceAxioms).
  Parameter StructuralCongruenceAxioms_preserves_swapR : (R_preserves_swapR StructuralCongruenceAxioms).
  Parameter StructuralCongruenceAxioms_preserves_low_subst : (R_preserves_low_subst StructuralCongruenceAxioms).
  Parameter StructuralCongruence_symmetric : symmetric _ (Rcongruence StructuralCongruenceAxioms).

  Definition preserves_mem_fn (R:relation A.P.t) := forall (p q:A.P.t),(R p q)->forall (i d:nat),(deBruijn.mem_fn A.P.deBruijn_fn_base i d p)=(deBruijn.mem_fn A.P.deBruijn_fn_base i d q).

  Parameter Rcongruence_mem_fn : forall (R:relation A.P.t),(preserves_mem_fn R)->(preserves_mem_fn (Rcongruence R)).

  Parameter SumMonoid_preserves_mem_fn : preserves_mem_fn SumMonoid.
  Parameter ParMonoid_preserves_mem_fn : preserves_mem_fn ParMonoid.
  Parameter ScopeExtrusion_preserves_mem_fn : preserves_mem_fn ScopeExtrusion.
  Parameter NewSwap_preserves_mem_fn : preserves_mem_fn NewSwap.
  Parameter NewAxioms_preserves_mem_fn : preserves_mem_fn NewAxioms.
  Parameter BangAxioms_preserves_mem_fn : preserves_mem_fn BangAxioms.
  Parameter StructuralCongruenceAxioms_preserves_mem_fn : preserves_mem_fn StructuralCongruenceAxioms.

  Inductive Acongruence (R:relation A.P.t) : relation A.t :=
    | ACong : forall (p q:A.P.t),(Rcongruence R p q)->(Acongruence R (A.Pr p) (A.Pr q))
    | ACongTrans : forall (a b:A.t),(Acongruence R a b)->forall (c:A.t),(Acongruence R b c)->(Acongruence R a c)
    | AAbs : forall (p q:A.P.t),(Rcongruence R p q)->(Acongruence R (A.F (A.Abs p)) (A.F (A.Abs q)))
    | AConc : forall (p q:A.P.t),(Rcongruence R p q)->forall (e:A.P.E.t),(Acongruence R (A.C (A.Conc 0 e p)) (A.C (A.Conc 0 e q)))
    | ACongNew : forall (a b:A.t),(Acongruence R a b)->(Acongruence R (A.new a) (A.new b))
    | ANewSwapL : forall (a:A.t),(A.well_formed a)->(Acongruence R (A.new (A.new a)) (A.new (A.new (deBruijn.swapR A.deBruijn_base 0 1 a)))).

  Parameter Acongruence_increase : forall (R R':relation A.P.t),(inclusion A.P.t R R')->(inclusion A.t (Acongruence R) (Acongruence R')).

  Parameter Acongruence_Pr : forall (R:relation A.P.t),(inclusion A.P.t NewSwap R)->forall (a b:A.t),(Acongruence R a b)->forall (p:A.P.t),(a = A.Pr p)->(exists q:A.P.t,(b = A.Pr q)/\(Rcongruence R p q)).

  Implicit Arguments Acongruence_Pr [R a b].

  Parameter Acongruence_F : forall (R:relation A.P.t),(inclusion A.P.t NewSwap R)->(R_preserves_swapR (Rcongruence R))->forall (a b:A.t),(Acongruence R a b)->forall (p:A.P.t),(a = A.F (A.Abs p))->(exists q:A.P.t,(b = A.F (A.Abs q))/\(Rcongruence R p q)).

  Implicit Arguments Acongruence_F [R a b].

  Parameter Acongruence_C : forall (R:relation A.P.t),(inclusion A.P.t NewSwap R)->(inclusion A.P.t ScopeExtrusion R)->forall (a b:A.t),(Acongruence R a b)->forall (n:nat)(m:A.P.E.t)(p:A.P.t),(a = A.C (A.Conc n m p))->(exists m':A.P.E.t,exists q:A.P.t,b = A.C (A.Conc n m' q) /\ (forall (i:nat),(deBruijn.mem_fn A.P.E.deBruijn_fn_base i n m)=(deBruijn.mem_fn A.P.E.deBruijn_fn_base i n m')) /\ forall (k:nat)(f:A.abstraction),(Acongruence R (A.pseudo_appl_r (A.Conc (k+n) m p) f) (A.pseudo_appl_r (A.Conc (k+n) m' q) f))).

  Implicit Arguments Acongruence_C [R a b].

  Parameter AConcR : forall (R:relation A.P.t)(p q:A.P.t),(Rcongruence R p q)->forall (m:A.P.E.t),forall (n:nat),(forall (i:nat),(i<n)->deBruijn.mem_fn A.P.E.deBruijn_fn_base i 0 m = true)->(Acongruence R (A.C (A.Conc n m p)) (A.C (A.Conc n m q))).

  Parameter ACongRefl : forall (R:relation A.P.t)(a:A.t),(A.well_formed a)->(Acongruence R a a).
  Parameter ACongSym : forall (R:relation A.P.t),(symmetric A.P.t (Rcongruence R))->(symmetric A.t (Acongruence R)).

  Parameter AScopeExtrusionL_r : forall (R:relation A.P.t),(inclusion A.P.t ScopeExtrusion R)->forall (a:A.t)(q:A.P.t),(A.well_formed a)->(Acongruence R (A.par_r (A.new a) q) (A.new (A.par_r a (deBruijn.lift A.P.deBruijn_base 0 1 q)))).
  Parameter AScopeExtrusionL_l : forall (R:relation A.P.t),(inclusion A.P.t ScopeExtrusion R)->forall (p:A.P.t)(a:A.t),(A.well_formed a)->(Acongruence R (A.par_l (A.P.New p) a) (A.new (A.par_l p (deBruijn.lift A.deBruijn_base 0 1 a)))).
  Parameter AScopeExtrusionL_pal : forall (R:relation A.P.t),(inclusion A.P.t ScopeExtrusion R)->(inclusion A.P.t NewSwap R)->forall (f:A.abstraction)(n:nat)(m:A.P.E.t)(q:A.P.t),(Acongruence R (A.pseudo_appl_l (A.new_abs f) (A.Conc n m q)) (A.new (A.pseudo_appl_l f (A.Conc n (deBruijn.lift A.P.E.deBruijn_base n 1 m) (deBruijn.lift A.P.deBruijn_base n 1 q))))).
  Parameter AScopeExtrusionL_par : forall (R:relation A.P.t),(inclusion A.P.t ScopeExtrusion R)->(inclusion A.P.t NewSwap R)->(inclusion A.P.t ParMonoid R)->forall (c:A.concretion)(q:A.P.t),(Acongruence R (A.pseudo_appl_r (A.new_conc c) (A.Abs q)) (A.new (A.pseudo_appl_r c (A.Abs (deBruijn.lift A.P.deBruijn_base 1 1 q))))).
  Parameter AScopeExtrusionR_r : forall (R:relation A.P.t),(inclusion A.P.t ScopeExtrusion R)->forall (a:A.t)(q:A.P.t),(A.well_formed a)->(Acongruence R (A.new (A.par_r a (deBruijn.lift A.P.deBruijn_base 0 1 q))) (A.par_r (A.new a) q)).
  Parameter AScopeExtrusionR_l : forall (R:relation A.P.t),(inclusion A.P.t ScopeExtrusion R)->forall (p:A.P.t)(a:A.t),(A.well_formed a)->(Acongruence R (A.new (A.par_l p (deBruijn.lift A.deBruijn_base 0 1 a))) (A.par_l (A.P.New p) a)).
  Parameter AScopeExtrusionR_pal : forall (R:relation A.P.t),(inclusion A.P.t ScopeExtrusion R)->(inclusion A.P.t NewSwap R)->forall (f:A.abstraction)(n:nat)(m:A.P.E.t)(q:A.P.t),(Acongruence R (A.new (A.pseudo_appl_l f (A.Conc n (deBruijn.lift A.P.E.deBruijn_base n 1 m) (deBruijn.lift A.P.deBruijn_base n 1 q)))) (A.pseudo_appl_l (A.new_abs f) (A.Conc n m q))).
  Parameter AScopeExtrusionR_par : forall (R:relation A.P.t),(inclusion A.P.t ScopeExtrusion R)->(inclusion A.P.t NewSwap R)->(inclusion A.P.t ParMonoid R)->forall (c:A.concretion)(q:A.P.t),(Acongruence R (A.new (A.pseudo_appl_r c (A.Abs (deBruijn.lift A.P.deBruijn_base 1 1 q)))) (A.pseudo_appl_r (A.new_conc c) (A.Abs q))).

  Parameter AParZeroL : forall (R:relation A.P.t),(inclusion A.P.t ParMonoid R)->forall (a:A.t),(A.well_formed a)->(Acongruence R (A.par_l A.P.Zero a) a).
  Parameter AParZeroR : forall (R:relation A.P.t),(inclusion A.P.t ParMonoid R)->forall (a:A.t),(A.well_formed a)->(Acongruence R a (A.par_l A.P.Zero a)).

  Parameter AParComm_r_l : forall (R:relation A.P.t),(inclusion A.P.t ParMonoid R)->forall (a:A.t)(q:A.P.t),(A.well_formed a)->(Acongruence R (A.par_r a q) (A.par_l q a)).
  Parameter AParComm_l_r : forall (R:relation A.P.t),(inclusion A.P.t ParMonoid R)->forall (p:A.P.t)(a:A.t),(A.well_formed a)->(Acongruence R (A.par_l p a) (A.par_r a p)).

  Parameter AParAssocL_r_r : forall (R:relation A.P.t),(inclusion A.P.t ParMonoid R)->forall (a:A.t)(q r:A.P.t),(A.well_formed a)->(Acongruence R (A.par_r (A.par_r a q) r) (A.par_r a (A.P.Par q r))).
  Parameter AParAssocL_r_l : forall (R:relation A.P.t),(inclusion A.P.t ParMonoid R)->forall (p:A.P.t)(a:A.t)(q:A.P.t),(A.well_formed a)->(Acongruence R (A.par_r (A.par_l p a) q) (A.par_l p (A.par_r a q))).
  Parameter AParAssocL_l : forall (R:relation A.P.t),(inclusion A.P.t ParMonoid R)->forall (p q:A.P.t)(a:A.t),(A.well_formed a)->(Acongruence R (A.par_l (A.P.Par p q) a) (A.par_l p (A.par_l q a))).

  Parameter APseudoApplComm_l : forall (R:relation A.P.t),(inclusion A.P.t ParMonoid R)->forall (f:A.abstraction)(c:A.concretion),(Acongruence R (A.pseudo_appl_l f c) (A.pseudo_appl_r c f)).
  Parameter APseudoApplComm_r : forall (R:relation A.P.t),(inclusion A.P.t ParMonoid R)->forall (f:A.abstraction)(c:A.concretion),(Acongruence R (A.pseudo_appl_r c f) (A.pseudo_appl_l f c)).

  Parameter AParAssocL_r_pal : forall (R:relation A.P.t),(inclusion A.P.t ParMonoid R)->(inclusion A.P.t ScopeExtrusion R)->forall (f:A.abstraction)(n:nat)(m:A.P.E.t)(q r:A.P.t),(Acongruence R (A.par_r (A.pseudo_appl_l f (A.Conc n m q)) r) (A.pseudo_appl_l f (A.Conc n m (A.P.Par q (deBruijn.lift A.P.deBruijn_base 0 n r))))).
  Parameter AParAssocL_r_par : forall (R:relation A.P.t),(inclusion A.P.t ParMonoid R)->(inclusion A.P.t ScopeExtrusion R)->forall (c:A.concretion)(f:A.abstraction)(r:A.P.t),(Acongruence R (A.par_r (A.pseudo_appl_r c f) r) (A.pseudo_appl_r c (A.abs_par_r f r))).
  Parameter AParAssocL_pal_abs_r : forall (R:relation A.P.t),(inclusion A.P.t ParMonoid R)->forall (f:A.abstraction)(q:A.P.t)(c:A.concretion),(Acongruence R (A.pseudo_appl_l (A.abs_par_r f q) c) (A.pseudo_appl_l f (A.conc_par_l q c))).
  Parameter AParAssocL_pal_abs_l : forall (R:relation A.P.t),(inclusion A.P.t ParMonoid R)->(inclusion A.P.t ScopeExtrusion R)->forall (p:A.P.t)(f:A.abstraction)(c:A.concretion),(Acongruence R (A.pseudo_appl_l (A.abs_par_l p f) c) (A.par_l p (A.pseudo_appl_l f c))).
  Parameter AParAssocL_pal_conc_r : forall (R:relation A.P.t),(inclusion A.P.t ParMonoid R)->forall (c:A.concretion)(q:A.P.t)(f:A.abstraction),(Acongruence R (A.pseudo_appl_r (A.conc_par_r c q) f) (A.pseudo_appl_r c (A.abs_par_l q f))).
  Parameter AParAssocL_pal_conc_l : forall (R:relation A.P.t),(inclusion A.P.t ParMonoid R)->(inclusion A.P.t ScopeExtrusion R)->forall (p:A.P.t)(c:A.concretion)(f:A.abstraction),(Acongruence R (A.pseudo_appl_r (A.conc_par_l p c) f) (A.par_l p (A.pseudo_appl_r c f))).

  Parameter Acongruence_well_formed : forall (R:relation A.P.t)(a b:A.t),(Acongruence R a b)->(A.well_formed a)/\(A.well_formed b).

  Parameter ACongParL_par_r : forall (R:relation A.P.t),(inclusion A.P.t NewSwap R)->(inclusion A.P.t ScopeExtrusion R)->(R_preserves_swapR (Rcongruence R))->forall (a b:A.t),(Acongruence R a b)->forall (q:A.P.t),(Acongruence R (A.par_r a q) (A.par_r b q)).
  Parameter ACongParL_par_l : forall (R:relation A.P.t),(R_preserves_lift (Rcongruence R))->forall (p q:A.P.t),(Rcongruence R p q)->forall (a:A.t),(A.well_formed a)->(Acongruence R (A.par_l p a) (A.par_l q a)).
  Parameter ACongParR_par_r : forall (R:relation A.P.t),(inclusion A.P.t NewSwap R)->(inclusion A.P.t ScopeExtrusion R)->(inclusion _ ParMonoid R)->(R_preserves_lift (Rcongruence R))->(R_preserves_swapR (Rcongruence R))->forall (p q:A.P.t),(Rcongruence R p q)->forall (a:A.t),(A.well_formed a)->(Acongruence R (A.par_r a p) (A.par_r a q)).

  Parameter ACongParL_pseudo_appl_l : forall (R:relation A.P.t),(R_preserves_low_subst (Rcongruence R))->forall (p q:A.P.t),(Rcongruence R p q)->forall (c:A.concretion),(Acongruence R (A.pseudo_appl_l (A.Abs p) c) (A.pseudo_appl_l (A.Abs q) c)).

  Parameter ANewGarbageL : forall (R:relation A.P.t),(inclusion _ (union _ (union _ ParMonoid ScopeExtrusion) NewAxioms) R)->forall (a:A.t),(A.well_formed a)->(Acongruence R a (A.new (deBruijn.lift A.deBruijn_base 0 1 a))).
  Parameter ANewGarbageR : forall (R:relation A.P.t),(inclusion _ (union _ (union _ ParMonoid ScopeExtrusion) NewAxioms) R)->forall (a:A.t),(A.well_formed a)->(Acongruence R (A.new (deBruijn.lift A.deBruijn_base 0 1 a)) a).

  Parameter Acongruence_C_permutation : forall (R:relation A.P.t),(inclusion A.P.t NewSwap R)->(R_preserves_swapR (Rcongruence R))->(R_preserves_low_subst (Rcongruence R))->forall (a b:A.t),(Acongruence R a b)->forall (n:nat)(m:A.P.E.t)(p:A.P.t),(a = A.C (A.Conc n m p))->(exists m':A.P.E.t,exists q:A.P.t,b = A.C (A.Conc n m' q) /\ (forall (i:nat),(deBruijn.mem_fn A.P.E.deBruijn_fn_base i n m)=(deBruijn.mem_fn A.P.E.deBruijn_fn_base i n m')) /\ (forall (es:list A.P.E.t),(length es = n)->exists fs:list A.P.E.t,(is_permutation _ es fs) /\ (deBruijn.low_subst A.P.E.deBruijn_subst 0 0 es m = deBruijn.low_subst A.P.E.deBruijn_subst 0 0 fs m') /\ (Rcongruence R (deBruijn.low_subst A.P.deBruijn_subst 0 0 es p) (deBruijn.low_subst A.P.deBruijn_subst 0 0 fs q)))).
End CongruenceType.
