(* A formalisation of the spi calculus in Coq *)
(* Copyright (C) 2005-2007 Sebastien Briais *)
(* http://lamp.epfl.ch/~sbriais/ *)

(* This library is free software; you can redistribute it and/or modify *)
(* it under the terms of the GNU Lesser General Public License as *)
(* published by the Free Software Foundation; either version 2.1 of the *)
(* License, or (at your option) any later version. *)

(* This library is distributed in the hope that it will be useful, but *)
(* WITHOUT ANY WARRANTY; without even the implied warranty of *)
(* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU *)
(* Lesser General Public License for more details. *)

(* You should have received a copy of the GNU Lesser General Public *)
(* License along with this library; if not, write to the Free Software *)
(* Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA *)
(* 02110-1301 USA *)

Require Import FSets.
Require Import Arith.
Require Import List.
Require Import missing.
Require Import tactics.

Module OrderedNat <: OrderedType with Definition t := nat with Definition eq := @eq nat.
  Definition t := nat.
  Definition eq := @eq t.
  Definition lt (n m:t) := n < m.

  Lemma eq_refl : forall (x:t),(eq x x).
    unfold eq;reflexivity.
  Qed.

  Lemma eq_sym : forall (x y:t),(eq x y)->(eq y x).
    unfold eq;auto.
  Qed.
  
  Lemma eq_trans : forall (x y z:t),(eq x y)->(eq y z)->(eq x z).
    unfold eq;intros;congruence.
  Qed.

  Lemma lt_trans : forall (x y z:t),(lt x y)->(lt y z)->(lt x z).
    unfold lt.
    apply lt_trans.
  Qed.

  Lemma lt_not_eq : forall (x y:t),(lt x y)->~(eq x y).
    unfold lt, eq.
    intros.
    red;intro.
    elim (lt_irrefl x).
    rewrite <- H0 in H;trivial.
  Qed.
  
  Definition compare : forall (x y:t),(Compare lt eq x y).
    unfold lt, eq.
    intros.
    case (lt_eq_lt_dec x y);intro.
    case s;clear s;intro.
    apply LT;trivial.
    apply EQ;congruence.
    apply GT;trivial.
  Defined.

  Hint Immediate eq_sym.
  Hint Resolve eq_refl eq_trans lt_not_eq lt_trans.
End OrderedNat.

Module DeBruijnNat.
  Definition lift (d k:nat)(n:nat) := 
    match le_lt_dec d n with
      | left _ => (n+k)
      | right _ => n
    end.

  Definition swapR (d k:nat)(n:nat) :=
    match (le_lt_dec d n) with
      | left _ =>
	(match (lt_eq_lt_dec n (d+k)) with
           | inleft s =>
             (match s with
		| left _ => (S n)
		| right _ => d
              end)
           | _ => n
	 end)
      | right _ => n
    end.

  Definition low1 (d:nat)(n:nat) :=
    match (le_lt_dec d n) with
      | (left _) => (pred n)
      | (right _) => n
    end.

  (* let compute_Ns (k,n1,n2) = *)
  (*   if n1 < n2 then (n2-1,n1) *)
  (*   else (n2,n1+1) *)

  Definition occur_swap (n1 n2:nat) : nat * nat :=
    match (le_lt_dec n2 n1) with
      (* n2 <= n1 *)
      | left _ => (n2,S n1)
      (* n1 < n2 *)
      | right _ => (pred n2,n1)
    end.

  Definition low_subst_generic_aux (t:Set)(lift_t:nat->nat->t->t)(j:nat->t)(d k p:nat)(ts:list t)(H:length ts = p)(n:nat) : t :=
    match (le_lt_dec d n) with
      | (left _) =>
	(match (nth_error ts (n-d)) with
	   | (Some x) => lift_t 0%nat d x
	   | None => j (n-p+k)
	 end)
      | (right _) => j n
    end.

  Definition low_subst_generic (t:Set)(lift:nat->nat->t->t)(j:nat->t)(d k:nat)(ts:list t)(n:nat) : t :=
    low_subst_generic_aux t lift j d k (length ts) ts (refl_equal (length ts)) n.

  Definition mem_fn (k:nat)(d:nat)(n:nat) : bool :=
    match eq_nat_dec (d+k) n with
      | left _ => true
      | right _ => false
    end.

  Lemma low1_injective : forall (d:nat)(x y:nat),(mem_fn 0 d x = false)->(mem_fn 0 d y = false)->(low1 d x = low1 d y)->x=y.
    unfold low1, mem_fn.
    intros d x y.
    rewrite plus_0_r.
    case eq_nat_dec;intros H H0;try (discriminate H0).
    case eq_nat_dec;intros H1 H2;try (discriminate H2).
    repeat (case le_lt_dec;intro);try omega.
  Qed.

  Lemma lift_injective : forall (d k:nat),(injective (lift d k)).
    intros d k;unfold injective;intros x y;unfold lift;repeat (case le_lt_dec);intros;injection H;intros;try_absurd_arith;try congruence.
    assert (x=y);try omega;congruence.
  Qed.
  
  Lemma lift_0 : forall (d:nat)(x:nat),(lift d 0 x) = x.
    intros.
    unfold lift.
    case le_lt_dec;intro;try omega.
  Qed.
  
  Lemma lift_plus : forall (d:nat)(k k':nat)(x:nat),(lift d k (lift d k' x)) = (lift d (k+k') x).
    intros.
    unfold lift.
    repeat (case le_lt_dec;try_absurd_arith);intros;try omega.
  Qed.

  Lemma lift_comm : forall (d d' k k':nat)(x:nat),(lift (d+d'+k) k' (lift d k x))=(lift d k (lift (d+d') k' x)).
    intros.
    unfold lift.
    repeat (case le_lt_dec;try_absurd_arith);intros;try omega.
  Qed.

  Lemma lift_compose : forall (d d' k k':nat)(x:nat),(lift (d+d') k (lift d (d'+k') x))=(lift d (d'+k+k') x).
    intros.
    unfold lift.
    repeat (case le_lt_dec;try_absurd_arith);intros;try omega.
  Qed.

  Lemma swapR_injective : forall (d k:nat),(injective (swapR d k)).
    intros.
    unfold injective.
    intros x y.
    unfold swapR.
    repeat (case le_lt_dec || (case lt_eq_lt_dec;try (destruct s)));intros;try (injection H;intros);try_absurd_arith;try congruence.
  Qed.

  Lemma swapR_0 : forall (d:nat)(x:nat),(swapR d 0 x) = x.
    intros.
    unfold swapR.
    repeat (case le_lt_dec || (case lt_eq_lt_dec;try (intro;case s;clear s));trivial;try_absurd_arith);intros;try omega.
  Qed.

  Lemma swapR_comm : forall (d d' k k':nat)(x:nat),(swapR (d+k+(S d')) k' (swapR d k x))=(swapR d k (swapR (d+k+(S d')) k' x)).
    intros.
    unfold swapR.
    repeat (case le_lt_dec || (case lt_eq_lt_dec;try (intro;case s;clear s));trivial;try_absurd_arith);intros;try omega.
  Qed.

  Lemma swapR_1_idem : forall (d:nat)(x:nat),(swapR d 1 (swapR d 1 x))=x.
    intros.
    unfold swapR.
    repeat (case le_lt_dec || (case lt_eq_lt_dec;try (intro;case s;clear s));trivial;try_absurd_arith);intros;try omega.
  Qed.

  Lemma swapR_simpl : forall (d k k':nat)(x:nat),(swapR d k (swapR (d+k) k' x))=(swapR d (k+k') x).
    intros.
    unfold swapR.
    repeat (case le_lt_dec || (case lt_eq_lt_dec;try (intro;case s;clear s));trivial;try_absurd_arith);intros;try omega.
  Qed.

  Lemma swapR_compose : forall (d k k':nat)(x:nat),(swapR d (k+(S k')) (swapR d k x))=(swapR (S d) k (swapR d (k+(S k')) x)).
    intros.
    unfold swapR.
    repeat (case le_lt_dec || (case lt_eq_lt_dec;try (intro;case s;clear s));trivial;try_absurd_arith);intros;try omega.
  Qed.

  Lemma swapR_inside : forall (d d' k:nat)(x:nat),(swapR d (d'+S k) (swapR (d+d') k x))=(swapR (d+S d') k (swapR d (d'+S k) x)).
    intros.
    unfold swapR.
    repeat (case le_lt_dec || (case lt_eq_lt_dec;try (intro;case s;clear s));trivial;try_absurd_arith);intros;try omega.
  Qed.

  Lemma lift_swapR_comm : forall (d d' k k':nat)(x:nat),(lift (d+k+(S d')) k' (swapR d k x))=(swapR d k (lift (d+k+(S d')) k' x)).
    intros.
    unfold lift, swapR.
    repeat (case le_lt_dec || (case lt_eq_lt_dec;try (intro;case s;clear s));trivial;try_absurd_arith);intros;try omega.
  Qed.

  Lemma swapR_lift_comm : forall (d d' k k':nat)(x:nat),(swapR (d+k+d') k' (lift d k x))=(lift d k (swapR (d+d') k' x)).
    intros.
    unfold lift, swapR.
    repeat (case le_lt_dec || (case lt_eq_lt_dec;try (intro;case s;clear s));trivial;try_absurd_arith);intros;try omega.
  Qed.

  Lemma swapR_lift_idem : forall (d d' k k':nat)(x:nat),(swapR (d+d') k (lift d (d'+k+(S k')) x))=(lift d (d'+k+(S k')) x).
    intros.
    unfold lift, swapR.
    repeat (case le_lt_dec || (case lt_eq_lt_dec;try (intro;case s;clear s));trivial;try_absurd_arith);intros;try omega.
  Qed.

  Lemma swapR_lift_simpl : forall (d k:nat)(x:nat),(swapR d k (lift d k x))=(lift (S d) k x).
    intros.
    unfold lift, swapR.
    repeat (case le_lt_dec || (case lt_eq_lt_dec;try (intro;case s;clear s));trivial;try_absurd_arith);intros;try omega.
  Qed.

  Lemma swapR_lift_1 : forall (d k:nat)(x:nat),(swapR d k (lift (d+k) 1 x))=(lift d 1 x).
    intros.
    unfold lift, swapR.
    repeat (case le_lt_dec || (case lt_eq_lt_dec;try (intro;case s;clear s));trivial;try_absurd_arith);intros;try omega.
  Qed.

  Lemma low1_lift_lift : forall (d d' k:nat)(x:nat),(low1 (d+d') (lift d (d'+S k) x))=(lift d (d'+k) x).
    intros.
    unfold low1, lift.
    repeat (case le_lt_dec;try_absurd_arith);intros;try omega.    
  Qed.

  Lemma low1_lift_simpl : forall (d k:nat)(x:nat),(low1 d (lift d (S k) x))=(lift d k x).
    intros.
    generalize (low1_lift_lift d 0 k x);simpl;intros.
    rewrite plus_0_r in H;auto.
  Qed.

  Lemma swapR_occur_swap : forall (k:nat)(n1 n2:nat),(n1<=k)->(n2<=S k)->forall (n'1 n'2:nat),(occur_swap n1 n2 = (n'1,n'2))->forall (d:nat)(x:nat),(swapR (d+n2) (S k-n2) (swapR (d+n1) (k-n1) x))=(swapR (d+n'2) (S k-n'2) (swapR (d+n'1) (S (k-n'1)) x)).
    intros k n1 n2 H1 H2 n'1 n'2.
    unfold occur_swap.
    case le_lt_dec;intros;injection H;clear H;intros;subst n'1;subst n'2.
    replace (S k - S n1) with (k - n1);try omega.
    unfold swapR.
    replace (d+n1+(k-n1)) with (d+k);try omega.
    replace (d+n2+S (k-n2)) with (d+S k);try omega. 
    replace (d + S n1 + (k - n1)) with (d+S k);try omega.
    repeat (case le_lt_dec || (case lt_eq_lt_dec;try (intro;case s;clear s));trivial;try_absurd_arith);intros;try omega.
    destruct n2.
    inversion l.
    replace (pred (S n2)) with n2;try reflexivity.
    replace (S k - S n2) with (k - n2);try omega.
    unfold swapR.
    replace (d+n1+(k-n1)) with (d+k);try omega.
    replace (d+S n2+(k-n2)) with (d+S k);try omega. 
    replace (d+n2+S (k-n2)) with (d+S k);try omega. 
    replace (d+n1+(S k-n1)) with (d+S k);try omega.
    repeat (case le_lt_dec || (case lt_eq_lt_dec;try (intro;case s;clear s));trivial;try_absurd_arith);intros;try omega.
  Qed.

  Lemma lift_low1_comm : forall (d d' k:nat)(x:nat),(mem_fn 0 d x=false)->(low1 d (lift (d+S d') k x))=(lift (d+d') k (low1 d x)).
    unfold mem_fn, low1, lift.
    intros d d' k x.
    rewrite plus_0_r.
    case eq_nat_dec;intros;try (discriminate H).
    repeat (case le_lt_dec);intros;try_absurd_arith;try omega.
  Qed.

  Lemma lift_1_low1_simpl : forall (d k:nat)(x:nat),(mem_fn 0 (d+k) x = false)->(lift d 1 (low1 (d+k) x))=(swapR d k x).
    unfold mem_fn, low1, lift, swapR.
    intros d k x.
    rewrite plus_0_r.
    case eq_nat_dec;intros;try (discriminate H).
    repeat (case le_lt_dec);intros;try_absurd_arith;try omega.
    case lt_eq_lt_dec;intro;try (case s;clear s;intro);try_absurd_arith;try omega.
    case lt_eq_lt_dec;intro;try (case s;clear s;intro);try_absurd_arith;try omega.
  Qed.

  Lemma lift_fn : forall (d k:nat)(x:nat),forall (i:nat),(mem_fn (lift 0 k i) d (lift d k x))=(mem_fn i d x).
    unfold mem_fn, lift.
    intros.
    repeat (case le_lt_dec);intros;try_absurd_arith;repeat (case eq_nat_dec;intro);auto;try_absurd_arith.
  Qed.
    
  Lemma lift_In_fn : forall (d':nat)(i:nat),(i<=d')->forall (d k:nat)(x:nat),(mem_fn i d x = mem_fn i d (lift (d+S d') k x)).
    unfold mem_fn, lift;intros.
    case le_lt_dec;intro;repeat (case eq_nat_dec;intro);try_absurd_arith;trivial.
  Qed.
  
  Lemma lift_not_fn : forall (d k:nat)(i:nat),(i<k)->forall (x:nat),(mem_fn i d (lift d k x))=false.
    unfold mem_fn, lift;intros.
    case le_lt_dec;intro;case eq_nat_dec;intro;try_absurd_arith;auto.
  Qed.

  Lemma swapR_low1_comm : forall (d d' k:nat)(x:nat),(mem_fn 0 d x = false)->(low1 d (swapR (d+S d') k x))=(swapR (d+d') k (low1 d x)).
    unfold mem_fn, low1, swapR.
    intros d d' k x.
    rewrite plus_0_r.
    case eq_nat_dec;intros;try (discriminate H).
    repeat (case le_lt_dec || case lt_eq_lt_dec;try (intro;case s;clear s;intro));intros;try_absurd_arith;try omega.
  Qed.

  Lemma swapR_In_fn : forall (d':nat)(i:nat),(i<=d')->forall (d k:nat)(x:nat),(mem_fn i d x = mem_fn i d (swapR (d+S d') k x)).
    unfold mem_fn, swapR;intros.
    case lt_eq_lt_dec;intro;try (case s;clear s;intro);case le_lt_dec;intro;try_absurd_arith;repeat (case eq_nat_dec;intro);try_absurd_arith;auto.
  Qed.

  Lemma swapR_eq_cond : forall (d k:nat)(x:nat),(forall (i:nat),(i <= k)->(mem_fn i d x = false))->(swapR d k x) = x.
    intros.
    unfold swapR.
    case le_lt_dec;intro;trivial.
    case lt_eq_lt_dec;intro;try (case s;clear s;intro);auto.
    assert (mem_fn (x-d) d x = true).
    unfold mem_fn.
    rewrite <- le_plus_minus;try omega.
    case eq_nat_dec;intro;try_absurd_arith;auto.
    rewrite H in H0;try (discriminate H0).
    omega.
    assert (mem_fn (x-d) d x = true).
    unfold mem_fn.
    rewrite <- le_plus_minus;try omega.
    case eq_nat_dec;intro;try_absurd_arith;auto.
    rewrite H in H0;try (discriminate H0);try omega.
  Qed.
    
  Lemma swapR_fn : forall (d k:nat)(x:nat),forall (i:nat),(mem_fn (swapR 0 k i) d (swapR d k x))=(mem_fn i d x).
    unfold mem_fn, swapR.
    intros.
    repeat (case le_lt_dec || (case lt_eq_lt_dec;intro;try (case s;clear s;intro)));intros;try_absurd_arith;repeat (case eq_nat_dec;intro);auto;try_absurd_arith.
  Qed.

  Lemma swapR_fn_same : forall (d d' k:nat)(x:nat),forall (i:nat),(mem_fn i (d+k+S d') (swapR d k x))=(mem_fn i (d+k+S d') x).
    unfold mem_fn, swapR.
    intros.
    repeat (case le_lt_dec || (case lt_eq_lt_dec;intro;try (case s;clear s;intro)));intros;try_absurd_arith;repeat (case eq_nat_dec;intro);auto;try_absurd_arith.
  Qed.
    
  Lemma swapR_low1_comm2 : forall (d d' k:nat)(x:nat),(mem_fn 0 d x = false)->(swapR (d+d') k (low1 d x))=(low1 d (swapR (d+S d') k x)).
    unfold mem_fn, swapR, low1.
    intros d d' k x.
    rewrite plus_0_r.
    case eq_nat_dec;intros;try (discriminate H).
    repeat (case le_lt_dec || (case lt_eq_lt_dec;intro;try (case s;clear s;intro)));intros;try_absurd_arith;repeat (case eq_nat_dec;intro);auto;try_absurd_arith.
    try omega.
    try omega.
  Qed.

  Lemma low1_swapR_comm : forall (d d' k:nat)(x:nat),(mem_fn 0 (d+k+S d') x = false)->(low1 (d+k+(S d')) (swapR d k x))=(swapR d k (low1 (d+k+(S d')) x)).
    unfold mem_fn, swapR, low1.
    intros d d' k x.
    rewrite plus_0_r.
    case eq_nat_dec;intros;try (discriminate H).
    repeat (case le_lt_dec || (case lt_eq_lt_dec;intro;try (case s;clear s;intro)));intros;try_absurd_arith;repeat (case eq_nat_dec;intro);auto;try_absurd_arith.
  Qed.

  Lemma lift_low1_fn : forall (d:nat)(x:nat),(mem_fn 0 d x = false)->forall (i:nat),(mem_fn i d (low1 d x))=(mem_fn (lift 0 1 i) d x).
    unfold mem_fn, low1, lift.
    intros d x.
    rewrite plus_0_r.
    case eq_nat_dec;intros;try (discriminate H).
    repeat (case le_lt_dec);intros;try_absurd_arith;repeat (case eq_nat_dec;intro);auto;try_absurd_arith.
  Qed.

  Lemma low1_In_fn : forall (d':nat)(i:nat),(i<=d')->forall (d:nat)(x:nat),(mem_fn 0 (d+S d') x = false)->(mem_fn i d x = mem_fn i d (low1 (d+S d') x)).
    intros.
    unfold mem_fn, low1.
    repeat (case le_lt_dec);intros;try_absurd_arith;repeat (case eq_nat_dec;intro);auto;try_absurd_arith.
    rewrite <- H0.
    unfold mem_fn.
    rewrite plus_0_r.
    case eq_nat_dec;intro;try_absurd_arith.
    auto.
  Qed.

  Lemma low1_In_fn_cond : forall (d d':nat)(x:nat)(i:nat),(mem_fn (lift 0 1 i) (d+d') x = true)->(mem_fn i (d+d') (low1 d' x))=true.
    intros.
    unfold mem_fn, low1.
    repeat (case le_lt_dec);intros;try_absurd_arith;repeat (case eq_nat_dec;intro);auto;try_absurd_arith.
    rewrite <- H.
    unfold mem_fn, lift.
    repeat (case le_lt_dec);intros;try_absurd_arith;repeat (case eq_nat_dec;intro);auto;try_absurd_arith.
    rewrite <- H.
    unfold mem_fn, lift.
    repeat (case le_lt_dec);intros;try_absurd_arith;repeat (case eq_nat_dec;intro);auto;try_absurd_arith.
  Qed.

  Lemma fn_plus_lift : forall (d d':nat)(x:nat),forall (i:nat),(mem_fn (lift 0 d' i) d x) = (mem_fn i (d+d') x).
    intros.
    unfold mem_fn, lift.
    repeat (case le_lt_dec);intros;try_absurd_arith;repeat (case eq_nat_dec;intro);auto;try_absurd_arith.
  Qed.

  Lemma low1_comm : forall (d d':nat)(x:nat),(mem_fn 0 (d+S d') x = false)->(low1 d (low1 (d+S d') x))=(low1 (d+d') (low1 d x)).
    intros.
    unfold low1.
    repeat (case le_lt_dec);intros;try_absurd_arith;auto.
    unfold mem_fn in H.
    rewrite plus_0_r in H.
    destruct x.
    simpl in l0.
    reflexivity.
    simpl.
    simpl in l0.
    assert (x=d+S d').
    omega.
    assert (true = false).
    rewrite <- H.
    case eq_nat_dec;intro;try_absurd_arith.
    discriminate H1.
  Qed.

  Lemma low1_swapR_simpl : forall (d k:nat)(x:nat),(mem_fn 0 (d+k) x = false)->(low1 d (swapR d k x))=(low1 (d+k) x).
    intros.
    unfold low1, swapR.
    repeat (case le_lt_dec || (case lt_eq_lt_dec;intro;try (case s;clear s;intro)));intros;try_absurd_arith;repeat (case eq_nat_dec;intro);auto;try_absurd_arith.
    destruct x.
    inversion l0.
    reflexivity.
    simpl.
    unfold mem_fn in H.
    rewrite plus_0_r in H.
    assert (true=false).
    rewrite <- H.
    case eq_nat_dec;intro;try_absurd_arith.
    trivial.
    discriminate H0.
  Qed.

  Section low_subst_gen.
    Definition t:= nat.
    Hypothesis to:Set.
    Hypothesis lift_to : nat->nat->to->to.
    Hypothesis swapR_to : nat->nat->to->to.
    Hypothesis j : t -> to.

    Hypothesis j_lift : forall (d k:nat)(x:t),(lift_to d k (j  x))=(j (lift d k x)).
    Hypothesis j_swapR : forall (d k:nat)(x:t),(swapR_to d k (j x))=(j (swapR d k x)).
    Hypothesis lift_to_compose : forall (d d' k k':nat)(x:to),(lift_to (d+d') k (lift_to d (d'+k') x))=(lift_to d (d'+k+k') x).

    Definition low_subst_gen := low_subst_generic to lift_to j.

    Lemma low_subst_gen_lift_comm : forall (d d' k k':nat)(es:list to)(x:t),(low_subst_gen (d+d'+k) k' es (lift d k x))=(lift_to d k (low_subst_gen (d+d') k' es x)).
      intros d d' k k' es n.
      unfold low_subst_gen, lift, low_subst_generic, low_subst_generic_aux.
      generalize (List_nth_Some to es);intro Hnth.
      repeat (simpl;case le_lt_dec);intros;try_absurd_arith.
      rewrite Arith_plus_minus_plus.
      elim (Hnth (n-(d+d')));intros.
      case (le_lt_dec (length es) (n-(d+d')));intro.
      assert (nth_error es (n-(d+d')) = None).
      destruct (nth_error es (n-(d+d')));trivial.
      elim (lt_irrefl (length es)).
      eapply le_lt_trans.
      apply l2.
      apply H0.
      exists t0;trivial.
      rewrite H1.
      rewrite j_lift.
      simpl.
      unfold lift.
      case le_lt_dec;intro;try_absurd_arith.
      rewrite (le_plus_minus (length es) n);try omega.
      rewrite (plus_comm (length es) (n-length es)).
      rewrite Arith_plus_minus.
      rewrite <- plus_assoc.
      rewrite (plus_comm (length es)).
      rewrite plus_assoc.
      rewrite Arith_plus_minus.
      rewrite <- plus_assoc.
      rewrite (plus_comm k).
      rewrite plus_assoc;reflexivity.
      elim (H l2);intro x;intros.
      rewrite H1.
      generalize (lift_to_compose 0 d k d');simpl;intro.
      rewrite H2.
      rewrite <- plus_assoc.
      rewrite (plus_comm d').
      rewrite plus_assoc;trivial.
      rewrite j_lift.
      simpl;unfold lift;case le_lt_dec;intro;try_absurd_arith;trivial.
      rewrite j_lift.
      simpl;unfold lift;case le_lt_dec;intro;try_absurd_arith;trivial.
    Qed.

    Hypothesis lift_to_comm : forall (d d' k k':nat)(x:to),(lift_to (d+d'+k) k' (lift_to d k x))=(lift_to d k (lift_to (d+d') k' x)).

    Lemma lift_low_subst_gen_succ : forall (d d' k k':nat)(es:list to)(x:t),(lift_to (d+d') k' (low_subst_gen d (d'+k) es x))=(low_subst_gen d (k+k') (map (lift_to d' k') es) (lift (d+length es) d' x)).
      intros d d' k k' es n.
      generalize (List_nth_Some to es);intro.
      generalize (List_nth_map to to (lift_to d' k'));intro.
      unfold lift, low_subst_gen, low_subst_generic, low_subst_generic_aux.
      repeat (simpl;unfold lift;case le_lt_dec);intros;try_absurd_arith.
      assert (length es <= n-d);try omega.
      generalize (H (n-d)).
      destruct (nth_error es (n-d)).
      intros.
      elim H2;clear H2;intros.
      elim (lt_irrefl (length es)).
      eapply le_lt_trans.
      apply H1.
      apply H3.
      exists t0;trivial.
      intro H2;clear H2.
      rewrite H0.
      generalize (H (n+d'-d)).
      destruct (nth_error es (n+d'-d)).
      intros.
      elim H2;clear H2;intros.
      elim (lt_irrefl (length es)).
      apply le_lt_trans with (n+d'-d).
      omega.
      apply H3.
      exists t0;trivial.
      intro H2;clear H2.
      simpl.
      rewrite List_length_map.
      rewrite j_lift.
      simpl;unfold lift;case le_lt_dec;intro;try_absurd_arith.
      rewrite (le_plus_minus (length es) n);try omega.
      rewrite (plus_comm (length es)).
      rewrite Arith_plus_minus.
      rewrite <- (plus_assoc (n-length es) (length es)).
      rewrite (plus_comm (length es)).
      repeat (rewrite plus_assoc).
      rewrite Arith_plus_minus.
      trivial.
      assert (n-d < length es);try omega.
      elim (H (n-d));intros.
      elim (H2 H1);intro x;intro.
      rewrite H0.
      rewrite H4.
      simpl.
      generalize (lift_to_comm 0 d' d k');simpl;intro.
      rewrite <- H5.
      rewrite plus_comm;trivial.
      rewrite j_lift.
      simpl;unfold lift;case le_lt_dec;intro;try_absurd_arith;trivial.
    Qed.

    Lemma low_subst_gen_lift_nil : forall (d k k':nat)(es:list to)(x:t),(low_subst_gen d k es (lift d (k'+length es) x))=(lift_to d (k+k') (j x)).
      intros d k k' es n.
      unfold lift, low_subst_gen, low_subst_generic, low_subst_generic_aux.
      repeat (simpl;unfold lift;case le_lt_dec);intros;try_absurd_arith.
      generalize (List_nth_Some to es (n+(k'+length es)-d)).
      destruct (nth_error es (n+(k'+length es)-d)).
      intros.
      elim H;clear H;intros.
      assert (n+(k'+length es)-d < length es).
      apply H0.
      exists t0;trivial.
      try_absurd_arith.
      intro.
      rewrite plus_assoc.
      rewrite Arith_plus_minus.
      rewrite j_lift.
      simpl;unfold lift.
      case le_lt_dec;intro;try_absurd_arith.
      rewrite (plus_comm k k');rewrite plus_assoc;reflexivity.
      rewrite j_lift.
      simpl;unfold lift.
      case le_lt_dec;intro;try_absurd_arith.
      reflexivity.
    Qed.

    Lemma lift_low_subst_gen_comm : forall (d d' k k':nat)(es:list to)(x:t),(lift_to (d+d'+k) k' (low_subst_gen d' k es x))=(low_subst_gen d' k (map (lift_to (d+k) k') es) (lift (d'+d+length es) k' x)).
      intros d d' k k' es n.
      unfold lift, low_subst_gen, low_subst_generic, low_subst_generic_aux.
      repeat (simpl;unfold lift;case le_lt_dec);intros;try_absurd_arith.
      generalize (List_nth_Some to es (n-d')).
      destruct (nth_error es (n-d'));intros.
      assert (n-d' < length es).
      elim H;intros;apply H1.
      exists t0;trivial.
      try_absurd_arith.
      rewrite List_nth_map.
      generalize (List_nth_Some to es (n+k'-d')).
      destruct (nth_error es (n+k'-d')).
      intros.
      assert (n+k'-d' < length es).
      elim H0;intros;apply H2.
      exists t0;trivial.
      try_absurd_arith.
      intros;simpl.
      rewrite List_length_map.
      rewrite j_lift.
      simpl;unfold lift.
      case le_lt_dec;intro;try_absurd_arith.
      rewrite (le_plus_minus (length es) n);try omega.
      rewrite (plus_comm (length es)).
      rewrite Arith_plus_minus.
      rewrite <- (plus_assoc (n-length es) (length es));rewrite (plus_comm (length es));rewrite plus_assoc.
      rewrite Arith_plus_minus.
      rewrite <- plus_assoc;rewrite (plus_comm k);rewrite plus_assoc;reflexivity.
      rewrite List_nth_map.
      generalize (List_nth_Some to es (n-d')).
      destruct (nth_error es (n-d')).
      intros.
      assert (n-d' < length es).
      elim H;intros;apply H1.
      exists t0;trivial.
      simpl.
      generalize (lift_to_comm 0 (d+k) d' k').
      simpl.
      intros.
      rewrite (plus_comm d d').
      rewrite <- plus_assoc.
      rewrite (plus_comm d' (d+k)).
      apply H1.
      intros.
      simpl.
      rewrite List_length_map.
      rewrite j_lift.
      assert (length es <= n-d').
      case (le_lt_dec (length es) (n-d'));auto.
      intro.
      elim H;intros.
      elim (H0 l2);intros.
      discriminate H2.
      simpl;unfold lift.
      case le_lt_dec;intro;try_absurd_arith.
      reflexivity.
      rewrite j_lift.
      simpl;unfold lift.
      case le_lt_dec;intro;try_absurd_arith.
      reflexivity.
    Qed.

    Lemma low_subst_gen_lift_succ_simpl :  forall (d k k':nat)(es:list to)(x:t),(low_subst_gen d k es (lift (d+length es) k' x))=(low_subst_gen d (k+k') es x).
      intros d k k' es n.
      unfold lift, low_subst_gen, low_subst_generic, low_subst_generic_aux.
      simpl;repeat ((case le_lt_dec || (case lt_eq_lt_dec;try (intro;case s;clear s;intro))));intros;try_absurd_arith;try (rewrite List_nth_map).
      elim (List_nth_Some to es (n-d));intros;simpl.
      destruct (nth_error es (n-d)).
      elim (lt_irrefl (n-d)).
      eapply lt_le_trans.
      apply H0.
      exists t0;trivial.
      omega.
      elim (List_nth_Some to es (n+k'-d));intros;simpl.
      destruct (nth_error es (n+k'-d)).
      elim (lt_irrefl (n+k'-d)).
      eapply lt_le_trans.
      apply H2.
      exists t0;trivial.
      omega.
      rewrite (le_plus_minus (length es) n);try omega.
      rewrite (plus_comm (length es) (n-length es)).
      rewrite Arith_plus_minus.
      rewrite <- plus_assoc.
      rewrite (plus_comm (length es) k').
      rewrite plus_assoc.
      rewrite Arith_plus_minus.
      rewrite (plus_comm k k').
      rewrite plus_assoc;reflexivity.
      elim (List_nth_Some to es (n-d)).
      destruct (nth_error es (n-d));intros;trivial.
      assert (n-d < length es);try omega.
      elim (H H1).
      intros.
      discriminate H2.
      reflexivity.
    Qed.

    Hypothesis swapR_lift_to_idem : forall (d d' k k':nat)(x:to),(swapR_to (d+d') k (lift_to d (d'+k+(S k')) x))=(lift_to d (d'+k+(S k')) x).

    Lemma low_subst_gen_swapR_comm : forall (d d' k k':nat)(es:list to)(x:t),(low_subst_gen (d+k+(S d')) k' es (swapR d k x))=(swapR_to d k (low_subst_gen (d+k+(S d')) k' es x)).
      intros d d' k k' es n.
      unfold lift, swapR, low_subst_gen, low_subst_generic, low_subst_generic_aux.
      repeat (simpl;(case le_lt_dec || (case lt_eq_lt_dec;try (intro;case s;clear s;intro))));intros;try_absurd_arith;auto.
      generalize (List_nth_Some to es (n-(d+k+S d'))).
      destruct (nth_error es (n-(d+k+S d')));intro.
      generalize (swapR_lift_to_idem 0%nat d k d');simpl;intro.
      rewrite H0.
      trivial.
      rewrite j_swapR.
      elim H;intros.
      case (le_lt_dec (length es) (n-(d+k+S d')));intro.
      simpl;unfold swapR;repeat (simpl;(case le_lt_dec || (case lt_eq_lt_dec;try (intro;case s;clear s;intro))));intros;try_absurd_arith;auto.
      elim (H0 l3);intro x;intros.
      discriminate H2.
      rewrite j_swapR.
      simpl;unfold swapR;repeat (simpl;(case le_lt_dec || (case lt_eq_lt_dec;try (intro;case s;clear s;intro))));intros;try_absurd_arith;auto.
      rewrite j_swapR.
      simpl;unfold swapR;repeat (simpl;(case le_lt_dec || (case lt_eq_lt_dec;try (intro;case s;clear s;intro))));intros;try_absurd_arith;auto.
      rewrite j_swapR.
      simpl;unfold swapR;repeat (simpl;(case le_lt_dec || (case lt_eq_lt_dec;try (intro;case s;clear s;intro))));intros;try_absurd_arith;auto.
      rewrite j_swapR.
      simpl;unfold swapR;repeat (simpl;(case le_lt_dec || (case lt_eq_lt_dec;try (intro;case s;clear s;intro))));intros;try_absurd_arith;auto.
    Qed.

    Hypothesis swapR_lift_to_comm : forall (d d' k k':nat)(x:to),(swapR_to (d+k+d') k' (lift_to d k x))=(lift_to d k (swapR_to (d+d') k' x)).

    Lemma swapR_low_subst_gen_idem : forall (d k k':nat)(es:list to)(x:t),(swapR_to d k (low_subst_gen d (k+S k') es x))=(low_subst_gen d (k+S k') (map (swapR_to 0 k) es) x).
      intros d k k' es n.
      unfold lift, swapR, low_subst_gen, low_subst_generic, low_subst_generic_aux.
      repeat (simpl;unfold lift, swapR;(case le_lt_dec || (case lt_eq_lt_dec;try (intro;case s;clear s;intro))));intros;try_absurd_arith;auto.
      generalize (List_nth_map to to (swapR_to 0%nat k));intro.
      rewrite H.
      generalize (List_nth_Some to es (n-d)).
      case (le_lt_dec (length es) (n-d));intro.
      destruct (nth_error es (n-d)).
      intro.
      elim H0;clear H0;intros.
      elim (lt_irrefl (length es)).
      eapply le_lt_trans.
      apply l0.
      apply H1.
      exists t0;trivial.
      intro.
      simpl.
      rewrite List_length_map.
      rewrite j_swapR.
      repeat (simpl;unfold lift, swapR;(case le_lt_dec || (case lt_eq_lt_dec;try (intro;case s;clear s;intro))));intros;try_absurd_arith;auto.
      intro.
      elim H0;clear H0;intros.
      elim (H0 l0);intro x;intros.
      rewrite H2.
      simpl.
      generalize (swapR_lift_to_comm 0%nat 0%nat d k);simpl;rewrite plus_0_r.
      auto.
      rewrite j_swapR.
      simpl;unfold swapR;case le_lt_dec;intro;try_absurd_arith;trivial.
    Qed.

    Hypothesis swapR_lift_to_1 : forall (d k:nat)(x:to),(swapR_to d k (lift_to (d+k) 1 x))=(lift_to d 1 x).
    Hypothesis lift_to_plus : forall (d:nat)(k k':nat)(x:to),(lift_to d k (lift_to d k' x)) = (lift_to d (k+k') x).

    Lemma low_subst_gen_swapR_simpl : forall (d k:nat)(es:list to)(x:t),(low_subst_gen (S d) k es (swapR d (length es) x))=(swapR_to d k (low_subst_gen d k (map (lift_to k 1) es) x)).
      intros d k es n.
      unfold lift, swapR, low_subst_gen, low_subst_generic, low_subst_generic_aux.
      repeat ((case le_lt_dec || (case lt_eq_lt_dec;try (intro;case s;clear s;intro))));intros;try_absurd_arith;auto;try (rewrite List_nth_map).
      elim (List_nth_Some to es (n-d));intros;simpl.
      assert (n-d < length es);try omega.
      elim (H H1).
      intros.
      rewrite H2;unfold option_map.
      generalize (swapR_lift_to_comm 0 0 d k);simpl.
      rewrite plus_0_r;intro.
      simpl.
      rewrite H3.
      generalize (swapR_lift_to_1 0 k);simpl;intro.
      rewrite H4.
      rewrite (lift_to_plus);rewrite plus_comm;reflexivity.
      elim (List_nth_Some to es (n-d)).
      destruct (nth_error es (n-d));intros.
      assert (n-d < length es);try_absurd_arith.
      apply H0.
      exists t0;trivial.
      unfold option_map.
      rewrite List_length_map.
      rewrite j_swapR.
      elim (List_nth_Some to es (n-S d)).
      destruct (nth_error es (n-S d)).
      intros.
      elim (lt_irrefl (n-S d)).
      eapply lt_le_trans.
      apply H2.
      exists t0;trivial.
      omega.
      intros.
      simpl;unfold swapR.
      case le_lt_dec;intro;try_absurd_arith.
      case lt_eq_lt_dec;intro;try (case s;clear s;intro);try_absurd_arith;trivial.
      elim (List_nth_Some to es (n-d)).
      destruct (nth_error es (n-d));intros.
      assert (n - d < length es);try_absurd_arith.
      apply H0.
      exists t0;trivial.
      unfold option_map.
      rewrite j_swapR.
      elim (List_nth_Some to es (n-S d)).
      destruct (nth_error es (n-S d));intros.
      rewrite List_length_map.
      simpl;unfold swapR.
      case le_lt_dec;intro;try_absurd_arith.
      case lt_eq_lt_dec;intro;try (case s;clear s;intro);try_absurd_arith;auto.
      rewrite List_length_map.
      simpl.
      unfold swapR.
      case le_lt_dec;intro;try_absurd_arith.
      case lt_eq_lt_dec;intro;try (case s;clear s;intro);try_absurd_arith;auto.
      rewrite j_swapR.
      simpl;unfold swapR.
      case le_lt_dec;intro;try_absurd_arith;auto.
    Qed.

    Lemma swapR_low_subst_gen_comm : forall (d d' k k':nat)(es:list to)(x:t),(swapR_to (d+d'+k) k' (low_subst_gen d' k es x))=(low_subst_gen d' k (map (swapR_to (d+k) k') es) (swapR (d'+d+length es) k' x)).
      intros d d' k k' es n.
      unfold lift, swapR, low_subst_gen, low_subst_generic, low_subst_generic_aux.
      repeat ((case le_lt_dec || (case lt_eq_lt_dec;try (intro;case s;clear s;intro))));intros;try_absurd_arith;try (rewrite List_nth_map);try (rewrite List_length_map).
      elim (List_nth_Some to es (n-d')).
      destruct (nth_error es (n-d'));intros.
      elim (lt_irrefl (n-d')).
      eapply lt_le_trans.
      apply H0.
      exists t0;trivial.
      omega.
      elim (List_nth_Some to es (S n -d')).
      destruct (nth_error es (S n-d'));intros.
      elim (lt_irrefl (S n-d')).
      eapply lt_le_trans.
      apply H2.
      exists t0;trivial.
      omega.
      unfold option_map.
      rewrite j_swapR.
      replace (S n - length es + k) with (S (n-length es+k));try omega.
      simpl;unfold swapR.
      case le_lt_dec;intro;try_absurd_arith.
      case lt_eq_lt_dec;intro;try (case s;clear s;intro);try_absurd_arith.
      trivial.
      elim (List_nth_Some to es (n-d')).
      destruct (nth_error es (n-d'));intros.
      elim (lt_irrefl (n-d')).
      eapply lt_le_trans.
      apply H0.
      exists t0;trivial.
      omega.
      elim (List_nth_Some to es (d'+d+length es-d')).
      destruct (nth_error es (d'+d+length es-d'));intros.
      elim (lt_irrefl (d'+d+length es-d')).
      eapply lt_le_trans.
      apply H2.
      exists t0;trivial.
      omega.
      unfold option_map.
      subst n.
      repeat (rewrite <- plus_assoc).
      rewrite (plus_comm (length es) k').
      repeat (rewrite plus_assoc).
      repeat (rewrite Arith_plus_minus).
      rewrite j_swapR.
      simpl;unfold swapR.
      case le_lt_dec;intro;try_absurd_arith.
      case lt_eq_lt_dec;intro;try (case s;clear s;intro);try_absurd_arith.
      repeat (rewrite <- plus_assoc).
      rewrite plus_comm.
      rewrite (plus_comm d k).
      rewrite plus_assoc;trivial.
      elim (List_nth_Some to es (n-d')).
      destruct (nth_error es (n-d'));intros.
      unfold option_map.
      generalize (swapR_lift_to_comm 0 (d+k) d' k');simpl.
      rewrite plus_assoc.
      rewrite (plus_comm d' d);auto.
      unfold option_map.
      rewrite j_swapR.
      simpl;unfold swapR.
      case le_lt_dec;intro;try_absurd_arith.
      case lt_eq_lt_dec;intro;try (case s;clear s;intro);try_absurd_arith.
      trivial.
      elim (List_nth_Some to es (n-d')).
      destruct (nth_error es (n-d'));intros;unfold option_map.
      generalize (swapR_lift_to_comm 0 (d+k) d' k');simpl. 
      rewrite plus_assoc.
      rewrite (plus_comm d' d);auto.
      rewrite j_swapR.
      simpl;unfold swapR.
      case le_lt_dec;intro;try_absurd_arith.
      case lt_eq_lt_dec;intro;try (case s;clear s;intro);try_absurd_arith.
      elim H;try omega.
      intros.
      discriminate H1.
      rewrite e.
      elim H;try omega.
      intros.
      discriminate H1.
      trivial.
      rewrite j_swapR.
      simpl;unfold swapR.
      case le_lt_dec;intro;try_absurd_arith.
      trivial.
    Qed.

    Lemma swapR_low_subst_gen_comm_idem : forall (d d' k k':nat)(es:list to)(x:t),(swapR_to (d+d') k' (low_subst_gen d (d'+k+S k') es x))=(low_subst_gen d (d'+k+S k') (map (swapR_to d' k') es) x).
      intros d d' k k' es n.
      unfold lift, swapR, low_subst_gen, low_subst_generic, low_subst_generic_aux.
      simpl;repeat ((case le_lt_dec || case lt_eq_lt_dec);intro;try (case s;intro;clear s);trivial;try_absurd_arith);try (rewrite List_nth_map);try (rewrite List_length_map).
      elim (List_nth_Some to es (n-d)).
      destruct (nth_error es (n-d));unfold option_map;try (rewrite j_swapR);intros.
      generalize (swapR_lift_to_comm 0 d' d k');simpl;auto.
      assert (length es <= n-d).
      case (le_lt_dec (length es) (n-d));intro;auto.
      elim H;auto.
      intros.
      discriminate H1.
      simpl;unfold swapR.
      case le_lt_dec;intro;try_absurd_arith.
      case lt_eq_lt_dec;intro;try (case s;clear s;intro);try_absurd_arith.
      auto.
      rewrite j_swapR.
      simpl;unfold swapR.
      case le_lt_dec;intro;try_absurd_arith.
      trivial.
    Qed.

    Lemma low_subst_swapR_rotate : forall (d k:nat)(e:to)(es:list to)(x:t),(low_subst_gen d k (List_right_rotate (cons e es)) (swapR d (length es) x))=(low_subst_gen d k (cons e es) x).
      intros.
      unfold swapR, low_subst_gen, low_subst_generic, low_subst_generic_aux.
      repeat (rewrite <- List_length_right_rotate).
      case (le_lt_dec d x);intro.
      case lt_eq_lt_dec;intro;try (case s;clear s;intro);try_absurd_arith.
      case le_lt_dec;intro;try_absurd_arith.
      replace (S x - d) with (S (x-d));try omega.
      replace (length (e::es)) with (S (length es));try reflexivity.
      assert (S (x-d) < length (cons e es)).
      replace (length (e::es)) with (S (length es));try reflexivity.
      omega.
      rewrite List_right_rotate_nth_Some;try omega.
      unfold pred.
      elim (List_nth_Some _  (cons e es) (x-d));intros.
      assert (x-d  < length (e :: es));try omega.
      elim (H0 H2).
      intro ei.
      intro.
      rewrite H3;auto.
      case le_lt_dec;intro;try_absurd_arith.
      rewrite <- minus_n_n.
      rewrite List_right_rotate_nth_O.
      subst x.
      rewrite minus_plus.
      elim (List_nth_Some _  (cons e es) (length es));intros.
      replace (length (e::es)) with (S (length es));try reflexivity.
      unfold pred.
      elim H.
      intro ei;intros.
      rewrite H1;auto.
      simpl;omega.
      red;intro.
      discriminate H.
      case le_lt_dec;intro;try_absurd_arith.
      rewrite List_nth_None.
      rewrite List_nth_None.
      auto.
      simpl.
      omega.
      rewrite <- List_length_right_rotate.
      simpl;omega.
      case le_lt_dec;intro;try_absurd_arith.
      auto.
    Qed.
  End low_subst_gen.
End DeBruijnNat.
