(* A formalisation of the spi calculus in Coq *)
(* Copyright (C) 2005-2007 Sebastien Briais *)
(* http://lamp.epfl.ch/~sbriais/ *)

(* This library is free software; you can redistribute it and/or modify *)
(* it under the terms of the GNU Lesser General Public License as *)
(* published by the Free Software Foundation; either version 2.1 of the *)
(* License, or (at your option) any later version. *)

(* This library is distributed in the hope that it will be useful, but *)
(* WITHOUT ANY WARRANTY; without even the implied warranty of *)
(* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU *)
(* Lesser General Public License for more details. *)

(* You should have received a copy of the GNU Lesser General Public *)
(* License along with this library; if not, write to the Free Software *)
(* Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA *)
(* 02110-1301 USA *)

Require Import Arith.
Require Import name.
Require Import messageSpi.
Require Import expressionSpi.
Require Import formulaeSpi.
Require Import FSets.
Require Import missing.
Require Import deBruijnType.
Require Import tactics.
Require Import opSpi.

Module EvalSpi.
  Import OpSpi.

  Fixpoint expr_of_mesg (m:MessageSpi.t) : ExpressionSpi.t :=
    match m with
      | (MessageSpi.MNam n) => (ExpressionSpi.ENam n)
      | (MessageSpi.MPair m n) => (ExpressionSpi.EPair (expr_of_mesg m) (expr_of_mesg n))
      | (MessageSpi.MEnc_s m n) => (ExpressionSpi.EEnc_s (expr_of_mesg m) (expr_of_mesg n))
      | (MessageSpi.MEnc_a m n) => (ExpressionSpi.EEnc_a (expr_of_mesg m) (expr_of_mesg n))
      | (MessageSpi.MOp o m) => (ExpressionSpi.EOp o (expr_of_mesg m))
    end.

  Lemma expr_of_mesg_injective : injective expr_of_mesg.
    unfold injective.
    induction x;simpl;intros.
    destruct y;try (discriminate H).
    simpl in H.
    injection H;intros.
    subst t0;auto.
    destruct y;try (discriminate H).
    simpl in H.
    injection H;intros.
    rewrite (IHx1 y1);auto.
    rewrite (IHx2 y2);auto.
    destruct y;try (discriminate H).
    simpl in H.
    injection H;intros.
    rewrite (IHx1 y1);auto.
    rewrite (IHx2 y2);auto.
    destruct y;try (discriminate H).
    simpl in H.
    injection H;intros.
    rewrite (IHx1 y1);auto.
    rewrite (IHx2 y2);auto.
    destruct y;try (discriminate H).
    simpl in H.
    injection H;intros.
    subst o0.
    rewrite (IHx y);auto.
  Qed.

  Lemma mem_fn_expr_of_msg : forall (d k:nat)(m:MessageSpi.t),(deBruijn.mem_fn ExpressionSpi.deBruijn_fn_base k d (expr_of_mesg m))=(deBruijn.mem_fn MessageSpi.deBruijn_fn_base k d m).
    induction m;lazy beta delta [expr_of_mesg] iota;fold expr_of_mesg;intros.
    reflexivity.
    rewrite ExpressionSpi.mem_fn_EPair.
    rewrite MessageSpi.mem_fn_MPair.
    rewrite IHm1;rewrite IHm2;auto.
    rewrite ExpressionSpi.mem_fn_EEnc_s.
    rewrite MessageSpi.mem_fn_MEnc_s.
    rewrite IHm1;rewrite IHm2;auto.
    rewrite ExpressionSpi.mem_fn_EEnc_a.
    rewrite MessageSpi.mem_fn_MEnc_a.
    rewrite IHm1;rewrite IHm2;auto.
    rewrite ExpressionSpi.mem_fn_EOp.
    rewrite MessageSpi.mem_fn_MOp.
    rewrite IHm;auto.
  Qed.

  Lemma map_expr_of_msg : forall (f:nat->Name.t->Name.t)(d:nat)(m:MessageSpi.t),(ExpressionSpi.t_map f d (expr_of_mesg m))=(expr_of_mesg (MessageSpi.t_map f d m)).
    induction m;simpl;repeat (rewrite IHm1 || rewrite IHm2 || rewrite IHm);try reflexivity.
  Qed.

  (* concrete evaluation *)
  Fixpoint evalc (e:ExpressionSpi.t) : option MessageSpi.t :=
    match e with
      | (ExpressionSpi.ENam n) => Some (MessageSpi.MNam n)
      | (ExpressionSpi.EPair e f) =>
	(match ((evalc e),(evalc f)) with
	   | ((Some m),(Some n)) => Some (MessageSpi.MPair m n)
	   | _ => None
	 end)
      | (ExpressionSpi.EEnc_s e f) =>
	(match ((evalc e),(evalc f)) with
	   | ((Some m),(Some n)) => Some (MessageSpi.MEnc_s m n)
	   | _ => None
	 end)
      | (ExpressionSpi.EEnc_a e f) =>
	(match ((evalc e),(evalc f)) with
	   | ((Some m),(Some n)) => Some (MessageSpi.MEnc_a m n)
	   | _ => None
	 end)
      | (ExpressionSpi.EFst e) =>
	(match (evalc e) with
	   | (Some (MessageSpi.MPair m _)) => Some m
	   | _ => None
	 end)
      | (ExpressionSpi.ESnd e) =>
	(match (evalc e) with
	   | (Some (MessageSpi.MPair _ m)) => Some m
	   | _ => None
	 end)
      | (ExpressionSpi.EDec_s e f) =>
	(match ((evalc e),(evalc f)) with
	   | (Some (MessageSpi.MEnc_s m n),Some k) =>
	     (match MessageSpi.Ord.compare n k with
		| EQ _ => Some m
		| _ => None
	      end)
	   | _ => None
	 end)
      | (ExpressionSpi.EDec_a e f) =>
	(match ((evalc e),(evalc f)) with
	   | (Some (MessageSpi.MEnc_a m n),Some k) =>
	     (match MessageSpi.inverse_key n with 
		| Some k' => (match MessageSpi.Ord.compare k' k with
				| EQ _ => Some m
				| _ => None
			      end)
		| _ => None
	      end)
	   | _ => None
	 end)
      | (ExpressionSpi.EOp o e) =>
	(match (evalc e) with
	   | Some m => Some (MessageSpi.MOp o m)
	   | _ => None
	 end)
    end.

  Lemma evalc_EPair : forall (e1 e2:ExpressionSpi.t),(evalc (ExpressionSpi.EPair e1 e2))=(match (evalc e1) with Some m1 => match (evalc e2) with Some m2 => Some (MessageSpi.MPair m1 m2) | _ => None end | _ => None end).
    intros.
    reflexivity.
  Qed.

  Lemma evalc_EEnc_s : forall (e1 e2:ExpressionSpi.t),(evalc (ExpressionSpi.EEnc_s e1 e2))=(match (evalc e1) with Some m1 => match (evalc e2) with Some m2 => Some (MessageSpi.MEnc_s m1 m2) | _ => None end | _ => None end).
    intros.
    reflexivity.
  Qed.

  Lemma evalc_EEnc_a : forall (e1 e2:ExpressionSpi.t),(evalc (ExpressionSpi.EEnc_a e1 e2))=(match (evalc e1) with Some m1 => match (evalc e2) with Some m2 => Some (MessageSpi.MEnc_a m1 m2) | _ => None end | _ => None end).
    intros.
    reflexivity.
  Qed.

  Lemma evalc_EFst : forall (e:ExpressionSpi.t),(evalc (ExpressionSpi.EFst e))=(match (evalc e) with Some (MessageSpi.MPair m1 m2) => Some m1 | _ => None end).
    intros.
    reflexivity.
  Qed.

  Lemma evalc_ESnd : forall (e:ExpressionSpi.t),(evalc (ExpressionSpi.ESnd e))=(match (evalc e) with Some (MessageSpi.MPair m1 m2) => Some m2 | _ => None end).
    intros.
    reflexivity.
  Qed.

  Lemma evalc_EDec_s : forall (e1 e2:ExpressionSpi.t),(evalc (ExpressionSpi.EDec_s e1 e2))=(match (evalc e1) with Some (MessageSpi.MEnc_s m1 m2) => match (evalc e2) with Some m3 => match MessageSpi.Ord.compare m2 m3 with EQ _ => Some m1 | _ => None end | _ => None end | _ => None end).
    intros.
    reflexivity.
  Qed.

  Lemma evalc_EDec_a : forall (e1 e2:ExpressionSpi.t),(evalc (ExpressionSpi.EDec_a e1 e2))=(match (evalc e1) with Some (MessageSpi.MEnc_a m1 m2) => match (evalc e2) with Some m3 => match (MessageSpi.inverse_key m2) with Some m4 => match MessageSpi.Ord.compare m4 m3 with EQ _ => Some m1 | _ => None end | _ => None end | _ => None end | _ => None end).
    intros.
    reflexivity.
  Qed.

  Lemma evalc_EOp : forall (o:op)(e:ExpressionSpi.t),(evalc (ExpressionSpi.EOp o e))=(match (evalc e) with Some m => Some (MessageSpi.MOp o m) | _ => None end).
    intros;reflexivity.
  Qed.

  Lemma evalc_mem_fn_true : forall (e:ExpressionSpi.t)(m:MessageSpi.t),(evalc e = Some m)->forall (d k:nat),(deBruijn.mem_fn MessageSpi.deBruijn_fn_base k d m)=true->(deBruijn.mem_fn ExpressionSpi.deBruijn_fn_base k d e = true).
    induction e.
    intros.
    simpl in H.
    injection H;intros;subst m.
    rewrite MessageSpi.mem_fn_MNam in H0.
    rewrite ExpressionSpi.mem_fn_ENam.
    auto.
    rewrite evalc_EPair;intros.
    destruct (evalc e1);try (discriminate H).
    destruct (evalc e2);try (discriminate H).
    injection H;intros;subst m.
    rewrite MessageSpi.mem_fn_MPair in H0.
    rewrite ExpressionSpi.mem_fn_EPair.
    case (orb_prop (deBruijn.mem_fn MessageSpi.deBruijn_fn_base k d t) (deBruijn.mem_fn MessageSpi.deBruijn_fn_base k d t0) H0);intro.
    rewrite (IHe1 t);auto.
    rewrite (IHe2 t0);auto.
    rewrite orb_b_true;auto.
    rewrite evalc_EEnc_s;intros.
    destruct (evalc e1);try (discriminate H).
    destruct (evalc e2);try (discriminate H).
    injection H;intros;subst m.
    rewrite MessageSpi.mem_fn_MEnc_s in H0.
    rewrite ExpressionSpi.mem_fn_EEnc_s.
    case (orb_prop (deBruijn.mem_fn MessageSpi.deBruijn_fn_base k d t) (deBruijn.mem_fn MessageSpi.deBruijn_fn_base k d t0) H0);intro.
    rewrite (IHe1 t);auto.
    rewrite (IHe2 t0);auto.
    rewrite orb_b_true;auto.
    rewrite evalc_EEnc_a;intros.
    destruct (evalc e1);try (discriminate H).
    destruct (evalc e2);try (discriminate H).
    injection H;intros;subst m.
    rewrite MessageSpi.mem_fn_MEnc_a in H0.
    rewrite ExpressionSpi.mem_fn_EEnc_a.
    case (orb_prop (deBruijn.mem_fn MessageSpi.deBruijn_fn_base k d t) (deBruijn.mem_fn MessageSpi.deBruijn_fn_base k d t0) H0);intro.
    rewrite (IHe1 t);auto.
    rewrite (IHe2 t0);auto.
    rewrite orb_b_true;auto.
    rewrite evalc_EFst;intros.
    destruct (evalc e);try (discriminate H).
    rewrite ExpressionSpi.mem_fn_EFst.
    destruct t;try (discriminate H).
    injection H;intros;subst m.
    eapply IHe.
    reflexivity.
    rewrite MessageSpi.mem_fn_MPair.
    rewrite H0;auto.
    rewrite evalc_ESnd;intros.
    destruct (evalc e);try (discriminate H).
    rewrite ExpressionSpi.mem_fn_ESnd.
    destruct t;try (discriminate H).
    injection H;intros;subst m.
    eapply IHe.
    reflexivity.
    rewrite MessageSpi.mem_fn_MPair.
    rewrite H0;auto.
    rewrite orb_b_true;auto.
    rewrite evalc_EDec_s;intros.
    destruct (evalc e1);try (discriminate H).
    destruct t;try (discriminate H).
    destruct (evalc e2);try (discriminate H).
    destruct (MessageSpi.Ord.compare t2 t);try (discriminate H).
    injection H;intros;subst m.
    rewrite ExpressionSpi.mem_fn_EDec_s.
    rewrite (IHe1 (MessageSpi.MEnc_s t1 t2));auto.
    rewrite MessageSpi.mem_fn_MEnc_s.
    rewrite H0;auto.
    rewrite evalc_EDec_a;intros.
    destruct (evalc e1);try (discriminate H).
    destruct t;try (discriminate H).
    destruct (evalc e2);try (discriminate H).
    destruct t2;try (discriminate H).
    destruct o;try (discriminate H);unfold MessageSpi.inverse_key in H.
    destruct (MessageSpi.Ord.compare (MessageSpi.MOp Priv t2) t);try (discriminate H).
    injection H;intros;subst m.
    rewrite ExpressionSpi.mem_fn_EDec_a.
    rewrite (IHe1 (MessageSpi.MEnc_a t1 (MessageSpi.MOp Pub t2)));auto.
    rewrite MessageSpi.mem_fn_MEnc_a.
    rewrite H0;auto.
    destruct (MessageSpi.Ord.compare (MessageSpi.MOp Pub t2) t);try (discriminate H).
    injection H;intros;subst m.
    rewrite ExpressionSpi.mem_fn_EDec_a.
    rewrite (IHe1 (MessageSpi.MEnc_a t1 (MessageSpi.MOp Priv t2)));auto.
    rewrite MessageSpi.mem_fn_MEnc_a.
    rewrite H0;auto.
    rewrite evalc_EOp;intros.
    destruct (evalc e);try (discriminate H).
    rewrite ExpressionSpi.mem_fn_EOp.
    injection H;intros;subst m.
    rewrite MessageSpi.mem_fn_MOp in H0.
    eauto.
  Qed.

  Lemma evalc_mem_fn_false : forall (e:ExpressionSpi.t)(m:MessageSpi.t),(evalc e = Some m)->forall (d k:nat),(deBruijn.mem_fn ExpressionSpi.deBruijn_fn_base k d e = false)->(deBruijn.mem_fn MessageSpi.deBruijn_fn_base k d m = false).
    intros.
    case (Bool_elim_bool (deBruijn.mem_fn MessageSpi.deBruijn_fn_base k d m));intro;auto.
    rewrite (evalc_mem_fn_true e m H d k) in H0;auto.
    discriminate H0.
  Qed.

  Lemma evalc_map : forall (f:nat->Name.t->Name.t),(forall (d:nat),(injective (f d)))->forall (d:nat)(e:ExpressionSpi.t),(evalc (ExpressionSpi.t_map f d e))=(option_map (MessageSpi.t_map f d) (evalc e)).
    intros.
    induction e;simpl;try reflexivity;repeat (rewrite IHe || rewrite IHe1 || rewrite IHe2);try (destruct (evalc e));try (destruct (evalc e1));try (destruct (evalc e2));simpl;try reflexivity;try (destruct t);simpl;try reflexivity.
    generalize (deBruijn.map_injective MessageSpi.t_map_struct f H);intro.
    repeat (case (MessageSpi.Ord.compare);intro);try reflexivity;try (rewrite e in l;elim (MessageSpi.Ord.lt_not_eq (MessageSpi.t_map f d t0) (MessageSpi.t_map f d t0));auto);generalize (H0 d t2 t0 e);intro;subst t2;elim (MessageSpi.Ord.lt_not_eq t0 t0);auto.
    rewrite MessageSpi.inverse_key_map.
    destruct (MessageSpi.inverse_key t2);try reflexivity.
    simpl.
    generalize (deBruijn.map_injective MessageSpi.t_map_struct f H);intro.
    repeat (case (MessageSpi.Ord.compare);intro);try reflexivity;try (rewrite e in l;elim (MessageSpi.Ord.lt_not_eq (MessageSpi.t_map f d t0) (MessageSpi.t_map f d t0));auto);generalize (H0 d m t0 e);intro;subst m;elim (MessageSpi.Ord.lt_not_eq t0 t0);auto.
  Qed.

  Lemma evalc_lift : forall (d k:nat)(e:ExpressionSpi.t),(evalc (deBruijn.lift ExpressionSpi.deBruijn_base d k e))=(option_map (deBruijn.lift MessageSpi.deBruijn_base d k) (evalc e)).
    intros.
    simpl.
    apply evalc_map.
    intros.
    apply (deBruijn.lift_injective Name.deBruijn_base).
  Qed.

  Lemma evalc_swapR : forall (d k:nat)(e:ExpressionSpi.t),(evalc (deBruijn.swapR ExpressionSpi.deBruijn_base d k e))=(option_map (deBruijn.swapR MessageSpi.deBruijn_base d k) (evalc e)).
    intros.
    simpl.
    apply evalc_map.
    intros.
    apply (deBruijn.swapR_injective Name.deBruijn_base).
  Qed.

  Lemma evalc_map_success : forall (f:nat->Name.t->Name.t),forall (d:nat)(e:ExpressionSpi.t)(m:MessageSpi.t),(evalc e = Some m)->(evalc (ExpressionSpi.t_map f d e))=(Some (MessageSpi.t_map f d m)).
    induction e;simpl;intros.
    injection H;intros;subst m.
    reflexivity.
    destruct (evalc e1);try (discriminate H).
    destruct (evalc e2);try (discriminate H).
    injection H;intros;subst m.
    rewrite (IHe1 t);auto.
    rewrite (IHe2 t0);auto.
    destruct (evalc e1);try (discriminate H).
    destruct (evalc e2);try (discriminate H).
    injection H;intros;subst m.
    rewrite (IHe1 t);auto.
    rewrite (IHe2 t0);auto.
    destruct (evalc e1);try (discriminate H).
    destruct (evalc e2);try (discriminate H).
    injection H;intros;subst m.
    rewrite (IHe1 t);auto.
    rewrite (IHe2 t0);auto.
    destruct (evalc e);try (discriminate H).
    rewrite (IHe t);auto.
    destruct t;try (discriminate H).
    injection H;intros;subst m.
    reflexivity.
    destruct (evalc e);try (discriminate H).
    rewrite (IHe t);auto.
    destruct t;try (discriminate H).
    injection H;intros;subst m.
    reflexivity.
    destruct (evalc e1);try (discriminate H).
    destruct t;try (discriminate H).
    destruct (evalc e2);try (discriminate H).
    rewrite (IHe1 (MessageSpi.MEnc_s t1 t2));auto.
    rewrite (IHe2 t);auto.
    destruct (MessageSpi.Ord.compare t2 t);try (discriminate H).
    simpl.
    rewrite e.
    injection H;intros;subst m.
    destruct MessageSpi.Ord.compare;try reflexivity;elim (MessageSpi.Ord.lt_not_eq (MessageSpi.t_map f d t) (MessageSpi.t_map f d t));auto.
    destruct (evalc e1);try (discriminate H).
    destruct t;try (discriminate H).
    destruct (evalc e2);try (discriminate H).
    rewrite (IHe1 (MessageSpi.MEnc_a t1 t2));auto.
    rewrite (IHe2 t);auto.
    destruct t2;try (discriminate H).
    destruct o;try (discriminate H);unfold MessageSpi.inverse_key in H.
    destruct (MessageSpi.Ord.compare (MessageSpi.MOp Priv t2) t);try (discriminate H).
    unfold MessageSpi.inverse_key.
    rewrite <- e.
    lazy beta delta [MessageSpi.t_map] iota;fold MessageSpi.t_map.
    injection H;intros;subst m.
    destruct MessageSpi.Ord.compare;try reflexivity;elim (MessageSpi.Ord.lt_not_eq (MessageSpi.MOp Priv (MessageSpi.t_map f d t2)) (MessageSpi.MOp Priv (MessageSpi.t_map f d t2)));auto.
    destruct (MessageSpi.Ord.compare (MessageSpi.MOp Pub t2) t);try (discriminate H).
    unfold MessageSpi.inverse_key.
    rewrite <- e.
    lazy beta delta [MessageSpi.t_map] iota;fold MessageSpi.t_map.
    injection H;intros;subst m.
    destruct MessageSpi.Ord.compare;try reflexivity;elim (MessageSpi.Ord.lt_not_eq (MessageSpi.MOp Pub (MessageSpi.t_map f d t2)) (MessageSpi.MOp Pub (MessageSpi.t_map f d t2)));auto.
    destruct (evalc e);try (discriminate H).
    injection H;intros;subst m.
    rewrite (IHe t);auto.
  Qed.

  Lemma evalc_low1 : forall (d:nat)(e:ExpressionSpi.t),(deBruijn.mem_fn ExpressionSpi.deBruijn_fn_base 0 d e = false)->(evalc (deBruijn.low1 ExpressionSpi.deBruijn_base d e))=(option_map (deBruijn.low1 MessageSpi.deBruijn_base d) (evalc e)).
    induction e.
    intros.
    reflexivity.
    rewrite ExpressionSpi.mem_fn_EPair.
    intros.
    elim (orb_false_elim (deBruijn.mem_fn ExpressionSpi.deBruijn_fn_base 0 d e1) (deBruijn.mem_fn ExpressionSpi.deBruijn_fn_base 0 d e2) H);simpl;intros.
    simpl in IHe1;rewrite IHe1;auto.
    simpl in IHe2;rewrite IHe2;auto.
    destruct (evalc e1);destruct (evalc e2);try reflexivity.
    rewrite ExpressionSpi.mem_fn_EEnc_s.
    intros.
    elim (orb_false_elim (deBruijn.mem_fn ExpressionSpi.deBruijn_fn_base 0 d e1) (deBruijn.mem_fn ExpressionSpi.deBruijn_fn_base 0 d e2) H);simpl;intros.
    simpl in IHe1;rewrite IHe1;auto.
    simpl in IHe2;rewrite IHe2;auto.
    destruct (evalc e1);destruct (evalc e2);try reflexivity.
    rewrite ExpressionSpi.mem_fn_EEnc_a.
    intros.
    elim (orb_false_elim (deBruijn.mem_fn ExpressionSpi.deBruijn_fn_base 0 d e1) (deBruijn.mem_fn ExpressionSpi.deBruijn_fn_base 0 d e2) H);simpl;intros.
    simpl in IHe1;rewrite IHe1;auto.
    simpl in IHe2;rewrite IHe2;auto.
    destruct (evalc e1);destruct (evalc e2);try reflexivity.
    rewrite ExpressionSpi.mem_fn_EFst.
    simpl;intros.
    simpl in IHe;rewrite IHe;auto.
    destruct (evalc e);try reflexivity.
    destruct t;try reflexivity.
    rewrite ExpressionSpi.mem_fn_ESnd.
    simpl;intros.
    simpl in IHe;rewrite IHe;auto.
    destruct (evalc e);try reflexivity.
    destruct t;try reflexivity.
    rewrite ExpressionSpi.mem_fn_EDec_s.
    intros.
    elim (orb_false_elim (deBruijn.mem_fn ExpressionSpi.deBruijn_fn_base 0 d e1) (deBruijn.mem_fn ExpressionSpi.deBruijn_fn_base 0 d e2) H);intros.
    generalize (evalc_mem_fn_false e1);generalize (evalc_mem_fn_false e2);intros.
    simpl.
    simpl in IHe1;rewrite IHe1;auto.
    simpl in IHe2;rewrite IHe2;auto.
    destruct (evalc e1);try reflexivity.
    destruct t;try reflexivity;simpl.
    destruct (evalc e2);try reflexivity.
    simpl.
    generalize (H2 t (refl_equal (Some t)) d 0%nat H1).
    generalize (H3 (MessageSpi.MEnc_s t1 t2) (refl_equal (Some (MessageSpi.MEnc_s t1 t2))) d 0%nat H0).
    clear H2;clear H3.
    rewrite MessageSpi.mem_fn_MEnc_s;intros.
    elim (orb_false_elim (deBruijn.mem_fn MessageSpi.deBruijn_fn_base 0 d t1) (deBruijn.mem_fn MessageSpi.deBruijn_fn_base 0 d t2) H2);clear H2;intros.
    generalize (deBruijn.low1_injective MessageSpi.deBruijn_fn_base d t2 t H4 H3);simpl;intro.
    destruct (MessageSpi.Ord.compare t2 t);destruct (MessageSpi.Ord.compare);try reflexivity;try (elim (MessageSpi.Ord.lt_not_eq t2 t);auto;rewrite H5;auto;fail);try (elim (MessageSpi.Ord.lt_not_eq t t2);auto;rewrite H5;auto;fail);try (rewrite e in l;elim (MessageSpi.Ord.lt_not_eq (MessageSpi.t_map (fun d => Name.low1 d) d t) (MessageSpi.t_map (fun d => Name.low1 d) d t));auto;fail).
    rewrite ExpressionSpi.mem_fn_EDec_a.
    intros.
    elim (orb_false_elim (deBruijn.mem_fn ExpressionSpi.deBruijn_fn_base 0 d e1) (deBruijn.mem_fn ExpressionSpi.deBruijn_fn_base 0 d e2) H);intros.
    generalize (evalc_mem_fn_false e1);generalize (evalc_mem_fn_false e2);intros.
    simpl.
    simpl in IHe1;rewrite IHe1;auto.
    simpl in IHe2;rewrite IHe2;auto.
    destruct (evalc e1);try reflexivity.
    destruct t;try reflexivity;simpl.
    destruct (evalc e2);try reflexivity.
    simpl.
    generalize (H2 t (refl_equal (Some t)) d 0%nat H1).
    generalize (H3 (MessageSpi.MEnc_a t1 t2) (refl_equal (Some (MessageSpi.MEnc_a t1 t2))) d 0%nat H0).
    clear H2;clear H3.
    rewrite MessageSpi.mem_fn_MEnc_a;intros.
    elim (orb_false_elim (deBruijn.mem_fn MessageSpi.deBruijn_fn_base 0 d t1) (deBruijn.mem_fn MessageSpi.deBruijn_fn_base 0 d t2) H2);clear H2;intros.
    destruct t2;try reflexivity.
    rewrite MessageSpi.mem_fn_MOp in H4.
    simpl.
    destruct o;try reflexivity.
    generalize (deBruijn.low1_injective MessageSpi.deBruijn_fn_base d (MessageSpi.MOp Priv t2) t H4 H3);intro H5;simpl in H5.
    destruct (MessageSpi.Ord.compare (MessageSpi.MOp Priv t2) t);destruct (MessageSpi.Ord.compare);try reflexivity;try (elim (MessageSpi.Ord.lt_not_eq (MessageSpi.MOp Priv t2) t);auto;rewrite H5;auto;fail);try (elim (MessageSpi.Ord.lt_not_eq t (MessageSpi.MOp Priv t2));auto;rewrite H5;auto;fail);try (rewrite <- e in l;simpl in l;elim (MessageSpi.Ord.lt_not_eq (MessageSpi.MOp Priv (MessageSpi.t_map (fun d => Name.low1 d) d t2)) (MessageSpi.MOp Priv (MessageSpi.t_map (fun d => Name.low1 d) d t2)));simpl;auto;fail).
    generalize (deBruijn.low1_injective MessageSpi.deBruijn_fn_base d (MessageSpi.MOp Pub t2) t H4 H3);intro H5;simpl in H5.
    destruct (MessageSpi.Ord.compare (MessageSpi.MOp Pub t2) t);destruct (MessageSpi.Ord.compare);try reflexivity;try (elim (MessageSpi.Ord.lt_not_eq (MessageSpi.MOp Pub t2) t);auto;rewrite H5;auto;fail);try (elim (MessageSpi.Ord.lt_not_eq t (MessageSpi.MOp Pub t2));auto;rewrite H5;auto;fail);try (rewrite <- e in l;simpl in l;elim (MessageSpi.Ord.lt_not_eq (MessageSpi.MOp Pub (MessageSpi.t_map (fun d => Name.low1 d) d t2)) (MessageSpi.MOp Pub (MessageSpi.t_map (fun d => Name.low1 d) d t2)));simpl;auto;fail).
    rewrite ExpressionSpi.mem_fn_EOp.
    intros.
    simpl.
    simpl in IHe;rewrite IHe;auto.
    destruct (evalc e);try reflexivity.
  Qed.

  Lemma evalc_msg : forall (m:MessageSpi.t),(evalc (expr_of_mesg m))=(Some m).
    induction m;simpl;intros.
    reflexivity.
    rewrite IHm1.
    rewrite IHm2.
    auto.
    rewrite IHm1.
    rewrite IHm2.
    auto.
    rewrite IHm1.
    rewrite IHm2.
    auto.
    rewrite IHm;auto.
  Qed.

  (* abstract evaluation *)
  Fixpoint evala (e:ExpressionSpi.t) {struct e} : ExpressionSpi.t :=
    match e with
      | (ExpressionSpi.ENam n) => e
      | (ExpressionSpi.EPair e f) => ExpressionSpi.EPair (evala e) (evala f)
      | (ExpressionSpi.EEnc_s e f) => ExpressionSpi.EEnc_s (evala e) (evala f)
      | (ExpressionSpi.EEnc_a e f) => ExpressionSpi.EEnc_a (evala e) (evala f)
      | (ExpressionSpi.EFst e) =>
	(match (evala e) with
	   | (ExpressionSpi.EPair e1 e2) => e1
	   | f => (ExpressionSpi.EFst f)
	 end)
      | (ExpressionSpi.ESnd e) =>
	(match (evala e) with
	   | (ExpressionSpi.EPair e1 e2) => e2
	   | f => (ExpressionSpi.ESnd f)
	 end)
      | (ExpressionSpi.EDec_s e f) =>
	(match (evala e) with
	   | (ExpressionSpi.EEnc_s e1 e2) => e1
	   | e' => (ExpressionSpi.EDec_s e' (evala f))
	 end)
      | (ExpressionSpi.EDec_a e f) =>
	(match (evala e) with
	   | (ExpressionSpi.EEnc_a e1 e2) => e1
	   | e' => (ExpressionSpi.EDec_a e' (evala f))
	 end)
      | (ExpressionSpi.EOp o e) => ExpressionSpi.EOp o (evala e)
    end.

  Lemma evala_EPair : forall (e1 e2:ExpressionSpi.t),(evala (ExpressionSpi.EPair e1 e2))=(ExpressionSpi.EPair (evala e1) (evala e2)).
    intros;reflexivity.
  Qed.

  Lemma evala_EEnc_s : forall (e1 e2:ExpressionSpi.t),(evala (ExpressionSpi.EEnc_s e1 e2))=(ExpressionSpi.EEnc_s (evala e1) (evala e2)).
    intros;reflexivity.
  Qed.

  Lemma evala_EEnc_a : forall (e1 e2:ExpressionSpi.t),(evala (ExpressionSpi.EEnc_a e1 e2))=(ExpressionSpi.EEnc_a (evala e1) (evala e2)).
    intros;reflexivity.
  Qed.

  Lemma evala_EFst : forall (e:ExpressionSpi.t),(evala (ExpressionSpi.EFst e))=(match (evala e) with ExpressionSpi.EPair e1 e2 => e1 | _ => (ExpressionSpi.EFst (evala e)) end).
    intros;reflexivity.
  Qed.

  Lemma evala_ESnd : forall (e:ExpressionSpi.t),(evala (ExpressionSpi.ESnd e))=(match (evala e) with ExpressionSpi.EPair e1 e2 => e2 | _ => (ExpressionSpi.ESnd (evala e)) end).
    intros;reflexivity.
  Qed.

  Lemma evala_EDec_s : forall (e1 e2:ExpressionSpi.t),(evala (ExpressionSpi.EDec_s e1 e2))=(match (evala e1) with ExpressionSpi.EEnc_s f1 f2 => f1 | _ => (ExpressionSpi.EDec_s (evala e1) (evala e2)) end).
    intros;reflexivity.
  Qed.

  Lemma evala_EDec_a : forall (e1 e2:ExpressionSpi.t),(evala (ExpressionSpi.EDec_a e1 e2))=(match (evala e1) with ExpressionSpi.EEnc_a f1 f2 => f1 | _ => (ExpressionSpi.EDec_a (evala e1) (evala e2)) end).
    intros;reflexivity.
  Qed.

  Lemma evala_EOp : forall (o:op)(e:ExpressionSpi.t),(evala (ExpressionSpi.EOp o e))=(ExpressionSpi.EOp o (evala e)).
    intros;reflexivity.
  Qed.

  Lemma evala_mem_fn_true : forall (e:ExpressionSpi.t),forall (d k:nat),(deBruijn.mem_fn ExpressionSpi.deBruijn_fn_base k d (evala e))=true->(deBruijn.mem_fn ExpressionSpi.deBruijn_fn_base k d e = true).
    induction e.
    auto.
    rewrite evala_EPair;intros d k.
    repeat (rewrite ExpressionSpi.mem_fn_EPair);intro.
    case (orb_prop (deBruijn.mem_fn ExpressionSpi.deBruijn_fn_base k d (evala e1)) (deBruijn.mem_fn ExpressionSpi.deBruijn_fn_base k d (evala e2)) H);intro.
    rewrite IHe1;auto.
    rewrite IHe2;auto.
    rewrite orb_b_true;auto.
    rewrite evala_EEnc_s;intros d k.
    repeat (rewrite ExpressionSpi.mem_fn_EEnc_s);intro.
    case (orb_prop (deBruijn.mem_fn ExpressionSpi.deBruijn_fn_base k d (evala e1)) (deBruijn.mem_fn ExpressionSpi.deBruijn_fn_base k d (evala e2)) H);intro.
    rewrite IHe1;auto.
    rewrite IHe2;auto.
    rewrite orb_b_true;auto.
    rewrite evala_EEnc_a;intros d k.
    repeat (rewrite ExpressionSpi.mem_fn_EEnc_a);intro.
    case (orb_prop (deBruijn.mem_fn ExpressionSpi.deBruijn_fn_base k d (evala e1)) (deBruijn.mem_fn ExpressionSpi.deBruijn_fn_base k d (evala e2)) H);intro.
    rewrite IHe1;auto.
    rewrite IHe2;auto.
    rewrite orb_b_true;auto. 
    rewrite evala_EFst;intros.
    rewrite ExpressionSpi.mem_fn_EFst.
    apply IHe.
    destruct (evala e);try (rewrite ExpressionSpi.mem_fn_EFst in H);auto.
    rewrite ExpressionSpi.mem_fn_EPair.
    rewrite H;auto.
    rewrite evala_ESnd;intros.
    rewrite ExpressionSpi.mem_fn_ESnd.
    apply IHe.
    destruct (evala e);try (rewrite ExpressionSpi.mem_fn_ESnd in H);auto.
    rewrite ExpressionSpi.mem_fn_EPair.
    rewrite H;auto.
    rewrite orb_b_true;auto.
    rewrite evala_EDec_s;intros.
    rewrite ExpressionSpi.mem_fn_EDec_s.
    case (Bool_elim_bool (deBruijn.mem_fn ExpressionSpi.deBruijn_fn_base k d (evala e2)));intro.
    rewrite IHe2;auto.
    rewrite orb_b_true;auto.
    destruct (evala e1);try (rewrite ExpressionSpi.mem_fn_EDec_s in H);try (rewrite H0 in H;rewrite orb_b_false in H;rewrite IHe1);auto.
    rewrite IHe1;auto.
    rewrite ExpressionSpi.mem_fn_EEnc_s.
    rewrite H;auto.
    rewrite evala_EDec_a;intros.
    rewrite ExpressionSpi.mem_fn_EDec_a.
    case (Bool_elim_bool (deBruijn.mem_fn ExpressionSpi.deBruijn_fn_base k d (evala e2)));intro.
    rewrite IHe2;auto.
    rewrite orb_b_true;auto.
    destruct (evala e1);try (rewrite ExpressionSpi.mem_fn_EDec_a in H);try (rewrite H0 in H;rewrite orb_b_false in H;rewrite IHe1);auto.
    rewrite IHe1;auto.
    rewrite ExpressionSpi.mem_fn_EEnc_a.
    rewrite H;auto.
    rewrite evala_EOp;intros d k.
    repeat (rewrite ExpressionSpi.mem_fn_EOp);intro.
    auto.
  Qed.

  Lemma evala_mem_fn_false : forall (e:ExpressionSpi.t),forall (d k:nat),(deBruijn.mem_fn ExpressionSpi.deBruijn_fn_base k d e = false)->(deBruijn.mem_fn ExpressionSpi.deBruijn_fn_base k d (evala e))=false.
    intros.
    case (Bool_elim_bool (deBruijn.mem_fn ExpressionSpi.deBruijn_fn_base k d (evala e)));intro;auto.
    rewrite evala_mem_fn_true in H;auto.
    discriminate H.
  Qed.

  Lemma evala_map : forall (f:nat->Name.t->Name.t),forall (d:nat)(e:ExpressionSpi.t),(evala (ExpressionSpi.t_map f d e))=(ExpressionSpi.t_map f d (evala e)).
    intros.
    induction e;simpl;try reflexivity;repeat (rewrite IHe || rewrite IHe1 || rewrite IHe2);auto;try (destruct (evala e));try (destruct (evala e1));try (destruct (evala e2));simpl;try reflexivity;try (destruct t);simpl;try reflexivity.
  Qed.

  Definition eval (phi:FormulaeSpi.t) : bool :=
    match phi with
      | (FormulaeSpi.FName e) => 
	(match (evalc e) with
	   | Some (MessageSpi.MNam n) => true
	   | _ => false
	 end)
      | (FormulaeSpi.FEqual e f) =>
	(match ((evalc e),(evalc f)) with
	   | (Some m, Some n) => 
	     (match MessageSpi.Ord.compare m n with
		| EQ _ => true 
		| _ => false
	      end)
	   | _ => false
	 end)
    end.
  
  Lemma eval_map : forall (f:nat->Name.t->Name.t),(forall (d:nat),(injective (f d)))->forall (d:nat)(phi:FormulaeSpi.t),(eval (FormulaeSpi.t_map (fun d => ExpressionSpi.t_map f d) d phi))=(eval phi).
    intros.
    induction phi;simpl;try reflexivity;repeat (rewrite evalc_map;auto);try (destruct (evalc t);simpl;try reflexivity);try (destruct (evalc t0);simpl;try reflexivity).
    destruct t0;try reflexivity.
    generalize (deBruijn.map_injective MessageSpi.t_map_struct f H);intro.
    repeat (case (MessageSpi.Ord.compare);intro);try reflexivity;try (rewrite e in l;elim (MessageSpi.Ord.lt_not_eq (MessageSpi.t_map f d t2) (MessageSpi.t_map f d t2));auto);generalize (H0 d t1 t2 e);intro;subst t2;elim (MessageSpi.Ord.lt_not_eq t1 t1);auto.  
  Qed.

  Lemma eval_lift : forall (d k:nat)(phi:FormulaeSpi.t),(eval (deBruijn.lift FormulaeSpi.deBruijn_base d k phi))=(eval phi).
    intros.
    simpl.
    apply eval_map.
    intros.
    apply (deBruijn.lift_injective Name.deBruijn_base).
  Qed.

  Lemma eval_swapR : forall (d k:nat)(phi:FormulaeSpi.t),(eval (deBruijn.swapR FormulaeSpi.deBruijn_base d k phi))=(eval phi).
    intros.
    simpl.
    apply eval_map.
    intros.
    apply (deBruijn.swapR_injective Name.deBruijn_base).
  Qed.

  Lemma eval_low1 : forall (d:nat)(phi:FormulaeSpi.t),(deBruijn.mem_fn FormulaeSpi.deBruijn_fn_base 0 d phi = false)->(eval (deBruijn.low1 FormulaeSpi.deBruijn_base d phi))=(eval phi).
    intros.
    destruct phi.
    generalize (evalc_low1 d t H).
    rewrite FormulaeSpi.low1_FName.
    simpl;intros.
    rewrite H0.
    destruct (evalc t);try reflexivity.
    destruct t0;try reflexivity.
    rewrite FormulaeSpi.low1_FEqual.
    rewrite FormulaeSpi.mem_fn_FEqual in H.
    elim (orb_false_elim (deBruijn.mem_fn ExpressionSpi.deBruijn_fn_base 0 d t) (deBruijn.mem_fn ExpressionSpi.deBruijn_fn_base 0 d t0) H);clear H;intros.
    generalize (evalc_low1 d t H).
    generalize (evalc_low1 d t0 H0).
    generalize (evalc_mem_fn_false t).
    generalize (evalc_mem_fn_false t0).
    intros H1 H2.
    simpl;intros.
    rewrite H3.
    rewrite H4.
    destruct (evalc t);try reflexivity.
    simpl.
    destruct (evalc t0);try reflexivity.
    simpl.
    generalize (deBruijn.low1_injective MessageSpi.deBruijn_fn_base d t1 t2 (H2 t1 (refl_equal (Some t1)) d 0%nat H) (H1 t2 (refl_equal (Some t2)) d 0%nat H0)).
    intros.
    simpl in H5.
    destruct (MessageSpi.Ord.compare t1 t2);destruct (MessageSpi.Ord.compare);try reflexivity;try (elim (MessageSpi.Ord.lt_not_eq t1 t2);auto;rewrite H5;auto;fail);try (elim (MessageSpi.Ord.lt_not_eq t2 t1);auto;rewrite H5;auto;fail);try (rewrite e in l;elim (MessageSpi.Ord.lt_not_eq (MessageSpi.t_map (fun d => Name.low1 d) d t2) (MessageSpi.t_map (fun d => Name.low1 d) d t2));auto;fail).
  Qed.

  Definition eval_nc (phi:FormulaeSpi.t) : bool*list Name.t :=
    match phi with
      | (FormulaeSpi.FName e) => 
	(match (evalc e) with
	   | Some (MessageSpi.MNam n) => (true,cons n nil)
	   | _ => (false,nil)
	 end)
      | (FormulaeSpi.FEqual e f) =>
	(match ((evalc e),(evalc f)) with
	   | (Some m, Some n) => 
	     (match MessageSpi.Ord.compare m n with
		| EQ _ => (true,nil)
		| _ => (false,nil)
	      end)
	   | _ => (false,nil)
	 end)
    end.

  Lemma eval_nc_mem_fn_true : forall (phi:FormulaeSpi.t)(d k:nat),(deBruijn.mem_fn Name.deBruijn_fn_base_list k d (snd (eval_nc phi))=true)->(deBruijn.mem_fn FormulaeSpi.deBruijn_fn_base k d phi = true).
    induction phi;intros.
    unfold eval_nc in H.
    generalize (evalc_mem_fn_true t);intros.
    destruct (evalc t);try (discriminate H).
    destruct t0;try (discriminate H).
    simpl in H.
    simpl;simpl in H0.
    eapply H0.
    reflexivity.
    simpl.
    auto.
    simpl in H.
    destruct (evalc t);destruct (evalc t0);try (discriminate H).
    destruct (MessageSpi.Ord.compare t1 t2);try (discriminate H).
  Qed.

  Lemma eval_nc_mem_fn_false : forall (phi:FormulaeSpi.t)(d k:nat),(deBruijn.mem_fn FormulaeSpi.deBruijn_fn_base k d phi = false)->(deBruijn.mem_fn Name.deBruijn_fn_base_list k d (snd (eval_nc phi))=false).
    intros.
    case (Bool_elim_bool (deBruijn.mem_fn Name.deBruijn_fn_base_list k d (snd (eval_nc phi))));intro;auto.
    rewrite eval_nc_mem_fn_true in H;auto.
    discriminate H.
  Qed.

  Lemma eval_nc_eval : forall (phi:FormulaeSpi.t),(fst (eval_nc phi))=(eval phi).
    intros.
    induction phi;simpl;try reflexivity.
    destruct (evalc t);try reflexivity.
    destruct t0;try reflexivity.
    destruct (evalc t);try reflexivity.
    destruct (evalc t0);try reflexivity.
    destruct MessageSpi.Ord.compare;try reflexivity.
  Qed.

  Lemma eval_nc_map_snd : forall (f:nat->Name.t->Name.t),(forall (d:nat),(injective (f d)))->forall (d:nat)(phi:FormulaeSpi.t),(snd (eval_nc (FormulaeSpi.t_map (fun d => ExpressionSpi.t_map f d) d phi)))=(deBruijn.map_list f d (snd (eval_nc phi))).
    intros.
    induction phi;simpl;try reflexivity;repeat (rewrite evalc_map;auto);try (destruct (evalc t);simpl;try reflexivity);try (destruct (evalc t0);simpl;try reflexivity).
    destruct t0;try reflexivity.
    generalize (deBruijn.map_injective MessageSpi.t_map_struct f H);intro.
    repeat (case (MessageSpi.Ord.compare);intro);try reflexivity;try (rewrite e in l;elim (MessageSpi.Ord.lt_not_eq (MessageSpi.t_map f d t2) (MessageSpi.t_map f d t2));auto);generalize (H0 d t1 t2 e);intro;subst t2;elim (MessageSpi.Ord.lt_not_eq t1 t1);auto.  
  Qed.

  Lemma eval_nc_map : forall (f:nat->Name.t->Name.t),(forall (d:nat),(injective (f d)))->forall (d:nat)(phi:FormulaeSpi.t),(eval_nc (FormulaeSpi.t_map (fun d => ExpressionSpi.t_map f d) d phi))=(fst (eval_nc phi),(deBruijn.map_list f d (snd (eval_nc phi)))).
    intros.
    rewrite (pair_split (eval_nc (FormulaeSpi.t_map (fun d => ExpressionSpi.t_map f d) d phi))).
    rewrite eval_nc_map_snd;trivial.
    repeat (rewrite eval_nc_eval).
    rewrite eval_map;auto.
  Qed.

  Lemma eval_nc_lift : forall (d k:nat)(phi:FormulaeSpi.t),(eval_nc (deBruijn.lift FormulaeSpi.deBruijn_base d k phi))=((fst (eval_nc phi),(deBruijn.lift Name.deBruijn_base_list d k (snd (eval_nc phi))))).
    intros.
    simpl.
    apply eval_nc_map.
    intros.
    apply (deBruijn.lift_injective Name.deBruijn_base).
  Qed.

  Lemma eval_nc_swapR : forall (d k:nat)(phi:FormulaeSpi.t),(eval_nc (deBruijn.swapR FormulaeSpi.deBruijn_base d k phi))=((fst (eval_nc phi),(deBruijn.swapR Name.deBruijn_base_list d k (snd (eval_nc phi))))).
    intros.
    simpl.
    apply eval_nc_map.
    intros.
    apply (deBruijn.swapR_injective Name.deBruijn_base).
  Qed.

  Lemma eval_nc_map_success : forall (f:nat->Name.t->Name.t),forall (d:nat)(phi:FormulaeSpi.t)(l:list Name.t),(eval_nc phi = (true,l))->(eval_nc (FormulaeSpi.t_map (fun d => ExpressionSpi.t_map f d) d phi))=(true,(deBruijn.map_list f d l)).
    intros.
    destruct phi.
    simpl.
    simpl in H.
    generalize (evalc_map_success f d t).
    intro.
    destruct (evalc t);try (discriminate H).
    rewrite (H0 t0);auto.
    destruct t0;try (discriminate H).
    injection H;intros.
    subst l.
    reflexivity.
    simpl in H.
    generalize (evalc_map_success f d t).
    generalize (evalc_map_success f d t0).
    intros.
    destruct (evalc t);try (discriminate H).
    destruct (evalc t0);try (discriminate H).
    simpl.
    rewrite (H0 t2);auto.
    rewrite (H1 t1);auto.
    destruct (MessageSpi.Ord.compare t1 t2);try (discriminate H).
    rewrite <- e.
    injection H;intros;subst l.
    destruct MessageSpi.Ord.compare;try reflexivity;elim (MessageSpi.Ord.lt_not_eq (MessageSpi.t_map f d t1) (MessageSpi.t_map f d t1));auto.
  Qed.

  Lemma eval_nc_map_success_eval : forall (f:nat->Name.t->Name.t),forall (d:nat)(phi:FormulaeSpi.t),(eval phi = true)->(eval_nc (FormulaeSpi.t_map (fun d => ExpressionSpi.t_map f d) d phi))=(true,(deBruijn.map_list f d (snd (eval_nc phi)))).
    intros.
    generalize (eval_nc_eval phi);intro.
    generalize (eval_nc_map_success f d phi).
    intros.
    destruct (eval_nc phi).
    simpl in H0.
    subst b.
    rewrite H in H1.
    simpl.
    auto.
  Qed.

  Lemma eval_nc_low1 : forall (d:nat)(phi:FormulaeSpi.t),(deBruijn.mem_fn FormulaeSpi.deBruijn_fn_base 0 d phi = false)->(eval_nc (deBruijn.low1 FormulaeSpi.deBruijn_base d phi))=(fst (eval_nc phi),(deBruijn.map_list (fun d => deBruijn.low1 Name.deBruijn_base d) d (snd (eval_nc phi)))).
    intros.
    induction phi.
    rewrite FormulaeSpi.mem_fn_FName in H.
    rewrite FormulaeSpi.low1_FName.
    generalize (evalc_low1 d t H);simpl;intro.
    simpl.
    rewrite H0.
    destruct (evalc t);try reflexivity.
    simpl.
    destruct t0;try reflexivity.
    rewrite FormulaeSpi.low1_FEqual.
    rewrite FormulaeSpi.mem_fn_FEqual in H.
    elim (orb_false_elim (deBruijn.mem_fn ExpressionSpi.deBruijn_fn_base 0 d t) (deBruijn.mem_fn ExpressionSpi.deBruijn_fn_base 0 d t0) H);clear H;intros.
    generalize (evalc_low1 d t H).
    generalize (evalc_low1 d t0 H0).
    generalize (evalc_mem_fn_false t).
    generalize (evalc_mem_fn_false t0).
    intros H1 H2.
    simpl;intros.
    rewrite H3.
    rewrite H4.
    destruct (evalc t);try reflexivity.
    simpl.
    destruct (evalc t0);try reflexivity.
    simpl.
    generalize (deBruijn.low1_injective MessageSpi.deBruijn_fn_base d t1 t2 (H2 t1 (refl_equal (Some t1)) d 0%nat H) (H1 t2 (refl_equal (Some t2)) d 0%nat H0)).
    intros.
    simpl in H5.
    destruct (MessageSpi.Ord.compare t1 t2);destruct (MessageSpi.Ord.compare);try reflexivity;try (elim (MessageSpi.Ord.lt_not_eq t1 t2);auto;rewrite H5;auto;fail);try (elim (MessageSpi.Ord.lt_not_eq t2 t1);auto;rewrite H5;auto;fail);try (rewrite e in l;elim (MessageSpi.Ord.lt_not_eq (MessageSpi.t_map (fun d => Name.low1 d) d t2) (MessageSpi.t_map (fun d => Name.low1 d) d t2));auto;fail).
  Qed.

  (* evaluation of lists *)
  Definition evalc_succeeds (e:ExpressionSpi.t) := exists m:MessageSpi.t,evalc e = Some m.

  Fixpoint evalc_list (es:list ExpressionSpi.t) : list MessageSpi.t :=
    match es with 
      | nil => nil
      | cons x xs =>
	(match evalc x with
	   | None => nil
	   | Some m => cons m (evalc_list xs)
	 end)
    end.

  Lemma evalc_succeeds_list_Some : forall (es:list ExpressionSpi.t),(List_forall evalc_succeeds es)->forall (i:nat)(e:ExpressionSpi.t),(nth_error es i = Some e)->exists m:MessageSpi.t,nth_error (evalc_list es) i = Some m /\ (evalc e = Some m).
    induction es;intros.
    destruct i;discriminate H0.
    inversion H.
    destruct i.
    subst x;subst xs.
    generalize (List_forall_nth evalc_succeeds (cons a es) H 0 e H0).
    unfold evalc_succeeds.
    intro.
    ex_elim H1 m.
    simpl.
    simpl in H0.
    injection H0;intros;subst a.
    rewrite H1.
    exists m.
    auto.
    subst x;subst xs.
    simpl.
    simpl in H0.
    elim (IHes H4 i e H0).
    intro m;intros;and_elim.
    elim H3.
    intros.
    rewrite H5.
    rewrite H1.
    exists m;auto.
  Qed.
    
  Lemma evalc_succeeds_list_None : forall (es:list ExpressionSpi.t),(List_forall evalc_succeeds es)->forall (i:nat),(nth_error es i = None)->nth_error (evalc_list es) i = None.
    induction es;intros.
    simpl.
    destruct i;reflexivity.
    inversion H.
    subst x;subst xs.
    destruct i;try (discriminate H0).
    simpl in H0.
    unfold evalc_succeeds in H3.
    ex_elim H3 m.
    simpl.
    rewrite H3.
    auto.
  Qed.

  Lemma evalc_succeeds_list_length : forall (es:list ExpressionSpi.t),(List_forall evalc_succeeds es)->(length es = length (evalc_list es)).
    induction es;intros.
    reflexivity.
    simpl.
    inversion H.
    unfold evalc_succeeds in H2.
    ex_elim H2 m.
    rewrite H2.
    simpl.
    rewrite IHes;auto.
  Qed.

  Lemma evalc_succeeds_msg : forall (ms:list MessageSpi.t),(List_forall evalc_succeeds (map expr_of_mesg ms)).
    induction ms.
    simpl.
    apply List_forall_nil.
    simpl.
    apply List_forall_cons;auto.
    unfold evalc_succeeds.
    rewrite evalc_msg.
    exists a;auto.
  Qed.

  Lemma evalc_list_msg : forall (ms:list MessageSpi.t),(evalc_list (map expr_of_mesg ms))=ms.
    induction ms;try reflexivity.
    simpl.
    rewrite evalc_msg.
    rewrite IHms;auto.
  Qed.

  (* evalc and evala *)
  Lemma evalc_impl_evala : forall (e:ExpressionSpi.t)(m:MessageSpi.t),(evalc e = Some m)->(evala e)=(expr_of_mesg m).
    induction e;simpl;intros.
    injection H;intros;subst m.
    reflexivity.
    destruct (evalc e1);try (discriminate H).
    destruct (evalc e2);try (discriminate H).
    injection H;intros;subst m.
    rewrite (IHe1 t);trivial.
    rewrite (IHe2 t0);trivial.
    destruct (evalc e1);try (discriminate H).
    destruct (evalc e2);try (discriminate H).
    injection H;intros;subst m.
    rewrite (IHe1 t);trivial.
    rewrite (IHe2 t0);trivial.
    destruct (evalc e1);try (discriminate H).
    destruct (evalc e2);try (discriminate H).
    injection H;intros;subst m.
    rewrite (IHe1 t);trivial.
    rewrite (IHe2 t0);trivial.
    destruct (evalc e);try (discriminate H).
    rewrite (IHe t);trivial.
    destruct t;try (discriminate H).
    simpl.
    injection H;intros;subst m;trivial.
    destruct (evalc e);try (discriminate H).
    rewrite (IHe t);trivial.
    destruct t;try (discriminate H).
    simpl.
    injection H;intros;subst m;trivial.
    destruct (evalc e1);try (discriminate H).
    rewrite (IHe1 t);trivial.
    destruct t;try (discriminate H).
    simpl.
    destruct (evalc e2);try (discriminate H).
    destruct (MessageSpi.Ord.compare t2 t);try (discriminate H).
    injection H;intros;subst m;trivial.
    destruct (evalc e1);try (discriminate H).
    rewrite (IHe1 t);trivial.
    destruct t;try (discriminate H).
    simpl.
    destruct (evalc e2);try (discriminate H).
    destruct (MessageSpi.inverse_key t2);try (discriminate H).
    destruct (MessageSpi.Ord.compare m0 t);try (discriminate H).
    injection H;intros;subst m;trivial.
    destruct (evalc e);try (discriminate H).
    injection H;intros;subst m.
    rewrite (IHe t);auto.
  Qed.

  (* abstract evaluation and redexes *)
  Fixpoint nb_redex (e:ExpressionSpi.t) : nat :=
    match e with
      | ExpressionSpi.ENam _ => 0%nat
      | ExpressionSpi.EPair e1 e2 => (nb_redex e1) + (nb_redex e2)
      | ExpressionSpi.EEnc_s e1 e2 => (nb_redex e1) + (nb_redex e2)
      | ExpressionSpi.EEnc_a e1 e2 => (nb_redex e1) + (nb_redex e2)
      | ExpressionSpi.EOp _ e => (nb_redex e)
      | ExpressionSpi.EFst (ExpressionSpi.EPair e1 e2) => 1 + (nb_redex e1) + (nb_redex e2)
      | ExpressionSpi.EFst e => (nb_redex e)
      | ExpressionSpi.ESnd (ExpressionSpi.EPair e1 e2) => 1 + (nb_redex e1) + (nb_redex e2)
      | ExpressionSpi.ESnd e => (nb_redex e)
      | ExpressionSpi.EDec_s (ExpressionSpi.EEnc_s e1 e2) e3 => 1 + (nb_redex e1) + (nb_redex e2) + (nb_redex e3)
      | ExpressionSpi.EDec_s e1 e2 => (nb_redex e1) + (nb_redex e2)
      | ExpressionSpi.EDec_a (ExpressionSpi.EEnc_a e1 e2) e3 => 1 + (nb_redex e1) + (nb_redex e2) + (nb_redex e3)
      | ExpressionSpi.EDec_a e1 e2 => (nb_redex e1) + (nb_redex e2)
    end.

  Lemma evala_no_redex : forall (e:ExpressionSpi.t),(nb_redex (evala e) = 0%nat).
    induction e;simpl;try omega.
    destruct (evala e);simpl in IHe;simpl;try reflexivity;try omega.
    destruct (evala e);simpl in IHe;simpl;try reflexivity;try omega.
    destruct (evala e1);simpl in IHe1;simpl;try reflexivity;try omega.
    destruct (evala e1);simpl in IHe1;simpl;try reflexivity;try omega.
  Qed.

  Lemma no_redex_evala : forall (e:ExpressionSpi.t),(nb_redex e = 0%nat)->(evala e = e).
    induction e;simpl;intros;repeat (rewrite IHe || rewrite IHe1 || rewrite IHe2);auto;try omega;try (destruct e;try (discriminate H);auto).
    destruct e1;try (discriminate H);auto.
    destruct e1;try (discriminate H);try omega.
    destruct e1;try (discriminate H);try omega.
    destruct e1;try (discriminate H);auto.
    destruct e1;try (discriminate H);try omega.
    destruct e1;try (discriminate H);try omega.
  Qed.

  Lemma evala_id_iff_no_redex : forall (e:ExpressionSpi.t),(evala e = e)<->(nb_redex e = 0%nat).
    split;intros.
    rewrite <- H.
    apply evala_no_redex.
    apply no_redex_evala;auto.
  Qed.

  Lemma nb_redex_map : forall (f:nat->Name.t->Name.t),forall (d:nat)(e:ExpressionSpi.t),(nb_redex (ExpressionSpi.t_map f d e))=(nb_redex e).
    intros.
    induction e;simpl;try reflexivity;repeat (rewrite IHe || rewrite IHe1 || rewrite IHe2);auto.
    destruct e;simpl;simpl in IHe;auto.
    destruct e;simpl;simpl in IHe;auto.
    destruct e1;simpl;simpl in IHe1;auto.
    destruct e1;simpl;simpl in IHe1;auto.
  Qed.

  Lemma nb_redex_lift : forall (d k:nat)(e:ExpressionSpi.t),(nb_redex (deBruijn.lift ExpressionSpi.deBruijn_base d k e))=(nb_redex e).
    intros.
    simpl.
    rewrite nb_redex_map;auto.
  Qed.

  Lemma nb_redex_swapR : forall (d k:nat)(e:ExpressionSpi.t),(nb_redex (deBruijn.swapR ExpressionSpi.deBruijn_base d k e))=(nb_redex e).
    intros.
    simpl.
    rewrite nb_redex_map;auto.
  Qed.

  Lemma nb_redex_low1 : forall (d:nat)(e:ExpressionSpi.t),(nb_redex (deBruijn.low1 ExpressionSpi.deBruijn_base d e))=(nb_redex e).
    intros.
    simpl.
    rewrite nb_redex_map;auto.
  Qed.
End EvalSpi.


