(* A formalisation of the spi calculus in Coq *)
(* Copyright (C) 2005-2007 Sebastien Briais *)
(* http://lamp.epfl.ch/~sbriais/ *)

(* This library is free software; you can redistribute it and/or modify *)
(* it under the terms of the GNU Lesser General Public License as *)
(* published by the Free Software Foundation; either version 2.1 of the *)
(* License, or (at your option) any later version. *)

(* This library is distributed in the hope that it will be useful, but *)
(* WITHOUT ANY WARRANTY; without even the implied warranty of *)
(* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU *)
(* Lesser General Public License for more details. *)

(* You should have received a copy of the GNU Lesser General Public *)
(* License along with this library; if not, write to the Free Software *)
(* Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA *)
(* 02110-1301 USA *)

Require Import missing.
Require Import name.
Require Import expressionType.
Require Import deBruijnType.
Require Import Arith.
Require Import substLemma.
Require Import substSplit.

Module ExpressionPi <: ExpressionType.
  Inductive expression : Set := ENam : Name.t -> expression.

  Definition t := expression.
  Definition j (x:Name.t) := ENam x.

  Module Ord <: OrderedType with Definition t := t with Definition eq := @eq t.
    Definition t := t.
    Definition eq := @eq t.
    Definition lt (e1 e2:t) := 
      match (e1,e2) with
	| (ENam n1,ENam n2) => Name.Ord.lt n1 n2
      end.
    
    Lemma eq_refl : forall (x:t),(eq x x).
      unfold eq;reflexivity.
    Qed.

    Lemma eq_sym : forall (x y:t),(eq x y)->(eq y x).
      unfold eq;auto.
    Qed.

    Lemma eq_trans : forall (x y z:t),(eq x y)->(eq y z)->(eq x z).
      unfold eq;intros;congruence.
    Qed.

    Lemma lt_trans : forall (x y z:t),(lt x y)->(lt y z)->(lt x z).
      unfold lt.
      destruct x;destruct y;destruct z.
      apply Name.Ord.lt_trans.
    Qed.

    Lemma lt_not_eq : forall (x y:t),(lt x y)->~(eq x y).
      unfold lt, eq.
      destruct x;destruct y.
      intro;red;intro.
      injection H0;intro.
      elim (Name.Ord.lt_not_eq t0 t1);auto.
    Qed.

    Definition compare : forall (x y:t),(Compare lt eq x y).
      unfold lt, eq.
      destruct x; destruct y.
      case (Name.Ord.compare t0 t1);intro.
      apply LT;trivial.
      apply EQ.
      rewrite e;trivial.
      apply GT;trivial.
    Defined.

    Hint Immediate eq_sym.
    Hint Resolve eq_refl eq_trans lt_not_eq lt_trans.
  End Ord.

  Definition t_map (f:nat->Name.t->Name.t) (d:nat)(x:t) :=
    match x with
      | (ENam x) => j (f d x)
    end.

  Lemma t_map_struct : deBruijn.map_struct Name.t t t_map.
    split;unfold t_map;try (destruct x;unfold j;repeat (rewrite H);trivial).
    unfold injective.
    destruct x;destruct y;unfold j;simpl.
    intro.
    injection H0;intros.
    rewrite (H d t0 t1);auto.
  Qed.

  Definition deBruijn_base := deBruijn.deBruijn_map_struct Name.t Name.deBruijn_base t t_map t_map_struct.

  Definition t_map2 (f:nat->Name.t->t) (d:nat)(x:t) :=
    match x with
      | (ENam x) => f d x
    end.

  Definition deBruijn_extension : deBruijn.deBruijn_extension Name.t Name.deBruijn_base t deBruijn_base.
    split with j;intros;simpl;try reflexivity.
    unfold injective.
    unfold j;intros.
    injection H;auto.
  Defined.

  Lemma t_map2_extension : deBruijn.map_extension Name.t Name.deBruijn_base t deBruijn_base deBruijn_extension t_map t_map_struct t_map2.
    split;intros;simpl;try reflexivity.
    unfold t_map2;destruct x;auto.
    unfold t_map2;destruct x;auto.
  Qed.

  Definition deBruijn_subst := deBruijn.deBruijn_subst_gen_subst Name.t Name.deBruijn_base t t_map t_map_struct deBruijn_extension (Name.deBruijn_subst_gen t deBruijn_base deBruijn_extension) t_map2 t_map2_extension.

  Definition t_fold (A:Set)(f:nat->Name.t->A->A)(d:nat)(x:t)(a:A) : A := match x with ENam n => f d n a end.

  Lemma t_fold_struct : deBruijn.fold_struct Name.t t t_map t_map_struct t_fold.
    split;destruct x;intros;simpl;auto.
    simpl in H0.
    rewrite H;auto.
    (* fold_orb_eq_map_inversion *)
    destruct y;try (discriminate H2);simpl in H2.
    injection H2;intros;auto.
    simpl in H0;simpl in H1.
    rewrite (H d t0 t1);auto.
  Qed.

  Definition deBruijn_fn_base := deBruijn.deBruijn_fn_map_struct Name.t t t_map t_map_struct t_fold Name.deBruijn_base Name.deBruijn_fn_base t_fold_struct.

  Lemma deBruijn_fn_extension : deBruijn.deBruijn_extension_fn Name.t Name.deBruijn_base Name.deBruijn_fn_base t deBruijn_base deBruijn_fn_base deBruijn_extension.
    split.
    intros.
    reflexivity.
  Qed.

  Lemma t_fold_extension : deBruijn.fold_extension Name.t Name.deBruijn_base t deBruijn_base deBruijn_extension t_map t_map_struct t_map2 t_fold t_fold_struct.
    split.
    intros.
    reflexivity.
    (* fold_map_gen_simpl *)
    induction x;simpl;intros.
    reflexivity.
    (* fold_orb_eq_gen *)
    induction x;simpl;intros.
    apply H.
    auto.
  Qed.

  Definition deBruijn_fn_subst := deBruijn.deBruijn_s_gen_fn_fold_struct Name.t Name.deBruijn_base Name.deBruijn_fn_base t t_map t_map_struct deBruijn_extension (Name.deBruijn_subst_gen t deBruijn_base deBruijn_extension) t_map2 t_map2_extension t_fold t_fold_struct t_fold_extension deBruijn_fn_extension (Name.deBruijn_subst_gen_fn t deBruijn_base deBruijn_fn_base deBruijn_extension deBruijn_fn_extension).

  Lemma low_subst_lemma : forall (x:t)(d d' k k':nat)(es fs:list t),deBruijn.low_subst deBruijn_subst (d+d'+k) k' es (deBruijn.low_subst deBruijn_subst d k fs x)=deBruijn.low_subst deBruijn_subst d k (map (deBruijn.low_subst deBruijn_subst (d'+k) k' es) fs) (deBruijn.low_subst deBruijn_subst (d+d'+length fs) k' es x).
    induction x;intros.
    generalize (SubstLemmaGen.low_subst_lemma t t_map t_map_struct deBruijn_extension t_map2 t_map2_extension t0 d d' k k' es fs).
    simpl.
    auto.
    Opaque deBruijn.deBruijn_map_struct.
  Qed.
  Transparent deBruijn.deBruijn_map_struct.

  Lemma low_subst_split_lemma : forall (x:t)(d k:nat)(es fs:list t),deBruijn.low_subst deBruijn_subst d 0 es (deBruijn.low_subst deBruijn_subst (d+length es) k fs x)=deBruijn.low_subst deBruijn_subst d k (app es fs) x.
    induction x;intros.
    generalize (SubstSplitLemmaGen.low_subst_split_lemma t t_map t_map_struct deBruijn_extension t_map2 t_map2_extension t0 d k es fs).
    simpl.
    auto.
    Opaque deBruijn.deBruijn_map_struct.
  Qed.
  Transparent deBruijn.deBruijn_map_struct.

  Lemma low_subst_snoc_lemma : forall (x:t)(d k:nat)(e:t)(es:list t),deBruijn.low_subst deBruijn_subst d k (app es (cons e nil)) x =  deBruijn.low_subst deBruijn_subst d 0 (cons e nil) (deBruijn.swapR deBruijn_base d k (deBruijn.low_subst deBruijn_subst d k (map (deBruijn.lift deBruijn_base k 1) es) x)).
    intros.
    apply SubstLemmaFacts.low_subst_snoc_lemma.
    intros.
    apply low_subst_split_lemma.
  Qed.

  Definition compute_fn (d:nat)(x:t) : ListSet.set nat :=
    match x with
      | ENam n => Name.compute_fn d n
    end.
  
  Lemma compute_fn_ok : forall (k:nat)(d:nat)(x:t),(deBruijn.mem_fn deBruijn_fn_base k d x = ListSet.set_mem eq_nat_dec k (compute_fn d x)).
    destruct x;simpl;intros.
    apply Name.compute_fn_ok.
  Qed.
End ExpressionPi.



