(* A formalisation of the spi calculus in Coq *)
(* Copyright (C) 2005-2007 Sebastien Briais *)
(* http://lamp.epfl.ch/~sbriais/ *)

(* This library is free software; you can redistribute it and/or modify *)
(* it under the terms of the GNU Lesser General Public License as *)
(* published by the Free Software Foundation; either version 2.1 of the *)
(* License, or (at your option) any later version. *)

(* This library is distributed in the hope that it will be useful, but *)
(* WITHOUT ANY WARRANTY; without even the implied warranty of *)
(* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU *)
(* Lesser General Public License for more details. *)

(* You should have received a copy of the GNU Lesser General Public *)
(* License along with this library; if not, write to the Free Software *)
(* Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA *)
(* 02110-1301 USA *)

Require Import Arith.
Require Import missing.
Require Import name.
Require Import expressionType.
Require Import deBruijnType.
Require Import tactics.
Require Import opSpi.
Require Import substLemma.
Require Import substSplit.

Module ExpressionSpi <: ExpressionType.
  Import OpSpi.

  Inductive expression : Set := 
    | ENam : Name.t -> expression
    | EPair : expression -> expression -> expression
    | EEnc_s : expression -> expression -> expression
    | EEnc_a : expression -> expression -> expression
    | EFst : expression -> expression
    | ESnd : expression -> expression
    | EDec_s : expression -> expression -> expression
    | EDec_a : expression -> expression -> expression
    | EOp : op -> expression -> expression.

  Definition t := expression.
  Definition j (x:Name.t) := ENam x.

  Module Ord <: OrderedType with Definition t := t with Definition eq := @eq t.
    Definition t := t.
    Definition eq := @eq t.
    Fixpoint lt (e f:t) {struct e} : Prop := 
      match (e,f) with
	| (ENam n1,ENam n2) => Name.Ord.lt n1 n2
	| (ENam _,_) => True
	| (_,ENam _) => False
	| (EPair e1 e2,EPair f1 f2) => (lt e1 f1) \/ ((eq e1 f1) /\ (lt e2 f2))
	| (EPair _ _,_) => True
	| (_,EPair _ _) => False
	| (EEnc_s e1 e2,EEnc_s f1 f2) => (lt e1 f1) \/ ((eq e1 f1) /\ (lt e2 f2))
	| (EEnc_s _ _,_) => True
	| (_,EEnc_s _ _) => False
	| (EEnc_a e1 e2,EEnc_a f1 f2) => (lt e1 f1) \/ ((eq e1 f1) /\ (lt e2 f2))
	| (EEnc_a _ _,_) => True
	| (_,EEnc_a _ _) => False
	| (EFst e,EFst f) => lt e f
	| (EFst _,_) => True
	| (_,EFst _) => False
	| (ESnd e,ESnd f) => lt e f
	| (ESnd _,_) => True
	| (_,ESnd _) => False
	| (EOp o1 e1,EOp o2 f1) => (OpSpi.Ord.lt o1 o2) \/ ((OpSpi.Ord.eq o1 o2) /\ (lt e1 f1))
	| (EOp _ _,_) => True
	| (_,EOp _ _) => False
	| (EDec_s e1 e2,EDec_s f1 f2) => (lt e1 f1) \/ ((eq e1 f1) /\ (lt e2 f2))
	| (EDec_s _ _,_) => True
	| (_,EDec_s _ _) => False
	| (EDec_a e1 e2,EDec_a f1 f2) => (lt e1 f1) \/ ((eq e1 f1) /\ (lt e2 f2))
      end.
    
    Lemma eq_refl : forall (x:t),(eq x x).
      unfold eq;reflexivity.
    Qed.

    Lemma eq_sym : forall (x y:t),(eq x y)->(eq y x).
      unfold eq;auto.
    Qed.

    Lemma eq_trans : forall (x y z:t),(eq x y)->(eq y z)->(eq x z).
      unfold eq;intros;congruence.
    Qed.

    Lemma lt_trans : forall (x y z:t),(lt x y)->(lt y z)->(lt x z).
      induction x;destruct y;destruct z;simpl;intros;and_elim;try tauto;eauto.
      case H;clear H;intros;and_elim;case H0;clear H0;intros;and_elim;try (left;eauto;fail);try (left;(rewrite <- H0 || rewrite H);auto;fail);right;split;eauto;eapply eq_trans;eauto.
      case H;clear H;intros;and_elim;case H0;clear H0;intros;and_elim;try (left;eauto;fail);try (left;(rewrite <- H0 || rewrite H);auto;fail);right;split;eauto;eapply eq_trans;eauto.
      case H;clear H;intros;and_elim;case H0;clear H0;intros;and_elim;try (left;eauto;fail);try (left;(rewrite <- H0 || rewrite H);auto;fail);right;split;eauto;eapply eq_trans;eauto.
      case H;clear H;intros;and_elim;case H0;clear H0;intros;and_elim;try (left;eauto;fail);try (left;(rewrite <- H0 || rewrite H);auto;fail);right;split;eauto;eapply eq_trans;eauto.
      case H;clear H;intros;and_elim;case H0;clear H0;intros;and_elim;try (left;eauto;fail);try (left;(rewrite <- H0 || rewrite H);auto;fail);right;split;eauto;eapply eq_trans;eauto.
      case H;clear H;intros;and_elim;case H0;clear H0;intros;and_elim;try (left;eauto;fail);try (left;(rewrite <- H0 || rewrite H);auto;fail);right;split;eauto;eapply eq_trans;eauto.
    Qed.

    Lemma lt_not_eq : forall (x y:t),(lt x y)->~(eq x y).
      unfold eq.
      induction x;destruct y;simpl;intros;and_elim;try tauto;try (intro H0;discriminate H0);red;intros;injection H0;clear H0;intros.
      elim (Name.Ord.lt_not_eq t0 t1 H);auto.
      case H;intro;and_elim;[elim (IHx1 y1) | elim (IHx2 y2)];auto.
      case H;intro;and_elim;[elim (IHx1 y1) | elim (IHx2 y2)];auto.
      case H;intro;and_elim;[elim (IHx1 y1) | elim (IHx2 y2)];auto.
      elim (IHx y);auto.
      elim (IHx y);auto.
      case H;intro;and_elim;[elim (IHx1 y1) | elim (IHx2 y2)];auto.
      case H;intro;and_elim;[elim (IHx1 y1) | elim (IHx2 y2)];auto.
      case H;intro;and_elim.
      elim (OpSpi.Ord.lt_not_eq H2);rewrite H1;auto.
      elim (IHx y);auto.
    Qed.

    Definition compare : forall (x y:t),(Compare lt eq x y).
      induction x;destruct y;try (apply LT;simpl;auto;fail);try (apply GT;simpl;auto;fail).
      case (Name.Ord.compare t0 t1);intro;[apply LT | apply EQ | apply GT];simpl;auto;unfold eq;rewrite e;auto.
      case (IHx1 y1);intro;[apply LT | idtac | apply GT];simpl;auto.
      case (IHx2 y2);intro;[apply LT | apply EQ | apply GT];simpl;auto.
      unfold eq;rewrite e;rewrite e0;auto.
      right;split;auto.
      rewrite e;unfold eq;auto.
      case (IHx1 y1);intro;[apply LT | idtac | apply GT];simpl;auto.
      case (IHx2 y2);intro;[apply LT | apply EQ | apply GT];simpl;auto.
      unfold eq;rewrite e;rewrite e0;auto.
      right;split;auto.
      rewrite e;unfold eq;auto.
      case (IHx1 y1);intro;[apply LT | idtac | apply GT];simpl;auto.
      case (IHx2 y2);intro;[apply LT | apply EQ | apply GT];simpl;auto.
      unfold eq;rewrite e;rewrite e0;auto.
      right;split;auto.
      rewrite e;unfold eq;auto.
      case (IHx y);intro;[apply LT | apply EQ | apply GT];simpl;auto.
      rewrite e;unfold eq;auto.
      case (IHx y);intro;[apply LT | apply EQ | apply GT];simpl;auto.
      rewrite e;unfold eq;auto.    
      case (IHx1 y1);intro;[apply LT | idtac | apply GT];simpl;auto.
      case (IHx2 y2);intro;[apply LT | apply EQ | apply GT];simpl;auto.
      unfold eq;rewrite e;rewrite e0;auto.
      right;split;auto.
      rewrite e;unfold eq;auto.
      case (IHx1 y1);intro;[apply LT | idtac | apply GT];simpl;auto.
      case (IHx2 y2);intro;[apply LT | apply EQ | apply GT];simpl;auto.
      unfold eq;rewrite e;rewrite e0;auto.
      right;split;auto.
      rewrite e;unfold eq;auto.
      case (OpSpi.Ord.compare o o0);intro;[apply LT | idtac | apply GT];simpl;try tauto.
      case (IHx y);intro;[apply LT | apply EQ | apply GT];simpl;try tauto.
      unfold eq.
      rewrite e0.
      rewrite e.
      auto.
      right.
      split;auto.
    Defined.

    Hint Immediate eq_sym.
    Hint Resolve eq_refl eq_trans lt_not_eq lt_trans.
  End Ord.

  Fixpoint t_map (f:nat->Name.t->Name.t) (d:nat)(x:t) {struct x} : t :=
    match x with
      | (ENam x) => j (f d x)
      | (EPair e1 e2) => EPair (t_map f d e1) (t_map f d e2)
      | (EEnc_s e1 e2) => EEnc_s (t_map f d e1) (t_map f d e2)
      | (EEnc_a e1 e2) => EEnc_a (t_map f d e1) (t_map f d e2)
      | (EFst e) => EFst (t_map f d e)
      | (ESnd e) => ESnd (t_map f d e)
      | (EDec_s e1 e2) => EDec_s (t_map f d e1) (t_map f d e2)
      | (EDec_a e1 e2) => EDec_a (t_map f d e1) (t_map f d e2)
      | (EOp o e) => EOp o (t_map f d e)
    end.

  Lemma t_map_struct : deBruijn.map_struct Name.t t t_map.
    split;try (induction x;unfold j;simpl;repeat (rewrite IHx || rewrite IHx1 || rewrite IHx2 || rewrite H);trivial).
    intros.
    unfold injective.
    intro x;generalize d;clear d.
    induction x;destruct y;simpl;intros;try (discriminate H0);injection H0;intros;repeat (rewrite (IHx1 d y1) || rewrite (IHx2 d y2) || rewrite (IHx d y));auto.
    rewrite (H d t0 t1);auto.
    subst o0;auto.
  Qed.

  Definition deBruijn_base := deBruijn.deBruijn_map_struct Name.t Name.deBruijn_base t t_map t_map_struct.

  Fixpoint t_map2 (f:nat->Name.t->t) (d:nat)(x:t) {struct x} : t :=
    match x with
      | (ENam x) => f d x
      | (EPair e1 e2) => EPair (t_map2 f d e1) (t_map2 f d e2)
      | (EEnc_s e1 e2) => EEnc_s (t_map2 f d e1) (t_map2 f d e2)
      | (EEnc_a e1 e2) => EEnc_a (t_map2 f d e1) (t_map2 f d e2)
      | (EFst e) => EFst (t_map2 f d e)
      | (ESnd e) => ESnd (t_map2 f d e)
      | (EDec_s e1 e2) => EDec_s (t_map2 f d e1) (t_map2 f d e2)
      | (EDec_a e1 e2) => EDec_a (t_map2 f d e1) (t_map2 f d e2)
      | (EOp o e) => EOp o (t_map2 f d e)
    end.

  Definition deBruijn_extension : deBruijn.deBruijn_extension Name.t Name.deBruijn_base t deBruijn_base.
    split with j;intros;simpl;try reflexivity.
    unfold injective;intros.
    injection H;auto.
  Defined.

  Lemma t_map2_extension : deBruijn.map_extension Name.t Name.deBruijn_base t deBruijn_base deBruijn_extension t_map t_map_struct t_map2.
    split;intros;induction x;simpl;repeat (rewrite IHx1 || rewrite IHx2 || rewrite IHx || rewrite H);auto.
  Qed.

  Definition deBruijn_subst := deBruijn.deBruijn_subst_gen_subst Name.t Name.deBruijn_base t t_map t_map_struct deBruijn_extension (Name.deBruijn_subst_gen t deBruijn_base deBruijn_extension) t_map2 t_map2_extension.

  Fixpoint t_fold (A:Set)(f:nat->Name.t->A->A)(d:nat)(x:t)(a:A) {struct x} : A :=
    match x with
      | ENam n => f d n a
      | EPair e1 e2 => t_fold A f d e2 (t_fold A f d e1 a)
      | EEnc_s e1 e2 => t_fold A f d e2 (t_fold A f d e1 a)
      | EEnc_a e1 e2 => t_fold A f d e2 (t_fold A f d e1 a)
      | EFst e => t_fold A f d e a
      | ESnd e => t_fold A f d e a
      | EDec_s e1 e2 => t_fold A f d e2 (t_fold A f d e1 a)
      | EDec_a e1 e2 => t_fold A f d e2 (t_fold A f d e1 a)
      | EOp _ e => t_fold A f d e a
    end.

  Implicit Arguments t_fold [A].

  Lemma t_fold_false_inv : forall (f:nat->Name.t->bool)(d:nat)(x:t)(b:bool),(t_fold (fun d x b => b || f d x) d x b = false)->b=false.
    induction x;simpl;intros;auto.
    elim (orb_false_elim b (f d t0) H);auto.
  Qed.

  Implicit Arguments t_fold_false_inv [f d x b].

  Lemma t_fold_true : forall (f:nat->Name.t->bool)(d:nat)(x:t),(t_fold (fun d x b => b || f d x) d x true = true).
    induction x;simpl;intros;repeat (rewrite IHx1 || rewrite IHx2);auto.
  Qed.

  Lemma t_fold_invariant : forall (A:Set)(P:A->Prop)(f:nat->Name.t->A->A),(forall (d:nat)(x:Name.t)(a:A),(P a)->(P (f d x a)))->forall (d:nat)(x:t)(a:A),(P a)->(P (t_fold f d x a)).
    intros.
    generalize H0;clear H0.
    generalize a;clear a.
    induction x;simpl;intros;auto.
  Qed.

  Definition t_fold_struct : deBruijn.fold_struct Name.t t t_map t_map_struct t_fold.
    split.
    (* fold_orb_eq *)
    induction x;intros;simpl;repeat (rewrite H || rewrite IHx || rewrite IHx1 || rewrite IHx2);simpl in H0;auto;intros;try (generalize (t_fold_false_inv (H0 i H1));intro;generalize (H0 i H1);intro;rewrite H2 in H3;auto).
    (* fold_exists *)
    induction x;simpl;intros;auto.
    destruct (t_fold (fun d x b => b || f d x) (d1+d) x1 false);auto.
    rewrite IHx1;trivial;apply t_fold_true.
    destruct (t_fold (fun d x b => b || g d x) (d2+d) x1 false);auto;try (apply t_fold_true).
    destruct (t_fold (fun d x b => b || f d x) (d1+d) x1 false);auto.
    rewrite IHx1;trivial;apply t_fold_true.
    destruct (t_fold (fun d x b => b || g d x) (d2+d) x1 false);auto;try (apply t_fold_true).
    destruct (t_fold (fun d x b => b || f d x) (d1+d) x1 false);auto.
    rewrite IHx1;trivial;apply t_fold_true.
    destruct (t_fold (fun d x b => b || g d x) (d2+d) x1 false);auto;try (apply t_fold_true).
    destruct (t_fold (fun d x b => b || f d x) (d1+d) x1 false);auto.
    rewrite IHx1;trivial;apply t_fold_true.
    destruct (t_fold (fun d x b => b || g d x) (d2+d) x1 false);auto;try (apply t_fold_true).
    destruct (t_fold (fun d x b => b || f d x) (d1+d) x1 false);auto.
    rewrite IHx1;trivial;apply t_fold_true.
    destruct (t_fold (fun d x b => b || g d x) (d2+d) x1 false);auto;try (apply t_fold_true).
    (* fold_map_simpl *)
    induction x;simpl;intros;repeat (rewrite IHx || rewrite IHx1 || rewrite IHx2);auto.
    (* fold_equal_cond_invariant *)
    induction x;simpl;intros;repeat (rewrite <- IHx || rewrite <- IHx1 || rewrite <- IHx2);intros;auto;try (generalize (H1 i H3);intro H4;rewrite (t_fold_false_inv H4) in H4;auto;fail);try (apply t_fold_invariant;auto;fail);try (apply (t_fold_false_inv (H1 i H3));auto;fail).
    (* fold_invariant *)
    intros.
    apply t_fold_invariant;auto.
    (* fold_orb_eq_map_inversion *)
    induction x;simpl;intros;auto;destruct y;try (discriminate H2);simpl in H2;injection H2;clear H2;intros;try (subst o0);simpl in H0;simpl in H1;repeat (rewrite (IHx1 y1) || rewrite (IHx2 y2) || rewrite (IHx y));intros;auto;try (generalize (H0 i H4);intro H5;rewrite (t_fold_false_inv H5) in H5;auto;fail);try (generalize (H1 i H4);intro H5;rewrite (t_fold_false_inv H5) in H5;auto;fail);try (apply (t_fold_false_inv (H0 i H4)));try (apply (t_fold_false_inv (H1 i H4))).
    rewrite (H d t0 t1);auto.
  Qed.

  Definition deBruijn_fn_base := deBruijn.deBruijn_fn_map_struct Name.t t t_map t_map_struct t_fold Name.deBruijn_base Name.deBruijn_fn_base t_fold_struct.

  Lemma low_subst_EPair : forall (d k:nat)(sigma:list ExpressionSpi.t)(e1 e2:ExpressionSpi.t),(deBruijn.low_subst deBruijn_subst d k sigma (ExpressionSpi.EPair e1 e2))=(ExpressionSpi.EPair (deBruijn.low_subst deBruijn_subst d k sigma e1) (deBruijn.low_subst deBruijn_subst d k sigma e2)).
    intros.
    reflexivity.
  Qed.

  Lemma low_subst_EEnc_s : forall (d k:nat)(sigma:list ExpressionSpi.t)(e1 e2:ExpressionSpi.t),(deBruijn.low_subst deBruijn_subst d k sigma (ExpressionSpi.EEnc_s e1 e2))=(ExpressionSpi.EEnc_s (deBruijn.low_subst deBruijn_subst d k sigma e1) (deBruijn.low_subst deBruijn_subst d k sigma e2)).
    intros.
    reflexivity.
  Qed.

  Lemma low_subst_EEnc_a : forall (d k:nat)(sigma:list ExpressionSpi.t)(e1 e2:ExpressionSpi.t),(deBruijn.low_subst deBruijn_subst d k sigma (ExpressionSpi.EEnc_a e1 e2))=(ExpressionSpi.EEnc_a (deBruijn.low_subst deBruijn_subst d k sigma e1) (deBruijn.low_subst deBruijn_subst d k sigma e2)).
    intros.
    reflexivity.
  Qed.

  Lemma low_subst_EDec_s : forall (d k:nat)(sigma:list ExpressionSpi.t)(e1 e2:ExpressionSpi.t),(deBruijn.low_subst deBruijn_subst d k sigma (ExpressionSpi.EDec_s e1 e2))=(ExpressionSpi.EDec_s (deBruijn.low_subst deBruijn_subst d k sigma e1) (deBruijn.low_subst deBruijn_subst d k sigma e2)).
    intros.
    reflexivity.
  Qed.

  Lemma low_subst_EDec_a : forall (d k:nat)(sigma:list ExpressionSpi.t)(e1 e2:ExpressionSpi.t),(deBruijn.low_subst deBruijn_subst d k sigma (ExpressionSpi.EDec_a e1 e2))=(ExpressionSpi.EDec_a (deBruijn.low_subst deBruijn_subst d k sigma e1) (deBruijn.low_subst deBruijn_subst d k sigma e2)).
    intros.
    reflexivity.
  Qed.

  Lemma low_subst_EFst : forall (d k:nat)(sigma:list ExpressionSpi.t)(e:ExpressionSpi.t),(deBruijn.low_subst deBruijn_subst d k sigma (ExpressionSpi.EFst e))=(ExpressionSpi.EFst (deBruijn.low_subst deBruijn_subst d k sigma e)).
    intros.
    reflexivity.
  Qed.

  Lemma low_subst_ESnd : forall (d k:nat)(sigma:list ExpressionSpi.t)(e:ExpressionSpi.t),(deBruijn.low_subst deBruijn_subst d k sigma (ExpressionSpi.ESnd e))=(ExpressionSpi.ESnd (deBruijn.low_subst deBruijn_subst d k sigma e)).
    intros.
    reflexivity.
  Qed.

  Lemma low_subst_EOp : forall (d k:nat)(sigma:list ExpressionSpi.t)(o:op)(e:ExpressionSpi.t),(deBruijn.low_subst deBruijn_subst d k sigma (ExpressionSpi.EOp o e))=(ExpressionSpi.EOp o (deBruijn.low_subst deBruijn_subst d k sigma e)).
    intros.
    reflexivity.
  Qed.

  Lemma mem_fn_ENam : forall (d k:nat)(n:Name.t),(deBruijn.mem_fn deBruijn_fn_base k d (ENam n))=(deBruijn.mem_fn Name.deBruijn_fn_base k d n).
    intros;reflexivity.
  Qed.

  Lemma mem_fn_EPair : forall (d k:nat)(m1 m2:t),(deBruijn.mem_fn deBruijn_fn_base k d (EPair m1 m2))=(orb (deBruijn.mem_fn deBruijn_fn_base k d m1) (deBruijn.mem_fn deBruijn_fn_base k d m2)).
    intros.
    case (Bool_elim_bool (deBruijn.mem_fn deBruijn_fn_base k d m1));simpl;intro H;rewrite H;auto.
    rewrite t_fold_true;auto.
  Qed.

  Lemma mem_fn_EEnc_s : forall (d k:nat)(m1 m2:t),(deBruijn.mem_fn deBruijn_fn_base k d (EEnc_s m1 m2))=(orb (deBruijn.mem_fn deBruijn_fn_base k d m1) (deBruijn.mem_fn deBruijn_fn_base k d m2)).
    intros.
    case (Bool_elim_bool (deBruijn.mem_fn deBruijn_fn_base k d m1));simpl;intro H;rewrite H;auto.
    rewrite t_fold_true;auto.
  Qed.

  Lemma mem_fn_EEnc_a : forall (d k:nat)(m1 m2:t),(deBruijn.mem_fn deBruijn_fn_base k d (EEnc_a m1 m2))=(orb (deBruijn.mem_fn deBruijn_fn_base k d m1) (deBruijn.mem_fn deBruijn_fn_base k d m2)).
    intros.
    case (Bool_elim_bool (deBruijn.mem_fn deBruijn_fn_base k d m1));simpl;intro H;rewrite H;auto.
    rewrite t_fold_true;auto.
  Qed.

  Lemma mem_fn_EDec_s : forall (d k:nat)(m1 m2:t),(deBruijn.mem_fn deBruijn_fn_base k d (EDec_s m1 m2))=(orb (deBruijn.mem_fn deBruijn_fn_base k d m1) (deBruijn.mem_fn deBruijn_fn_base k d m2)).
    intros.
    case (Bool_elim_bool (deBruijn.mem_fn deBruijn_fn_base k d m1));simpl;intro H;rewrite H;auto.
    rewrite t_fold_true;auto.
  Qed.

  Lemma mem_fn_EDec_a : forall (d k:nat)(m1 m2:t),(deBruijn.mem_fn deBruijn_fn_base k d (EDec_a m1 m2))=(orb (deBruijn.mem_fn deBruijn_fn_base k d m1) (deBruijn.mem_fn deBruijn_fn_base k d m2)).
    intros.
    case (Bool_elim_bool (deBruijn.mem_fn deBruijn_fn_base k d m1));simpl;intro H;rewrite H;auto.
    rewrite t_fold_true;auto.
  Qed.

  Lemma mem_fn_EOp : forall (d k:nat)(o:op)(m:t),(deBruijn.mem_fn deBruijn_fn_base k d (EOp o m))=(deBruijn.mem_fn deBruijn_fn_base k d m).
    intros.
    reflexivity.
  Qed.

  Lemma mem_fn_EFst : forall (d k:nat)(m:t),(deBruijn.mem_fn deBruijn_fn_base k d (EFst m))=(deBruijn.mem_fn deBruijn_fn_base k d m).
    intros.
    reflexivity.
  Qed.

  Lemma mem_fn_ESnd : forall (d k:nat)(m:t),(deBruijn.mem_fn deBruijn_fn_base k d (ESnd m))=(deBruijn.mem_fn deBruijn_fn_base k d m).
    intros.
    reflexivity.
  Qed.

  Lemma deBruijn_fn_extension : deBruijn.deBruijn_extension_fn Name.t Name.deBruijn_base Name.deBruijn_fn_base t deBruijn_base deBruijn_fn_base deBruijn_extension.
    split.
    intros.
    reflexivity.
  Qed.

  Lemma t_fold_extension : deBruijn.fold_extension Name.t Name.deBruijn_base t deBruijn_base deBruijn_extension t_map t_map_struct t_map2 t_fold t_fold_struct.
    split.
    intros.
    reflexivity.
    (* fold_map_gen_simpl *)
    induction x;simpl;intros.
    generalize a;clear a.
    generalize (g (d2+d) t0).
    induction t1;intros;simpl;auto;try (rewrite IHt1_1);try (rewrite IHt1_2);auto.
    rewrite IHx2;rewrite IHx1;auto.
    rewrite IHx2;rewrite IHx1;auto.
    rewrite IHx2;rewrite IHx1;auto.
    rewrite IHx;auto.
    rewrite IHx;auto.
    rewrite IHx2;rewrite IHx1;auto.
    rewrite IHx2;rewrite IHx1;auto.
    rewrite IHx;auto.
    (* fold_orb_eq_gen *)
    induction x;intros;simpl;repeat (rewrite H || rewrite IHx || rewrite IHx1 || rewrite IHx2);simpl in H0;auto;intros;try (generalize (t_fold_false_inv (H0 i H1));intro;generalize (H0 i H1);intro;rewrite H2 in H3;auto).
  Qed.

  Definition deBruijn_fn_subst := deBruijn.deBruijn_s_gen_fn_fold_struct Name.t Name.deBruijn_base Name.deBruijn_fn_base t t_map t_map_struct deBruijn_extension (Name.deBruijn_subst_gen t deBruijn_base deBruijn_extension) t_map2 t_map2_extension t_fold t_fold_struct t_fold_extension deBruijn_fn_extension (Name.deBruijn_subst_gen_fn t deBruijn_base deBruijn_fn_base deBruijn_extension deBruijn_fn_extension).

  Lemma low_subst_ENam : forall (d k:nat)(sigma:list ExpressionSpi.t)(x:Name.t),(deBruijn.low_subst deBruijn_subst d k sigma (ExpressionSpi.ENam x))=(Name.low_subst_generic t (fun d k => t_map (fun d' => Name.lift d' k) d) j d k) sigma x.
    intros.
    reflexivity.
  Qed.

  Lemma low_subst_lemma : forall (x:t)(d d' k k':nat)(es fs:list t),deBruijn.low_subst deBruijn_subst (d+d'+k) k' es (deBruijn.low_subst deBruijn_subst d k fs x)=deBruijn.low_subst deBruijn_subst d k (map (deBruijn.low_subst deBruijn_subst (d'+k) k' es) fs) (deBruijn.low_subst deBruijn_subst (d+d'+length fs) k' es x).
    induction x;intros.
    generalize (SubstLemmaGen.low_subst_lemma t t_map t_map_struct deBruijn_extension t_map2 t_map2_extension t0 d d' k k' es fs).
    simpl;auto.
    repeat (rewrite low_subst_EPair);try (rewrite IHx1);try (rewrite IHx2);auto.
    repeat (rewrite low_subst_EEnc_s);try (rewrite IHx1);try (rewrite IHx2);auto.
    repeat (rewrite low_subst_EEnc_a);try (rewrite IHx1);try (rewrite IHx2);auto.
    repeat (rewrite low_subst_EFst);try (rewrite IHx);auto.
    repeat (rewrite low_subst_ESnd);try (rewrite IHx);auto.
    repeat (rewrite low_subst_EDec_s);try (rewrite IHx1);try (rewrite IHx2);auto.
    repeat (rewrite low_subst_EDec_a);try (rewrite IHx1);try (rewrite IHx2);auto.
    repeat (rewrite low_subst_EOp);try (rewrite IHx);auto.
    Opaque deBruijn.deBruijn_map_struct.
  Qed.
  Transparent deBruijn.deBruijn_map_struct.

  Lemma low_subst_split_lemma : forall (x:t)(d k:nat)(es fs:list t),deBruijn.low_subst deBruijn_subst d 0 es (deBruijn.low_subst deBruijn_subst (d+length es) k fs x)=deBruijn.low_subst deBruijn_subst d k (app es fs) x.
    induction x;intros.
    generalize (SubstSplitLemmaGen.low_subst_split_lemma t t_map t_map_struct deBruijn_extension t_map2 t_map2_extension t0 d k es fs).
    simpl;auto.
    repeat (rewrite low_subst_EPair);try (rewrite IHx1);try (rewrite IHx2);auto.
    repeat (rewrite low_subst_EEnc_s);try (rewrite IHx1);try (rewrite IHx2);auto.
    repeat (rewrite low_subst_EEnc_a);try (rewrite IHx1);try (rewrite IHx2);auto.
    repeat (rewrite low_subst_EFst);try (rewrite IHx);auto.
    repeat (rewrite low_subst_ESnd);try (rewrite IHx);auto.
    repeat (rewrite low_subst_EDec_s);try (rewrite IHx1);try (rewrite IHx2);auto.
    repeat (rewrite low_subst_EDec_a);try (rewrite IHx1);try (rewrite IHx2);auto.
    repeat (rewrite low_subst_EOp);try (rewrite IHx);auto.
    Opaque deBruijn.deBruijn_map_struct.
  Qed.
  Transparent deBruijn.deBruijn_map_struct.

  Lemma low_subst_snoc_lemma : forall (x:t)(d k:nat)(e:t)(es:list t),deBruijn.low_subst deBruijn_subst d k (app es (cons e nil)) x =  deBruijn.low_subst deBruijn_subst d 0 (cons e nil) (deBruijn.swapR deBruijn_base d k (deBruijn.low_subst deBruijn_subst d k (map (deBruijn.lift deBruijn_base k 1) es) x)).
    intros.
    apply SubstLemmaFacts.low_subst_snoc_lemma.
    intros.
    apply low_subst_split_lemma.
  Qed.

  Fixpoint compute_fn (d:nat)(x:t) {struct x} : set nat :=
  match x with
    | ENam n => Name.compute_fn d n
    | EPair m1 m2 => set_union eq_nat_dec (compute_fn d m1) (compute_fn d m2)
    | EEnc_s m1 m2 => set_union eq_nat_dec (compute_fn d m1) (compute_fn d m2)
    | EEnc_a m1 m2 => set_union eq_nat_dec (compute_fn d m1) (compute_fn d m2)
    | EOp _ m => compute_fn d m
    | EFst m => compute_fn d m
    | ESnd m => compute_fn d m
    | EDec_s m1 m2 => set_union eq_nat_dec (compute_fn d m1) (compute_fn d m2)
    | EDec_a m1 m2 => set_union eq_nat_dec (compute_fn d m1) (compute_fn d m2)
  end.

  Lemma compute_fn_ok : forall (k:nat)(d:nat)(x:t),(deBruijn.mem_fn deBruijn_fn_base k d x = set_mem eq_nat_dec k (compute_fn d x)).
    induction x;repeat (rewrite mem_fn_EPair || rewrite mem_fn_EEnc_s || rewrite mem_fn_EEnc_a || rewrite mem_fn_EDec_s || rewrite mem_fn_EDec_a || rewrite mem_fn_EOp);lazy beta delta [compute_fn] iota;fold compute_fn;try (rewrite ListSet_mem_union;rewrite IHx1;rewrite IHx2;auto);auto.
    simpl;apply Name.compute_fn_ok.
  Qed.
End ExpressionSpi.