(* A formalisation of the spi calculus in Coq *)
(* Copyright (C) 2005-2007 Sebastien Briais *)
(* http://lamp.epfl.ch/~sbriais/ *)

(* This library is free software; you can redistribute it and/or modify *)
(* it under the terms of the GNU Lesser General Public License as *)
(* published by the Free Software Foundation; either version 2.1 of the *)
(* License, or (at your option) any later version. *)

(* This library is distributed in the hope that it will be useful, but *)
(* WITHOUT ANY WARRANTY; without even the implied warranty of *)
(* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU *)
(* Lesser General Public License for more details. *)

(* You should have received a copy of the GNU Lesser General Public *)
(* License along with this library; if not, write to the Free Software *)
(* Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA *)
(* 02110-1301 USA *)

Require Import Arith.
Require Import missing.
Require Import formulaeType.
Require Import expressionPi.
Require Import List.
Require Import tactics.
Require Import deBruijnType.
Require Import substLemma.
Require Import substSplit.

Module FormulaePi <: FormulaeType with Module E := ExpressionPi.
  Module E := ExpressionPi.

  Inductive formulae : Set := 
    | FEqual : E.t -> E.t -> formulae.

  Definition t := formulae.

  Module Ord <: OrderedType with Definition t := t with Definition eq := @eq t.
    Definition t := t.
    Definition eq := @eq t.
    Definition lt (phi1 phi2:t) := 
      match (phi1,phi2) with
	| (FEqual e1 e'1,FEqual e2 e'2) => (E.Ord.lt e1 e2) \/ ((E.Ord.eq e1 e2) /\ (E.Ord.lt e'1 e'2))
      end.
    
    Lemma eq_refl : forall (x:t),(eq x x).
      unfold eq;reflexivity.
    Qed.

    Lemma eq_sym : forall (x y:t),(eq x y)->(eq y x).
      unfold eq;auto.
    Qed.

    Lemma eq_trans : forall (x y z:t),(eq x y)->(eq y z)->(eq x z).
      unfold eq;intros;congruence.
    Qed.

    Lemma lt_trans : forall (x y z:t),(lt x y)->(lt y z)->(lt x z).
      unfold lt.
      destruct x;destruct y;destruct z;intros.
      case H;case H0;intros;and_elim.
      left;eapply E.Ord.lt_trans;eauto.
      left;rewrite <- H1;trivial.
      left;rewrite H2;trivial.
      right.
      split.
      rewrite H2;auto.
      eapply E.Ord.lt_trans;eauto.
    Qed.

    Lemma lt_not_eq : forall (x y:t),(lt x y)->~(eq x y).
      unfold lt, eq.
      destruct x;destruct y;intros.
      red;intro.
      injection H0;intros.
      case H;intro;and_elim.
      elim (E.Ord.lt_not_eq t0 t2);auto.
      elim (E.Ord.lt_not_eq t1 t3);auto.
    Qed.

    Definition compare : forall (x y:t),(Compare lt eq x y).
      unfold lt, eq.
      destruct x; destruct y.
      case (E.Ord.compare t0 t2);intro.
      apply LT;tauto.
      case (E.Ord.compare t1 t3);intro.
      apply LT;tauto.
      apply EQ.
      rewrite e;rewrite e0;trivial.
      apply GT.
      right.
      split;auto.
      apply GT.
      tauto.
    Defined.

    Hint Immediate eq_sym.
    Hint Resolve eq_refl eq_trans lt_not_eq lt_trans.
  End Ord.

  Definition t_map (f:nat->E.t->E.t) (d:nat)(phi:t) :=
    match phi with
      | FEqual e1 e2 => FEqual (f d e1) (f d e2)
    end.

  Lemma t_map_struct : deBruijn.map_struct E.t t t_map.
    split;unfold t_map;try (destruct x;repeat (rewrite H);trivial).
    intros.
    unfold injective.
    destruct x;destruct y;intros.
    injection H0;intros.
    rewrite (H d t0 t2);auto.
    rewrite (H d t1 t3);auto.
  Qed.

  Definition deBruijn_base := deBruijn.deBruijn_map_struct E.t E.deBruijn_base t t_map t_map_struct.

  Definition deBruijn_subst := deBruijn.deBruijn_s_map_struct E.t E.deBruijn_base E.t E.deBruijn_base E.deBruijn_subst t t_map t_map_struct.

  Definition t_fold (A:Set)(f:nat->E.t->A->A)(d:nat)(x:t)(a:A) :=
    match x with FEqual e1 e2 => (f d e2 (f d e1 a)) end.

  Lemma t_fold_false_inv : forall (f:nat->E.t->bool)(d:nat)(x:t)(b:bool),(t_fold bool (fun d x b => b || f d x) d x b = false)->b=false.
    destruct x;simpl;intros;auto.
    elim (orb_false_elim (b || f d t0) (f d t1) H).
    intros.
    elim (orb_false_elim b (f d t0) H0);auto.
  Qed.

  Implicit Arguments t_fold_false_inv [f d x b].

  Lemma t_fold_true : forall (f:nat->E.t->bool)(d:nat)(x:t),(t_fold bool (fun d x b => b || f d x) d x true = true).
    destruct x;simpl;intros;auto.
  Qed.

  Lemma t_fold_struct : deBruijn.fold_struct E.t t t_map t_map_struct t_fold.
    split;intros;destruct x;simpl;auto.
    repeat (rewrite H);intros;auto.
    simpl in H0.
    generalize (H0 i H1);intro.
    destruct (fi i (d1+d) t0);simpl in H2;try (discriminate H2);auto.
    simpl in H0.
    generalize (H0 i H1);intro.
    elim (orb_false_elim (fi i (d1+d) t0) (fi i (d1+d) t1) H2);auto.
    simpl in H0.
    case (Bool_elim_bool (f (d1+d) t0));intro.
    rewrite (H d t0);simpl;auto.
    rewrite H1 in H0;simpl in H0.
    rewrite (H d t1 H0).
    rewrite orb_b_true;auto.
    repeat (rewrite <- H);auto;intros;simpl in H0.
    generalize (H1 i H3);intro.
    simpl in H4.
    elim (orb_false_elim (fi i (d1+d) t0) (fi i (d1+d) t1) H4);auto.
    generalize (H1 i H3);intro.
    simpl in H4.
    elim (orb_false_elim (fi i (d1+d) t0) (fi i (d1+d) t1) H4);auto.
    generalize (H1 i H3);intro.
    simpl in H4.
    elim (orb_false_elim (fi i (d1+d) t0) (fi i (d1+d) t1) H4);auto.
    (* fold_orb_eq_map_inversion *)
    destruct y;try (discriminate H2).
    simpl in H2.
    simpl in H0;simpl in H1.
    injection H2;clear H2;intros.
    rewrite (H d t1 t3);intros;auto.
    rewrite (H d t0 t2);intros;auto.
    generalize (H0 i H4);intros.
    elim (orb_false_elim (fi i (d1 + d) t0) (fi i (d1 + d) t1));auto. 
    elim (orb_false_elim (fi i (d1 + d) t2) (fi i (d1 + d) t3));auto. 
    elim (orb_false_elim (fi i (d1 + d) t0) (fi i (d1 + d) t1));auto. 
    elim (orb_false_elim (fi i (d1 + d) t2) (fi i (d1 + d) t3));auto. 
  Qed.
  
  Definition deBruijn_fn_base := deBruijn.deBruijn_fn_map_struct E.t t t_map t_map_struct t_fold E.deBruijn_base E.deBruijn_fn_base t_fold_struct.
  Definition deBruijn_fn_subst := deBruijn.deBruijn_s_fn_fold_struct E.t E.deBruijn_base E.deBruijn_fn_base E.t E.deBruijn_base E.deBruijn_fn_base E.deBruijn_subst E.deBruijn_fn_subst t t_map t_map_struct t_fold t_fold_struct.

  Lemma low_subst_lemma : forall (x:t)(d d' k k':nat)(es fs:list E.t),deBruijn.low_subst deBruijn_subst (d+d'+k) k' es (deBruijn.low_subst deBruijn_subst d k fs x)=deBruijn.low_subst deBruijn_subst d k (map (deBruijn.low_subst E.deBruijn_subst (d'+k) k' es) fs) (deBruijn.low_subst deBruijn_subst (d+d'+length fs) k' es x).
    intros.
    apply (SubstLemmaMap.low_subst_lemma E.t E.deBruijn_base E.t E.deBruijn_base E.deBruijn_subst t t_map t_map_struct E.deBruijn_subst E.low_subst_lemma).
  Qed.

  Lemma low_subst_split_lemma : forall (x:t)(d k:nat)(es fs:list E.t),deBruijn.low_subst deBruijn_subst d 0 es (deBruijn.low_subst deBruijn_subst (d+length es) k fs x)=deBruijn.low_subst deBruijn_subst d k (app es fs) x.
    intros.
    apply (SubstSplitLemmaMap.low_subst_split_lemma E.t E.deBruijn_base E.t E.deBruijn_base E.deBruijn_subst t t_map t_map_struct E.low_subst_split_lemma).
  Qed.

  Lemma mem_fn_FEqual : forall (d k:nat)(e f:ExpressionPi.t),(deBruijn.mem_fn deBruijn_fn_base k d (FEqual e f))=(orb (deBruijn.mem_fn ExpressionPi.deBruijn_fn_base k d e) (deBruijn.mem_fn ExpressionPi.deBruijn_fn_base k d f)).
    intros;reflexivity.
  Qed.

  Definition compute_fn (d:nat)(x:t) : ListSet.set nat :=
    match x with
      | FEqual e1 e2 => ListSet.set_union eq_nat_dec (E.compute_fn d e1) (E.compute_fn d e2)
    end.

  Lemma compute_fn_ok : forall (k:nat)(d:nat)(x:t),(deBruijn.mem_fn deBruijn_fn_base k d x = ListSet.set_mem eq_nat_dec k (compute_fn d x)).
    destruct x;intros.
    rewrite mem_fn_FEqual.
    lazy beta delta [compute_fn] iota;fold compute_fn.
    rewrite ListSet_mem_union.
    repeat (rewrite E.compute_fn_ok);auto.
  Qed.
End FormulaePi.




