(* A formalisation of the spi calculus in Coq *)
(* Copyright (C) 2005-2007 Sebastien Briais *)
(* http://lamp.epfl.ch/~sbriais/ *)

(* This library is free software; you can redistribute it and/or modify *)
(* it under the terms of the GNU Lesser General Public License as *)
(* published by the Free Software Foundation; either version 2.1 of the *)
(* License, or (at your option) any later version. *)

(* This library is distributed in the hope that it will be useful, but *)
(* WITHOUT ANY WARRANTY; without even the implied warranty of *)
(* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU *)
(* Lesser General Public License for more details. *)

(* You should have received a copy of the GNU Lesser General Public *)
(* License along with this library; if not, write to the Free Software *)
(* Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA *)
(* 02110-1301 USA *)

Require Import Arith.
Require Import missing.
Require Import formulaeType.
Require Import expressionSpi.
Require Import List.
Require Import tactics.
Require Import deBruijnType.
Require Import name.
Require Import substLemma.
Require Import substSplit.

Module FormulaeSpi <: FormulaeType with Module E := ExpressionSpi.
  Module E := ExpressionSpi.

  Inductive formulae : Set := 
    | FName : E.t -> formulae
    | FEqual : E.t -> E.t -> formulae.

  Definition t := formulae.

  Definition Wff (e:E.t) : t := (FEqual e e).

  Module Ord <: OrderedType with Definition t := t with Definition eq := @eq t.
    Definition t := t.
    Definition eq := @eq t.
    Definition lt (phi1 phi2:t) := 
      match (phi1,phi2) with
	| (FName e,FName f) => E.Ord.lt e f
	| (FName _,_) => True
	| (_,FName _) => False
	| (FEqual e1 e'1,FEqual e2 e'2) => (E.Ord.lt e1 e2) \/ ((E.Ord.eq e1 e2) /\ (E.Ord.lt e'1 e'2))
      end.
    
    Lemma eq_refl : forall (x:t),(eq x x).
      unfold eq;reflexivity.
    Qed.

    Lemma eq_sym : forall (x y:t),(eq x y)->(eq y x).
      unfold eq;auto.
    Qed.

    Lemma eq_trans : forall (x y z:t),(eq x y)->(eq y z)->(eq x z).
      unfold eq;intros;congruence.
    Qed.

    Lemma lt_trans : forall (x y z:t),(lt x y)->(lt y z)->(lt x z).
      unfold lt.
      destruct x;destruct y;destruct z;simpl;try tauto;intros;eauto.
      case H;case H0;intros;and_elim;try (left;eauto;fail);try (left;(rewrite <- H1 || rewrite H2);auto;fail).
      right;split;eauto.
    Qed.

    Lemma lt_not_eq : forall (x y:t),(lt x y)->~(eq x y).
      unfold lt, eq.
      destruct x;destruct y;intros;try (intro H0;discriminate H0);red;intro;injection H0;clear H0;intros.
      elim (E.Ord.lt_not_eq t0 t1);auto.
      case H;clear H;intros;and_elim.
      elim (E.Ord.lt_not_eq t0 t2);auto.
      elim (E.Ord.lt_not_eq t1 t3);auto.
    Qed.

    Definition compare : forall (x y:t),(Compare lt eq x y).
      unfold lt, eq.
      destruct x; destruct y;try (apply LT;simpl;auto;fail);try (apply GT;simpl;auto;fail).
      case (E.Ord.compare t0 t1);intro;[apply LT | apply EQ | apply GT];auto;rewrite e;auto.
      case (E.Ord.compare t0 t2);intro;[apply LT | idtac | apply GT];auto.
      case (E.Ord.compare t1 t3);intro;[apply LT | idtac | apply GT];auto.
      apply EQ;rewrite e;rewrite e0;auto.
    Defined.

    Hint Immediate eq_sym.
    Hint Resolve eq_refl eq_trans lt_not_eq lt_trans.
  End Ord.

  Definition t_map (f:nat->E.t->E.t) (d:nat)(phi:t) :=
    match phi with
      | FName e => FName (f d e)
      | FEqual e1 e2 => FEqual (f d e1) (f d e2)
    end.

  Lemma t_map_struct : deBruijn.map_struct E.t t t_map.
    split;unfold t_map;try (destruct x;repeat (rewrite H);trivial).
    intros.
    unfold injective.
    destruct x;destruct y;intros;try (discriminate H0);injection H0;intros.
    rewrite (H d t0 t1);auto.
    rewrite (H d t0 t2);auto.
    rewrite (H d t1 t3);auto.
  Qed.

  Definition deBruijn_base := deBruijn.deBruijn_map_struct E.t E.deBruijn_base t t_map t_map_struct.

  Definition deBruijn_subst := deBruijn.deBruijn_s_map_struct E.t E.deBruijn_base E.t E.deBruijn_base E.deBruijn_subst t t_map t_map_struct.

  Definition t_fold (A:Set)(f:nat->E.t->A->A)(d:nat)(x:t)(a:A) : A :=
    match x with
      | FName e => f d e a
      | FEqual e1 e2 => f d e2 (f d e1 a)
    end.

  Lemma t_fold_struct : deBruijn.fold_struct E.t t t_map t_map_struct t_fold.
    split;intros.
    (* fold_orb_eq *)
    destruct x;simpl;repeat (rewrite H);intros;simpl in H0;auto.
    generalize (H0 i H1);intro.
    destruct (fi i (d1+d) t0);simpl in H2;try (discriminate H2);auto.
    generalize (H0 i H1);intro.
    elim (orb_false_elim (fi i (d1+d) t0) (fi i (d1+d) t1) H2);auto.
    (* fold_exists *)
    destruct x;simpl;intros;simpl in H0;auto.
    generalize (H d t0);generalize (H d t1);intros.
    destruct (f (d1+d) t0);simpl in H0;try (discriminate H0).
    rewrite H2;simpl;auto.
    destruct (g (d2+d) t0);simpl;auto.
    (* fold_map_simpl *)
    destruct x;simpl;intros;auto.
    (* fold_equal_cond_invariant *)
    destruct x;simpl;repeat (rewrite <- H);intros;simpl in H0;auto;try (generalize (H1 i H3);intro H4;simpl in H4);auto.
    elim (orb_false_elim (fi i (d1+d) t0) (fi i (d1+d) t1) H4);auto.
    elim (orb_false_elim (fi i (d1+d) t0) (fi i (d1+d) t1) H4);auto.
    elim (orb_false_elim (fi i (d1+d) t0) (fi i (d1+d) t1) H4);auto.
    (* fold_invariant *)
    destruct x;simpl;intros;auto.
    (* fold_orb_eq_map_inversion *)
    destruct x;destruct y;try (discriminate H2);simpl in H0;simpl in H1;simpl in H2;injection H2;clear H2;intros.
    rewrite (H d t0 t1);intros;auto.
    rewrite (H d t1 t3);intros;auto;try (elim (orb_false_elim (fi i (d1 + d) t0) (fi i (d1 + d) t1) (H0 i H4));auto;fail);try (elim (orb_false_elim (fi i (d1 + d) t2) (fi i (d1 + d) t3) (H1 i H4));auto;fail).
    rewrite (H d t0 t2);intros;auto;try (elim (orb_false_elim (fi i (d1 + d) t0) (fi i (d1 + d) t1) (H0 i H4));auto;fail);try (elim (orb_false_elim (fi i (d1 + d) t2) (fi i (d1 + d) t3) (H1 i H4));auto;fail).
  Qed.

  Definition deBruijn_fn_base := deBruijn.deBruijn_fn_map_struct E.t t t_map t_map_struct t_fold E.deBruijn_base E.deBruijn_fn_base t_fold_struct.

  Lemma lift_FName : forall (d k:nat)(e:ExpressionSpi.t),(deBruijn.lift deBruijn_base d k (FormulaeSpi.FName e))=(FName (deBruijn.lift ExpressionSpi.deBruijn_base d k e)).
    intros;reflexivity.
  Qed.

  Lemma lift_FEqual : forall (d k:nat)(e f:ExpressionSpi.t),(deBruijn.lift deBruijn_base d k (FormulaeSpi.FEqual e f))=(FEqual (deBruijn.lift ExpressionSpi.deBruijn_base d k e) (deBruijn.lift ExpressionSpi.deBruijn_base d k f)).
    intros;reflexivity.
  Qed.

  Lemma swapR_FName : forall (d k:nat)(e:ExpressionSpi.t),(deBruijn.swapR deBruijn_base d k (FormulaeSpi.FName e))=(FName (deBruijn.swapR ExpressionSpi.deBruijn_base d k e)).
    intros;reflexivity.
  Qed.

  Lemma swapR_FEqual : forall (d k:nat)(e f:ExpressionSpi.t),(deBruijn.swapR deBruijn_base d k (FormulaeSpi.FEqual e f))=(FEqual (deBruijn.swapR ExpressionSpi.deBruijn_base d k e) (deBruijn.swapR ExpressionSpi.deBruijn_base d k f)).
    intros;reflexivity.
  Qed.

  Lemma low1_FName : forall (d:nat)(e:ExpressionSpi.t),(deBruijn.low1 deBruijn_base d (FormulaeSpi.FName e))=(FName (deBruijn.low1 ExpressionSpi.deBruijn_base d e)).
    intros;reflexivity.
  Qed.

  Lemma low1_FEqual : forall (d:nat)(e f:ExpressionSpi.t),(deBruijn.low1 deBruijn_base d (FormulaeSpi.FEqual e f))=(FEqual (deBruijn.low1 ExpressionSpi.deBruijn_base d e) (deBruijn.low1 ExpressionSpi.deBruijn_base d f)).
    intros;reflexivity.
  Qed.

  Lemma low_subst_FName : forall (d k:nat)(sigma:list ExpressionSpi.t)(e:ExpressionSpi.t),(deBruijn.low_subst deBruijn_subst d k sigma (FormulaeSpi.FName e))=(FormulaeSpi.FName (deBruijn.low_subst ExpressionSpi.deBruijn_subst d k sigma e)).
    intros;simpl.
    auto.
  Qed.

  Lemma low_subst_FEqual : forall (d k:nat)(sigma:list ExpressionSpi.t)(e1 e2:ExpressionSpi.t),(deBruijn.low_subst deBruijn_subst d k sigma (FormulaeSpi.FEqual e1 e2))=(FormulaeSpi.FEqual (deBruijn.low_subst ExpressionSpi.deBruijn_subst d k sigma e1) (deBruijn.low_subst ExpressionSpi.deBruijn_subst d k sigma e2)).
    intros;simpl.
    auto.
  Qed.

  Lemma low_subst_Wff : forall (d k:nat)(es:list ExpressionSpi.t)(e:ExpressionSpi.t),(deBruijn.low_subst deBruijn_subst d k es (Wff e))=(Wff (deBruijn.low_subst ExpressionSpi.deBruijn_subst d k es e)).
    intros.
    simpl.
    auto.
  Qed.

  Lemma mem_fn_FName : forall (d k:nat)(e:ExpressionSpi.t),(deBruijn.mem_fn deBruijn_fn_base k d (FName e))=(deBruijn.mem_fn ExpressionSpi.deBruijn_fn_base k d e).
    intros;reflexivity.
  Qed.
    
  Lemma mem_fn_FEqual : forall (d k:nat)(e f:ExpressionSpi.t),(deBruijn.mem_fn deBruijn_fn_base k d (FEqual e f))=(orb (deBruijn.mem_fn ExpressionSpi.deBruijn_fn_base k d e) (deBruijn.mem_fn ExpressionSpi.deBruijn_fn_base k d f)).
    intros;reflexivity.
  Qed.

  Definition deBruijn_fn_subst := deBruijn.deBruijn_s_fn_fold_struct E.t E.deBruijn_base E.deBruijn_fn_base E.t E.deBruijn_base E.deBruijn_fn_base E.deBruijn_subst E.deBruijn_fn_subst t t_map t_map_struct t_fold t_fold_struct.

  Lemma low_subst_lemma : forall (x:t)(d d' k k':nat)(es fs:list E.t),deBruijn.low_subst deBruijn_subst (d+d'+k) k' es (deBruijn.low_subst deBruijn_subst d k fs x)=deBruijn.low_subst deBruijn_subst d k (map (deBruijn.low_subst E.deBruijn_subst (d'+k) k' es) fs) (deBruijn.low_subst deBruijn_subst (d+d'+length fs) k' es x).
    intros.
    apply (SubstLemmaMap.low_subst_lemma E.t E.deBruijn_base E.t E.deBruijn_base E.deBruijn_subst t t_map t_map_struct E.deBruijn_subst E.low_subst_lemma).
  Qed.

  Lemma low_subst_split_lemma : forall (x:t)(d k:nat)(es fs:list E.t),deBruijn.low_subst deBruijn_subst d 0 es (deBruijn.low_subst deBruijn_subst (d+length es) k fs x)=deBruijn.low_subst deBruijn_subst d k (app es fs) x.
    intros.
    apply (SubstSplitLemmaMap.low_subst_split_lemma E.t E.deBruijn_base E.t E.deBruijn_base E.deBruijn_subst t t_map t_map_struct E.low_subst_split_lemma).
  Qed.

  Definition compute_fn (d:nat)(x:t) : ListSet.set nat :=
    match x with
      | FEqual e1 e2 => ListSet.set_union eq_nat_dec (E.compute_fn d e1) (E.compute_fn d e2)
      | FName e => E.compute_fn d e
    end.

  Lemma compute_fn_ok : forall (k:nat)(d:nat)(x:t),(deBruijn.mem_fn deBruijn_fn_base k d x = ListSet.set_mem eq_nat_dec k (compute_fn d x)).
    destruct x;intros;repeat (rewrite mem_fn_FEqual || rewrite mem_fn_FName);lazy beta delta [compute_fn] iota;fold compute_fn;try (rewrite ListSet_mem_union);repeat (rewrite E.compute_fn_ok);auto.
  Qed.
End FormulaeSpi.

