(* A formalisation of the spi calculus in Coq *)
(* Copyright (C) 2005-2007 Sebastien Briais *)
(* http://lamp.epfl.ch/~sbriais/ *)

(* This library is free software; you can redistribute it and/or modify *)
(* it under the terms of the GNU Lesser General Public License as *)
(* published by the Free Software Foundation; either version 2.1 of the *)
(* License, or (at your option) any later version. *)

(* This library is distributed in the hope that it will be useful, but *)
(* WITHOUT ANY WARRANTY; without even the implied warranty of *)
(* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU *)
(* Lesser General Public License for more details. *)

(* You should have received a copy of the GNU Lesser General Public *)
(* License along with this library; if not, write to the Free Software *)
(* Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA *)
(* 02110-1301 USA *)

Require Import deBruijnType.
Require Import name.
Require Import messageSpi.
Require Import expressionSpi.
Require Import missing.
Require Import ordmissing.
Require Import tactics.
Require Import FSets.
Require Import hedges.
Require Import calculi.
Require Import evalSpi.
Require Import substSpi.

Module HedgedRelations (H:Hedge).
  (* Auxiliary definitions for defining hedged bisimulations *)
  (** Definition of early hedged bisimulation *)
  Definition subst_input (m:MessageSpi.message)(p:AgentSpi.P.t) := deBruijn.low_subst AgentSpi.P.deBruijn_subst 0 0 (cons (EvalSpi.expr_of_mesg m) nil) p.

  Definition instantiate_message (l:list Name.t)(m:MessageSpi.message) := deBruijn.low_subst MessageSpi.deBruijn_subst 0 0 (map MessageSpi.j l) m.

  Lemma instantiate_message_nil : forall (m:MessageSpi.message),(instantiate_message nil m = m).
    unfold instantiate_message.
    intros.
    generalize (deBruijn.low_subst_lift_nil MessageSpi.deBruijn_subst 0 0 0 nil).
    repeat (rewrite plus_0_l).
    lazy beta delta [map length] iota.
    fold MessageSpi.deBruijn_base.
    intros.
    generalize (H m).
    rewrite deBruijn.lift_0.
    auto.
  Qed.

  Lemma instantiate_message_expr : forall (l:list Name.t)(m:MessageSpi.message),EvalSpi.expr_of_mesg (instantiate_message l m) = deBruijn.low_subst ExpressionSpi.deBruijn_subst 0 0 (map ExpressionSpi.j l) (EvalSpi.expr_of_mesg m).
    unfold instantiate_message.
    intros.
    rewrite (SubstSpi.low_subst_expr_of_msg m 0 0 (map MessageSpi.j l)).
    unfold SubstSpi.msg_subst.
    rewrite map_map.
    reflexivity.
  Qed.

  Definition instantiate_process (l:list Name.t)(p:AgentSpi.P.t) := deBruijn.low_subst AgentSpi.P.deBruijn_subst 0 0 (map ExpressionSpi.j l) p.

  Lemma instantiate_process_nil : forall (p:AgentSpi.P.t),(instantiate_process nil p = p).
    unfold instantiate_process.
    intros.
    generalize (deBruijn.low_subst_lift_nil AgentSpi.P.deBruijn_subst 0 0 0 nil).
    repeat (rewrite plus_0_l).
    lazy beta delta [map length] iota.
    intros.
    generalize (H p).
    rewrite deBruijn.lift_0.
    auto.
  Qed.

  (* Involve hedges *)
  Module HOp := HedgeOp(H).

  Definition possible_inputs (h:H.t)(B:H.t)(m n:MessageSpi.message) := 
    HOp.are_fresh_names h B /\
    HOp.hedge_synthesis_relation (HOp.relation_of_hedge (H.union h B)) m n /\
    forall (b1 b2:nat),
      (H.In (MessageSpi.MNam (Name.Ref b1),MessageSpi.MNam (Name.Ref b2)) B)
      ->(deBruijn.mem_fn MessageSpi.deBruijn_fn_base b1 0 m = true \/ deBruijn.mem_fn MessageSpi.deBruijn_fn_base b2 0 n = true).

  Lemma possible_inputs_Subset : forall (h B:H.t)(h' B':H.t),(H.Subset h h')->(HOp.are_fresh_names h' B')->(H.Equal B B')->forall (m n:MessageSpi.message),(possible_inputs h B m n)->(possible_inputs h' B' m n).
    unfold possible_inputs;intros;and_elim.
    split;intros;auto.
    split;intros;auto.
    generalize HOp.incl_impl_le_h.
    unfold HOp.le_h.
    intros.
    eapply H5;eauto.
    unfold HOp.relation_of_hedge.
    red;intros.
    case (H.union_1 H6);intros.
    apply H.union_2;auto.
    apply H.union_3;auto.
    elim (H1 (x,y));auto.
    apply H4.
    elim (H1 (MessageSpi.MNam (Name.Ref b1), MessageSpi.MNam (Name.Ref b2))).
    intros;auto.
  Qed.

  Lemma possible_inputs_mem_fn : forall (h:H.t)(B:H.t)(m n:MessageSpi.message),(possible_inputs h B m n)->forall (b1 b2:nat),(H.In (MessageSpi.MNam (Name.Ref b1),MessageSpi.MNam (Name.Ref b2)) B)->(deBruijn.mem_fn MessageSpi.deBruijn_fn_base b1 0 m = true /\ deBruijn.mem_fn MessageSpi.deBruijn_fn_base b2 0 n = true).
    intros.
    unfold possible_inputs in H.
    elim H;clear H.
    intro Hfresh;intros.
    and_elim.
    generalize (H1 b1 b2 H0);clear H1;intro.
    induction H.
    unfold HOp.relation_of_hedge in H.
    case (H.union_1 H).
    intros.
    unfold HOp.are_fresh_names in Hfresh.
    generalize Hfresh;clear Hfresh.
    intros;and_elim.
    elim (H4 b1 b2 H0 m n);auto;intros.
    case H1;intro.
    rewrite H5 in H7;discriminate H7.
    rewrite H6 in H7;discriminate H7.
    intro.
    elim Hfresh;clear Hfresh;intros;and_elim.
    elim (H3 m n H2);intro a1;intros.
    ex_elim H6 a2.
    case H1;auto;intro.
    rewrite H8;split;auto.
    subst m.
    rewrite MessageSpi.mem_fn_MNam in H8.
    destruct a1 as [a1].
    generalize (Name.fn_0_name_unique b1 a1 H8);intro.
    subst a1.
    elim (H4 (MessageSpi.MNam (Name.Ref b1)) (MessageSpi.MNam (Name.Ref b2)) H0 (MessageSpi.MNam (Name.Ref b1)) n H2).
    intros.
    rewrite <- H6;auto.
    rewrite MessageSpi.mem_fn_MNam.
    apply Name.fn_0_name.
    rewrite H8;split;auto.
    subst n.
    rewrite MessageSpi.mem_fn_MNam in H8.
    destruct a2 as [a2].
    generalize (Name.fn_0_name_unique b2 a2 H8);intro.
    subst a2.
    elim (H4 (MessageSpi.MNam (Name.Ref b1)) (MessageSpi.MNam (Name.Ref b2)) H0 m (MessageSpi.MNam (Name.Ref b2)) H2).
    intros.
    rewrite <- H9;auto.
    rewrite MessageSpi.mem_fn_MNam.
    apply Name.fn_0_name.
    generalize H1.
    repeat (rewrite MessageSpi.mem_fn_MEnc_s).
    intros.
    case H3;clear H3;intro.
    case (orb_prop _ _ H3);clear H3;intro.
    rewrite H3.
    split;auto.
    elim IHhedge_synthesis_relation1;intros.
    rewrite H5;auto.
    left;auto.
    rewrite H3.
    rewrite orb_b_true;split;auto.
    elim IHhedge_synthesis_relation2;intros.
    rewrite H5.
    rewrite orb_b_true;auto.
    left;auto.
    case (orb_prop _ _ H3);clear H3;intro.
    elim IHhedge_synthesis_relation1;intros;auto.
    rewrite H4;rewrite H5;auto.
    elim IHhedge_synthesis_relation2;intros;auto.
    rewrite H4;rewrite H5;repeat (rewrite orb_b_true);auto.
    generalize H1.
    repeat (rewrite MessageSpi.mem_fn_MEnc_a).
    intros.
    case H3;clear H3;intro.
    case (orb_prop _ _ H3);clear H3;intro.
    elim IHhedge_synthesis_relation1;intros;auto.
    rewrite H4;rewrite H5;auto.
    elim IHhedge_synthesis_relation2;intros;auto.
    rewrite H4;rewrite H5.
    repeat (rewrite orb_b_true);auto.
    case (orb_prop _ _ H3);clear H3;intro.
    elim IHhedge_synthesis_relation1;intros;auto.
    rewrite H4;rewrite H5;auto.
    elim IHhedge_synthesis_relation2;intros;auto.
    rewrite H4;rewrite H5;repeat (rewrite orb_b_true);auto.
    generalize H1.
    repeat (rewrite MessageSpi.mem_fn_MPair).
    intros.
    case H3;clear H3;intro.
    case (orb_prop _ _ H3);clear H3;intro.
    elim IHhedge_synthesis_relation1;intros;auto.
    rewrite H4;rewrite H5;auto.
    elim IHhedge_synthesis_relation2;intros;auto.
    rewrite H4;rewrite H5.
    repeat (rewrite orb_b_true);auto.
    case (orb_prop _ _ H3);clear H3;intro.
    elim IHhedge_synthesis_relation1;intros;auto.
    rewrite H4;rewrite H5;auto.
    elim IHhedge_synthesis_relation2;intros;auto.
    rewrite H4;rewrite H5;repeat (rewrite orb_b_true);auto.
    generalize H1.
    repeat (rewrite MessageSpi.mem_fn_MOp).
    intros.
    apply IHhedge_synthesis_relation.
    auto.
  Qed.

  Lemma possible_inputs_transpose : forall (h:H.t)(B:H.t)(m n:MessageSpi.message),(possible_inputs h B m n)->(possible_inputs (HOp.hedge_transpose h) (HOp.hedge_transpose B) n m).
    unfold possible_inputs;intros.
    elim H;clear H.
    intro Hfresh;intros;and_elim.
    split.
    apply HOp.are_fresh_names_transpose.
    apply Hfresh.
    split.
    apply HOp.hedge_in_synthesis_impl_hedge_synthesis_spec.
    rewrite <- (HOp.in_synthesis_Equal _ _ (HOp.hedge_transpose_union h B)).
    rewrite HOp.in_synthesis_transpose.
    apply HOp.hedge_synthesis_spec_impl_hedge_in_synthesis.
    auto.
    intros.
    elim (HOp.hedge_transpose_prop B (MessageSpi.MNam (Name.Ref b2)) (MessageSpi.MNam (Name.Ref b1))).
    intros.
    generalize (H0 _ _ (H3 H1)).
    tauto.
  Qed.

  (* hedged relations *)
  Definition hedged_relation_type := H.t * AgentSpi.P.t * AgentSpi.P.t -> Prop.

  Definition empty_hedged_relation : hedged_relation_type := fun x => False.

  Definition inclusion_hedged_relation (R R':hedged_relation_type) := 
    forall (h:H.t)(p q:AgentSpi.P.t),(R (h,p,q))->(exists h':H.t,(H.Equal h h')/\(R' (h',p,q))).

  Lemma inclusion_empty : forall (R:hedged_relation_type),(inclusion_hedged_relation empty_hedged_relation R).
    unfold inclusion_hedged_relation;intros.
    elim H.
  Qed.

  Definition same_hedged_relation (R R':hedged_relation_type) := (inclusion_hedged_relation R R')/\(inclusion_hedged_relation R' R).

  Lemma inclusion_hedged_relation_refl : forall (R:hedged_relation_type),(inclusion_hedged_relation R R).
    unfold inclusion_hedged_relation;intros.
    exists h.
    split;auto.
    apply HOp.HedgeProp.equal_refl.
  Qed.

  Lemma inclusion_hedged_relation_trans : forall (R1 R2 R3:hedged_relation_type),(inclusion_hedged_relation R1 R2)->(inclusion_hedged_relation R2 R3)->(inclusion_hedged_relation R1 R3).
    unfold inclusion_hedged_relation;intros.
    elim (H _ _ _ H1);intro h1;intros.
    and_elim.
    elim (H0 _ _ _ H3);intro h2;intros.
    and_elim.
    exists h2.
    split;auto.
    eapply HOp.HedgeProp.equal_trans;eauto.
  Qed.

  Lemma inclusion_hedged_relation_antisym : forall (R R':hedged_relation_type),(inclusion_hedged_relation R R')->(inclusion_hedged_relation R' R)->(same_hedged_relation R R').
    intros.
    unfold same_hedged_relation.
    split;auto.
  Qed.

  Lemma same_hedged_relation_refl : forall (R:hedged_relation_type),(same_hedged_relation R R).
    unfold same_hedged_relation.
    intros.
    split;apply inclusion_hedged_relation_refl.
  Qed.

  Lemma same_hedged_relation_sym : forall (R R':hedged_relation_type),(same_hedged_relation R R')->(same_hedged_relation R' R).
    unfold same_hedged_relation;intros;and_elim.
    split;auto.
  Qed.

  Lemma same_hedged_relation_trans : forall (R1 R2 R3:hedged_relation_type),(same_hedged_relation R1 R2)->(same_hedged_relation R2 R3)->(same_hedged_relation R1 R3).
    unfold same_hedged_relation;intros;and_elim.
    split;eapply inclusion_hedged_relation_trans;eauto.
  Qed.

  Definition hedged_relation_family_type := nat -> hedged_relation_type.

  Definition union_hedged_relation_family (Rn:hedged_relation_family_type) : hedged_relation_type := 
    fun x => exists n:nat,Rn n x.

  Lemma inclusion_union_hedged_relation_family : forall (Rn:hedged_relation_family_type),forall (n:nat),(inclusion_hedged_relation (Rn n) (union_hedged_relation_family Rn)).
    unfold inclusion_hedged_relation;intros.
    exists h.
    split.
    apply HOp.HedgeProp.equal_refl.
    unfold union_hedged_relation_family.
    exists n.
    auto.
  Qed.

  Definition union_hedged_relation (R R':hedged_relation_type) : hedged_relation_type := fun x => (R x) \/ (R' x).

  Lemma inclusion_union_1 : forall (R R':hedged_relation_type),inclusion_hedged_relation R (union_hedged_relation R R').
    unfold inclusion_hedged_relation, union_hedged_relation;intros.
    exists h.
    split;auto.
    apply HOp.HedgeProp.equal_refl.
  Qed.

  Lemma inclusion_union_2 : forall (R R':hedged_relation_type),inclusion_hedged_relation R' (union_hedged_relation R R').
    unfold inclusion_hedged_relation, union_hedged_relation;intros.
    exists h.
    split;auto.
    apply HOp.HedgeProp.equal_refl.
  Qed.

  Lemma inclusion_union_3 : forall (R R' S:hedged_relation_type),inclusion_hedged_relation R S->inclusion_hedged_relation R' S->inclusion_hedged_relation (union_hedged_relation R R') S.
    unfold inclusion_hedged_relation;intros.
    case H1;intro;auto.
  Qed.

  Lemma union_hedged_relation_comm : forall (R R':hedged_relation_type),(same_hedged_relation (union_hedged_relation R R') (union_hedged_relation R' R)).
    unfold same_hedged_relation.
    intros.
    split.
    apply inclusion_union_3.
    apply inclusion_union_2.
    apply inclusion_union_1.
    apply inclusion_union_3.
    apply inclusion_union_2.
    apply inclusion_union_1.
  Qed.

  Lemma same_hedged_relation_union_hedged_relation_family : forall (R R':hedged_relation_type),(same_hedged_relation (union_hedged_relation_family (fun n => match n with O => R | S O => R' | _ => empty_hedged_relation end)) (union_hedged_relation R R')).
    intros.
    split;unfold inclusion_hedged_relation;intros.
    elim H.
    intro i;intros.
    destruct i.
    exists h.
    split.
    apply HOp.HedgeProp.equal_refl.
    left.
    auto.
    destruct i.
    exists h.
    split.
    apply HOp.HedgeProp.equal_refl.
    right;auto.
    elim H0.
    exists h.
    split.
    apply HOp.HedgeProp.equal_refl.
    case H.
    intro.
    exists 0.
    auto.
    intro.
    exists 1.
    auto.
  Qed.

  (** Auxiliary definitions for constraining hedged relations *)

  (** symmetric hedged relation *)
  Definition symmetric_hedged_relation (R:hedged_relation_type) :=
    forall (h:H.t)(p q:AgentSpi.P.t),(R (h,p,q))->exists hinv:H.t,(H.Equal hinv (HOp.hedge_transpose h))/\(R (hinv,q,p)).

  Lemma empty_hedged_relation_symmetric : symmetric_hedged_relation empty_hedged_relation.
    unfold symmetric_hedged_relation;intros.
    elim H.
  Qed.

  Lemma same_hedged_relation_symmetric : forall (R R':hedged_relation_type),(same_hedged_relation R R')->(symmetric_hedged_relation R)->(symmetric_hedged_relation R').
    unfold symmetric_hedged_relation.
    intros.
    elim H.
    intros.
    elim (H3 _ _ _ H1).
    intro h';intros;and_elim.
    elim (H0 _ _ _ H5).
    intro h'inv;intros;and_elim.
    elim (H2 _ _ _ H7);intro hinv;intros;and_elim.
    exists hinv.
    split;auto.
    eapply HOp.HedgeProp.equal_trans.
    apply HOp.HedgeProp.equal_sym.
    apply H8.
    eapply HOp.HedgeProp.equal_trans.
    apply H6.
    apply HOp.hedge_transpose_morphism.
    apply HOp.HedgeProp.equal_sym.
    auto.
  Qed.

  Lemma union_hedged_relation_family_symmetric : forall (R:hedged_relation_family_type),(forall (n:nat),symmetric_hedged_relation (R n))->symmetric_hedged_relation (union_hedged_relation_family R).
    unfold symmetric_hedged_relation;intros.
    elim H0.
    intro n;intros.
    elim (H _ _ _ _ H1).
    intro hinv;intros;and_elim.
    exists hinv.
    split;auto.
    exists n.
    auto.
  Qed.

  Lemma union_hedged_relation_symmetric : forall (R R':hedged_relation_type),(symmetric_hedged_relation R)->(symmetric_hedged_relation R')->symmetric_hedged_relation (union_hedged_relation R R').
    intros.
    eapply same_hedged_relation_symmetric.
    apply same_hedged_relation_union_hedged_relation_family.
    apply union_hedged_relation_family_symmetric.
    intro n.
    destruct n.
    auto.
    destruct n.
    auto.
    apply empty_hedged_relation_symmetric.
  Qed.

  (** hedged relation (adding contraints on free names) *)

  Definition fn (x:nat)(p:AgentSpi.P.t) := deBruijn.mem_fn AgentSpi.P.deBruijn_fn_base x 0 p = true.

  Definition hedged_relation (R:hedged_relation_type) := 
    forall (h:H.t)(p q:AgentSpi.P.t),(R (h,p,q))->(forall (x:nat),(fn x p)->HOp.is_left_name_spec x h) /\ (forall (x:nat),(fn x q)->HOp.is_right_name_spec x h).

  Lemma empty_hedged_relation_hedged : hedged_relation empty_hedged_relation.
    unfold hedged_relation.
    intros.
    elim H.
  Qed.

  Lemma inclusion_hedged_relation_hedged_relation : forall (R R':hedged_relation_type),(inclusion_hedged_relation R R')->(hedged_relation R')->(hedged_relation R).
    unfold hedged_relation;intros.
    elim (H _ _ _ H1).
    intro h';intros;and_elim.
    elim (H0 _ _ _ H3);intros.
    split;intros.
    generalize (HOp.same_is_left_name _ _ (HOp.Equal_impl_same _ _ H2));intros.
    elim (H7 x);intros.
    apply H9.
    apply H4.
    auto.
    generalize (HOp.same_is_right_name _ _ (HOp.Equal_impl_same _ _ H2));intros.
    elim (H7 x);intros.
    apply H9.
    apply H5.
    auto.
  Qed.

  Lemma union_hedged_relation_family_hedged : forall (R:hedged_relation_family_type),(forall (n:nat),hedged_relation (R n))->hedged_relation (union_hedged_relation_family R).
    unfold hedged_relation;intros.
    elim H0.
    intro n;intros.
    eauto.
  Qed.

  Lemma union_hedged_relation_hedged : forall (R R':hedged_relation_type),(hedged_relation R)->(hedged_relation R')->hedged_relation (union_hedged_relation R R').
    intros.
    eapply inclusion_hedged_relation_hedged_relation.
    elim (same_hedged_relation_union_hedged_relation_family R R').
    intros.
    apply H2.
    apply union_hedged_relation_family_hedged.
    intro n.
    destruct n;auto.
    destruct n;auto.
    apply empty_hedged_relation_hedged.
  Qed.

  Definition left_hedged_relation (R:hedged_relation_type) := 
    forall (h:H.t)(p q:AgentSpi.P.t),(R (h,p,q))->(forall (x:nat),(fn x p)->HOp.is_left_name_spec x h).

  Lemma empty_hedged_relation_left_hedged : left_hedged_relation empty_hedged_relation.
    unfold left_hedged_relation;intros.
    elim H.
  Qed.

  Lemma inclusion_hedged_relation_left_hedged_relation : forall (R R':hedged_relation_type),(inclusion_hedged_relation R R')->(left_hedged_relation R')->(left_hedged_relation R).
    unfold left_hedged_relation;intros.
    elim (H _ _ _ H1).
    intro h';intros;and_elim.
    generalize (H0 _ _ _ H4 _ H2);intros.
    generalize (HOp.same_is_left_name _ _ (HOp.Equal_impl_same _ _ H3));intros.
    elim (H6 x);intros.
    apply H8.
    apply (H0 _ _ _ H4 _ H2).
  Qed.

  Lemma symmetric_left_hedged_relation_is_hedged_relation : forall (R:hedged_relation_type),(symmetric_hedged_relation R)->(left_hedged_relation R)->(hedged_relation R).
    unfold left_hedged_relation, symmetric_hedged_relation, hedged_relation;intros.
    split;intros.
    eauto.
    elim (HOp.is_right_name_spec_iff_is_left_name_transpose x h).
    intros.
    apply H4.
    elim (H _ _ _ H1);intro hinv;intros.
    and_elim.
    elim (HOp.same_is_left_name _ _ (HOp.Equal_impl_same _ _ H5) x);intros.
    apply H7.
    apply (H0 _ _ _ H6 _ H2).
  Qed.

  Lemma union_hedged_relation_family_left_hedged : forall (R:hedged_relation_family_type),(forall (n:nat),left_hedged_relation (R n))->left_hedged_relation (union_hedged_relation_family R).
    unfold left_hedged_relation;intros.
    elim H0.
    intro n;intros.
    eapply H;eauto.
  Qed.

  Lemma union_hedged_relation_left_hedged : forall (R R':hedged_relation_type),(left_hedged_relation R)->(left_hedged_relation R')->left_hedged_relation (union_hedged_relation R R').
    intros.
    eapply inclusion_hedged_relation_left_hedged_relation.
    elim (same_hedged_relation_union_hedged_relation_family R R').
    intros.
    apply H2.
    apply union_hedged_relation_family_left_hedged.
    intro n.
    destruct n;auto.
    destruct n;auto.
    apply empty_hedged_relation_left_hedged.
  Qed.
   
  (** consistent and left consistent hedged relation *)
  Definition consistent_hedged_relation (R:hedged_relation_type) := 
    forall (h:H.t)(p q:AgentSpi.P.t),(R (h,p,q))->(HOp.is_consistent h).

  Lemma inclusion_hedged_relation_consistent : forall (R R':hedged_relation_type),(inclusion_hedged_relation R R')->(consistent_hedged_relation R')->(consistent_hedged_relation R).
    unfold consistent_hedged_relation;intros.
    elim (H _ _ _ H1);intro h';intros.
    and_elim.
    unfold HOp.is_consistent.
    eapply HOp.same_impl_is_consistent_relation_impl.
    apply HOp.Equal_impl_same.
    apply HOp.HedgeProp.equal_sym.
    apply H2.
    apply (H0 _ _ _ H3).
  Qed.

  Definition left_consistent_hedged_relation (R:hedged_relation_type) := 
    forall (h:H.t)(p q:AgentSpi.P.t),(R (h,p,q))->(HOp.is_left_consistent h).

  Lemma empty_hedged_relation_left_consistent : left_consistent_hedged_relation empty_hedged_relation.
    unfold left_consistent_hedged_relation.
    intros.
    elim H.
  Qed.

  Lemma inclusion_hedged_relation_left_consistent : forall (R R':hedged_relation_type),(inclusion_hedged_relation R R')->(left_consistent_hedged_relation R')->(left_consistent_hedged_relation R).
    unfold left_consistent_hedged_relation;intros.
    elim (H _ _ _ H1);intro h';intros.
    and_elim.
    unfold HOp.is_left_consistent.
    eapply HOp.same_impl_is_left_consistent_relation_impl.
    apply HOp.Equal_impl_same.
    apply HOp.HedgeProp.equal_sym.
    apply H2.
    apply (H0 _ _ _ H3).
  Qed.

  Lemma union_hedged_relation_family_left_consistent : forall (R:hedged_relation_family_type),(forall (n:nat),left_consistent_hedged_relation (R n))->left_consistent_hedged_relation (union_hedged_relation_family R).
    unfold left_consistent_hedged_relation;intros.
    elim H0.
    intro n;intros.
    eapply H;eauto.
  Qed.

  Lemma union_hedged_relation_left_consistent : forall (R R':hedged_relation_type),(left_consistent_hedged_relation R)->(left_consistent_hedged_relation R')->left_consistent_hedged_relation (union_hedged_relation R R').
    intros.
    eapply inclusion_hedged_relation_left_consistent.
    elim (same_hedged_relation_union_hedged_relation_family R R').
    intros.
    apply H2.
    apply union_hedged_relation_family_left_consistent.
    intro n.
    destruct n;auto.
    destruct n;auto.
    apply empty_hedged_relation_left_consistent.
  Qed.

  Lemma symmetric_left_consistent_hedged_relation_is_consistent_hedged_relation : 
    forall (R:hedged_relation_type),(symmetric_hedged_relation R)->(left_consistent_hedged_relation R)->(consistent_hedged_relation R).
    unfold left_consistent_hedged_relation, symmetric_hedged_relation, consistent_hedged_relation;intros.
    elim (HOp.is_consistent_iff h);intros.
    apply H3.
    split;intros.
    eauto.
    elim (H _ _ _ H1);intro hinv.
    intros;and_elim.
    generalize (H0 _ _ _ H5);intro.
    unfold HOp.is_left_consistent.
    eapply HOp.same_impl_is_left_consistent_relation_impl.
    apply HOp.Equal_impl_same;eauto.
    auto.
  Qed.  

  Inductive mk_symm_hedged_relation (R:hedged_relation_type) : hedged_relation_type :=
  | inR : forall (h:H.t)(p q:AgentSpi.P.t),(R (h,p,q))->(mk_symm_hedged_relation R (h,p,q))
  | inInvR : forall (h:H.t)(p q:AgentSpi.P.t),(R (h,p,q))->(mk_symm_hedged_relation R (HOp.hedge_transpose h,q,p)).

  Lemma mk_symm_hedged_relation_symmetric : forall (R:hedged_relation_type),(symmetric_hedged_relation (mk_symm_hedged_relation R)).
    unfold symmetric_hedged_relation.
    intros.
    inversion H.
    subst h0;subst p0;subst q0.
    exists (HOp.hedge_transpose h).
    split.
    apply HOp.HedgeProp.equal_refl.
    apply inInvR.
    auto.
    subst p0;subst q0.
    generalize (HOp.hedge_transpose_involutive h0).
    intro.
    exists h0.
    split;auto.
    apply inR.
    auto.
  Qed.

  Lemma inclusion_hedged_relation_mk_symm_hedged_relation : forall (R R':hedged_relation_type),(inclusion_hedged_relation R R')->(inclusion_hedged_relation (mk_symm_hedged_relation R) (mk_symm_hedged_relation R')).
    intros.
    red;intros.
    inversion H0.
    subst h0;subst p0;subst q0.
    elim (H _ _ _ H2);intro h';intros.
    and_elim.
    exists h'.
    split;auto.
    apply inR.
    auto.
    subst p0;subst q0.
    elim (H _ _ _ H2);intro h0';intros.
    and_elim.
    exists (HOp.hedge_transpose h0').
    split.
    apply HOp.hedge_transpose_morphism.
    auto.
    apply inInvR.
    auto.
  Qed.

  Definition symmetric_left_consistent_hedged_relation (R:hedged_relation_type) := (symmetric_hedged_relation R)/\(left_hedged_relation R)/\(left_consistent_hedged_relation R).

  Lemma empty_hedged_relation_symmetric_left_consistent : symmetric_left_consistent_hedged_relation empty_hedged_relation.
    split.
    apply empty_hedged_relation_symmetric.
    split.
    apply empty_hedged_relation_left_hedged.
    apply empty_hedged_relation_left_consistent.
  Qed.

  Lemma inclusion_hedged_relation_symmetric_left_consistent_hedged_relation : 
    forall (R S:hedged_relation_type),(inclusion_hedged_relation R S)->(symmetric_left_consistent_hedged_relation S)->(symmetric_hedged_relation R)->(symmetric_left_consistent_hedged_relation R).
    unfold symmetric_left_consistent_hedged_relation;intros;and_elim.
    split;[idtac | split].
    auto.
    eapply inclusion_hedged_relation_left_hedged_relation;eauto.
    eapply inclusion_hedged_relation_left_consistent;eauto.
  Qed.

  Lemma same_hedged_relation_symmetric_left_consistent_hedged_relation : forall (R R':hedged_relation_type),(same_hedged_relation R R')->(symmetric_left_consistent_hedged_relation R)->(symmetric_left_consistent_hedged_relation R').
    intros.
    elim H;intros.
    eapply inclusion_hedged_relation_symmetric_left_consistent_hedged_relation;eauto.
    eapply same_hedged_relation_symmetric.
    apply H.
    elim H0;auto.
  Qed.

  Lemma union_hedged_relation_family_symmetric_left_consistent_hedged : forall (R:hedged_relation_family_type),(forall (n:nat),symmetric_left_consistent_hedged_relation (R n))->symmetric_left_consistent_hedged_relation (union_hedged_relation_family R).
    unfold symmetric_left_consistent_hedged_relation;intros.
    split.
    apply union_hedged_relation_family_symmetric.
    intro.
    elim (H n);auto.
    split.
    apply union_hedged_relation_family_left_hedged.
    intro.
    elim (H n);tauto.
    apply union_hedged_relation_family_left_consistent.
    intro.
    elim (H n);tauto.
  Qed.

  Lemma union_hedged_relation_symmetric_left_consistent_hedged : forall (R R':hedged_relation_type),(symmetric_left_consistent_hedged_relation R)->(symmetric_left_consistent_hedged_relation R')->symmetric_left_consistent_hedged_relation (union_hedged_relation R R').
    intros.
    eapply same_hedged_relation_symmetric_left_consistent_hedged_relation.
    apply same_hedged_relation_union_hedged_relation_family.
    apply union_hedged_relation_family_symmetric_left_consistent_hedged.
    intro n.
    destruct n;auto.
    destruct n;auto.
    apply empty_hedged_relation_symmetric_left_consistent.
  Qed.

  Definition hedged_relation_operator := hedged_relation_type -> hedged_relation_type.

  Definition hedged_relation_operator_correct (f:hedged_relation_operator) := forall (R:hedged_relation_type),(symmetric_left_consistent_hedged_relation R)->(symmetric_left_consistent_hedged_relation (f R)).

  Fixpoint hedged_relation_operator_iter (f:hedged_relation_operator)(R:hedged_relation_type)(n:nat) {struct n} : hedged_relation_type :=
    match n with
      | O => R
      | (S n) => union_hedged_relation (hedged_relation_operator_iter f R n) (f (hedged_relation_operator_iter f R n))
    end.

  Definition hedged_relation_operator_limit (f:hedged_relation_operator)(R:hedged_relation_type) : hedged_relation_type := 
    union_hedged_relation_family (hedged_relation_operator_iter f R).

  Lemma inclusion_hedged_relation_operator_iter : forall (f:hedged_relation_operator)(R:hedged_relation_type)(n:nat),(inclusion_hedged_relation (hedged_relation_operator_iter f R n) (hedged_relation_operator_iter f R (S n))).
    intros.
    simpl.
    apply inclusion_union_1.
  Qed.

  Lemma hedged_relation_operator_correct_iter : forall (f:hedged_relation_operator),(hedged_relation_operator_correct f)->forall (R:hedged_relation_type),(symmetric_left_consistent_hedged_relation R)->forall (n:nat),(symmetric_left_consistent_hedged_relation (hedged_relation_operator_iter f R n)).
    intros.
    induction n.
    simpl;auto.
    simpl.
    apply union_hedged_relation_symmetric_left_consistent_hedged.
    auto.
    apply H.
    auto.
  Qed.

  Lemma hedged_relation_operator_correct_limit : forall (f:hedged_relation_operator),(hedged_relation_operator_correct f)->forall (R:hedged_relation_type),(symmetric_left_consistent_hedged_relation R)->(symmetric_left_consistent_hedged_relation (hedged_relation_operator_limit f R)).
    intros.
    unfold hedged_relation_operator_limit.
    apply union_hedged_relation_family_symmetric_left_consistent_hedged.
    intros.
    apply hedged_relation_operator_correct_iter;auto.
  Qed.

  Definition f_structural_congruence (R:hedged_relation_type) : hedged_relation_type := 
    fun hpq =>
      match hpq with
        (h,p',q') =>
        exists p:AgentSpi.P.t,exists q:AgentSpi.P.t,
          (R (h,p,q))/\
          CongruenceSpi.Rcongruence CongruenceSpi.StructuralCongruenceAxioms p' p/\
          CongruenceSpi.Rcongruence CongruenceSpi.StructuralCongruenceAxioms q' q
      end.

  Lemma f_structural_congruence_correct : hedged_relation_operator_correct f_structural_congruence.
    unfold hedged_relation_operator_correct.
    unfold symmetric_left_consistent_hedged_relation.
    intros;and_elim.
    split.
    unfold symmetric_hedged_relation.
    intros.
    elim H2.
    intro p'.
    intro.
    ex_elim H3 q'.
    elim (H _ _ _ H3).
    intro hinv;intros.
    exists hinv.
    and_elim.
    split;auto.
    exists q'.
    exists p'.
    split;auto.
    split.
    unfold left_hedged_relation.
    intros.
    elim H2.
    intro p';intros.
    ex_elim H4 q'.
    apply (H0 _ _ _ H4 x).
    unfold fn.
    rewrite <- (CongruenceSpi.Rcongruence_mem_fn _ (CongruenceSpi.StructuralCongruenceAxioms_preserves_mem_fn) _ _ H5).
    apply H3.
    unfold left_consistent_hedged_relation.
    intros.
    elim H2.
    intro p';intros.
    ex_elim H3 q'.
    apply (H1 _ _ _ H3).
  Qed.

  Lemma f_structural_congruence_inclusion : forall (R S:hedged_relation_type),(inclusion_hedged_relation R S)->(inclusion_hedged_relation (f_structural_congruence R) (f_structural_congruence S)).
    unfold inclusion_hedged_relation;intros.
    elim H0.
    intro p';intros.
    ex_elim H1 q'.
    elim (H _ _ _ H1).
    intro h';intros.
    and_elim.
    exists h'.
    split;auto.
    exists p'.
    exists q'.
    split;auto.
  Qed.
End HedgedRelations.










