(* A formalisation of the spi calculus in Coq *)
(* Copyright (C) 2005-2007 Sebastien Briais *)
(* http://lamp.epfl.ch/~sbriais/ *)

(* This library is free software; you can redistribute it and/or modify *)
(* it under the terms of the GNU Lesser General Public License as *)
(* published by the Free Software Foundation; either version 2.1 of the *)
(* License, or (at your option) any later version. *)

(* This library is distributed in the hope that it will be useful, but *)
(* WITHOUT ANY WARRANTY; without even the implied warranty of *)
(* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU *)
(* Lesser General Public License for more details. *)

(* You should have received a copy of the GNU Lesser General Public *)
(* License along with this library; if not, write to the Free Software *)
(* Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA *)
(* 02110-1301 USA *)

Require Export ListSet.
Require Export List.
Require Export Bool.
Require Export Relations.
Require Import Omega.
Require Import Max.
Require Import tactics.

Definition injective (t u:Set)(f:t->u) := forall (x y:t),(f x = f y)->x = y.
Implicit Arguments injective [t u].

Lemma Arith_minus_minus : forall (n m p:nat),(n-m-p)=(n-(m+p)).
  induction n;intros;simpl;auto.
  destruct m;destruct p;simpl;auto.
Qed.

Lemma Arith_plus_minus : forall (n m:nat),n+m-m = n.
intros.
induction m.
rewrite plus_comm;simpl.
auto with arith.
rewrite plus_comm;simpl.
rewrite plus_comm;trivial.
Qed.

Lemma Arith_plus_minus_plus : forall (n m p:nat),(n+p)-(m+p) = n-m.
induction p.
repeat (rewrite plus_comm;simpl).
trivial.
rewrite plus_comm.
rewrite (plus_comm m).
simpl.
rewrite plus_comm.
rewrite (plus_comm p);trivial.
Qed.

Lemma Arith_le_minus_le : forall (n m p:nat),(n+m <= p)->n<=(p-m).
  intros.
  omega.
Qed.

Lemma Arith_le_plus_minus : forall (n m p:nat),(n-p <= n+m-p).
  intros.
  omega.
Qed.

Lemma Arith_plus_lt : forall (n m p:nat),(p+n < p+m) -> n < m.
  induction p;intros.
  simpl in H;trivial.
  simpl in H;trivial.
  apply IHp.
  inversion H.
  auto with arith.
  auto with arith.
Qed.

Lemma ListSet_mem_union : forall (A:Set)(Aeq_dec: forall x y:A, {x = y} + {x <> y})(x:A)(l m:set A),(set_mem Aeq_dec x (set_union Aeq_dec l m))=orb (set_mem Aeq_dec x l) (set_mem Aeq_dec x m).
intros.
generalize (set_mem_correct1 Aeq_dec x (set_union Aeq_dec l m));intro.
generalize (set_mem_complete1 Aeq_dec x (set_union Aeq_dec l m));intro.
destruct (set_mem Aeq_dec x (set_union Aeq_dec l m)).
assert ((set_In x l)\/(set_In x m)).
eapply set_union_elim;eauto.
generalize (set_mem_correct2 Aeq_dec x);intro.
case H1.
intro.
rewrite (H2 l);try reflexivity;trivial.
intro.
rewrite (H2 m);trivial.
rewrite orb_b_true;trivial.
generalize (set_mem_correct1 Aeq_dec x l);intro.
generalize (set_mem_correct1 Aeq_dec x m);intro.
destruct (set_mem Aeq_dec x l);destruct (set_mem Aeq_dec x m);trivial;elim H0;trivial;apply set_union_intro.
left;apply H1;trivial.
left;apply H1;trivial.
right;apply H2;trivial.
Qed.

Lemma ListSet_mem_add : forall (A:Set)(Aeq_dec: forall x y:A, {x = y} + {x <> y})(l:set A)(x:A)(y:A),(set_mem Aeq_dec x (set_add Aeq_dec y l))=orb (if Aeq_dec x y then true else false) (set_mem Aeq_dec x l).
intros.
generalize (set_mem_correct1 Aeq_dec x (set_add Aeq_dec y l)).
generalize (set_mem_complete1 Aeq_dec x (set_add Aeq_dec y l)).
intros.
destruct (set_mem Aeq_dec x (set_add Aeq_dec y l)).
assert ((x=y)\/(set_In x l)).
eapply set_add_elim;eauto.
case H1;intro.
case Aeq_dec;intro;simpl;trivial.
elim n;trivial.
generalize (set_mem_correct2 Aeq_dec x l);intro.
rewrite H3;trivial.
rewrite orb_b_true;trivial.
generalize (set_mem_correct1 Aeq_dec x l);intro.
case Aeq_dec;intro;simpl.
elim H;trivial.
apply set_add_intro2;trivial.
destruct (set_mem Aeq_dec x l);trivial.
elim H;trivial.
apply set_add_intro1;auto.
Qed.

Lemma ListSet_mem_empty : forall (A:Set)(Aeq_dec: forall x y:A, {x = y} + {x <> y})(x:A),(set_mem Aeq_dec x (empty_set A))=false.
  intros;reflexivity.
Qed.

Lemma Bool_elim_bool : forall (b:bool),b=true \/ b=false.
  destruct b;try tauto.
Qed.

Lemma Bool_eq_dec : forall (b:bool),{b=true}+{b=false}.
  destruct b;[left | right];trivial.
Qed.

Lemma Pair_eq_dec : forall (A B:Set),(forall (x y:A),{x=y}+{x<>y})->(forall (x y:B),{x=y}+{x<>y})->forall (x y:A*B),{x=y}+{x<>y}.
intros.
destruct x.
destruct y.
case (H a a0);case (H0 b b0);intros;try (left;congruence);right;intro H1;injection H1;auto.
Qed.

Require Import Arith.
Require Import Max.

Lemma Max_lt_max_lt : forall (n m p:nat),(n<p)->(m<p)->(max n m)<p.
  intros.
  apply max_case2;trivial.
Qed.

Lemma Max_assoc : forall (n m p:nat),(max (max n m) p)=(max n (max m p)).
  intros.
  case (le_lt_dec n m);intro.
  rewrite (max_r n m);trivial.
  assert (n<=(max m p)).
  eapply le_trans;eauto with arith.
  rewrite (max_r n (max m p));trivial.
  rewrite (max_l n m);auto with arith.
  case (le_lt_dec m p);intro.
  rewrite (max_r m p);trivial.
  assert (p <= n);eauto with arith.
  rewrite (max_l n p);trivial.
  rewrite (max_l m p);auto with arith.
  rewrite (max_l n m);auto with arith.
Qed.

Require Import Setoid.

Definition Bool_setoid : Setoid.Setoid_Theory bool (@eq bool).
  split;intros;auto.
  congruence.
Qed.
  
Lemma Option_elim_option : forall (A:Set)(x:option A),{y:A | x = (Some y)}+{x=None}.
  intros.
  destruct x.
  left;exists a;trivial.
  right;trivial.
Qed.

Lemma case_plus_gt_0 : forall (n m:nat),(n+m> 0)->(n>0)\/(m>0).
  intros.
  case (le_lt_dec n 0);intro;try tauto.
  case (le_lt_dec m 0);intro;try tauto.
  inversion l.
  inversion l0.
  rewrite H0 in H.
  rewrite H1 in H.
  simpl in H.
  inversion H.
Qed.

Lemma List_nth_Some : forall (A:Set)(l:list A)(i:nat),(i < length l)<->exists x:A, nth_error l i = Some x.
  split.
  generalize i;clear i.
  induction l;simpl;intros.
  inversion H.
  destruct i.
  simpl.
  exists a;trivial.
  simpl.
  apply IHl;omega.
  generalize l;clear l.
  induction i;simpl;intros.
  destruct l.
  elim H;intros.
  discriminate H0.
  simpl;auto with arith.
  destruct l.
  elim H;intros.
  discriminate H0.
  simpl.
  generalize (IHi l H).
  omega.
Qed.

Lemma List_nth_None : forall (A:Set)(l:list A)(i:nat),(length l <= i)->(nth_error l i = None).
  intros.
  elim (List_nth_Some _ l i);intros.
  destruct (nth_error l i);auto.
  assert (i < length l).
  apply H1.
  exists a;auto.
  try_absurd_arith.
Qed.

Lemma List_nth_nil : forall (A:Set)(i:nat),(nth_error (A:=A) nil i = None).
  destruct i;reflexivity.
Qed.

Lemma List_length_map : forall (A B:Set)(f:A->B)(l:list A),(length (map f l))=(length l).
intros.
induction l;simpl.
trivial.
congruence.
Qed.

Definition option_map (A B:Set)(f:A->B)(x:option A) : option B :=
  match x with
    | None => None
    | (Some x) => (Some (f x))
  end.

Implicit Arguments option_map [A B].

Lemma List_nth_map : forall (A B:Set)(f:A->B)(i:nat)(l:list A),(nth_error (map f l) i)=(option_map f (nth_error l i)).
  induction i;intros;simpl.
  destruct l;reflexivity.
  destruct l;try reflexivity.
  simpl.
  apply IHi.
Qed.

Definition List_left_rotate (A:Set)(l:list A) : list A :=
  match l with
    | nil => nil
    | (cons x xs) => app xs (cons x nil)
  end.

Fixpoint List_right_rotate (A:Set)(l:list A) {struct l} : list A :=
  match l with 
    | nil => nil
    | (cons x xs) => 
      (match (List_right_rotate A xs) with
	 | nil => (cons x nil)
	 | (cons y ys) => (cons y (cons x ys))
       end)
  end.

Implicit Arguments List_left_rotate [A].
Implicit Arguments List_right_rotate [A].

Lemma List_right_rotate_simpl : forall (A:Set)(a:A)(l:list A),(List_right_rotate (app l (cons a nil)))=(cons a l).
  intros.
  generalize a;clear a.
  induction l;simpl;try reflexivity.
  intros.
  rewrite IHl.
  trivial.
Qed.

Lemma List_right_left_rotate_id : forall (A:Set)(l:list A),(List_right_rotate (List_left_rotate l))=l.
  induction l;simpl;trivial.
  rewrite List_right_rotate_simpl.
  trivial.
Qed.

Lemma List_left_right_rotate_id : forall (A:Set)(l:list A),(List_left_rotate (List_right_rotate l))=l.
  intros.
  rewrite <- (rev_involutive l).
  elim (rev l) using rev_list_ind.
  reflexivity.
  simpl;intros.
  rewrite List_right_rotate_simpl.
  reflexivity.
Qed.

Lemma List_length_append : forall (A:Set)(l l':list A),(length (app l l'))=(length l)+(length l').
  induction l;simpl;intros;trivial.
  rewrite IHl.
  trivial.
Qed.
  
Implicit Arguments List_length_append [A].

Lemma List_length_left_rotate : forall (A:Set)(l:list A),(length l)=(length (List_left_rotate l)).
  destruct l;simpl;intros;trivial.
  rewrite List_length_append.
  simpl.
  rewrite plus_comm;reflexivity.
Qed.

Lemma List_length_right_rotate : forall (A:Set)(l:list A),(length l)=(length (List_right_rotate l)).
  intros.
  rewrite (List_length_left_rotate A (List_right_rotate l)).
  rewrite List_left_right_rotate_id.
  trivial.
Qed.

Lemma List_left_rotate_nth_O : forall (A:Set)(l:list A),l<>nil->(nth_error l 0)=(nth_error (List_left_rotate l) (pred (length l))).
  simpl.
  induction l;try reflexivity;intros.
  simpl.
  clear IHl.
  clear H.
  induction l.
  reflexivity.
  simpl.
  rewrite <- IHl.
  trivial.
Qed.

Lemma List_nth_app_left : forall (A:Set)(l l':list A)(i:nat),(i < length l)->(nth_error (app l l') i)=(nth_error l i).
  induction l.
  simpl;intros.
  inversion H.
  simpl;intros.
  induction i.
  reflexivity.
  simpl.
  rewrite IHl;trivial;try omega.
Qed.

Lemma List_nth_app_right : forall (A:Set)(l l':list A)(i:nat),(nth_error (app l l') (i+length l))=(nth_error l' i).
  induction l;intros;rewrite plus_comm.
  simpl;intros.
  auto.
  simpl.
  rewrite plus_comm;auto.
Qed.

Lemma List_left_rotate_nth_Some : forall (A:Set)(l:list A)(i:nat),(0 < i)->(i < length l)->(nth_error l i)=(nth_error (List_left_rotate l) (pred i)).
  induction l;simpl;intros.
  elim (lt_irrefl 0);omega.
  destruct i.
  inversion H.
  simpl.
  assert (i < length l);try omega.
  inversion H.
  subst i.
  destruct l.
  simpl in H1.
  elim (lt_irrefl 0);omega.
  reflexivity.
  rewrite List_nth_app_left;trivial.
Qed.

Lemma List_right_rotate_nth_O : forall (A:Set)(l:list A),l<>nil->(nth_error (List_right_rotate l) 0)=(nth_error l (pred (length l))).
  intros.
  rewrite List_left_rotate_nth_O.
  rewrite List_left_right_rotate_id.
  rewrite <- List_length_right_rotate;trivial.
  generalize (List_length_right_rotate A l).
  destruct l;intros.
  simpl;trivial.
  intro.
  assert ((length (List_right_rotate (a::l))) = (length (A:=A) nil)).
  rewrite H1.
  trivial.
  rewrite <- H0 in H2.
  discriminate H2.
Qed.

Lemma List_right_rotate_nth_Some : forall (A:Set)(l:list A)(i:nat),(0 < i)->(i < length l)->(nth_error (List_right_rotate l) i)=(nth_error l (pred i)).
  intros.
  rewrite List_left_rotate_nth_Some.
  rewrite List_left_right_rotate_id.
  trivial.
  trivial.
  rewrite <- List_length_right_rotate;trivial.
Qed.

Lemma List_nil_dec : forall (A:Set)(l:list A),{l=nil}+{l<>nil}.
  destruct l;intros.
  left;trivial.
  right.
  intro.
  discriminate H.
Qed.

Lemma max_assoc : forall (n m p:nat),(max (max n m) p)=(max n (max m p)).
  intros.
  case (max_dec n m);case (max_dec n p);case (max_dec m p);intros;repeat (rewrite e || rewrite e0 || rewrite e1);trivial;generalize (le_max_l n m);generalize (le_max_l n p);generalize (le_max_l m p);generalize (le_max_r n m);generalize (le_max_r n p);generalize (le_max_r m p);intros.
  eapply le_antisym;omega.
  eapply le_antisym;omega.
Qed.  

Lemma union_symmetric : forall (t:Set)(R R':relation t),(symmetric t R)->(symmetric t R')->(symmetric t (union t R R')).
  unfold symmetric;unfold union;intros.
  case H1;[left | right];auto.
Qed.

Lemma same_relation_symmetric : forall (t:Set)(R R':relation t),(same_relation t R R')->(symmetric t R)->(symmetric t R').
  unfold symmetric;unfold same_relation;unfold inclusion;intros;and_elim;auto.
Qed.

Lemma same_relation_sym : forall (t:Set)(R R':relation t),(same_relation t R R')->(same_relation t R' R).
  unfold same_relation;intros;and_elim;tauto.
Qed.
  
Lemma same_relation_trans : forall (A:Set)(f g h:relation A),(same_relation A f g)->(same_relation A g h)->(same_relation A f h).
  unfold same_relation, inclusion;intros;and_elim.
  split;auto.
Qed.

Lemma union_comm : forall (A:Set)(f g:relation A),(same_relation _ (union _ f g) (union _ g f)).
  unfold same_relation, inclusion, union.
  split;intros;try tauto.
Qed.

Lemma apply_f : forall (t u:Set)(f:t->u)(x y:t),x=y->(f x=f y).
  intros.
  subst y;trivial.
Qed.

Implicit Arguments apply_f [t u x y].

Lemma pair_split : forall (t u:Set)(x:t*u),x=(fst x,snd x).
  destruct x;try reflexivity.
Qed.

Implicit Arguments pair_split [t u].

Lemma if_simpl : forall (A:Set)(b:bool)(th:A)(el:A)(res:A),(if b then th else el)=res -> (b = true /\ res = th) \/ (b = false /\ res = el).
  intros.
  destruct b.
  left.
  split;auto.
  right.
  split;auto.
Qed.

Implicit Arguments if_simpl [A b th el res].

Definition R_preserves_f (A:Set)(R:relation A) (f:A->A) := forall (x y:A),(R x y)->(R (f x) (f y)).

Implicit Arguments R_preserves_f [A].

Lemma R_preserves_f_union : forall (A:Set)(R R':relation A)(f:A->A),(R_preserves_f R f)->(R_preserves_f R' f)->(R_preserves_f (union A R R') f).
  unfold R_preserves_f;intros.
  case H1;intro;[left | right];auto.
Qed.

Implicit Arguments R_preserves_f_union [A].

Inductive List_forall (A:Set)(P:A->Prop) : list A -> Prop :=
  | List_forall_nil : List_forall A P nil
  | List_forall_cons : forall (x:A),(P x)->forall (xs:list A),(List_forall A P xs)->(List_forall A P (cons x xs)).

Lemma List_forall_nth : forall (A:Set)(P:A->Prop)(xs:list A),(List_forall A P xs)->forall (i:nat)(x:A),(nth_error xs i)=(Some x)->(P x).
  induction xs.
  intros.
  destruct i;try (discriminate H0).
  intros.
  inversion H.
  destruct i;simpl in H0.
  injection H0;intros.
  subst x;auto.
  eapply IHxs;eauto.
Qed.

Lemma List_nth_forall : forall (A:Set)(P:A->Prop)(xs:list A),(forall (i:nat)(x:A),(nth_error xs i)=(Some x)->(P x))->(List_forall A P xs).
  induction xs;intros.
  apply List_forall_nil.
  apply List_forall_cons.
  apply (H 0).
  reflexivity.
  apply IHxs.
  intros.
  apply (H (S i)).
  simpl.
  auto.
Qed.

Inductive List_relation (A:Set)(R:A->A->Prop) : list A -> list A -> Prop :=
  | List_relation_nil : List_relation A R nil nil
  | List_relation_cons : forall (x y:A),(R x y)->forall (xs ys:list A),(List_relation A R xs ys)->(List_relation A R (cons x xs) (cons y ys)).

Lemma List_relation_length : forall (A:Set)(R:A->A->Prop)(xs ys:list A),(List_relation A R xs ys)->(length xs = length ys).
  induction 1.
  reflexivity.
  simpl.
  rewrite IHList_relation;auto.
Qed.

Lemma List_relation_nth_Some : forall (A:Set)(R:A->A->Prop)(xs ys:list A),(List_relation A R xs ys)->forall (i:nat)(x:A),(nth_error xs i)=(Some x)->exists y:A,(R x y)/\(nth_error ys i)=(Some y).
  induction 1.
  intros.
  destruct i;discriminate H.
  intros.
  destruct i.
  simpl in H1.
  exists y.
  simpl.
  injection H1;intros;subst x0.
  auto.
  simpl in H1.
  elim (IHList_relation i x0 H1).
  intro y0;intros;and_elim.
  exists y0.
  simpl.
  auto.
Qed.

Lemma List_relation_nth_None : forall (A:Set)(R:A->A->Prop)(xs ys:list A),(List_relation A R xs ys)->forall (i:nat),(nth_error xs i)=None->(nth_error ys i)=None.
  induction 1;intros.
  auto.
  destruct i;try (discriminate H1).
  simpl in H1.
  simpl.
  auto.
Qed.

Implicit Arguments List_forall [A].
Implicit Arguments List_forall_nth [A].
Implicit Arguments List_nth_forall [A].

Implicit Arguments List_relation [A].
Implicit Arguments List_relation_length [A].
Implicit Arguments List_relation_nth_Some [A].
Implicit Arguments List_relation_nth_None [A].

Fixpoint max_nat_set (x:set nat) : nat :=
  match x with
    | nil => 0
    | cons x xs =>
      max x (max_nat_set xs)
  end.

Lemma max_nat_set_is_ge : forall (x:set nat),forall (y:nat),(In y x)->(le y (max_nat_set x)).
  intros.
  induction x.
  inversion H.
  simpl in H.
  simpl.
  case H.
  intros.
  subst a.
  apply le_max_l.
  intros.
  eapply le_trans.
  apply IHx.
  auto.
  apply le_max_r.
Qed.

Definition fresh_nat_exists : forall (x:set nat),{y:nat | ~In y x}.
  intros.
  exists (S (max_nat_set x)).
  red;intro.
  generalize (max_nat_set_is_ge _ _ H).
  intro.
  omega.
Defined.

Lemma List_length_S_inversion : forall (A:Set)(xss:list A)(n:nat),(length xss = S n)->exists xs:list A,exists x:A,xss = app xs (cons x nil).
  induction xss.
  intros.
  discriminate H.
  intros.
  simpl in H.
  injection H;intros.
  destruct n.
  destruct xss;try (discriminate H0).
  exists (nil (A:=A)).
  exists a.
  reflexivity.
  elim (IHxss _ H0).
  intro xs.
  intro.
  ex_elim H1 x.
  exists (cons a xs).
  exists x.
  rewrite H1.
  reflexivity.
Qed.
