(* A formalisation of the spi calculus in Coq *)
(* Copyright (C) 2005-2007 Sebastien Briais *)
(* http://lamp.epfl.ch/~sbriais/ *)

(* This library is free software; you can redistribute it and/or modify *)
(* it under the terms of the GNU Lesser General Public License as *)
(* published by the Free Software Foundation; either version 2.1 of the *)
(* License, or (at your option) any later version. *)

(* This library is distributed in the hope that it will be useful, but *)
(* WITHOUT ANY WARRANTY; without even the implied warranty of *)
(* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU *)
(* Lesser General Public License for more details. *)

(* You should have received a copy of the GNU Lesser General Public *)
(* License along with this library; if not, write to the Free Software *)
(* Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA *)
(* 02110-1301 USA *)

Require Import Arith.
Require Import FSets.
Require Import missing.
Require Import tactics.
Require Import deBruijnType.
Require Import ListSet.

Module Name.
  Inductive name : Set := Ref : nat -> name.
  Definition t := name.

  Module Ord <: OrderedType with Definition t := t with Definition eq := @eq t.
    Definition t := t.
    Definition eq := @eq t.
    Definition lt (n m:t) :=
      match n with (Ref n) => match m with (Ref m) => OrderedNat.lt n m end end.

    Lemma eq_refl : forall (x:t),(eq x x).
      unfold eq;reflexivity.
    Qed.

    Lemma eq_sym : forall (x y:t),(eq x y)->(eq y x).
      unfold eq;auto.
    Qed.

    Lemma eq_trans : forall (x y z:t),(eq x y)->(eq y z)->(eq x z).
      unfold eq;intros;congruence.
    Qed.

    Lemma lt_trans : forall (x y z:t),(lt x y)->(lt y z)->(lt x z).
      unfold lt.
      destruct x;destruct y;destruct z.
      apply OrderedNat.lt_trans.
    Qed.

    Lemma lt_not_eq : forall (x y:t),(lt x y)->~(eq x y).
      unfold lt, eq.
      destruct x;destruct y.
      intro;red;intro.
      elim (OrderedNat.lt_not_eq n n0);auto.
      injection H0;intros;subst n0.
      auto.
    Qed.

    Definition compare : forall (x y:t),(Compare lt eq x y).
      unfold lt, eq.
      destruct x;destruct y.
      case (OrderedNat.compare n n0);intro;[apply LT | apply EQ | apply GT];auto.
    Defined.

    Hint Immediate eq_sym.
    Hint Resolve eq_refl eq_trans lt_not_eq lt_trans.
  End Ord.

  Lemma name_equal : forall (n m:nat),(n=m)->(Ref n = Ref m).
    intros.
    congruence.
  Qed.

  Definition lift (d k:nat)(n:t) := match n with (Ref n) => Ref (DeBruijnNat.lift d k n) end.
    
  Definition swapR (d k:nat)(n:t) := match n with (Ref n) => Ref (DeBruijnNat.swapR d k n) end.

  Definition low1 (d:nat)(n:t) : t := match n with (Ref n) => Ref (DeBruijnNat.low1 d n) end.

  (* de Bruijn properties *)
  Definition deBruijn_base : deBruijn.deBruijn_base t.
    split with lift swapR low1;intros;try (destruct x;unfold lift, swapR, low1).
    (* lift_injective *)
    unfold injective.
    destruct x;destruct y;simpl;intros.
    injection H;clear H;intros.
    rewrite (DeBruijnNat.lift_injective d k n n0);auto.
    (* lift_0 *)
    rewrite DeBruijnNat.lift_0;auto.
    (* lift_plus *)
    rewrite DeBruijnNat.lift_plus;auto.
    (* lift_comm *)
    rewrite DeBruijnNat.lift_comm;auto.
    (* lift_compose *)
    rewrite DeBruijnNat.lift_compose;auto.
    (* swapR_injective *)
    unfold injective.
    destruct x;destruct y;simpl;intros.
    injection H;clear H;intros.
    rewrite (DeBruijnNat.swapR_injective d k n n0);auto.
    (* swapR_0 *)
    rewrite DeBruijnNat.swapR_0;auto.
    (* swapR_comm *)
    rewrite DeBruijnNat.swapR_comm;auto.
    (* swapR_1_idem *)
    rewrite DeBruijnNat.swapR_1_idem;auto.
    (* swapR_simpl *)
    rewrite DeBruijnNat.swapR_simpl;auto.
    (* swapR_compose *)
    rewrite DeBruijnNat.swapR_compose;auto.
    (* swapR_inside *)
    rewrite DeBruijnNat.swapR_inside;auto.
    (* lift_swapR_comm *)
    rewrite DeBruijnNat.lift_swapR_comm;auto.
    (* swapR_lift_comm *)
    rewrite DeBruijnNat.swapR_lift_comm;auto.
    (* swapR_lift_idem *)
    rewrite DeBruijnNat.swapR_lift_idem;auto.
    (* swapR_lift_simpl *)
    rewrite DeBruijnNat.swapR_lift_simpl;auto.
    (* swapR_lift_1 *)
    rewrite DeBruijnNat.swapR_lift_1;auto.
    (* low1_lift_simpl *)
    rewrite DeBruijnNat.low1_lift_lift;auto.
    (* swapR_occur_swap *)
    rewrite (DeBruijnNat.swapR_occur_swap k n1 n2 H H0 n'1 n'2 H1 d n);auto.
  Defined.

  Definition low_subst_generic (to:Set)(lift_to:nat->nat->to->to)(j:t->to)(d k:nat)(ts:list to)(n:t) : to :=
    DeBruijnNat.low_subst_gen to lift_to (fun n => j (Ref n)) d k ts (match n with Ref n => n end). 

  Definition deBruijn_subst_gen (to:Set)(deBruijn_to:deBruijn.deBruijn_base to)(Hextension:deBruijn.deBruijn_extension t deBruijn_base to deBruijn_to) : deBruijn.deBruijn_subst_gen t deBruijn_base to deBruijn_to Hextension.
    intros.
    split with (low_subst_generic to (deBruijn.lift deBruijn_to) (deBruijn.j Hextension)).
    (* low_subst_gen_lift_comm *)
    unfold low_subst_generic;destruct x;intros;simpl.
    apply DeBruijnNat.low_subst_gen_lift_comm;intros.
    rewrite deBruijn.j_lift;reflexivity.
    apply deBruijn.lift_compose.
    (* lift_low_subst_gen_succ *)
    unfold low_subst_generic;destruct x;intros;simpl.
    apply DeBruijnNat.lift_low_subst_gen_succ;intros.
    rewrite deBruijn.j_lift;reflexivity.
    apply deBruijn.lift_comm.
    (* low_subst_gen_lift_nil *)
    unfold low_subst_generic;destruct x;intros;simpl.
    rewrite DeBruijnNat.low_subst_gen_lift_nil;intros;auto.
    rewrite deBruijn.j_lift;reflexivity.
    (* lift_low_subst_gen_comm *)
    unfold low_subst_generic;destruct x;intros;simpl.
    rewrite DeBruijnNat.lift_low_subst_gen_comm;intros;auto.
    rewrite deBruijn.j_lift;reflexivity.
    apply deBruijn.lift_comm.
    (* low_subst_gen_lift_succ_simpl *)
    unfold low_subst_generic;destruct x;intros;simpl.
    rewrite DeBruijnNat.low_subst_gen_lift_succ_simpl;intros;auto.
    (* low_subst_gen_swapR_comm *)
    unfold low_subst_generic;destruct x;intros;simpl.
    apply DeBruijnNat.low_subst_gen_swapR_comm;intros.
    rewrite deBruijn.j_swapR;reflexivity.
    apply deBruijn.swapR_lift_idem.
    (* swapR_low_subst_gen_idem *)
    unfold low_subst_generic;destruct x;intros;simpl.
    apply DeBruijnNat.swapR_low_subst_gen_idem;intros.
    rewrite deBruijn.j_swapR;reflexivity.
    apply deBruijn.swapR_lift_comm.
    (* low_subst_gen_swapR_simpl *)
    unfold low_subst_generic;destruct x;intros;simpl.
    apply DeBruijnNat.low_subst_gen_swapR_simpl;intros.
    rewrite deBruijn.j_swapR;reflexivity.
    apply deBruijn.swapR_lift_comm.
    apply deBruijn.swapR_lift_1.
    apply deBruijn.lift_plus.
    (* swapR_low_subst_gen_comm *)
    unfold low_subst_generic;destruct x;intros;simpl.
    apply DeBruijnNat.swapR_low_subst_gen_comm;intros.
    rewrite deBruijn.j_swapR;reflexivity.
    apply deBruijn.swapR_lift_comm.
    (* swapR_low_subst_gen_comm_idem *)
    unfold low_subst_generic;destruct x;intros;simpl.
    apply DeBruijnNat.swapR_low_subst_gen_comm_idem;intros.
    rewrite deBruijn.j_swapR;reflexivity.
    apply deBruijn.swapR_lift_comm.
    (* low_subst_gen_swapR_rotate *)
    unfold low_subst_generic;destruct x;intros;simpl.
    generalize DeBruijnNat.low_subst_swapR_rotate;simpl;intros;auto.
  Defined.

  Definition low_subst (d k:nat)(ns:list t)(n:t) : t := low_subst_generic t lift (fun x => x) d k ns n.

  Definition mem_fn k d n :=
    match n with Ref n => DeBruijnNat.mem_fn k d n end.

  Definition Nmin (d:nat)(k:nat)(n:name)(accu:bool*nat) :=
    let (b,kmin) := accu in
      if b
	then (b,kmin)
	else
	  match n with
	    | Ref n =>
	      match le_lt_dec d n with
		| left _ =>
		  match eq_nat_dec (d+k) n with
		    | left _ => (true,kmin)
		    | right _ =>
		      match le_lt_dec kmin (n-d) with
			| left _ => (b,kmin)
			| right _ => (b,n-d)
		      end
		  end
		| right _ => (b,kmin)
	      end
	  end.

  Definition deBruijn_fn_base : deBruijn.deBruijn_fn t deBruijn_base.
    split with mem_fn Nmin;destruct x;simpl;intros.
    (* lift_low1_comm *)
    rewrite DeBruijnNat.lift_low1_comm;auto.
    (* lift_1_low1_simpl *)
    rewrite DeBruijnNat.lift_1_low1_simpl;auto.
    (* lift_fn *)
    rewrite DeBruijnNat.lift_fn;auto.
    (* lift_In_fn *)
    apply DeBruijnNat.lift_In_fn;auto.
    (* lift_not_fn *)
    apply DeBruijnNat.lift_not_fn;auto.
    (* swapR_low1_comm *)
    rewrite DeBruijnNat.swapR_low1_comm;auto.
    (* swapR_In_fn *)
    apply DeBruijnNat.swapR_In_fn;auto.
    (* swapR_eq_cond *)
    rewrite DeBruijnNat.swapR_eq_cond;auto.
    (* swapR_fn *)
    rewrite DeBruijnNat.swapR_fn;auto.
    (* swapR_fn_same *)
    rewrite DeBruijnNat.swapR_fn_same;auto.
    (* swapR_low1_comm2 *)
    rewrite DeBruijnNat.swapR_low1_comm2;auto.
    (* low1_swapR_comm *)
    rewrite DeBruijnNat.low1_swapR_comm;auto.
    (* lift_low1_fn *)
    rewrite DeBruijnNat.lift_low1_fn;auto.
    (* low1_In_fn *)
    apply DeBruijnNat.low1_In_fn;auto.
    (* low1_In_fn_cond *)
    apply DeBruijnNat.low1_In_fn_cond;auto.
    (* fn_plus_lift *)
    rewrite DeBruijnNat.fn_plus_lift;auto.
    (* low1_comm *)
    rewrite DeBruijnNat.low1_comm;auto.
    (* low1_swapR_simpl *)
    rewrite DeBruijnNat.low1_swapR_simpl;auto.
    (* Nmin_true *)
    intros;reflexivity.  
    (* Nmin_le *)
    generalize H;clear H.
    case le_lt_dec;intro.
    case eq_nat_dec;intro.
    intros.
    injection H.
    intros.
    subst kmin';auto.
    case le_lt_dec;intros.
    injection H;intros.
    subst kmin';auto.
    injection H;intros.
    subst kmin';auto.
    omega.
    intro.
    injection H.
    intros.
    subst kmin';auto.
    (* Nmin_lift *)
    simpl;intros.
    unfold DeBruijnNat.lift.
    repeat (case le_lt_dec);repeat (case eq_nat_dec);intros;auto;try_absurd_arith.
    (* Nmin_swapR *)
    simpl;intros.
    unfold DeBruijnNat.swapR.
    repeat (case lt_eq_lt_dec;intro;try (case s;intro));repeat (case le_lt_dec);repeat (case eq_nat_dec);intros;auto;try_absurd_arith.
    (* Nmin_low1 *)
    generalize H0;clear H0.
    unfold DeBruijnNat.mem_fn.
    case eq_nat_dec;intros;try (discriminate H0);clear H0.
    rewrite plus_assoc in n0.
    unfold DeBruijnNat.low1.
    repeat (case le_lt_dec);repeat (case eq_nat_dec);intros;auto;try_absurd_arith.
    (* Nmin_swapR_gt *)
    destruct b;simpl;auto.
    unfold DeBruijnNat.swapR.   
    repeat (case lt_eq_lt_dec;intro;try (case s;intro));repeat (case le_lt_dec);repeat (case eq_nat_dec);intros;auto;try_absurd_arith.
    (* low1_injective *)
    destruct y;simpl in H1.
    injection H1;intros.
    simpl in H0.
    rewrite (DeBruijnNat.low1_injective d n n0);auto.
  Defined.

  Definition deBruijn_subst_gen_fn (to:Set)(deBruijn_to:deBruijn.deBruijn_base to)(deBruijn_fn_to:deBruijn.deBruijn_fn to deBruijn_to)(Hextension:deBruijn.deBruijn_extension t deBruijn_base to deBruijn_to)(Hextension_fn:deBruijn.deBruijn_extension_fn t deBruijn_base deBruijn_fn_base to deBruijn_to deBruijn_fn_to Hextension) : deBruijn.deBruijn_subst_gen_fn t deBruijn_base deBruijn_fn_base to deBruijn_to deBruijn_fn_to Hextension Hextension_fn (deBruijn_subst_gen to deBruijn_to Hextension).
    intros.
    split.
    (* low_subst_gen_fn_not_supp *)
    intros.
    destruct x.
    simpl.
    unfold low_subst_generic.
    unfold DeBruijnNat.low_subst_gen.
    unfold DeBruijnNat.low_subst_generic.
    unfold DeBruijnNat.low_subst_generic_aux.
    case le_lt_dec.
    intros.
    elim (List_nth_Some to es (n-(d+S d')));intros.
    destruct (nth_error es (n-(d+S d'))).
    assert (i+d < d+S d');try omega.
    generalize (deBruijn.lift_not_fn deBruijn_fn_to 0 (d+S d') (i+d) H2 t0).
    intro.
    generalize (deBruijn.fn_plus_lift deBruijn_fn_to 0 d).
    rewrite plus_0_l.
    unfold DeBruijnNat.lift.
    simpl.
    intro.
    rewrite H4 in H3.
    rewrite H3.
    unfold DeBruijnNat.mem_fn.
    case eq_nat_dec;intros;auto.
    try_absurd_arith.
    assert (length es <= n-(d+S d')).
    case (le_lt_dec (length es) (n-(d+S d')));auto.
    intro.
    elim H0;auto.
    intros.
    discriminate H2.
    rewrite <- (deBruijn.j_mem_fn Hextension_fn).
    simpl.
    unfold DeBruijnNat.mem_fn.
    case eq_nat_dec;case eq_nat_dec;intros;try_absurd_arith.
    auto.
    intros.
    rewrite <- (deBruijn.j_mem_fn Hextension_fn).
    reflexivity.
    (* low_subst_gen_low1_comm *)
    destruct x;simpl;intros.
    unfold low_subst_generic.
    unfold DeBruijnNat.low_subst_gen.
    unfold DeBruijnNat.low_subst_generic.
    unfold DeBruijnNat.low_subst_generic_aux.
    assert (n<>d).
    unfold DeBruijnNat.mem_fn in H.
    rewrite plus_0_r in H.
    generalize H;clear H.
    case eq_nat_dec;intros.
    discriminate H.
    auto.
    unfold DeBruijnNat.low1.
    case (le_lt_dec d n);intro.
    assert (d < n).
    omega.
    case (le_lt_dec (d + S d') n);intro.
    case le_lt_dec;intro;try_absurd_arith.
    destruct n.
    inversion H1.
    unfold pred.
    replace (S n - (d+S d')) with (n-(d+d'));try omega.
    elim (List_nth_Some to es (n-(d+d')));intros.
    destruct (nth_error es (n-(d+d'))).
    generalize (deBruijn.low1_lift_lift deBruijn_to 0 d d');simpl.
    intros.
    rewrite H4.
    auto.
    rewrite deBruijn.j_low1.
    assert (length es <= n-(d+d')).
    case (le_lt_dec (length es) (n-(d+d')));intros;auto.
    elim H2;auto.
    intros.
    discriminate H4.
    simpl.
    destruct (length es).
    simpl.
    replace (n-0+k) with (n+k);try omega.
    unfold DeBruijnNat.low1;simpl.
    case le_lt_dec;intro;try_absurd_arith.
    auto.
    unfold DeBruijnNat.low1;simpl.
    case le_lt_dec;intro;try_absurd_arith.
    replace (n-S n0+k) with (pred (n-n0+k));try omega.
    auto.
    case le_lt_dec;intro;try_absurd_arith.
    rewrite deBruijn.j_low1.
    simpl.
    unfold DeBruijnNat.low1;simpl.
    case le_lt_dec;intro;try_absurd_arith.
    auto.
    case (le_lt_dec (d+d') n);intro;try_absurd_arith.
    case (le_lt_dec);intro;try_absurd_arith.
    rewrite deBruijn.j_low1.
    simpl.
    unfold DeBruijnNat.low1;simpl.
    case le_lt_dec;intro;try_absurd_arith.
    auto.
    (* low_subst_gen_low1_same *)
    destruct x;simpl;intros.
    unfold low_subst_generic.
    unfold DeBruijnNat.low_subst_gen.
    unfold DeBruijnNat.low_subst_generic.
    unfold DeBruijnNat.low_subst_generic_aux.
    assert (n<>d).
    unfold DeBruijnNat.mem_fn in H.
    rewrite plus_0_r in H.
    generalize H;clear H.
    case eq_nat_dec;intros.
    discriminate H.
    auto.
    unfold DeBruijnNat.low1.
    case le_lt_dec;intro;try reflexivity.
    simpl.
    rewrite List_nth_None.
    assert (d < n).
    omega.
    destruct n.
    inversion H1.
    simpl.
    rewrite <- minus_n_O.
    rewrite plus_comm;simpl;auto.
    simpl.
    omega.
  Qed.

  Lemma fn_unique : forall (d:nat)(n m:nat)(x:t),(deBruijn.mem_fn deBruijn_fn_base n d x)=true->(deBruijn.mem_fn deBruijn_fn_base m d x)=true->n=m.
    destruct x;simpl.
    unfold DeBruijnNat.mem_fn.
    repeat (case eq_nat_dec;intro);intros;try omega;try (discriminate H || discriminate H0).
  Qed.

  Lemma fn_0_name : forall (n:nat),(deBruijn.mem_fn deBruijn_fn_base n 0 (Ref n))=true.
    intros;simpl.
    unfold DeBruijnNat.mem_fn.
    repeat (case eq_nat_dec;intro;try_absurd_arith);auto.
  Qed.

  Lemma fn_0_name_unique : forall (x n:nat),(deBruijn.mem_fn deBruijn_fn_base x 0 (Ref n))=true->x=n.
    intros.
    eapply fn_unique.
    apply H.
    apply fn_0_name.
  Qed.

  Definition deBruijn_base_list := deBruijn.deBruijn_base_list t deBruijn_base.
  Definition deBruijn_fn_base_list :=  deBruijn.deBruijn_fn_base_list t deBruijn_base deBruijn_fn_base.

  Definition compute_fn (d:nat)(x:name) : set nat :=
    match x with
      | Name.Ref(n) =>
        match le_lt_dec d n with
          | left _ => set_add eq_nat_dec (n-d) (empty_set _)
          | right _ => empty_set _
        end
    end.

  Lemma compute_fn_ok : forall (x:nat)(d:nat)(n:name),(deBruijn.mem_fn deBruijn_fn_base x d n = set_mem eq_nat_dec x (compute_fn d n)).
    destruct n.
    simpl;unfold DeBruijnNat.mem_fn;simpl.
    case le_lt_dec;intro;try_absurd_arith;simpl;auto.
    case eq_nat_dec;intro;case eq_nat_dec;intro;try_absurd_arith;auto.
    case eq_nat_dec;intro;auto.
    try_absurd_arith.
  Qed.
End Name.
