(* A formalisation of the spi calculus in Coq *)
(* Copyright (C) 2005-2007 Sebastien Briais *)
(* http://lamp.epfl.ch/~sbriais/ *)

(* This library is free software; you can redistribute it and/or modify *)
(* it under the terms of the GNU Lesser General Public License as *)
(* published by the Free Software Foundation; either version 2.1 of the *)
(* License, or (at your option) any later version. *)

(* This library is distributed in the hope that it will be useful, but *)
(* WITHOUT ANY WARRANTY; without even the implied warranty of *)
(* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU *)
(* Lesser General Public License for more details. *)

(* You should have received a copy of the GNU Lesser General Public *)
(* License along with this library; if not, write to the Free Software *)
(* Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA *)
(* 02110-1301 USA *)

Require Import FSets.
Require Import tactics.
Require Import missing.

Module OrderedList (X:OrderedType) <: OrderedType with Definition t := list X.t.
  Definition t := list X.t.

  Fixpoint eq (xs ys:t) {struct xs} : Prop :=
    match (xs,ys) with
      | (nil,nil) => True
      | (nil,_) => False
      | (_,nil) => False
      | (x::xs,y::ys) => (X.eq x y) /\ (eq xs ys)
    end.

  Fixpoint lt (xs ys:t) {struct xs} : Prop :=
    match (xs,ys) with
      | (nil,nil) => False
      | (nil,_) => True
      | (_,nil) => False
      | (x::xs,y::ys) => (X.lt x y) \/ ((X.eq x y) /\ (lt xs ys))
    end.

  Lemma eq_refl : forall (x:t),(eq x x).
    induction x;simpl;trivial.
    split;auto.
  Qed.

  Lemma eq_sym : forall (x y:t),(eq x y)->(eq y x).
    induction x;destruct y;simpl;auto.
    intros;and_elim.
    split;eauto.
  Qed.

  Lemma eq_trans : forall (x y z:t),(eq x y)->(eq y z)->(eq x z).
    induction x;destruct y;destruct z;simpl;try tauto.
    intros;and_elim.
    split;eauto.
  Qed.

  Lemma lt_not_eq : forall (x y:t),(lt x y)->~(eq x y).
    induction x;destruct y;simpl;intros;try tauto.
    case H;clear H;intro;and_elim.
    red;intro;and_elim.
    elim (X.lt_not_eq H);auto.
    red;intro;and_elim.
    elim (IHx y H0);auto.
  Qed.

  Lemma X_eq_lt_lt : forall (x y z:X.t),(X.eq x y)->(X.lt y z)->(X.lt x z).
    intros.
    case (X.compare x z);intro;auto.
    elim (X.lt_not_eq H0);eauto.
    elim (X.lt_not_eq (X.lt_trans H0 l));eauto.
  Qed.

  Lemma X_lt_eq_lt : forall (x y z:X.t),(X.lt x y)->(X.eq y z)->(X.lt x z).
    intros.
    case (X.compare x z);intro;auto.
    elim (X.lt_not_eq H);eauto.
    elim (X.lt_not_eq (X.lt_trans l H));eauto.
  Qed.

  Lemma lt_trans : forall (x y z:t),(lt x y)->(lt y z)->(lt x z).
    induction x;destruct y;destruct z;simpl;try tauto.
    intros;and_elim.
    case H;clear H;intro H;case H0;clear H0;intro H0;and_elim.
    left;eauto.
    left.
    eapply X_lt_eq_lt;eauto.
    left.
    eapply X_eq_lt_lt;eauto.
    right.
    split;eauto.
  Qed.

  Definition compare : forall (x y:t),(Compare lt eq x y).
    induction x;destruct y.
    apply EQ;simpl;trivial.
    apply LT;simpl;trivial.
    apply GT;simpl;trivial.
    case (X.compare a t0);intro.
    apply LT;simpl.
    left;trivial.
    case (IHx y);intro.
    apply LT;simpl.
    right;tauto.
    apply EQ;simpl;tauto.
    apply GT;simpl.
    right.
    split;trivial.
    eauto.
    apply GT.
    simpl.
    left;trivial.
  Defined.

  Hint Immediate eq_sym.
  Hint Resolve eq_refl eq_trans lt_not_eq lt_trans. 
End OrderedList.

Module OrderedPair (X:OrderedType) (Y:OrderedType) <: OrderedType with Definition t := (X.t * Y.t)%type.
  Definition t := (X.t * Y.t)%type.

  Definition eq (xy1:t) (xy2:t) := 
    let (x1,y1) := xy1 in
      let (x2,y2) := xy2 in
	(X.eq x1 x2) /\ (Y.eq y1 y2).

  Definition lt (xy1:t) (xy2:t) := 
    let (x1,y1) := xy1 in
      let (x2,y2) := xy2 in
	(X.lt x1 x2) \/ ((X.eq x1 x2) /\ (Y.lt y1 y2)).

  Lemma eq_refl : forall (x:t),(eq x x).
    destruct x.
    unfold eq.
    split;[apply X.eq_refl | apply Y.eq_refl].
  Qed.

  Lemma eq_sym : forall (x y:t),(eq x y)->(eq y x).
    destruct x;destruct y;unfold eq;intro.
    and_elim.
    split;[apply X.eq_sym | apply Y.eq_sym];trivial.
  Qed.

  Lemma eq_trans : forall (x y z:t),(eq x y)->(eq y z)->(eq x z).
    unfold eq;destruct x;destruct y;destruct z;intros.
    and_elim.
    split;[eapply X.eq_trans | eapply Y.eq_trans];eauto.
  Qed.

  Lemma lt_trans : forall (x y z:t),(lt x y)->(lt y z)->(lt x z).
    unfold lt;destruct x;destruct y;destruct z;intros.
    case H;clear H;intro.
    case H0;clear H0;intro.
    left.
    eapply X.lt_trans;eauto.
    and_elim.
    left.
    case (X.compare t0 t4);trivial;intros.
    generalize (X.eq_sym H0);intro.
    generalize (X.eq_trans e H2);intro.
    elim (X.lt_not_eq H H3).
    generalize (X.lt_trans l H);intro.
    generalize (X.eq_sym H0);intro.
    elim (X.lt_not_eq H2 H3).
    and_elim.
    case H0;clear H0;intro.
    left.
    case (X.compare t0 t4);trivial;intros.
    generalize (X.eq_sym H);intro.
    generalize (X.eq_trans H2 e);intro.
    elim (X.lt_not_eq H0 H3).
    generalize (X.lt_trans H0 l);intro.
    generalize (X.eq_sym H);intro.
    elim (X.lt_not_eq H2 H3).
    and_elim.
    right.
    split.
    eauto.
    eauto.
  Qed.

  Lemma lt_not_eq : forall (x y:t),(lt x y)->~(eq x y).
    unfold lt, eq;destruct x;destruct y;intro;intro.
    and_elim.
    case H.
    intro.
    apply (X.lt_not_eq H2 H0).
    intro.
    and_elim.
    apply (Y.lt_not_eq H3 H1).
  Qed.

  Definition compare : forall (x y:t),(Compare lt eq x y).
    destruct x;destruct y.
    case (X.compare t0 t2);intro.
    apply LT.
    left;trivial.
    case (Y.compare t1 t3);intro.
    apply LT.
    right.
    tauto.
    apply EQ.
    split;trivial.
    apply GT.
    right;auto.
    apply GT.
    left;trivial.
  Defined.

  Hint Immediate eq_sym.
  Hint Resolve eq_refl eq_trans lt_not_eq lt_trans. 
End OrderedPair.

Module MissingSetProperties (setModule:S).
  Module setEqProps := FSetEqProperties.EqProperties setModule.
  Module setProps := FSetProperties.Properties setModule.
  Module setEltProps := OrderedTypeFacts setModule.E.

  Lemma In_union_empty_l : forall (s:setModule.t)(x:setModule.elt),(setModule.In x (setModule.union s setModule.empty))->(setModule.In x s).
    intros.
    case (setModule.union_1 H);intro;auto.
    elim (setModule.empty_1 H0).
  Qed.

  Lemma In_union_empty_r : forall (s:setModule.t)(x:setModule.elt),(setModule.In x (setModule.union setModule.empty s))->(setModule.In x s).
    intros.
    case (setModule.union_1 H);intro;auto.
    elim (setModule.empty_1 H0).
  Qed.

  Lemma In_union_distrib_l : forall (s t u:setModule.t)(x:setModule.elt),(setModule.In x (setModule.union s u))->(setModule.In x (setModule.union (setModule.union s t) u)).
    intros.
    case (setModule.union_1 H);intro.
    apply setModule.union_2.
    apply setModule.union_2;trivial.
    apply setModule.union_3;trivial.
  Qed.

  Lemma In_union_distrib_r : forall (s t u:setModule.t)(x:setModule.elt),(setModule.In x (setModule.union t u))->(setModule.In x (setModule.union (setModule.union s t) u)).
    intros.
    case (setModule.union_1 H);intro.
    apply setModule.union_2.
    apply setModule.union_3;trivial.
    apply setModule.union_3;trivial.
  Qed.

  Lemma singleton_equal_add : forall (x:setModule.elt),(setModule.Equal (setModule.singleton x) (setModule.add x setModule.empty)).
    intro.
    red;intros.
    split;intro.
    generalize (setModule.singleton_1 H).
    intro.
    eapply setModule.In_1.
    apply H0.
    apply setModule.add_1.
    auto.
    generalize (setProps.Add_add (setModule.empty) x).
    unfold setProps.Add.
    intros.
    elim (H0 a).
    intros.
    case (H1 H);intro.
    apply setModule.singleton_2.
    auto.
    elim (setModule.empty_1 H3).
  Qed.

  Lemma empty_mem_false : forall (s:setModule.t),(setModule.Empty s)->forall (x:setModule.elt),(setModule.mem x s)=false.
    intros.
    unfold setModule.Empty in H.
    rewrite(setEqProps.mem_3 s x);trivial.
  Qed.

  Lemma mem_union : forall (s t:setModule.t)(x:setModule.elt),(setModule.mem x (setModule.union s t))=(orb (setModule.mem x s) (setModule.mem x t)).
    intros.
    case (Bool_elim_bool (setModule.mem x (setModule.union s t)));intro.
    generalize (setModule.mem_2 H);intro.
    case (setModule.union_1 H0);intro;rewrite H;rewrite (setModule.mem_1 H1);auto with bool.
    rewrite H.
    case (Bool_elim_bool (setModule.mem x s));intro.
    generalize (setModule.mem_2 H0);intro.
    rewrite setModule.mem_1 in H.
    discriminate H.
    apply setModule.union_2.
    trivial.
    case (Bool_elim_bool (setModule.mem x t));intro;rewrite H0;rewrite H1;try reflexivity.
    generalize (setModule.mem_2 H1);intro.
    rewrite setModule.mem_1 in H.
    discriminate H.
    apply setModule.union_3.
    trivial.
  Qed.

  Lemma add_mem_simpl_true : forall (s s':setModule.t)(x:setModule.elt),(setProps.Add x s s')->forall (y:setModule.elt),(setModule.mem y s')=true<->(setModule.E.eq y x)\/(setModule.mem y s)=true.
    unfold setProps.Add;intros.
    elim (H y);intros.
    split;intro.
    case (H0 (setModule.mem_2 H2));try tauto;auto.
    case H2;intro;auto.
  Qed.

  Lemma add_mem_simpl_false : forall (s s':setModule.t)(x:setModule.elt),(setProps.Add x s s')->forall (y:setModule.elt),(setModule.mem y s')=false<->~(setModule.E.eq y x)/\(setModule.mem y s)=false.
    intros.
    elim (add_mem_simpl_true s s' x H y);intros.
    unfold setProps.Add in H.
    elim (H y);intros.
    split;intro.
    generalize (setEqProps.mem_4 s' y H4);intro.
    split.
    intro.
    rewrite H1 in H4.
    discriminate H4.
    tauto.
    generalize H1.
    case (setModule.mem y s).
    intros.
    rewrite H6 in H4;trivial.
    right;trivial.
    trivial.
    and_elim.
    generalize H0.
    case (setModule.mem y s');intros.
    case H6;trivial.
    intro.
    elim H4;trivial.
    intro.
    rewrite H5 in H7.
    discriminate H7.
    trivial.
  Qed.

  Definition equal (x y:setModule.elt) : bool.
    intros.
    case (setModule.E.compare x y);intros.
    exact false.
    exact true.
    exact false.
  Defined.

  Lemma equal_is_eq : forall (x y:setModule.elt),(setModule.E.eq x y)<->(equal x y)=true.
    intros.
    unfold equal.
    case (setModule.E.compare x y);intro;split;intro H;try (discriminate H);trivial.
    elim (setModule.E.lt_not_eq l);trivial.
    elim (setModule.E.lt_not_eq l);auto.
  Qed.

  Lemma equal_is_eq_false : forall (x y:setModule.elt),~(setModule.E.eq x y)<->(equal x y)=false.
    intros.
    unfold equal.
    case (setModule.E.compare x y);intro;split;intro H;trivial.
    apply setModule.E.lt_not_eq;trivial.
    elim H;trivial.
    discriminate H.
    intro.
    apply (setModule.E.lt_not_eq l);auto.
  Qed.

  Lemma add_mem_simpl : forall (s s':setModule.t)(x:setModule.elt),(setProps.Add x s s')->forall (y:setModule.elt),(setModule.mem y s')=(orb (equal x y) (setModule.mem y s)).
    intros.
    case (Bool_elim_bool (setModule.mem y s'));intro.
    elim (add_mem_simpl_true s s' x H y);intros.
    case (H1 H0);intro.
    elim (equal_is_eq x y);intros.
    rewrite H0.
    rewrite H4;auto.
    rewrite H0;rewrite H3;rewrite orb_b_true;trivial.
    elim (add_mem_simpl_false s s' x H y);intros.
    elim (H1 H0);intros.
    and_elim.
    rewrite H4.
    rewrite orb_b_false.
    elim (equal_is_eq_false x y);intros.
    rewrite H5;auto.
  Qed.

  Lemma add_In_simpl : forall (s:setModule.t)(x:setModule.elt),forall (y:setModule.elt),(setModule.In y (setModule.add x s))<->((setModule.E.eq x y)\/(setModule.In y s)).
    intros.
    elim (setProps.Add_add s x y);unfold setProps.Add;intros.
    split;intro.
    case (H H1);try tauto.
    case H1;try tauto.
  Qed.

  Definition setSetoid : Setoid.Setoid_Theory setModule.t setModule.Equal.
    split;intros.
    apply setProps.equal_refl.
    apply setProps.equal_sym;trivial.
    eapply setProps.equal_trans;eauto.
  Qed.

  Lemma compat_op_add : (compat_op setModule.E.eq setModule.Equal setModule.add).
    unfold compat_op;intros.
    unfold setModule.Equal;intro.
    elim (add_In_simpl y x a);intros.
    elim (add_In_simpl y' x' a);intros.
    elim (H0 a);clear H0;intros.
    split;intro.
    apply H4.
    case (H1 H6);intro.
    left;eauto.
    right;auto.
    apply H2.
    case (H3 H6);intro.
    left;eauto.
    right;auto.
  Qed.

  Lemma compat_op_union : (compat_op setModule.Equal setModule.Equal setModule.union).
    unfold compat_op;intros.
    unfold setModule.Equal;intro.
    split;intro.
    case (setModule.union_1 H1);intro.
    apply setModule.union_2.
    elim (H a);auto.
    apply setModule.union_3.
    elim (H0 a);auto.
    case (setModule.union_1 H1);intro.
    apply setModule.union_2.
    elim (H a);auto.
    apply setModule.union_3.
    elim (H0 a);auto.
  Qed.

  Lemma transpose_add : (transpose setModule.Equal setModule.add).
    unfold transpose;intros.
    unfold setModule.Equal;intro.
    elim (add_In_simpl (setModule.add y z) x a);intros.
    elim (add_In_simpl (setModule.add x z) y a);intros.
    elim (add_In_simpl z y a);intros.
    elim (add_In_simpl z x a);intros.
    split;intro.
    apply H2.
    case (H H7);intro.
    right.
    apply H6;auto.
    case (H3 H8);intro.
    left;trivial.
    right;auto.
    apply H0.
    case(H1 H7);intro.
    right.
    apply H4;tauto.
    case (H5 H8);intro.
    left;trivial.
    right;auto.
  Qed.

  Lemma transpose_union : transpose setModule.Equal setModule.union.
    unfold transpose.
    intros.
    unfold setModule.Equal.
    split;intros.
    case (setModule.union_1 H);intro.
    apply setModule.union_3;apply setModule.union_2;auto.
    case (setModule.union_1 H0);intro.
    apply setModule.union_2;auto.
    apply setModule.union_3;apply setModule.union_3;auto.
    case (setModule.union_1 H);intro.
    apply setModule.union_3;apply setModule.union_2;auto.
    case (setModule.union_1 H0);intro.
    apply setModule.union_2;auto.
    apply setModule.union_3;apply setModule.union_3;auto.
  Qed.

  Lemma compat_op_remove : (compat_op setModule.E.eq setModule.Equal setModule.remove).
    unfold compat_op.
    intros x y s t;intros.
    case (setProps.In_dec x s);intro.
    assert (setModule.In y t).
    eapply setModule.In_1.
    apply H.
    elim (H0 x);auto.
    assert (setProps.Add x (setModule.remove x s) s);auto.
    assert (setProps.Add y (setModule.remove y t) t);auto.
    split;intros.
    assert (~setModule.E.eq a x).
    intro.
    elim (setModule.remove_1 (setModule.E.eq_sym H5) H4).
    generalize (setModule.remove_3 H4);intro.
    apply setModule.remove_2.
    intro.
    apply H5.
    eauto.
    elim (H0 a);auto.
    assert (~setModule.E.eq a y).
    intro.
    elim (setModule.remove_1 (setModule.E.eq_sym H5) H4).
    generalize (setModule.remove_3 H4);intro.
    apply setModule.remove_2.
    intro.
    apply H5.
    eauto.
    elim (H0 a);auto.
    eapply setProps.equal_trans.
    apply setProps.remove_equal;trivial.
    eapply setProps.equal_trans.
    apply H0.
    apply setProps.equal_sym.
    apply setProps.remove_equal.
    intro.
    apply n.
    eapply setModule.In_1.
    apply setModule.E.eq_sym.
    apply H.
    elim (H0 y);auto.
  Qed.

  Lemma transpose_remove : (transpose setModule.Equal setModule.remove).
    unfold transpose.
    intros x y s.
    case (setEltProps.eq_dec x y);intro.
    generalize (setModule.remove_1 (setModule.E.eq_sym e) (s:=s));intro.
    eapply setProps.equal_trans.
    apply setProps.remove_equal;trivial.
    apply setProps.equal_sym.
    eapply setProps.equal_trans.
    apply setProps.remove_equal.
    apply setModule.remove_1.
    auto.
    apply compat_op_remove;auto.
    apply setProps.equal_refl.
    split;intros;generalize (setModule.remove_3 H);intro;generalize (setModule.remove_3 H0);intro;apply setModule.remove_2.
    intro.
    apply (setModule.remove_1 H2 H0).
    apply setModule.remove_2;trivial.
    intro.
    apply (setModule.remove_1 H2 H).
    intro.
    apply (setModule.remove_1 H2 H0).
    apply setModule.remove_2;trivial.
    intro.
    apply (setModule.remove_1 H2 H).    
  Qed.

  Lemma empty_is_empty : forall (s:setModule.t),(setModule.Empty s)->(setModule.Equal s (setModule.empty)).
    intros.
    split;intro.
    elim (H a);trivial.
    generalize (setModule.empty_1);intro.
    elim (H1 a);trivial.
  Qed.

  Lemma Equal_empty_empty : forall (s:setModule.t),(setModule.Equal s (setModule.empty))->(setModule.Empty s).
    intros.
    intro;intro.
    generalize (setModule.empty_1);intro.
    elim (H1 a).
    elim (H a);auto.
  Qed.

  Lemma add_is_add : forall (s s':setModule.t)(x:setModule.elt),(setProps.Add x s s')->(setModule.Equal s' (setModule.add x s)).
    intros s s' x.
    generalize (setProps.Add_add s x).
    unfold setProps.Add;intros.
    split;intro.
    elim (H a);intros.
    apply H3.
    elim (H0 a);auto.
    elim (H0 a);intros.
    apply H3.
    elim (H a);auto.
  Qed.

  Lemma subset_add_elim : forall (s s' t:setModule.t)(x:setModule.elt),(setProps.Add x s s')->~(setModule.In x s)->(setModule.Subset t s')->(setModule.In x t)\/(setModule.Subset t s).
    intros.
    case (setProps.In_dec x t).
    left;trivial.
    intro.
    right.
    intro.
    intro.
    elim (H a);intros.
    case H3.
    apply H1;trivial.
    intro.
    elim n.
    eapply setModule.In_1.
    apply setModule.E.eq_sym.
    apply H5.
    trivial.
    trivial.
  Qed.

  Lemma subset_add_remove_subset : forall (s s' t:setModule.t)(x:setModule.elt),(setProps.Add x s s')->~(setModule.In x s)->(setModule.Subset t s')->(setModule.Subset (setModule.remove x t) s).
    intros.
    intro;intros.
    assert (setModule.In a t).
    eapply setModule.remove_3.
    apply H2.
    assert (setModule.In a s').
    apply H1;auto.
    elim (H a);intros.
    case (H5 H4);trivial.
    intro.
    elim (setModule.remove_1 H7 (s:=t));trivial.
  Qed.

  Lemma filter_Add_true : forall (s s':setModule.t)(x:setModule.elt),(setProps.Add x s s')->~(setModule.In x s)->forall (p:setModule.elt->bool),(compat_bool setModule.E.eq p)->(p x)=true->(setProps.Add x (setModule.filter p s) (setModule.filter p s'))/\~(setModule.In x (setModule.filter p s)).
    split;[split | red];intros.
    generalize (setModule.filter_1 H1 H3);intro.
    elim (H y);intros.
    case (H5 H4);intro;try tauto.
    right.
    apply setModule.filter_3;auto.
    eapply setModule.filter_2;eauto.
    case H3;intro.
    apply setModule.filter_3;trivial.
    elim (H y);auto.
    rewrite <- H2.
    apply H1;auto.
    apply setModule.filter_3;trivial.
    elim (H y);intros.
    apply H6.
    right.
    eapply setModule.filter_1;eauto.
    eapply setModule.filter_2;eauto.
    apply H0.
    eapply setModule.filter_1;eauto.
  Qed.

  Lemma filter_Add_false : forall (s s':setModule.t)(x:setModule.elt),(setProps.Add x s s')->~(setModule.In x s)->forall (p:setModule.elt->bool),(compat_bool setModule.E.eq p)->(p x)=false->(setModule.Equal (setModule.filter p s) (setModule.filter p s')).
    split;intros.
    apply setModule.filter_3;trivial.
    elim (H a);intros.
    apply H5.
    right.
    eapply setModule.filter_1;eauto.
    eapply setModule.filter_2;eauto.
    apply setModule.filter_3;auto.
    elim (H a);intros.
    assert (setModule.In a s').
    eapply setModule.filter_1;eauto.
    case (H4 H6);intro;trivial.
    assert (p a=true).
    eapply setModule.filter_2;eauto.
    generalize (H1 a x (setModule.E.eq_sym H7)).
    rewrite H2;rewrite H8;intro.
    discriminate H9.
    eapply setModule.filter_2;eauto.
  Qed.

  Require Import Wf_nat.

  Lemma induction_type_cardinal : forall (P:setModule.t->Type),(forall (s:setModule.t),((forall (s':setModule.t),(setModule.cardinal s')<(setModule.cardinal s)->(P s'))->(P s)))->(forall (s:setModule.t),(P s)).
    intros.
    eapply well_founded_induction_type.
    apply (well_founded_ltof setModule.t  (fun s => setModule.cardinal s)).
    unfold ltof.
    trivial.
  Defined.

  Lemma induction_cardinal : forall (P:setModule.t->Prop),(forall (s:setModule.t),((forall (s':setModule.t),(setModule.cardinal s')<(setModule.cardinal s)->(P s'))->(P s)))->(forall (s:setModule.t),(P s)).
    intros;apply induction_type_cardinal;auto.
  Qed.

  Lemma recursion_cardinal : forall (P:setModule.t->Set),(forall (s:setModule.t),((forall (s':setModule.t),(setModule.cardinal s')<(setModule.cardinal s)->(P s'))->(P s)))->(forall (s:setModule.t),(P s)).
    intros;apply induction_type_cardinal;auto.
  Defined.

  Lemma empty_dec : forall (s:setModule.t),{setModule.Equal s setModule.empty}+{~(setModule.Equal s setModule.empty)}.
    intro s.
    case (Bool_eq_dec (setModule.equal s setModule.empty));intro;[left | right].
    auto.
    intro.
    assert (setModule.equal s setModule.empty =true).
    auto.
    rewrite H0 in e;discriminate e.
  Defined.

  Lemma remove_In_neq : forall (s:setModule.t)(x y:setModule.elt),(~setModule.E.eq x y)->(setModule.mem x s)=(setModule.mem x (setModule.remove y s)).
    intros.
    case (Bool_elim_bool (setModule.mem x s));intro.
    assert (setModule.In x s);auto.
    assert (setModule.In x (setModule.remove y s));auto.
    rewrite H0;symmetry.
    auto.
    assert (~setModule.In x s).
    intro.
    assert (setModule.mem x s=true);auto.
    rewrite H0 in H2;discriminate H2.
    case (Bool_elim_bool (setModule.mem x (setModule.remove y s)));intro.
    assert (setModule.In x (setModule.remove y s));auto.
    elim H1.
    eapply setModule.remove_3;eauto.
    congruence.
  Qed.

  Lemma filter_Empty : forall (s:setModule.t)(p:setModule.elt->bool),(compat_bool setModule.E.eq p)->(setModule.Empty s)->(setModule.Empty (setModule.filter p s)).
    intros.
    intro;intro.
    apply (H0 a).
    apply (setModule.filter_1 H H1).
  Qed.

  Lemma filter_equiv : forall (s:setModule.t)(p q:setModule.elt->bool),(compat_bool setModule.E.eq p)->(compat_bool setModule.E.eq q)->(forall (x:setModule.elt),(setModule.In x s)->(p x = q x))->(setModule.Equal (setModule.filter p s) (setModule.filter q s)).
    intro s.
    apply setProps.set_induction with (P:=(fun s => forall (p q:setModule.elt->bool),(compat_bool setModule.E.eq p)->(compat_bool setModule.E.eq q)->(forall (x:setModule.elt),(setModule.In x s)->(p x = q x))->(setModule.Equal (setModule.filter p s) (setModule.filter q s))));clear s;intros.
    generalize (empty_is_empty s H);intro.
    assert (setModule.Empty (setModule.filter p s)).
    intro;intro.
    apply (H a).
    apply (setModule.filter_1 H0 H4).
    assert (setModule.Empty (setModule.filter q s)).
    intro;intro.
    apply (H a).
    apply (setModule.filter_1 H1 H5).
    generalize (empty_is_empty (setModule.filter p s) H4);generalize (empty_is_empty (setModule.filter q s) H5);intros.
    eapply setProps.equal_trans.
    apply H7.
    apply setProps.equal_sym.
    apply H6.
    assert (setModule.In x s').
    elim (H1 x);auto.
    generalize (H4 x H5);intro.
    assert (setModule.Equal (setModule.filter p s) (setModule.filter q s)).
    apply H;trivial.
    intros.
    apply H4.
    elim (H1 x0);auto.
    case (Bool_elim_bool (p x));intro.
    elim (filter_Add_true s s' x H1 H0 p H2 H8);intros.
    rewrite H6 in H8.
    elim (filter_Add_true s s' x H1 H0 q H3 H8);intros.
    split;intros.
    elim (H11 a);intros.
    apply H15.
    elim (H9 a);intros.
    case (H16 H13);intro;try tauto.
    right;elim (H7 a);auto.
    elim (H9 a);intros.
    apply H15.
    elim (H11 a);intros.
    case (H16 H13);intro;try tauto.
    right;elim (H7 a);auto.
    generalize (filter_Add_false s s' x H1 H0 p H2 H8);intros.
    rewrite H6 in H8.
    generalize (filter_Add_false s s' x H1 H0 q H3 H8);intros.
    eapply setProps.equal_trans.
    apply setProps.equal_sym.
    apply H9.
    eapply setProps.equal_trans.
    apply H7.
    trivial.
  Qed.

  Lemma compat_bool_and : forall (p q:setModule.elt->bool),(compat_bool setModule.E.eq p)->(compat_bool setModule.E.eq q)->(compat_bool setModule.E.eq (fun x => andb (p x) (q x))).
    intros.
    unfold compat_bool.
    intros.
    rewrite (H x y);trivial.
    rewrite (H0 x y);trivial.
  Qed.

  Lemma eq_In : forall (s:setModule.t)(x y:setModule.elt),(setModule.E.eq x y)->(setModule.In x s)->(setModule.In y s).
    intros.
    eapply setModule.In_1;eauto.
  Qed.

  Lemma filter_Equal : forall (s s':setModule.t)(p:setModule.elt->bool),(setModule.Equal s s')->(compat_bool setModule.E.eq p)->(setModule.Equal (setModule.filter p s) (setModule.filter p s')).
    intro s.
    apply setProps.set_induction with (P:=(fun s => forall (s':setModule.t)(p:setModule.elt->bool),(setModule.Equal s s')->(compat_bool setModule.E.eq p)->(setModule.Equal (setModule.filter p s) (setModule.filter p s'))));clear s;intros.
    eapply setProps.equal_trans.
    apply empty_is_empty.
    apply filter_Empty;trivial.
    apply setProps.equal_sym.
    apply empty_is_empty.
    apply filter_Empty;trivial.
    intro;intro.
    apply (H a).
    elim (H0 a);auto.
    assert (setModule.Equal s (setModule.remove x s'0)).
    split;intros.
    elim (H1 a).
    intros.
    assert (setModule.In a s').
    apply H6.
    right;trivial.
    apply setModule.remove_2.
    intro.
    apply H0.
    eapply eq_In.
    apply setModule.E.eq_sym.
    apply H8.
    trivial.
    elim (H2 a);auto.
    generalize (setModule.remove_3 H4);intro.
    elim (H1 a);intros.
    case H6;trivial.
    elim (H2 a);auto.
    intro.
    generalize (eq_In (setModule.remove x s'0) a x (setModule.E.eq_sym H8) H4);intro.
    elim (setModule.remove_1 H8 H4).
    assert (~setModule.In x (setModule.remove x s'0)).
    apply setModule.remove_1;auto.
    assert (setProps.Add x (setModule.remove x s'0) s'0).
    split;intros.
    elim (H2 y);intros.
    generalize (H8 H6);intro.
    elim (H1 y);intros.
    case (H10 H9);intro;[left;tauto | right].
    elim (H4 y);auto.
    elim (H1 y);intros.
    assert (setModule.In y s').
    apply H8.
    case H6;[left;tauto | right].
    elim (H4 y);auto.
    elim (H2 y);auto.
    generalize (H (setModule.remove x s'0) p H4 H3);intro.
    case (Bool_elim_bool (p x));intro.
    elim (filter_Add_true s s' x H1 H0 p H3 H8);intros.
    elim (filter_Add_true (setModule.remove x s'0) s'0 x H6 H5 p H3 H8);intros.
    split;intros.
    elim (H9 a);intros.
    case (H14 H13);intro.
    elim (H11 a);auto.
    elim (H11 a);intros.
    apply H18.
    right.
    elim (H7 a);auto.
    elim (H9 a);intros.
    apply H15.
    elim (H11 a);intros.
    case (H16 H13);[left;tauto | right].
    elim (H7 a);auto.
    generalize (filter_Add_false s s' x H1 H0 p H3 H8);intro.
    generalize (filter_Add_false (setModule.remove x s'0) s'0 x H6 H5 p H3 H8);intro.
    eapply setProps.equal_trans.
    apply setProps.equal_sym.
    apply H9.
    eapply setProps.equal_trans.
    apply H7.
    auto.
  Qed.

  Lemma filter_and : forall (s:setModule.t)(p q:setModule.elt->bool),(compat_bool setModule.E.eq p)->(compat_bool setModule.E.eq q)->(setModule.Equal (setModule.filter p (setModule.filter q s)) (setModule.filter (fun x => andb (p x) (q x)) s)).
    intro.
    apply setProps.set_induction with (P:=(fun s => forall (p q:setModule.elt->bool),(compat_bool setModule.E.eq p)->(compat_bool setModule.E.eq q)->(setModule.Equal (setModule.filter p (setModule.filter q s)) (setModule.filter (fun x => andb (p x) (q x)) s))));clear s;intros.
    eapply setProps.equal_trans.
    apply empty_is_empty.
    apply filter_Empty;trivial.
    apply filter_Empty;trivial.
    apply setProps.equal_sym.
    apply empty_is_empty.
    apply filter_Empty;trivial.
    apply compat_bool_and;auto.
    generalize (compat_bool_and p q H2 H3);intro.
    generalize (H p q H2 H3);intro.
    case (Bool_elim_bool (q x));intro.
    elim (filter_Add_true s s' x H1 H0 q H3 H6);intros.
    case (Bool_elim_bool (p x));intro.
    elim (filter_Add_true (setModule.filter q s) (setModule.filter q s') x H7 H8 p H2 H9);intros.
    assert (andb (p x) (q x)=true).
    rewrite H9;rewrite H6;reflexivity.
    elim (filter_Add_true s s' x H1 H0 (fun x => p x && q x) H4 H12);intros.
    split;intros.
    elim (H13 a);intros.
    apply H17.
    elim (H10 a);intros.
    case (H18 H15);intro;[left;tauto | right].
    elim (H5 a);auto.
    elim (H10 a);intros.
    apply H17.
    elim (H13 a);intros.
    case (H18 H15);intro;[left;tauto | right].
    elim (H5 a);auto.
    assert (p x && q x = false).
    rewrite H6;rewrite H9;reflexivity.
    generalize (filter_Add_false (setModule.filter q s) (setModule.filter q s') x H7 H8 p H2 H9);intro.
    eapply setProps.equal_trans.
    apply setProps.equal_sym.
    apply H11.
    generalize (filter_Add_false s s' x H1 H0 (fun x => p x && q x) H4 H10);intro.
    eapply setProps.equal_trans.
    apply H5.
    trivial.
    generalize (filter_Add_false s s' x H1 H0 q H3 H6);intro.
    assert (p x && q x = false).
    rewrite H6;destruct (p x);reflexivity.
    generalize (filter_Add_false s s' x H1 H0 (fun x => p x && q x) H4 H8);intro.
    apply setProps.equal_sym.
    eapply setProps.equal_trans.
    apply setProps.equal_sym.
    apply H9.
    eapply setProps.equal_trans.
    apply setProps.equal_sym.
    apply H5.
    apply filter_Equal;trivial.
  Qed.

  Definition select_such_that (R:setModule.elt->setModule.elt->Prop) (Rtotaldec:forall (x y:setModule.elt),{R x y}+{R y x}) (Rrefl:forall (x y:setModule.elt),(setModule.E.eq x y)->(R x y)) (Rtrans:forall (x y z:setModule.elt),(R x y)->(R y z)->(R x z)) (s:setModule.t) : {x:setModule.elt | (setModule.In x s) /\ (forall (y:setModule.elt),(setModule.In y s)->(R x y))}+{setModule.Empty s}.
    intros.
    case (empty_dec s);intro.
    right.
    intro;intro.
    assert (setModule.Empty setModule.empty);auto.
    apply (H0 a).
    elim (e a);auto.
    left.
    generalize n;clear n.
    apply (recursion_cardinal (fun s => ~(setModule.Equal s setModule.empty) -> {x:setModule.elt | setModule.In x s /\ (forall y : setModule.elt, setModule.In y s -> R x y)}));clear s;intros.
    case (Option_elim_option setModule.elt (setModule.choose s)).
    intro.
    ex_elim s0 a.
    assert (setModule.In a s);auto.
    assert (setModule.Equal (setModule.add a (setModule.remove a s)) s).
    apply setProps.add_remove;trivial.
    assert (setProps.Add a (setModule.remove a s) s);auto.
    case (empty_dec (setModule.remove a s)).
    intro.
    exists a.
    split;trivial.
    intros.
    elim (H3 y);intros.
    case (H5 H4);intro.
    apply Rrefl;auto.
    elim (e y);intros.
    generalize (H8 H7);intro.
    assert (setModule.Empty setModule.empty);auto.
    elim (H11 y);auto.
    intro.
    assert (~(setModule.In a (setModule.remove a s)));auto.
    assert (setModule.cardinal (setModule.remove a s) < setModule.cardinal s).
    rewrite (setProps.cardinal_2 H4 H3).
    auto with arith.
    elim (H (setModule.remove a s) H5).
    intro b;intros.
    and_elim.
    case (Rtotaldec a b);intro.
    exists a.
    split;trivial.
    intros.
    elim (H3 y);intros.
    case (H8 H7);intro.
    apply Rrefl;eauto.
    eapply Rtrans.
    apply r.
    apply H6;trivial.
    exists b.
    split.
    eapply setModule.remove_3.
    apply p.
    intros.
    elim (H3 y);intros.
    case (H8 H7);intro.
    eapply Rtrans.
    apply r.
    apply Rrefl;eauto.
    apply H6;trivial.
    trivial.
    intros.
    elim H0.
    apply empty_is_empty.
    auto.
  Qed.

  Theorem induction_type_such_that : forall (R:setModule.elt->setModule.elt->Prop) (Rtotaldec:forall (x y:setModule.elt),{R x y}+{R y x}) (Rrefl:forall (x y:setModule.elt),(setModule.E.eq x y)->(R x y)) (Rtrans:forall (x y z:setModule.elt),(R x y)->(R y z)->(R x z)), forall (P:setModule.t->Type),(forall (s:setModule.t),(setModule.Empty s)->(P s))->(forall (s s':setModule.t),(P s)->(forall (x:setModule.elt),~(setModule.In x s)->(setProps.Add x s s')->(setModule.For_all (fun y => (R x y)) s)->(P s')))->(forall (s:setModule.t),(P s)).
    intros R Rtotaldec Rrefl Rtrans P H H0 s.
    apply induction_type_cardinal;clear s;intros s H1.
    case (empty_dec s).
    intros.
    apply H.
    intro;intro.
    assert (setModule.Empty setModule.empty);auto.
    apply (H3 a).
    elim (e a);auto.
    intros.
    assert (~setModule.Empty s).
    intro.
    apply n.
    apply empty_is_empty;trivial.
    elim (select_such_that R Rtotaldec Rrefl Rtrans s);intro.
    ex_elim a x.
    assert (setProps.Add x (setModule.remove x s) s);auto.
    assert (~(setModule.In x (setModule.remove x s)));auto.
    assert (setModule.cardinal (setModule.remove x s) < setModule.cardinal s).
    rewrite (setProps.cardinal_2 H5 H4).
    auto with arith.
    eapply H0.
    apply H1.
    apply H6.
    apply H5.
    trivial.
    unfold setModule.For_all.
    intros.
    apply H3.
    eapply setModule.remove_3.
    apply H7.
    elim H2;trivial.
  Defined.

  Theorem induction_such_that : forall (R:setModule.elt->setModule.elt->Prop) (Rtotaldec:forall (x y:setModule.elt),{R x y}+{R y x}) (Rrefl:forall (x y:setModule.elt),(setModule.E.eq x y)->(R x y)) (Rtrans:forall (x y z:setModule.elt),(R x y)->(R y z)->(R x z)), forall (P:setModule.t->Prop),(forall (s:setModule.t),(setModule.Empty s)->(P s))->(forall (s s':setModule.t),(P s)->(forall (x:setModule.elt),~(setModule.In x s)->(setProps.Add x s s')->(setModule.For_all (fun y => (R x y)) s)->(P s')))->(forall (s:setModule.t),(P s)).
    intros.
    apply (induction_type_such_that R Rtotaldec Rrefl Rtrans);auto.
  Qed.

  Theorem recursion_such_that : forall (R:setModule.elt->setModule.elt->Prop) (Rtotaldec:forall (x y:setModule.elt),{R x y}+{R y x}) (Rrefl:forall (x y:setModule.elt),(setModule.E.eq x y)->(R x y)) (Rtrans:forall (x y z:setModule.elt),(R x y)->(R y z)->(R x z)), forall (P:setModule.t->Set),(forall (s:setModule.t),(setModule.Empty s)->(P s))->(forall (s s':setModule.t),(P s)->(forall (x:setModule.elt),~(setModule.In x s)->(setProps.Add x s s')->(setModule.For_all (fun y => (R x y)) s)->(P s')))->(forall (s:setModule.t),(P s)).
    intros.
    apply (induction_type_such_that R Rtotaldec Rrefl Rtrans);auto.
  Defined.

  Lemma filter_union : forall (s s':setModule.t)(p:setModule.elt->bool),(compat_bool setModule.E.eq p)->(setModule.Equal (setModule.filter p (setModule.union s s')) (setModule.union (setModule.filter p s) (setModule.filter p s'))).
    split;intros.
    generalize (setModule.filter_1 H H0);intro.
    generalize (setModule.filter_2 H H0);intro.
    case (setModule.union_1 H1);intro.
    apply setModule.union_2;auto.
    apply setModule.union_3;auto.
    case (setModule.union_1 H0);intro.
    generalize (setModule.filter_1 H H1);intro.
    generalize (setModule.filter_2 H H1);intro.
    auto.
    generalize (setModule.filter_1 H H1);intro.
    generalize (setModule.filter_2 H H1);intro.
    auto.
  Qed.
  
  Lemma filter_subset : forall (s:setModule.t)(p q:setModule.elt->bool),(compat_bool setModule.E.eq p)->(compat_bool setModule.E.eq q)->(forall (x:setModule.elt),(p x)=true->(q x)=true)->(setModule.Subset (setModule.filter p s) (setModule.filter q s)).
    intros.
    unfold setModule.Subset;intros.
    generalize (setModule.filter_1 H H2);intro.
    generalize (setModule.filter_2 H H2);intro.
    generalize (H1 a H4);intro;auto.
  Qed.

  Definition strict_Subset (s t:setModule.t) := (setModule.Subset s t)/\~(setModule.Equal s t).

  Lemma strict_Subset_irrefl : forall (x:setModule.t),~(strict_Subset x x).
    unfold strict_Subset.
    intros.
    intro.
    and_elim.
    apply H0.
    apply setProps.equal_refl.
  Qed.

  Lemma Subset_strict_Subset : forall (x y z:setModule.t),(setModule.Subset x y)->(strict_Subset y z)->(strict_Subset x z).
    unfold strict_Subset.
    intros.
    and_elim.
    split.
    eapply setProps.subset_trans.
    apply H.
    apply H0.
    intro.
    apply H1.
    apply setProps.subset_antisym;trivial.
    eapply setProps.subset_trans.
    apply setProps.subset_equal.
    apply setProps.equal_sym.
    apply H2.
    trivial.
  Qed.

  Lemma Equal_strict_Subset : forall (x y z:setModule.t),(setModule.Equal x y)->(strict_Subset y z)->(strict_Subset x z).
    intros.
    eapply Subset_strict_Subset.
    apply setProps.subset_equal.
    apply H.
    trivial.
  Qed.

  Lemma strict_Subset_Subset : forall (x y z:setModule.t),(strict_Subset x y)->(setModule.Subset y z)->(strict_Subset x z).
    unfold strict_Subset.
    intros;and_elim.
    split.
    eapply setProps.subset_trans.
    apply H.
    apply H0.
    intro.
    apply H1.
    apply setProps.subset_antisym;trivial.
    eapply setProps.subset_trans.
    apply H0.
    apply setProps.subset_equal.
    apply setProps.equal_sym;trivial.
  Qed.

  Lemma strict_Subset_Equal : forall (x y z:setModule.t),(strict_Subset x y)->(setModule.Equal y z)->(strict_Subset x z).
    intros.
    eapply strict_Subset_Subset.
    apply H.
    apply setProps.subset_equal;trivial.
  Qed.

  Lemma strict_Subset_trans : forall (x y z:setModule.t),(strict_Subset x y)->(strict_Subset y z)->(strict_Subset x z).
    intros.
    eapply Subset_strict_Subset.
    unfold strict_Subset in H.
    elim H;intros.
    apply H1.
    trivial.
  Qed.

  Lemma strict_Subset_iff_1 : forall (x y:setModule.t),(strict_Subset x y)<->(setModule.Subset x y)/\(~setModule.Empty (setModule.diff y x)).
    unfold strict_Subset;split;intros.
    and_elim.
    split;trivial.
    intro.
    apply H0.
    apply setProps.subset_antisym;trivial.
    intro;intro.
    case (setProps.In_dec a x);intro;trivial.
    elim (H1 a);auto.
    and_elim.
    split;trivial.
    intro.
    assert (setModule.Subset y x).
    apply setProps.subset_equal.
    apply setProps.equal_sym;trivial.
    generalize (setProps.diff_subset_equal H2);intro.
    assert (setModule.Empty setModule.empty).
    apply setModule.empty_1.
    apply H0.
    intro;intro.
    elim (H3 a);intros.
    elim (H4 a);auto.
  Qed.

  Lemma subset_complement_union : forall (x y:setModule.t),(setModule.Subset x y)->(setModule.Equal (setModule.union x (setModule.diff y x)) y).
    intros.
    eapply setProps.equal_trans.
    apply setProps.union_sym.
    apply setProps.equal_sym.
    eapply setProps.equal_trans.
    apply setProps.equal_sym.
    apply setProps.diff_inter_all.
    apply setProps.union_equal_2.
    eapply setProps.equal_trans.
    apply setProps.inter_sym.
    apply setProps.inter_subset_equal.
    trivial.
  Qed.

  Lemma subset_complement_cardinal : forall (x y:setModule.t),(setModule.Subset x y)->(setModule.cardinal x) + (setModule.cardinal (setModule.diff y x)) = setModule.cardinal y.
    intros.
    rewrite <- (setProps.Equal_cardinal (subset_complement_union x y H)).
    rewrite <- setProps.union_inter_cardinal.
    assert (setModule.Empty (setModule.inter x (setModule.diff y x))).
    intro.
    intro.
    generalize (setModule.inter_1 H0);intro.
    generalize (setModule.inter_2 H0);intro.
    apply (setModule.diff_2 H2);trivial.
    rewrite (setProps.cardinal_1 H0).
    auto with arith.
  Qed.

  Require Import Arith.

  Lemma strict_Subset_cardinal : forall (x y:setModule.t),(strict_Subset x y)->(setModule.cardinal x) < (setModule.cardinal y).
    intros.
    elim (strict_Subset_iff_1 x y);intros.
    elim (H0 H);intros.
    rewrite <- (subset_complement_cardinal x y H2).
    case (eq_nat_dec (setModule.cardinal (setModule.diff y x)) 0);intro.
    elim H3.
    apply setProps.cardinal_inv_1;trivial.
    destruct (setModule.cardinal (setModule.diff y x)).
    elim n;trivial.
    rewrite plus_comm.
    simpl.
    auto with arith.
  Qed.
    
  Lemma induction_type_strict_Subset : forall (P:setModule.t->Type),(forall (s:setModule.t),((forall (s':setModule.t),(strict_Subset s' s)->(P s'))->(P s)))->(forall (s:setModule.t),(P s)).
    intros.
    apply induction_type_cardinal;clear s.
    intros.
    apply X.
    intros.
    apply X0.
    apply strict_Subset_cardinal.
    trivial.
  Defined.

  Lemma strict_Subset_Empty : forall (y:setModule.t),(setModule.Empty y)->forall (x:setModule.t),~(strict_Subset x y).
    intros.
    intro.
    generalize (strict_Subset_cardinal x y H0);intro.
    rewrite (setProps.cardinal_1 H) in H1.
    inversion H1.
  Qed.
  
  Lemma strict_Subset_remove : forall (s t:setModule.t)(x:setModule.elt),(strict_Subset s t)->(setModule.In x s)->(strict_Subset (setModule.remove x s) (setModule.remove x t)).
    unfold strict_Subset.
    intros;and_elim.
    assert (setModule.In x t).
    apply H;trivial.
    assert (setProps.Add x (setModule.remove x s) s);auto.
    assert (setProps.Add x (setModule.remove x t) t);auto.
    split.
    intro;intro.
    assert (~setModule.E.eq a x).
    intro.
    apply (setModule.remove_1 (setModule.E.eq_sym H6) H5).
    generalize (setModule.remove_3 H5);intro.
    generalize (H a H7);intro.
    apply setModule.remove_2;auto.
    intro.
    apply H1.
    eapply setProps.equal_trans.
    apply setProps.equal_sym.
    apply setProps.add_remove.
    apply H0.
    apply setProps.equal_sym.
    eapply setProps.equal_trans.
    apply setProps.equal_sym.
    apply setProps.add_remove.
    apply H2.
    apply setProps.equal_sym.
    apply compat_op_add;trivial.
  Qed.

  Lemma strict_Subset_Add : forall (s t:setModule.t)(x:setModule.elt),(~setModule.In x s)->(setProps.Add x s t)->(strict_Subset s t).
    intros.
    unfold strict_Subset.
    split.
    intro;intro.
    elim (H0 a);intros.
    apply H3.
    right;trivial.
    intro.
    apply H.
    assert (setModule.In x t).
    elim (H0 x);auto.
    elim (H1 x);auto.
  Qed.

  Lemma strict_Subset_remove_Subset : forall (s t:setModule.t)(x:setModule.elt),(~setModule.In x s)->(setModule.In x t)->(strict_Subset s t)->(setModule.Subset s (setModule.remove x t)).
    intros.
    unfold strict_Subset in H1;and_elim.
    intro;intro.
    assert (~setModule.E.eq a x).
    intro.
    apply H.
    eapply setModule.In_1.
    apply H4.
    trivial.
    apply setModule.remove_2.
    auto.
    apply H1.
    trivial.
  Qed.

  Lemma remove_strict_Subset : forall (s:setModule.t)(x:setModule.elt),(setModule.In x s)->(strict_Subset (setModule.remove x s) s).
    intros.
    assert (~setModule.In x (setModule.remove x s));auto.
    assert (setProps.Add x (setModule.remove x s) s);auto.
    eapply strict_Subset_Add.
    apply H0.
    trivial.
  Qed.

  Lemma remove_add_comm : forall (s:setModule.t)(x y:setModule.elt),(~setModule.E.eq x y)->(setModule.Equal (setModule.remove x (setModule.add y s)) (setModule.add y (setModule.remove x s))).
    intros.
    split;intros.
    generalize (setModule.remove_3 H0);intro.
    case (setEltProps.eq_dec y a);intro.
    apply setModule.add_1;auto.
    generalize (setModule.add_3 n H1);intro.
    apply setModule.add_2.
    apply setModule.remove_2.
    intro.
    elim (setModule.remove_1 H3 H0).
    trivial.
    case (setEltProps.eq_dec y a);intro.
    apply setModule.remove_2.
    intro.
    apply H;eauto.
    auto.
    generalize (setModule.add_3 n H0);intro.
    apply setModule.remove_2.
    intro.
    elim (setModule.remove_1 H2 H1).
    apply setModule.add_2.
    eapply setModule.remove_3.
    apply H1.
  Qed.

  Lemma in_not_in_neq : forall (s:setModule.t)(x y:setModule.elt),(setModule.In x s)->(~setModule.In y s)->(~setModule.E.eq x y).
    intros.
    intro.
    apply H0.
    eapply setModule.In_1.
    apply H1.
    trivial.
  Qed.

  Lemma remove_union : forall (s t:setModule.t)(x:setModule.elt),(setModule.Equal (setModule.union (setModule.remove x s) (setModule.remove x t)) (setModule.remove x (setModule.union s t))).
    intros.
    split;intros.
    case (setModule.union_1 H);intro.
    apply setModule.remove_2.
    intro.
    elim (setModule.remove_1 H1 H0).
    apply setModule.union_2.
    eapply setModule.remove_3.
    apply H0.
    apply setModule.remove_2.
    intro.
    elim (setModule.remove_1 H1 H0).
    apply setModule.union_3.
    eapply setModule.remove_3.
    apply H0.
    generalize (setModule.remove_3 H);intro.
    case (setModule.union_1 H0);intro.
    apply setModule.union_2.
    apply setModule.remove_2.
    intro.
    elim (setModule.remove_1 H2 H).
    trivial.
    apply setModule.union_3.
    apply setModule.remove_2.
    intro.
    elim (setModule.remove_1 H2 H).
    trivial.
  Qed.

  Lemma Subset_diff_union : forall (s t:setModule.t),(setModule.Subset s t)->(setModule.Equal (setModule.union (setModule.diff t s) s) t).
    intros.
    assert (setModule.Equal s (setModule.inter t s)).
    split;intros.
    apply setModule.inter_3.
    apply H;trivial.
    trivial.
    eapply setModule.inter_2.
    apply H0.
    eapply setProps.equal_trans.
    apply setProps.union_equal_2.
    apply H0.
    apply setProps.diff_inter_all.
  Qed.

  Lemma Subset_diff_diff : forall (s t:setModule.t),(setModule.Subset s t)->(setModule.Equal (setModule.diff t (setModule.diff t s)) s).
    intros;split;intros.
    generalize (setModule.diff_1 H0);intro.
    generalize (setModule.diff_2 H0);intro.
    case (setProps.In_dec a s);intro;trivial.
    elim H2.
    apply setModule.diff_3;trivial.
    apply setModule.diff_3;auto.
    intro.
    apply (setModule.diff_2 H1);trivial.
  Qed.

  Lemma Subset_union_Equal_Subset : forall (s t u:setModule.t),(setModule.Subset s t)->(setModule.Subset (setModule.union s u) (setModule.union t u)).
    intros.
    intro;intro.
    case (setModule.union_1 H0);intro.
    apply setModule.union_2.
    auto.
    auto.
  Qed.

  Lemma Subset_union_Subset_Subset : forall (s s' t t':setModule.t),(setModule.Subset s s')->(setModule.Subset t t')->(setModule.Subset (setModule.union s t) (setModule.union s' t')).
    intros.
    intro;intros.
    case (setModule.union_1 H1);intro;auto.
  Qed.

  Lemma Equal_union_Equal_Equal : forall  (s s' t t':setModule.t),(setModule.Equal s s')->(setModule.Equal t t')->(setModule.Equal (setModule.union s t) (setModule.union s' t')).
    intros.
    apply setProps.subset_antisym.
    apply Subset_union_Subset_Subset;apply setProps.subset_equal;trivial.
    apply Subset_union_Subset_Subset;apply setProps.subset_equal;apply setProps.equal_sym;trivial.
  Qed.

  Lemma empty_Subset : forall (s:setModule.t),(setModule.Subset setModule.empty s).
    red;intros.
    elim (setModule.empty_1 H).
  Qed.

  Lemma Subset_Union_Union_Subset : forall (s s' t t' u:setModule.t),(setModule.Subset s (setModule.union t u))->(setModule.Subset s' (setModule.union t' u))->(setModule.Subset (setModule.union s s') (setModule.union (setModule.union t t') u)).
    intros.
    red;intros.
    case (setModule.union_1 H1);clear H1;intro.
    case (setModule.union_1 (H a H1));intro.
    apply setModule.union_2;apply setModule.union_2;auto.
    apply setModule.union_3;auto.
    case (setModule.union_1 (H0 a H1));intro.
    apply setModule.union_2;apply setModule.union_3;auto.
    apply setModule.union_3;auto.  
  Qed.

  Lemma Subset_Union : forall (s t u:setModule.t),(setModule.Subset s u)->(setModule.Subset t u)->(setModule.Subset (setModule.union s t) u).
    intros.
    red;intros.
    case (setModule.union_1 H1);intro.
    apply H;auto.
    apply H0;auto.
  Qed.

  Definition disjoint_sets (s s':setModule.t) := setModule.Subset (setModule.inter s s') setModule.empty.

  Lemma disjoint_sets_impl : forall (s s':setModule.t),(disjoint_sets s s')->forall (x:setModule.elt),(setModule.In x s)->~(setModule.In x s').
    intros.
    intro.
    unfold disjoint_sets in H.
    generalize setModule.empty_1;intro.
    elim (H2 x).
    apply H.
    apply setModule.inter_3;trivial.
  Qed.

  Lemma disjoint_sets_sym : forall (s s':setModule.t),(disjoint_sets s s')->(disjoint_sets s' s).
    unfold disjoint_sets.
    intros.
    red;intros.
    apply H.
    apply setModule.inter_3;[eapply setModule.inter_2 | eapply setModule.inter_1];eauto.
  Qed.

  Lemma disjoint_sets_empty : forall (s:setModule.t),(disjoint_sets setModule.empty s).
    unfold disjoint_sets;intros.
    red;intros.
    generalize setModule.empty_1;intro.
    eapply setModule.inter_1;eauto.
  Qed.

  Lemma disjoint_sets_cond : forall (s t:setModule.t),(forall (x:setModule.elt),(setModule.In x s)->~(setModule.In x t))->(disjoint_sets s t).
    intros.
    unfold disjoint_sets.
    red;intros.
    generalize (setModule.inter_1 H0);intro.
    generalize (setModule.inter_2 H0);intro.
    elim (H a);auto.
  Qed.

  Lemma disjoint_sets_union_1 : forall (s t1 t2:setModule.t),(disjoint_sets s (setModule.union t1 t2))->(disjoint_sets s t1).
    unfold disjoint_sets;intros.
    red;intros.
    apply H.
    apply setModule.inter_3.
    eapply setModule.inter_1;eauto.
    apply setModule.union_2.
    eapply setModule.inter_2;eauto.
  Qed.

  Lemma disjoint_sets_union_2 : forall (s t1 t2:setModule.t),(disjoint_sets s (setModule.union t1 t2))->(disjoint_sets s t2).
    unfold disjoint_sets;intros.
    red;intros.
    apply H.
    apply setModule.inter_3.
    eapply setModule.inter_1;eauto.
    apply setModule.union_3.
    eapply setModule.inter_2;eauto.
  Qed.

  Lemma disjoint_sets_neq : forall (s t:setModule.t),(disjoint_sets s t)->forall (x y:setModule.elt),(setModule.In x s)->(setModule.In y t)->~(setModule.E.eq x y).
    intros.
    red;intro.
    elim (disjoint_sets_impl s t H x H0).
    eapply setModule.In_1.
    apply setModule.E.eq_sym;apply H2.
    trivial.
  Qed.

  Lemma disjoint_sets_singleton_1 : forall (x:setModule.elt)(s:setModule.t),(disjoint_sets (setModule.singleton x) s)->~(setModule.In x s).
    intros.
    eapply disjoint_sets_impl.
    apply H.
    apply setModule.singleton_2.
    auto.
  Qed.

  Lemma disjoint_sets_singleton_2 : forall (x:setModule.elt)(s:setModule.t),~(setModule.In x s)->(disjoint_sets (setModule.singleton x) s).
    intros.
    unfold disjoint_sets.
    red;intros.
    generalize (setModule.inter_1 H0);intro.
    generalize (setModule.inter_2 H0);intro.
    generalize (setModule.singleton_1 H1);intro.
    elim H.
    eapply setModule.In_1.
    apply setModule.E.eq_sym.
    apply H3.
    trivial.
  Qed.

  Lemma disjoint_sets_subset : forall (s s' t:setModule.t),(setModule.Subset s s')->(disjoint_sets s' t)->(disjoint_sets s t).
    unfold disjoint_sets.
    intros.
    eapply setProps.subset_trans.
    eapply setProps.inter_subset_3.
    eapply setProps.subset_trans.
    apply setProps.inter_subset_1.
    apply H.
    apply setProps.inter_subset_2.
    trivial.
  Qed.

  Lemma disjoint_sets_add_1 : forall (s t:setModule.t)(x:setModule.elt),~(setModule.In x s)->(disjoint_sets (setModule.add x s) t)->(disjoint_sets s (setModule.add x t)).
    intros.
    unfold disjoint_sets.
    intro;intros.
    unfold disjoint_sets in H0.
    apply H0.
    generalize (setModule.inter_1 H1).
    generalize (setModule.inter_2 H1).
    intros.
    generalize setProps.Add_add.
    unfold setProps.Add.
    intros.
    elim (H4 t x a);intros.
    case (H5 H2);intro.
    apply setModule.inter_3.
    apply setModule.add_1.
    trivial.
    elim H.
    eapply setModule.In_1.
    apply (setModule.E.eq_sym H7).
    trivial.
    apply setModule.inter_3.
    apply setModule.add_2.
    trivial.
    trivial.
  Qed.

  Lemma subset_empty_union : forall (s t:setModule.t),(setModule.Subset s t)->(setModule.Subset (setModule.union setModule.empty s) t).
    intros.
    red;intros.
    apply H.
    case (setModule.union_1 H0);auto.
    intro.
    elim (setModule.empty_1 H1).
  Qed.

  Definition option_eq (x x':option setModule.elt) :=
    match (x,x') with
      | (None,None) => True
      | (Some x,Some x') => setModule.E.eq x x'
      | _ => False
    end.

  Lemma option_eq_refl : forall (x:option setModule.elt),(option_eq x x).
    destruct x;unfold option_eq.
    apply setModule.E.eq_refl.
    trivial.
  Qed.

  Lemma option_eq_sym : forall (x y:option setModule.elt),(option_eq x y)->(option_eq y x).
    destruct x;destruct y;unfold option_eq;intros;try tauto.
    apply setModule.E.eq_sym.
    trivial.
  Qed.

  Definition compat_f (f:setModule.elt->option setModule.elt) := forall (x x':setModule.elt),(setModule.E.eq x x')->(option_eq (f x) (f x')).

  Definition add_option_f (f:setModule.elt->option setModule.elt)(x:setModule.elt)(s:setModule.t) := match f x with None => s | Some y => setModule.add y s end.
  
  Definition map (f:setModule.elt->option setModule.elt)(s:setModule.t) := setModule.fold (add_option_f f) s setModule.empty.

  Lemma compat_op_map_add_f : forall (f:setModule.elt->option setModule.elt),(compat_f f)->(compat_op setModule.E.eq setModule.Equal (add_option_f f)).
    unfold compat_op, add_option_f;intros.
    generalize (H x x' H0).
    destruct (f x);destruct (f x');unfold option_eq;try tauto.
    intros.
    apply compat_op_add;auto.
  Qed.

  Lemma transpose_map_add_f : forall (f:setModule.elt->option setModule.elt),(transpose setModule.Equal (add_option_f f)).
    unfold transpose, add_option_f.
    intros.
    destruct (f x);destruct (f y);try (apply setProps.equal_refl).
    apply transpose_add;auto.
  Qed.

  Lemma In_map_impl : forall (f:setModule.elt->option setModule.elt),(compat_f f)->forall (s:setModule.t)(y:setModule.elt),(setModule.In y (map f s))->(exists x:setModule.elt, (setModule.In x s) /\ (option_eq (f x) (Some y))).
    intros f H s.
    elim s using setProps.set_induction;unfold map;intros.
    generalize (setProps.fold_1 (s:=s0) setSetoid setModule.empty (add_option_f f) H0);intro.
    elim (H2 y);intros.
    elim (setModule.empty_1 (H3 H1)).
    generalize (setProps.fold_2 (s:=s0) (s':=s') (x:=x) setSetoid setModule.empty (compat_op_map_add_f f H) (transpose_map_add_f f) H1 H2);intro.
    elim (H4 y);intros.
    generalize (H5 H3);clear H5;clear H6;intros.
    unfold add_option_f in H5;fold (add_option_f f) in H5.
    case (Option_elim_option setModule.elt (f x));intro.
    ex_elim s1 y0.
    rewrite s1 in H5.
    generalize (setProps.Add_add (setModule.fold (add_option_f f) s0 setModule.empty) y0).
    unfold setProps.Add.
    intros.
    elim (H6 y);intros.
    case (H7 H5);clear H7;clear H8;intros.
    exists x.
    split.
    elim (H2 x);auto.
    rewrite s1.
    unfold option_eq.
    auto.
    elim (H0 y H7).
    intro x0;intros;and_elim.
    exists x0.
    split;trivial.
    elim (H2 x0);auto.
    rewrite e in H5.
    elim (H0 y H5).
    intro x0;intros;and_elim.
    exists x0.
    split;trivial.
    elim (H2 x0);auto.
  Qed.

  Lemma impl_In_map : forall (f:setModule.elt->option setModule.elt),(compat_f f)->forall (s:setModule.t)(x y:setModule.elt),(setModule.In x s)->(option_eq (f x) (Some y))->(setModule.In y (map f s)).
    intros.
    assert (~setModule.In x (setModule.remove x s));auto.
    assert (setProps.Add x (setModule.remove x s) s);auto.
    generalize (setProps.fold_2 (s:=setModule.remove x s) (s':=s) (x:=x) setSetoid setModule.empty (compat_op_map_add_f f H) (transpose_map_add_f f) H2 H3);intro. 
    unfold map.
    elim (H4 y);intros.
    apply H6.
    unfold add_option_f.
    destruct (f x).
    apply setModule.add_1.
    unfold option_eq in H1.
    auto.
    unfold option_eq in H1.
    elim H1.
  Qed.

  Lemma map_empty : forall (f:setModule.elt->option setModule.elt),(compat_f f)->forall (s:setModule.t),(setModule.Empty s)->(setModule.Empty (map f s)).
    intros.
    generalize (setProps.fold_1 (s:=s) setSetoid setModule.empty (add_option_f f) H0);intro.
    unfold map.
    unfold setModule.Empty;intros.
    red;intros.
    elim (H1 a);intros.
    generalize (H3 H2);intro.
    elim (setModule.empty_1 H5).
  Qed.

  Lemma map_Subset : forall (f g:setModule.elt->option setModule.elt),(compat_f f)->(compat_f g)->(forall (x:setModule.elt),(option_eq (f x) (g x)))->forall (s:setModule.t),(setModule.Subset (map f s) (map g s)).
    intros.
    elim s using setProps.set_induction;intros;unfold setModule.Subset;intros.
    generalize (map_empty f H s0 H2);intro.
    elim (H4 a);trivial.
    generalize (setProps.fold_2 (s:=s0) (s':=s') (x:=x) setSetoid setModule.empty (compat_op_map_add_f f H) (transpose_map_add_f f) H3 H4);intro.
    generalize (setProps.fold_2 (s:=s0) (s':=s') (x:=x) setSetoid setModule.empty (compat_op_map_add_f g H0) (transpose_map_add_f g) H3 H4);intro.
    unfold add_option_f in H6;fold (add_option_f f) in H6.
    unfold add_option_f in H7;fold (add_option_f g) in H7.
    generalize (H1 x);intro.
    destruct (f x);destruct (g x);unfold option_eq in H8;try tauto.
    elim (H7 a);intros.
    apply H10.
    clear H9;clear H10.
    elim (H6 a);intros.
    generalize (H9 H5).
    clear H9;clear H10.
    intro.
    fold (map f s0) in H9.
    fold (map g s0).
    generalize (setProps.Add_add (map f s0) e).
    unfold setProps.Add;intros.
    elim (H10 a);intros.
    case (H11 H9);clear H10;clear H11;clear H12.
    intro.
    apply setModule.add_1.
    eapply setModule.E.eq_trans.
    apply (setModule.E.eq_sym H8).
    trivial.
    intro.
    apply setModule.add_2.
    apply H2;trivial.
    fold (map f s0) in H6.
    fold (map g s0) in H7.
    fold (map f s') in H6.
    fold (map g s') in H7.
    elim (H7 a);intros.
    apply H10.
    apply H2.
    elim (H6 a);auto.
  Qed.
    
  Lemma map_Equal : forall (f g:setModule.elt->option setModule.elt),(compat_f f)->(compat_f g)->(forall (x:setModule.elt),(option_eq (f x) (g x)))->forall (s:setModule.t),(setModule.Equal (map f s) (map g s)).    
    intros.
    apply setProps.subset_antisym.
    apply map_Subset;auto.
    apply map_Subset;auto.
    intros.
    apply option_eq_sym.
    apply H1.
  Qed.

  Definition union_option_f (f:setModule.elt->setModule.elt->option setModule.elt)(t:setModule.t)(x:setModule.elt)(s:setModule.t) := setModule.union (map (f x) t) s.

  Definition compat_f2 (f:setModule.elt->setModule.elt->option setModule.elt) := forall (x x' y y':setModule.elt),(setModule.E.eq x x')->(setModule.E.eq y y')->(option_eq (f x y) (f x' y')).

  Lemma compat_f2_impl_compat_f : forall (f:setModule.elt->setModule.elt->option setModule.elt),(compat_f2 f)->forall (x:setModule.elt),(compat_f (f x)).
    unfold compat_f, compat_f2;intros.
    apply H.
    apply setModule.E.eq_refl.
    auto.
  Qed.

  Definition map2 (f:setModule.elt->setModule.elt->option setModule.elt)(s t:setModule.t) := setModule.fold (union_option_f f t) s setModule.empty.

  Lemma compat_op_map2_add_f : forall (f:setModule.elt->setModule.elt->option setModule.elt),(compat_f2 f)->forall (t:setModule.t)(x:setModule.elt),(compat_op setModule.E.eq setModule.Equal (union_option_f f t)).
    unfold compat_op, union_option_f;intros.
    apply Equal_union_Equal_Equal;auto.
    apply map_Equal;try (apply compat_f2_impl_compat_f;auto).
    intros.
    apply H;auto.
  Qed.

  Lemma transpose_map2_add_f : forall (f:setModule.elt->setModule.elt->option setModule.elt)(t:setModule.t),(transpose setModule.Equal (union_option_f f t)).
    unfold transpose, union_option_f.
    intros.
    eapply setProps.equal_trans.
    apply setProps.equal_sym.
    apply setProps.union_assoc.
    apply setProps.equal_sym.
    eapply setProps.equal_trans.
    apply setProps.equal_sym.
    apply setProps.union_assoc.
    apply Equal_union_Equal_Equal.
    apply setProps.union_sym.
    apply setProps.equal_refl.
  Qed.

  Lemma In_map2_impl : forall (f:setModule.elt->setModule.elt->option setModule.elt),(compat_f2 f)->forall (s t:setModule.t)(z:setModule.elt),(setModule.In z (map2 f s t))->(exists x:setModule.elt,exists y:setModule.elt,(setModule.In x s)/\(setModule.In y t)/\(option_eq (f x y) (Some z))).
    intros f H s.
    elim s using setProps.set_induction;unfold map2;intros.
    generalize (setProps.fold_1 (s:=s0) setSetoid setModule.empty (union_option_f f t) H0);intro.
    elim (H2 z);intros.
    elim (setModule.empty_1 (H3 H1)).
    generalize (setProps.fold_2 (s:=s0) (s':=s') (x:=x) setSetoid setModule.empty (compat_op_map2_add_f f H t x) (transpose_map2_add_f f t) H1 H2);intro.
    elim (H4 z);intros.
    generalize (H5 H3);clear H5;clear H6;intros.
    unfold union_option_f in H5.
    case (setModule.union_1 H5);intro.
    elim (In_map_impl (f x) (compat_f2_impl_compat_f f H x) t z H6);intro y;intros;and_elim.
    exists x.
    exists y.
    split.
    elim (H2 x);auto.
    split;trivial.
    fold (union_option_f f t) in H6.
    elim (H0 t z H6).
    intro x0;intro.
    ex_elim H7 y.
    exists x0.
    exists y.
    split.
    elim (H2 x0);auto.
    split;trivial.
  Qed.

  Lemma impl_In_map2 : forall (f:setModule.elt->setModule.elt->option setModule.elt),(compat_f2 f)->forall (s t:setModule.t)(x y z:setModule.elt),(setModule.In x s)->(setModule.In y t)->(option_eq (f x y) (Some z))->(setModule.In z (map2 f s t)).
    intros.
    unfold map2.
    assert (~setModule.In x (setModule.remove x s));auto.
    assert (setProps.Add x (setModule.remove x s) s);auto.
    generalize (setProps.fold_2 (s:=setModule.remove x s) (s':=s) (x:=x) setSetoid setModule.empty (compat_op_map2_add_f f H t x) (transpose_map2_add_f f t) H3 H4);intro.
    elim (H5 z).
    intros.
    apply H7.
    clear H6;clear H7.
    unfold union_option_f.
    apply setModule.union_2.
    eapply impl_In_map;eauto.
    apply compat_f2_impl_compat_f.
    trivial.
  Qed.

  Lemma mem_Equal : forall (s t:setModule.t),(setModule.Equal s t)->forall (x:setModule.elt),(setModule.mem x s)=(setModule.mem x t).
    intros.
    case (Bool_elim_bool (setModule.mem x s));intro;rewrite H0;symmetry.
    generalize (setModule.mem_2 H0).
    intro.
    apply setModule.mem_1.
    elim (H x);auto.
    case (Bool_elim_bool (setModule.mem x t));intro.
    rewrite setModule.mem_1 in H0.
    discriminate H0.
    elim (H x);intros.
    apply H3.
    apply setModule.mem_2.
    auto.
    auto.
  Qed.

  Lemma mem_In_iff : forall (x y:setModule.elt)(s t:setModule.t),(setModule.In x s<->setModule.In y t)->(setModule.mem x s = setModule.mem y t).
    intros.
    case (Bool_elim_bool (setModule.mem x s));intro H0;rewrite H0;case (Bool_elim_bool (setModule.mem y t));intro H1;rewrite H1;trivial;elim H;clear H;intros.
    rewrite setModule.mem_1 in H1;try (discriminate H1).
    apply H.
    apply setModule.mem_2;auto.
    rewrite setModule.mem_1 in H0;try (discriminate H0).
    apply H2.
    apply setModule.mem_2;auto.
  Qed.

  Lemma In_mem_false : forall (x:setModule.elt)(s:setModule.t),(~setModule.In x s)->(setModule.mem x s = false).
    intros.
    case (Bool_elim_bool (setModule.mem x s));intro;auto.
    elim H.
    apply setModule.mem_2;trivial.
  Qed.
    
  Lemma mem_In_false : forall (x:setModule.elt)(s:setModule.t),(setModule.mem x s = false)->~(setModule.In x s).
    intros.
    red;intro.
    rewrite (setModule.mem_1 H0) in H.
    discriminate H.
  Qed.

  Lemma In_impl_singleton_Subset : forall (x:setModule.elt)(s t:setModule.t),(setModule.In x s -> setModule.In x t)<->(setModule.Subset (setModule.inter (setModule.singleton x) s) (setModule.inter (setModule.singleton x) t)).
    split;intros.
    red;intros.
    apply setModule.inter_3.
    eapply setModule.inter_1.
    apply H0.
    assert (setModule.E.eq x a).
    apply setModule.singleton_1.
    eapply setModule.inter_1;eauto.
    eapply setModule.In_1.
    apply H1.
    apply H.
    eapply setModule.In_1.
    apply setModule.E.eq_sym.
    apply H1.
    eapply setModule.inter_2;eauto.
    eapply setModule.inter_2.
    apply H.
    apply setModule.inter_3.
    apply setModule.singleton_2.
    auto.
    trivial.
  Qed.

  Lemma In_iff_singleton_Equal : forall (x:setModule.elt)(s t:setModule.t),(setModule.In x s <-> setModule.In x t)<->(setModule.Equal (setModule.inter (setModule.singleton x) s) (setModule.inter (setModule.singleton x) t)).
    intros.
    elim (In_impl_singleton_Subset x s t);intros.
    elim (In_impl_singleton_Subset x t s);intros.
    split;intros.
    apply setProps.subset_antisym;try tauto.
    split.
    apply H0.
    apply setProps.subset_equal;auto.
    apply H2.
    apply setProps.subset_equal.
    apply setProps.equal_sym.
    auto.
  Qed.

  Lemma inter_distrib_union : forall (s t u:setModule.t),(setModule.Equal (setModule.inter s (setModule.union t u)) (setModule.union (setModule.inter s t) (setModule.inter s u))).
    intros.
    split;intros.
    generalize (setModule.inter_1 H).
    generalize (setModule.inter_2 H);intros.
    case (setModule.union_1 H0);intro.
    apply setModule.union_2.
    apply setModule.inter_3;auto.
    apply setModule.union_3.
    apply setModule.inter_3;auto.
    case (setModule.union_1 H);intro;apply setModule.inter_3;auto.
    eapply setModule.inter_1;eauto.
    apply setModule.union_2.
    eapply setModule.inter_2;eauto.
    eapply setModule.inter_1;eauto.
    apply setModule.union_3.
    eapply setModule.inter_2;eauto.  
  Qed.

  Definition big_union (f:setModule.elt->setModule.t)(s:setModule.t) : setModule.t := setModule.fold (fun x s => setModule.union s (f x)) s setModule.empty.

  Lemma compat_op_big_union : forall (f:setModule.elt->setModule.t),(forall (x y:setModule.elt),(setModule.E.eq x y)->(setModule.Equal (f x) (f y)))->(compat_op setModule.E.eq setModule.Equal (fun (x : setModule.elt) (s : setModule.t) => setModule.union s (f x))).
    intros.
    unfold compat_op.
    intros.
    split;intros.
    case (setModule.union_1 H2);intro.
    apply setModule.union_2.
    elim (H1 a);auto.
    apply setModule.union_3.
    elim (H x x' H0 a);auto.
    case (setModule.union_1 H2);intro.
    apply setModule.union_2.
    elim (H1 a);auto.
    apply setModule.union_3.
    elim (H x x' H0 a);auto.
  Qed.

  Lemma transpose_big_union : forall (f:setModule.elt->setModule.t),transpose setModule.Equal (fun (x : setModule.elt) (s : setModule.t) => setModule.union s (f x)).
    unfold transpose.
    intros.
    split;intros.
    case (setModule.union_1 H);intro;auto.
    case (setModule.union_1 H0);intro;auto.
    case (setModule.union_1 H);intro;auto.
    case (setModule.union_1 H0);intro;auto.
  Qed.
    
  Lemma In_big_union : forall (f:setModule.elt->setModule.t),(forall (x y:setModule.elt),(setModule.E.eq x y)->(setModule.Equal (f x) (f y)))->forall (s:setModule.t)(y:setModule.elt),(setModule.In y (big_union f s))<->exists x:setModule.elt,(setModule.In x s)/\(setModule.In y (f x)).
    unfold big_union.
    intros f Hf.
    split.
    generalize y;clear y.
    elim s using setProps.set_induction.
    intros.
    elim (setProps.fold_1 (s:=s0) setSetoid setModule.empty (fun x s => setModule.union s (f x)) H y);intros.
    elim (setModule.empty_1 (H1 H0)).
    intros.
    elim (setProps.fold_2 (s:=s0) (s':=s') (x:=x) setSetoid setModule.empty (compat_op_big_union f Hf) (transpose_big_union f) H0 H1 y);intros.
    generalize (H3 H2);clear H3;clear H4;clear H2.
    intro.
    case (setModule.union_1 H2);clear H2;intro.
    elim (H y H2).
    intro x0;intros;and_elim.
    exists x0.
    split;auto.
    elim (H1 x0);auto.
    exists x.
    split;auto.
    elim (H1 x);auto.
    intros.
    ex_elim H x.
    assert (~setModule.In x (setModule.remove x s));auto.
    assert (setProps.Add x (setModule.remove x s) s);auto.
    elim (setProps.fold_2 (s:=(setModule.remove x s)) (s':=s) (x:=x) setSetoid setModule.empty (compat_op_big_union f Hf) (transpose_big_union f) H1 H2 y);intros.
    apply H4.
    apply setModule.union_3.
    auto.
  Qed.

  Lemma big_union_Subset : forall (f:setModule.elt->setModule.t),(forall (x y:setModule.elt),(setModule.E.eq x y)->(setModule.Equal (f x) (f y)))->forall (s t:setModule.t),(setModule.Subset s t)->(setModule.Subset (big_union f s) (big_union f t)).
    intros.
    red;intros.
    elim (In_big_union f H s a).
    elim (In_big_union f H t a).
    intros.
    apply H3.
    elim (H4 H1).
    intro x;intros;and_elim.
    exists x.
    split;auto.
  Qed.

  Lemma big_union_Equal : forall (f:setModule.elt->setModule.t),(forall (x y:setModule.elt),(setModule.E.eq x y)->(setModule.Equal (f x) (f y)))->forall (s t:setModule.t),(setModule.Equal s t)->(setModule.Equal (big_union f s) (big_union f t)).
    intros.
    apply setProps.subset_antisym.
    apply big_union_Subset;auto.
    red;intros.
    elim (H0 a);auto.
    apply big_union_Subset;auto.
    red;intros.
    elim (H0 a);auto.
  Qed.

  Hint Resolve empty_is_empty add_is_add empty_Subset.
End MissingSetProperties.


