(* A formalisation of the spi calculus in Coq *)
(* Copyright (C) 2005-2007 Sebastien Briais *)
(* http://lamp.epfl.ch/~sbriais/ *)

(* This library is free software; you can redistribute it and/or modify *)
(* it under the terms of the GNU Lesser General Public License as *)
(* published by the Free Software Foundation; either version 2.1 of the *)
(* License, or (at your option) any later version. *)

(* This library is distributed in the hope that it will be useful, but *)
(* WITHOUT ANY WARRANTY; without even the implied warranty of *)
(* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU *)
(* Lesser General Public License for more details. *)

(* You should have received a copy of the GNU Lesser General Public *)
(* License along with this library; if not, write to the Free Software *)
(* Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA *)
(* 02110-1301 USA *)

Require Export processType.
Require Import deBruijnType.
Require Import Arith.
Require Import missing.
Require Import tactics.
Require Import substLemma.
Require Import substSplit.

Module Process(Fo:FormulaeType) <: ProcessType with Module E := Fo.E with Module F := Fo.
  Module E := Fo.E.
  Module F := Fo.

  Inductive t : Set :=
    | Zero : t
    | Par : t -> t -> t
    | Sum : t -> t -> t
    | Bang : t -> t
    | New : t -> t
    | Tau : t -> t
    | Input : E.t -> t -> t
    | Output : E.t -> E.t -> t -> t
    | IfThen : F.t -> t -> t.

  Module Ord <: OrderedType with Definition t := t with Definition eq := @eq t.
    Definition t := t.
    Definition eq := @eq t.

    Fixpoint lt (p q:t) {struct p} : Prop :=
      match (p,q) with
	| (Zero,Zero) => False
	| (Zero,_) => True
	| (_,Zero) => False
	| (Par p1 p2,Par q1 q2) => (lt p1 q1) \/ ((eq p1 q1) /\ (lt p2 q2))
	| (Par _ _,_) => True
	| (_,Par _ _) => False
	| (Sum p1 p2,Sum q1 q2) => (lt p1 q1) \/ ((eq p1 q1) /\ (lt p2 q2))
	| (Sum _ _,_) => True
	| (_,Sum _ _) => False
	| (Bang p,Bang q) => (lt p q)
	| (Bang _,_) => True
	| (_,Bang _) => False
	| (New p,New q) => (lt p q)
	| (New _,_) => True
	| (_,New _) => False
	| (Tau p,Tau q) => (lt p q)
	| (Tau _,_) => True
	| (_,Tau _) => False
	| (Input a p,Input b q) => (E.Ord.lt a b) \/ ((E.Ord.eq a b) /\ (lt p q))
	| (Input _ _,_) => True
	| (_,Input _ _) => False
	| (Output a m p,Output b n q) => (E.Ord.lt a b) \/ ((E.Ord.eq a b) /\ (E.Ord.lt m n)) \/ ((E.Ord.eq a b) /\ (E.Ord.eq m n) /\ (lt p q))
	| (Output _ _ _,_) => True
	| (_,Output _ _ _) => False
	| (IfThen phi p,IfThen psi q) => (F.Ord.lt phi psi) \/ ((F.Ord.eq phi psi) /\ (lt p q))
      end.

    Lemma eq_refl : forall (x:t),(eq x x).
      unfold eq;trivial.
    Qed.

    Lemma eq_sym : forall (x y:t),(eq x y)->(eq y x).
      unfold eq;auto.
    Qed.

    Lemma eq_trans : forall (x y z:t),(eq x y)->(eq y z)->(eq x z).
      unfold eq;intros;congruence.
    Qed.

    Lemma lt_not_eq : forall (x y:t),(lt x y)->~(eq x y).
      unfold eq.
      induction x;destruct y;try tauto;intro;red;intros;try (discriminate H0);simpl in H;repeat (case H;clear H;intro H);and_elim;injection H0;clear H0;intros;try (elim (IHx y);auto;fail);try (elim (IHx1 y1 H);auto;fail);try (elim (IHx2 y2 H);auto;fail).
      elim (E.Ord.lt_not_eq H);auto.
      elim (E.Ord.lt_not_eq H);auto.
      elim (E.Ord.lt_not_eq H);auto.
      elim (F.Ord.lt_not_eq H);auto.
    Qed.

    Lemma lt_eq_trans : forall (x y z:t),(lt x y)->(eq y z)->(lt x z).
      intros.
      rewrite H0 in H;auto.
    Qed.

    Lemma eq_lt_trans : forall (x y z:t),(eq x y)->(lt y z)->(lt x z).
      intros.
      rewrite H;auto.
    Qed.

    Lemma lt_trans : forall (x y z:t),(lt x y)->(lt y z)->(lt x z).
      induction x;destruct y;destruct z;simpl;try tauto;intros;eauto.
      case H;case H0;intros;and_elim.
      left;eauto.
      left;eapply lt_eq_trans;eauto.
      left;eapply eq_lt_trans;eauto.
      right.
      split;eauto.
      eapply eq_trans;eauto.
      case H;case H0;intros;and_elim.
      left;eauto.
      left;eapply lt_eq_trans;eauto.
      left;eapply eq_lt_trans;eauto.
      right.
      split;eauto.
      eapply eq_trans;eauto.
      case H;case H0;intros;and_elim.
      left;eauto.
      left;case (E.Ord.compare t0 t2);intro;auto;elim (E.Ord.lt_not_eq H2);eauto;elim (E.Ord.lt_not_eq (E.Ord.lt_trans l H2));auto.
      left;case (E.Ord.compare t0 t2);intro;auto;elim (E.Ord.lt_not_eq H1);eauto;elim (E.Ord.lt_not_eq (E.Ord.lt_trans H1 l));auto.
      right;split;eauto.
      repeat (case H;clear H;intro H;and_elim);repeat (case H0;clear H0;intro H0;and_elim);try (left;eauto;fail);try (left;(rewrite <- H0 || rewrite H);auto;fail);right;try (left;eauto;fail);try (left;(rewrite <- H2 || rewrite H1);eauto;fail);right.
      repeat (split;eauto).
      case H;case H0;intros;and_elim.
      left;eauto.
      left;case (F.Ord.compare t0 t2);intro;auto;elim (F.Ord.lt_not_eq H2);eauto;elim (F.Ord.lt_not_eq (F.Ord.lt_trans l H2));auto.
      left;case (F.Ord.compare t0 t2);intro;auto;elim (F.Ord.lt_not_eq H1);eauto;elim (F.Ord.lt_not_eq (F.Ord.lt_trans H1 l));auto.
      right;split;eauto.
    Qed.

    Definition compare : forall (x y:t),(Compare lt eq x y).
      unfold eq.
      induction x;destruct y;try (apply LT;simpl;trivial;fail);try (apply GT;simpl;trivial;fail);try (apply EQ;reflexivity).
      case (IHx1 y1);intro;[apply LT | (case (IHx2 y2);intro;[apply LT | apply EQ | apply GT];simpl;try (right;split;auto)) | apply GT];try (simpl;left;trivial);unfold eq;try congruence.
      case (IHx1 y1);intro;[apply LT | (case (IHx2 y2);intro;[apply LT | apply EQ | apply GT];simpl;try (right;split;auto)) | apply GT];try (simpl;left;trivial);unfold eq;try congruence.
      case (IHx y);intro;[apply LT | apply EQ | apply GT];auto;congruence.
      case (IHx y);intro;[apply LT | apply EQ | apply GT];auto;congruence.
      case (IHx y);intro;[apply LT | apply EQ | apply GT];auto;congruence.
      case (E.Ord.compare t0 t1);intro;[apply LT | idtac | apply GT];try (simpl;left;auto).
      case (IHx y);intro;[apply LT | apply EQ | apply GT];try (simpl;right;auto).
      unfold E.Ord.eq in e;congruence.
      case (E.Ord.compare t0 t2);intro;[apply LT | idtac | apply GT];try (simpl;left;auto).
      case (E.Ord.compare t1 t3);intro;[apply LT | idtac | apply GT];try (simpl;right;left;auto).
      case (IHx y);intro;[apply LT | apply EQ | apply GT];try (simpl;right;right;auto).
      unfold E.Ord.eq in e;unfold E.Ord.eq in e0;congruence.
      case (F.Ord.compare t0 t1);intro;[apply LT | idtac | apply GT];try (simpl;left;auto).
      case (IHx y);intro;[apply LT | apply EQ | apply GT];try (simpl;right;split;auto).
      unfold F.Ord.eq in e;congruence.
    Defined.

    Hint Immediate eq_sym.
    Hint Resolve eq_refl eq_trans lt_not_eq lt_trans.
  End Ord.

  Fixpoint mk_nu (k:nat)(p:t) {struct k} : t :=
    match k with
      | 0%nat => p
      | (S k) => (mk_nu k (New p))
    end.

  Lemma mk_nu_0 : forall (p:t),(mk_nu 0 p) = p.
    reflexivity.
  Qed.

  Lemma mk_nu_S_1 : forall (n:nat)(p:t),(mk_nu (S n) p) = New (mk_nu n p).
    induction n.
    reflexivity.
    simpl.
    intro.
    simpl in IHn.
    rewrite IHn.
    reflexivity.
  Qed.

  Lemma mk_nu_S_2 : forall (n:nat)(p:t),(mk_nu (S n) p) = (mk_nu n (New p)).
    reflexivity.
  Qed.

  Fixpoint t_map (fe:nat->E.t->E.t) (ff:nat->F.t->F.t) (d:nat) (p:t) {struct p} : t :=
    match p with
      | Zero => Zero
      | Par p q => Par (t_map fe ff d p) (t_map fe ff d q)
      | Sum p q => Sum (t_map fe ff d p) (t_map fe ff d q)
      | Bang p => Bang (t_map fe ff d p)
      | New p => New (t_map fe ff (S d) p)
      | Tau p => Tau (t_map fe ff d p)
      | Input e p => Input (fe d e) (t_map fe ff (S d) p)
      | Output e f p => Output (fe d e) (fe d f) (t_map fe ff d p)
      | IfThen phi p => IfThen (ff d phi) (t_map fe ff d p)
    end.

  Lemma t_map_struct : deBruijn.map2_struct E.t F.t t t_map.
    split;intros;try (generalize d;clear d;induction x;intro;simpl;repeat (rewrite IHx || rewrite IHx1 || rewrite IHx2);try reflexivity;try (generalize (IHx (S d));repeat (rewrite <- plus_Snm_nSm;simpl);intro IHx';rewrite IHx');repeat (rewrite H || rewrite H0);try reflexivity).
    unfold injective.
    intro x;generalize d;clear d.
    induction x;destruct y;simpl;intro H1;try (discriminate H1);try (injection H1;intros);auto;repeat (rewrite (IHx d y) || rewrite (IHx1 d y1) || rewrite
(IHx2 d y2));try reflexivity;auto;try (rewrite (IHx (S d) y);auto;fail).
    rewrite (H d t0 t1);auto.
    rewrite (H d t0 t2);auto.
    rewrite (H d t1 t3);auto.
    rewrite (H0 d t0 t1);auto.
  Qed.

  Definition deBruijn_base := deBruijn.deBruijn_map2_struct E.t E.deBruijn_base F.t F.deBruijn_base t t_map t_map_struct.

  Lemma lift_Zero : forall (d k:nat),(deBruijn.lift deBruijn_base d k Zero)=Zero.
    intros;reflexivity.
  Qed.

  Lemma lift_Par : forall (d k:nat)(p1 p2:t),(deBruijn.lift deBruijn_base d k (Par p1 p2))=(Par (deBruijn.lift deBruijn_base d k p1) (deBruijn.lift deBruijn_base d k p2)).
    intros;reflexivity.
  Qed.

  Lemma lift_Sum : forall (d k:nat)(p1 p2:t),(deBruijn.lift deBruijn_base d k (Sum p1 p2))=(Sum (deBruijn.lift deBruijn_base d k p1) (deBruijn.lift deBruijn_base d k p2)).
    intros;reflexivity.
  Qed.

  Lemma lift_Bang : forall (d k:nat)(p:t),(deBruijn.lift deBruijn_base d k (Bang p))=(Bang (deBruijn.lift deBruijn_base d k p)).
    intros;reflexivity.
  Qed.

  Lemma lift_New : forall (d k:nat)(p:t),(deBruijn.lift deBruijn_base d k (New p))=(New (deBruijn.lift deBruijn_base (S d) k p)).
    intros;reflexivity.
  Qed.

  Lemma lift_Tau : forall (d k:nat)(p:t),(deBruijn.lift deBruijn_base d k (Tau p))=(Tau (deBruijn.lift deBruijn_base d k p)).
    intros;reflexivity.
  Qed.

  Lemma lift_Input : forall (d k:nat)(e:E.t)(p:t),(deBruijn.lift deBruijn_base d k (Input e p))=(Input ((deBruijn.lift E.deBruijn_base) d k e) (deBruijn.lift deBruijn_base (S d) k p)).
    intros;reflexivity.
  Qed.

  Lemma lift_Output : forall (d k:nat)(e1 e2:E.t)(p:t),(deBruijn.lift deBruijn_base d k (Output e1 e2 p))=(Output ((deBruijn.lift E.deBruijn_base) d k e1) ((deBruijn.lift E.deBruijn_base) d k e2) (deBruijn.lift deBruijn_base d k p)).
    intros;reflexivity.
  Qed.

  Lemma lift_IfThen : forall (d k:nat)(phi:F.t)(p:t),(deBruijn.lift deBruijn_base d k (IfThen phi p))=(IfThen ((deBruijn.lift F.deBruijn_base) d k phi) (deBruijn.lift deBruijn_base d k p)).
    intros;reflexivity.
  Qed.

  Lemma lift_mk_nu : forall (n:nat)(d k:nat)(p:t),(deBruijn.lift deBruijn_base d k (mk_nu n p))=(mk_nu n (deBruijn.lift deBruijn_base (d+n) k p)).
    induction n;intros.
    repeat (rewrite mk_nu_0).
    rewrite plus_0_r;reflexivity.
    repeat (rewrite mk_nu_S_1).
    rewrite lift_New.
    rewrite IHn.
    simpl;rewrite plus_n_Sm;reflexivity.
  Qed.

  Lemma swapR_Zero : forall (d k:nat),(deBruijn.swapR deBruijn_base d k Zero)=Zero.
    intros;reflexivity.
  Qed.

  Lemma swapR_Par : forall (d k:nat)(p1 p2:t),(deBruijn.swapR deBruijn_base d k (Par p1 p2))=(Par (deBruijn.swapR deBruijn_base d k p1) (deBruijn.swapR deBruijn_base d k p2)).
    intros;reflexivity.
  Qed.

  Lemma swapR_Sum : forall (d k:nat)(p1 p2:t),(deBruijn.swapR deBruijn_base d k (Sum p1 p2))=(Sum (deBruijn.swapR deBruijn_base d k p1) (deBruijn.swapR deBruijn_base d k p2)). 
    intros;reflexivity.
  Qed.

  Lemma swapR_Bang : forall (d k:nat)(p:t),(deBruijn.swapR deBruijn_base d k (Bang p))=(Bang (deBruijn.swapR deBruijn_base d k p)).
    intros;reflexivity.
  Qed.

  Lemma swapR_New : forall (d k:nat)(p:t),(deBruijn.swapR deBruijn_base d k (New p))=(New (deBruijn.swapR deBruijn_base (S d) k p)).
    intros;reflexivity.
  Qed.

  Lemma swapR_Tau : forall (d k:nat)(p:t),(deBruijn.swapR deBruijn_base d k (Tau p))=(Tau (deBruijn.swapR deBruijn_base d k p)).
    intros;reflexivity.
  Qed.

  Lemma swapR_Input : forall (d k:nat)(e:E.t)(p:t),(deBruijn.swapR deBruijn_base d k (Input e p))=(Input ((deBruijn.swapR E.deBruijn_base) d k e) (deBruijn.swapR deBruijn_base (S d) k p)).
    intros;reflexivity.
  Qed.

  Lemma swapR_Output : forall (d k:nat)(e1 e2:E.t)(p:t),(deBruijn.swapR deBruijn_base d k (Output e1 e2 p))=(Output ((deBruijn.swapR E.deBruijn_base) d k e1) ((deBruijn.swapR E.deBruijn_base) d k e2) (deBruijn.swapR deBruijn_base d k p)).
    intros;reflexivity.
  Qed.

  Lemma swapR_IfThen : forall (d k:nat)(phi:F.t)(p:t),(deBruijn.swapR deBruijn_base d k (IfThen phi p))=(IfThen ((deBruijn.swapR F.deBruijn_base) d k phi) (deBruijn.swapR deBruijn_base d k p)).
    intros;reflexivity.
  Qed.

  Lemma swapR_mk_nu : forall (d k k':nat)(p:t),(deBruijn.swapR deBruijn_base d k (mk_nu k' p)) = (mk_nu k' (deBruijn.swapR deBruijn_base (d+k') k p)). 
    intros.
    generalize p;clear p;generalize d;clear d;generalize k;clear k.
    induction k';intros.
    repeat (rewrite mk_nu_0).
    rewrite plus_0_r;reflexivity.
    repeat (rewrite mk_nu_S_2).
    rewrite IHk'.
    rewrite swapR_New.
    rewrite <- plus_Snm_nSm;reflexivity.
  Qed.

  Definition deBruijn_subst := deBruijn.deBruijn_s_map2_struct E.t E.deBruijn_base F.t F.deBruijn_base E.t E.deBruijn_base E.deBruijn_subst F.deBruijn_subst t t_map t_map_struct.

  Lemma low_subst_Zero : forall (d k:nat)(es:list E.t),(deBruijn.low_subst deBruijn_subst d k es Zero)=Zero.
    intros;reflexivity.
  Qed.

  Lemma low_subst_Par : forall (d k:nat)(es:list E.t)(p1 p2:t),(deBruijn.low_subst deBruijn_subst d k es (Par p1 p2))=(Par (deBruijn.low_subst deBruijn_subst d k es p1) (deBruijn.low_subst deBruijn_subst d k es p2)).
    intros;reflexivity.
  Qed.

  Lemma low_subst_Sum : forall (d k:nat)(es:list E.t)(p1 p2:t),(deBruijn.low_subst deBruijn_subst d k es (Sum p1 p2))=(Sum (deBruijn.low_subst deBruijn_subst d k es p1) (deBruijn.low_subst deBruijn_subst d k es p2)).
    intros;reflexivity.
  Qed.

  Lemma low_subst_Bang : forall (d k:nat)(es:list E.t)(p:t),(deBruijn.low_subst deBruijn_subst d k es (Bang p))=(Bang (deBruijn.low_subst deBruijn_subst d k es p)).
    intros;reflexivity.
  Qed.

  Lemma low_subst_New : forall (d k:nat)(es:list E.t)(p:t),(deBruijn.low_subst deBruijn_subst d k es (New p))=(New (deBruijn.low_subst deBruijn_subst (S d) k es p)).
    intros;reflexivity.
  Qed.

  Lemma low_subst_Tau : forall (d k:nat)(es:list E.t)(p:t),(deBruijn.low_subst deBruijn_subst d k es (Tau p))=(Tau (deBruijn.low_subst deBruijn_subst d k es p)).
    intros;reflexivity.
  Qed.

  Lemma low_subst_Input : forall (d k:nat)(es:list E.t)(e:E.t)(p:t),(deBruijn.low_subst deBruijn_subst d k es (Input e p))=(Input (deBruijn.low_subst E.deBruijn_subst d k es e) (deBruijn.low_subst deBruijn_subst (S d) k es p)).
    intros;reflexivity.
  Qed.

  Lemma low_subst_Output : forall (d k:nat)(es:list E.t)(e1 e2:E.t)(p:t),(deBruijn.low_subst deBruijn_subst d k es (Output e1 e2 p))=(Output (deBruijn.low_subst E.deBruijn_subst d k es e1) (deBruijn.low_subst E.deBruijn_subst d k es e2) (deBruijn.low_subst deBruijn_subst d k es p)).
    intros;reflexivity.
  Qed.

  Lemma low_subst_IfThen : forall (d k:nat)(es:list E.t)(phi:F.t)(p:t),(deBruijn.low_subst deBruijn_subst d k es (IfThen phi p))=(IfThen (deBruijn.low_subst F.deBruijn_subst d k es phi) (deBruijn.low_subst deBruijn_subst d k es p)).
    intros;reflexivity.
  Qed.

  Lemma low_subst_mk_nu : forall (n:nat)(d k:nat)(es:list E.t)(p:t),(deBruijn.low_subst deBruijn_subst d k es (mk_nu n p))=(mk_nu n (deBruijn.low_subst deBruijn_subst (d+n) k es p)).
    induction n;intros.
    rewrite plus_0_r.
    reflexivity.
    rewrite mk_nu_S_1.
    rewrite low_subst_New.
    rewrite IHn.
    rewrite <- mk_nu_S_1.
    rewrite <- plus_n_Sm.
    reflexivity.
  Qed.

  Fixpoint t_fold (A:Set)(fe:nat->E.t->A->A)(ff:nat->F.t->A->A)(d:nat)(x:t)(a:A) {struct x} : A :=
    match x with
      | Zero => a
      | Par p q => t_fold A fe ff d q (t_fold A fe ff d p a)
      | Sum p q => t_fold A fe ff d q (t_fold A fe ff d p a)
      | Bang p => t_fold A fe ff d p a
      | New p => t_fold A fe ff (S d) p a
      | Tau p => t_fold A fe ff d p a
      | Input e p => t_fold A fe ff (S d) p (fe d e a)
      | Output e f p => t_fold A fe ff d p (fe d f (fe d e a))
      | IfThen phi p => t_fold A fe ff d p (ff d phi a)
    end.

  Lemma t_fold_false_inv : forall (fe:nat->E.t->bool)(ff:nat->F.t->bool)(d:nat)(x:t)(b:bool),(t_fold bool (fun d x b => b || fe d x) (fun d x b => b || (ff d x)) d x b = false)->b=false.
    intros.
    generalize H;clear H.
    generalize d;clear d.
    generalize b;clear b.
    induction x;simpl;intros;auto;try (eapply IHx1;eapply IHx2;apply H;fail);try (eapply IHx;apply H;fail).
    generalize (IHx (b || fe d t0) (S d) H);intro.
    elim (orb_false_elim b (fe d t0) H0);auto.
    generalize (IHx (b || fe d t0 || fe d t1) d H);intro.
    elim (orb_false_elim (b || fe d t0) (fe d t1) H0);intros.
    elim (orb_false_elim b (fe d t0) H1);auto.
    generalize (IHx (b || ff d t0) d H);intro.
    elim (orb_false_elim b (ff d t0) H0);auto.
  Qed.

  Implicit Arguments t_fold_false_inv [fe ff d x b].

  Lemma t_fold_true : forall (fe:nat->E.t->bool)(ff:nat->F.t->bool)(d:nat)(x:t),(t_fold bool (fun d x b => b || fe d x) (fun d x b => b || (ff d x)) d x true = true).
    intros.
    generalize d;clear d.
    induction x;simpl;intros;auto;repeat (rewrite IHx1 || rewrite IHx2);auto.
  Qed.

  Lemma t_fold_invariant : forall (A:Set)(P:A->Prop)(fe:nat->E.t->A->A)(ff:nat->F.t->A->A),(forall (d:nat)(x:E.t)(a:A),(P a)->(P (fe d x a)))->(forall (d:nat)(x:F.t)(a:A),(P a)->(P (ff d x a)))->forall (d:nat)(x:t)(a:A),(P a)->(P (t_fold A fe ff d x a)).
    intros.
    generalize H1;clear H1.
    generalize a;clear a.
    generalize d;clear d.
    induction x;simpl;intros;auto.
  Qed.

  Lemma t_fold_struct : deBruijn.fold2_struct E.t F.t t t_map t_map_struct t_fold.
    split;intros.
    (* fold_orb_eq *)
    generalize H1;clear H1.
    generalize d;clear d.
    induction x;simpl;intros;auto;repeat (rewrite plus_n_Sm || rewrite H || rewrite H0 || rewrite IHx || rewrite IHx1 || rewrite IHx2);auto;intros;try (rewrite <- plus_n_Sm);try (generalize (t_fold_false_inv (H1 i H2));intro;generalize (H1 i H2);rewrite H3;auto;fail);try (apply (t_fold_false_inv (H1 i H2));fail);auto.
    elim (orb_false_elim (fi i (d1+d) t0) (fi i (d1+d) t1) (t_fold_false_inv (H1 i H2)));auto.
    elim (orb_false_elim (fi i (d1+d) t0) (fi i (d1+d) t1) (t_fold_false_inv (H1 i H2)));auto.
    (* fold_exists *)
    generalize H1;clear H1.
    generalize d;clear d.
    induction x;simpl;intros;auto.
    case (Bool_elim_bool (t_fold bool (fun d x b => b || f1 d x) (fun d x b => b || f2 d x) (d1+d) x1 false));intro.
    rewrite IHx1;auto;apply t_fold_true.
    rewrite H2 in H1.
    case (Bool_elim_bool (t_fold bool (fun d x b => b || g1 d x) (fun d x b => b || g2 d x) (d2+d) x1 false));intro.
    rewrite H3;apply t_fold_true.
    rewrite H3;auto.
    case (Bool_elim_bool (t_fold bool (fun d x b => b || f1 d x) (fun d x b => b || f2 d x) (d1+d) x1 false));intro.
    rewrite IHx1;auto;apply t_fold_true.
    rewrite H2 in H1.
    case (Bool_elim_bool (t_fold bool (fun d x b => b || g1 d x) (fun d x b => b || g2 d x) (d2+d) x1 false));intro.
    rewrite H3;apply t_fold_true.
    rewrite H3;auto.
    rewrite plus_n_Sm.
    apply IHx.
    rewrite <- plus_n_Sm;auto.
    rewrite plus_n_Sm.
    case (Bool_elim_bool (f1 (d1+d) t0));intro.
    rewrite H;auto;apply t_fold_true.
    rewrite H2 in H1.
    case (Bool_elim_bool (g1 (d2+d) t0));intro.
    rewrite H3;apply t_fold_true.
    rewrite H3.
    apply IHx.
    rewrite <- plus_n_Sm.
    auto.
    case (Bool_elim_bool (f1 (d1+d) t0));intro.
    rewrite (H d t0);simpl;auto;apply t_fold_true.
    rewrite H2 in H1.
    simpl in H1.
    case (Bool_elim_bool (g1 (d2+d) t0));intro.
    rewrite H3;simpl;apply t_fold_true.
    rewrite H3;simpl.
    case (Bool_elim_bool (f1 (d1+d) t1));intro.
    rewrite (H d t1);simpl;auto;apply t_fold_true.
    rewrite H4 in H1.
    case (Bool_elim_bool (g1 (d2+d) t1));intro.
    rewrite H5;apply t_fold_true.
    rewrite H5;auto.
    case (Bool_elim_bool (f2 (d1+d) t0));intro.
    rewrite H0;auto;apply t_fold_true.
    rewrite H2 in H1.
    case (Bool_elim_bool (g2 (d2+d) t0));intro.
    rewrite H3;apply t_fold_true.
    rewrite H3;auto.
    (* fold_map_simpl *)
    generalize a;clear a.
    generalize d;clear d.
    induction x;simpl;intros;auto;repeat (rewrite plus_n_Sm);repeat (rewrite IHx1 || rewrite IHx2 || rewrite IHx);auto.
    (* fold_equal_cond *)
    generalize H4;clear H4.
    generalize H3;clear H3.
    generalize a;clear a.
    generalize d;clear d.
    induction x;simpl;intros;auto;repeat (rewrite plus_n_Sm);repeat (rewrite <- IHx1 || rewrite <- IHx2 || rewrite <- IHx || rewrite <- H || rewrite <- H0);intros;auto;try (rewrite <- plus_n_Sm);auto;try (generalize (H3 i H5);intro H6;rewrite (t_fold_false_inv H6) in H6;auto;fail);try (apply (t_fold_false_inv (H3 i H5));auto;fail);try (apply t_fold_invariant;auto;fail);generalize (H3 i H5);intro H6;try (destruct (fi i (d1+d) t0);auto;rewrite orb_true_b in H6;rewrite t_fold_true in H6;discriminate H6);try (destruct (fi i (d1+d) t1);auto;rewrite orb_b_true in H6;rewrite t_fold_true in H6;discriminate H6).
    (* fold_invariant *)
    apply t_fold_invariant;auto.
    (* fold_orb_eq_map_inversion *)
    generalize H3;clear H3.
    generalize H2;clear H2.
    generalize H1;clear H1.
    generalize d;clear d.
    generalize y;clear y.
    induction x;simpl;intros;auto;destruct y;try (discriminate H3);auto;simpl in H2;repeat (rewrite plus_n_Sm in H1);repeat (rewrite plus_n_Sm in H2);simpl in H3;injection H3;clear H3;repeat (rewrite plus_n_Sm in H3);intros;repeat (rewrite (IHx1 y1 d) || rewrite (IHx2 y2 d));intros;auto;try (generalize (H1 i H5);intro H6;rewrite (t_fold_false_inv H6) in H6;auto;fail);try (generalize (H2 i H5);intro H6;rewrite (t_fold_false_inv H6) in H6;auto;fail);try (apply (t_fold_false_inv (H1 i H5)));try (apply (t_fold_false_inv (H2 i H5))).
    rewrite (IHx y d);intros;auto.
    rewrite (IHx y (S d));intros;auto.
    repeat (rewrite plus_n_Sm in H3);auto.
    rewrite (IHx y d);intros;auto.
    repeat (rewrite plus_n_Sm in H3).
    rewrite (IHx y (S d));intros;auto.
    rewrite (H d t0 t1);intros;auto.
    apply (t_fold_false_inv (H1 i H5)).  
    apply (t_fold_false_inv (H2 i H5)).
    generalize (H1 i H5);intro H6;rewrite (t_fold_false_inv H6) in H6;auto.
    generalize (H2 i H5);intro H6;rewrite (t_fold_false_inv H6) in H6;auto.
    rewrite (H d t1 t3);intros;auto.
    rewrite (H d t0 t2);intros;auto.
    rewrite (IHx y d);intros;auto.
    generalize (H1 i H6);intro H7;rewrite (t_fold_false_inv H7) in H7;auto.
    generalize (H2 i H6);intro H7;rewrite (t_fold_false_inv H7) in H7;auto.
    generalize (t_fold_false_inv (H1 i H6));intros.
    elim (orb_false_elim (fi i (d1 + d) t0) (fi i (d1 + d) t1) H7);auto.
    generalize (t_fold_false_inv (H2 i H6));intros.
    elim (orb_false_elim (fi i (d1 + d) t2) (fi i (d1 + d) t3) H7);auto.
    generalize (t_fold_false_inv (H1 i H6));intros.
    elim (orb_false_elim (fi i (d1 + d) t0) (fi i (d1 + d) t1) H7);auto.
    generalize (t_fold_false_inv (H2 i H6));intros.
    elim (orb_false_elim (fi i (d1 + d) t2) (fi i (d1 + d) t3) H7);auto.
    rewrite (IHx y d);intros;auto.
    rewrite (H0 d t0 t1);intros;auto.
    apply (t_fold_false_inv (H1 i H5)).
    apply (t_fold_false_inv (H2 i H5)).
    generalize (H1 i H5);intro H6;rewrite (t_fold_false_inv H6) in H6;auto.
    generalize (H2 i H5);intro H6;rewrite (t_fold_false_inv H6) in H6;auto.
  Qed.
  
  Definition deBruijn_fn_base := deBruijn.deBruijn_fn_map2_struct E.t F.t t t_map t_map_struct t_fold E.deBruijn_base E.deBruijn_fn_base F.deBruijn_base F.deBruijn_fn_base t_fold_struct.

  Lemma fn_Zero : forall (k:nat)(d:nat),(deBruijn.mem_fn deBruijn_fn_base k d Zero)=false.
    intros;reflexivity.
  Qed.

  Lemma fn_Par : forall (k:nat)(d:nat)(p q:t),(deBruijn.mem_fn deBruijn_fn_base k d (Par p q))=(deBruijn.mem_fn deBruijn_fn_base k d p) || (deBruijn.mem_fn deBruijn_fn_base k d q).
    intros;simpl.
    destruct (t_fold bool (fun d x b => b || deBruijn.mem_fn E.deBruijn_fn_base k d x) (fun d x b => b || deBruijn.mem_fn F.deBruijn_fn_base k d x) d p false);simpl;auto.
    apply (t_fold_true (fun d x => deBruijn.mem_fn E.deBruijn_fn_base k d x) (fun d x => deBruijn.mem_fn F.deBruijn_fn_base k d x)).
  Qed.

  Lemma fn_Sum : forall (k:nat)(d:nat)(p q:t),(deBruijn.mem_fn deBruijn_fn_base k d (Sum p q))=(deBruijn.mem_fn deBruijn_fn_base k d p) || (deBruijn.mem_fn deBruijn_fn_base k d q).
    intros;simpl.
    destruct (t_fold bool (fun d x b => b || deBruijn.mem_fn E.deBruijn_fn_base k d x) (fun d x b => b || deBruijn.mem_fn F.deBruijn_fn_base k d x) d p false);simpl;auto.
    apply (t_fold_true (fun d x => deBruijn.mem_fn E.deBruijn_fn_base k d x) (fun d x => deBruijn.mem_fn F.deBruijn_fn_base k d x)).
  Qed.

  Lemma fn_Bang : forall (k:nat)(d:nat)(p:t),(deBruijn.mem_fn deBruijn_fn_base k d (Bang p))=(deBruijn.mem_fn deBruijn_fn_base k d p).
    intros;reflexivity.
  Qed.
    
  Lemma fn_New : forall (k:nat)(d:nat)(p:t),(deBruijn.mem_fn deBruijn_fn_base k d (New p))=(deBruijn.mem_fn deBruijn_fn_base k (S d) p).
    intros;reflexivity.
  Qed.

  Lemma fn_Tau : forall (k:nat)(d:nat)(p:t),(deBruijn.mem_fn deBruijn_fn_base k d (Tau p))=(deBruijn.mem_fn deBruijn_fn_base k d p).
    intros;reflexivity.
  Qed.

  Lemma fn_Input : forall (k:nat)(d:nat)(e:E.t)(p:t),(deBruijn.mem_fn deBruijn_fn_base k d (Input e p))=(deBruijn.mem_fn E.deBruijn_fn_base k d e) || (deBruijn.mem_fn deBruijn_fn_base k (S d) p).
    intros;simpl.
    destruct (deBruijn.mem_fn E.deBruijn_fn_base k d e);simpl;auto.
    apply (t_fold_true (fun d x => deBruijn.mem_fn E.deBruijn_fn_base k d x) (fun d x => deBruijn.mem_fn F.deBruijn_fn_base k d x)).
  Qed.
    
  Lemma fn_Output : forall (k:nat)(d:nat)(e f:E.t)(p:t),(deBruijn.mem_fn deBruijn_fn_base k d (Output e f p))=(deBruijn.mem_fn E.deBruijn_fn_base k d e) || (deBruijn.mem_fn E.deBruijn_fn_base k d f) || (deBruijn.mem_fn deBruijn_fn_base k d p).
    intros;simpl.
    destruct (deBruijn.mem_fn E.deBruijn_fn_base k d e);destruct (deBruijn.mem_fn E.deBruijn_fn_base k d f);simpl;auto.
    apply (t_fold_true (fun d x => deBruijn.mem_fn E.deBruijn_fn_base k d x) (fun d x => deBruijn.mem_fn F.deBruijn_fn_base k d x)).
    apply (t_fold_true (fun d x => deBruijn.mem_fn E.deBruijn_fn_base k d x) (fun d x => deBruijn.mem_fn F.deBruijn_fn_base k d x)).
    apply (t_fold_true (fun d x => deBruijn.mem_fn E.deBruijn_fn_base k d x) (fun d x => deBruijn.mem_fn F.deBruijn_fn_base k d x)).
  Qed.

  Lemma fn_IfThen : forall (k:nat)(d:nat)(phi:F.t)(p:t),(deBruijn.mem_fn deBruijn_fn_base k d (IfThen phi p))=(deBruijn.mem_fn F.deBruijn_fn_base k d phi) || (deBruijn.mem_fn deBruijn_fn_base k d p).
    intros;simpl.
    destruct (deBruijn.mem_fn F.deBruijn_fn_base k d phi);simpl;auto.
    apply (t_fold_true (fun d x => deBruijn.mem_fn E.deBruijn_fn_base k d x) (fun d x => deBruijn.mem_fn F.deBruijn_fn_base k d x)).
  Qed.

  Definition deBruijn_fn_subst := deBruijn.deBruijn_s_fn_fold2_struct E.t E.deBruijn_base E.deBruijn_fn_base F.t F.deBruijn_base F.deBruijn_fn_base E.t E.deBruijn_base E.deBruijn_fn_base E.deBruijn_subst E.deBruijn_fn_subst F.deBruijn_subst F.deBruijn_fn_subst t t_map t_map_struct t_fold t_fold_struct.

  Lemma low_subst_lemma : forall (x:t)(d d' k k':nat)(es fs:list E.t),deBruijn.low_subst deBruijn_subst (d+d'+k) k' es (deBruijn.low_subst deBruijn_subst d k fs x)=deBruijn.low_subst deBruijn_subst d k (map (deBruijn.low_subst E.deBruijn_subst (d'+k) k' es) fs) (deBruijn.low_subst deBruijn_subst (d+d'+length fs) k' es x).
    intros.
    apply (SubstLemmaMap2.low_subst_lemma E.t E.deBruijn_base F.t F.deBruijn_base E.t E.deBruijn_base E.deBruijn_subst F.deBruijn_subst t t_map t_map_struct E.deBruijn_subst).
    apply E.low_subst_lemma.
    apply F.low_subst_lemma.
  Qed.

  Lemma low_subst_split_lemma : forall (x:t)(d k:nat)(es fs:list E.t),deBruijn.low_subst deBruijn_subst d 0 es (deBruijn.low_subst deBruijn_subst (d+length es) k fs x)=deBruijn.low_subst deBruijn_subst d k (app es fs) x.
    intros.
    apply (SubstSplitLemmaMap2.low_subst_split_lemma E.t E.deBruijn_base F.t F.deBruijn_base E.t E.deBruijn_base E.deBruijn_subst F.deBruijn_subst t t_map t_map_struct).
    apply E.low_subst_split_lemma.
    apply F.low_subst_split_lemma.
  Qed.

  Lemma low_subst_snoc_lemma : forall (x:t)(d k:nat)(e:E.t)(es:list E.t),deBruijn.low_subst deBruijn_subst d k (app es (cons e nil)) x =  deBruijn.low_subst deBruijn_subst d 0 (cons e nil) (deBruijn.swapR deBruijn_base d k (deBruijn.low_subst deBruijn_subst d k (map (deBruijn.lift E.deBruijn_base k 1) es) x)).
    intros.
    apply SubstLemmaFacts.low_subst_snoc_lemma.
    intros.
    apply low_subst_split_lemma.
  Qed.

  Fixpoint compute_fn (d:nat)(p:t) {struct p} : ListSet.set nat :=
    match p with
      | Zero => ListSet.empty_set _
      | Par p q => ListSet.set_union eq_nat_dec (compute_fn d p) (compute_fn d q)
      | Sum p q => ListSet.set_union eq_nat_dec (compute_fn d p) (compute_fn d q)
      | Bang p => compute_fn d p
      | New p => compute_fn (S d) p
      | Tau p => compute_fn d p
      | Input e p => ListSet.set_union eq_nat_dec (E.compute_fn d e) (compute_fn (S d) p)
      | Output e f p => ListSet.set_union eq_nat_dec (E.compute_fn d e) (ListSet.set_union eq_nat_dec (E.compute_fn d f) (compute_fn d p))
      | IfThen phi p => ListSet.set_union eq_nat_dec (F.compute_fn d phi) (compute_fn d p)
    end.

  Lemma compute_fn_ok : forall (k:nat)(d:nat)(x:t),(deBruijn.mem_fn deBruijn_fn_base k d x = ListSet.set_mem eq_nat_dec k (compute_fn d x)).
    intros k d x.
    generalize d;clear d.
    induction x;intros;try reflexivity;auto;lazy beta delta [compute_fn] iota;fold compute_fn;repeat (rewrite fn_Par || rewrite fn_Sum || rewrite fn_Bang || rewrite fn_New || rewrite fn_Tau || rewrite fn_Input || rewrite fn_Output || rewrite fn_IfThen);repeat (rewrite ListSet_mem_union);repeat (rewrite IHx1 || rewrite IHx2 || rewrite IHx || rewrite E.compute_fn_ok || rewrite F.compute_fn_ok);auto.
    rewrite orb_assoc;auto.
  Qed.
End Process.



