(* A formalisation of the spi calculus in Coq *)
(* Copyright (C) 2005-2007 Sebastien Briais *)
(* http://lamp.epfl.ch/~sbriais/ *)

(* This library is free software; you can redistribute it and/or modify *)
(* it under the terms of the GNU Lesser General Public License as *)
(* published by the Free Software Foundation; either version 2.1 of the *)
(* License, or (at your option) any later version. *)

(* This library is distributed in the hope that it will be useful, but *)
(* WITHOUT ANY WARRANTY; without even the implied warranty of *)
(* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU *)
(* Lesser General Public License for more details. *)

(* You should have received a copy of the GNU Lesser General Public *)
(* License along with this library; if not, write to the Free Software *)
(* Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA *)
(* 02110-1301 USA *)

Require Import Arith.
Require Export expressionType.
Require Export formulaeType.
Require Import deBruijnType.

Module Type ProcessType.
  Declare Module E : ExpressionType.
  Declare Module F : FormulaeType with Module E := E.

  Inductive t : Set :=
    | Zero : t
    | Par : t -> t -> t
    | Sum : t -> t -> t
    | Bang : t -> t
    | New : t -> t
    | Tau : t -> t
    | Input : E.t -> t -> t
    | Output : E.t -> E.t -> t -> t
    | IfThen : F.t -> t -> t.

  Declare Module Ord : OrderedType with Definition t := t with Definition eq := @eq t.

  Parameter mk_nu : nat -> t -> t.

  Parameter mk_nu_0 : forall (p:t),(mk_nu 0 p) = p.
  Parameter mk_nu_S_1 : forall (n:nat)(p:t),(mk_nu (S n) p) = New (mk_nu n p).
  Parameter mk_nu_S_2 : forall (n:nat)(p:t),(mk_nu (S n) p) = (mk_nu n (New p)).
  
  Parameter deBruijn_base : deBruijn.deBruijn_base t.

  Parameter lift_Zero : forall (d k:nat),(deBruijn.lift deBruijn_base d k Zero)=Zero.
  Parameter lift_Par : forall (d k:nat)(p1 p2:t),(deBruijn.lift deBruijn_base d k (Par p1 p2))=(Par (deBruijn.lift deBruijn_base d k p1) (deBruijn.lift deBruijn_base d k p2)).
  Parameter lift_Sum : forall (d k:nat)(p1 p2:t),(deBruijn.lift deBruijn_base d k (Sum p1 p2))=(Sum (deBruijn.lift deBruijn_base d k p1) (deBruijn.lift deBruijn_base d k p2)). 
  Parameter lift_Bang : forall (d k:nat)(p:t),(deBruijn.lift deBruijn_base d k (Bang p))=(Bang (deBruijn.lift deBruijn_base d k p)).
  Parameter lift_New : forall (d k:nat)(p:t),(deBruijn.lift deBruijn_base d k (New p))=(New (deBruijn.lift deBruijn_base (S d) k p)).
  Parameter lift_Tau : forall (d k:nat)(p:t),(deBruijn.lift deBruijn_base d k (Tau p))=(Tau (deBruijn.lift deBruijn_base d k p)).
  Parameter lift_Input : forall (d k:nat)(e:E.t)(p:t),(deBruijn.lift deBruijn_base d k (Input e p))=(Input ((deBruijn.lift E.deBruijn_base) d k e) (deBruijn.lift deBruijn_base (S d) k p)).
  Parameter lift_Output : forall (d k:nat)(e1 e2:E.t)(p:t),(deBruijn.lift deBruijn_base d k (Output e1 e2 p))=(Output ((deBruijn.lift E.deBruijn_base) d k e1) ((deBruijn.lift E.deBruijn_base) d k e2) (deBruijn.lift deBruijn_base d k p)).
  Parameter lift_IfThen : forall (d k:nat)(phi:F.t)(p:t),(deBruijn.lift deBruijn_base d k (IfThen phi p))=(IfThen ((deBruijn.lift F.deBruijn_base) d k phi) (deBruijn.lift deBruijn_base d k p)).

  Parameter lift_mk_nu : forall (n:nat)(d k:nat)(p:t),(deBruijn.lift deBruijn_base d k (mk_nu n p))=(mk_nu n (deBruijn.lift deBruijn_base (d+n) k p)).

  Parameter swapR_Zero : forall (d k:nat),(deBruijn.swapR deBruijn_base d k Zero)=Zero.
  Parameter swapR_Par : forall (d k:nat)(p1 p2:t),(deBruijn.swapR deBruijn_base d k (Par p1 p2))=(Par (deBruijn.swapR deBruijn_base d k p1) (deBruijn.swapR deBruijn_base d k p2)).
  Parameter swapR_Sum : forall (d k:nat)(p1 p2:t),(deBruijn.swapR deBruijn_base d k (Sum p1 p2))=(Sum (deBruijn.swapR deBruijn_base d k p1) (deBruijn.swapR deBruijn_base d k p2)). 
  Parameter swapR_Bang : forall (d k:nat)(p:t),(deBruijn.swapR deBruijn_base d k (Bang p))=(Bang (deBruijn.swapR deBruijn_base d k p)).
  Parameter swapR_New : forall (d k:nat)(p:t),(deBruijn.swapR deBruijn_base d k (New p))=(New (deBruijn.swapR deBruijn_base (S d) k p)).
  Parameter swapR_Tau : forall (d k:nat)(p:t),(deBruijn.swapR deBruijn_base d k (Tau p))=(Tau (deBruijn.swapR deBruijn_base d k p)).
  Parameter swapR_Input : forall (d k:nat)(e:E.t)(p:t),(deBruijn.swapR deBruijn_base d k (Input e p))=(Input ((deBruijn.swapR E.deBruijn_base) d k e) (deBruijn.swapR deBruijn_base (S d) k p)).
  Parameter swapR_Output : forall (d k:nat)(e1 e2:E.t)(p:t),(deBruijn.swapR deBruijn_base d k (Output e1 e2 p))=(Output ((deBruijn.swapR E.deBruijn_base) d k e1) ((deBruijn.swapR E.deBruijn_base) d k e2) (deBruijn.swapR deBruijn_base d k p)).
  Parameter swapR_IfThen : forall (d k:nat)(phi:F.t)(p:t),(deBruijn.swapR deBruijn_base d k (IfThen phi p))=(IfThen ((deBruijn.swapR F.deBruijn_base) d k phi) (deBruijn.swapR deBruijn_base d k p)).

  Parameter swapR_mk_nu : forall (d k k':nat)(p:t),(deBruijn.swapR deBruijn_base d k (mk_nu k' p)) = (mk_nu k' (deBruijn.swapR deBruijn_base (d+k') k p)). 

  (* low_subst *)
  Parameter deBruijn_subst : deBruijn.deBruijn_subst t deBruijn_base E.t E.deBruijn_base.

  Parameter low_subst_Zero : forall (d k:nat)(es:list E.t),(deBruijn.low_subst deBruijn_subst d k es Zero)=Zero.
  Parameter low_subst_Par : forall (d k:nat)(es:list E.t)(p1 p2:t),(deBruijn.low_subst deBruijn_subst d k es (Par p1 p2))=(Par (deBruijn.low_subst deBruijn_subst d k es p1) (deBruijn.low_subst deBruijn_subst d k es p2)).
  Parameter low_subst_Sum : forall (d k:nat)(es:list E.t)(p1 p2:t),(deBruijn.low_subst deBruijn_subst d k es (Sum p1 p2))=(Sum (deBruijn.low_subst deBruijn_subst d k es p1) (deBruijn.low_subst deBruijn_subst d k es p2)). 
  Parameter low_subst_Bang : forall (d k:nat)(es:list E.t)(p:t),(deBruijn.low_subst deBruijn_subst d k es (Bang p))=(Bang (deBruijn.low_subst deBruijn_subst d k es p)).
  Parameter low_subst_New : forall (d k:nat)(es:list E.t)(p:t),(deBruijn.low_subst deBruijn_subst d k es (New p))=(New (deBruijn.low_subst deBruijn_subst (S d) k es p)).
  Parameter low_subst_Tau : forall (d k:nat)(es:list E.t)(p:t),(deBruijn.low_subst deBruijn_subst d k es (Tau p))=(Tau (deBruijn.low_subst deBruijn_subst d k es p)).
  Parameter low_subst_Input : forall (d k:nat)(es:list E.t)(e:E.t)(p:t),(deBruijn.low_subst deBruijn_subst d k es (Input e p))=(Input (deBruijn.low_subst E.deBruijn_subst d k es e) (deBruijn.low_subst deBruijn_subst (S d) k es p)).
  Parameter low_subst_Output : forall (d k:nat)(es:list E.t)(e1 e2:E.t)(p:t),(deBruijn.low_subst deBruijn_subst d k es (Output e1 e2 p))=(Output (deBruijn.low_subst E.deBruijn_subst d k es e1) (deBruijn.low_subst E.deBruijn_subst d k es e2) (deBruijn.low_subst deBruijn_subst d k es p)).
  Parameter low_subst_IfThen : forall (d k:nat)(es:list E.t)(phi:F.t)(p:t),(deBruijn.low_subst deBruijn_subst d k es (IfThen phi p))=(IfThen (deBruijn.low_subst F.deBruijn_subst d k es phi) (deBruijn.low_subst deBruijn_subst d k es p)).

  Parameter low_subst_mk_nu : forall (n:nat)(d k:nat)(es:list E.t)(p:t),(deBruijn.low_subst deBruijn_subst d k es (mk_nu n p))=(mk_nu n (deBruijn.low_subst deBruijn_subst (d+n) k es p)).

  (* free names *)
  Parameter deBruijn_fn_base : deBruijn.deBruijn_fn t deBruijn_base.

  Parameter fn_Zero : forall (k:nat)(d:nat),(deBruijn.mem_fn deBruijn_fn_base k d Zero)=false.
  Parameter fn_Par : forall (k:nat)(d:nat)(p q:t),(deBruijn.mem_fn deBruijn_fn_base k d (Par p q))=(deBruijn.mem_fn deBruijn_fn_base k d p) || (deBruijn.mem_fn deBruijn_fn_base k d q).
  Parameter fn_Sum : forall (k:nat)(d:nat)(p q:t),(deBruijn.mem_fn deBruijn_fn_base k d (Sum p q))=(deBruijn.mem_fn deBruijn_fn_base k d p) || (deBruijn.mem_fn deBruijn_fn_base k d q).
  Parameter fn_Bang : forall (k:nat)(d:nat)(p:t),(deBruijn.mem_fn deBruijn_fn_base k d (Bang p))=(deBruijn.mem_fn deBruijn_fn_base k d p).
  Parameter fn_New : forall (k:nat)(d:nat)(p:t),(deBruijn.mem_fn deBruijn_fn_base k d (New p))=(deBruijn.mem_fn deBruijn_fn_base k (S d) p).
  Parameter fn_Tau : forall (k:nat)(d:nat)(p:t),(deBruijn.mem_fn deBruijn_fn_base k d (Tau p))=(deBruijn.mem_fn deBruijn_fn_base k d p).
  Parameter fn_Input : forall (k:nat)(d:nat)(e:E.t)(p:t),(deBruijn.mem_fn deBruijn_fn_base k d (Input e p))=(deBruijn.mem_fn E.deBruijn_fn_base k d e) || (deBruijn.mem_fn deBruijn_fn_base k (S d) p).
  Parameter fn_Output : forall (k:nat)(d:nat)(e f:E.t)(p:t),(deBruijn.mem_fn deBruijn_fn_base k d (Output e f p))=(deBruijn.mem_fn E.deBruijn_fn_base k d e) || (deBruijn.mem_fn E.deBruijn_fn_base k d f) || (deBruijn.mem_fn deBruijn_fn_base k d p).
  Parameter fn_IfThen : forall (k:nat)(d:nat)(phi:F.t)(p:t),(deBruijn.mem_fn deBruijn_fn_base k d (IfThen phi p))=(deBruijn.mem_fn F.deBruijn_fn_base k d phi) || (deBruijn.mem_fn deBruijn_fn_base k d p).

  Parameter deBruijn_fn_subst : deBruijn.deBruijn_subst_fn t deBruijn_base deBruijn_fn_base E.t E.deBruijn_base E.deBruijn_fn_base deBruijn_subst.

  Parameter low_subst_lemma : forall (x:t)(d d' k k':nat)(es fs:list E.t),deBruijn.low_subst deBruijn_subst (d+d'+k) k' es (deBruijn.low_subst deBruijn_subst d k fs x)=deBruijn.low_subst deBruijn_subst d k (map (deBruijn.low_subst E.deBruijn_subst (d'+k) k' es) fs) (deBruijn.low_subst deBruijn_subst (d+d'+length fs) k' es x).

  Parameter low_subst_split_lemma : forall (x:t)(d k:nat)(es fs:list E.t),deBruijn.low_subst deBruijn_subst d 0 es (deBruijn.low_subst deBruijn_subst (d+length es) k fs x)=deBruijn.low_subst deBruijn_subst d k (app es fs) x.

  Parameter low_subst_snoc_lemma : forall (x:t)(d k:nat)(e:E.t)(es:list E.t),deBruijn.low_subst deBruijn_subst d k (app es (cons e nil)) x =  deBruijn.low_subst deBruijn_subst d 0 (cons e nil) (deBruijn.swapR deBruijn_base d k (deBruijn.low_subst deBruijn_subst d k (map (deBruijn.lift E.deBruijn_base k 1) es) x)).

  Parameter compute_fn : nat -> t -> ListSet.set nat. 
  Parameter compute_fn_ok : forall (k:nat)(d:nat)(x:t),(deBruijn.mem_fn deBruijn_fn_base k d x = ListSet.set_mem eq_nat_dec k (compute_fn d x)).
End ProcessType.
