(* A formalisation of the spi calculus in Coq *)
(* Copyright (C) 2005-2007 Sebastien Briais *)
(* http://lamp.epfl.ch/~sbriais/ *)

(* This library is free software; you can redistribute it and/or modify *)
(* it under the terms of the GNU Lesser General Public License as *)
(* published by the Free Software Foundation; either version 2.1 of the *)
(* License, or (at your option) any later version. *)

(* This library is distributed in the hope that it will be useful, but *)
(* WITHOUT ANY WARRANTY; without even the implied warranty of *)
(* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU *)
(* Lesser General Public License for more details. *)

(* You should have received a copy of the GNU Lesser General Public *)
(* License along with this library; if not, write to the Free Software *)
(* Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA *)
(* 02110-1301 USA *)

Require Import expressionType.
Require Import formulaeType.
Require Import processType.
Require Import actionType.
Require Import tactics.
Require Import deBruijnType.
Require Import missing.
Require Import reduction.
Require Import Arith.
Require Import congruenceType.

Module RedCong(Action:ActionType)(C:CongruenceType with Module A := Action.A).
  Module R := Reduction(Action).
  Export R.
  Export Action. 

  Definition R_reduction (R:relation A.t)(p q:Action.A.P.t) := forall (mu:Action.t)(a:A.t),(reduction p mu a)->(exists mu':Action.t,exists b:A.t,(Action.equal mu mu')/\(reduction q mu' b)/\(R a b)).
  
  Lemma R_reduction_inclusion : forall (R R':relation A.t),(inclusion A.t R R')->(inclusion Action.A.P.t (R_reduction R) (R_reduction R')).
    unfold inclusion;unfold R_reduction.
    intros.
    elim (H0 mu a H1);intro mu';intro.
    ex_elim H2 b.
    exists mu'.
    exists b.
    split;auto.
  Qed.

  Lemma Acongruence_reduction_CongRefl : forall (R:relation A.P.t),(reflexive A.P.t (R_reduction (C.Acongruence R))).
    unfold reflexive.
    unfold R_reduction.
    intros.
    exists mu.
    exists a.
    split.
    apply Action.equal_refl.
    split;auto.
    apply C.ACongRefl.
    eapply well_formed_reduction;eauto.
  Qed.
  
  Lemma Acongruence_reduction_CongTrans : forall (R:relation A.P.t),(transitive A.P.t (R_reduction (C.Acongruence R))).
    unfold transitive.
    unfold R_reduction.
    intros.
    elim (H mu a H1);intro mu';intros.
    ex_elim H2 b.
    elim (H0 mu' b H3);intro mu'';intros.
    ex_elim H5 c.
    exists mu''.
    exists c.
    split.
    eapply Action.equal_trans;eauto.
    split;auto.
    eapply C.ACongTrans;eauto.
  Qed.

  Lemma Acongruence_reduction_CongInput : forall (R:relation A.P.t)(p q:A.P.t),(C.Rcongruence R p q)->forall (e:E.t),(R_reduction (C.Acongruence R) (A.P.Input e p) (A.P.Input e q)).
    unfold R_reduction.
    intros.
    inversion H0.
    subst a.
    exists mu.
    exists (A.F (A.Abs q)).
    split.
    apply Action.equal_refl.
    split.
    apply RedInput.
    auto.
    apply C.AAbs;auto.
  Qed.
  
  Lemma Acongruence_reduction_CongOutput : forall (R:relation A.P.t)(p q:A.P.t),(C.Rcongruence R p q)->forall (e f:E.t),(R_reduction (C.Acongruence R) (A.P.Output e f p) (A.P.Output e f q)).
    unfold R_reduction.
    intros.
    inversion H0.
    exists mu.
    exists (A.C (A.Conc 0 m q)).
    split.
    apply Action.equal_refl.
    split.
    apply RedOutput;auto.
    apply C.AConc;auto.
  Qed.
  
  Lemma Acongruence_reduction_CongIfThen : forall (R:relation A.P.t)(p q:A.P.t),(R_reduction (C.Acongruence R) p q)->forall (phi:F.t),(R_reduction (C.Acongruence R) (A.P.IfThen phi p) (A.P.IfThen phi q)).
    unfold R_reduction.
    intros.
    inversion H0.
    subst phi0.
    subst p0.
    subst mu'.
    subst a0.
    elim (H mu0 a H3).
    intro mu'0.
    intro.
    ex_elim H1 b.
    elim (Action.mk_if_equal phi mu0 mu H6 mu'0 H1).
    intro mu';intros;and_elim.
    exists mu'.
    exists b.
    split;auto.
    split;auto.
    eapply RedIfThen;eauto.
  Qed.

  Lemma Acongruence_reduction_CongNew : forall (R:relation A.P.t)(p q:A.P.t),(R_reduction (C.Acongruence R) p q)->(R_reduction (C.Acongruence R) (A.P.New p) (A.P.New q)).
    unfold R_reduction.
    intros.
    inversion H0.
    elim (H mu0 a0 H2);intro mu'0;intros.
    ex_elim H6 b0.
    elim (Action.mk_res_equal mu0 mu H3 mu'0 H6).
    intro mu''0;intros;and_elim.
    exists mu''0.
    exists (A.new b0).
    split;auto.
    split.
    eapply RedRes;eauto.
    apply C.ACongNew.
    auto.
  Qed.

  Lemma Acongruence_reduction_CongSumL :  forall (R:relation A.P.t),forall (p q:A.P.t),(R_reduction (C.Acongruence R) p q)->forall (r:A.P.t),(R_reduction (C.Acongruence R) (A.P.Sum p r) (A.P.Sum q r)).
    intros.
    unfold R_reduction.
    intros.
    inversion H0.
    elim (H mu a H5);intro mu';intro.
    ex_elim H6 b.
    exists mu'.
    exists b.
    split;auto.
    split;auto.
    apply RedSumL.
    auto.
    exists mu.
    exists a.
    split.
    apply Action.equal_refl.
    split.
    apply RedSumR;auto.
    apply C.ACongRefl.
    eapply well_formed_reduction;eauto.
  Qed.

  Lemma SumMonoid_reduction : (inclusion A.P.t C.SumMonoid (R_reduction (C.Acongruence C.SumMonoid))).
    unfold inclusion;unfold R_reduction.
    intros x y H.
    induction H;intros.
    inversion H.
    inversion H4.
    exists mu.
    exists a.
    split.
    apply Action.equal_refl.
    split;auto.
    apply C.ACongRefl.
    eapply well_formed_reduction;eauto.
    exists mu.
    exists a.
    split.
    apply Action.equal_refl.
    split.
    apply RedSumR;auto.
    apply C.ACongRefl.
    eapply well_formed_reduction;eauto.
    inversion H.
    exists mu.
    exists a.
    split.
    apply Action.equal_refl.
    split.
    apply RedSumR;auto.
    apply C.ACongRefl.
    eapply well_formed_reduction;eauto.
    exists mu.
    exists a.
    split.
    apply Action.equal_refl.
    split.
    apply RedSumL;auto.
    apply C.ACongRefl.
    eapply well_formed_reduction;eauto.
    inversion H.
    inversion H4.
    exists mu.
    exists a.
    split.
    apply Action.equal_refl.
    split.
    apply RedSumL;auto.
    apply C.ACongRefl.
    eapply well_formed_reduction;eauto.
    exists mu.
    exists a.
    split.
    apply Action.equal_refl.
    split.
    apply RedSumR.
    apply RedSumL;auto.
    apply C.ACongRefl.
    eapply well_formed_reduction;eauto.
    exists mu.
    exists a.
    split.
    apply Action.equal_refl.
    split.
    apply RedSumR.
    apply RedSumR;auto.
    apply C.ACongRefl.
    eapply well_formed_reduction;eauto.
  Qed.

  Lemma Acongruence_reduction_CongParL : forall (R:relation A.P.t),(inclusion A.P.t C.NewSwap R)->(inclusion A.P.t C.ScopeExtrusion R)->(C.R_preserves_swapR (C.Rcongruence R))->(C.R_preserves_lift (C.Rcongruence R))->(C.R_preserves_low_subst (C.Rcongruence R))->forall (p q:A.P.t),(C.Rcongruence R p q)->(R_reduction (C.Acongruence R) p q)->forall (r:A.P.t),(R_reduction (C.Acongruence R) (A.P.Par p r) (A.P.Par q r)).
    intros R RNewSwap RScopeExtrusion RswapR Rlift Rlow_subst.
    unfold R_reduction;intros.
    inversion H1.
    (* p -mu-> a0 *)
    subst mu0;subst p0;subst q0.
    elim (H0 mu a0 H6);intro mu';intros.
    ex_elim H2 b0.
    exists mu'.
    exists (A.par_r b0 r).
    split;auto.
    split.
    eapply RedParL;auto.
    apply C.ACongParL_par_r;auto.
    (* r -mu-> a0 *)
    subst mu0;subst p0;subst q0.
    exists mu.
    exists (A.par_l q a0).
    split.
    apply Action.equal_refl.
    split.
    apply RedParR.
    auto.
    apply C.ACongParL_par_l;auto.
    eapply well_formed_reduction;eauto.
    (* p -inp-> f *)
    (* r -out-> c *)
    subst mu;subst p0;subst q0.
    elim (H0 inp (A.F f) H4).
    intro inp';intros.
    ex_elim H2 f'.
    destruct f.
    elim (C.Acongruence_F RNewSwap RswapR H6 t0);auto.
    intro t0';intros;and_elim.
    subst f'.
    elim (Action.mk_interact_equal inp out tau H8 inp' out H2 (Action.equal_refl out));intro tau';intros;and_elim.
    exists tau'.
    exists (A.pseudo_appl_l (A.Abs t0') c).
    split;auto.
    split.
    eapply RedCloseL;eauto.
    apply C.ACongParL_pseudo_appl_l;auto.
    (* p -out-> c *)
    (* r -inp-> f *)
    subst mu;subst p0;subst q0.
    elim (H0 out (A.C c) H4).
    intro out';intros.
    ex_elim H2 c'.
    destruct c.
    elim (C.Acongruence_C RNewSwap RScopeExtrusion H6 n t0 t1);auto.
    intro t0';intros.
    ex_elim H9 t1'.
    elim (Action.mk_interact_equal out inp tau H8 out' inp H2 (Action.equal_refl inp));intro tau';intros;and_elim.
    exists tau'.
    subst c'.
    exists (A.pseudo_appl_r (A.Conc n t0' t1') f).
    split;auto.
    split.
    eapply RedCloseR;eauto.
    generalize (H11 0%nat f);simpl;auto.
  Qed.

  Lemma ParMonoid_reduction : (inclusion A.P.t C.ParMonoid (R_reduction (C.Acongruence (union A.P.t C.ParMonoid C.ScopeExtrusion)))).
    unfold inclusion;unfold R_reduction.
    intros x y H.
    induction H;intros.
    (* 0 | P and P *)
    inversion H.
    inversion H4.
    subst p0;subst mu0;subst q.
    exists mu.
    exists a0.
    split.
    apply Action.equal_refl.
    split;auto.
    apply C.AParZeroL.
    unfold inclusion, union;intros;tauto.
    eapply well_formed_reduction;eauto.
    inversion H2.
    inversion H2.
    (* P and P | 0 *)
    exists mu.
    exists (A.par_l A.P.Zero a).
    split.
    apply Action.equal_refl.
    split.
    apply RedParR.
    auto.
    apply C.AParZeroR.
    unfold inclusion, union;intros;tauto.
    eapply well_formed_reduction;eauto.
    (* P | Q and Q | P *)
    inversion H.
    subst p0;subst q0;subst mu0.
    exists mu.
    exists (A.par_l q a0).
    split.
    apply Action.equal_refl.
    split.
    apply RedParR;auto.
    apply C.AParComm_r_l.
    unfold inclusion, union;intros;auto.
    eapply well_formed_reduction;eauto.
    subst p0;subst q0;subst mu0.
    exists mu.
    exists (A.par_r a0 p).
    split.
    apply Action.equal_refl.
    split.
    apply RedParL;auto.
    apply C.AParComm_l_r.
    unfold inclusion, union;intros;auto.
    eapply well_formed_reduction;eauto.
    subst p0;subst q0;subst mu.
    elim (Action.mk_interact_exg inp out tau H6);intro tau';intros;and_elim.
    exists tau'.
    exists (A.pseudo_appl_r c f).
    split;auto.
    split.
    eapply RedCloseR;eauto.
    apply C.APseudoApplComm_l.
    unfold inclusion, union;intros;auto.
    subst p0;subst q0;subst mu.
    elim (Action.mk_interact_exg out inp tau H6);intro tau';intros;and_elim.
    exists tau'.
    exists (A.pseudo_appl_l f c).
    split;auto.
    split.
    eapply RedCloseL;eauto.
    apply C.APseudoApplComm_r.
    unfold inclusion, union;intros;auto.
    (* (P | Q) | R and P | (Q | R) *)
    inversion H.
    subst p0;subst q0;subst mu0.
    inversion H4.
    subst p0;subst q0;subst mu0.
    exists mu.
    exists (A.par_r a1 (A.P.Par q r)).
    split.
    apply Action.equal_refl.
    split.
    apply RedParL;auto.
    apply C.AParAssocL_r_r.
    unfold inclusion, union;intros;auto.
    eapply well_formed_reduction;eauto.
    subst p0;subst q0;subst mu0.
    exists mu.
    exists (A.par_l p (A.par_r a1 r)).
    split.
    apply Action.equal_refl.
    split.
    apply RedParR.
    apply RedParL;auto.
    apply C.AParAssocL_r_l.
    unfold inclusion, union;intros;auto.
    eapply well_formed_reduction;eauto.
    subst p0;subst q0;subst mu.
    exists tau.
    destruct c.
    assert (reduction (A.P.Par q r) out (A.par_r (A.C (A.Conc n t0 t1)) r)).
    apply RedParL;auto.
    repeat (rewrite A.par_r_C in H0 || rewrite A.conc_par_r_def in H0).
    exists (A.pseudo_appl_l f (A.Conc n t0 (A.P.Par t1 (deBruijn.lift A.P.deBruijn_base 0 n r)))).
    split.
    apply Action.equal_refl.
    split.
    eapply RedCloseL;eauto.
    apply C.AParAssocL_r_pal.
    unfold inclusion, union;intros;auto.
    unfold inclusion, union;intros;auto.
    subst p0;subst q0;subst mu.
    exists tau.
    destruct f.
    assert (reduction (A.P.Par q r) inp (A.par_r (A.F (A.Abs t0)) r)).
    apply RedParL;auto.
    exists (A.pseudo_appl_r c (A.abs_par_r (A.Abs t0) r)).
    split.
    apply Action.equal_refl.
    split.
    rewrite A.par_r_F in H0.
    eapply RedCloseR;eauto.
    apply C.AParAssocL_r_par;unfold inclusion, union;intros;auto.
    exists mu.
    exists (A.par_l p (A.par_l q a0)).
    split.
    apply Action.equal_refl.
    split.
    apply RedParR.
    apply RedParR;auto.
    apply C.AParAssocL_l.
    unfold inclusion, union;intros;auto.
    eapply well_formed_reduction;eauto.
    subst p0;subst q0;subst mu.
    inversion H2.
    subst p0;subst q0;subst mu.
    exists tau.
    destruct c.
    assert (reduction (A.P.Par q r) out (A.par_l q (A.C (A.Conc n t0 t1)))).
    apply RedParR;auto.
    rewrite A.par_l_C in H0.
    destruct a0.
    rewrite A.par_r_Pr in H8;discriminate H8.
    destruct a0.
    rewrite A.par_r_F in H8.
    injection H8;clear H8;intros.
    subst f.
    exists (A.pseudo_appl_l (A.Abs t2) (A.conc_par_l q (A.Conc n t0 t1))).
    split.
    apply Action.equal_refl.
    split.
    eapply RedCloseL;eauto.
    apply C.AParAssocL_pal_abs_r;unfold inclusion, union;intros;auto.
    rewrite A.par_r_C in H8;discriminate H8.
    subst p0;subst q0;subst mu.
    exists tau.
    destruct a0.
    rewrite A.par_l_Pr in H8;discriminate H8.
    destruct a0.
    rewrite A.par_l_F in H8.
    destruct c.
    assert (reduction (A.P.Par q r) tau (A.pseudo_appl_l (A.Abs t0) (A.Conc n t1 t2))).
    eapply RedCloseL;eauto.
    injection H8;clear H8;intros.
    subst f.
    exists (A.par_l p (A.pseudo_appl_l (A.Abs t0) (A.Conc n t1 t2))).
    split.
    apply Action.equal_refl.
    split.
    apply RedParR.
    auto.
    apply C.AParAssocL_pal_abs_l;unfold inclusion, union;intros;auto.
    rewrite A.par_l_C in H8;discriminate H8.
    destruct f0;destruct c0;rewrite A.pseudo_appl_l_def in H4;discriminate H4.
    destruct f0;destruct c0;rewrite A.pseudo_appl_r_def in H4;discriminate H4.
    subst p0;subst q0;subst mu.
    inversion H2.
    subst p0;subst q0;subst mu.
    destruct a0.
    rewrite A.par_r_Pr in H8;discriminate H8.
    rewrite A.par_r_F in H8;discriminate H8.
    rewrite A.par_r_C in H8.
    destruct c0.
    injection H8;clear H8;intros.
    subst c.
    exists tau.
    assert (reduction (A.P.Par q r) inp (A.par_l q (A.F f))).
    apply RedParR;auto.
    destruct f.
    rewrite A.par_l_F in H0.
    exists (A.pseudo_appl_r (A.Conc n t0 t1) (A.abs_par_l q (A.Abs t2))).
    split.
    apply Action.equal_refl.
    split.
    eapply RedCloseR;eauto.
    apply C.AParAssocL_pal_conc_r;unfold inclusion, union;intros;auto.
    subst p0;subst q0;subst mu.
    exists tau.
    destruct a0.
    rewrite A.par_l_Pr in H8;discriminate H8.
    rewrite A.par_l_F in H8;discriminate H8.
    rewrite A.par_l_C in H8.
    destruct c0.
    injection H8;clear H8;intros.
    subst c.
    assert (reduction (A.P.Par q r) tau (A.pseudo_appl_r (A.Conc n t0 t1) f)).
    eapply RedCloseR;eauto.
    exists (A.par_l p (A.pseudo_appl_r (A.Conc n t0 t1) f)).
    split.
    apply Action.equal_refl.
    split.
    apply RedParR;auto.
    apply C.AParAssocL_pal_conc_l;unfold inclusion, union;intros;auto.
    destruct f0;destruct c0;rewrite A.pseudo_appl_l_def in H4;discriminate H4.
    destruct f0;destruct c0;rewrite A.pseudo_appl_r_def in H4;discriminate H4.
  Qed.

  Lemma ScopeExtrusion_reduction : (inclusion A.P.t C.ScopeExtrusion (R_reduction (C.Acongruence (union A.P.t (union A.P.t C.ParMonoid C.ScopeExtrusion) C.NewSwap)))).
    unfold inclusion;unfold R_reduction.
    intros x y H. 
    assert (symmetric A.P.t (C.Rcongruence (union A.P.t (union A.P.t C.ParMonoid C.ScopeExtrusion) C.NewSwap))).
    eapply same_relation_symmetric.
    apply C.Rcongruence_union.
    apply C.Rcongruence_sym.
    apply union_symmetric.
    eapply same_relation_symmetric.
    apply C.Rcongruence_union.
    apply C.Rcongruence_sym.
    apply union_symmetric.
    apply C.Rcongruence_ParMonoid_sym.
    apply C.ScopeExtrusion_sym.
    apply C.NewSwap_sym.
    generalize H0;clear H0.
    intro Hsym. 
    induction H;intros.
    (* ScopeExtrusionL *)
    inversion H.
    subst p0;subst q0;subst mu0.
    inversion H4.
    subst p0;subst mu';subst a0.
      (* p -mu0-> a1 *)
    exists mu.
    exists (A.new (A.par_r a1 (deBruijn.lift A.P.deBruijn_base 0 1 q))).
    split.
    apply Action.equal_refl.
    split.
    eapply RedRes;eauto.
    apply RedParL;auto.
    apply C.AScopeExtrusionL_r.
    unfold inclusion, union;intros;auto.
    eapply well_formed_reduction;eauto.
      (* q -mu-> a0 *)
    subst p0;subst q0;subst mu0.
    elim (reduction_lift q mu a0 H4 0 1).
    intro mu';intros;and_elim.
    elim (Action.mk_res_lift_1 mu);intro mu0;intros;and_elim.
    elim (Action.mk_res_equal (deBruijn.lift deBruijn_base 0 1 mu) mu0 H5 mu').
    intro mu0';intros;and_elim.
    exists mu0'.
    exists (A.new (A.par_l p (deBruijn.lift A.deBruijn_base 0 1 a0))).
    split.
    eapply Action.equal_trans;eauto.
    apply Action.equal_sym;auto.
    split.
    eapply RedRes;eauto.
    apply RedParR;auto.
    apply C.AScopeExtrusionL_l;auto.
    unfold inclusion, union;intros;auto.
    eapply well_formed_reduction;eauto.
    apply Action.equal_sym;auto.
    subst p0;subst q0;subst mu.
      (* q -out-> c *)
    inversion H2.
    subst p0;subst mu'.
    destruct a0.
    rewrite A.new_Pr in H1;discriminate H1.
    rewrite A.new_F in H1.
    injection H1;clear H1;intros;subst f.
      (* p -mu-> \t0 *)
    elim (reduction_lift q out (A.C c) H3 0 1);intro out';intros;and_elim.
    elim (Action.mk_interact_res_lift mu inp H8 out tau H6);intro tau0;intros.
    ex_elim H7 tau1.
    elim (Action.mk_interact_equal mu (deBruijn.lift deBruijn_base 0 1 out) tau0 H9 mu out').
    intro tau2;intros;and_elim.
    elim (Action.mk_res_equal tau0 tau1 H10 tau2);auto.
    intro tau';intros;and_elim.
    exists tau'.
    destruct c.
    rewrite A.lift_C in H1.
    simpl in H1.
    exists (A.new (A.pseudo_appl_l a0 (A.Conc n (deBruijn.lift A.P.E.deBruijn_base n 1 t0) (deBruijn.lift A.P.deBruijn_base n 1 t1)))).
    split.
    eapply Action.equal_trans;eauto.
    apply Action.equal_sym;auto.
    split.
    eapply RedRes;eauto.
    eapply RedCloseL;eauto.
    apply C.AScopeExtrusionL_pal.
    unfold inclusion, union;intros;auto.
    unfold inclusion, union;intros;auto.
    apply Action.equal_refl.
    apply Action.equal_sym;auto.
    rewrite A.new_C in H1;discriminate H1.
      (* q -inp-> f *)
    subst p0;subst q0;subst mu.
    inversion H2.
    subst p0;subst mu'.
    elim (reduction_lift q inp (A.F f) H3 0 1);intro inp';intros;and_elim.
    destruct a0.
    rewrite A.new_Pr in H1;discriminate H1.
    rewrite A.new_F in H1;discriminate H1.
    rewrite A.new_C in H1.
    injection H1;clear H1;intros;subst c.
      (* p -mu-> c0 *)
    elim (Action.mk_interact_res_lift mu out H8 inp tau H6).
    intro tau0;intros.
    ex_elim H1 tau1.
    elim (Action.mk_interact_equal mu (deBruijn.lift deBruijn_base 0 1 inp) tau0 H9 mu inp').
    intro tau2;intros;and_elim.
    elim (Action.mk_res_equal tau0 tau1 H10 tau2);auto.
    intro tau';intros;and_elim.
    exists tau'.
    destruct f.
    rewrite A.lift_F in H7.
    exists (A.new (A.pseudo_appl_r c0 (A.Abs (deBruijn.lift A.P.deBruijn_base 1 1 t0)))).
    split.
    eapply Action.equal_trans;eauto.
    apply Action.equal_sym;auto.
    split.
    eapply RedRes;eauto.
    eapply RedCloseR;eauto.
    apply C.AScopeExtrusionL_par;auto.
    unfold inclusion, union;intros;auto.
    unfold inclusion, union;intros;auto.
    unfold inclusion, union;intros;auto.
    apply Action.equal_refl.
    apply Action.equal_sym;auto.
    (* ScopeExtrusionR *)
    inversion H.
    subst p0;subst mu';subst a.
    inversion H1.
    subst p0;subst q0;subst mu1.
      (* p -mu0-> a *)
    exists mu.
    exists (A.par_r (A.new a) q).
    split.
    apply Action.equal_refl.
    split.
    apply RedParL.
    eapply RedRes;eauto.
    apply C.AScopeExtrusionR_r.
    unfold inclusion, union;intros;auto.
    eapply well_formed_reduction;eauto.
    subst p0;subst q0;subst mu1.
    elim (reduction_lift_inv (deBruijn.lift A.P.deBruijn_base 0 1 q) mu0 a H6 0 1 q);auto.
    intro mu'0;intros.
    ex_elim H0 a'.
      (* q -mu'0-> a' *)
    elim (Action.mk_res_equal mu0 mu H2 (deBruijn.lift deBruijn_base 0 1 mu'0));auto.
    intro mu';intros;and_elim.
    elim (Action.mk_res_lift_1 mu'0);intro mu1;intros;and_elim.
    rewrite H10 in H8.
    injection H8;clear H8;intros;subst mu1.
    exists mu'0.
    exists (A.par_l (A.P.New p) a').
    split.
    eapply Action.equal_trans;eauto.
    split.
    apply RedParR.
    auto.
    subst a.
    apply C.AScopeExtrusionR_l.
    unfold inclusion, union;intros;auto.
    eapply well_formed_reduction;eauto.
      (* p -inp-> f *)
    subst p0;subst q0;subst mu0.
    elim (reduction_lift_inv (deBruijn.lift A.P.deBruijn_base 0 1 q) out (A.C c) H5 0 1 q);auto.
    intro out';intros.
    ex_elim H0 c'.
    destruct c'.
    rewrite A.lift_Pr in H6;discriminate H6.
    destruct a;rewrite A.lift_F in H6;discriminate H6.
    destruct c0.
    rewrite A.lift_C in H6.
    injection H6;clear H6;intros;subst c.
    elim (Action.mk_interact_equal inp out tau H8 inp (deBruijn.lift deBruijn_base 0 1 out'));auto.
    intro tau0;intros;and_elim.
    elim (Action.mk_res_equal tau mu H2 tau0);auto.
    intro tau1;intros;and_elim.
    elim (Action.mk_interact_res_lift_inv inp out' tau0 H9 tau1 H11).
    intro inp';intros;and_elim.
    ex_elim H13 tau'.
    exists tau'.
    exists (A.pseudo_appl_l (A.new_abs f) (A.Conc n t0 t1)).
    split.
    elim (Action.mk_interact_res_lift inp inp' H12 out' tau' H13);intros.
    ex_elim H14 c'.
    rewrite H9 in H15.
    injection H15;intros;subst x.
    rewrite H11 in H16.
    injection H16;intros;subst c'.
    eapply Action.equal_trans;eauto.
    split.
    eapply RedCloseL;eauto.
    rewrite <- A.new_F.
    eapply RedRes;eauto.
    apply C.AScopeExtrusionR_pal.
    unfold inclusion, union;intros;auto.
    unfold inclusion, union;intros;auto.
    apply Action.equal_refl.
    subst p0;subst q0;subst mu0.
    elim (reduction_lift_inv (deBruijn.lift A.P.deBruijn_base 0 1 q) inp (A.F f) H5 0 1 q);auto.
    intro inp';intros.
    ex_elim H0 f'.
    destruct f'.
    rewrite A.lift_Pr in H6;discriminate H6.
    destruct a.
    rewrite A.lift_F in H6.
    injection H6;clear H6;intros.
    subst f.
    elim (Action.mk_interact_equal out inp tau H8 out (deBruijn.lift deBruijn_base 0 1 inp'));auto.
    intro tau0;intros;and_elim.
    elim (Action.mk_res_equal tau mu H2 tau0);auto.
    intro tau1;intros;and_elim.
    elim (Action.mk_interact_res_lift_inv out inp' tau0 H9 tau1);auto.
    intro out';intros;and_elim.
    ex_elim H13 tau'.
    exists tau'.
    exists (A.pseudo_appl_r (A.new_conc c) (A.Abs t0)).
    split.
    elim (Action.mk_interact_res_lift out out' H12 inp' tau' H13);intros.
    ex_elim H14 c'.
    rewrite H9 in H15.
    injection H15;intros;subst x.
    rewrite H11 in H16.
    injection H16;intros;subst c'.
    eapply Action.equal_trans;eauto.
    split.
    eapply RedCloseR;eauto.
    rewrite <- A.new_C.
    eapply RedRes;eauto.
    apply C.AScopeExtrusionR_par.
    unfold inclusion, union;intros;auto.
    unfold inclusion, union;intros;auto.
    unfold inclusion, union;intros;auto.
    apply Action.equal_refl.
    destruct c0;rewrite A.lift_C in H6;discriminate H6.
  Qed.

  Lemma NewSwap_reduction : (inclusion A.P.t C.NewSwap (R_reduction (C.Acongruence C.NewSwap))).
    unfold inclusion.
    unfold R_reduction.
    intros.
    induction H.
    inversion H0.
    subst p0;subst mu';subst a.
    inversion H1.
    subst p0;subst mu0;subst a0.
    elim (reduction_swapR p mu1 a H3 0 1);intro mu0;intros;and_elim.
    elim (Action.mk_res_swapR mu1 mu' H4 mu H2).
    intro mu2;intros.
    ex_elim H6 mu3.
    elim (Action.mk_res_equal (deBruijn.swapR deBruijn_base 0 1 mu1) mu2 H7 mu0).
    intro mu0';intros;and_elim.
    elim (Action.mk_res_equal mu2 mu3 H8 mu0').
    intro mu1';intros;and_elim.
    exists mu1'.
    exists (A.new (A.new (deBruijn.swapR A.deBruijn_base 0 1 a))).
    split.
    eapply Action.equal_trans;eauto.
    apply Action.equal_sym;auto.
    split.
    eapply RedRes;eauto.
    eapply RedRes;eauto.
    apply C.ANewSwapL.
    eapply well_formed_reduction;eauto.
    auto.
    apply Action.equal_sym;auto.
  Qed.

  Lemma NewAxioms_reduction : (inclusion A.P.t C.NewAxioms (R_reduction (C.Acongruence (union _ (union _ C.ParMonoid C.ScopeExtrusion) C.NewAxioms)))).
    unfold inclusion, R_reduction;intros.
    induction H.
    (* O = (^x)0 *)
    inversion H0.
    (* (^x)0 = 0 *)
    inversion H0.
    inversion H1.
    (* (^x)P + Q = (^x)(P + Q) *)
    inversion H0.
    inversion H4.
    exists mu.
    exists (A.new a1).
    split.
    apply Action.equal_refl.
    split.
    eapply RedRes;eauto.
    apply RedSumL;auto.
    apply C.ACongRefl.
    rewrite H9.
    eapply well_formed_reduction;eauto.
    elim (reduction_lift _ _ _ H4 0 1).
    intro mu1;intros;and_elim.
    elim (Action.mk_res_lift_1 mu);intro mu2;intros;and_elim.
    elim (Action.mk_res_equal _ _ H8 _ (Action.equal_sym _ _ H5));intro mu';intros;and_elim.
    exists mu'.
    exists (A.new (deBruijn.lift A.deBruijn_base 0 1 a)).
    split.
    eapply Action.equal_trans;eauto.
    apply Action.equal_sym;auto.
    split.
    eapply RedRes;eauto.
    apply RedSumR;auto.
    apply C.ANewGarbageL.
    unfold inclusion;auto.
    eapply well_formed_reduction;eauto.
    (* (^x)(P + Q) = (^x)P + Q *)
    inversion H0.
    inversion H1.
    subst p1;subst mu1;subst a1.
    exists mu.
    exists (A.new a0).
    split.
    apply Action.equal_refl.
    split.
    apply RedSumL.
    eapply RedRes;eauto.
    apply C.ACongRefl.
    rewrite H4.
    eapply well_formed_reduction;eauto.
    subst q0;subst mu1;subst a1;subst mu'.
    elim (reduction_lift_inv _ _ _ H9 _ _ _ (refl_equal (deBruijn.lift A.P.deBruijn_base 0 1 q))).
    intro mu1;intros;and_elim.
    ex_elim H3 b;intros.
    subst a0.
    elim (Action.mk_res_equal _ _ H2 _ H5).
    intro mu';intros;and_elim.
    elim (Action.mk_res_lift_1 mu1).
    intro mu2;intros;and_elim.
    rewrite H11 in H8.
    injection H8;clear H8;intros;subst mu2.
    exists mu1.
    exists b.
    split.
    eapply Action.equal_trans;eauto.
    split.
    apply RedSumR;auto.
    apply C.ANewGarbageR.
    unfold inclusion;auto.
    eapply well_formed_reduction;eauto.
  Qed.

  Lemma BangAxioms_reduction : (inclusion A.P.t C.BangAxioms (R_reduction (C.Acongruence (union _ (union _ (union _ C.NewSwap C.ScopeExtrusion) C.ParMonoid) C.BangAxioms)))).
    unfold inclusion;unfold R_reduction.
    intros x y H.
    induction H;intros.
    (* !P = P | !P *)
    inversion H.
    subst p0;subst mu0.
    exists mu.
    exists (A.par_r a0 (A.P.Bang p)).
    split.
    apply Action.equal_refl.
    split.
    apply RedParL.
    auto.
    apply C.ACongParL_par_r.
    unfold inclusion, union;intros;auto.
    unfold inclusion, union;intros;auto.
    apply C.Rcongruence_swapR.
    unfold C.R_preserves_swapR;intros.
    repeat (apply R_preserves_f_union).
    apply C.NewSwap_preserves_swapR.
    apply C.ScopeExtrusion_preserves_swapR.
    apply C.ParMonoid_preserves_swapR.
    apply C.BangAxioms_preserves_swapR.
    apply C.ACongRefl.
    eapply well_formed_reduction;eauto.
    subst mu;subst p0.
    assert (reduction (A.P.Bang p) out (A.par_r (A.C c) (A.P.Bang p))).
    apply RedBangAct;auto.
    destruct c as [n m q].
    rewrite A.par_r_C in H0.
    rewrite A.conc_par_r_def in H0.
    exists tau.
    exists (A.pseudo_appl_l f (A.Conc n m (A.P.Par q (deBruijn.lift A.P.deBruijn_base 0 n (A.P.Bang p))))).
    split.
    apply Action.equal_refl.
    split.
    eapply RedCloseL;eauto.
    apply C.AParAssocL_r_pal.
    unfold inclusion, union;intros;try tauto.
    unfold inclusion, union;intros;try tauto.
    (* P | !P = !P *)
    inversion H.
    exists mu.
    exists a.
    split.
    apply Action.equal_refl.
    split.
    subst a.
    apply RedBangAct.
    auto.
    rewrite H3.
    apply C.ACongRefl.
    eapply well_formed_reduction;eauto.
    subst p0;subst q;subst mu0.
    inversion H4.
    subst p0;subst mu0.
    exists mu.
    exists a0.
    split.
    apply Action.equal_refl.
    split.
    auto.
    subst a0.
    eapply C.ACongTrans.
    apply C.AParComm_l_r.
    unfold inclusion, union;intros;try tauto.
    apply A.well_formed_par_r.
    eapply well_formed_reduction;eauto.
    eapply C.ACongTrans.
    apply C.AParAssocL_r_r.
    unfold inclusion, union;intros;try tauto.
    eapply well_formed_reduction;eauto.
    apply C.ACongParR_par_r.
    unfold inclusion, union;intros;try tauto.
    left;auto.
    unfold inclusion, union;intros;try tauto.
    unfold inclusion, union;intros;try tauto.
    apply C.Rcongruence_lift.
    unfold C.R_preserves_lift;intros.
    repeat (apply R_preserves_f_union).
    apply C.NewSwap_preserves_lift;auto.
    apply C.ScopeExtrusion_preserves_lift;auto.
    apply C.ParMonoid_preserves_lift;auto.
    apply C.BangAxioms_preserves_lift;auto.
    apply C.Rcongruence_swapR.
    unfold C.R_preserves_swapR;intros.
    repeat (apply R_preserves_f_union).
    apply C.NewSwap_preserves_swapR;auto.
    apply C.ScopeExtrusion_preserves_swapR;auto.
    apply C.ParMonoid_preserves_swapR;auto.
    apply C.BangAxioms_preserves_swapR;auto.
    eapply C.CongTrans.
    apply C.Rcongruence_increase with C.ParMonoid.
    unfold inclusion, union;intros;try tauto.
    apply C.CongAxiom.
    apply C.ParComm.
    apply C.Rcongruence_increase with C.BangAxioms.
    unfold inclusion, union;intros;try tauto.
    apply C.CongAxiom.
    apply C.BangExpR.
    eapply well_formed_reduction;eauto.
    subst p0;subst mu.
    exists tau.
    exists (A.par_r (A.pseudo_appl_l f c) (A.P.Bang p)).
    split.
    apply Action.equal_refl.
    split.
    eapply RedBangCloseL;eauto.
    eapply C.ACongTrans.
    apply C.AParComm_l_r.
    unfold inclusion, union;intros;try tauto.
    apply A.well_formed_par_r.
    apply A.well_formed_pseudo_appl_l.
    eapply C.ACongTrans.
    apply C.AParAssocL_r_r.
    unfold inclusion, union;intros;try tauto.
    apply A.well_formed_pseudo_appl_l.
    apply C.ACongParR_par_r.
    unfold inclusion, union;intros;try tauto.
    left;auto.
    unfold inclusion, union;intros;try tauto.
    unfold inclusion, union;intros;try tauto.
    apply C.Rcongruence_lift.
    unfold C.R_preserves_lift;intros.
    repeat (apply R_preserves_f_union).
    apply C.NewSwap_preserves_lift;auto.
    apply C.ScopeExtrusion_preserves_lift;auto.
    apply C.ParMonoid_preserves_lift;auto.
    apply C.BangAxioms_preserves_lift;auto.
    apply C.Rcongruence_swapR.
    unfold C.R_preserves_swapR;intros.
    repeat (apply R_preserves_f_union).
    apply C.NewSwap_preserves_swapR;auto.
    apply C.ScopeExtrusion_preserves_swapR;auto.
    apply C.ParMonoid_preserves_swapR;auto.
    apply C.BangAxioms_preserves_swapR;auto.
    eapply C.CongTrans.
    apply C.Rcongruence_increase with C.ParMonoid.
    unfold inclusion, union;intros;try tauto.
    apply C.CongAxiom.
    apply C.ParComm.
    apply C.Rcongruence_increase with C.BangAxioms.
    unfold inclusion, union;intros;try tauto.
    apply C.CongAxiom.
    apply C.BangExpR.
    apply A.well_formed_pseudo_appl_l.
    subst mu;subst p0;subst q.
    inversion H3.
    subst p0;subst mu.
    destruct a0.
    rewrite A.par_r_Pr in H1;discriminate H1.
    destruct a0.
    rewrite A.par_r_F in H1;discriminate H1.
    rewrite A.par_r_C in H1.
    injection H1;clear H1;intros.
    destruct c0 as [n m q].
    rewrite A.conc_par_r_def in H0.
    exists tau.
    exists (A.par_r (A.pseudo_appl_l f (A.Conc n m q)) (A.P.Bang p)).
    split.
    apply Action.equal_refl.
    split.
    eapply RedBangCloseL;eauto.
    subst c.
    rewrite <- A.conc_par_r_def.
    eapply C.ACongTrans.
    apply C.APseudoApplComm_l.
    unfold inclusion, union;intros;try tauto.
    eapply C.ACongTrans.
    apply C.AParAssocL_pal_conc_r.
    unfold inclusion, union;intros;try tauto.
    eapply C.ACongTrans.
    apply C.APseudoApplComm_r.
    unfold inclusion, union;intros;try tauto.
    eapply C.ACongTrans.
    apply C.AParAssocL_pal_abs_l.
    unfold inclusion, union;intros;try tauto.
    unfold inclusion, union;intros;try tauto.
    apply C.AParComm_l_r.
    unfold inclusion, union;intros;try tauto.
    apply A.well_formed_pseudo_appl_l.
    subst p0;subst tau0.
    generalize (Action.mk_interact_kind _ _ _ H9);intro.
    elim (reduction_silent_process H3);auto;intros.
    discriminate H8.
    inversion H3.
    subst mu0;subst p1;subst mu;subst p0.
    elim (Action.mk_interact_exg _ _ _ H6).
    intro tau';intros;and_elim.
    exists tau'.
    destruct a0.
    rewrite A.par_r_Pr in H8.
    discriminate H8.
    subst q.
    destruct a0 as [q].
    rewrite A.par_r_F in H8.
    rewrite A.abs_par_r_def in H8.
    injection H8;clear H8;intros.
    subst f.
    exists (A.par_r (A.pseudo_appl_l (A.Abs q) c) (A.P.Bang p)).
    split.
    auto.
    split.
    eapply RedBangCloseL;eauto.
    rewrite <- A.abs_par_r_def.
    eapply C.ACongTrans.
    apply C.APseudoApplComm_r.
    unfold inclusion, union;intros;try tauto.
    eapply C.ACongTrans.
    apply C.AParAssocL_pal_abs_r.
    unfold inclusion, union;intros;try tauto.
    eapply C.ACongTrans.
    apply C.APseudoApplComm_l.
    unfold inclusion, union;intros;try tauto.
    eapply C.ACongTrans.
    apply C.AParAssocL_pal_conc_l.
    unfold inclusion, union;intros;try tauto.
    unfold inclusion, union;intros;try tauto.
    eapply C.ACongTrans.
    apply C.AParComm_l_r.
    unfold inclusion, union;intros;try tauto.
    apply A.well_formed_pseudo_appl_r.
    apply C.ACongParL_par_r.
    unfold inclusion, union;intros;try tauto.
    left;left;auto.
    unfold inclusion, union;intros;try tauto.
    apply C.Rcongruence_swapR.
    unfold C.R_preserves_swapR;intros.
    repeat (apply R_preserves_f_union).
    apply C.NewSwap_preserves_swapR;auto.
    apply C.ScopeExtrusion_preserves_swapR;auto.
    apply C.ParMonoid_preserves_swapR;auto.
    apply C.BangAxioms_preserves_swapR;auto.
    apply C.APseudoApplComm_r.
    unfold inclusion, union;intros;try tauto.
    rewrite A.par_r_C in H8.
    discriminate H8.
    subst p1;subst tau0;subst mu;subst p0.
    generalize (Action.mk_interact_kind _ _ _ H12);intro.
    elim (reduction_silent_process H3);auto;intros.
    discriminate H4.
  Qed.

  Theorem Rcongruence_reduction : forall (R:relation A.P.t),(inclusion A.P.t C.NewSwap R)->(inclusion A.P.t C.ScopeExtrusion R)->(inclusion A.P.t R (R_reduction (C.Acongruence R)))->(C.R_preserves_lift R)->(C.R_preserves_swapR R)->(C.R_preserves_low_subst R)->(inclusion A.P.t (C.Rcongruence R) (R_reduction (C.Acongruence R))).
    intros R RNewSwap RScopeExtrusion.
    unfold inclusion;intros.
    induction H3.
    apply H;trivial.
    apply Acongruence_reduction_CongRefl.
    generalize (Acongruence_reduction_CongTrans R);unfold transitive;intros.
    eapply H3;eauto.
    apply Acongruence_reduction_CongInput;trivial.
    apply Acongruence_reduction_CongOutput;trivial.
    apply Acongruence_reduction_CongIfThen;trivial.
    apply Acongruence_reduction_CongNew;trivial.
    apply Acongruence_reduction_CongSumL;trivial.
    apply Acongruence_reduction_CongParL;trivial.
    apply C.Rcongruence_swapR;trivial.
    apply C.Rcongruence_lift;trivial.
    apply C.Rcongruence_low_subst;trivial.
  Qed.

  Theorem StructuralCongruence_simulation : inclusion A.P.t (C.Rcongruence C.StructuralCongruenceAxioms) (R_reduction (C.Acongruence C.StructuralCongruenceAxioms)).
    apply Rcongruence_reduction.
    unfold inclusion.
    apply C.SCNewSwap.
    unfold inclusion.
    apply C.SCScopeExtrusion.
    unfold inclusion.
    induction 1.
    apply (R_reduction_inclusion (C.Acongruence C.SumMonoid)).
    apply C.Acongruence_increase.
    unfold inclusion.
    apply C.SCSumMonoid.
    apply SumMonoid_reduction;trivial.    
    apply (R_reduction_inclusion (C.Acongruence (union A.P.t C.ParMonoid C.ScopeExtrusion))).    
    apply C.Acongruence_increase.
    unfold inclusion, union;intros.
    case H0.
    apply C.SCParMonoid.
    apply C.SCScopeExtrusion.
    apply ParMonoid_reduction;trivial.
    apply (R_reduction_inclusion (C.Acongruence (union A.P.t (union A.P.t C.ParMonoid C.ScopeExtrusion) C.NewSwap))).
    apply C.Acongruence_increase.
    unfold inclusion, union;intros.
    case H0;clear H0;intro H0.
    case H0;clear H0;intro H0.
    apply C.SCParMonoid;auto.
    apply C.SCScopeExtrusion;auto.
    apply C.SCNewSwap;auto.
    apply ScopeExtrusion_reduction;trivial.
    apply (R_reduction_inclusion (C.Acongruence C.NewSwap)).
    apply C.Acongruence_increase.
    unfold inclusion.
    apply C.SCNewSwap.
    apply NewSwap_reduction;trivial.
    apply (R_reduction_inclusion (C.Acongruence (union _ (union _ C.ParMonoid C.ScopeExtrusion) C.NewAxioms))).
    apply C.Acongruence_increase.
    unfold inclusion, union;intros.
    case H0;clear H0;intro H0.
    case H0;clear H0;intro H0.
    apply C.SCParMonoid;auto.
    apply C.SCScopeExtrusion;auto.
    apply C.SCNewAxioms;auto.
    apply NewAxioms_reduction;auto.
    apply (R_reduction_inclusion (C.Acongruence (union _ (union _ (union _ C.NewSwap C.ScopeExtrusion) C.ParMonoid) C.BangAxioms))).
    apply C.Acongruence_increase.
    unfold inclusion, union;intros.
    case H0;clear H0;intro H0.
    case H0;clear H0;intro H0.
    case H0;clear H0;intro H0.
    apply C.SCNewSwap;auto.
    apply C.SCScopeExtrusion;auto.
    apply C.SCParMonoid;auto.
    apply C.SCBangAxioms;auto.
    apply BangAxioms_reduction;auto.
    apply C.StructuralCongruenceAxioms_preserves_lift.
    apply C.StructuralCongruenceAxioms_preserves_swapR.
    apply C.StructuralCongruenceAxioms_preserves_low_subst.
  Qed.

  Inductive input_barb (R:relation Action.A.P.t) : Action.A.P.t -> Action.t -> A.abstraction -> Prop :=
  | InpBarbInput : forall (e:E.t)(p:A.P.t),forall (inp:Action.t),(Action.mk_input e = Some inp)->(input_barb R (A.P.Input e p) inp (A.Abs p))
  | InpBarbPar : forall (p:A.P.t)(inp:Action.t)(f:A.abstraction),(input_barb R p inp f)->forall (r:A.P.t),(input_barb R (A.P.Par p r) inp (A.abs_par_r f r))
  | InpBarbSum : forall (p:A.P.t)(inp:Action.t)(f:A.abstraction),(input_barb R p inp f)->forall (r:A.P.t),(input_barb R (A.P.Sum p r) inp f)
  | InpBarbRes : forall (p:A.P.t)(inp:Action.t)(f:A.abstraction),(input_barb R p inp f)->forall (inp':Action.t),(Action.mk_res inp = Some inp')->(input_barb R (A.P.New p) inp' (A.new_abs f))
  | InpBarbIfThen : forall (p:A.P.t)(inp:Action.t)(f:A.abstraction),(input_barb R p inp f)->forall (phi:F.t)(inp':Action.t),(Action.mk_if phi inp = Some inp')->(input_barb R (A.P.IfThen phi p) inp' f)
  | InpBarbStruct : forall (p:A.P.t)(inp:Action.t)(f:A.abstraction),(input_barb R p inp f)->forall (p':A.P.t)(f':A.abstraction),(C.Rcongruence R p p')->(C.Acongruence R (A.F f) (A.F f'))->(input_barb R p' inp f').

  Lemma input_barb_In : forall (R:relation _)(p:A.P.t)(mu:Action.t)(f:A.abstraction),(input_barb R p mu f)->(Action.get_kind mu = Action.In).
    induction 1.
    apply (Action.mk_input_kind _ _ H).
    auto.
    auto.
    rewrite (Action.mk_res_kind _ _ H0);auto.
    rewrite (Action.mk_if_kind _ _ _ H0);auto.
    auto.
  Qed.

  Lemma input_barb_impl_input_reduction : forall (R:relation A.P.t),(inclusion _ (C.Rcongruence R) (R_reduction (C.Acongruence R)))->(inclusion _ C.NewSwap R)->(inclusion _ C.ScopeExtrusion R)->(C.R_preserves_swapR (C.Rcongruence R))->(symmetric _ (C.Rcongruence R))->forall (p:A.P.t)(inp:Action.t)(f:A.abstraction),(input_barb R p inp f)->exists inp':Action.t,exists f':A.abstraction,(Action.equal inp inp')/\(reduction p inp' (A.F f'))/\(C.Acongruence R (A.F f) (A.F f')).
    intros R Rsim RNewSwap RScopeExtrusion RswapR Rsymmetric.
    induction 1.
    (* InpBarbInput *)
    exists inp.
    exists (A.Abs p).
    split.
    apply Action.equal_refl.
    split.
    apply RedInput;auto.
    apply C.AAbs.
    apply C.CongRefl.
    (* InpBarbPar *)
    elim IHinput_barb.
    intro inp';intros.
    ex_elim H0 f'.
    exists inp'.
    exists (A.abs_par_r f' r).
    split;auto.
    rewrite <- A.par_r_F.
    split.
    apply RedParL.
    auto.
    rewrite <- A.par_r_F.
    apply C.ACongParL_par_r;auto.
    (* InpBarbSum *)
    elim IHinput_barb.
    intro inp';intros.
    ex_elim H0 f'.
    exists inp'.
    exists f'.
    split;auto.
    split;auto.
    apply RedSumL;auto.
    (* InpBarbRes *)
    elim IHinput_barb.
    intro inp0;intros.
    ex_elim H1 f'.
    elim (Action.mk_res_equal _ _ H0 _ H1).
    intro inp0';intros;and_elim.
    exists inp0'.
    exists (A.new_abs f').
    rewrite <- A.new_F.
    split;auto.
    split.
    eapply RedRes;eauto.
    rewrite <- A.new_F.
    apply C.ACongNew.
    auto.
    (* InpBarbIfThen *)
    elim IHinput_barb.
    intro inp0;intros.
    ex_elim H1 f'.
    elim (Action.mk_if_equal _ _ _ H0 _ H1).
    intro inp0';intros;and_elim.
    exists inp0'.
    exists f'.
    split;auto.
    split.
    eapply RedIfThen;eauto.
    auto.
    (* InpBarbStruct *)
    elim IHinput_barb.
    intro inp';intros.
    ex_elim H2 f''.
    elim (Rsim _ _ H0 _ _ H3).
    intro inp'';intros.
    ex_elim H5 b.
    destruct f'' as [q].
    elim (C.Acongruence_F RNewSwap RswapR H7 _ (refl_equal (C.A.F (C.A.Abs q)))).
    intro q';intros;and_elim.
    exists inp''.
    subst b.
    exists (C.A.Abs q').
    split.
    eapply Action.equal_trans;eauto.
    split;auto.
    eapply C.ACongTrans.
    apply C.ACongSym;eauto.
    eapply C.ACongTrans;eauto.
  Qed.

  Lemma StructuralCongruence_input_barb_impl_input_reduction : forall (p:A.P.t)(inp:Action.t)(f:A.abstraction),(input_barb C.StructuralCongruenceAxioms p inp f)->exists inp':Action.t,exists f':A.abstraction,(Action.equal inp inp')/\(reduction p inp' (A.F f'))/\(C.Acongruence C.StructuralCongruenceAxioms (A.F f) (A.F f')).
    intros.
    eapply input_barb_impl_input_reduction;eauto.
    apply StructuralCongruence_simulation.
    unfold inclusion.
    apply C.SCNewSwap.
    unfold inclusion.
    apply C.SCScopeExtrusion.
    apply C.Rcongruence_swapR.
    apply C.StructuralCongruenceAxioms_preserves_swapR.
    apply C.StructuralCongruence_symmetric.
  Qed.

  Lemma input_reduction_impl_input_barb : forall (p:A.P.t)(mu:Action.t)(a:A.t),(reduction p mu a)->forall (f:A.abstraction),a=(A.F f)->(input_barb C.StructuralCongruenceAxioms p mu f).
    induction 1;intros.
    (* RedSilent *)
    discriminate H.
    (* RedInput *)
    injection H0;clear H0;intros;subst f.
    apply InpBarbInput.
    apply H.
    (* RedOutput *)
    discriminate H0.
    (* RedIfThen *)
    eapply InpBarbIfThen.
    apply IHreduction;auto.
    auto.
    (* RedSumL *)
    apply InpBarbSum.
    auto.
    (* RedSumR *)
    eapply InpBarbStruct.
    apply InpBarbSum.
    eauto.
    apply C.CongAxiom.
    apply C.SCSumMonoid.
    apply C.SumComm.
    apply C.ACongRefl.
    rewrite <- H0.
    eapply well_formed_reduction;eauto.
    (* RedRes *)
    destruct a.
    rewrite A.new_Pr in H1;discriminate H1.
    rewrite A.new_F in H1.
    injection H1;clear H1;intros;subst f.
    eapply InpBarbRes;eauto.
    destruct c;rewrite A.new_C in H1;discriminate H1.
    (* RedParL *)
    destruct a.
    rewrite A.par_r_Pr in H0;discriminate H0.
    rewrite A.par_r_F in H0.
    injection H0;clear H0;intros.
    subst f.
    apply InpBarbPar.
    auto.
    rewrite A.par_r_C in H0;discriminate H0.
    (* RedParR *)
    destruct a.
    rewrite A.par_l_Pr in H0;discriminate H0.
    rewrite A.par_l_F in H0.
    injection H0;clear H0;intros;subst f.
    eapply InpBarbStruct.
    apply InpBarbPar.
    apply IHreduction;auto.
    apply C.CongAxiom.
    apply C.SCParMonoid.
    apply C.ParComm.
    rewrite <- A.par_r_F.
    rewrite <- A.par_l_F.
    apply C.AParComm_r_l.
    unfold inclusion.
    apply C.SCParMonoid.
    eapply well_formed_reduction;eauto.
    rewrite A.par_l_C in H0;discriminate H0.
    (* RedCloseL *)
    destruct f;destruct c;rewrite A.pseudo_appl_l_def in H2;discriminate H2.
    (* RedCloseR *)
    destruct f;destruct c;rewrite A.pseudo_appl_r_def in H2;discriminate H2.
    (* RedBangAct *)
    destruct a.
    rewrite A.par_r_Pr in H0;discriminate H0.
    rewrite A.par_r_F in H0;injection H0;clear H0;intros;subst f.
    eapply InpBarbStruct.
    apply InpBarbPar.
    eauto.
    apply C.CongAxiom.
    apply C.SCBangAxioms.
    apply C.BangExpR.
    apply C.ACongRefl.
    rewrite <- A.par_r_F.
    apply A.well_formed_par_r.
    eapply well_formed_reduction;eauto.
    rewrite A.par_r_C in H0.
    discriminate H0.
    (* RedBangCloseL *)
    destruct f;destruct c;rewrite A.pseudo_appl_l_def in H2.
    rewrite A.par_r_Pr in H2;discriminate H2.
  Qed.

  Inductive output_barb (R:relation Action.A.P.t) : Action.A.P.t -> Action.t -> A.concretion -> Prop :=
  | OutBarbOutput : forall (e f:E.t)(p:A.P.t),forall (out:Action.t)(m:E.t),(Action.mk_output e f  = Some (out,m))->(output_barb R (A.P.Output e f p) out (A.Conc 0%nat m p))
  | OutBarbPar : forall (p:A.P.t)(out:Action.t)(c:A.concretion),(output_barb R p out c)->forall (r:A.P.t),(output_barb R (A.P.Par p r) out (A.conc_par_r c r))
  | OutBarbSum : forall (p:A.P.t)(out:Action.t)(c:A.concretion),(output_barb R p out c)->forall (r:A.P.t),(output_barb R (A.P.Sum p r) out c)
  | OutBarbRes : forall (p:A.P.t)(out:Action.t)(c:A.concretion),(output_barb R p out c)->forall (out':Action.t),(Action.mk_res out = Some out')->(output_barb R (A.P.New p) out' (A.new_conc c))
  | OutBarbIfThen : forall (p:A.P.t)(out:Action.t)(c:A.concretion),(output_barb R p out c)->forall (phi:F.t)(out':Action.t),(Action.mk_if phi out = Some out')->(output_barb R (A.P.IfThen phi p) out' c)
  | OutBarbStruct : forall (p:A.P.t)(out:Action.t)(c:A.concretion),(output_barb R p out c)->forall (p':A.P.t)(c':A.concretion),(C.Rcongruence R p p')->(C.Acongruence R (A.C c) (A.C c'))->(output_barb R p' out c').

  Lemma output_barb_Out : forall (R:relation _)(p:A.P.t)(mu:Action.t)(c:A.concretion),(output_barb R p mu c)->(Action.get_kind mu = Action.Out).
    induction 1.
    apply (Action.mk_output_kind _ _ _ _ H).
    auto.
    auto.
    rewrite (Action.mk_res_kind _ _ H0);auto.
    rewrite (Action.mk_if_kind _ _ _ H0);auto.
    auto.
  Qed.

  Lemma output_barb_impl_output_reduction : forall (R:relation A.P.t),(inclusion _ (C.Rcongruence R) (R_reduction (C.Acongruence R)))->(inclusion _ C.NewSwap R)->(inclusion _ C.ScopeExtrusion R)->(C.R_preserves_swapR (C.Rcongruence R))->(symmetric _ (C.Rcongruence R))->forall (p:A.P.t)(out:Action.t)(c:A.concretion),(output_barb R p out c)->exists out':Action.t,exists c':A.concretion,(Action.equal out out')/\(reduction p out' (A.C c'))/\(C.Acongruence R (A.C c) (A.C c')).
    intros R Rsim RNewSwap RScopeExtrusion RswapR Rsymmetric.
    induction 1.
    (* OutBarbOutput *)
    exists out.
    exists (A.Conc 0 m p).
    split.
    apply Action.equal_refl.
    split.
    apply RedOutput;auto.
    apply C.AConc.
    apply C.CongRefl.
    (* OutBarbPar *)
    elim IHoutput_barb.
    intro out';intros.
    ex_elim H0 c'.
    exists out'.
    exists (A.conc_par_r c' r).
    split;auto.
    rewrite <- A.par_r_C.
    split.
    apply RedParL.
    auto.
    rewrite <- A.par_r_C.
    apply C.ACongParL_par_r;auto.
    (* OutBarbSum *)
    elim IHoutput_barb.
    intro out';intros.
    ex_elim H0 c'.
    exists out'.
    exists c'.
    split;auto.
    split;auto.
    apply RedSumL;auto.
    (* OutBarbRes *)
    elim IHoutput_barb.
    intro out0;intros.
    ex_elim H1 c'.
    elim (Action.mk_res_equal _ _ H0 _ H1).
    intro out0';intros;and_elim.
    exists out0'.
    exists (A.new_conc c').
    rewrite <- A.new_C.
    split;auto.
    split.
    eapply RedRes;eauto.
    rewrite <- A.new_C.
    apply C.ACongNew.
    auto.
    (* OutBarbIfThen *)
    elim IHoutput_barb.
    intro out0;intros.
    ex_elim H1 c'.
    elim (Action.mk_if_equal _ _ _ H0 _ H1).
    intro out0';intros;and_elim.
    exists out0'.
    exists c'.
    split;auto.
    split.
    eapply RedIfThen;eauto.
    auto.
    (* OutBarbStruct *)
    elim IHoutput_barb.
    intro out';intros.
    ex_elim H2 c''.
    elim (Rsim _ _ H0 _ _ H3).
    intro out'';intros.
    ex_elim H5 b.
    destruct c'' as [n m q].
    elim (C.Acongruence_C RNewSwap RScopeExtrusion H7 _ _ _ (refl_equal (C.A.C (C.A.Conc n m q)))).
    intro m';intros.
    ex_elim H8 q'.
    exists out''.
    subst b.
    exists (C.A.Conc n m' q').
    split.
    eapply Action.equal_trans;eauto.
    split;auto.
    eapply C.ACongTrans.
    apply C.ACongSym;eauto.
    eapply C.ACongTrans;eauto.
  Qed.

  Lemma StructuralCongruence_output_barb_impl_output_reduction : forall (p:A.P.t)(out:Action.t)(c:A.concretion),(output_barb C.StructuralCongruenceAxioms p out c)->exists out':Action.t,exists c':A.concretion,(Action.equal out out')/\(reduction p out' (A.C c'))/\(C.Acongruence C.StructuralCongruenceAxioms (A.C c) (A.C c')).
    intros.
    eapply output_barb_impl_output_reduction;eauto.
    apply StructuralCongruence_simulation.
    unfold inclusion.
    apply C.SCNewSwap.
    unfold inclusion.
    apply C.SCScopeExtrusion.
    apply C.Rcongruence_swapR.
    apply C.StructuralCongruenceAxioms_preserves_swapR.
    apply C.StructuralCongruence_symmetric.
  Qed.

  Lemma output_reduction_impl_output_barb : forall (p:A.P.t)(mu:Action.t)(a:A.t),(reduction p mu a)->forall (c:A.concretion),a=(A.C c)->(output_barb C.StructuralCongruenceAxioms p mu c).
    induction 1;intros.
    (* RedSilent *)
    discriminate H.
    (* RedInput *)
    discriminate H0.
    (* RedOutput *)
    injection H0;clear H0;intros;subst c.
    apply OutBarbOutput.
    apply H.
    (* RedIfThen *)
    eapply OutBarbIfThen.
    apply IHreduction;auto.
    auto.
    (* RedSumL *)
    apply OutBarbSum.
    auto.
    (* RedSumR *)
    eapply OutBarbStruct.
    apply OutBarbSum.
    eauto.
    apply C.CongAxiom.
    apply C.SCSumMonoid.
    apply C.SumComm.
    apply C.ACongRefl.
    rewrite <- H0.
    eapply well_formed_reduction;eauto.
    (* RedRes *)
    destruct a.
    rewrite A.new_Pr in H1;discriminate H1.
    rewrite A.new_F in H1;discriminate H1.
    rewrite A.new_C in H1.
    injection H1;clear H1;intros;subst c.
    eapply OutBarbRes;eauto.
    (* RedParL *)
    destruct a.
    rewrite A.par_r_Pr in H0;discriminate H0.
    rewrite A.par_r_F in H0;discriminate H0.
    rewrite A.par_r_C in H0.    
    injection H0;clear H0;intros.
    subst c.
    apply OutBarbPar.
    auto.
    (* RedParR *)
    destruct a.
    rewrite A.par_l_Pr in H0;discriminate H0.
    rewrite A.par_l_F in H0;discriminate H0.
    rewrite A.par_l_C in H0.    
    injection H0;clear H0;intros;subst c.
    eapply OutBarbStruct.
    apply OutBarbPar.
    apply IHreduction;auto.
    apply C.CongAxiom.
    apply C.SCParMonoid.
    apply C.ParComm.
    rewrite <- A.par_r_C.
    rewrite <- A.par_l_C.
    apply C.AParComm_r_l.
    unfold inclusion.
    apply C.SCParMonoid.
    eapply well_formed_reduction;eauto.
    (* RedCloseL *)
    destruct f;destruct c;rewrite A.pseudo_appl_l_def in H2;discriminate H2.
    (* RedCloseR *)
    destruct f;destruct c;rewrite A.pseudo_appl_r_def in H2;discriminate H2.
    (* RedBangAct *)
    destruct a.
    rewrite A.par_r_Pr in H0;discriminate H0.
    rewrite A.par_r_F in H0;discriminate H0.
    rewrite A.par_r_C in H0.
    injection H0;clear H0;intros;subst c.
    eapply OutBarbStruct.
    apply OutBarbPar.
    eauto.
    apply C.CongAxiom.
    apply C.SCBangAxioms.
    apply C.BangExpR.
    apply C.ACongRefl.
    rewrite <- A.par_r_C.
    apply A.well_formed_par_r.
    eapply well_formed_reduction;eauto.
    (* RedBangCloseL *)
    destruct f;destruct c;rewrite A.pseudo_appl_l_def in H2.
    rewrite A.par_r_Pr in H2;discriminate H2.
  Qed.

(*   The following definition is not correct because of guards. *)
(*   Inductive reaction (R:relation A.P.t) : A.P.t -> Action.t -> A.P.t -> Prop := *)
(*   | ReactComm : forall (e:E.t)(p1 p2:A.P.t)(inp:Action.t),(Action.mk_input e = Some inp)->forall (f g:E.t)(q1 q2:A.P.t)(out:Action.t)(m:E.t),(Action.mk_output f g)=(Some (out,m))->forall (tau:Action.t),(Action.mk_interact inp out = Some tau)->(reaction R (A.P.Par (A.P.Sum (A.P.Input e p1) p2) (A.P.Sum (A.P.Output f g q1) q2)) tau (A.P.Par (deBruijn.low_subst A.P.deBruijn_subst 0 0 (cons m nil) p1) q1)) *)
(*   | ReactSilent : forall (p:A.P.t),(reaction R (A.P.Tau p) Action.mk_silent p) *)
(*   | ReactPar : forall (p:A.P.t)(tau:Action.t)(q:A.P.t),(reaction R p tau q)->forall (r:A.P.t),(reaction R (A.P.Par p r) tau (A.P.Par q r)) *)
(*   | ReactSum : forall (p:A.P.t)(tau:Action.t)(q:A.P.t),(reaction R p tau q)->forall (r:A.P.t),(reaction R (A.P.Sum p r) tau q) *)
(*   | ReactRes : forall (p:A.P.t)(tau:Action.t)(q:A.P.t),(reaction R p tau q)->forall (tau':Action.t),(Action.mk_res tau = Some tau')->(reaction R (A.P.New p) tau' (A.P.New q)) *)
(*   | ReactIfThen : forall (p:A.P.t)(tau:Action.t)(q:A.P.t),(reaction R p tau q)->forall (phi:F.t)(tau':Action.t),(Action.mk_if phi tau = Some tau')->(reaction R (A.P.IfThen phi p) tau' q) *)
(*   | ReactStruct : forall (p:A.P.t)(tau:Action.t)(q:A.P.t),(reaction R p tau q)->forall (p':A.P.t)(q':A.P.t),(C.Rcongruence R p p')->(C.Rcongruence R q q')->(reaction R p' tau q'). *)

(*   Lemma reaction_silent : forall (R:relation _)(p:A.P.t)(mu:Action.t)(q:A.P.t),(reaction R p mu q)->(Action.get_kind mu = Action.Silent). *)
(*     induction 1. *)
(*     apply (Action.mk_interact_kind _ _ _ H1). *)
(*     apply Action.mk_silent_kind. *)
(*     auto. *)
(*     auto. *)
(*     rewrite (Action.mk_res_kind _ _ H0);auto. *)
(*     rewrite (Action.mk_if_kind _ _ _ H0);auto. *)
(*     auto. *)
(*   Qed. *)

(*   Lemma reaction_impl_internal_reduction : forall (R:relation A.P.t),(inclusion _ (C.Rcongruence R) (R_reduction (C.Acongruence R)))->(inclusion _ C.NewSwap R)->(symmetric _ (C.Rcongruence R))->forall (p:A.P.t)(tau:Action.t)(q:A.P.t),(reaction R p tau q)->exists tau':Action.t,exists q':A.P.t,(Action.equal tau tau')/\(internal_reduction p tau' q')/\(C.Rcongruence R q q'). *)
(*     unfold internal_reduction. *)
(*     intros R Rsim RNewSwap Rsymmetric. *)
(*     induction 1. *)
(*     (* ReactComm *) *)
(*     exists tau. *)
(*     exists (A.P.Par (deBruijn.low_subst A.P.deBruijn_subst 0 0 (m :: nil) p1) q1). *)
(*     split. *)
(*     apply Action.equal_refl. *)
(*     split. *)
(*     assert (reduction (A.P.Sum (A.P.Input e p1) p2) inp (A.F (A.Abs p1))). *)
(*     apply RedSumL. *)
(*     apply RedInput;auto. *)
(*     assert (reduction (A.P.Sum (A.P.Output f g q1) q2) out (A.C (A.Conc 0 m q1))). *)
(*     apply RedSumL. *)
(*     apply RedOutput;auto. *)
(*     generalize (RedCloseL _ _ _ H2 _ _ _ H3 _ H1). *)
(*     rewrite A.pseudo_appl_l_def. *)
(*     rewrite A.P.mk_nu_0. *)
(*     auto. *)
(*     apply C.CongRefl. *)
(*     (* ReactSilent *) *)
(*     exists Action.mk_silent. *)
(*     exists p. *)
(*     split. *)
(*     apply Action.equal_refl. *)
(*     split. *)
(*     apply RedSilent. *)
(*     apply C.CongRefl. *)
(*     (* ReactPar *) *)
(*     elim IHreaction. *)
(*     intro tau';intros. *)
(*     ex_elim H0 q'. *)
(*     exists tau'. *)
(*     exists (A.P.Par q' r). *)
(*     split;auto. *)
(*     split. *)
(*     rewrite <- A.par_r_Pr. *)
(*     apply RedParL. *)
(*     auto. *)
(*     apply C.CongParL. *)
(*     auto. *)
(*     (* ReactPar *) *)
(*     elim IHreaction. *)
(*     intro tau';intros. *)
(*     ex_elim H0 q'. *)
(*     exists tau'. *)
(*     exists q'. *)
(*     split;auto. *)
(*     split. *)
(*     apply RedSumL. *)
(*     auto. *)
(*     auto. *)
(*     (* ReactRes *) *)
(*     elim IHreaction. *)
(*     intro tau0;intros. *)
(*     ex_elim H1 q'. *)
(*     elim (Action.mk_res_equal _ _ H0 _ H1). *)
(*     intro tau0';intros;and_elim. *)
(*     exists tau0'. *)
(*     exists (A.P.New q'). *)
(*     rewrite <- A.new_Pr. *)
(*     split;auto. *)
(*     split. *)
(*     eapply RedRes;eauto. *)
(*     apply C.CongNew. *)
(*     auto. *)
(*     (* ReactIfThen *) *)
(*     elim IHreaction. *)
(*     intro tau0;intros. *)
(*     ex_elim H1 q'. *)
(*     elim (Action.mk_if_equal _ _ _ H0 _ H1). *)
(*     intro tau0';intros;and_elim. *)
(*     exists tau0'. *)
(*     exists q'. *)
(*     split;auto. *)
(*     split. *)
(*     eapply RedIfThen;eauto. *)
(*     auto. *)
(*     (* ReactStruct *) *)
(*     elim IHreaction. *)
(*     intro tau';intros. *)
(*     ex_elim H2 q''. *)
(*     elim (Rsim _ _ H0 _ _ H3). *)
(*     intro tau'';intros. *)
(*     ex_elim H5 b. *)
(*     elim (C.Acongruence_Pr RNewSwap H7 _ (refl_equal (A.Pr q''))). *)
(*     intro q3;intros;and_elim. *)
(*     subst b. *)
(*     exists tau''. *)
(*     exists q3. *)
(*     split. *)
(*     eapply Action.equal_trans;eauto. *)
(*     split;auto. *)
(*     eapply C.CongTrans. *)
(*     apply Rsymmetric;eauto. *)
(*     eapply C.CongTrans;eauto. *)
(*   Qed. *)

(*   Lemma StructuralCongruence_reaction_impl_internal_reduction : forall (p:A.P.t)(tau:Action.t)(q:A.P.t),(reaction C.StructuralCongruenceAxioms p tau q)->exists tau':Action.t,exists q':A.P.t,(Action.equal tau tau')/\(internal_reduction p tau' q')/\(C.Rcongruence C.StructuralCongruenceAxioms q q'). *)
(*     intros. *)
(*     apply reaction_impl_internal_reduction. *)
(*     apply StructuralCongruence_simulation. *)
(*     unfold inclusion. *)
(*     apply C.SCNewSwap. *)
(*     apply C.StructuralCongruence_symmetric. *)
(*     auto. *)
(*   Qed. *)
End RedCong.





