(* A formalisation of the spi calculus in Coq *)
(* Copyright (C) 2005-2007 Sebastien Briais *)
(* http://lamp.epfl.ch/~sbriais/ *)

(* This library is free software; you can redistribute it and/or modify *)
(* it under the terms of the GNU Lesser General Public License as *)
(* published by the Free Software Foundation; either version 2.1 of the *)
(* License, or (at your option) any later version. *)

(* This library is distributed in the hope that it will be useful, but *)
(* WITHOUT ANY WARRANTY; without even the implied warranty of *)
(* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU *)
(* Lesser General Public License for more details. *)

(* You should have received a copy of the GNU Lesser General Public *)
(* License along with this library; if not, write to the Free Software *)
(* Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA *)
(* 02110-1301 USA *)

Require Import expressionType.
Require Import formulaeType.
Require Import processType.
Require Import agent.
Require Import actionType.
Require Import tactics.
Require Import deBruijnType.
Require Import Arith.
Require Import missing.

Module Reduction(Action:ActionTypeLight).
  Import Action.
  Module E := A.P.E.
  Module F := A.P.F.

  Inductive reduction : A.P.t -> Action.t -> A.t -> Prop :=
    | RedSilent : 
      forall (p:A.P.t),
	(reduction (A.P.Tau p) Action.mk_silent (A.Pr p))
    | RedInput : 
      forall (p:A.P.t)(e:E.t)(mu:Action.t),
	(Action.mk_input e = Some mu)->
	(reduction (A.P.Input e p) mu (A.F (A.Abs p)))
    | RedOutput : 
      forall (p:A.P.t)(e f:E.t)(mu:Action.t)(m:E.t),
	(Action.mk_output e f = Some (mu,m))->
	(reduction (A.P.Output e f p) mu (A.C (A.Conc 0 m p)))
    | RedIfThen : 
      forall (p:A.P.t)(mu:Action.t)(a:A.t),
	(reduction p mu a)->
	forall (phi:F.t)(mu':Action.t),(Action.mk_if phi mu = Some mu')->
	  (reduction (A.P.IfThen phi p) mu' a)
    | RedSumL : 
      forall (p:A.P.t)(mu:Action.t)(a:A.t),
	(reduction p mu a)->
	forall (q:A.P.t),(reduction (A.P.Sum p q) mu a)
    | RedSumR : 
      forall (p:A.P.t)(mu:Action.t)(a:A.t),
	(reduction p mu a)->
	forall (q:A.P.t),(reduction (A.P.Sum q p) mu a)
    | RedRes :
      forall (p:A.P.t)(mu:Action.t)(a:A.t),
	(reduction p mu a)->
	forall (mu':Action.t),(Action.mk_res mu = Some mu')->
	  (reduction (A.P.New p) mu' (A.new a))
    | RedParL : 
      forall (p:A.P.t)(mu:Action.t)(a:A.t),
	(reduction p mu a)->
	forall (q:A.P.t),(reduction (A.P.Par p q) mu (A.par_r a q))
    | RedParR : 
      forall (q:A.P.t)(mu:Action.t)(a:A.t),
	(reduction q mu a)->
	forall (p:A.P.t),(reduction (A.P.Par p q) mu (A.par_l p a))
    | RedCloseL : 
      forall (p:A.P.t)(inp:Action.t)(f:A.abstraction),
	(reduction p inp (A.F f))->
	forall (q:A.P.t)(out:Action.t)(c:A.concretion),
	  (reduction q out (A.C c))->
	  forall (tau:Action.t),
	    ((Action.mk_interact inp out) = Some(tau))->
	    (reduction (A.P.Par p q) tau (A.pseudo_appl_l f c))
    | RedCloseR : 
      forall (p:A.P.t)(out:Action.t)(c:A.concretion),
	(reduction p out (A.C c))->
	forall (q:A.P.t)(inp:Action.t)(f:A.abstraction),
	  (reduction q inp (A.F f))->
	  forall (tau:Action.t),
	    ((Action.mk_interact out inp) = Some(tau))->
	    (reduction (A.P.Par p q) tau (A.pseudo_appl_r c f))
    | RedBangAct : 
      forall (p:A.P.t)(mu:Action.t)(a:A.t),
	(reduction p mu a)->
	(reduction (A.P.Bang p) mu (A.par_r a (A.P.Bang p)))
    | RedBangCloseL : 
      forall (p:A.P.t)(inp:Action.t)(f:A.abstraction),
	(reduction p inp (A.F f))->
	forall (out:Action.t)(c:A.concretion),
	  (reduction p out (A.C c))->
	  forall (tau:Action.t),
	    ((Action.mk_interact inp out) = Some(tau))->
	    (reduction (A.P.Bang p) tau (A.par_r (A.pseudo_appl_l f c) (A.P.Bang p))).

  Definition internal_reduction (p:A.P.t)(tau:Action.t)(q:A.P.t) := reduction p tau (Action.A.Pr q).

  (* well formed residues *)
  Lemma well_formed_reduction : forall (p:A.P.t)(mu:Action.t)(a:A.t),(reduction p mu a)->(A.well_formed a).
    induction 1;auto.
    apply A.WFProc.
    apply A.WFAbs.
    apply A.WFConc.
    intros.
    inversion H0.
    apply A.well_formed_new;auto.
    apply A.well_formed_par_r;auto.
    apply A.well_formed_par_l;auto.
    apply A.well_formed_pseudo_appl_l;auto.    
    apply A.well_formed_pseudo_appl_r;auto.    
    apply A.well_formed_par_r;auto.
    apply A.well_formed_par_r;auto.
    apply A.well_formed_pseudo_appl_l;auto.    
  Qed.

  Implicit Arguments well_formed_reduction [p mu a].

  (* abstraction <-> input *)
  (* concretion <-> output *)
  (* process <-> silent *)
  Lemma reduction_abstraction_input : forall (p:A.P.t)(mu:Action.t)(a:A.t),(reduction p mu a)->forall (f:A.abstraction),(a = A.F f)->(Action.get_kind mu = Action.In).
    induction 1;intros;try (discriminate H || discriminate H0).
    eapply Action.mk_input_kind;eauto.
    rewrite (Action.mk_if_kind phi mu);eauto.
    eauto.
    eauto.
    rewrite (Action.mk_res_kind mu);eauto.
    destruct a.
    rewrite A.new_Pr in H1;discriminate H1.
    eauto.
    rewrite A.new_C in H1;discriminate H1.
    destruct a.
    rewrite A.par_r_Pr in H0;discriminate H0.
    eauto.
    rewrite A.par_r_C in H0;discriminate H0.
    destruct a.
    rewrite A.par_l_Pr in H0;discriminate H0.
    eauto.
    rewrite A.par_l_C in H0;discriminate H0.
    destruct f;destruct c.
    rewrite A.pseudo_appl_l_def in H2;discriminate H2.
    destruct f;destruct c.
    rewrite A.pseudo_appl_r_def in H2;discriminate H2.
    destruct a.
    rewrite A.par_r_Pr in H0;discriminate H0.
    eauto.
    rewrite A.par_r_C in H0;discriminate H0.
    destruct f;destruct c.
    rewrite A.pseudo_appl_l_def in H2.
    rewrite A.par_r_Pr in H2;discriminate H2.
  Qed.

  Lemma reduction_concretion_output : forall (p:A.P.t)(mu:Action.t)(a:A.t),(reduction p mu a)->forall (c:A.concretion),(a = A.C c)->(Action.get_kind mu = Action.Out).
    induction 1;intros;try (discriminate H || discriminate H0).
    eapply Action.mk_output_kind;eauto.
    rewrite (Action.mk_if_kind phi mu);eauto.
    eauto.
    eauto.
    rewrite (Action.mk_res_kind mu);eauto.
    destruct a.
    rewrite A.new_Pr in H1;discriminate H1.
    rewrite A.new_F in H1;discriminate H1.
    eauto.
    destruct a.
    rewrite A.par_r_Pr in H0;discriminate H0.
    rewrite A.par_r_F in H0;discriminate H0.
    eauto.
    destruct a.
    rewrite A.par_l_Pr in H0;discriminate H0.
    rewrite A.par_l_F in H0;discriminate H0.
    eauto.
    destruct f;destruct c.
    rewrite A.pseudo_appl_l_def in H2;discriminate H2.
    destruct f;destruct c.
    rewrite A.pseudo_appl_r_def in H2;discriminate H2.
    destruct a.
    rewrite A.par_r_Pr in H0;discriminate H0.
    rewrite A.par_r_F in H0;discriminate H0.
    eauto.
    destruct f;destruct c.
    rewrite A.pseudo_appl_l_def in H2.
    rewrite A.par_r_Pr in H2;discriminate H2.
  Qed.

  Lemma reduction_process_silent : forall (p:A.P.t)(mu:Action.t)(a:A.t),(reduction p mu a)->forall (q:A.P.t),(a = A.Pr q)->(Action.get_kind mu = Action.Silent).
    induction 1;intros;try (discriminate H || discriminate H0).
    apply Action.mk_silent_kind.
    rewrite (Action.mk_if_kind phi mu);eauto.
    eauto.
    eauto.
    rewrite (Action.mk_res_kind mu);eauto.
    destruct a.
    eauto.
    rewrite A.new_F in H1;discriminate H1.
    rewrite A.new_C in H1;discriminate H1.
    destruct a.
    eauto.
    rewrite A.par_r_F in H0;discriminate H0.
    rewrite A.par_r_C in H0;discriminate H0.
    destruct a.
    eauto.
    rewrite A.par_l_F in H0;discriminate H0.
    rewrite A.par_l_C in H0;discriminate H0.
    eapply Action.mk_interact_kind;eauto.
    eapply Action.mk_interact_kind;eauto.
    destruct a.
    eauto.
    rewrite A.par_r_F in H0;discriminate H0.
    rewrite A.par_r_C in H0;discriminate H0.
    eapply Action.mk_interact_kind;eauto.
  Qed.

  Lemma reduction_input_abstraction : forall (p:A.P.t)(mu:Action.t)(a:A.t),(reduction p mu a)->(Action.get_kind mu = Action.In)->exists f:A.abstraction,a = A.F f.
    destruct a;intros.
    rewrite (reduction_process_silent p mu (A.Pr t0) H t0) in H0;auto;discriminate H0.
    exists a;auto.
    rewrite (reduction_concretion_output p mu (A.C c) H c) in H0;auto;discriminate H0.
  Qed.
    
  Lemma reduction_output_concretion : forall (p:A.P.t)(mu:Action.t)(a:A.t),(reduction p mu a)->(Action.get_kind mu = Action.Out)->exists c:A.concretion,a = A.C c.
    destruct a;intros.
    rewrite (reduction_process_silent p mu (A.Pr t0) H t0) in H0;auto;discriminate H0.
    rewrite (reduction_abstraction_input p mu (A.F a) H a) in H0;auto;discriminate H0.
    exists c;auto.
  Qed.

  Lemma reduction_silent_process : forall (p:A.P.t)(mu:Action.t)(a:A.t),(reduction p mu a)->(Action.get_kind mu = Action.Silent)->exists q:A.P.t,a = A.Pr q.
    destruct a;intros.
    exists t0;auto.
    rewrite (reduction_abstraction_input p mu (A.F a) H a) in H0;auto;discriminate H0.
    rewrite (reduction_concretion_output p mu (A.C c) H c) in H0;auto;discriminate H0.
  Qed.

  Implicit Arguments reduction_abstraction_input [p mu a].
  Implicit Arguments reduction_concretion_output [p mu a].
  Implicit Arguments reduction_process_silent [p mu a].
  Implicit Arguments reduction_input_abstraction [p mu a].
  Implicit Arguments reduction_output_concretion [p mu a].
  Implicit Arguments reduction_silent_process [p mu a].

  Lemma internal_reduction_silent : forall (p:A.P.t)(tau:Action.t)(q:A.P.t),(internal_reduction p tau q)->(Action.get_kind tau = Action.Silent).
    intros.
    unfold internal_reduction in H.
    eapply reduction_process_silent;eauto.
  Qed.

  (* computation of reduction relation *)
  Require Import missing.
  Require Import ordmissing. 

  Module Transition := OrderedPair Action.Ord A.Ord. 

  Module Type TransitionSet := FSetInterface.S with Module E := Transition.

  Module Make(T:TransitionSet).
    Module TOtherProps := MissingSetProperties T.

    Definition add_ParL (q:A.P.t)(ap : T.elt) := let (a,p) := ap in Some (a,A.par_l q p).
    Definition add_ParR (q:A.P.t)(ap : T.elt) := let (a,p) := ap in Some (a,A.par_r p q).

    Lemma add_ParL_compat : forall (q:A.P.t),TOtherProps.compat_f (add_ParL q).
      intros.
      unfold TOtherProps.compat_f.
      intros.
      destruct x.
      destruct x'.
      simpl in H.
      and_elim.
      unfold add_ParL.
      unfold TOtherProps.option_eq.
      simpl.
      split;trivial.
      rewrite H0;auto.
    Qed.

    Lemma add_ParR_compat : forall (q:A.P.t),TOtherProps.compat_f (add_ParR q).
      intros.
      unfold TOtherProps.compat_f.
      intros.
      destruct x.
      destruct x'.
      simpl in H.
      and_elim.
      unfold add_ParR.
      unfold TOtherProps.option_eq.
      simpl.
      split;trivial.
      rewrite H0;auto.
    Qed.

    Definition react (symm:bool)(ap:T.elt)(aq:T.elt) : option T.elt :=
      match (ap,aq) with
	| ((inp,A.F f),(out,A.C c)) =>
	  (match Action.mk_interact inp out with
	     | None => None
	     | Some(tau) => Some(tau,A.pseudo_appl_l f c)
	   end)
	| ((out,A.C c),(inp,A.F f)) =>
	  if symm 
	    then
	      (match Action.mk_interact out inp with
		 | None => None
		 | Some(tau) => Some(tau,A.pseudo_appl_r c f)
	       end)
	    else None
	| _ => None
      end.
	  
    Lemma react_compat: forall (b:bool),TOtherProps.compat_f2 (react b).
      unfold TOtherProps.compat_f2.
      intros.
      destruct x;destruct x'.
      simpl in H;and_elim.
      destruct y;destruct y'.
      simpl in H0;and_elim.
      rewrite H;rewrite H1;rewrite H0;rewrite H2.
      apply TOtherProps.option_eq_refl.
    Qed.

    Definition add_if (phi:F.t)(ap:T.elt) := 
      let (alpha,p) := ap in
	match Action.mk_if phi alpha with
	  | Some(beta) => Some(beta,p)
	  | None => None
	end.

    Lemma add_if_compat : forall (phi:F.t),TOtherProps.compat_f (add_if phi).
      intros.
      unfold TOtherProps.compat_f.
      intros.
      destruct x;destruct x';simpl.
      simpl in H.
      and_elim.
      rewrite H;rewrite H0.
      destruct (mk_if phi t2);simpl;intros;unfold TOtherProps.option_eq;simpl;auto.
    Qed.

    Definition add_open_res (ap:T.elt) :=
      let (alpha,p) := ap in
	match (Action.mk_res alpha) with
	  | Some beta => Some(beta,(A.new p))
	  | None => None
	end.

    Lemma add_open_res_compat : TOtherProps.compat_f add_open_res.
      intros.
      unfold TOtherProps.compat_f.
      destruct x;destruct x'.
      simpl;intros;and_elim.
      rewrite H;rewrite H0.
      apply TOtherProps.option_eq_refl.
    Qed.

    Fixpoint late_transitions (p:A.P.t) {struct p} : T.t :=
      match p with
	| A.P.Zero => T.empty
	| A.P.Par p q => 
	  let tps := late_transitions p in
	    let tqs := late_transitions q in
	      (T.union (T.union (TOtherProps.map (add_ParR q) tps) (TOtherProps.map (add_ParL p) tqs)) (TOtherProps.map2 (react true) tps tqs))
	| A.P.Sum p q => T.union (late_transitions p) (late_transitions q)
	| A.P.Bang p => 
	  let tps := late_transitions p in
            (T.union (TOtherProps.map (add_ParR (A.P.Bang p)) tps) (TOtherProps.map (add_ParR (A.P.Bang p)) (TOtherProps.map2 (react false) tps tps)))
	| A.P.New p => TOtherProps.map add_open_res (late_transitions p)
	| A.P.Tau p => T.singleton (Action.mk_silent,A.Pr p)
	| A.P.Input a p => 
	  (match Action.mk_input a with
	     | Some(alpha) => T.singleton (alpha,A.F (A.Abs p))
	     | None => T.empty
	   end)
	| A.P.Output a m p => 
	  (match Action.mk_output a m with
	     | Some(alpha,n) => T.singleton (alpha,A.C (A.Conc 0 n p))
	     | None => T.empty
	   end)
	| A.P.IfThen phi p => 
	  TOtherProps.map (add_if phi) (late_transitions p)
      end.

    Lemma late_transitions_sound : forall (p:A.P.t)(mu:Action.t)(a:A.t),(T.In (mu,a) (late_transitions p))->(reduction p mu a).
      induction p;simpl;intros.
      (* Zero *)
      elim (T.empty_1 H).
      (* Par p q *)
      case (T.union_1 H);intro.
      case (T.union_1 H0);intro.
      elim (TOtherProps.In_map_impl (add_ParR p2) (add_ParR_compat p2) (late_transitions p1) (mu,a));intros;auto.
      destruct x;and_elim.
      unfold TOtherProps.option_eq in H3;simpl in H3;and_elim.
      rewrite <- H4.
      apply RedParL.
      rewrite H3 in H2.
      auto.
      elim (TOtherProps.In_map_impl (add_ParL p1) (add_ParL_compat p1) (late_transitions p2) (mu,a));intros;auto.
      destruct x;and_elim.
      unfold TOtherProps.option_eq in H3;simpl in H3;and_elim.
      rewrite <- H4.
      rewrite H3 in H2.
      apply RedParR.
      auto.
      elim (TOtherProps.In_map2_impl (react true) (react_compat true) (late_transitions p1) (late_transitions p2) (mu,a));intros;auto.
      destruct x;intros;and_elim.
      ex_elim H1 y;destruct y.
      generalize H3;clear H3.
      unfold TOtherProps.option_eq.
      destruct t1;destruct t3;simpl;try tauto;unfold react.
      elim (Option_elim_option Action.t (Action.mk_interact t0 t2));intro.
      elim a1;intro tau;intro.
      rewrite p.
      simpl;intros;and_elim.
      rewrite <- H4.
      eapply RedCloseL;eauto.
      rewrite H3 in p;auto.
      rewrite b;tauto.
      elim (Option_elim_option Action.t (Action.mk_interact t0 t2));intro.
      elim a1;intro tau;intro.
      rewrite p.
      simpl;intros;and_elim.
      rewrite <- H4.
      eapply RedCloseR;eauto.
      rewrite H3 in p;auto.
      rewrite b;tauto.
      (* Sum p q *)
      case (T.union_1 H);intro.
      apply RedSumL;auto.
      apply RedSumR;auto.
      (* Bang p *)
      case (T.union_1 H);intro.
      elim (TOtherProps.In_map_impl (add_ParR (A.P.Bang p)) (add_ParR_compat (A.P.Bang p)) (late_transitions p) (mu,a));intros;auto.
      destruct x;and_elim.
      unfold TOtherProps.option_eq in H2;simpl in H2.
      and_elim.
      rewrite <- H2.
      rewrite <- H3.
      apply RedBangAct;auto.
      elim (TOtherProps.In_map_impl (add_ParR (A.P.Bang p)) (add_ParR_compat (A.P.Bang p)) (TOtherProps.map2 (react false) (late_transitions p) (late_transitions p)) (mu,a));intros;auto.
      destruct x;and_elim.
      elim (TOtherProps.In_map2_impl (react false) (react_compat false) (late_transitions p) (late_transitions p) (t0,t1));intros;auto.
      ex_elim H3 y.
      destruct x.
      destruct y.
      generalize H5;clear H5.
      unfold TOtherProps.option_eq.
      destruct t3;destruct t5;simpl;try tauto;unfold react.
      elim (Option_elim_option Action.t (Action.mk_interact t2 t4));intro.
      elim a1;intro tau;intro.
      rewrite p0.
      simpl;intros;and_elim.
      rewrite <- H6 in H2.
      simpl in H2.
      unfold TOtherProps.option_eq in H2;simpl in H2;and_elim.
      rewrite <- H7.
      rewrite H2 in H5.
      rewrite H5 in p0.
      eapply RedBangCloseL;eauto.
      rewrite b;tauto.
      (* New p *)
      elim (TOtherProps.In_map_impl add_open_res add_open_res_compat (late_transitions p) (mu,a));intros;auto.
      and_elim.
      destruct x.
      generalize H1;clear H1.
      unfold TOtherProps.option_eq. 
      unfold add_open_res.
      elim (Option_elim_option Action.t (Action.mk_res t0));intro.
      elim a0;intro;intro.
      rewrite p0.
      simpl;intros;and_elim.
      rewrite <- H2.
      rewrite H1 in p0.
      eapply RedRes;eauto.
      rewrite b;tauto.
      (* Tau p *)
      generalize (T.singleton_1 H).
      simpl;intros;and_elim.
      rewrite <- H1.
      rewrite <- H0.
      apply RedSilent;auto.
      (* Input a p *)
      elim (Option_elim_option Action.t (Action.mk_input t0));intro.
      elim a0;intros.
      rewrite p0 in H.
      generalize (T.singleton_1 H).
      simpl;intros;and_elim.
      rewrite <- H1.
      rewrite H0 in p0.
      eapply RedInput;eauto.
      rewrite b in H.
      elim (T.empty_1 H).
      (* Output a m p *)
      elim (Option_elim_option (Action.t * A.P.E.t) (Action.mk_output t0 t1));intro.
      elim a0;intros.
      destruct x.
      rewrite p0 in H.
      generalize (T.singleton_1 H).
      simpl;intros;and_elim.
      rewrite <- H1.
      rewrite H0 in p0.
      eapply RedOutput;eauto.
      rewrite b in H.
      elim (T.empty_1 H).
      (* IfThen phi p *)
      elim (TOtherProps.In_map_impl (add_if t0) (add_if_compat t0) (late_transitions p) (mu,a));intros;auto.
      and_elim.
      destruct x.
      generalize H1;clear H1.
      unfold TOtherProps.option_eq.
      simpl.
      elim (Option_elim_option Action.t (Action.mk_if t0 t1));intro.
      elim a0;intros.
      rewrite p0 in H1.
      simpl in H1;and_elim.
      rewrite <- H2.
      eapply RedIfThen;eauto.
      rewrite <- H1;auto.
      rewrite b;tauto.
    Qed.

    Lemma late_transitions_complete : forall (p:A.P.t)(mu:Action.t)(a:A.t),(reduction p mu a)->(T.In (mu,a) (late_transitions p)).
      induction p;intros;inversion H;simpl.
      apply T.union_2.
      apply T.union_2.
      eapply TOtherProps.impl_In_map with (mu,a0).
      apply add_ParR_compat.
      auto.
      apply TOtherProps.option_eq_refl.
      apply T.union_2.
      apply T.union_3.
      eapply TOtherProps.impl_In_map with (mu,a0);auto.
      apply add_ParL_compat.
      apply TOtherProps.option_eq_refl.
      apply T.union_3.
      eapply TOtherProps.impl_In_map2 with (inp,(A.F f)) (out,(A.C c));auto.
      apply react_compat.
      simpl.
      unfold react.
      rewrite H6.
      apply TOtherProps.option_eq_refl.
      apply T.union_3.
      eapply TOtherProps.impl_In_map2 with (out,(A.C c)) (inp,(A.F f));auto.
      apply react_compat.
      simpl.
      unfold react.
      rewrite H6.
      apply TOtherProps.option_eq_refl.
      apply T.union_2;auto.
      apply T.union_3;auto.
      apply T.union_2.
      eapply TOtherProps.impl_In_map with (mu,a0);auto.
      apply add_ParR_compat.
      apply TOtherProps.option_eq_refl.
      apply T.union_3.
      apply TOtherProps.impl_In_map with (mu,(A.pseudo_appl_l f c)).
      apply add_ParR_compat.
      apply TOtherProps.impl_In_map2 with (inp,(A.F f)) (out,(A.C c));auto.
      apply react_compat.
      unfold react.
      rewrite H3.
      apply TOtherProps.option_eq_refl.
      simpl.
      apply TOtherProps.option_eq_refl.
      apply TOtherProps.impl_In_map with (mu0,a0);auto.
      apply add_open_res_compat.
      simpl.
      rewrite H2.
      apply TOtherProps.option_eq_refl.
      apply T.singleton_2;auto.
      rewrite H4.
      apply T.singleton_2;auto.
      rewrite H5.
      apply T.singleton_2;auto.
      apply TOtherProps.impl_In_map with (mu0,a);auto.
      apply add_if_compat.
      simpl.
      rewrite H5.
      apply TOtherProps.option_eq_refl.
    Qed.
  End Make.

  Lemma reduction_lift : forall (p:A.P.t)(mu:Action.t)(a:A.t),(reduction p mu a)->forall (d k:nat),(exists mu':Action.t,(Action.equal mu' (deBruijn.lift Action.deBruijn_base d k mu))/\(reduction (deBruijn.lift A.P.deBruijn_base d k p) mu' (deBruijn.lift A.deBruijn_base d k a))).
    induction 1;intros.
    (* RedSilent *)
    rewrite A.P.lift_Tau.
    exists Action.mk_silent.
    split.
    apply Action.lift_mk_silent.
    rewrite A.lift_Pr.
    apply RedSilent.
    (* RedInput *)
    rewrite A.P.lift_Input.
    elim (Action.lift_mk_input e mu H d k).
    intro mu';intros;and_elim.
    exists mu'.
    rewrite A.lift_F.
    split;auto.
    apply RedInput;auto.
    (* RedOutput *)
    rewrite A.P.lift_Output.
    elim (Action.lift_mk_output e f mu m H d k).
    intro mu';intros;and_elim.
    exists mu'.
    split;auto.
    rewrite A.lift_C.
    rewrite plus_0_r.
    apply RedOutput;auto.
    (* RedIfThen *)
    rewrite A.P.lift_IfThen.
    elim (IHreduction d k).
    intro mu'';intros;and_elim.
    elim (Action.lift_mk_if phi mu mu' H0 d k).
    intro mu3;intros;and_elim.
    elim (Action.mk_if_equal (deBruijn.lift A.P.F.deBruijn_base d k phi) (deBruijn.lift Action.deBruijn_base d k mu) mu3 H4 mu'').
    intro mu3';intros;and_elim.
    exists mu3'.
    split.
    eapply Action.equal_trans.
    apply Action.equal_sym.
    apply H5.
    auto.
    eapply RedIfThen;eauto.
    apply Action.equal_sym;auto.
    (* RedSumL *)
    rewrite A.P.lift_Sum.
    elim (IHreduction d k);intro mu';intros;and_elim.
    exists mu'.
    split;auto.
    apply RedSumL;auto.
    (* RedSumR *)
    rewrite A.P.lift_Sum.
    elim (IHreduction d k);intro mu';intros;and_elim.
    exists mu'.
    split;auto.
    apply RedSumR;auto.
    (* RedRes *)
    repeat (rewrite A.P.lift_New).
    rewrite A.lift_new.
    elim (IHreduction (S d) k);intro mu'';intros;and_elim.
    elim (Action.lift_mk_res mu mu' H0 d k);intro mu3;intros;and_elim.
    elim (Action.mk_res_equal (deBruijn.lift deBruijn_base (S d) k mu) mu3 H4 mu'').
    intro mu3';intros;and_elim.
    exists mu3'.
    split.
    eapply Action.equal_trans.
    apply Action.equal_sym.
    apply H5.
    auto.
    eapply RedRes;eauto.
    apply Action.equal_sym;auto.
    (* RedParL *)
    repeat (rewrite A.P.lift_Par).
    rewrite A.lift_par_r.
    elim (IHreduction d k);intro mu';intros;and_elim.
    exists mu'.
    split;auto.
    apply RedParL;auto.
    (* RedParR *)
    repeat (rewrite A.P.lift_Par).
    rewrite A.lift_par_l.
    elim (IHreduction d k);intro mu';intros;and_elim.
    exists mu'.
    split;auto.
    apply RedParR;auto.
    (* RedCloseL *)
    elim (IHreduction1 d k);intro inp';intros;and_elim.
    elim (IHreduction2 d k);intro out';intros;and_elim.
    elim (Action.lift_mk_interact inp out tau H1 d k);intro tau';intros;and_elim.
    elim (Action.mk_interact_equal (deBruijn.lift deBruijn_base d k inp) (deBruijn.lift deBruijn_base d k out) tau' H7 inp' out').
    intro tau'';intros;and_elim.
    exists tau''.
    split.
    eapply Action.equal_trans.
    apply Action.equal_sym.
    apply H8.
    trivial.
    destruct f;destruct c.
    rewrite A.lift_pseudo_appl_l.
    rewrite A.P.lift_Par.
    rewrite A.lift_F in H3.
    rewrite A.lift_C in H5.
    eapply RedCloseL;eauto.
    apply Action.equal_sym;auto.
    apply Action.equal_sym;auto.
    (* RedCloseR *)
    elim (IHreduction1 d k);intro out';intros;and_elim.
    elim (IHreduction2 d k);intro inp';intros;and_elim.
    elim (Action.lift_mk_interact out inp tau H1 d k);intro tau';intros;and_elim.
    elim (Action.mk_interact_equal (deBruijn.lift deBruijn_base d k out) (deBruijn.lift deBruijn_base d k inp) tau' H7 out' inp').
    intro tau'';intros;and_elim.
    exists tau''.
    split.
    eapply Action.equal_trans.
    apply Action.equal_sym.
    apply H8.
    trivial.
    destruct f;destruct c.
    rewrite A.lift_pseudo_appl_r.
    rewrite A.P.lift_Par.
    rewrite A.lift_C in H3.
    rewrite A.lift_F in H5.
    eapply RedCloseR;eauto.
    apply Action.equal_sym;auto.
    apply Action.equal_sym;auto.
    (* RedBangAct *)
    repeat (rewrite A.P.lift_Par || rewrite A.lift_par_r || rewrite A.P.lift_Bang).
    elim (IHreduction d k);intro mu';intros;and_elim.
    exists mu'.
    split;auto.
    apply RedBangAct.
    auto.
    (* RedBangCloseL *)
    repeat (rewrite A.P.lift_Par || rewrite A.lift_par_r || rewrite A.P.lift_mk_nu).
    elim (IHreduction1 d k);intro inp';intros;and_elim.
    elim (IHreduction2 d k);intro out';intros;and_elim.
    elim (Action.lift_mk_interact inp out tau H1 d k);intro tau';intros;and_elim.
    elim (Action.mk_interact_equal (deBruijn.lift deBruijn_base d k inp) (deBruijn.lift deBruijn_base d k out) tau' H7 inp' out').
    intro tau'';intros;and_elim.
    exists tau''.
    split.
    eapply Action.equal_trans.
    apply Action.equal_sym.
    apply H8.
    trivial.
    destruct f;destruct c.
    rewrite A.lift_pseudo_appl_l.
    rewrite A.P.lift_Bang.
    rewrite A.lift_F in H3.
    rewrite A.lift_C in H5.
    eapply RedBangCloseL;eauto.
    apply Action.equal_sym;auto.
    apply Action.equal_sym;auto.
  Qed.
  
  Lemma reduction_lift_inv : forall (q:A.P.t)(mu':Action.t)(b:A.t),(reduction q mu' b)->forall (d k:nat)(p:A.P.t),q=(deBruijn.lift A.P.deBruijn_base d k p)->(exists mu:Action.t,exists a:A.t,(reduction p mu a)/\(Action.equal mu' (deBruijn.lift Action.deBruijn_base d k mu))/\b=(deBruijn.lift A.deBruijn_base d k a)).
    induction 1;intros d k p0;destruct p0;repeat (rewrite A.P.lift_Zero || rewrite A.P.lift_Par || rewrite A.P.lift_Sum || rewrite A.P.lift_Bang || rewrite A.P.lift_New || rewrite A.P.lift_Tau || rewrite A.P.lift_Input || rewrite A.P.lift_Output || rewrite A.P.lift_IfThen);intro Heq;try (discriminate Heq);injection Heq;intros.
    (* RedSilent *)
    exists Action.mk_silent.
    exists (A.Pr p0).
    split.
    eapply RedSilent.
    split;trivial.
    apply Action.lift_mk_silent.
    rewrite H.
    rewrite A.lift_Pr.
    auto.
    (* RedInput *)
    subst e.
    elim (Action.lift_mk_input_inv d k t0 mu H);intro mu';intros;and_elim.
    exists mu'.
    exists (A.F (A.Abs p0)).
    split.
    apply RedInput;auto.
    split;auto.
    elim (Action.lift_mk_input t0 mu' H1 d k);intros;and_elim.
    rewrite H in H2.
    injection H2;intros;subst x.
    auto.
    rewrite A.lift_F.
    rewrite H0.
    auto.
    (* RedOutput *)
    subst e;subst f.
    elim (Action.lift_mk_output_inv d k t0 t1 mu m H).
    intro mu';intros.
    ex_elim H1 m'.
    exists mu'.
    exists (A.C (A.Conc 0 m' p0)).
    split.
    apply RedOutput;auto.
    split;auto.
    elim (Action.lift_mk_output t0 t1 mu' m' H1 d k);intros;and_elim.
    rewrite H in H3.
    injection H3;intros.
    subst x.
    auto.
    rewrite A.lift_C.
    rewrite plus_0_r.
    rewrite H0.
    rewrite H2.
    auto.
    (* RedIfThen *)
    elim (IHreduction d k p0 H1);intro mu'';intros.
    ex_elim H3 a'.
    elim (Action.mk_if_equal phi mu mu' H0 (deBruijn.lift deBruijn_base d k mu'') H4);intro mu3;intros;and_elim.
    subst phi.
    elim (Action.lift_mk_if_inv d k t0 mu'' mu3 H7);intro mu'3;intros;and_elim.
    exists mu'3.
    exists a'.
    split.
    eapply RedIfThen;eauto.
    split;auto.
    elim (Action.lift_mk_if t0 mu'' mu'3 H2 d k);intros;and_elim.
    rewrite H7 in H9.
    injection H9;intros;subst x.
    eapply Action.equal_trans;eauto.
    (* RedSumL *)
    elim (IHreduction d k p0_1 H1);intro mu'';intros.
    ex_elim H2 a'.
    exists mu''.
    exists a'.
    split.
    apply RedSumL;auto.
    split;auto.
    (* RedSumR *)
    elim (IHreduction d k p0_2 H0);intro mu'';intros.
    ex_elim H2 a'.
    exists mu''.
    exists a'.
    split.
    apply RedSumR;auto.
    split;auto.
    (* RedRes *)
    elim (IHreduction (S d) k p0 H1);intro mu'';intros.
    ex_elim H2 a0.
    subst a.
    elim (Action.mk_res_equal mu mu' H0 (deBruijn.lift deBruijn_base (S d) k mu'') H3).
    intro mu3;intros;and_elim.
    elim (Action.lift_mk_res_inv d k mu'' mu3 H5).
    intro mu3';intros;and_elim.
    exists mu3'.
    exists (A.new a0).
    split.
    eapply RedRes;eauto.
    split.
    elim (Action.lift_mk_res mu'' mu3' H6 d k);intros;and_elim.
    rewrite H5 in H8.
    injection H8;intros;subst x.
    eapply Action.equal_trans;eauto.
    rewrite A.lift_new;auto.
    (* RedParL *)
    elim (IHreduction d k p0_1 H1);intro mu'';intros.
    ex_elim H2 a0.
    exists mu''.
    exists (A.par_r a0 p0_2).
    split.
    apply RedParL;trivial.
    split;trivial.
    rewrite A.lift_par_r.
    rewrite H4.
    rewrite H0.
    auto.
    (* RedParR *)
    elim (IHreduction d k p0_2 H0);intro mu'';intros.
    ex_elim H2 a0.
    exists mu''.
    exists (A.par_l p0_1 a0).
    split.
    apply RedParR;trivial.
    split;trivial.
    rewrite A.lift_par_l.
    rewrite H4.
    rewrite H1.
    auto.
    (* RedCloseL *)
    elim (IHreduction1 d k p0_1 H3);intro inp';intros.
    ex_elim H4 f'.
    destruct f'.
    rewrite A.lift_Pr in H6;discriminate H6.
    elim (IHreduction2 d k p0_2 H2);intro out';intros.
    ex_elim H7 c'.
    destruct c'.
    rewrite A.lift_Pr in H9;discriminate H9.
    destruct a0;rewrite A.lift_F in H9;discriminate H9.
    destruct c0.
    rewrite A.lift_C in H9.
    injection H9;intros.
    destruct a.
    rewrite A.lift_F in H6.
    injection H6;intros.
    elim (Action.mk_interact_equal inp out tau H1 (deBruijn.lift deBruijn_base d k inp') (deBruijn.lift deBruijn_base d k out'));auto.
    intro tau';intros;and_elim.
    elim (Action.lift_mk_interact_inv d k inp' out' tau' H13).
    intro tau'';intros;and_elim.
    exists tau''.
    exists (A.pseudo_appl_l (A.Abs t2) (A.Conc n t0 t1)).
    split.
    eapply RedCloseL;eauto.
    split.
    elim (Action.lift_mk_interact inp' out' tau'' H14 d k);intros;and_elim.
    rewrite H13 in H16.
    injection H16;intros;subst x.
    eapply Action.equal_trans;eauto.
    subst f;subst c.
    rewrite A.lift_pseudo_appl_l;auto.
    destruct c0;rewrite A.lift_C in H6;discriminate H6.
    (* RedCloseR *)
    elim (IHreduction1 d k p0_1 H3);intro out';intros.
    ex_elim H4 c'.
    destruct c'.
    rewrite A.lift_Pr in H6;discriminate H6.
    destruct a;rewrite A.lift_F in H6;discriminate H6.
    destruct c0.
    rewrite A.lift_C in H6.
    elim (IHreduction2 d k p0_2 H2);intro inp';intros.
    ex_elim H7 f'.
    destruct f'.
    rewrite A.lift_Pr in H9;discriminate H9.
    destruct a.
    rewrite A.lift_F in H9.
    injection H6;intros.
    injection H9;intros.
    elim (Action.mk_interact_equal out inp tau H1 (deBruijn.lift deBruijn_base d k out') (deBruijn.lift deBruijn_base d k inp'));auto.
    intro tau';intros;and_elim.
    elim (Action.lift_mk_interact_inv d k out' inp' tau' H13).
    intro tau'';intros;and_elim.
    exists tau''.
    exists (A.pseudo_appl_r (A.Conc n t0 t1) (A.Abs t2)).
    split.
    eapply RedCloseR;eauto.
    split.
    elim (Action.lift_mk_interact out' inp' tau'' H14 d k);intros;and_elim.
    rewrite H13 in H16;injection H16;intros;subst x.
    eapply Action.equal_trans;eauto.
    subst f;subst c.
    rewrite A.lift_pseudo_appl_r;auto.
    destruct c0;rewrite A.lift_C in H9;discriminate H9.
    (* RedBangAct *)
    elim (IHreduction d k p0 H0);intro mu';intros.
    ex_elim H1 a0.
    exists mu'.
    exists (A.par_r a0 (A.P.Bang p0)).
    split.
    apply RedBangAct;auto.
    split;auto.
    rewrite A.lift_par_r.
    rewrite A.P.lift_Bang.
    rewrite H3.
    rewrite H0;auto.
    (* RedBangCloseL *)
    elim (IHreduction1 d k p0 H2);intro inp';intros.
    ex_elim H3 f'.
    destruct f'.
    rewrite A.lift_Pr in H5;discriminate H5.
    elim (IHreduction2 d k p0 H2);intro out';intros.
    ex_elim H6 c'.
    destruct c'.
    rewrite A.lift_Pr in H8;discriminate H8.
    destruct a0.
    rewrite A.lift_F in H8;discriminate H8.
    destruct a;destruct c0.
    rewrite A.lift_C in H8.
    rewrite A.lift_F in H5.
    injection H5;intros.
    injection H8;intros.
    elim (Action.mk_interact_equal inp out tau H1 (deBruijn.lift deBruijn_base d k inp') (deBruijn.lift deBruijn_base d k out'));auto.
    intro tau';intros;and_elim.
    elim (Action.lift_mk_interact_inv d k inp' out' tau' H12).
    intro tau'';intros;and_elim.
    exists tau''.
    exists (A.par_r (A.pseudo_appl_l (A.Abs t0) (A.Conc n t1 t2)) (A.P.Bang p0)).
    split.
    eapply RedBangCloseL;eauto.
    split.
    elim (Action.lift_mk_interact inp' out' tau'' H13 d k);intros;and_elim.
    rewrite H12 in H15;injection H15;intros;subst x.
    eapply Action.equal_trans;eauto.
    subst f;subst c.
    rewrite A.lift_par_r.
    rewrite A.lift_pseudo_appl_l.
    rewrite A.P.lift_Bang.
    rewrite H2.
    auto.
    destruct c0;rewrite A.lift_C in H5;discriminate H5.
  Qed.

  Lemma reduction_swapR : forall (p:A.P.t)(mu:Action.t)(a:A.t),(reduction p mu a)->forall (d k:nat),(exists mu':Action.t,(Action.equal mu' (deBruijn.swapR Action.deBruijn_base d k mu))/\(reduction (deBruijn.swapR A.P.deBruijn_base d k p) mu' (deBruijn.swapR A.deBruijn_base d k a))).
    induction 1;intros.
    (* RedSilent *)
    rewrite A.P.swapR_Tau.
    exists Action.mk_silent.
    split.
    apply Action.swapR_mk_silent.
    rewrite A.swapR_Pr.
    apply RedSilent.
    (* RedInput *)
    rewrite A.P.swapR_Input.
    elim (Action.swapR_mk_input e mu H d k).
    intro mu';intros;and_elim.
    exists mu'.
    rewrite A.swapR_F.
    split;auto.
    apply RedInput;auto.
    (* RedOutput *)
    rewrite A.P.swapR_Output.
    elim (Action.swapR_mk_output e f mu m H d k).
    intro mu';intros;and_elim.
    exists mu'.
    split;auto.
    rewrite A.swapR_C.
    rewrite plus_0_r.
    apply RedOutput;auto.
    (* RedIfThen *)
    rewrite A.P.swapR_IfThen.
    elim (IHreduction d k).
    intro mu'';intros;and_elim.
    elim (Action.swapR_mk_if phi mu mu' H0 d k).
    intro mu3;intros;and_elim.
    elim (Action.mk_if_equal (deBruijn.swapR A.P.F.deBruijn_base d k phi) (deBruijn.swapR Action.deBruijn_base d k mu) mu3 H4 mu'').
    intro mu3';intros;and_elim.
    exists mu3'.
    split.
    eapply Action.equal_trans.
    apply Action.equal_sym.
    apply H5.
    auto.
    eapply RedIfThen;eauto.
    apply Action.equal_sym;auto.
    (* RedSumL *)
    rewrite A.P.swapR_Sum.
    elim (IHreduction d k);intro mu';intros;and_elim.
    exists mu'.
    split;auto.
    apply RedSumL;auto.
    (* RedSumR *)
    rewrite A.P.swapR_Sum.
    elim (IHreduction d k);intro mu';intros;and_elim.
    exists mu'.
    split;auto.
    apply RedSumR;auto.
    (* RedRes *)
    repeat (rewrite A.P.swapR_New).
    rewrite A.swapR_new.
    elim (IHreduction (S d) k);intro mu'';intros;and_elim.
    elim (Action.swapR_mk_res mu mu' H0 d k);intro mu3;intros;and_elim.
    elim (Action.mk_res_equal (deBruijn.swapR deBruijn_base (S d) k mu) mu3 H4 mu'').
    intro mu3';intros;and_elim.
    exists mu3'.
    split.
    eapply Action.equal_trans.
    apply Action.equal_sym.
    apply H5.
    auto.
    eapply RedRes;eauto.
    apply Action.equal_sym;auto.
    (* RedParL *)
    repeat (rewrite A.P.swapR_Par).
    rewrite A.swapR_par_r.
    elim (IHreduction d k);intro mu';intros;and_elim.
    exists mu'.
    split;auto.
    apply RedParL;auto.
    (* RedParR *)
    repeat (rewrite A.P.swapR_Par).
    rewrite A.swapR_par_l.
    elim (IHreduction d k);intro mu';intros;and_elim.
    exists mu'.
    split;auto.
    apply RedParR;auto.
    (* RedCloseL *)
    elim (IHreduction1 d k);intro inp';intros;and_elim.
    elim (IHreduction2 d k);intro out';intros;and_elim.
    elim (Action.swapR_mk_interact inp out tau H1 d k);intro tau';intros;and_elim.
    elim (Action.mk_interact_equal (deBruijn.swapR deBruijn_base d k inp) (deBruijn.swapR deBruijn_base d k out) tau' H7 inp' out').
    intro tau'';intros;and_elim.
    exists tau''.
    split.
    eapply Action.equal_trans.
    apply Action.equal_sym.
    apply H8.
    trivial.
    destruct f;destruct c.
    rewrite A.swapR_pseudo_appl_l.
    rewrite A.P.swapR_Par.
    rewrite A.swapR_F in H3.
    rewrite A.swapR_C in H5.
    eapply RedCloseL;eauto.
    apply Action.equal_sym;auto.
    apply Action.equal_sym;auto.
    (* RedCloseR *)
    elim (IHreduction1 d k);intro out';intros;and_elim.
    elim (IHreduction2 d k);intro inp';intros;and_elim.
    elim (Action.swapR_mk_interact out inp tau H1 d k);intro tau';intros;and_elim.
    elim (Action.mk_interact_equal (deBruijn.swapR deBruijn_base d k out) (deBruijn.swapR deBruijn_base d k inp) tau' H7 out' inp').
    intro tau'';intros;and_elim.
    exists tau''.
    split.
    eapply Action.equal_trans.
    apply Action.equal_sym.
    apply H8.
    trivial.
    destruct f;destruct c.
    rewrite A.swapR_pseudo_appl_r.
    rewrite A.P.swapR_Par.
    rewrite A.swapR_F in H5.
    rewrite A.swapR_C in H3.
    eapply RedCloseR;eauto.
    apply Action.equal_sym;auto.
    apply Action.equal_sym;auto.
    (* RedBangAct *)
    repeat (rewrite A.P.swapR_Par || rewrite A.swapR_par_r || rewrite A.P.swapR_Bang).
    elim (IHreduction d k);intro mu';intros;and_elim.
    exists mu'.
    split;auto.
    apply RedBangAct.
    auto.
    (* RedBangCloseL *)
    repeat (rewrite A.P.swapR_Par || rewrite A.swapR_par_r || rewrite A.P.swapR_mk_nu).
    elim (IHreduction1 d k);intro inp';intros;and_elim.
    elim (IHreduction2 d k);intro out';intros;and_elim.
    elim (Action.swapR_mk_interact inp out tau H1 d k);intro tau';intros;and_elim.
    elim (Action.mk_interact_equal (deBruijn.swapR deBruijn_base d k inp) (deBruijn.swapR deBruijn_base d k out) tau' H7 inp' out').
    intro tau'';intros;and_elim.
    exists tau''.
    split.
    eapply Action.equal_trans.
    apply Action.equal_sym.
    apply H8.
    trivial.
    destruct f;destruct c.
    rewrite A.swapR_pseudo_appl_l.
    rewrite A.P.swapR_Bang.
    rewrite A.swapR_F in H3.
    rewrite A.swapR_C in H5.
    eapply RedBangCloseL;eauto.
    apply Action.equal_sym;auto.
    apply Action.equal_sym;auto.
  Qed.
End Reduction.




