(* A formalisation of the spi calculus in Coq *)
(* Copyright (C) 2005-2007 Sebastien Briais *)
(* http://lamp.epfl.ch/~sbriais/ *)

(* This library is free software; you can redistribute it and/or modify *)
(* it under the terms of the GNU Lesser General Public License as *)
(* published by the Free Software Foundation; either version 2.1 of the *)
(* License, or (at your option) any later version. *)

(* This library is distributed in the hope that it will be useful, but *)
(* WITHOUT ANY WARRANTY; without even the implied warranty of *)
(* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU *)
(* Lesser General Public License for more details. *)

(* You should have received a copy of the GNU Lesser General Public *)
(* License along with this library; if not, write to the Free Software *)
(* Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA *)
(* 02110-1301 USA *)

Require Import name.
Require Import deBruijnType.
Require Import evalSpi.
Require Import opSpi.
Require Import expressionSpi.
Require Import messageSpi.
Require Import formulaeSpi.
Require Import calculi.
Require Import missing.
Require Import tactics.
Require Import substSpi.
Require Import reductionSpi.
Require Import actionSpiNC.
Require Import Arith.
Require Import workAround.

Module RoSpi.
  Import OpSpi.
  Import EvalSpi.
  Import SubstSpi.
  Import calculi.

  (* precongruence on expressions *)
  Inductive ero : ExpressionSpi.t-> ExpressionSpi.t-> Prop :=
    | EROrefl : forall (e:ExpressionSpi.t),(ero e e)
    | EROtrans : forall (e f g:ExpressionSpi.t),(ero e f)->(ero f g)->(ero e g) 
    | EROpair : forall (e1 f1 e2 f2:ExpressionSpi.t),(ero e1 f1)->(ero e2 f2)->(ero (ExpressionSpi.EPair e1 e2) (ExpressionSpi.EPair f1 f2))
    | EROenc_s : forall (e1 f1 e2 f2:ExpressionSpi.t),(ero e1 f1)->(ero e2 f2)->(ero (ExpressionSpi.EEnc_s e1 e2) (ExpressionSpi.EEnc_s f1 f2))
    | EROdec_s : forall (e1 f1 e2 f2:ExpressionSpi.t),(ero e1 f1)->(ero e2 f2)->(ero (ExpressionSpi.EDec_s e1 e2) (ExpressionSpi.EDec_s f1 f2))
    | EROenc_a : forall (e1 f1 e2 f2:ExpressionSpi.t),(ero e1 f1)->(ero e2 f2)->(ero (ExpressionSpi.EEnc_a e1 e2) (ExpressionSpi.EEnc_a f1 f2))
    | EROdec_a : forall (e1 f1 e2 f2:ExpressionSpi.t),(ero e1 f1)->(ero e2 f2)->(ero (ExpressionSpi.EDec_a e1 e2) (ExpressionSpi.EDec_a f1 f2))
    | EROop : forall (e f:ExpressionSpi.t),(ero e f)->forall (o:op),(ero (ExpressionSpi.EOp o e) (ExpressionSpi.EOp o f))
    | EROfst : forall (e f:ExpressionSpi.t),(ero e f)->(ero (ExpressionSpi.EFst e) (ExpressionSpi.EFst f))
    | EROsnd : forall (e f:ExpressionSpi.t),(ero e f)->(ero (ExpressionSpi.ESnd e) (ExpressionSpi.ESnd f))
    | EROfstpair : forall (e1 e2:ExpressionSpi.t),(evalc e2 <> None)->(ero (ExpressionSpi.EFst (ExpressionSpi.EPair e1 e2)) e1)
    | EROsndpair : forall (e1 e2:ExpressionSpi.t),(evalc e1 <> None)->(ero (ExpressionSpi.ESnd (ExpressionSpi.EPair e1 e2)) e2)
    | EROdecenc_s : forall (e1 e2:ExpressionSpi.t),(evalc e2 <> None)->(ero (ExpressionSpi.EDec_s (ExpressionSpi.EEnc_s e1 e2) e2) e1)
    | EROdecenc_a_pub : forall (e1 e2:ExpressionSpi.t),(evalc e2 <> None)->(ero (ExpressionSpi.EDec_a (ExpressionSpi.EEnc_a e1 (ExpressionSpi.EOp Pub e2)) (ExpressionSpi.EOp Priv e2)) e1)
    | EROdecenc_a_priv : forall (e1 e2:ExpressionSpi.t),(evalc e2 <> None)->(ero (ExpressionSpi.EDec_a (ExpressionSpi.EEnc_a e1 (ExpressionSpi.EOp Priv e2)) (ExpressionSpi.EOp Pub e2)) e1).

  Lemma evalc_impl_ero : forall (e:ExpressionSpi.t)(m:MessageSpi.t),(evalc e = Some m)->(ero e (expr_of_mesg m)).
    induction e;simpl;intros.
    injection H;clear H;intros;subst m.
    simpl.
    apply EROrefl.
    case (Option_elim_option MessageSpi.t (evalc e1));intro.
    ex_elim s m1.
    rewrite s in H.
    case (Option_elim_option MessageSpi.t (evalc e2));intro.
    ex_elim s0 m2.
    rewrite s0 in H.
    injection H;clear H;intros;subst m.
    simpl.
    apply EROpair;auto.
    rewrite e in H;discriminate H.
    rewrite e in H;discriminate H.
    case (Option_elim_option MessageSpi.t (evalc e1));intro.
    ex_elim s m1.
    rewrite s in H.
    case (Option_elim_option MessageSpi.t (evalc e2));intro.
    ex_elim s0 m2.
    rewrite s0 in H.
    injection H;clear H;intros;subst m.
    simpl.
    apply EROenc_s;auto.
    rewrite e in H;discriminate H.
    rewrite e in H;discriminate H.
    case (Option_elim_option MessageSpi.t (evalc e1));intro.
    ex_elim s m1.
    rewrite s in H.
    case (Option_elim_option MessageSpi.t (evalc e2));intro.
    ex_elim s0 m2.
    rewrite s0 in H.
    injection H;clear H;intros;subst m.
    simpl.
    apply EROenc_a;auto.
    rewrite e in H;discriminate H.
    rewrite e in H;discriminate H.
    case (Option_elim_option MessageSpi.t (evalc e));intro.
    ex_elim s m0.
    rewrite s in H.
    destruct m0;try (discriminate H).
    injection H;clear H;intros;subst m.
    generalize (IHe (MessageSpi.MPair m0_1 m0_2) s).
    simpl.
    intro.
    eapply EROtrans.
    apply EROfst.
    apply H.
    apply EROfstpair.
    rewrite evalc_msg.
    red;intro.
    discriminate H0.
    rewrite e0 in H;discriminate H.
    case (Option_elim_option MessageSpi.t (evalc e));intro.
    ex_elim s m0.
    rewrite s in H.
    destruct m0;try (discriminate H).
    injection H;clear H;intros;subst m.
    generalize (IHe (MessageSpi.MPair m0_1 m0_2) s).
    simpl.
    intro.
    eapply EROtrans.
    apply EROsnd.
    apply H.
    apply EROsndpair.
    rewrite evalc_msg.
    red;intro.
    discriminate H0.
    rewrite e0 in H;discriminate H.
    case (Option_elim_option MessageSpi.t (evalc e1));intro.
    ex_elim s m1.
    rewrite s in H.
    destruct m1;try (discriminate H).
    case (Option_elim_option MessageSpi.t (evalc e2));intro.
    ex_elim s0 m2.
    rewrite s0 in H.
    destruct (MessageSpi.Ord.compare m1_2 m2);try (discriminate H).
    rewrite e in s.
    generalize (IHe1 (MessageSpi.MEnc_s m1_1 m2) s).
    simpl;intro.
    generalize (IHe2 m2 s0).
    intro.
    injection H;clear H;intros;subst m.
    eapply EROtrans.
    apply EROdec_s.
    apply H0.
    apply H1.
    apply EROdecenc_s.
    rewrite evalc_msg.
    red;intro.
    discriminate H.
    rewrite e in H;discriminate H.
    rewrite e in H;discriminate H.
    case (Option_elim_option MessageSpi.t (evalc e1));intro.
    ex_elim s m1.
    rewrite s in H.
    destruct m1;try (discriminate H).
    case (Option_elim_option MessageSpi.t (evalc e2));intro.
    ex_elim s0 m2.
    rewrite s0 in H.
    case (Option_elim_option MessageSpi.t (MessageSpi.inverse_key m1_2));intro.
    ex_elim s1 m1'_2.
    rewrite s1 in H.
    destruct (MessageSpi.Ord.compare m1'_2 m2);try (discriminate H).
    destruct m1_2;try (discriminate s1).
    destruct o;try (discriminate s1);simpl in s1.
    injection s1;intros.
    rewrite e in H0.
    destruct m2;try (discriminate H0).
    destruct o;try (discriminate H0).
    injection H0;intros.
    subst m1_2.
    generalize (IHe1 (MessageSpi.MEnc_a m1_1 (MessageSpi.MOp Pub m2)) s).
    simpl;intro.
    generalize (IHe2 (MessageSpi.MOp Priv m2) s0).
    intro.
    injection H;clear H;intros;subst m.
    eapply EROtrans.
    apply EROdec_a.
    apply H1.
    apply H2.
    simpl.
    apply EROdecenc_a_pub.
    rewrite evalc_msg.
    red;intro.
    discriminate H.
    injection s1;intros.
    rewrite e in H0.
    destruct m2;try (discriminate H0).
    destruct o;try (discriminate H0).
    injection H0;intros.
    subst m1_2.
    generalize (IHe1 (MessageSpi.MEnc_a m1_1 (MessageSpi.MOp Priv m2)) s).
    simpl;intro.
    generalize (IHe2 (MessageSpi.MOp Pub m2) s0).
    intro.
    injection H;clear H;intros;subst m.
    eapply EROtrans.
    apply EROdec_a.
    apply H1.
    apply H2.
    simpl.
    apply EROdecenc_a_priv.
    rewrite evalc_msg.
    red;intro.
    discriminate H.
    rewrite e in H;discriminate H.
    rewrite e in H;discriminate H.
    rewrite e in H;discriminate H.
    destruct (evalc e);try (discriminate H).
    injection H;intros;subst m.
    simpl.
    apply EROop.
    apply IHe;auto.
  Qed.

  Lemma ero_evalc : forall (e f:ExpressionSpi.t),(ero e f)->forall (m:MessageSpi.t),(evalc f = Some m)->(evalc e = Some m).
    induction 1;simpl;intros;auto.
    destruct (evalc f1);try (discriminate H1).
    destruct (evalc f2);try (discriminate H1).
    rewrite (IHero1 t);auto.
    rewrite (IHero2 t0);auto.
    destruct (evalc f1);try (discriminate H1).
    destruct (evalc f2);try (discriminate H1).
    rewrite (IHero1 t);auto.
    rewrite (IHero2 t0);auto.
    destruct (evalc f1);try (discriminate H1).
    destruct t;try (discriminate H1).
    rewrite (IHero1 (MessageSpi.MEnc_s t1 t2));auto.
    destruct (evalc f2);try (discriminate H1).
    rewrite (IHero2 t);auto.
    destruct (evalc f1);try (discriminate H1).
    destruct (evalc f2);try (discriminate H1).
    rewrite (IHero1 t);auto.
    rewrite (IHero2 t0);auto.
    destruct (evalc f1);try (discriminate H1).
    destruct t;try (discriminate H1).
    destruct (evalc f2);try (discriminate H1).
    rewrite (IHero1 (MessageSpi.MEnc_a t1 t2));auto.
    rewrite (IHero2 t);auto.
    destruct (evalc f);try (discriminate H0).
    rewrite (IHero t);auto.
    destruct (evalc f);try (discriminate H0).
    rewrite (IHero t);auto.
    destruct (evalc f);try (discriminate H0).
    rewrite (IHero t);auto.
    rewrite H0.
    destruct (evalc e2);auto.
    elim H;auto.
    rewrite H0.
    destruct (evalc e1);auto.
    elim H;auto.
    rewrite H0.
    destruct (evalc e2);auto.
    destruct (MessageSpi.Ord.compare t t);auto;elim (MessageSpi.Ord.lt_not_eq t t);auto.
    elim H;auto.
    destruct (evalc e1);auto;try (discriminate H0).
    destruct (evalc e2);auto.
    unfold MessageSpi.inverse_key.
    destruct (MessageSpi.Ord.compare (MessageSpi.MOp Priv t0) (MessageSpi.MOp Priv t0));auto;elim (MessageSpi.Ord.lt_not_eq (MessageSpi.MOp Priv t0) (MessageSpi.MOp Priv t0));auto.
    elim H;auto.
    destruct (evalc e1);auto;try (discriminate H0).
    destruct (evalc e2);auto.
    unfold MessageSpi.inverse_key.
    destruct (MessageSpi.Ord.compare (MessageSpi.MOp Pub t0) (MessageSpi.MOp Pub t0));auto;elim (MessageSpi.Ord.lt_not_eq (MessageSpi.MOp Pub t0) (MessageSpi.MOp Pub t0));auto.
    elim H;auto.
  Qed.

  Lemma evalc_ero : forall (e f:ExpressionSpi.t),(ero e f)->forall (m:MessageSpi.t),(evalc e = Some m)->(evalc f = Some m).
    induction 1;simpl;intros;auto.
    destruct (evalc e1);try (discriminate H1).
    destruct (evalc e2);try (discriminate H1).
    rewrite (IHero1 t);auto.
    rewrite (IHero2 t0);auto.
    destruct (evalc e1);try (discriminate H1).
    destruct (evalc e2);try (discriminate H1).
    rewrite (IHero1 t);auto.
    rewrite (IHero2 t0);auto.
    destruct (evalc e1);try (discriminate H1).
    destruct t;try (discriminate H1).
    rewrite (IHero1 (MessageSpi.MEnc_s t1 t2));auto.
    destruct (evalc e2);try (discriminate H1).
    rewrite (IHero2 t);auto.
    destruct (evalc e1);try (discriminate H1).
    destruct (evalc e2);try (discriminate H1).
    rewrite (IHero1 t);auto.
    rewrite (IHero2 t0);auto.
    destruct (evalc e1);try (discriminate H1).
    destruct t;try (discriminate H1).
    destruct (evalc e2);try (discriminate H1).
    rewrite (IHero1 (MessageSpi.MEnc_a t1 t2));auto.
    rewrite (IHero2 t);auto.
    destruct (evalc e);try (discriminate H0).
    rewrite (IHero t);auto.
    destruct (evalc e);try (discriminate H0).
    rewrite (IHero t);auto.
    destruct (evalc e);try (discriminate H0).
    rewrite (IHero t);auto.
    destruct (evalc e1);try (discriminate H0).
    destruct (evalc e2);try (discriminate H0);auto.
    destruct (evalc e1);try (discriminate H0).
    destruct (evalc e2);try (discriminate H0);auto.
    destruct (evalc e1);try (discriminate H0).
    destruct (evalc e2);try (discriminate H0).
    destruct (MessageSpi.Ord.compare t0 t0);auto;elim (MessageSpi.Ord.lt_not_eq t0 t0);auto.
    destruct (evalc e1);try (discriminate H0).
    destruct (evalc e2);try (discriminate H0);auto.
    unfold MessageSpi.inverse_key in H0.
    destruct (MessageSpi.Ord.compare (MessageSpi.MOp Priv t0) (MessageSpi.MOp Priv t0));auto;elim (MessageSpi.Ord.lt_not_eq (MessageSpi.MOp Priv t0) (MessageSpi.MOp Priv t0));auto.
    destruct (evalc e1);try (discriminate H0).
    destruct (evalc e2);try (discriminate H0);auto.
    unfold MessageSpi.inverse_key in H0.
    destruct (MessageSpi.Ord.compare (MessageSpi.MOp Pub t0) (MessageSpi.MOp Pub t0));auto;elim (MessageSpi.Ord.lt_not_eq (MessageSpi.MOp Pub t0) (MessageSpi.MOp Pub t0));auto.
  Qed.

  Lemma ero_evala : forall (e f:ExpressionSpi.t),(ero e f)->(evala e = evala f).
    induction 1;simpl;intros;auto;try congruence.
    rewrite IHero1;rewrite IHero2;auto.
    rewrite IHero1;rewrite IHero2;auto.
    rewrite IHero;auto.
    rewrite IHero;auto.
  Qed.

  Lemma ero_list_succeeds : forall (es fs:list ExpressionSpi.t),(List_relation ero es fs)->(List_forall evalc_succeeds es)->(List_forall evalc_succeeds fs).
    induction 1.
    auto.
    intros.
    inversion H1.
    apply List_forall_cons;auto.
    unfold evalc_succeeds in H4.
    unfold evalc_succeeds.
    ex_elim H4 m.
    rewrite (evalc_ero x y H m H4).
    exists m;auto.
  Qed.
    
  Lemma succeeds_ero_list : forall (es fs:list ExpressionSpi.t),(List_relation ero es fs)->(List_forall evalc_succeeds fs)->(List_forall evalc_succeeds es).
    induction 1;auto.
    intro.
    inversion H1.
    apply List_forall_cons;auto.
    unfold evalc_succeeds in H4.
    unfold evalc_succeeds.
    ex_elim H4 m.
    rewrite (ero_evalc x y H m H4).
    exists m;auto.
  Qed.

  Inductive fro : FormulaeSpi.t -> FormulaeSpi.t -> Prop :=
    | FROname : forall (e f:ExpressionSpi.t),(ero e f)->(fro (FormulaeSpi.FName e) (FormulaeSpi.FName f))
    | FROequal : forall (e1 f1 e2 f2:ExpressionSpi.t),(ero e1 f1)->(ero e2 f2)->(fro (FormulaeSpi.FEqual e1 e2) (FormulaeSpi.FEqual f1 f2)).

  Lemma FROrefl : forall (phi:FormulaeSpi.t),(fro phi phi).
    destruct phi.
    apply FROname.
    apply EROrefl.
    apply FROequal.
    apply EROrefl.
    apply EROrefl.
  Qed.

  Lemma FROtrans : forall (x y z:FormulaeSpi.t),(fro x y)->(fro y z)->(fro x z).
    intros x y z H.
    generalize z;clear z.
    induction H;intros.
    inversion H0.
    apply FROname.
    eapply EROtrans;eauto.
    inversion H1.
    apply FROequal.
    eapply EROtrans;eauto.
    eapply EROtrans;eauto.
  Qed.

  Lemma fro_eval : forall (phi psi:FormulaeSpi.t),(fro phi psi)->(eval psi = true)->(eval phi = true).
    induction 1.
    generalize (ero_evalc e f H);simpl;intros.
    destruct (evalc f);try (discriminate H1).
    rewrite (H0 t);auto.
    generalize (ero_evalc e1 f1 H).
    generalize (ero_evalc e2 f2 H0);simpl;intros.
    destruct (evalc f1);try (discriminate H3).
    destruct (evalc f2);try (discriminate H3).
    rewrite (H1 t0);auto.
    rewrite (H2 t);auto.
  Qed.

  Lemma eval_fro : forall (phi psi:FormulaeSpi.t),(fro phi psi)->(eval phi = true)->(eval psi = true).
    induction 1.
    generalize (evalc_ero e f H);simpl;intros.
    destruct (evalc e);try (discriminate H1).
    rewrite (H0 t);auto.
    generalize (evalc_ero e1 f1 H).
    generalize (evalc_ero e2 f2 H0);simpl;intros.
    destruct (evalc e1);try (discriminate H3).
    destruct (evalc e2);try (discriminate H3).
    rewrite (H1 t0);auto.
    rewrite (H2 t);auto.
  Qed.

  Lemma fro_eval_nc : forall (phi psi:FormulaeSpi.t),(fro phi psi)->forall (l:list Name.t),(eval_nc psi = (true,l))->(eval_nc phi = (true,l)).
    induction 1.
    generalize (ero_evalc e f H);simpl;intros.
    destruct (evalc f);try (discriminate H1).
    rewrite (H0 t);auto.
    generalize (ero_evalc e1 f1 H).
    generalize (ero_evalc e2 f2 H0);simpl;intros.
    destruct (evalc f1);try (discriminate H3).
    destruct (evalc f2);try (discriminate H3).
    rewrite (H1 t0);auto.
    rewrite (H2 t);auto.
  Qed.

  Lemma eval_nc_fro : forall (phi psi:FormulaeSpi.t),(fro phi psi)->forall (l:list Name.t),(eval_nc phi = (true,l))->(eval_nc psi = (true,l)).
    induction 1.
    generalize (evalc_ero e f H);simpl;intros.
    destruct (evalc e);try (discriminate H1).
    rewrite (H0 t);auto.
    generalize (evalc_ero e1 f1 H).
    generalize (evalc_ero e2 f2 H0);simpl;intros.
    destruct (evalc e1);try (discriminate H3).
    destruct (evalc e2);try (discriminate H3).
    rewrite (H1 t0);auto.
    rewrite (H2 t);auto.
  Qed.

  Inductive pro : SpiCalcul.t -> SpiCalcul.t -> Prop :=
    | PROzero : (pro SpiCalcul.Zero SpiCalcul.Zero)
    | PROpar : forall (p1 q1 p2 q2:SpiCalcul.t),(pro p1 q1)->(pro p2 q2)->(pro (SpiCalcul.Par p1 p2) (SpiCalcul.Par q1 q2))
    | PROsum : forall (p1 q1 p2 q2:SpiCalcul.t),(pro p1 q1)->(pro p2 q2)->(pro (SpiCalcul.Sum p1 p2) (SpiCalcul.Sum q1 q2))
    | PRObang : forall (p q:SpiCalcul.t),(pro p q)->(pro (SpiCalcul.Bang p) (SpiCalcul.Bang q))
    | PROnew : forall (p q:SpiCalcul.t),(pro p q)->(pro (SpiCalcul.New p) (SpiCalcul.New q))
    | PROtau : forall (p q:SpiCalcul.t),(pro p q)->(pro (SpiCalcul.Tau p) (SpiCalcul.Tau q))
    | PROinput : forall (e f:ExpressionSpi.t)(p q:SpiCalcul.t),(ero e f)->(pro p q)->(pro (SpiCalcul.Input e p) (SpiCalcul.Input f q))
    | PROoutput : forall (e1 f1 e2 f2:ExpressionSpi.t)(p q:SpiCalcul.t),(ero e1 f1)->(ero e2 f2)->(pro p q)->(pro (SpiCalcul.Output e1 e2 p) (SpiCalcul.Output f1 f2 q))
    | PROifthen : forall (phi psi:FormulaeSpi.t)(p q:SpiCalcul.t),(fro phi psi)->(pro p q)->(pro (SpiCalcul.IfThen phi p) (SpiCalcul.IfThen psi q)).

  Lemma PROrefl : forall (p:SpiCalcul.t),(pro p p).
    induction p.
    apply PROzero.
    apply PROpar;auto.
    apply PROsum;auto.
    apply PRObang;auto.
    apply PROnew;auto.
    apply PROtau;auto.
    apply PROinput;auto.
    apply EROrefl.
    apply PROoutput;auto.
    apply EROrefl.
    apply EROrefl.
    apply PROifthen;auto.
    apply FROrefl.
  Qed.

  Lemma PROtrans : forall (p q r:SpiCalcul.t),(pro p q)->(pro q r)->(pro p r).
    intros p q r H.
    generalize r;clear r.
    induction H;intros;auto.
    inversion H1.
    apply PROpar;auto.
    inversion H1.
    apply PROsum;auto.
    inversion H0.
    apply PRObang;auto.
    inversion H0.
    apply PROnew;auto.
    inversion H0.
    apply PROtau;auto.
    inversion H1.
    apply PROinput;auto.
    eapply EROtrans;eauto.
    inversion H2.
    apply PROoutput;auto.
    eapply EROtrans;eauto.
    eapply EROtrans;eauto.
    inversion H1.
    apply PROifthen;auto.
    eapply FROtrans;eauto.
  Qed.

  Lemma PROmk_nu : forall (p q:SpiCalcul.t),(pro p q)->forall (n:nat),(pro (SpiCalcul.mk_nu n p) (SpiCalcul.mk_nu n q)).
    intros.
    induction n.
    repeat (rewrite SpiCalcul.mk_nu_0);auto.
    repeat (rewrite SpiCalcul.mk_nu_S_1).
    apply PROnew.
    apply IHn.
  Qed.

  Inductive aro : AgentSpi.t -> AgentSpi.t -> Prop :=
    | AROproc : forall (p q:SpiCalcul.t),(pro p q)->(aro (AgentSpi.Pr p) (AgentSpi.Pr q))
    | AROabs : forall (p q:SpiCalcul.t),(pro p q)->(aro (AgentSpi.F (AgentSpi.Abs p)) (AgentSpi.F (AgentSpi.Abs q)))
    | AROconc : forall (n:nat)(m:ExpressionSpi.t)(p q:SpiCalcul.t),(pro p q)->(aro (AgentSpi.C (AgentSpi.Conc n m p)) (AgentSpi.C (AgentSpi.Conc n m q))).

  Lemma AROrefl : forall (a:AgentSpi.t),(aro a a).
    destruct a.
    apply AROproc.
    apply PROrefl.
    destruct a.
    apply AROabs.
    apply PROrefl.
    destruct c.
    apply AROconc.
    apply PROrefl.
  Qed.

  Lemma AROtrans : forall (a b c:AgentSpi.t),(aro a b)->(aro b c)->(aro a c).
    intros a b c H.
    generalize c;clear c.
    induction H;intros.
    inversion H0.
    apply AROproc.
    eapply PROtrans;eauto.
    inversion H0.
    apply AROabs.
    eapply PROtrans;eauto.
    inversion H0.
    apply AROconc.
    eapply PROtrans;eauto.
  Qed.

  Definition preserves_evalc (f:nat->Name.t->Name.t) := forall (d:nat)(e:ExpressionSpi.t),(evalc (ExpressionSpi.t_map f d e))=(option_map (MessageSpi.t_map f d) (evalc e)).
  
  Lemma ero_preserves : forall (e f:ExpressionSpi.t),(ero e f)->forall (g:nat->Name.t->Name.t),(preserves_evalc g)->forall (d:nat),(ero (ExpressionSpi.t_map g d e) (ExpressionSpi.t_map g d f)).
    induction 1;intros.
    apply EROrefl.
    eapply EROtrans;eauto.
    simpl;apply EROpair;auto.
    simpl;apply EROenc_s;auto.
    simpl;apply EROdec_s;auto.
    simpl;apply EROenc_a;auto.
    simpl;apply EROdec_a;auto.
    simpl;apply EROop;auto.
    simpl;apply EROfst;auto.
    simpl;apply EROsnd;auto.
    simpl;apply EROfstpair.
    red;intro.
    apply H.
    rewrite (H0 d e2) in H1.
    destruct (evalc e2);try (discriminate H1);auto.
    simpl;apply EROsndpair.
    red;intro.
    apply H.
    rewrite (H0 d e1) in H1.
    destruct (evalc e1);try (discriminate H1);auto.
    simpl;apply EROdecenc_s.
    red;intro.
    apply H.
    rewrite (H0 d e2) in H1.
    destruct (evalc e2);try (discriminate H1);auto.
    simpl;apply EROdecenc_a_pub.
    red;intro.
    apply H.
    rewrite (H0 d e2) in H1.
    destruct (evalc e2);try (discriminate H1);auto.
    simpl;apply EROdecenc_a_priv.
    red;intro.
    apply H.
    rewrite (H0 d e2) in H1.
    destruct (evalc e2);try (discriminate H1);auto.
  Qed.

  Lemma ero_lift : forall (e f:ExpressionSpi.t),(ero e f)->forall (d k:nat),(ero (deBruijn.lift ExpressionSpi.deBruijn_base d k e) (deBruijn.lift ExpressionSpi.deBruijn_base d k f)).
    intros.
    simpl.
    apply ero_preserves;auto.
    unfold preserves_evalc.
    intros.
    apply evalc_map.
    intros.
    apply (deBruijn.lift_injective Name.deBruijn_base).
  Qed.

  Lemma ero_swapR : forall (e f:ExpressionSpi.t),(ero e f)->forall (d k:nat),(ero (deBruijn.swapR ExpressionSpi.deBruijn_base d k e) (deBruijn.swapR ExpressionSpi.deBruijn_base d k f)).
    intros.
    simpl.
    apply ero_preserves;auto.
    unfold preserves_evalc.
    intros.
    apply evalc_map.
    intros.
    apply (deBruijn.swapR_injective Name.deBruijn_base).
  Qed.

  Lemma ero_low_subst : forall (e f:ExpressionSpi.t),(ero e f)->forall (d k:nat)(es:list ExpressionSpi.t),(List_forall evalc_succeeds es)->(ero (deBruijn.low_subst ExpressionSpi.deBruijn_subst d k es e) (deBruijn.low_subst ExpressionSpi.deBruijn_subst d k es f)).
    induction 1;intros.
    apply EROrefl.
    eapply EROtrans;eauto.
    repeat (rewrite ExpressionSpi.low_subst_EPair).
    apply EROpair;auto.
    repeat (rewrite ExpressionSpi.low_subst_EEnc_s).
    apply EROenc_s;auto.
    repeat (rewrite ExpressionSpi.low_subst_EDec_s).
    apply EROdec_s;auto.
    repeat (rewrite ExpressionSpi.low_subst_EEnc_a).
    apply EROenc_a;auto.
    repeat (rewrite ExpressionSpi.low_subst_EDec_a).
    apply EROdec_a;auto.
    repeat (rewrite ExpressionSpi.low_subst_EOp).
    apply EROop;auto.
    repeat (rewrite ExpressionSpi.low_subst_EFst).
    apply EROfst;auto.
    repeat (rewrite ExpressionSpi.low_subst_ESnd).
    apply EROsnd;auto.
    repeat (rewrite ExpressionSpi.low_subst_EFst || rewrite ExpressionSpi.low_subst_EPair).
    apply EROfstpair.
    red;intro.
    elim H.
    case (Option_elim_option MessageSpi.t (evalc e2));intro;auto.
    ex_elim s m2.
    rewrite (low_subst_evalc es H0 e2 m2 s d k) in H1;discriminate H1.
    repeat (rewrite ExpressionSpi.low_subst_ESnd || rewrite ExpressionSpi.low_subst_EPair).
    apply EROsndpair.
    red;intro.
    elim H.
    case (Option_elim_option MessageSpi.t (evalc e1));intro;auto.
    ex_elim s m1.
    rewrite (low_subst_evalc es H0 e1 m1 s d k) in H1;discriminate H1.
    repeat (rewrite ExpressionSpi.low_subst_EDec_s || rewrite ExpressionSpi.low_subst_EEnc_s).
    apply EROdecenc_s.
    red;intro.
    elim H.
    case (Option_elim_option MessageSpi.t (evalc e2));intro;auto.
    ex_elim s m2.
    rewrite (low_subst_evalc es H0 e2 m2 s d k) in H1;discriminate H1.
    repeat (rewrite ExpressionSpi.low_subst_EDec_a || rewrite ExpressionSpi.low_subst_EEnc_a || rewrite ExpressionSpi.low_subst_EOp).
    apply EROdecenc_a_pub.
    red;intro.
    elim H.
    case (Option_elim_option MessageSpi.t (evalc e2));intro;auto.
    ex_elim s m2.
    rewrite (low_subst_evalc es H0 e2 m2 s d k) in H1;discriminate H1.
    repeat (rewrite ExpressionSpi.low_subst_EDec_a || rewrite ExpressionSpi.low_subst_EEnc_a || rewrite ExpressionSpi.low_subst_EOp).
    apply EROdecenc_a_priv.
    red;intro.
    elim H.
    case (Option_elim_option MessageSpi.t (evalc e2));intro;auto.
    ex_elim s m2.
    rewrite (low_subst_evalc es H0 e2 m2 s d k) in H1;discriminate H1.
  Qed.

  Lemma ero_low_subst_list_ero : forall  (e:ExpressionSpi.t)(es fs:list ExpressionSpi.t),(List_relation ero es fs)->forall (d k:nat),(ero (deBruijn.low_subst ExpressionSpi.deBruijn_subst d k es e) (deBruijn.low_subst ExpressionSpi.deBruijn_subst d k fs e)).
    induction e;intros.
    destruct t;simpl.
    unfold Name.low_subst_generic.
    unfold DeBruijnNat.low_subst_gen.
    unfold DeBruijnNat.low_subst_generic.
    unfold DeBruijnNat.low_subst_generic_aux.
    case le_lt_dec.
    case (Option_elim_option ExpressionSpi.t (nth_error es (n-d))).
    intros.
    ex_elim s e'.
    elim (List_relation_nth_Some ero es fs H (n-d) e');auto.
    intro f';intros;and_elim.
    rewrite s.
    rewrite H1.
    generalize (ero_lift e' f' H0 0 d).
    simpl.
    auto.
    intros.
    rewrite e.
    rewrite (List_relation_nth_None ero es fs H (n-d));auto.
    rewrite (List_relation_length ero es fs H).
    apply EROrefl.
    intro.
    apply EROrefl.
    repeat (rewrite ExpressionSpi.low_subst_EPair).
    apply EROpair;auto.
    repeat (rewrite ExpressionSpi.low_subst_EEnc_s).
    apply EROenc_s;auto.
    repeat (rewrite ExpressionSpi.low_subst_EEnc_a).
    apply EROenc_a;auto.
    repeat (rewrite ExpressionSpi.low_subst_EFst).
    apply EROfst;auto.
    repeat (rewrite ExpressionSpi.low_subst_ESnd).
    apply EROsnd;auto.
    repeat (rewrite ExpressionSpi.low_subst_EDec_s).
    apply EROdec_s;auto.
    repeat (rewrite ExpressionSpi.low_subst_EDec_a).
    apply EROdec_a;auto.
    repeat (rewrite ExpressionSpi.low_subst_EOp).
    apply EROop;auto.
  Qed.

  Lemma ero_low_subst_list : forall  (e f:ExpressionSpi.t),(ero e f)->forall (es fs:list ExpressionSpi.t),(List_forall evalc_succeeds es)->(List_relation ero es fs)->forall (d k:nat),(ero (deBruijn.low_subst ExpressionSpi.deBruijn_subst d k es e) (deBruijn.low_subst ExpressionSpi.deBruijn_subst d k fs f)).
    intros.
    eapply EROtrans.
    apply (ero_low_subst e f H d k es H0).
    apply ero_low_subst_list_ero.
    auto.
  Qed.

  Lemma fro_preserves : forall (phi psi:FormulaeSpi.t),(fro phi psi)->forall (g:nat->Name.t->Name.t),(preserves_evalc g)->forall (d:nat),(fro (FormulaeSpi.t_map (fun d => ExpressionSpi.t_map g d) d phi) (FormulaeSpi.t_map (fun d => ExpressionSpi.t_map g d) d psi)).
    induction 1;intros;simpl.
    apply FROname.
    apply ero_preserves;auto.
    apply FROequal.
    apply ero_preserves;auto.
    apply ero_preserves;auto.
  Qed.
  
  Lemma fro_lift : forall (phi psi:FormulaeSpi.t),(fro phi psi)->forall (d k:nat),(fro (deBruijn.lift FormulaeSpi.deBruijn_base d k phi) (deBruijn.lift FormulaeSpi.deBruijn_base d k psi)).
    intros.
    simpl.
    apply fro_preserves;auto.
    unfold preserves_evalc;intros.
    apply evalc_map.  
    intros.
    apply (deBruijn.lift_injective Name.deBruijn_base).
  Qed.
  
  Lemma fro_swapR : forall (phi psi:FormulaeSpi.t),(fro phi psi)->forall (d k:nat),(fro (deBruijn.swapR FormulaeSpi.deBruijn_base d k phi) (deBruijn.swapR FormulaeSpi.deBruijn_base d k psi)).
    intros.
    simpl.
    apply fro_preserves;auto.
    unfold preserves_evalc;intros.
    apply evalc_map.  
    intros.
    apply (deBruijn.swapR_injective Name.deBruijn_base).
  Qed.

  Lemma fro_low_subst : forall (phi psi:FormulaeSpi.t),(fro phi psi)->forall (d k:nat)(es:list ExpressionSpi.t),(List_forall evalc_succeeds es)->(fro (deBruijn.low_subst FormulaeSpi.deBruijn_subst d k es phi) (deBruijn.low_subst FormulaeSpi.deBruijn_subst d k es psi)).
    induction 1;intros.
    repeat (rewrite FormulaeSpi.low_subst_FName).
    apply FROname.
    apply ero_low_subst;auto.
    repeat (rewrite FormulaeSpi.low_subst_FEqual).
    apply FROequal.
    apply ero_low_subst;auto.
    apply ero_low_subst;auto.
  Qed.

  Lemma fro_low_subst_list_ero : forall  (phi:FormulaeSpi.t)(es fs:list ExpressionSpi.t),(List_relation ero es fs)->forall (d k:nat),(fro (deBruijn.low_subst FormulaeSpi.deBruijn_subst d k es phi) (deBruijn.low_subst FormulaeSpi.deBruijn_subst d k fs phi)).
    induction phi;intros.
    repeat (rewrite FormulaeSpi.low_subst_FName).
    apply FROname.
    apply ero_low_subst_list_ero;auto.
    repeat (rewrite FormulaeSpi.low_subst_FEqual).
    apply FROequal.
    apply ero_low_subst_list_ero;auto.
    apply ero_low_subst_list_ero;auto.
  Qed.

  Lemma fro_low_subst_list : forall  (phi psi:FormulaeSpi.t),(fro phi psi)->forall (es fs:list ExpressionSpi.t),(List_forall evalc_succeeds es)->(List_relation ero es fs)->forall (d k:nat),(fro (deBruijn.low_subst FormulaeSpi.deBruijn_subst d k es phi) (deBruijn.low_subst FormulaeSpi.deBruijn_subst d k fs psi)).
    intros.
    eapply FROtrans.
    apply (fro_low_subst phi psi H d k es H0).
    apply fro_low_subst_list_ero.
    auto.
  Qed.

  Lemma pro_preserves : forall (p q:SpiCalcul.t),(pro p q)->forall (g:nat->Name.t->Name.t),(preserves_evalc g)->forall (d:nat),(pro (SpiCalcul.t_map (fun d => ExpressionSpi.t_map g d) (fun d => FormulaeSpi.t_map (fun d => ExpressionSpi.t_map g d) d) d p) (SpiCalcul.t_map (fun d => ExpressionSpi.t_map g d) (fun d => FormulaeSpi.t_map (fun d => ExpressionSpi.t_map g d) d) d q)).
    induction 1;intros;simpl.
    apply PROzero.
    apply PROpar;auto.
    apply PROsum;auto.
    apply PRObang;auto.
    apply PROnew;auto.
    apply PROtau;auto.
    apply PROinput;auto.
    apply ero_preserves;auto.
    apply PROoutput;auto.
    apply ero_preserves;auto.
    apply ero_preserves;auto.
    apply PROifthen;auto.
    apply fro_preserves;auto.
  Qed.
    
  Lemma pro_lift : forall (p q:SpiCalcul.t),(pro p q)->forall (d k:nat),(pro (deBruijn.lift SpiCalcul.deBruijn_base d k p) (deBruijn.lift SpiCalcul.deBruijn_base d k q)).
    intros.
    simpl.
    apply pro_preserves;auto.
    unfold preserves_evalc;intros.
    apply evalc_map.  
    intros.
    apply (deBruijn.lift_injective Name.deBruijn_base).
  Qed.

  Lemma pro_swapR : forall (p q:SpiCalcul.t),(pro p q)->forall (d k:nat),(pro (deBruijn.swapR SpiCalcul.deBruijn_base d k p) (deBruijn.swapR SpiCalcul.deBruijn_base d k q)).
    intros.
    simpl.
    apply pro_preserves;auto.
    unfold preserves_evalc;intros.
    apply evalc_map.  
    intros.
    apply (deBruijn.swapR_injective Name.deBruijn_base).
  Qed.

  Lemma pro_low_subst : forall (p q:SpiCalcul.t),(pro p q)->forall (d k:nat)(es:list ExpressionSpi.t),(List_forall evalc_succeeds es)->(pro (deBruijn.low_subst SpiCalcul.deBruijn_subst d k es p) (deBruijn.low_subst SpiCalcul.deBruijn_subst d k es q)).
    induction 1;intros.
    repeat (rewrite SpiCalcul.low_subst_Zero).
    apply PROzero.
    repeat (rewrite SpiCalcul.low_subst_Par).
    apply PROpar;auto.
    repeat (rewrite SpiCalcul.low_subst_Sum).
    apply PROsum;auto.
    repeat (rewrite SpiCalcul.low_subst_Bang).
    apply PRObang;auto.
    repeat (rewrite SpiCalcul.low_subst_New).
    apply PROnew;auto.
    repeat (rewrite SpiCalcul.low_subst_Tau).
    apply PROtau;auto.
    repeat (rewrite SpiCalcul_low_subst_Input_aux).
    apply PROinput;auto.
    apply ero_low_subst;auto.
    repeat (rewrite SpiCalcul_low_subst_Output_aux).
    apply PROoutput;auto.
    apply ero_low_subst;auto.
    apply ero_low_subst;auto.
    repeat (rewrite SpiCalcul_low_subst_IfThen_aux).
    apply PROifthen;auto.
    apply fro_low_subst;auto.
  Qed.

  Lemma pro_low_subst_list_ero : forall  (p:SpiCalcul.t)(es fs:list ExpressionSpi.t),(List_relation ero es fs)->forall (d k:nat),(pro (deBruijn.low_subst SpiCalcul.deBruijn_subst d k es p) (deBruijn.low_subst SpiCalcul.deBruijn_subst d k fs p)).
    induction p;intros.
    repeat (rewrite SpiCalcul.low_subst_Zero).
    apply PROzero.
    repeat (rewrite SpiCalcul.low_subst_Par).
    apply PROpar;auto.
    repeat (rewrite SpiCalcul.low_subst_Sum).
    apply PROsum;auto.
    repeat (rewrite SpiCalcul.low_subst_Bang).
    apply PRObang;auto.
    repeat (rewrite SpiCalcul.low_subst_New).
    apply PROnew;auto.
    repeat (rewrite SpiCalcul.low_subst_Tau).
    apply PROtau;auto.
    repeat (rewrite SpiCalcul_low_subst_Input_aux).
    apply PROinput;auto.
    apply ero_low_subst_list_ero;auto.
    repeat (rewrite SpiCalcul_low_subst_Output_aux).
    apply PROoutput;auto.
    apply ero_low_subst_list_ero;auto.
    apply ero_low_subst_list_ero;auto.
    repeat (rewrite SpiCalcul_low_subst_IfThen_aux).
    apply PROifthen;auto.
    apply fro_low_subst_list_ero;auto.
  Qed.

  Lemma pro_low_subst_list : forall  (p q:SpiCalcul.t),(pro p q)->forall (es fs:list ExpressionSpi.t),(List_forall evalc_succeeds es)->(List_relation ero es fs)->forall (d k:nat),(pro (deBruijn.low_subst SpiCalcul.deBruijn_subst d k es p) (deBruijn.low_subst SpiCalcul.deBruijn_subst d k fs q)).
    intros.
    eapply PROtrans.
    apply (pro_low_subst p q H d k es H0).
    apply pro_low_subst_list_ero.
    auto.
  Qed.

  Lemma aro_lift : forall (a b:AgentSpi.t),(aro a b)->forall (d k:nat),(aro (deBruijn.lift AgentSpi.deBruijn_base d k a) (deBruijn.lift AgentSpi.deBruijn_base d k b)).
    induction 1;intros.
    simpl.
    apply AROproc.
    apply (pro_lift p q H d k).
    simpl.
    apply AROabs.
    apply (pro_lift p q H (S d) k).
    simpl.
    apply AROconc.
    apply (pro_lift p q H (d+n) k).
  Qed.
  
  Lemma aro_swapR : forall (a b:AgentSpi.t),(aro a b)->forall (d k:nat),(aro (deBruijn.swapR AgentSpi.deBruijn_base d k a) (deBruijn.swapR AgentSpi.deBruijn_base d k b)).
    induction 1;intros.
    simpl.
    apply AROproc.
    apply (pro_swapR p q H d k).
    simpl.
    apply AROabs.
    apply (pro_swapR p q H (S d) k).
    simpl.
    apply AROconc.
    apply (pro_swapR p q H (d+n) k).
  Qed.

  Lemma aro_new : forall (a b:AgentSpi.t),(aro a b)->(aro (AgentSpi.new a) (AgentSpi.new b)).
    induction 1.
    repeat (rewrite AgentSpi.new_Pr).
    apply AROproc.
    apply PROnew;auto.
    repeat (rewrite AgentSpi.new_F).
    repeat (rewrite AgentSpi.new_abs_def).
    apply AROabs.
    apply PROnew.
    apply pro_swapR;auto.
    repeat (rewrite AgentSpi.new_C).
    repeat (rewrite AgentSpi.new_conc_def).
    case deBruijn.mem_fn.
    apply AROconc;auto.
    apply AROconc.
    apply PROnew.
    apply pro_swapR;auto.
  Qed.

  Lemma aro_par_l : forall (a b:AgentSpi.t)(p q:SpiCalcul.t),(aro a b)->(pro p q)->(aro (AgentSpi.par_l p a) (AgentSpi.par_l q b)).
    induction 1;intros.
    repeat (rewrite AgentSpi.par_l_Pr).
    apply AROproc.
    apply PROpar;auto.
    repeat (rewrite AgentSpi.par_l_F).
    repeat (rewrite AgentSpi.abs_par_l_def).
    apply AROabs.
    apply PROpar;auto.
    apply pro_lift;auto.
    repeat (rewrite AgentSpi.par_l_C).
    repeat (rewrite AgentSpi.conc_par_l_def).
    apply AROconc.
    apply PROpar;auto.
    apply pro_lift;auto.
  Qed.
  
  Lemma aro_par_r : forall (a b:AgentSpi.t)(p q:SpiCalcul.t),(aro a b)->(pro p q)->(aro (AgentSpi.par_r a p) (AgentSpi.par_r b q)).
    induction 1;intros.
    repeat (rewrite AgentSpi.par_r_Pr).
    apply AROproc.
    apply PROpar;auto.
    repeat (rewrite AgentSpi.par_r_F).
    repeat (rewrite AgentSpi.abs_par_r_def).
    apply AROabs.
    apply PROpar;auto.
    apply pro_lift;auto.
    repeat (rewrite AgentSpi.par_r_C).
    repeat (rewrite AgentSpi.conc_par_r_def).
    apply AROconc.
    apply PROpar;auto.
    apply pro_lift;auto.
  Qed.

  Lemma aro_pseudo_appl_l : forall (p1 q1 p2 q2:SpiCalcul.t)(n:nat)(m:MessageSpi.t),(pro p1 p2)->(pro q1 q2)->(aro (AgentSpi.pseudo_appl_l (AgentSpi.Abs p1) (AgentSpi.Conc n (EvalSpi.expr_of_mesg m) q1)) (AgentSpi.pseudo_appl_l (AgentSpi.Abs p2) (AgentSpi.Conc n (EvalSpi.expr_of_mesg m) q2))).
    intros.
    repeat (rewrite AgentSpi.pseudo_appl_l_def).
    apply AROproc.
    apply PROmk_nu.
    apply PROpar;auto.
    apply pro_low_subst;auto.
    apply List_forall_cons.
    unfold evalc_succeeds.
    rewrite evalc_msg.
    exists m;auto.
    apply List_forall_nil.
  Qed.

  Lemma aro_pseudo_appl_r : forall (p1 q1 p2 q2:SpiCalcul.t)(n:nat)(m:MessageSpi.t),(pro p1 p2)->(pro q1 q2)->(aro (AgentSpi.pseudo_appl_r (AgentSpi.Conc n (EvalSpi.expr_of_mesg m) p1) (AgentSpi.Abs q1)) (AgentSpi.pseudo_appl_r (AgentSpi.Conc n (EvalSpi.expr_of_mesg m) p2) (AgentSpi.Abs q2))).
    intros.
    repeat (rewrite AgentSpi.pseudo_appl_r_def).
    apply AROproc.
    apply PROmk_nu.
    apply PROpar;auto.
    apply pro_low_subst;auto.
    apply List_forall_cons.
    unfold evalc_succeeds.
    rewrite evalc_msg.
    exists m;auto.
    apply List_forall_nil.
  Qed.

  (* tweak the module type system *)
  Inductive pro_aux : ActionSpiNC.A.P.t -> ActionSpiNC.A.P.t -> Prop :=
    | PROXzero : (pro_aux ActionSpiNC.A.P.Zero ActionSpiNC.A.P.Zero)
    | PROXpar : forall (p1 q1 p2 q2:ActionSpiNC.A.P.t),(pro_aux p1 q1)->(pro_aux p2 q2)->(pro_aux (ActionSpiNC.A.P.Par p1 p2) (ActionSpiNC.A.P.Par q1 q2))
    | PROXsum : forall (p1 q1 p2 q2:ActionSpiNC.A.P.t),(pro_aux p1 q1)->(pro_aux p2 q2)->(pro_aux (ActionSpiNC.A.P.Sum p1 p2) (ActionSpiNC.A.P.Sum q1 q2))
    | PROXbang : forall (p q:ActionSpiNC.A.P.t),(pro_aux p q)->(pro_aux (ActionSpiNC.A.P.Bang p) (ActionSpiNC.A.P.Bang q))
    | PROXnew : forall (p q:ActionSpiNC.A.P.t),(pro_aux p q)->(pro_aux (ActionSpiNC.A.P.New p) (ActionSpiNC.A.P.New q))
    | PROXtau : forall (p q:ActionSpiNC.A.P.t),(pro_aux p q)->(pro_aux (ActionSpiNC.A.P.Tau p) (ActionSpiNC.A.P.Tau q))
    | PROXinput : forall (e f:ExpressionSpi.t)(p q:ActionSpiNC.A.P.t),(ero e f)->(pro_aux p q)->(pro_aux (ActionSpiNC.A.P.Input e p) (ActionSpiNC.A.P.Input f q))
    | PROXoutput : forall (e1 f1 e2 f2:ExpressionSpi.t)(p q:ActionSpiNC.A.P.t),(ero e1 f1)->(ero e2 f2)->(pro_aux p q)->(pro_aux (ActionSpiNC.A.P.Output e1 e2 p) (ActionSpiNC.A.P.Output f1 f2 q))
    | PROXifthen : forall (phi psi:FormulaeSpi.t)(p q:ActionSpiNC.A.P.t),(fro phi psi)->(pro_aux p q)->(pro_aux (ActionSpiNC.A.P.IfThen phi p) (ActionSpiNC.A.P.IfThen psi q)).

  Lemma pro_impl_pro_aux : forall (p q:SpiCalcul.t),(pro p q)->(pro_aux p q).
    induction 1.
    apply PROXzero.
    apply PROXpar;auto.
    apply PROXsum;auto.
    apply PROXbang;auto.
    apply PROXnew;auto.
    apply PROXtau;auto.
    apply PROXinput;auto.
    apply PROXoutput;auto.
    apply PROXifthen;auto.
  Qed.

  Lemma pro_aux_impl_pro : forall (p q:SpiCalcul.t),(pro_aux p q)->(pro p q).
    induction 1.
    apply PROzero.
    apply PROpar;auto.
    apply PROsum;auto.
    apply PRObang;auto.
    apply PROnew;auto.
    apply PROtau;auto.
    apply PROinput;auto.
    apply PROoutput;auto.
    apply PROifthen;auto.
  Qed.

  Inductive aro_aux : ActionSpiNC.A.t -> ActionSpiNC.A.t -> Prop :=
    | AROXproc : forall (p q:ActionSpiNC.A.P.t),(pro_aux p q)->(aro_aux (ActionSpiNC.A.Pr p) (ActionSpiNC.A.Pr q))
    | AROXabs : forall (p q:ActionSpiNC.A.P.t),(pro_aux p q)->(aro_aux (ActionSpiNC.A.F (ActionSpiNC.A.Abs p)) (ActionSpiNC.A.F (ActionSpiNC.A.Abs q)))
    | AROXconc : forall (n:nat)(m:ExpressionSpi.t)(p q:ActionSpiNC.A.P.t),(pro_aux p q)->(aro_aux (ActionSpiNC.A.C (ActionSpiNC.A.Conc n m p)) (ActionSpiNC.A.C (ActionSpiNC.A.Conc n m q))).

  Lemma aro_impl_aro_aux : forall (a b:AgentSpi.t),(aro a b)->(aro_aux a b).
    induction 1.
    apply AROXproc.
    apply pro_impl_pro_aux;auto.
    apply AROXabs.
    apply pro_impl_pro_aux;auto.
    apply AROXconc.
    apply pro_impl_pro_aux;auto.
  Qed.

  Lemma reduction_impl_reduction_pro : forall (p:SpiCalcul.t)(mu:ActionSpiNC.t)(a:AgentSpi.t),(RedSpiNC.reduction p mu a)->forall (q:SpiCalcul.t),(pro p q)->exists b:AgentSpi.t,(aro a b)/\(RedSpiNC.reduction q mu b).
    intros p mu a H q.
    intro H0.
    generalize (pro_impl_pro_aux p q H0);clear H0.
    generalize q;clear q.
    generalize H;clear H.
    generalize a;clear a.
    generalize mu;clear mu.
    generalize p;clear p.
    induction 1;intros.
    (* RedSilent *)
    inversion H.
    exists (ActionSpiNC.A.Pr q0).
    split.
    apply AROproc.
    apply pro_aux_impl_pro;auto.
    apply RedSpiNC.RedSilent.
    (* RedInput *)
    inversion H0.
    exists (ActionSpiNC.A.F (ActionSpiNC.A.Abs q0)).
    split.
    apply AROabs.
    apply pro_aux_impl_pro;auto.
    apply RedSpiNC.RedInput.
    unfold ActionSpiNC.mk_input in H.
    unfold ActionSpiNC.mk_input.
    generalize (evalc_ero e f H3).
    destruct (evalc e);try (discriminate H).
    destruct t;try (discriminate H).
    intros.
    rewrite (H6 (MessageSpi.MNam t));auto.
    (* RedOutput *)
    inversion H0.
    subst e;subst f.
    exists (ActionSpiNC.A.C (ActionSpiNC.A.Conc 0 m q0)).
    split.
    apply AROconc.
    apply pro_aux_impl_pro;auto.
    apply RedSpiNC.RedOutput.
    unfold ActionSpiNC.mk_output in H.
    unfold ActionSpiNC.mk_output.
    generalize (evalc_ero e1 f1 H4).
    generalize (evalc_ero e2 f2 H6).
    intros.
    destruct (evalc e1);try (discriminate H).
    rewrite (H2 t);auto.
    destruct t;try (discriminate H).
    destruct (evalc e2);try (discriminate H).
    rewrite (H1 t0);auto.
    (* RedIfThen *)
    inversion H1.
    elim (IHreduction q0 H6).
    intro b;intros;and_elim.
    exists b.
    split;auto.
    eapply RedSpiNC.RedIfThen;eauto.
    subst phi0.
    generalize (eval_nc_fro phi psi H4).
    intros.
    unfold ActionSpiNC.mk_if in H0.
    unfold ActionSpiNC.mk_if.
    destruct mu.
    destruct (eval_nc phi).
    destruct b0;try (discriminate H0).
    rewrite (H2 l);auto.
    (* RedSumL *)
    inversion H0.
    elim (IHreduction q1);auto.
    intro b;intros;and_elim.
    exists b.
    split;auto.
    apply RedSpiNC.RedSumL;auto.
    (* RedSumR *)
    inversion H0.
    elim (IHreduction q2);auto.
    intro b;intros;and_elim.
    exists b.
    split;auto.
    apply RedSpiNC.RedSumR;auto.
    (* RedRes *)
    inversion H1.
    elim (IHreduction q0);auto.
    intro b;intros;and_elim.
    exists (ActionSpiNC.A.new b).
    split.
    apply aro_new;auto.
    eapply RedSpiNC.RedRes;eauto.
    (* RedParL *)
    inversion H0.
    elim (IHreduction q1);auto.
    intro b;intros;and_elim.
    exists (ActionSpiNC.A.par_r b q2).
    split.
    apply aro_par_r;auto.
    apply pro_aux_impl_pro;auto.
    apply RedSpiNC.RedParL;auto.
    (* RedParR *)
    inversion H0.
    elim (IHreduction q2);auto.
    intro b;intros;and_elim.
    exists (ActionSpiNC.A.par_l q1 b).
    split.
    apply aro_par_l;auto.
    apply pro_aux_impl_pro;auto.
    apply RedSpiNC.RedParR;auto.
    (* RedCloseL *)
    inversion H2.
    subst p;subst q.
    elim (IHreduction1 q1);auto.
    intro f1;intros;and_elim.
    elim (IHreduction2 q2);auto.
    intro c1;intros;and_elim.
    generalize (aro_impl_aro_aux (ActionSpiNC.A.F f) f1 H3).
    generalize (aro_impl_aro_aux (ActionSpiNC.A.C c) c1 H8).
    intros.
    inversion H10.
    inversion H11.
    subst c1;subst f1.
    exists (ActionSpiNC.A.pseudo_appl_l (ActionSpiNC.A.Abs q3) (ActionSpiNC.A.Conc n m q)). 
    split.
    subst c.
    elim (reduction_spi_nc_conc_message p2 out (ActionSpiNC.A.C (ActionSpiNC.A.Conc n m p)) H0 n m p (refl_equal (ActionSpiNC.A.C (ActionSpiNC.A.Conc n m p)))).
    intro x;intros.
    subst m.
    apply aro_pseudo_appl_l.
    apply pro_aux_impl_pro;auto.
    apply pro_aux_impl_pro;auto.
    eapply RedSpiNC.RedCloseL;eauto.
    (* RedCloseR *)
    inversion H2.
    subst p;subst q.
    elim (IHreduction1 q1);auto.
    intro c1;intros;and_elim.
    elim (IHreduction2 q2);auto.
    intro f1;intros;and_elim.
    generalize (aro_impl_aro_aux (ActionSpiNC.A.F f) f1 H8).
    generalize (aro_impl_aro_aux (ActionSpiNC.A.C c) c1 H3).
    intros.
    inversion H10.
    inversion H11.
    subst c1;subst f1.
    exists (ActionSpiNC.A.pseudo_appl_r (ActionSpiNC.A.Conc n m q) (ActionSpiNC.A.Abs q3)). 
    split.
    subst c.
    elim (reduction_spi_nc_conc_message p1 out (ActionSpiNC.A.C (ActionSpiNC.A.Conc n m p)) H n m p (refl_equal (ActionSpiNC.A.C (ActionSpiNC.A.Conc n m p)))).
    intro x;intros.
    subst m.
    apply aro_pseudo_appl_r.
    apply pro_aux_impl_pro;auto.
    apply pro_aux_impl_pro;auto.
    eapply RedSpiNC.RedCloseR;eauto.
    (* RedBangAct *)
    inversion H0.
    elim (IHreduction q0);auto.
    intro b;intros;and_elim.
    exists (ActionSpiNC.A.par_r b (ActionSpiNC.A.P.Bang q0)).
    split.
    apply aro_par_r;auto.
    apply PRObang.
    apply pro_aux_impl_pro;auto.
    apply RedSpiNC.RedBangAct;auto.
    (* RedBangCloseL *)
    inversion H2.
    subst p;subst q.
    elim (IHreduction1 q0);auto.
    intro f1;intros;and_elim.
    elim (IHreduction2 q0);auto.
    intro c1;intros;and_elim.
    generalize (aro_impl_aro_aux (ActionSpiNC.A.F f) f1 H3).
    generalize (aro_impl_aro_aux (ActionSpiNC.A.C c) c1 H6).
    intros.
    inversion H8.
    inversion H9.
    subst c1;subst f1.
    exists (ActionSpiNC.A.par_r (ActionSpiNC.A.pseudo_appl_l (ActionSpiNC.A.Abs q1) (ActionSpiNC.A.Conc n m q)) (ActionSpiNC.A.P.Bang q0)). 
    split.
    subst c.
    elim (reduction_spi_nc_conc_message p0 out (ActionSpiNC.A.C (ActionSpiNC.A.Conc n m p)) H0 n m p (refl_equal (ActionSpiNC.A.C (ActionSpiNC.A.Conc n m p)))).
    intro x;intros.
    subst m.
    apply aro_par_r.
    apply aro_pseudo_appl_l.
    apply pro_aux_impl_pro;auto.
    apply pro_aux_impl_pro;auto.
    apply PRObang.
    apply pro_aux_impl_pro;auto.
    eapply RedSpiNC.RedBangCloseL;eauto.
  Qed.
  
  Lemma reduction_pro_impl_reduction : forall (q:SpiCalcul.t)(mu:ActionSpiNC.t)(b:AgentSpi.t),(RedSpiNC.reduction q mu b)->forall (p:SpiCalcul.t),(pro p q)->exists a:AgentSpi.t,(aro a b)/\(RedSpiNC.reduction p mu a).
    intros q mu b H p.
    intro H0.
    generalize (pro_impl_pro_aux p q H0);clear H0.
    generalize p;clear p.
    generalize H;clear H.
    generalize b;clear b.
    generalize mu;clear mu.
    generalize q;clear q.
    induction 1;intros.
    (* RedSilent *)
    inversion H.
    exists (ActionSpiNC.A.Pr p1).
    split.
    apply AROproc.
    apply pro_aux_impl_pro;auto.
    apply RedSpiNC.RedSilent.
    (* RedInput *)
    inversion H0.
    subst f;subst q.
    exists (ActionSpiNC.A.F (ActionSpiNC.A.Abs p1)).
    split.
    apply AROabs.
    apply pro_aux_impl_pro;auto.
    apply RedSpiNC.RedInput.
    unfold ActionSpiNC.mk_input in H.
    unfold ActionSpiNC.mk_input.
    generalize (ero_evalc e0 e H4);intro.
    destruct (evalc e);try (discriminate H).
    rewrite (H1 t);auto.
    (* RedOutput *)
    inversion H0.
    subst f1;subst f2;subst q.
    exists (ActionSpiNC.A.C (ActionSpiNC.A.Conc 0 m p1)).
    split.
    apply AROconc.
    apply pro_aux_impl_pro;auto.
    apply RedSpiNC.RedOutput.
    unfold ActionSpiNC.mk_output in H.
    unfold ActionSpiNC.mk_output.
    generalize (ero_evalc e1 e H5);intro.
    generalize (ero_evalc e2 f H6);intro.
    destruct (evalc e);try (discriminate H).
    rewrite (H1 t);auto.
    destruct t;try (discriminate H).
    destruct (evalc f);try (discriminate H).
    rewrite (H2 t0);auto.
    (* RedIfThen *)
    inversion H1.
    subst phi;subst q.
    elim (IHreduction p1 H6);intro a0;intros;and_elim.
    exists a0.
    split;auto.
    eapply RedSpiNC.RedIfThen;eauto.
    unfold ActionSpiNC.mk_if in H0.
    unfold ActionSpiNC.mk_if.
    generalize (fro_eval_nc phi0 psi H5);intro.
    destruct mu.
    destruct (eval_nc psi).
    destruct b;try (discriminate H0).
    rewrite (H7 l);auto.
    (* RedSumL *)
    inversion H0.
    subst p;subst q.
    elim (IHreduction p1);auto.
    intro a0;intros;and_elim.
    exists a0.
    split;auto.
    apply RedSpiNC.RedSumL;auto.
    (* RedSumR *)
    inversion H0.
    subst p;subst q.
    elim (IHreduction p2);auto.
    intro a0;intros;and_elim.
    exists a0.
    split;auto.
    apply RedSpiNC.RedSumR;auto.
    (* RedRes *)
    inversion H1.
    subst q;subst p0.
    elim (IHreduction p1);auto.
    intro a0;intros;and_elim.
    exists (ActionSpiNC.A.new a0).
    split.
    apply aro_new;auto.
    eapply RedSpiNC.RedRes;eauto.
    (* RedParL *)
    inversion H0.
    subst p;subst q.
    elim (IHreduction p1);auto.
    intro a0;intros;and_elim.
    exists (ActionSpiNC.A.par_r a0 p2).
    split.
    apply aro_par_r;auto.
    apply pro_aux_impl_pro;auto.
    apply RedSpiNC.RedParL;auto.
    (* RedParR *)
    inversion H0.
    subst p;subst q.
    elim (IHreduction p2);auto.
    intro a0;intros;and_elim.
    exists (ActionSpiNC.A.par_l p1 a0).
    split.
    apply aro_par_l;auto.
    apply pro_aux_impl_pro;auto.
    apply RedSpiNC.RedParR;auto.
    (* RedCloseL *)
    inversion H2.
    subst p;subst q.
    elim (IHreduction1 p1);auto.
    intro f1;intros;and_elim.
    elim (IHreduction2 p2);auto.
    intro c1;intros;and_elim.
    generalize (aro_impl_aro_aux f1 (ActionSpiNC.A.F f) H3).
    generalize (aro_impl_aro_aux c1 (ActionSpiNC.A.C c) H8).
    intros.
    inversion H10.
    inversion H11.
    subst c1;subst f1.
    exists (ActionSpiNC.A.pseudo_appl_l (ActionSpiNC.A.Abs p3) (ActionSpiNC.A.Conc n m p)). 
    split.
    subst c.
    elim (reduction_spi_nc_conc_message q2 out (ActionSpiNC.A.C (ActionSpiNC.A.Conc n m q)) H0 n m q (refl_equal (ActionSpiNC.A.C (ActionSpiNC.A.Conc n m q)))).
    intro x;intros.
    subst m.
    apply aro_pseudo_appl_l.
    apply pro_aux_impl_pro;auto.
    apply pro_aux_impl_pro;auto.
    eapply RedSpiNC.RedCloseL;eauto.
    (* RedCloseR *)
    inversion H2.
    subst p;subst q.
    elim (IHreduction1 p1);auto.
    intro c1;intros;and_elim.
    elim (IHreduction2 p2);auto.
    intro f1;intros;and_elim.
    generalize (aro_impl_aro_aux f1 (ActionSpiNC.A.F f) H8).
    generalize (aro_impl_aro_aux c1 (ActionSpiNC.A.C c) H3).
    intros.
    inversion H10.
    inversion H11.
    subst c1;subst f1.
    exists (ActionSpiNC.A.pseudo_appl_r (ActionSpiNC.A.Conc n m p) (ActionSpiNC.A.Abs p3)). 
    split.
    subst c.
    elim (reduction_spi_nc_conc_message q1 out (ActionSpiNC.A.C (ActionSpiNC.A.Conc n m q)) H n m q (refl_equal (ActionSpiNC.A.C (ActionSpiNC.A.Conc n m q)))).
    intro x;intros.
    subst m.
    apply aro_pseudo_appl_r.
    apply pro_aux_impl_pro;auto.
    apply pro_aux_impl_pro;auto.
    eapply RedSpiNC.RedCloseR;eauto.
    (* RedBangAct *)
    inversion H0.
    subst q.
    elim (IHreduction p1);auto.
    intro b;intros;and_elim.
    exists (ActionSpiNC.A.par_r b (ActionSpiNC.A.P.Bang p1)).
    split.
    apply aro_par_r;auto.
    apply PRObang.
    apply pro_aux_impl_pro;auto.
    apply RedSpiNC.RedBangAct;auto.
    (* RedBangCloseL *)
    inversion H2.
    subst p;subst p0.
    elim (IHreduction1 p1);auto.
    intro f1;intros;and_elim.
    elim (IHreduction2 p1);auto.
    intro c1;intros;and_elim.
    generalize (aro_impl_aro_aux f1 (ActionSpiNC.A.F f) H3).
    generalize (aro_impl_aro_aux c1 (ActionSpiNC.A.C c) H6).
    intros.
    inversion H8.
    inversion H9.
    subst c1;subst f1.
    exists (ActionSpiNC.A.par_r (ActionSpiNC.A.pseudo_appl_l (ActionSpiNC.A.Abs p0) (ActionSpiNC.A.Conc n m p)) (ActionSpiNC.A.P.Bang p1)). 
    split.
    subst c.
    elim (reduction_spi_nc_conc_message q out (ActionSpiNC.A.C (ActionSpiNC.A.Conc n m q0)) H0 n m q0 (refl_equal (ActionSpiNC.A.C (ActionSpiNC.A.Conc n m q0)))).
    intro x;intros.
    subst m.
    apply aro_par_r.
    apply aro_pseudo_appl_l.
    apply pro_aux_impl_pro;auto.
    apply pro_aux_impl_pro;auto.
    apply PRObang.
    apply pro_aux_impl_pro;auto.
    eapply RedSpiNC.RedBangCloseL;eauto.
  Qed.
End RoSpi.
