(* A formalisation of the spi calculus in Coq *)
(* Copyright (C) 2005-2007 Sebastien Briais *)
(* http://lamp.epfl.ch/~sbriais/ *)

(* This library is free software; you can redistribute it and/or modify *)
(* it under the terms of the GNU Lesser General Public License as *)
(* published by the Free Software Foundation; either version 2.1 of the *)
(* License, or (at your option) any later version. *)

(* This library is distributed in the hope that it will be useful, but *)
(* WITHOUT ANY WARRANTY; without even the implied warranty of *)
(* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU *)
(* Lesser General Public License for more details. *)

(* You should have received a copy of the GNU Lesser General Public *)
(* License along with this library; if not, write to the Free Software *)
(* Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA *)
(* 02110-1301 USA *)

Require Import Arith.
Require Import List.
Require Import Bool.
Require Import missing.
Require Export deBruijnNat.

Module deBruijn.
  Record deBruijn_base (t:Set) : Set := { 
    lift : nat -> nat -> t -> t;
    swapR : nat -> nat -> t -> t;
    low1 : nat -> t -> t;
    (* lift properties *)
    lift_injective : forall (d k:nat),(injective (lift d k));
    lift_0 : forall (d:nat)(x:t),(lift d 0 x) = x;
    lift_plus : forall (d:nat)(k k':nat)(x:t),(lift d k (lift d k' x)) = (lift d (k+k') x);
    lift_comm : forall (d d' k k':nat)(x:t),(lift (d+d'+k) k' (lift d k x))=(lift d k (lift (d+d') k' x));
    lift_compose : forall (d d' k k':nat)(x:t),(lift (d+d') k (lift d (d'+k') x))=(lift d (d'+k+k') x);
    (* swapR properties *)
    swapR_injective : forall (d k:nat),(injective (swapR d k));
    swapR_0 : forall (d:nat)(x:t),(swapR d 0 x) = x;
    swapR_comm : forall (d d' k k':nat)(x:t),(swapR (d+k+(S d')) k' (swapR d k x))=(swapR d k (swapR (d+k+(S d')) k' x));
    swapR_1_idem : forall (d:nat)(x:t),(swapR d 1 (swapR d 1 x))=x;
    swapR_simpl : forall (d k k':nat)(x:t),(swapR d k (swapR (d+k) k' x))=(swapR d (k+k') x);
    swapR_compose : forall (d k k':nat)(x:t),(swapR d (k+(S k')) (swapR d k x))=(swapR (S d) k (swapR d (k+(S k')) x));
    swapR_inside : forall (d d' k:nat)(x:t),(swapR d (d'+S k) (swapR (d+d') k x))=(swapR (d+S d') k (swapR d (d'+S k) x));
    (* lift and swapR interaction *)
    lift_swapR_comm : forall (d d' k k':nat)(x:t),(lift (d+k+(S d')) k' (swapR d k x))=(swapR d k (lift (d+k+(S d')) k' x));
    swapR_lift_comm : forall (d d' k k':nat)(x:t),(swapR (d+k+d') k' (lift d k x))=(lift d k (swapR (d+d') k' x));
    swapR_lift_idem : forall (d d' k k':nat)(x:t),(swapR (d+d') k (lift d (d'+k+(S k')) x))=(lift d (d'+k+(S k')) x);
    swapR_lift_simpl : forall (d k:nat)(x:t),(swapR d k (lift d k x))=(lift (S d) k x);
    swapR_lift_1 : forall (d k:nat)(x:t),(swapR d k (lift (d+k) 1 x))=(lift d 1 x);
    (* low1 and lift *)
    low1_lift_lift :  forall (d d' k:nat)(x:t),(low1 (d+d') (lift d (d'+S k) x))=(lift d (d'+k) x);
    (* swapR and occur *)
    swapR_occur_swap : forall (k:nat)(n1 n2:nat),(n1<=k)->(n2<=S k)->forall (n'1 n'2:nat),(DeBruijnNat.occur_swap n1 n2 = (n'1,n'2))->forall (d:nat)(x:t),(swapR (d+n2) (S k-n2) (swapR (d+n1) (k-n1) x))=(swapR (d+n'2) (S k-n'2) (swapR (d+n'1) (S (k-n'1)) x))
  }.
  
  Implicit Arguments lift [t].
  Implicit Arguments swapR [t].
  Implicit Arguments low1 [t].
  Implicit Arguments lift_injective [t].
  Implicit Arguments lift_0 [t].
  Implicit Arguments lift_plus [t].
  Implicit Arguments lift_comm [t].
  Implicit Arguments lift_compose [t].
  Implicit Arguments swapR_injective [t].
  Implicit Arguments swapR_0 [t].
  Implicit Arguments swapR_comm [t].
  Implicit Arguments swapR_1_idem [t].
  Implicit Arguments swapR_simpl [t].
  Implicit Arguments swapR_compose [t].
  Implicit Arguments swapR_inside [t].
  Implicit Arguments lift_swapR_comm [t].
  Implicit Arguments swapR_lift_comm [t].
  Implicit Arguments swapR_lift_idem [t].
  Implicit Arguments swapR_lift_simpl [t].
  Implicit Arguments swapR_lift_1 [t].
  Implicit Arguments low1_lift_lift [t].
  Implicit Arguments swapR_occur_swap [t].

  Lemma low1_lift_simpl (t:Set)(deBruijn_t:deBruijn_base t) : forall (d k:nat)(x:t),(low1 deBruijn_t d (lift deBruijn_t d (S k) x))=(lift deBruijn_t d k x).
    intros.
    generalize (low1_lift_lift deBruijn_t d 0 k x);simpl;intros.
    rewrite plus_0_r in H.
    auto.
  Qed.

  Implicit Arguments low1_lift_simpl [t].

  Fixpoint swapL (t:Set)(deBruijn_t:deBruijn_base t)(d k:nat)(x:t) {struct k} : t :=
    match k with
      | 0 => x
      | (S k) => swapR deBruijn_t (d+k) 1 (swapL t deBruijn_t d k x)
    end.

  Implicit Arguments swapL [t].

  Lemma swapL_0 : forall (t:Set)(deBruijn_t:deBruijn_base t)(d:nat)(x:t),(swapL deBruijn_t d 0 x)=x.
    intros.
    reflexivity.
  Qed.

  Lemma swapR_swapL_simpl : forall (t:Set)(deBruijn_t:deBruijn_base t)(d k:nat)(x:t),(swapR deBruijn_t (d+k) 1 (swapL deBruijn_t d k x))=(swapL deBruijn_t d (S k) x).
    intros.
    reflexivity.
  Qed.

  Lemma swapL_swapR_1 : forall (t:Set)(deBruijn_t:deBruijn_base t)(d:nat)(x:t),(swapL deBruijn_t d 1 x)=(swapR deBruijn_t d 1 x).
    intros.
    simpl.
    rewrite plus_0_r;reflexivity.
  Qed.

  Lemma swapL_swapR_id : forall (t:Set)(deBruijn_t:deBruijn_base t)(d k:nat)(x:t),(swapL deBruijn_t d k (swapR deBruijn_t d k x))=x.
    induction k;intros;simpl.
    apply swapR_0.
    generalize (swapR_simpl deBruijn_t d k 1).
    rewrite <- plus_n_Sm;rewrite plus_0_r;simpl;intro.
    rewrite <- H.
    rewrite IHk.
    apply swapR_1_idem.
  Qed.

  Lemma swapR_swapL_id : forall (t:Set)(deBruijn_t:deBruijn_base t)(d k:nat)(x:t),(swapR deBruijn_t d k (swapL deBruijn_t d k x))=x.
    induction k;intros;simpl.
    apply swapR_0.
    generalize (swapR_simpl deBruijn_t d k 1).
    rewrite <- plus_n_Sm;rewrite plus_0_r;simpl;intro.
    rewrite <- H.
    rewrite swapR_1_idem.
    apply IHk.
  Qed.

  Implicit Arguments swapL_0 [t].
  Implicit Arguments swapR_swapL_simpl [t].
  Implicit Arguments swapL_swapR_1 [t].
  Implicit Arguments swapL_swapR_id [t].
  Implicit Arguments swapR_swapL_id [t].
  
  Record deBruijn_extension (t:Set)(deBruijn_t:deBruijn_base t)(t':Set)(deBruijn_t':deBruijn_base t') : Set := {
    j : t -> t';
    j_injective : injective j;
    j_lift : forall (d k:nat)(x:t),(lift deBruijn_t' d k (j  x))=(j (lift deBruijn_t d k x));
    j_swapR : forall (d k:nat)(x:t),(swapR deBruijn_t' d k (j x))=(j (swapR deBruijn_t d k x));
    j_low1 : forall (d:nat)(x:t),(low1 deBruijn_t' d (j x))=(j (low1 deBruijn_t d x))
  }.

  Implicit Arguments j [t deBruijn_t t' deBruijn_t'].
  Implicit Arguments j_injective [t deBruijn_t t' deBruijn_t'].
  Implicit Arguments j_lift [t deBruijn_t t' deBruijn_t'].
  Implicit Arguments j_swapR [t deBruijn_t t' deBruijn_t'].
  Implicit Arguments j_low1 [t deBruijn_t t' deBruijn_t'].

  Definition deBruijn_id_extension (t:Set)(deBruijn_t:deBruijn_base t) : (deBruijn_extension t deBruijn_t t deBruijn_t).
    intros.
    split with (fun x:t => x);auto.
    unfold injective;auto.
  Defined.

  Record map_struct (t t':Set)(map_fun:(nat->t->t)->nat->t'->t') : Prop := {
    map_id : forall (d:nat)(x:t'),(map_fun (fun d x => x) d x)=x;
    map_injective : forall (f:nat->t->t),(forall (d:nat),(injective (f d)))->forall (d:nat),(injective (map_fun f d));
    map_compose: forall (d1 d2 d3 d4:nat) (f1 g1 f2 g2:nat->t->t),(forall (d:nat)(x:t),(f1 (d1+d) (g1 (d2+d) x))=(f2 (d3+d) (g2 (d4+d) x)))->forall (d:nat)(x:t'),(map_fun f1 (d1+d) (map_fun g1 (d2+d) x))=(map_fun f2 (d3+d) (map_fun g2  (d4+d) x));
    map_equal : forall (d1 d2:nat)(f g:nat->t->t),(forall (d:nat)(x:t),(f (d1+d) x)=(g (d2+d) x))->forall (d:nat)(x:t'),(map_fun f (d1+d) x)=(map_fun g (d2+d) x)
  }.

  Implicit Arguments map_id [t t' map_fun d].
  Implicit Arguments map_injective [t t' map_fun].
  Implicit Arguments map_equal [t t' map_fun].
  Implicit Arguments map_compose [t t' map_fun].

  Lemma map_compose_simpl : forall (t t':Set)(map_fun:(nat->t->t)->nat->t'->t')(map_ext:map_struct t t' map_fun),forall (d1 d2 d3:nat)(f g h:nat->t->t),(forall (d:nat)(x:t),(f (d1+d) (g (d2+d) x))=(h (d3+d) x))->forall (d:nat)(x:t'),(map_fun f (d1+d) (map_fun g (d2+d) x))=(map_fun h (d3+d) x).
    intros.
    generalize (map_compose map_ext d1 d2 d3 0 f g h (fun d x => x) H d x);intros.
    simpl in H0.
    rewrite (map_id map_ext (d:=d) x) in H0.
    trivial.
  Qed.
    
  Implicit Arguments map_compose_simpl [t t' map_fun].

  Definition deBruijn_map_struct (t:Set)(deBruijn_t:deBruijn_base t)(t':Set)(map_fun:(nat->t->t)->nat->t'->t')(map_ext:map_struct t t' map_fun) : (deBruijn_base t').
    intros.
    split with (fun d k => map_fun (fun d => (lift deBruijn_t d k)) d) (fun d k => map_fun (fun d => (swapR deBruijn_t d k)) d) (map_fun (fun d => low1 deBruijn_t d));intros.
    (* lift_injective *)
    apply (map_injective map_ext);intros.
    apply lift_injective.
    (* lift_0 *)
    generalize (map_equal map_ext 0 0 (fun d => (lift deBruijn_t d 0)) (fun d x => x));simpl;intros.
    rewrite H;try (intros;apply lift_0).
    apply (map_id map_ext).
    (* lift_plus *)
    generalize (map_compose_simpl map_ext 0 0 0 (fun d => lift deBruijn_t d k) (fun d => lift deBruijn_t d k') (fun d => lift deBruijn_t d (k+k')));simpl;intros.
    apply H;intros;apply lift_plus.
    (* lift_comm *)
    generalize (map_compose map_ext (d'+k) 0 0 d' (fun d => lift deBruijn_t d k') (fun d => lift deBruijn_t d k) (fun d => lift deBruijn_t d k) (fun d => lift deBruijn_t d k'));simpl;intros.
    rewrite <- plus_assoc;repeat (rewrite (plus_comm d)).
    apply H;intros;rewrite (plus_comm (d'+k));rewrite plus_assoc;rewrite (plus_comm d' d0);apply lift_comm.
    (* lift_compose *)
    generalize (map_compose_simpl map_ext d' 0 0 (fun d => lift deBruijn_t d k) (fun d => lift deBruijn_t d (d'+k')) (fun d => lift deBruijn_t d (d'+k+k')));simpl;intros.
    rewrite (plus_comm d).
    apply H;intros;rewrite (plus_comm d' d0);apply lift_compose.
    (* swapR_injective *)
    apply (map_injective map_ext);intros.
    apply swapR_injective.
    (* swapR_0 *)
    generalize (map_equal map_ext 0 0 (fun d => (swapR deBruijn_t d 0)) (fun d x => x));simpl;intros.
    rewrite H;try (intros;apply swapR_0).
    apply (map_id map_ext).
    (* swapR_comm *)
    generalize (map_compose map_ext (k+S d') 0 0 (k+S d') (fun d => swapR deBruijn_t d k') (fun d => swapR deBruijn_t d k) (fun d => swapR deBruijn_t d k) (fun d => swapR deBruijn_t d k'));simpl;intros.
    repeat (rewrite <- plus_assoc);repeat (rewrite (plus_comm d)).
    apply H;intros;repeat (rewrite (plus_comm (k+S d')));repeat (rewrite plus_assoc);apply swapR_comm.
    (* swapR_1_idem *)
    generalize (map_compose_simpl map_ext 0 0 0 (fun d => swapR deBruijn_t d 1) (fun d => (swapR deBruijn_t d 1)) (fun d x => x));simpl;intros.
    rewrite H;try (intros;apply swapR_1_idem).
    apply (map_id map_ext).
    (* swapR_simpl *)
    generalize (map_compose_simpl map_ext 0 k 0 (fun d => swapR deBruijn_t d k) (fun d => swapR deBruijn_t d k') (fun d => swapR deBruijn_t d (k+k')));simpl;intros.
    rewrite (plus_comm d).
    apply H;intros;rewrite (plus_comm k d0);apply swapR_simpl.
    (* swapR_compose *)
    generalize (map_compose map_ext 0 0 1 0 (fun d => swapR deBruijn_t d (k+S k')) (fun d => swapR deBruijn_t d k) (fun d => swapR deBruijn_t d k) (fun d => swapR deBruijn_t d (k+S k')));simpl;intros.
    apply H;intros;apply swapR_compose.
    (* swapR_inside *)
    generalize (map_compose map_ext 0 d' (S d') 0 (fun d => swapR deBruijn_t d (d'+S k)) (fun d => swapR deBruijn_t d k) (fun d => swapR deBruijn_t d k) (fun d => swapR deBruijn_t d (d'+S k)));simpl;intros.
    rewrite (plus_comm d d').
    rewrite (plus_comm d (S d'));simpl.
    apply H;intros;rewrite (plus_comm d' d0);rewrite plus_n_Sm;apply swapR_inside.
    (* lift_swapR_comm *)
    generalize (map_compose map_ext (k+S d') 0 0 (k+S d') (fun d => lift deBruijn_t d k') (fun d => swapR deBruijn_t d k) (fun d => swapR deBruijn_t d k) (fun d => lift deBruijn_t d k'));simpl;intros.
    repeat (rewrite <- plus_assoc);repeat (rewrite (plus_comm d)).
    apply H;intros;repeat (rewrite (plus_comm (k+S d')));repeat (rewrite plus_assoc);apply lift_swapR_comm.
    (* swapR_lift_comm *)
    generalize (map_compose map_ext (k+d') 0 0 d' (fun d => swapR deBruijn_t d k') (fun d => lift deBruijn_t d k) (fun d => lift deBruijn_t d k) (fun d => swapR deBruijn_t d k'));simpl;intros.
    repeat (rewrite <- plus_assoc);repeat (rewrite (plus_comm d)).
    apply H;intros;repeat (rewrite (plus_comm (k+d')));repeat (rewrite plus_assoc);rewrite (plus_comm d' d0);apply swapR_lift_comm.
    (* swapR_lift_idem *)
    generalize (map_compose_simpl map_ext d' 0 0 (fun d => swapR deBruijn_t d k) (fun d => lift deBruijn_t d (d'+k+S k')) (fun d => lift deBruijn_t d (d'+k+S k')));simpl;intros.
    rewrite (plus_comm d).
    apply H;intros;rewrite (plus_comm d' d0);apply swapR_lift_idem.
    (* swapR_lift_simpl *)
    generalize (map_compose_simpl map_ext 0 0 1 (fun d => swapR deBruijn_t d k) (fun d => lift deBruijn_t d k) (fun d => lift deBruijn_t d k));simpl;intros.
    apply H;intros;apply swapR_lift_simpl.
    (* swapR_lift_1 *)
    generalize (map_compose_simpl map_ext 0 k 0 (fun d => swapR deBruijn_t d k) (fun d => lift deBruijn_t d 1) (fun d => lift deBruijn_t d 1));simpl;intros.
    rewrite plus_comm.
    apply H;intros;rewrite plus_comm;apply swapR_lift_1.
    (* low1_lift_lift *)
    generalize (map_compose_simpl map_ext d' 0 0 (fun d => low1 deBruijn_t d) (fun d => lift deBruijn_t d (d'+S k)) (fun d => lift deBruijn_t d (d'+k)));simpl;intros.
    rewrite plus_comm;apply H;intros.
    rewrite plus_comm;apply low1_lift_lift.
    (* swapR_occur_swap *)
    generalize (map_compose map_ext n2 n1 n'2 n'1 (fun d => swapR deBruijn_t d (S k - n2)) (fun d => swapR deBruijn_t d (k-n1)) (fun d => swapR deBruijn_t d (S k-n'2)) (fun d => swapR deBruijn_t d (S (k-n'1))));intros.
    repeat (rewrite (plus_comm d)).
    apply H2;intros;rewrite (plus_comm n2);rewrite (plus_comm n1);rewrite (plus_comm n'2);rewrite (plus_comm n'1);apply swapR_occur_swap;auto.
  Defined.

  Record map2_struct (t1 t2 t:Set)(map2_fun:(nat->t1->t1)->(nat->t2->t2)->nat->t->t) : Prop := {
    map2_id : forall (d:nat)(x:t),(map2_fun (fun d x => x) (fun d x => x) d x)=x;
    map2_injective : forall (f1:nat->t1->t1)(f2:nat->t2->t2),(forall (d:nat),(injective (f1 d)))->(forall (d:nat),(injective (f2 d)))->forall (d:nat),(injective (map2_fun f1 f2 d));
    map2_compose: forall (d1 d2 d3 d4:nat)(f1 g1 f2 g2:nat->t1->t1)(f1' g1' f2' g2':nat->t2->t2),(forall (d:nat)(x:t1),(f1 (d1+d) (g1 (d2+d) x))=(f2 (d3+d) (g2 (d4+d) x)))->(forall (d:nat)(x:t2),(f1' (d1+d) (g1' (d2+d) x))=(f2' (d3+d) (g2' (d4+d) x)))->forall (d:nat)(x:t),(map2_fun f1 f1' (d1+d) (map2_fun g1 g1' (d2+d) x))=(map2_fun f2 f2' (d3+d) (map2_fun g2 g2' (d4+d) x));
    map2_equal : forall (d1 d2:nat)(f g:nat->t1->t1)(f' g':nat->t2->t2),(forall (d:nat)(x:t1),(f (d1+d) x)=(g (d2+d) x))->(forall (d:nat)(x:t2),(f' (d1+d) x)=(g' (d2+d) x))->forall (d:nat)(x:t),(map2_fun f f' (d1+d) x)=(map2_fun g g' (d2+d) x)
  }.

  Implicit Arguments map2_id [t1 t2 t map2_fun d].
  Implicit Arguments map2_injective [t1 t2 t map2_fun].
  Implicit Arguments map2_equal [t1 t2 t map2_fun].
  Implicit Arguments map2_compose [t1 t2 t map2_fun].

  Lemma map2_compose_simpl : forall (t1 t2 t:Set)(map2_fun:(nat->t1->t1)->(nat->t2->t2)->nat->t->t)(map2_ext:map2_struct t1 t2 t map2_fun),forall (d1 d2 d3:nat)(f g h:nat->t1->t1)(f' g' h':nat->t2->t2),(forall (d:nat)(x:t1),(f (d1+d) (g (d2+d) x))=(h (d3+d) x))->(forall (d:nat)(x:t2),(f' (d1+d) (g' (d2+d) x))=(h' (d3+d) x))->forall (d:nat)(x:t),(map2_fun f f' (d1+d) (map2_fun g g' (d2+d) x))=(map2_fun h h' (d3+d) x).
    intros.
    generalize (map2_compose map2_ext d1 d2 d3 0 f g h (fun d x => x) f' g' h' (fun d x => x) H H0 d x);intro.
    rewrite (map2_id map2_ext) in H1.
    trivial.
  Qed.

  Implicit Arguments map2_compose_simpl [t1 t2 t map2_fun].

  Definition deBruijn_map2_struct (t1:Set)(deBruijn_t1:deBruijn_base t1)(t2:Set)(deBruijn_t2:deBruijn_base t2)(t:Set)(map2_fun:(nat->t1->t1)->(nat->t2->t2)->nat->t->t)(map2_ext:map2_struct t1 t2 t map2_fun) : (deBruijn_base t).
    intros.
    split with (fun d k => map2_fun (fun d => (lift deBruijn_t1 d k)) (fun d => (lift deBruijn_t2 d k)) d) (fun d k => map2_fun (fun d => (swapR deBruijn_t1 d k)) (fun d => (swapR deBruijn_t2 d k)) d) (map2_fun (fun d => low1 deBruijn_t1 d) (fun d => low1 deBruijn_t2 d));intros.
    (* lift_injective *)
    apply (map2_injective map2_ext);intros;apply lift_injective.
    (* lift_0 *)
    generalize (map2_equal map2_ext 0 0 (fun d => (lift deBruijn_t1 d 0)) (fun d x => x) (fun d => (lift deBruijn_t2 d 0)) (fun d x => x));simpl;intros.
    rewrite H;try (intros;apply lift_0).
    apply (map2_id map2_ext).
    (* lift_plus *)
    generalize (map2_compose_simpl map2_ext 0 0 0 (fun d => lift deBruijn_t1 d k) (fun d => lift deBruijn_t1 d k') (fun d => lift deBruijn_t1 d (k+k')) (fun d => lift deBruijn_t2 d k) (fun d => lift deBruijn_t2 d k') (fun d => lift deBruijn_t2 d (k+k')));simpl;intros.
    apply H;intros;apply lift_plus.
    (* lift_comm *)
    generalize (map2_compose map2_ext (d'+k) 0 0 d' (fun d => lift deBruijn_t1 d k') (fun d => lift deBruijn_t1 d k) (fun d => lift deBruijn_t1 d k) (fun d => lift deBruijn_t1 d k') (fun d => lift deBruijn_t2 d k') (fun d => lift deBruijn_t2 d k) (fun d => lift deBruijn_t2 d k) (fun d => lift deBruijn_t2 d k'));simpl;intros.
    rewrite <- plus_assoc;repeat (rewrite (plus_comm d)).
    apply H;intros;rewrite (plus_comm (d'+k));rewrite plus_assoc;rewrite (plus_comm d' d0);apply lift_comm.
    (* lift_compose *)
    generalize (map2_compose_simpl map2_ext d' 0 0 (fun d => lift deBruijn_t1 d k) (fun d => lift deBruijn_t1 d (d'+k')) (fun d => lift deBruijn_t1 d (d'+k+k')) (fun d => lift deBruijn_t2 d k) (fun d => lift deBruijn_t2 d (d'+k')) (fun d => lift deBruijn_t2 d (d'+k+k')));simpl;intros.
    rewrite (plus_comm d).
    apply H;intros;rewrite (plus_comm d' d0);apply lift_compose.
    (* swapR_injective *)
    apply (map2_injective map2_ext);intros;apply swapR_injective.
    (* swapR_0 *)
    generalize (map2_equal map2_ext 0 0 (fun d => (swapR deBruijn_t1 d 0)) (fun d x => x) (fun d => (swapR deBruijn_t2 d 0)) (fun d x => x));simpl;intros.
    rewrite H;try (intros;apply swapR_0).
    apply (map2_id map2_ext).
    (* swapR_comm *)
    generalize (map2_compose map2_ext (k+S d') 0 0 (k+S d') (fun d => swapR deBruijn_t1 d k') (fun d => swapR deBruijn_t1 d k) (fun d => swapR deBruijn_t1 d k) (fun d => swapR deBruijn_t1 d k') (fun d => swapR deBruijn_t2 d k') (fun d => swapR deBruijn_t2 d k) (fun d => swapR deBruijn_t2 d k) (fun d => swapR deBruijn_t2 d k'));simpl;intros.
    repeat (rewrite <- plus_assoc);repeat (rewrite (plus_comm d)).
    apply H;intros;repeat (rewrite (plus_comm (k+S d')));repeat (rewrite plus_assoc);apply swapR_comm.
    (* swapR_1_idem *)
    generalize (map2_compose_simpl map2_ext 0 0 0 (fun d => swapR deBruijn_t1 d 1) (fun d => (swapR deBruijn_t1 d 1)) (fun d x => x) (fun d => swapR deBruijn_t2 d 1) (fun d => (swapR deBruijn_t2 d 1)) (fun d x => x));simpl;intros.
    rewrite H;try (intros;apply swapR_1_idem).
    apply (map2_id map2_ext).
    (* swapR_simpl *)
    generalize (map2_compose_simpl map2_ext 0 k 0 (fun d => swapR deBruijn_t1 d k) (fun d => swapR deBruijn_t1 d k') (fun d => swapR deBruijn_t1 d (k+k')) (fun d => swapR deBruijn_t2 d k) (fun d => swapR deBruijn_t2 d k') (fun d => swapR deBruijn_t2 d (k+k')));simpl;intros.
    rewrite (plus_comm d).
    apply H;intros;rewrite (plus_comm k d0);apply swapR_simpl.
    (* swapR_compose *)
    generalize (map2_compose map2_ext 0 0 1 0 (fun d => swapR deBruijn_t1 d (k+S k')) (fun d => swapR deBruijn_t1 d k) (fun d => swapR deBruijn_t1 d k) (fun d => swapR deBruijn_t1 d (k+S k')) (fun d => swapR deBruijn_t2 d (k+S k')) (fun d => swapR deBruijn_t2 d k) (fun d => swapR deBruijn_t2 d k) (fun d => swapR deBruijn_t2 d (k+S k')));simpl;intros.
    apply H;intros;apply swapR_compose.
    (* swapR_inside *)
   generalize (map2_compose map2_ext 0 d' (S d') 0 (fun d => swapR deBruijn_t1 d (d'+S k)) (fun d => swapR deBruijn_t1 d k) (fun d => swapR deBruijn_t1 d k) (fun d => swapR deBruijn_t1 d (d'+S k)) (fun d => swapR deBruijn_t2 d (d'+S k)) (fun d => swapR deBruijn_t2 d k) (fun d => swapR deBruijn_t2 d k) (fun d => swapR deBruijn_t2 d (d'+S k)));simpl;intros.
    rewrite (plus_comm d d').
    rewrite (plus_comm d (S d'));simpl.
    apply H;intros;rewrite (plus_comm d' d0);rewrite plus_n_Sm;apply swapR_inside.
    (* lift_swapR_comm *)
    generalize (map2_compose map2_ext (k+S d') 0 0 (k+S d') (fun d => lift deBruijn_t1 d k') (fun d => swapR deBruijn_t1 d k) (fun d => swapR deBruijn_t1 d k) (fun d => lift deBruijn_t1 d k') (fun d => lift deBruijn_t2 d k') (fun d => swapR deBruijn_t2 d k) (fun d => swapR deBruijn_t2 d k) (fun d => lift deBruijn_t2 d k'));simpl;intros.
    repeat (rewrite <- plus_assoc);repeat (rewrite (plus_comm d)).
    apply H;intros;repeat (rewrite (plus_comm (k+S d')));repeat (rewrite plus_assoc);apply lift_swapR_comm.
    (* swapR_lift_comm *)
    generalize (map2_compose map2_ext (k+d') 0 0 d' (fun d => swapR deBruijn_t1 d k') (fun d => lift deBruijn_t1 d k) (fun d => lift deBruijn_t1 d k) (fun d => swapR deBruijn_t1 d k') (fun d => swapR deBruijn_t2 d k') (fun d => lift deBruijn_t2 d k) (fun d => lift deBruijn_t2 d k) (fun d => swapR deBruijn_t2 d k'));simpl;intros.
    repeat (rewrite <- plus_assoc);repeat (rewrite (plus_comm d)).
    apply H;intros;repeat (rewrite (plus_comm (k+d')));repeat (rewrite plus_assoc);rewrite (plus_comm d' d0);apply swapR_lift_comm.
    (* swapR_lift_idem *)
    generalize (map2_compose_simpl map2_ext d' 0 0 (fun d => swapR deBruijn_t1 d k) (fun d => lift deBruijn_t1 d (d'+k+S k')) (fun d => lift deBruijn_t1 d (d'+k+S k')) (fun d => swapR deBruijn_t2 d k) (fun d => lift deBruijn_t2 d (d'+k+S k')) (fun d => lift deBruijn_t2 d (d'+k+S k')));simpl;intros.
    rewrite (plus_comm d).
    apply H;intros;rewrite (plus_comm d' d0);apply swapR_lift_idem.
    (* swapR_lift_simpl *)
    generalize (map2_compose_simpl map2_ext 0 0 1 (fun d => swapR deBruijn_t1 d k) (fun d => lift deBruijn_t1 d k) (fun d => lift deBruijn_t1 d k) (fun d => swapR deBruijn_t2 d k) (fun d => lift deBruijn_t2 d k) (fun d => lift deBruijn_t2 d k));simpl;intros.
    apply H;intros;apply swapR_lift_simpl.
    (* swapR_lift_1 *)
    generalize (map2_compose_simpl map2_ext 0 k 0 (fun d => swapR deBruijn_t1 d k) (fun d => lift deBruijn_t1 d 1) (fun d => lift deBruijn_t1 d 1) (fun d => swapR deBruijn_t2 d k) (fun d => lift deBruijn_t2 d 1) (fun d => lift deBruijn_t2 d 1));simpl;intros.
    rewrite plus_comm.
    apply H;intros;rewrite plus_comm;apply swapR_lift_1.
    (* low1_lift_simpl *)
    generalize (map2_compose_simpl map2_ext d' 0 0 (fun d => low1 deBruijn_t1 d) (fun d => lift deBruijn_t1 d (d'+S k)) (fun d => lift deBruijn_t1 d (d'+k)) (fun d => low1 deBruijn_t2 d) (fun d => lift deBruijn_t2 d (d'+S k)) (fun d => lift deBruijn_t2 d (d'+k)));simpl;intros.
    rewrite plus_comm;apply H;intros.
    rewrite plus_comm;apply low1_lift_lift.
    rewrite plus_comm;apply low1_lift_lift.
    (* swapR_occur_swap *)
    generalize (map2_compose map2_ext n2 n1 n'2 n'1 (fun d => swapR deBruijn_t1 d (S k - n2)) (fun d => swapR deBruijn_t1 d (k-n1)) (fun d => swapR deBruijn_t1 d (S k-n'2)) (fun d => swapR deBruijn_t1 d (S (k-n'1))) (fun d => swapR deBruijn_t2 d (S k - n2)) (fun d => swapR deBruijn_t2 d (k-n1)) (fun d => swapR deBruijn_t2 d (S k-n'2)) (fun d => swapR deBruijn_t2 d (S (k-n'1))));intros.
    repeat (rewrite (plus_comm d)).
    apply H2;intros;rewrite (plus_comm n2);rewrite (plus_comm n1);rewrite (plus_comm n'2);rewrite (plus_comm n'1);apply swapR_occur_swap;auto.
  Defined.

  Record deBruijn_subst (t:Set)(deBruijn_t:deBruijn_base t)(to:Set)(deBruijn_to:deBruijn_base to) : Set := {
    low_subst : nat -> nat -> list to -> t -> t;
    (* low_subst and lift *)
    low_subst_lift_comm : forall (d d' k k':nat)(es:list to)(x:t),(low_subst (d+d'+k) k' es (lift deBruijn_t d k x))=(lift deBruijn_t d k (low_subst (d+d') k' es x));
    lift_low_subst_succ : forall (d d' k k':nat)(es:list to)(x:t),(lift deBruijn_t (d+d') k' (low_subst d (d'+k) es x))=(low_subst d (k+k') (map (lift deBruijn_to d' k') es) (lift deBruijn_t (d+length es) d' x));
    low_subst_lift_nil : forall (d k k':nat)(es:list to)(x:t),(low_subst d k es (lift deBruijn_t d (k'+length es) x))=(lift deBruijn_t d (k+k') x);
    lift_low_subst_comm : forall (d d' k k':nat)(es:list to)(x:t),(lift deBruijn_t (d+d'+k) k' (low_subst d' k es x))=(low_subst d' k (map (lift deBruijn_to (d+k) k') es) (lift deBruijn_t (d'+d+length es) k' x));
    low_subst_lift_succ_simpl : forall (d k k':nat)(es:list to)(x:t),(low_subst d k es (lift deBruijn_t (d+length es) k' x))=(low_subst d (k+k') es x);
    (* low_subst and swapR *)
    low_subst_swapR_comm : forall (d d' k k':nat)(es:list to)(x:t),(low_subst (d+k+(S d')) k' es (swapR deBruijn_t d k x))=(swapR deBruijn_t d k (low_subst (d+k+(S d')) k' es x));
    swapR_low_subst_idem : forall (d k k':nat)(es:list to)(x:t),(swapR deBruijn_t d k (low_subst d (k+S k') es x))=(low_subst d (k+S k') (map (swapR deBruijn_to 0 k) es) x);
    low_subst_swapR_simpl : forall (d k:nat)(es:list to)(x:t),(low_subst (S d) k es (swapR deBruijn_t d (length es) x))=(swapR deBruijn_t d k (low_subst d k (map (lift deBruijn_to k 1) es) x));
    swapR_low_subst_comm : forall (d d' k k':nat)(es:list to)(x:t),(swapR deBruijn_t (d+d'+k) k' (low_subst d' k es x))=(low_subst d' k (map (swapR deBruijn_to (d+k) k') es) (swapR deBruijn_t (d'+d+length es) k' x));
    swapR_low_subst_comm_idem : forall (d d' k k':nat)(es:list to)(x:t),(swapR deBruijn_t (d+d') k' (low_subst d (d'+k+S k') es x))=(low_subst d (d'+k+S k') (map (swapR deBruijn_to d' k') es) x);
    low_subst_swapR_rotate : forall (d k:nat)(e:to)(es:list to)(x:t),(low_subst d k (List_right_rotate (cons e es)) (swapR deBruijn_t d (length es) x))=(low_subst d k (cons e es) x)
  }.

  Implicit Arguments low_subst [t deBruijn_t to deBruijn_to].
  Implicit Arguments low_subst_lift_comm [t deBruijn_t to deBruijn_to].
  Implicit Arguments lift_low_subst_succ [t deBruijn_t to deBruijn_to].
  Implicit Arguments low_subst_lift_nil [t deBruijn_t to deBruijn_to].
  Implicit Arguments lift_low_subst_comm [t deBruijn_t to deBruijn_to].
  Implicit Arguments low_subst_lift_succ_simpl [t deBruijn_t to deBruijn_to].
  Implicit Arguments low_subst_swapR_comm [t deBruijn_t to deBruijn_to].
  Implicit Arguments swapR_low_subst_idem [t deBruijn_t to deBruijn_to].
  Implicit Arguments low_subst_swapR_simpl [t deBruijn_t to deBruijn_to].
  Implicit Arguments swapR_low_subst_comm [t deBruijn_t to deBruijn_to].
  Implicit Arguments swapR_low_subst_comm_idem [t deBruijn_t to deBruijn_to].
  Implicit Arguments low_subst_swapR_rotate [t deBruijn_t to deBruijn_to].

  Definition deBruijn_s_map_struct (t:Set)(deBruijn_t:deBruijn_base t)(to:Set)(deBruijn_to:deBruijn_base to)(t_subst_to:deBruijn_subst t deBruijn_t to deBruijn_to)(t':Set)(map_fun:(nat->t->t)->nat->t'->t')(map_ext:map_struct t t' map_fun) : (deBruijn_subst t' (deBruijn_map_struct t deBruijn_t t' map_fun map_ext) to deBruijn_to).
    intros.
    split with (fun d k es => map_fun (fun d => low_subst t_subst_to d k es) d);intros;simpl.
    (* low_subst_lift_comm *)
    generalize (map_compose map_ext (d'+k) 0 0 d' (fun d => low_subst t_subst_to d k' es) (fun d => lift deBruijn_t d k) (fun d => lift deBruijn_t d k) (fun d => low_subst t_subst_to d k' es));simpl;intros.
    repeat (rewrite (plus_comm d d'));repeat (rewrite <- plus_assoc);repeat (rewrite (plus_comm d k));repeat (rewrite plus_assoc).
    apply H;intros;repeat (rewrite <- plus_assoc);repeat (rewrite (plus_comm k d0));repeat (rewrite plus_assoc);repeat (rewrite (plus_comm d' d0));apply low_subst_lift_comm.
    (* lift_low_subst_succ *)
    generalize (map_compose map_ext d' 0 0 (length es) (fun d => lift deBruijn_t d k') (fun d => low_subst t_subst_to d (d'+k) es) (fun d => low_subst t_subst_to d (k+k') (map (lift deBruijn_to d' k') es)) (fun d => lift deBruijn_t d d'));simpl;intros.
    repeat (rewrite (plus_comm d d'));repeat (rewrite (plus_comm d (length es))).
    apply H;intros;repeat (rewrite (plus_comm d' d0));repeat (rewrite (plus_comm (length es) d0));apply lift_low_subst_succ.
    (* low_subst_lift_nil *)
    generalize (map_compose_simpl map_ext 0 0 0 (fun d => low_subst t_subst_to d k es) (fun d => lift deBruijn_t d (k'+length es)) (fun d => lift deBruijn_t d (k+k')));simpl;intros.
    apply H;intros;apply low_subst_lift_nil.
    (* lift_low_subst_comm *)
    generalize (map_compose map_ext (d+k) 0 0 (d+length es) (fun d' => lift deBruijn_t d' k') (fun d' => low_subst t_subst_to d' k es) (fun d' => low_subst t_subst_to d' k (map (lift deBruijn_to (d+k) k') es)) (fun d' => lift deBruijn_t d' k'));simpl;intros.
    repeat (rewrite <- plus_assoc);repeat (rewrite (plus_comm d'));repeat (rewrite plus_assoc);simpl.
    apply H;intros;repeat (rewrite (plus_comm (d+k) d0));rewrite plus_assoc;rewrite (plus_comm d0 d);repeat (rewrite (plus_comm (d+length es) d0));repeat (rewrite plus_assoc);apply lift_low_subst_comm.
    (* low_subst_lift_succ_simpl *)
    generalize (map_compose_simpl map_ext 0 (length es) 0 (fun d => low_subst t_subst_to d k es) (fun d => lift deBruijn_t d k') (fun d => low_subst t_subst_to d (k+k') es));simpl;intros.
    repeat (rewrite (plus_comm d (length es))).
    apply H;intros;repeat (rewrite (plus_comm (length es) d0));apply low_subst_lift_succ_simpl.
    (* low_subst_swapR_comm *)
    generalize (map_compose map_ext (k+S d') 0 0 (k+S d') (fun d => low_subst t_subst_to d k' es) (fun d => swapR deBruijn_t d k) (fun d => swapR deBruijn_t d k) (fun d => low_subst t_subst_to d k' es));simpl;intros.
    repeat (rewrite <- (plus_assoc));repeat (rewrite (plus_comm d)).
    apply H;intros;repeat (rewrite (plus_comm (k+S d')));repeat (rewrite plus_assoc);apply low_subst_swapR_comm.
    (* swapR_low_subst_idem *)
    generalize (map_compose_simpl map_ext 0 0 0 (fun d => swapR deBruijn_t d k) (fun d => low_subst t_subst_to d (k+S k') es) (fun d => low_subst t_subst_to d (k+S k') (map (swapR deBruijn_to 0 k) es)));simpl;intros.
    apply H;intros;apply swapR_low_subst_idem.
    (* low_subst_swapR_simpl *)
    generalize (map_compose map_ext 1 0 0 0 (fun d => low_subst t_subst_to d k es) (fun d => swapR deBruijn_t d (length es)) (fun d => swapR deBruijn_t d k) (fun d => low_subst t_subst_to d k (map (lift deBruijn_to k 1) es)));simpl;intros.
    apply H;intros;apply low_subst_swapR_simpl.
    (* swapR_low_subst_comm *)
    generalize (map_compose map_ext (d+k) 0 0 (d+length es) (fun d0 => swapR deBruijn_t d0 k') (fun d0 => low_subst t_subst_to d0 k es) (fun d0 => low_subst t_subst_to d0 k (map (swapR deBruijn_to (d+k) k') es)) (fun d0 => swapR deBruijn_t d0 k'));simpl;intros.
    repeat (rewrite <- (plus_assoc));repeat (rewrite (plus_comm d'));repeat (rewrite plus_assoc);simpl.
    apply H;intros;repeat (rewrite <- (plus_assoc));repeat (rewrite (plus_comm k d0));repeat (rewrite plus_assoc);repeat (rewrite (plus_comm (d+length es) d0));repeat (rewrite plus_assoc);simpl;apply swapR_low_subst_comm.
    (* swapR_low_subst_comm_idem *)
    generalize (map_compose_simpl map_ext d' 0 0 (fun d => swapR deBruijn_t d k') (fun d => low_subst t_subst_to d (d'+k+S k') es) (fun d => low_subst t_subst_to d (d'+k+S k') (map (swapR deBruijn_to d' k') es)));simpl;intros.
    rewrite (plus_comm d d').
    apply H;intros;rewrite (plus_comm d' d0);apply swapR_low_subst_comm_idem.
    (* low_subst_swapR_rotate *)
    generalize (map_compose_simpl map_ext 0 0 0 (fun d => low_subst t_subst_to d k (List_right_rotate (cons e es))) (fun d => swapR deBruijn_t d (length es)) (fun d => low_subst t_subst_to d k (cons e es)));simpl;intros.
    apply H.
    generalize (low_subst_swapR_rotate t_subst_to);simpl;intros;auto.
  Defined.

  Definition deBruijn_s_map2_struct (t1:Set)(deBruijn_t1:deBruijn_base t1)(t2:Set)(deBruijn_t2:deBruijn_base t2)(to:Set)(deBruijn_to:deBruijn_base to)(t1_subst_to:deBruijn_subst t1 deBruijn_t1 to deBruijn_to)(t2_subst_to:deBruijn_subst t2 deBruijn_t2 to deBruijn_to)(t:Set)(map2_fun:(nat->t1->t1)->(nat->t2->t2)->nat->t->t)(map2_ext:map2_struct t1 t2 t map2_fun) : (deBruijn_subst t (deBruijn_map2_struct t1 deBruijn_t1 t2 deBruijn_t2 t map2_fun map2_ext) to deBruijn_to).
    intros.
    split with (fun d k es => map2_fun (fun d => low_subst t1_subst_to d k es) (fun d => low_subst t2_subst_to d k es) d);intros;simpl.
    (* low_subst_lift_comm *)
    generalize (map2_compose map2_ext (d'+k) 0 0 d' (fun d => low_subst t1_subst_to d k' es) (fun d => lift deBruijn_t1 d k) (fun d => lift deBruijn_t1 d k) (fun d => low_subst t1_subst_to d k' es) (fun d => low_subst t2_subst_to d k' es) (fun d => lift deBruijn_t2 d k) (fun d => lift deBruijn_t2 d k) (fun d => low_subst t2_subst_to d k' es));simpl;intros.
    repeat (rewrite (plus_comm d d'));repeat (rewrite <- plus_assoc);repeat (rewrite (plus_comm d k));repeat (rewrite plus_assoc).
    apply H;intros;repeat (rewrite <- plus_assoc);repeat (rewrite (plus_comm k d0));repeat (rewrite plus_assoc);repeat (rewrite (plus_comm d' d0));apply low_subst_lift_comm.
    (* lift_low_subst_succ *)
    generalize (map2_compose map2_ext d' 0 0 (length es) (fun d => lift deBruijn_t1 d k') (fun d => low_subst t1_subst_to d (d'+k) es) (fun d => low_subst t1_subst_to d (k+k') (map (lift deBruijn_to d' k') es)) (fun d => lift deBruijn_t1 d d') (fun d => lift deBruijn_t2 d k') (fun d => low_subst t2_subst_to d (d'+k) es) (fun d => low_subst t2_subst_to d (k+k') (map (lift deBruijn_to d' k') es)) (fun d => lift deBruijn_t2 d d'));simpl;intros.
    repeat (rewrite (plus_comm d d'));repeat (rewrite (plus_comm d (length es))).
    apply H;intros;repeat (rewrite (plus_comm d' d0));repeat (rewrite (plus_comm (length es) d0));apply lift_low_subst_succ.
    (* low_subst_lift_nil *)
    generalize (map2_compose_simpl map2_ext 0 0 0 (fun d => low_subst t1_subst_to d k es) (fun d => lift deBruijn_t1 d (k'+length es)) (fun d => lift deBruijn_t1 d (k+k')) (fun d => low_subst t2_subst_to d k es) (fun d => lift deBruijn_t2 d (k'+length es)) (fun d => lift deBruijn_t2 d (k+k')));simpl;intros.
    apply H;intros;apply low_subst_lift_nil.
   (* lift_low_subst_comm *)
    generalize (map2_compose map2_ext (d+k) 0 0 (d+length es) (fun d' => lift deBruijn_t1 d' k') (fun d' => low_subst t1_subst_to d' k es) (fun d' => low_subst t1_subst_to d' k (map (lift deBruijn_to (d+k) k') es)) (fun d' => lift deBruijn_t1 d' k') (fun d' => lift deBruijn_t2 d' k') (fun d' => low_subst t2_subst_to d' k es) (fun d' => low_subst t2_subst_to d' k (map (lift deBruijn_to (d+k) k') es)) (fun d' => lift deBruijn_t2 d' k'));simpl;intros.
    repeat (rewrite <- plus_assoc);repeat (rewrite (plus_comm d'));repeat (rewrite plus_assoc);simpl.
    apply H;intros;repeat (rewrite (plus_comm (d+k) d0));rewrite plus_assoc;rewrite (plus_comm d0 d);repeat (rewrite (plus_comm (d+length es) d0));repeat (rewrite plus_assoc);apply lift_low_subst_comm.
    (* low_subst_lift_succ_simpl *)
    generalize (map2_compose_simpl map2_ext 0 (length es) 0 (fun d => low_subst t1_subst_to d k es) (fun d => lift deBruijn_t1 d k') (fun d => low_subst t1_subst_to d (k+k') es) (fun d => low_subst t2_subst_to d k es) (fun d => lift deBruijn_t2 d k') (fun d => low_subst t2_subst_to d (k+k') es));simpl;intros.
    repeat (rewrite (plus_comm d (length es))).
    apply H;intros;repeat (rewrite (plus_comm (length es) d0));apply low_subst_lift_succ_simpl.
    (* low_subst_swapR_comm *)
    generalize (map2_compose map2_ext (k+S d') 0 0 (k+S d') (fun d => low_subst t1_subst_to d k' es) (fun d => swapR deBruijn_t1 d k) (fun d => swapR deBruijn_t1 d k) (fun d => low_subst t1_subst_to d k' es) (fun d => low_subst t2_subst_to d k' es) (fun d => swapR deBruijn_t2 d k) (fun d => swapR deBruijn_t2 d k) (fun d => low_subst t2_subst_to d k' es));simpl;intros.
    repeat (rewrite <- (plus_assoc));repeat (rewrite (plus_comm d)).
    apply H;intros;repeat (rewrite (plus_comm (k+S d')));repeat (rewrite plus_assoc);apply low_subst_swapR_comm.
    (* swapR_low_subst_idem *)
    generalize (map2_compose_simpl map2_ext 0 0 0 (fun d => swapR deBruijn_t1 d k) (fun d => low_subst t1_subst_to d (k+S k') es) (fun d => low_subst t1_subst_to d (k+S k') (map (swapR deBruijn_to 0 k) es)) (fun d => swapR deBruijn_t2 d k) (fun d => low_subst t2_subst_to d (k+S k') es) (fun d => low_subst t2_subst_to d (k+S k') (map (swapR deBruijn_to 0 k) es)));simpl;intros.
    apply H;intros;apply swapR_low_subst_idem.
     (* low_subst_swapR_simpl *)
    generalize (map2_compose map2_ext 1 0 0 0 (fun d => low_subst t1_subst_to d k es) (fun d => swapR deBruijn_t1 d (length es)) (fun d => swapR deBruijn_t1 d k) (fun d => low_subst t1_subst_to d k (map (lift deBruijn_to k 1) es)) (fun d => low_subst t2_subst_to d k es) (fun d => swapR deBruijn_t2 d (length es)) (fun d => swapR deBruijn_t2 d k) (fun d => low_subst t2_subst_to d k (map (lift deBruijn_to k 1) es)));simpl;intros.
    apply H;intros;apply low_subst_swapR_simpl.
     (* swapR_low_subst_comm *)
    generalize (map2_compose map2_ext (d+k) 0 0 (d+length es) (fun d0 => swapR deBruijn_t1 d0 k') (fun d0 => low_subst t1_subst_to d0 k es) (fun d0 => low_subst t1_subst_to d0 k (map (swapR deBruijn_to (d+k) k') es)) (fun d0 => swapR deBruijn_t1 d0 k') (fun d0 => swapR deBruijn_t2 d0 k') (fun d0 => low_subst t2_subst_to d0 k es) (fun d0 => low_subst t2_subst_to d0 k (map (swapR deBruijn_to (d+k) k') es)) (fun d0 => swapR deBruijn_t2 d0 k'));simpl;intros.
    repeat (rewrite <- (plus_assoc));repeat (rewrite (plus_comm d'));repeat (rewrite plus_assoc);simpl.
    apply H;intros;repeat (rewrite <- (plus_assoc));repeat (rewrite (plus_comm k d0));repeat (rewrite plus_assoc);repeat (rewrite (plus_comm (d+length es) d0));repeat (rewrite plus_assoc);simpl;apply swapR_low_subst_comm.
      (* swapR_low_subst_comm_idem *)
    generalize (map2_compose_simpl map2_ext d' 0 0 (fun d => swapR deBruijn_t1 d k') (fun d => low_subst t1_subst_to d (d'+k+S k') es) (fun d => low_subst t1_subst_to d (d'+k+S k') (map (swapR deBruijn_to d' k') es)) (fun d => swapR deBruijn_t2 d k') (fun d => low_subst t2_subst_to d (d'+k+S k') es) (fun d => low_subst t2_subst_to d (d'+k+S k') (map (swapR deBruijn_to d' k') es)));simpl;intros.
    rewrite (plus_comm d d').
    apply H;intros;rewrite (plus_comm d' d0);apply swapR_low_subst_comm_idem.
    (* low_subst_swapR_rotate *)
    generalize (map2_compose_simpl map2_ext 0 0 0 (fun d => low_subst t1_subst_to d k (List_right_rotate (cons e es))) (fun d => swapR deBruijn_t1 d (length es)) (fun d => low_subst t1_subst_to d k (cons e es)) (fun d => low_subst t2_subst_to d k (List_right_rotate (cons e es))) (fun d => swapR deBruijn_t2 d (length es)) (fun d => low_subst t2_subst_to d k (cons e es)));simpl;intros.
    apply H.
    generalize (low_subst_swapR_rotate t1_subst_to);simpl;intros;auto.
    generalize (low_subst_swapR_rotate t2_subst_to);simpl;intros;auto.
  Defined.

  Record deBruijn_subst_gen (t:Set)(deBruijn_t:deBruijn_base t)(to:Set)(deBruijn_to:deBruijn_base to)(deBruijn_ext:deBruijn_extension t deBruijn_t to deBruijn_to) : Set := {
    low_subst_gen : nat -> nat -> list to -> t -> to;
    (* low_subst and lift *)
    low_subst_gen_lift_comm : forall (d d' k k':nat)(es:list to)(x:t),(low_subst_gen (d+d'+k) k' es (lift deBruijn_t d k x))=(lift deBruijn_to d k (low_subst_gen (d+d') k' es x));
    lift_low_subst_gen_succ : forall (d d' k k':nat)(es:list to)(x:t),(lift deBruijn_to (d+d') k' (low_subst_gen d (d'+k) es x))=(low_subst_gen d (k+k') (map (lift deBruijn_to d' k') es) (lift deBruijn_t (d+length es) d' x));
    low_subst_gen_lift_nil : forall (d k k':nat)(es:list to)(x:t),(low_subst_gen d k es (lift deBruijn_t d (k'+length es) x))=(lift deBruijn_to d (k+k') (j deBruijn_ext x));
    lift_low_subst_gen_comm : forall (d d' k k':nat)(es:list to)(x:t),(lift deBruijn_to (d+d'+k) k' (low_subst_gen d' k es x))=(low_subst_gen d' k (map (lift deBruijn_to (d+k) k') es) (lift deBruijn_t (d'+d+length es) k' x));
    low_subst_gen_lift_succ_simpl :  forall (d k k':nat)(es:list to)(x:t),(low_subst_gen d k es (lift deBruijn_t (d+length es) k' x))=(low_subst_gen d (k+k') es x);
    (* low_subst and swapR *)
    low_subst_gen_swapR_comm : forall (d d' k k':nat)(es:list to)(x:t),(low_subst_gen (d+k+(S d')) k' es (swapR deBruijn_t d k x))=(swapR deBruijn_to d k (low_subst_gen (d+k+(S d')) k' es x));
    swapR_low_subst_gen_idem : forall (d k k':nat)(es:list to)(x:t),(swapR deBruijn_to d k (low_subst_gen d (k+S k') es x))=(low_subst_gen d (k+S k') (map (swapR deBruijn_to 0 k) es) x);
    low_subst_gen_swapR_simpl : forall (d k:nat)(es:list to)(x:t),(low_subst_gen (S d) k es (swapR deBruijn_t d (length es) x))=(swapR deBruijn_to d k (low_subst_gen d k (map (lift deBruijn_to k 1) es) x));
    swapR_low_subst_gen_comm : forall (d d' k k':nat)(es:list to)(x:t),(swapR deBruijn_to (d+d'+k) k' (low_subst_gen d' k es x))=(low_subst_gen d' k (map (swapR deBruijn_to (d+k) k') es) (swapR deBruijn_t (d'+d+length es) k' x));
    swapR_low_subst_gen_comm_idem : forall (d d' k k':nat)(es:list to)(x:t),(swapR deBruijn_to (d+d') k' (low_subst_gen d (d'+k+S k') es x))=(low_subst_gen d (d'+k+S k') (map (swapR deBruijn_to d' k') es) x);
    low_subst_gen_swapR_rotate : forall (d k:nat)(e:to)(es:list to)(x:t),(low_subst_gen d k (List_right_rotate (cons e es)) (swapR deBruijn_t d (length es) x))=(low_subst_gen d k (cons e es) x)
  }.

  Implicit Arguments low_subst_gen [t deBruijn_t to deBruijn_to deBruijn_ext].
  Implicit Arguments low_subst_gen_lift_comm [t deBruijn_t to deBruijn_to deBruijn_ext].
  Implicit Arguments lift_low_subst_gen_succ [t deBruijn_t to deBruijn_to deBruijn_ext].
  Implicit Arguments low_subst_gen_lift_nil [t deBruijn_t to deBruijn_to deBruijn_ext].
  Implicit Arguments lift_low_subst_gen_comm [t deBruijn_t to deBruijn_to deBruijn_ext].
  Implicit Arguments low_subst_gen_lift_succ_simpl [t deBruijn_t to deBruijn_to deBruijn_ext].
  Implicit Arguments low_subst_gen_swapR_comm [t deBruijn_t to deBruijn_to deBruijn_ext].
  Implicit Arguments swapR_low_subst_gen_idem [t deBruijn_t to deBruijn_to deBruijn_ext].
  Implicit Arguments low_subst_gen_swapR_simpl [t deBruijn_t to deBruijn_to deBruijn_ext].
  Implicit Arguments swapR_low_subst_gen_comm [t deBruijn_t to deBruijn_to deBruijn_ext].
  Implicit Arguments swapR_low_subst_gen_comm_idem [t deBruijn_t to deBruijn_to deBruijn_ext].
  Implicit Arguments low_subst_gen_swapR_rotate [t deBruijn_t to deBruijn_to deBruijn_ext].

  Record map_extension (t:Set)(deBruijn_t:deBruijn_base t)(t':Set)(deBruijn_t':deBruijn_base t')(deBruijn_ext:deBruijn_extension t deBruijn_t t' deBruijn_t')(map_fun:(nat->t->t)->nat->t'->t')(deBruijn_map_ext:map_struct t t' map_fun)(map_gen_fun:(nat->t->t')->nat->t'->t') : Prop := {
    map_gen_j_map : forall (f:nat->t->t)(d:nat)(x:t'),(map_fun f d x)=(map_gen_fun (fun d x => j deBruijn_ext (f d x)) d x);
    map_gen_j : forall (f:nat->t->t')(d:nat)(x:t),(map_gen_fun f d (j deBruijn_ext x))=(f d x);
    map_gen_equal : forall (d1 d2:nat)(f g:nat->t->t'),(forall (d:nat)(x:t),(f (d1+d) x)=(g (d2+d) x))->forall (d:nat)(x:t'),(map_gen_fun f (d1+d) x)=(map_gen_fun g (d2+d) x);
    map_gen_compose : forall (d1 d2:nat)(f g:nat->t->t'),forall (d:nat)(x:t'),(map_gen_fun f (d1+d) (map_gen_fun g (d2+d) x))=(map_gen_fun (fun d x => map_gen_fun f (d1+d) (g (d2+d) x)) d x)
  }.

  Implicit Arguments map_gen_j_map [t deBruijn_t t' deBruijn_t' deBruijn_ext map_fun deBruijn_map_ext map_gen_fun].
  Implicit Arguments map_gen_j [t deBruijn_t t' deBruijn_t' deBruijn_ext map_fun deBruijn_map_ext map_gen_fun].
  Implicit Arguments map_gen_equal [t deBruijn_t t' deBruijn_t' deBruijn_ext map_fun deBruijn_map_ext map_gen_fun].
  Implicit Arguments map_gen_compose [t deBruijn_t t' deBruijn_t' deBruijn_ext map_fun deBruijn_map_ext map_gen_fun].

  Definition deBruijn_subst_gen_subst (t:Set)(deBruijn_t:deBruijn_base t)(to:Set)(map_fun:(nat->t->t)->nat->to->to)(map_str:map_struct t to map_fun)(deBruijn_ext:deBruijn_extension t deBruijn_t to (deBruijn_map_struct t deBruijn_t to map_fun map_str))(deBruijn_s_gen:deBruijn_subst_gen t deBruijn_t to (deBruijn_map_struct t deBruijn_t to map_fun map_str) deBruijn_ext)(map_gen_fun:(nat->t->to)->nat->to->to)(map_ext:map_extension t deBruijn_t to (deBruijn_map_struct t deBruijn_t to map_fun map_str) deBruijn_ext map_fun map_str map_gen_fun) : (deBruijn_subst to (deBruijn_map_struct t deBruijn_t to map_fun map_str) to (deBruijn_map_struct t deBruijn_t to map_fun map_str)).
    intros.
    split with (fun d k es => map_gen_fun (fun d => low_subst_gen deBruijn_s_gen d k es) d);intro;simpl;intros.
    (* low_subst_lift_comm *)
    repeat (rewrite (map_gen_j_map map_ext)).
    generalize (map_gen_compose map_ext (d'+k) 0);simpl;intro.
    generalize (map_gen_compose map_ext 0 d');simpl;intro.
    repeat (rewrite <- plus_assoc);repeat (rewrite (plus_comm d));repeat (rewrite plus_assoc).
    rewrite H;rewrite H0.
    generalize (map_gen_equal map_ext 0 0);simpl;intro.
    apply H1;intros.
    rewrite <- (map_gen_j_map map_ext (fun d => lift deBruijn_t d k)).
    rewrite (map_gen_j map_ext).
    repeat (rewrite (plus_comm (d'+k)));repeat (rewrite (plus_comm d' d0));repeat (rewrite plus_assoc).
    generalize (low_subst_gen_lift_comm deBruijn_s_gen);intro.
    simpl in H2;auto.
    (* lift_low_subst_succ *)
    repeat (rewrite (map_gen_j_map map_ext)).
    generalize (map_gen_compose map_ext d' 0);simpl;intro.
    generalize (map_gen_compose map_ext 0 (length es));simpl;intro.
    repeat (rewrite (plus_comm d)).
    rewrite H;rewrite H0.
    generalize (map_gen_equal map_ext 0 0);simpl;intro.
    apply H1;intros.
    rewrite <- (map_gen_j_map map_ext (fun d => lift deBruijn_t d k')).
    rewrite (map_gen_j map_ext).
    repeat (rewrite (plus_comm d' d0));repeat (rewrite (plus_comm (length es) d0)).
    generalize (lift_low_subst_gen_succ deBruijn_s_gen);intro.
    simpl in H2;auto.
    (* low_subst_lift_nil *)
    repeat (rewrite (map_gen_j_map map_ext)).
    generalize (map_gen_compose map_ext 0 0);simpl;intros.
    rewrite H.
    generalize (map_gen_equal map_ext 0 0);simpl;intros.
    apply H0.
    intros.
    rewrite (map_gen_j map_ext).
    rewrite (low_subst_gen_lift_nil deBruijn_s_gen).
    apply j_lift.
    (* lift_low_subst_comm *)
    repeat (rewrite (map_gen_j_map map_ext)).
    generalize (map_gen_compose map_ext (d+k) 0);simpl;intro.
    generalize (map_gen_compose map_ext 0 (d+length es));simpl;intro.
    repeat (rewrite <- plus_assoc);repeat (rewrite (plus_comm d'));repeat (rewrite plus_assoc);simpl.
    rewrite H;rewrite H0.
    generalize (map_gen_equal map_ext 0 0);simpl;intro.
    apply H1;intros.
    rewrite <- (map_gen_j_map map_ext (fun d => lift deBruijn_t d k')).
    rewrite (map_gen_j map_ext).
    repeat (rewrite (plus_comm (d+k) d0));rewrite plus_assoc;rewrite (plus_comm d0 d);repeat (rewrite (plus_comm (d+length es) d0));repeat (rewrite plus_assoc).
    generalize (lift_low_subst_gen_comm deBruijn_s_gen);simpl;intros.
    apply H2.
    (* low_subst_lift_succ_simpl *)
    repeat (rewrite (map_gen_j_map map_ext)).
    generalize (map_gen_compose map_ext 0 (length es));simpl;intro.
    rewrite (plus_comm d (length es)).
    rewrite H.
    generalize (map_gen_equal map_ext 0 0);simpl;intro.
    apply H0;intros.
    rewrite (plus_comm (length es) d0).
    rewrite (map_gen_j map_ext).
    apply low_subst_gen_lift_succ_simpl.
    (* low_subst_swapR_comm *)
    repeat (rewrite (map_gen_j_map map_ext)).
    generalize (map_gen_compose map_ext (k+S d') 0);simpl;intro.
    generalize (map_gen_compose map_ext 0 (k+S d'));simpl;intro.
    repeat (rewrite <- plus_assoc);repeat (rewrite (plus_comm d));repeat (rewrite plus_assoc).
    rewrite H;rewrite H0.
    generalize (map_gen_equal map_ext 0 0);simpl;intro.
    apply H1;intros.
    rewrite <- (map_gen_j_map map_ext (fun d => swapR deBruijn_t d k)).
    rewrite (map_gen_j map_ext).
    repeat (rewrite (plus_comm (k+S d') d0));repeat (rewrite plus_assoc).
    generalize (low_subst_gen_swapR_comm deBruijn_s_gen);intro.
    simpl in H2;auto.
    (* swapR_low_subst_idem *)
    repeat (rewrite (map_gen_j_map map_ext)).
    generalize (map_gen_compose map_ext 0 0);simpl;intro.
    rewrite H.
    generalize (map_gen_equal map_ext 0 0);simpl;intro.
    apply H0;intros.
    rewrite <- (map_gen_j_map map_ext (fun d => swapR deBruijn_t d k)).
    generalize (swapR_low_subst_gen_idem deBruijn_s_gen);intro.
    simpl in H1;auto.
    (* low_subst_swapR_simpl *)
    repeat (rewrite (map_gen_j_map map_ext)).
    generalize (map_gen_compose map_ext 1 0);simpl;intro.
    generalize (map_gen_compose map_ext 0 0);simpl;intro.
    rewrite H.
    rewrite H0.
    generalize (map_gen_equal map_ext 0 0);simpl;intro.
    apply H1;intros.
    rewrite <- (map_gen_j_map map_ext (fun d => swapR deBruijn_t d k)).
    rewrite (map_gen_j map_ext).
    generalize (low_subst_gen_swapR_simpl deBruijn_s_gen);intro.
    simpl in H2;auto.
    (* swapR_low_subst_comm *)
    repeat (rewrite (map_gen_j_map map_ext)).
    generalize (map_gen_compose map_ext (d+k) 0);simpl;intro.
    generalize (map_gen_compose map_ext 0 (d+length es));simpl;intro.
    repeat (rewrite <- (plus_assoc));repeat (rewrite (plus_comm d'));repeat (rewrite plus_assoc);simpl.
    rewrite H;rewrite H0.
    generalize (map_gen_equal map_ext 0 0);simpl;intro.
    apply H1;intros.
    rewrite <- (map_gen_j_map map_ext (fun d0 => swapR deBruijn_t d0 k')).
    rewrite (map_gen_j map_ext).
    repeat (rewrite <- (plus_assoc));repeat (rewrite (plus_comm k d0));repeat (rewrite plus_assoc);repeat (rewrite (plus_comm (d+length es) d0));repeat (rewrite plus_assoc);simpl.
    generalize (swapR_low_subst_gen_comm deBruijn_s_gen);intro.
    simpl in H2;auto.
    (* swapR_low_subst_comm_idem *)
    repeat (rewrite (map_gen_j_map map_ext)).
    generalize (map_gen_compose map_ext d' 0);simpl;intro.
    rewrite (plus_comm d d').
    rewrite H.
    generalize (map_gen_equal map_ext 0 0);simpl;intro.
    apply H0;intros.
    rewrite <- (map_gen_j_map map_ext (fun d => swapR deBruijn_t d k')).
    generalize (swapR_low_subst_gen_comm_idem deBruijn_s_gen);intro.
    simpl in H1.
    rewrite (plus_comm d' d0);auto.
    (* low_subst_swapR_rotate *)
    repeat (rewrite (map_gen_j_map map_ext)).
    generalize (map_gen_compose map_ext 0 0);simpl;intro.
    rewrite H.
    generalize (map_gen_equal map_ext 0 0);simpl;intro.
    apply H0;intros.
    rewrite (map_gen_j map_ext).
    generalize (low_subst_gen_swapR_rotate deBruijn_s_gen);simpl;intros;auto.
  Defined.

  Record deBruijn_fn (t:Set)(deBruijn_t:deBruijn_base t) : Set := {
    (* mem_fn k d x = true iff (Ref k) is a free name of x considered at depth d *)
    mem_fn : nat -> nat -> t -> bool;
    (* Nmin d k x (b,kmin) = kmin if b = true, otherwise is the minimum k' <= kmin appearing before k *)
    Nmin : nat -> nat ->  t -> (bool * nat) -> bool * nat;
    (* properties of mem_fn *)
    lift_low1_comm : forall (d d' k:nat)(x:t),(mem_fn 0 d x=false)->(low1 deBruijn_t d (lift deBruijn_t (d+S d') k x))=(lift deBruijn_t (d+d') k (low1 deBruijn_t d x));
    lift_1_low1_simpl : forall (d k:nat)(x:t),(mem_fn 0 (d+k) x = false)->(lift deBruijn_t d 1 (low1 deBruijn_t (d+k) x))=(swapR deBruijn_t d k x);
    lift_fn : forall (d k:nat)(x:t),forall (i:nat),(mem_fn (DeBruijnNat.lift 0 k i) d (lift deBruijn_t d k x))=(mem_fn i d x);
    lift_In_fn : forall (d':nat)(i:nat),(i<=d')->forall (d k:nat)(x:t),(mem_fn i d x = mem_fn i d (lift deBruijn_t (d+S d') k x));
    lift_not_fn : forall (d k:nat)(i:nat),(i<k)->forall (x:t),(mem_fn i d (lift deBruijn_t d k x))=false;
    swapR_low1_comm : forall (d d' k:nat)(x:t),(mem_fn 0 d x = false)->(low1 deBruijn_t d (swapR deBruijn_t (d+S d') k x))=(swapR deBruijn_t (d+d') k (low1 deBruijn_t d x));
    swapR_In_fn : forall (d':nat)(i:nat),(i<=d')->forall (d k:nat)(x:t),(mem_fn i d x = mem_fn i d (swapR deBruijn_t (d+S d') k x));
    swapR_eq_cond : forall (d k:nat)(x:t),(forall (i:nat),(i <= k)->(mem_fn i d x = false))->(swapR deBruijn_t d k x) = x;
    swapR_fn : forall (d k:nat)(x:t),forall (i:nat),(mem_fn (DeBruijnNat.swapR 0 k i) d (swapR deBruijn_t d k x))=(mem_fn i d x);
    swapR_fn_same : forall (d d' k:nat)(x:t),forall (i:nat),(mem_fn i (d+k+S d') (swapR deBruijn_t d k x))=(mem_fn i (d+k+S d') x);
    swapR_low1_comm2 : forall (d d' k:nat)(x:t),(mem_fn 0 d x = false)->(swapR deBruijn_t (d+d') k (low1 deBruijn_t d x))=(low1 deBruijn_t d (swapR deBruijn_t (d+S d') k x));
    low1_swapR_comm : forall (d d' k:nat)(x:t),(mem_fn 0 (d+k+S d') x = false)->(low1 deBruijn_t (d+k+(S d')) (swapR deBruijn_t d k x))=(swapR deBruijn_t d k (low1 deBruijn_t (d+k+(S d')) x));
    lift_low1_fn : forall (d:nat)(x:t),(mem_fn 0 d x = false)->forall (i:nat),(mem_fn i d (low1 deBruijn_t d x))=(mem_fn (DeBruijnNat.lift 0 1 i) d x);
    low1_In_fn : forall (d':nat)(i:nat),(i<=d')->forall (d:nat)(x:t),(mem_fn 0 (d+S d') x = false)->(mem_fn i d x = mem_fn i d (low1 deBruijn_t (d+S d') x));
    low1_In_fn_cond : forall (d d':nat)(x:t)(i:nat),(mem_fn (DeBruijnNat.lift 0 1 i) (d+d') x = true)->(mem_fn i (d+d') (low1 deBruijn_t d' x))=true;
    fn_plus_lift : forall (d d':nat)(x:t),forall (i:nat),(mem_fn (DeBruijnNat.lift 0 d' i) d x) = (mem_fn i (d+d') x);
    low1_comm : forall (d d':nat)(x:t),(mem_fn 0 (d+S d') x = false)->(low1 deBruijn_t d (low1 deBruijn_t (d+S d') x))=(low1 deBruijn_t (d+d') (low1 deBruijn_t d x));
    low1_swapR_simpl : forall (d k:nat)(x:t),(mem_fn 0 (d+k) x = false)->(low1 deBruijn_t d (swapR deBruijn_t d k x))=(low1 deBruijn_t (d+k) x);
    (* properties of Nmin *)
    Nmin_true : forall (d k:nat)(x:t)(kmin:nat),Nmin d k x (true,kmin) = (true,kmin);
    Nmin_le_false : forall (d k:nat)(x:t)(kmin:nat)(b:bool)(kmin':nat),Nmin d k x (false,kmin)=(b,kmin')->(kmin' <= kmin);
    Nmin_lift : forall (d d' k k':nat)(x:t)(b:bool)(kmin:nat),(kmin <= k)->(Nmin d k x (b,kmin))=(Nmin d k (lift deBruijn_t (d+k+S d') k' x) (b,kmin));
    Nmin_swapR : forall (d d' k k':nat)(x:t)(b:bool)(kmin:nat),(kmin <= k)->(Nmin d k x (b,kmin))=(Nmin d k (swapR deBruijn_t (d+k+S d') k' x) (b,kmin));
    Nmin_low1 : forall (d d' k:nat)(x:t)(b:bool)(kmin:nat),(kmin <= k)->(mem_fn (k+S d') d x=false)->(Nmin d k x (b,kmin))=(Nmin d k (low1 deBruijn_t (d+k+S d') x) (b,kmin));
    Nmin_swapR_gt : forall (d k k1 k2:nat)(x:t)(b:bool)(kmin:nat),(k1 <= k2)->(k < k1)->(kmin <= k)->(Nmin d k (swapR deBruijn_t (d+k1) (k2-k1) x) (b,kmin))=(Nmin d k x (b,kmin));
    (* low1 is almost injective *)
    low1_injective : forall (d:nat)(x y:t),(mem_fn 0 d x = false)->(mem_fn 0 d y = false)->(low1 deBruijn_t d x = low1 deBruijn_t d y)->x = y
  }.

  Implicit Arguments mem_fn [t deBruijn_t].
  Implicit Arguments Nmin [t deBruijn_t].
  Implicit Arguments lift_low1_comm [t deBruijn_t].
  Implicit Arguments lift_1_low1_simpl [t deBruijn_t].
  Implicit Arguments lift_fn [t deBruijn_t].
  Implicit Arguments lift_In_fn [t deBruijn_t].
  Implicit Arguments lift_not_fn [t deBruijn_t].
  Implicit Arguments swapR_low1_comm [t deBruijn_t].
  Implicit Arguments swapR_In_fn [t deBruijn_t].
  Implicit Arguments swapR_eq_cond [t deBruijn_t].
  Implicit Arguments swapR_fn [t deBruijn_t].
  Implicit Arguments swapR_fn_same [t deBruijn_t].
  Implicit Arguments swapR_low1_comm2 [t deBruijn_t].
  Implicit Arguments low1_swapR_comm [t deBruijn_t].
  Implicit Arguments lift_low1_fn [t deBruijn_t].
  Implicit Arguments low1_In_fn [t deBruijn_t].
  Implicit Arguments low1_In_fn_cond [t deBruijn_t].
  Implicit Arguments fn_plus_lift [t deBruijn_t].
  Implicit Arguments low1_comm [t deBruijn_t].
  Implicit Arguments low1_swapR_simpl [t deBruijn_t].
  Implicit Arguments Nmin_true [t deBruijn_t].
  Implicit Arguments Nmin_le_false [t deBruijn_t].
  Implicit Arguments Nmin_lift [t deBruijn_t].
  Implicit Arguments Nmin_swapR [t deBruijn_t].
  Implicit Arguments Nmin_low1 [t deBruijn_t].
  Implicit Arguments Nmin_swapR_gt [t deBruijn_t].
  Implicit Arguments low1_injective [t deBruijn_t].

  Lemma Nmin_le : forall (t:Set)(deBruijn_t:deBruijn_base t)(deBruijn_fn_t:deBruijn_fn t deBruijn_t),forall (d k:nat)(x:t)(b:bool)(kmin:nat),(snd (Nmin deBruijn_fn_t d k x (b,kmin))) <= kmin. 
    intros.
    destruct b.
    rewrite Nmin_true;simpl;auto.
    eapply (Nmin_le_false deBruijn_fn_t d k x kmin).
    rewrite (pair_split (Nmin deBruijn_fn_t d k x (false,kmin))).
    reflexivity.
  Qed.

  Implicit Arguments Nmin_le [t deBruijn_t].

  Record fold_struct (t t':Set)(map_fun:(nat->t->t)->nat->t'->t')(map_ext:map_struct t t' map_fun)(fold_fun:forall (A:Set),(nat->t->A->A)->nat->t'->A->A) : Prop := {
    fold_orb_eq : forall (P:nat->Prop)(fi:nat->nat->t->bool)(g h:nat->t->t)(d1 d2 d3:nat),(forall (d:nat)(x:t),(forall (i:nat),(P i)->(fi i (d1+d) x)=false)->g (d2+d) x = h (d3+d) x)->(forall (d:nat)(x:t'),(forall (i:nat),(P i)->(fold_fun bool (fun d x b => b || (fi i d x)) (d1+d) x false = false))->(map_fun g (d2+d) x)=(map_fun h (d3+d) x));
    fold_exists : forall (f g:nat->t->bool)(d1 d2:nat),(forall (d:nat)(x:t),(f (d1+d) x = true)->(g (d2+d) x = true))->forall (d:nat)(x:t'),(fold_fun bool (fun d x b => b || f d x) (d1+d) x false = true)->(fold_fun bool (fun d x b => b || g d x) (d2+d) x false = true);
    fold_map_simpl : forall (A:Set)(f:nat->t->A->A)(g:nat->t->t)(d1 d2:nat),forall (d:nat)(x:t')(a:A),(fold_fun A f (d1+d) (map_fun g (d2+d) x) a)=(fold_fun A (fun d x a => f (d1+d) (g (d2+d) x) a) d x a);
    fold_equal_cond_invariant : forall (A:Set)(Q:A->Prop)(P:nat->Prop)(fi:nat->nat->t->bool)(g h:nat->t->A->A)(d1 d2 d3:nat),(forall (d:nat)(x:t)(a:A),(forall (i:nat),(P i)->(fi i (d1+d) x = false))->(Q a)->(g (d2+d) x a = h (d3+d) x a))->(forall (d:nat)(x:t)(a:A),(Q a)->(Q (g d x a)))->forall (d:nat)(x:t')(a:A),(forall (i:nat),(P i)->(fold_fun bool (fun d x b => b || (fi i d x)) (d1+d) x false) = false)->(Q a)->(fold_fun A g (d2+d) x a = fold_fun A h (d3+d) x a);
    fold_invariant : forall (A:Set)(P:A->Prop)(f:nat->t->A->A),(forall (d:nat)(x:t)(a:A),(P a)->(P (f d x a)))->forall (d:nat)(x:t')(a:A),(P a)->(P (fold_fun A f d x a));
    fold_orb_eq_map_inversion : forall (P:nat->Prop)(fi:nat->nat->t->bool)(g h:nat->t->t)(d1 d2 d3:nat),(forall (d:nat)(x y:t),(forall (i:nat),(P i)->(fi i (d1+d) x)=false)->(forall (i:nat),(P i)->(fi i (d1+d) y)=false)->g (d2+d) x = h (d3+d) y->x = y)->(forall (d:nat)(x y:t'),(forall (i:nat),(P i)->(fold_fun bool (fun d x b => b || (fi i d x)) (d1+d) x false = false))->(forall (i:nat),(P i)->(fold_fun bool (fun d x b => b || (fi i d x)) (d1+d) y false = false))->(map_fun g (d2+d) x)=(map_fun h (d3+d) y)->x = y)
  }.

  Implicit Arguments fold_orb_eq [t t' map_fun fold_fun].
  Implicit Arguments fold_exists [t t' map_fun fold_fun].
  Implicit Arguments fold_map_simpl [t t' map_fun fold_fun A].
  Implicit Arguments fold_equal_cond_invariant [t t' map_fun fold_fun A].
  Implicit Arguments fold_invariant [t t' map_fun fold_fun A].
  Implicit Arguments fold_orb_eq_map_inversion [t t' map_fun fold_fun].

  Lemma fold_neutral : forall (t t':Set)(map_fun:(nat->t->t)->nat->t'->t')(map_ext:map_struct t t' map_fun)(fold_fun:forall (A:Set),(nat->t->A->A)->nat->t'->A->A)(fold_ext:fold_struct t t' map_fun map_ext fold_fun),forall (A:Set)(f:nat->t->A->A)(a:A),(forall (d:nat)(x:t),(f d x a = a))->forall (d:nat)(x:t'),(fold_fun A f d x a)=a.
    intros.
    apply (fold_invariant map_ext fold_ext (fun x => x = a));auto.
    intros.
    rewrite H0.
    auto.
  Qed.

  Implicit Arguments fold_neutral [t t' map_fun fold_fun A].

  Lemma fold_equal_cond : forall (t t':Set)(map_fun:(nat->t->t)->nat->t'->t')(map_ext:map_struct t t' map_fun)(fold_fun:forall (A:Set),(nat->t->A->A)->nat->t'->A->A)(fold_ext:fold_struct t t' map_fun map_ext fold_fun),forall (A:Set)(P:nat->Prop)(fi:nat->nat->t->bool)(g h:nat->t->A->A)(d1 d2 d3:nat),(forall (d:nat)(x:t)(a:A),(forall (i:nat),(P i)->(fi i (d1+d) x = false))->(g (d2+d) x a = h (d3+d) x a))->forall (d:nat)(x:t')(a:A),(forall (i:nat),(P i)->(fold_fun bool (fun d x b => b || (fi i d x)) (d1+d) x false) = false)->(fold_fun A g (d2+d) x a = fold_fun A h (d3+d) x a).
    intros.
    apply (fold_equal_cond_invariant map_ext fold_ext (A:=A) (fun x => True) P fi g h d1 d2 d3);auto.
  Qed.

  Implicit Arguments fold_equal_cond [t t' map_fun fold_fun A].

  Lemma fold_equal_invariant : forall (t t':Set)(map_fun:(nat->t->t)->nat->t'->t')(map_ext:map_struct t t' map_fun)(fold_fun:forall (A:Set),(nat->t->A->A)->nat->t'->A->A)(fold_ext:fold_struct t t' map_fun map_ext fold_fun),forall (A:Set)(P:A->Prop)(f g:nat->t->A->A),(forall (d:nat)(x:t)(a:A),(P a)->(f d x a = g d x a))->(forall (d:nat)(x:t)(a:A),(P a)->(P (f d x a)))->forall (d:nat)(x:t')(a:A),(P a)->(fold_fun A f d x a)=(fold_fun A g d x a).
    intros.
    generalize (fold_equal_cond_invariant map_ext fold_ext (A:=A) P (fun i => False) (fun i d x => false) f g 0 0 0);simpl;intros;auto.
    apply H2;auto.
    intros.
    apply (fold_neutral map_ext fold_ext (fun d x b => b || false) false);auto.
  Qed.

  Implicit Arguments fold_equal_invariant [t t' map_fun fold_fun A].

  Lemma fold_equal : forall (t t':Set)(map_fun:(nat->t->t)->nat->t'->t')(map_ext:map_struct t t' map_fun)(fold_fun:forall (A:Set),(nat->t->A->A)->nat->t'->A->A)(fold_ext:fold_struct t t' map_fun map_ext fold_fun),forall (A:Set)(f g:nat->t->A->A),(forall (d:nat)(x:t)(a:A),(f d x a = g d x a))->forall (d:nat)(x:t')(a:A),(fold_fun A f d x a = fold_fun A g d x a).
    intros.
    generalize (fold_equal_cond map_ext fold_ext (fun x => False) (fun i d x => false) f g 0 0 0);simpl.
    intros.
    apply H0.
    intros.
    apply H.
    intros.
    tauto.
  Qed.

  Implicit Arguments fold_equal [t t' map_fun fold_fun A].
  
  Definition deBruijn_fn_map_struct (t t':Set)(map_fun:(nat->t->t)->nat->t'->t')(map_ext:map_struct t t' map_fun)(fold_fun:forall (A:Set),(nat->t->A->A)->nat->t'->A->A)(deBruijn_t:deBruijn_base t)(deBruijn_fn_t:deBruijn_fn t deBruijn_t)(fold_ext:fold_struct t t' map_fun map_ext fold_fun) : (deBruijn_fn t' (deBruijn_map_struct t deBruijn_t t' map_fun map_ext)).
    intros.
    split with (fun k d y => fold_fun bool (fun d x b => orb b (mem_fn deBruijn_fn_t k d x)) d y false) (fun d k y accu => fold_fun ((bool*nat)%type) (fun d x accu => Nmin deBruijn_fn_t d k x accu) d y accu).
    (* lift_low1_comm *)
    intros;simpl.
    generalize (map_compose_simpl map_ext 0 (S d') 0 (fun d => low1 deBruijn_t d) (fun d => lift deBruijn_t d k) (fun d x => low1 deBruijn_t d (lift deBruijn_t (d+S d') k x)));simpl;intros.
    rewrite <- plus_n_Sm.
    rewrite plus_comm.
    rewrite H0;try (intros;rewrite <- plus_n_Sm;rewrite plus_comm;reflexivity).
    generalize (map_compose_simpl map_ext d' 0 0 (fun d => lift deBruijn_t d k) (fun d => low1 deBruijn_t d) (fun d x => lift deBruijn_t (d+d') k (low1 deBruijn_t d x)));simpl;intros.
    rewrite H1;try (intros;rewrite plus_comm;reflexivity).
    generalize (fold_orb_eq map_ext fold_ext (fun _ => True) (fun _ d x => mem_fn deBruijn_fn_t 0 d x) (fun d x => low1 deBruijn_t d (lift deBruijn_t (d+S d')k x)) (fun d x => lift deBruijn_t (d+d') k (low1 deBruijn_t d x)) 0 0 0);simpl;intros.
    apply H2;intros;auto.
    eapply lift_low1_comm;eauto.
    (* lift_1_low1_simpl *)
    intros;simpl.
    generalize (map_compose_simpl map_ext 0 k 0 (fun d => lift deBruijn_t d 1) (fun d => low1 deBruijn_t d) (fun d x => lift deBruijn_t d 1 (low1 deBruijn_t (d+k) x)));simpl;intros.
    rewrite plus_comm.
    rewrite H0;try (intros;rewrite plus_comm;reflexivity).
    generalize (fold_orb_eq map_ext fold_ext (fun _ => True) (fun _ d x => mem_fn deBruijn_fn_t 0 d x) (fun d x => lift deBruijn_t d 1 (low1 deBruijn_t (d+k) x)) (fun d => swapR deBruijn_t d k) k 0 0);simpl;intros.
    apply H1;intros;auto.
    apply (lift_1_low1_simpl deBruijn_fn_t).
    rewrite plus_comm;auto.
    rewrite plus_comm;auto.
    (* lift_fn *)
    intros;simpl.
    generalize (fold_map_simpl map_ext fold_ext (fun d x b => b || mem_fn deBruijn_fn_t (DeBruijnNat.lift 0 k i) d x) (fun d => lift deBruijn_t d k) 0 0);simpl;intros.
    rewrite H.
    apply (fold_equal map_ext fold_ext);intros.
    rewrite lift_fn;auto.
    (* lift_In_fn *)
    intros;simpl.
    generalize (fold_map_simpl map_ext fold_ext (fun d x b => b || mem_fn deBruijn_fn_t i d x) (fun d => lift deBruijn_t d k) 0 (S d'));simpl;intros.
    rewrite plus_comm;simpl.
    rewrite H0.
    apply (fold_equal map_ext fold_ext);intros.
    rewrite (lift_In_fn deBruijn_fn_t d' i H d0 k x0);auto;rewrite plus_comm;simpl;auto.
    (* lift_not_fn *)
    intros;simpl.
    generalize (fold_map_simpl map_ext fold_ext (fun d x b => b || mem_fn deBruijn_fn_t i d x) (fun d => lift deBruijn_t d k) 0 0);simpl;intros.
    rewrite H0.
    apply (fold_neutral map_ext fold_ext).
    intros;simpl.
    apply lift_not_fn;auto.
    (* swapR_low1_comm *)
    intros.
    simpl.
    generalize (map_compose_simpl map_ext 0 (S d') 0 (fun d => low1 deBruijn_t d) (fun d => swapR deBruijn_t d k) (fun d x => low1 deBruijn_t d (swapR deBruijn_t (d+S d') k x)));simpl;intros.
    rewrite <- plus_n_Sm.
    rewrite plus_comm.
    rewrite H0;try (intros;rewrite <- plus_n_Sm;rewrite plus_comm;reflexivity).
    generalize (map_compose_simpl map_ext d' 0 0 (fun d => swapR deBruijn_t d k) (fun d => low1 deBruijn_t d) (fun d x => swapR deBruijn_t (d+d') k (low1 deBruijn_t d x)));simpl;intros.
    rewrite H1;try (intros;rewrite plus_comm;reflexivity).
    generalize (fold_orb_eq map_ext fold_ext (fun _ => True) (fun _ d x => mem_fn deBruijn_fn_t 0 d x) (fun d x => low1 deBruijn_t d (swapR deBruijn_t (d+S d') k x)) (fun d x => swapR deBruijn_t (d+d') k (low1 deBruijn_t d x)) 0 0 0);simpl;intros.
    apply H2;intros;auto.
    apply (swapR_low1_comm deBruijn_fn_t);auto.
    (* swapR_in_fn *)
    intros;simpl.
    generalize (fold_map_simpl map_ext fold_ext (fun d x b => b || mem_fn deBruijn_fn_t i d x) (fun d => swapR deBruijn_t d k) 0 (S d'));simpl;intros.
    rewrite plus_comm;simpl.
    rewrite H0.
    apply (fold_equal map_ext fold_ext).
    intros.
    rewrite (swapR_In_fn deBruijn_fn_t d' i H d0 k x0);auto;rewrite plus_comm;simpl;auto.
    (* swapR_eq_cond *)
    intros;simpl.
    generalize (fold_orb_eq map_ext fold_ext (fun i => i <= k) (fun i d x => mem_fn deBruijn_fn_t i d x) (fun d => swapR deBruijn_t d k) (fun d x => x) 0 0 0);simpl;intros.
    rewrite H0.
    apply (map_id map_ext).
    intros.
    eapply swapR_eq_cond.
    intros.
    apply H1;trivial.
    auto.
    (* swapR_fn *)
    intros;simpl.
    generalize (fold_map_simpl map_ext fold_ext (fun d x b => b || mem_fn deBruijn_fn_t (DeBruijnNat.swapR 0 k i) d x) (fun d => swapR deBruijn_t d k) 0 0);simpl;intros.
    rewrite H.
    apply (fold_equal map_ext fold_ext);intros.
    rewrite swapR_fn;auto.
    (* swapR_fn_same *)
    intros;simpl.
    generalize (fold_map_simpl map_ext fold_ext (fun d x b => b || mem_fn deBruijn_fn_t i d x) (fun d => swapR deBruijn_t d k) (k+S d') 0);simpl;intros.
    rewrite <- plus_assoc;rewrite plus_comm.
    rewrite H.
    generalize (fold_map_simpl map_ext fold_ext (fun d x b => b || mem_fn deBruijn_fn_t i d x) (fun d x => x) (k+S d') 0);simpl;intros.
    generalize (H0 d x false).
    rewrite (map_id map_ext);intro.
    rewrite H1.
    apply (fold_equal map_ext fold_ext);intros.
    rewrite (plus_comm (k+S d'));rewrite plus_assoc.
    rewrite swapR_fn_same;auto.
    (* swapR_low1_comm2 *)
    intros.
    simpl.
    generalize (map_compose_simpl map_ext d' 0 0 (fun d => swapR deBruijn_t d k) (fun d => low1 deBruijn_t d) (fun d x => swapR deBruijn_t (d+d') k (low1 deBruijn_t d x)));simpl;intros.
    rewrite plus_comm.
    rewrite H0;try (intros;rewrite plus_comm;reflexivity).
    generalize (map_compose_simpl map_ext 0 (S d') 0 (fun d => low1 deBruijn_t d) (fun d => swapR deBruijn_t d k) (fun d x => low1 deBruijn_t d (swapR deBruijn_t (d+S d') k x)));simpl;intros.
    rewrite <- plus_n_Sm.
    rewrite plus_comm.
    rewrite H1;try (intros;rewrite <- plus_n_Sm;rewrite plus_comm;reflexivity).
    generalize (fold_orb_eq map_ext fold_ext (fun i => True) (fun _ d x => mem_fn deBruijn_fn_t 0 d x) (fun d x => swapR deBruijn_t (d+d') k (low1 deBruijn_t d x)) (fun d x => low1 deBruijn_t d (swapR deBruijn_t (d+S d') k x)) 0 0 0);simpl;intros.
    apply H2;intros;auto.
    eapply swapR_low1_comm2;eauto.
    (* low1_swapR_comm *)
    intros;simpl.
    generalize (map_compose_simpl map_ext (k+S d') 0 0 (fun d => low1 deBruijn_t d) (fun d => swapR deBruijn_t d k) (fun d x => low1 deBruijn_t (d+k+S d') (swapR deBruijn_t d k x)));simpl;intros.
    rewrite <- plus_assoc;rewrite plus_comm.
    rewrite H0;try (intros;rewrite plus_comm;rewrite plus_assoc;reflexivity).
    generalize (map_compose_simpl map_ext 0 (k+S d') 0 (fun d => swapR deBruijn_t d k) (fun d => low1 deBruijn_t d) (fun d x => swapR deBruijn_t d k (low1 deBruijn_t (d+k+S d') x)));simpl;intros.
    rewrite H1;try (intros;rewrite plus_comm;rewrite plus_assoc;reflexivity).
    generalize (fold_orb_eq map_ext fold_ext (fun i => True) (fun _ d x => mem_fn deBruijn_fn_t 0 d x) (fun d x => low1 deBruijn_t (d+k+S d') (swapR deBruijn_t d k x)) (fun d x => swapR deBruijn_t d k (low1 deBruijn_t (d+k+S d') x)) (k+S d') 0 0 );simpl;intros.
    apply H2;intros;auto.
    apply (low1_swapR_comm deBruijn_fn_t);auto.
    rewrite <- plus_assoc;rewrite plus_comm;auto.
    rewrite (plus_comm (k+S d'));rewrite plus_assoc;auto.
    (* lift_low1_fn *)
    intros;simpl.
    generalize (fold_map_simpl map_ext fold_ext (fun d x b => b || mem_fn deBruijn_fn_t i d x) (fun d => low1 deBruijn_t d) 0 0);simpl;intros.
    rewrite H0.
    generalize (fold_equal_cond map_ext fold_ext (fun i => True) (fun i d x => mem_fn deBruijn_fn_t 0 d x) (fun d x b => b || mem_fn deBruijn_fn_t i d (low1 deBruijn_t d x)) (fun d x b => b || mem_fn deBruijn_fn_t (DeBruijnNat.lift 0 1 i) d x) 0 0 0);simpl;intros.
    apply H1;intros;auto.
    rewrite lift_low1_fn;auto.
    (* low1_In_fn *)
    intros;simpl.
    generalize (fold_map_simpl map_ext fold_ext (fun d x b => b || mem_fn deBruijn_fn_t i d x) (fun d => low1 deBruijn_t d) 0 (S d'));simpl;intros.
    rewrite plus_comm;simpl.
    rewrite H1.
    generalize (fold_equal_cond map_ext fold_ext (fun i => True) (fun _ d x => mem_fn deBruijn_fn_t 0 d x) (fun d x b => b || mem_fn deBruijn_fn_t i d x) (fun d x b => b || mem_fn deBruijn_fn_t i d (low1 deBruijn_t (S (d'+d)) x)) (S d') 0 0);simpl;intros.
    apply H2;intros;auto.
    rewrite plus_comm;rewrite plus_n_Sm.
    rewrite <- low1_In_fn;auto.
    rewrite plus_comm in H3;rewrite plus_n_Sm in H3;auto.
    rewrite plus_comm;rewrite plus_n_Sm;auto.
    (* low1_In_fn_cond *)
    intros;simpl.
    generalize (fold_map_simpl map_ext fold_ext (fun d x b => b || mem_fn deBruijn_fn_t i d x) (fun d => low1 deBruijn_t d) d 0);simpl;intros.
    rewrite H0.
    generalize (fold_exists map_ext fold_ext (fun d x => mem_fn deBruijn_fn_t (DeBruijnNat.lift 0 1 i) d x) (fun d0 x => mem_fn deBruijn_fn_t i (d+d0) (low1 deBruijn_t d0 x)) d 0);simpl;intros.
    apply H1;intros;auto.
    apply low1_In_fn_cond;auto.
    (* fn_plus_lift *)
    intros.
    generalize (fold_map_simpl map_ext fold_ext (fun d x b => b || mem_fn deBruijn_fn_t i d x) (fun d x => x) d' 0 d x false);simpl;intros.
    rewrite (map_id map_ext) in H.
    rewrite plus_comm.
    rewrite H.
    apply (fold_equal map_ext fold_ext);intros.
    rewrite fn_plus_lift;rewrite plus_comm;auto.
    (* low1_comm *)
    intros;simpl.
    generalize (map_compose_simpl map_ext 0 (S d') 0 (fun d => low1 deBruijn_t d) (fun d => low1 deBruijn_t d) (fun d x => low1 deBruijn_t d (low1 deBruijn_t (d+S d') x)));simpl;intros.
    rewrite <- plus_n_Sm.
    rewrite plus_comm.
    rewrite H0;try (intros;rewrite <- plus_n_Sm;rewrite plus_comm;reflexivity).
    generalize (map_compose_simpl map_ext d' 0 0 (fun d => low1 deBruijn_t d) (fun d => low1 deBruijn_t d) (fun d x => low1 deBruijn_t (d+d') (low1 deBruijn_t d x)));simpl;intros.
    rewrite H1;try (intros;rewrite plus_comm;reflexivity).
    generalize (fold_orb_eq map_ext fold_ext (fun _ => True) (fun _ d x => mem_fn deBruijn_fn_t 0 d x) (fun d x => low1 deBruijn_t d (low1 deBruijn_t (d+S d') x)) (fun d x => low1 deBruijn_t (d+d') (low1 deBruijn_t d x)) (S d') 0 0);simpl;intros.
    apply H2;intros;auto.
    eapply low1_comm.
    rewrite plus_comm;simpl;eauto.
    rewrite plus_comm;rewrite plus_n_Sm;auto.
    (* low1_swapR_simpl *)
    intros;simpl.
    generalize (map_compose_simpl map_ext 0 0 0 (fun d => low1 deBruijn_t d) (fun d => swapR deBruijn_t d k) (fun d x => low1 deBruijn_t d (swapR deBruijn_t d k x)));simpl;intros.
    rewrite H0;trivial.
    generalize (fold_orb_eq map_ext fold_ext (fun _ => True) (fun _ d x => mem_fn deBruijn_fn_t 0 d x) (fun d x => low1 deBruijn_t d (swapR deBruijn_t d k x)) (fun d => low1 deBruijn_t d) k 0 k);simpl;intros.
    rewrite plus_comm;apply H1;intros;auto.
    rewrite plus_comm.
    rewrite (low1_swapR_simpl deBruijn_fn_t);auto.
    rewrite plus_comm;auto.
    rewrite plus_comm;auto.
    (* Nmin_true *)
    intros.
    apply (fold_neutral map_ext fold_ext).
    intros.
    apply Nmin_true.
    (* Nmin_le_false *)
    intros.
    generalize (fold_invariant map_ext fold_ext (fun accu:bool*nat => let (b,k) := accu in k <= kmin) (fun d x accu => Nmin deBruijn_fn_t d k x accu)).
    intros.
    assert (forall (d:nat)(x:t)(a:bool*nat),(let (_,k) := a in k <= kmin)->(let (_,k) := Nmin deBruijn_fn_t d k x a in k <= kmin)).
    intros.
    destruct a as [ba ka].
    destruct ba.
    rewrite Nmin_true;auto.
    generalize (Nmin_le_false deBruijn_fn_t d0 k x0 ka).
    destruct (Nmin deBruijn_fn_t d0 k x0 (false,ka));intros.
    generalize (H2 b0 n (refl_equal (b0,n)));intros.
    omega.
    generalize (H0 H1 d x (false,kmin)).
    rewrite H;auto.
    (* Nmin_lift *)
    intros.
    simpl.
    generalize (fold_map_simpl map_ext fold_ext (fun d x accu => Nmin deBruijn_fn_t d k x accu) (fun d => lift deBruijn_t d k') 0 (k+S d'));simpl;intros.
    rewrite <- plus_assoc;rewrite plus_comm.
    rewrite H0.
    apply (fold_equal_invariant map_ext fold_ext (fun accu:bool*nat => let (_,km) := accu in km <= k));auto.
    intros.
    destruct a.
    rewrite (plus_comm (k+S d'));rewrite plus_assoc.
    apply (Nmin_lift deBruijn_fn_t);auto.
    intros.
    destruct a.
    destruct b0.
    rewrite (Nmin_true deBruijn_fn_t);auto.
    generalize (Nmin_le_false deBruijn_fn_t d0 k x0 n).
    destruct (Nmin deBruijn_fn_t d0 k x0 (false,n)).
    intros.
    generalize (H2 b0 n0 (refl_equal (b0,n0)));intro.
    omega.
    (* Nmin_swapR *)
    intros.
    simpl.
    generalize (fold_map_simpl map_ext fold_ext (fun d x accu => Nmin deBruijn_fn_t d k x accu) (fun d => swapR deBruijn_t d k') 0 (k+S d'));simpl;intros.
    rewrite <- plus_assoc;rewrite plus_comm.
    rewrite H0.
    apply (fold_equal_invariant map_ext fold_ext (fun accu:bool*nat => let (_,km) := accu in km <= k));auto.
    intros.
    destruct a.
    rewrite (plus_comm (k+S d'));rewrite plus_assoc.
    apply (Nmin_swapR deBruijn_fn_t);auto.
    intros.
    destruct a.
    destruct b0.
    rewrite (Nmin_true deBruijn_fn_t);auto.
    generalize (Nmin_le_false deBruijn_fn_t d0 k x0 n).
    destruct (Nmin deBruijn_fn_t d0 k x0 (false,n)).
    intros.
    generalize (H2 b0 n0 (refl_equal (b0,n0)));intro.
    omega.
    (* Nmin_low1 *)
    intros.
    simpl.
    generalize (fold_map_simpl map_ext fold_ext (fun d x accu => Nmin deBruijn_fn_t d k x accu) (fun d => low1 deBruijn_t d) 0 (k+S d'));simpl;intros.
    rewrite <- plus_assoc;rewrite plus_comm.
    rewrite H1.
    generalize (fold_equal_cond_invariant map_ext fold_ext (fun accu : bool*nat => let (_,km) := accu in km <= k) (fun _ => True) (fun _ d x => mem_fn deBruijn_fn_t (k+S d') d x) (fun d x accu => Nmin deBruijn_fn_t d k x accu) (fun d x accu => Nmin deBruijn_fn_t d k (low1 deBruijn_t (k+S d'+d) x) accu) 0 0 0);simpl;intro.
    apply H2;intros;auto.
    rewrite (plus_comm (k+S d'));rewrite plus_assoc.
    destruct a.
    apply (Nmin_low1 deBruijn_fn_t);auto.
    destruct a.
    destruct b0.
    rewrite Nmin_true;auto.
    generalize (Nmin_le_false deBruijn_fn_t d0 k x0 n).
    destruct (Nmin deBruijn_fn_t d0 k x0 (false,n));intros.
    generalize (H4 b0 n0 (refl_equal (b0,n0)));intro;omega.
    (* Nmin_swapR_gt *)
    intros;simpl.
    generalize (fold_map_simpl map_ext fold_ext (fun d x accu => Nmin deBruijn_fn_t d k x accu) (fun d => swapR deBruijn_t d (k2-k1)) 0 k1);simpl;intros.
    rewrite plus_comm.
    rewrite H2.
    apply (fold_equal_invariant map_ext fold_ext (fun accu : bool*nat => let (_,km) := accu in km <= k));intros;auto.
    rewrite plus_comm.
    destruct a.
    rewrite (Nmin_swapR_gt deBruijn_fn_t d0 k k1 k2);auto.
    destruct a.
    rewrite (pair_split (Nmin deBruijn_fn_t d0 k (swapR deBruijn_t (k1+d0) (k2-k1) x0) (b0,n))).
    apply le_trans with n.
    apply Nmin_le;auto.
    auto.
    (* low1_injective *)
    simpl;intros.
    eapply (fold_orb_eq_map_inversion map_ext fold_ext (fun x => True) (fun i d x => mem_fn deBruijn_fn_t 0 d x) (fun d => low1 deBruijn_t d) (fun d => low1 deBruijn_t d) 0 0 0);simpl;intros;eauto.
    eapply (low1_injective deBruijn_fn_t);eauto.
  Defined.

  Record fold2_struct (t1 t2 t:Set)(map2_fun:(nat->t1->t1)->(nat->t2->t2)->nat->t->t)(map2_ext:map2_struct t1 t2 t map2_fun)(fold2_fun:forall (A:Set),(nat->t1->A->A)->(nat->t2->A->A)->nat->t->A->A) : Prop := {
    fold2_orb_eq : forall (P:nat->Prop)(fi:nat->nat->t1->bool)(g1 h1:nat->t1->t1)(f'i:nat->nat->t2->bool)(g2 h2:nat->t2->t2)(d1 d2 d3:nat),(forall (d:nat)(x:t1),(forall (i:nat),(P i)->(fi i (d1+d) x)=false)->g1 (d2+d) x = h1 (d3+d) x)->(forall (d:nat)(x:t2),(forall (i:nat),(P i)->(f'i i (d1+d) x)=false)->g2 (d2+d) x = h2 (d3+d) x)->(forall (d:nat)(x:t),(forall (i:nat),(P i)->(fold2_fun bool (fun d x b => b || (fi i d x)) (fun d x b => b || (f'i i d x)) (d1+d) x false = false))->(map2_fun g1 g2 (d2+d) x)=(map2_fun h1 h2 (d3+d) x));
    fold2_exists : forall (f1 g1:nat->t1->bool)(f2 g2:nat->t2->bool)(d1 d2:nat),(forall (d:nat)(x:t1),(f1 (d1+d) x = true)->(g1 (d2+d) x = true))->(forall (d:nat)(x:t2),(f2 (d1+d) x = true)->(g2 (d2+d) x)=true)->forall (d:nat)(x:t),(fold2_fun bool (fun d x b => b || f1 d x) (fun d x b => b || f2 d x) (d1+d) x false = true)->(fold2_fun bool (fun d x b => b || g1 d x) (fun d x b => b || g2 d x) (d2+d) x false = true);
    fold2_map2_simpl : forall (A:Set)(f1:nat->t1->A->A)(g1:nat->t1->t1)(f2:nat->t2->A->A)(g2:nat->t2->t2)(d1 d2:nat),forall (d:nat)(x:t)(a:A),(fold2_fun A f1 f2 (d1+d) (map2_fun g1 g2 (d2+d) x) a)=(fold2_fun A (fun d x a => f1 (d1+d) (g1 (d2+d) x) a) (fun d x a => f2 (d1+d) (g2 (d2+d) x) a) d x a);
    fold2_equal_cond_invariant : forall (A:Set)(Q:A->Prop)(P:nat->Prop)(fi:nat->nat->t1->bool)(g1 h1:nat->t1->A->A)(f'i:nat->nat->t2->bool)(g2 h2:nat->t2->A->A)(d1 d2 d3:nat),(forall (d:nat)(x:t1)(a:A),(forall (i:nat),(P i)->(fi i (d1+d) x = false))->(Q a)->(g1 (d2+d) x a = h1 (d3+d) x a))->(forall (d:nat)(x:t2)(a:A),(forall (i:nat),(P i)->(f'i i (d1+d) x = false))->(Q a)->(g2 (d2+d) x a = h2 (d3+d) x a))->(forall (d:nat)(x:t1)(a:A),(Q a)->(Q (g1 d x a)))->(forall (d:nat)(x:t2)(a:A),(Q a)->(Q (g2 d x a)))->forall (d:nat)(x:t)(a:A),(forall (i:nat),(P i)->(fold2_fun bool (fun d x b => b || (fi i d x)) (fun d x b => b || (f'i i d x)) (d1+d) x false) = false)->(Q a)->(fold2_fun A g1 g2 (d2+d) x a = fold2_fun A h1 h2 (d3+d) x a);
    fold2_invariant : forall (A:Set)(P:A->Prop)(f1:nat->t1->A->A)(f2:nat->t2->A->A),(forall (d:nat)(x:t1)(a:A),(P a)->(P (f1 d x a)))->(forall (d:nat)(x:t2)(a:A),(P a)->(P (f2 d x a)))->forall (d:nat)(x:t)(a:A),(P a)->(P (fold2_fun A f1 f2 d x a));
    fold2_orb_eq_map2_inversion : forall (P:nat->Prop)(fi:nat->nat->t1->bool)(g1 h1:nat->t1->t1)(f'i:nat->nat->t2->bool)(g2 h2:nat->t2->t2)(d1 d2 d3:nat),(forall (d:nat)(x y:t1),(forall (i:nat),(P i)->(fi i (d1+d) x)=false)->(forall (i:nat),(P i)->(fi i (d1+d) y)=false)->g1 (d2+d) x = h1 (d3+d) y->x=y)->(forall (d:nat)(x y:t2),(forall (i:nat),(P i)->(f'i i (d1+d) x)=false)->(forall (i:nat),(P i)->(f'i i (d1+d) y)=false)->g2 (d2+d) x = h2 (d3+d) y->x=y)->(forall (d:nat)(x y:t),(forall (i:nat),(P i)->(fold2_fun bool (fun d x b => b || (fi i d x)) (fun d x b => b || (f'i i d x)) (d1+d) x false = false))->(forall (i:nat),(P i)->(fold2_fun bool (fun d x b => b || (fi i d x)) (fun d x b => b || (f'i i d x)) (d1+d) y false = false))->(map2_fun g1 g2 (d2+d) x)=(map2_fun h1 h2 (d3+d) y)->x=y)
  }.

  Implicit Arguments fold2_orb_eq [t1 t2 t map2_fun fold2_fun].
  Implicit Arguments fold2_exists [t1 t2 t map2_fun fold2_fun].
  Implicit Arguments fold2_map2_simpl [t1 t2 t map2_fun fold2_fun A].
  Implicit Arguments fold2_equal_cond_invariant [t1 t2 t map2_fun fold2_fun A].
  Implicit Arguments fold2_invariant [t1 t2 t map2_fun fold2_fun A].
  Implicit Arguments fold2_orb_eq_map2_inversion [t1 t2 t map2_fun fold2_fun].

  Lemma fold2_neutral : forall (t1 t2 t:Set)(map2_fun:(nat->t1->t1)->(nat->t2->t2)->nat->t->t)(map2_ext:map2_struct t1 t2 t map2_fun)(fold2_fun:forall (A:Set),(nat->t1->A->A)->(nat->t2->A->A)->nat->t->A->A)(fold2_ext:fold2_struct t1 t2 t map2_fun map2_ext fold2_fun),forall (A:Set)(f1:nat->t1->A->A)(f2:nat->t2->A->A)(a:A),(forall (d:nat)(x:t1),(f1 d x a = a))->(forall (d:nat)(x:t2),(f2 d x a = a))->forall (d:nat)(x:t),(fold2_fun A f1 f2 d x a)=a.
    intros.
    apply (fold2_invariant map2_ext fold2_ext (fun x => x = a));auto;intros;rewrite H1;auto.
  Qed.

  Implicit Arguments fold2_neutral [t1 t2 t map2_fun fold2_fun A].

  Lemma fold2_equal_cond : forall (t1 t2 t:Set)(map2_fun:(nat->t1->t1)->(nat->t2->t2)->nat->t->t)(map2_ext:map2_struct t1 t2 t map2_fun)(fold2_fun:forall (A:Set),(nat->t1->A->A)->(nat->t2->A->A)->nat->t->A->A)(fold2_ext:fold2_struct t1 t2 t map2_fun map2_ext fold2_fun),forall (A:Set)(P:nat->Prop)(fi:nat->nat->t1->bool)(g1 h1:nat->t1->A->A)(f'i:nat->nat->t2->bool)(g2 h2:nat->t2->A->A)(d1 d2 d3:nat),(forall (d:nat)(x:t1)(a:A),(forall (i:nat),(P i)->(fi i (d1+d) x = false))->(g1 (d2+d) x a = h1 (d3+d) x a))->(forall (d:nat)(x:t2)(a:A),(forall (i:nat),(P i)->(f'i i (d1+d) x = false))->(g2 (d2+d) x a = h2 (d3+d) x a))->forall (d:nat)(x:t)(a:A),(forall (i:nat),(P i)->(fold2_fun bool (fun d x b => b || (fi i d x)) (fun d x b => b || (f'i i d x)) (d1+d) x false) = false)->(fold2_fun A g1 g2 (d2+d) x a = fold2_fun A h1 h2 (d3+d) x a).
    intros.
    apply (fold2_equal_cond_invariant map2_ext fold2_ext (A:=A) (fun x => True) P fi g1 h1 f'i g2 h2 d1 d2 d3);auto.
  Qed.

  Implicit Arguments fold2_equal_cond [t1 t2 t map2_fun fold2_fun A].

  Lemma fold2_equal_invariant : forall (t1 t2 t:Set)(map2_fun:(nat->t1->t1)->(nat->t2->t2)->nat->t->t)(map2_ext:map2_struct t1 t2 t map2_fun)(fold2_fun:forall (A:Set),(nat->t1->A->A)->(nat->t2->A->A)->nat->t->A->A)(fold2_ext:fold2_struct t1 t2 t map2_fun map2_ext fold2_fun),forall (A:Set)(P:A->Prop)(f1 g1:nat->t1->A->A)(f2 g2:nat->t2->A->A),(forall (d:nat)(x:t1)(a:A),(P a)->(f1 d x a = g1 d x a))->(forall (d:nat)(x:t2)(a:A),(P a)->(f2 d x a = g2 d x a))->(forall (d:nat)(x:t1)(a:A),(P a)->(P (f1 d x a)))->(forall (d:nat)(x:t2)(a:A),(P a)->(P (f2 d x a)))->forall (d:nat)(x:t)(a:A),(P a)->(fold2_fun A f1 f2 d x a)=(fold2_fun A g1 g2 d x a).
    intros.
    generalize (fold2_equal_cond_invariant map2_ext fold2_ext (A:=A) P (fun i => False) (fun i d x => false) f1 g1 (fun i d x => false) f2 g2 0 0 0);simpl;intros;auto.
    apply H4;auto.
    intros.
    apply (fold2_neutral map2_ext fold2_ext (fun d x b => b || false) (fun d x b => b || false) false);auto.
  Qed.

  Implicit Arguments fold2_equal_invariant [t1 t2 t map2_fun fold2_fun A].

  Lemma fold2_equal : forall (t1 t2 t:Set)(map2_fun:(nat->t1->t1)->(nat->t2->t2)->nat->t->t)(map2_ext:map2_struct t1 t2 t map2_fun)(fold2_fun:forall (A:Set),(nat->t1->A->A)->(nat->t2->A->A)->nat->t->A->A)(fold2_ext:fold2_struct t1 t2 t map2_fun map2_ext fold2_fun),forall (A:Set)(f1 g1:nat->t1->A->A)(f2 g2:nat->t2->A->A),(forall (d:nat)(x:t1)(a:A),(f1 d x a = g1 d x a))->(forall (d:nat)(x:t2)(a:A),(f2 d x a = g2 d x a))->forall (d:nat)(x:t)(a:A),(fold2_fun A f1 f2 d x a = fold2_fun A g1 g2 d x a).
    intros.
    generalize (fold2_equal_cond map2_ext fold2_ext (fun x => False) (fun i d x => false) f1 g1 (fun i d x => false) f2 g2 0 0 0);simpl.
    intros.
    apply H1;intros;auto.
    tauto.
  Qed.

  Implicit Arguments fold2_equal [t1 t2 t map2_fun fold2_fun A].
  
  Definition deBruijn_fn_map2_struct (t1 t2 t:Set)(map2_fun:(nat->t1->t1)->(nat->t2->t2)->nat->t->t)(map2_ext:map2_struct t1 t2 t map2_fun)(fold2_fun:forall (A:Set),(nat->t1->A->A)->(nat->t2->A->A)->nat->t->A->A)(deBruijn_t1:deBruijn_base t1)(deBruijn_fn_t1:deBruijn_fn t1 deBruijn_t1)(deBruijn_t2:deBruijn_base t2)(deBruijn_fn_t2:deBruijn_fn t2 deBruijn_t2)(fold2_ext:fold2_struct t1 t2 t map2_fun map2_ext fold2_fun) : (deBruijn_fn t (deBruijn_map2_struct t1 deBruijn_t1 t2 deBruijn_t2 t map2_fun map2_ext)).
    intros.
    split with (fun k d y => fold2_fun bool (fun d x b => orb b (mem_fn deBruijn_fn_t1 k d x)) (fun d x b => orb b (mem_fn deBruijn_fn_t2 k d x)) d y false) (fun d k y accu => fold2_fun ((bool*nat)%type) (fun d x accu => Nmin deBruijn_fn_t1 d k x accu) (fun d x accu => Nmin deBruijn_fn_t2 d k x accu) d y accu).
    (* lift_low1_comm *)
    intros;simpl.
    generalize (map2_compose_simpl map2_ext 0 (S d') 0 (fun d => low1 deBruijn_t1 d) (fun d => lift deBruijn_t1 d k) (fun d x => low1 deBruijn_t1 d (lift deBruijn_t1 (d+S d') k x)) (fun d => low1 deBruijn_t2 d) (fun d => lift deBruijn_t2 d k) (fun d x => low1 deBruijn_t2 d (lift deBruijn_t2 (d+S d') k x)));simpl;intros.
    repeat (rewrite <- plus_n_Sm).
    rewrite plus_comm.
    rewrite H0;try (intros;repeat (rewrite <- plus_n_Sm);rewrite plus_comm;reflexivity).
    generalize (map2_compose_simpl map2_ext d' 0 0 (fun d => lift deBruijn_t1 d k) (fun d => low1 deBruijn_t1 d) (fun d x => lift deBruijn_t1 (d+d') k (low1 deBruijn_t1 d x)) (fun d => lift deBruijn_t2 d k) (fun d => low1 deBruijn_t2 d) (fun d x => lift deBruijn_t2 (d+d') k (low1 deBruijn_t2 d x)));simpl;intros.
    rewrite H1;try (intros;rewrite plus_comm;reflexivity).
    generalize (fold2_orb_eq map2_ext fold2_ext (fun _ => True) (fun _ d x => mem_fn deBruijn_fn_t1 0 d x) (fun d x => low1 deBruijn_t1 d (lift deBruijn_t1 (d+S d')k x)) (fun d x => lift deBruijn_t1 (d+d') k (low1 deBruijn_t1 d x)) (fun _ d x => mem_fn deBruijn_fn_t2 0 d x) (fun d x => low1 deBruijn_t2 d (lift deBruijn_t2 (d+S d')k x)) (fun d x => lift deBruijn_t2 (d+d') k (low1 deBruijn_t2 d x)) 0 0 0);simpl;intros.
    apply H2;intros;auto.
    eapply lift_low1_comm;eauto.
    eapply lift_low1_comm;eauto.
    (* lift_1_low1_simpl *)
    intros;simpl.
    generalize (map2_compose_simpl map2_ext 0 k 0 (fun d => lift deBruijn_t1 d 1) (fun d => low1 deBruijn_t1 d) (fun d x => lift deBruijn_t1 d 1 (low1 deBruijn_t1 (d+k) x)) (fun d => lift deBruijn_t2 d 1) (fun d => low1 deBruijn_t2 d) (fun d x => lift deBruijn_t2 d 1 (low1 deBruijn_t2 (d+k) x)));simpl;intros. 
    rewrite plus_comm.
    rewrite H0;try (intros;rewrite plus_comm;reflexivity).
    generalize (fold2_orb_eq map2_ext fold2_ext (fun _ => True) (fun _ d x => mem_fn deBruijn_fn_t1 0 d x) (fun d x => lift deBruijn_t1 d 1 (low1 deBruijn_t1 (d+k) x)) (fun d => swapR deBruijn_t1 d k) (fun _ d x => mem_fn deBruijn_fn_t2 0 d x) (fun d x => lift deBruijn_t2 d 1 (low1 deBruijn_t2 (d+k) x)) (fun d => swapR deBruijn_t2 d k) k 0 0);simpl;intros.
    apply H1;intros;auto.
    eapply lift_1_low1_simpl.
    rewrite plus_comm;eauto.
    eapply lift_1_low1_simpl.
    rewrite plus_comm;eauto.
    rewrite plus_comm;auto.
    (* lift_fn *)
    intros;simpl.
    generalize (fold2_map2_simpl map2_ext fold2_ext (fun d x b => b || mem_fn deBruijn_fn_t1 (DeBruijnNat.lift 0 k i) d x) (fun d => lift deBruijn_t1 d k) (fun d x b => b || mem_fn deBruijn_fn_t2 (DeBruijnNat.lift 0 k i) d x) (fun d => lift deBruijn_t2 d k) 0 0);simpl;intros.
    rewrite H.
    apply (fold2_equal map2_ext fold2_ext);intros.
    rewrite lift_fn;auto.
    rewrite lift_fn;auto.
    (* lift_In_fn *)
    intros;simpl.
    generalize (fold2_map2_simpl map2_ext fold2_ext (fun d x b => b || mem_fn deBruijn_fn_t1 i d x) (fun d => lift deBruijn_t1 d k) (fun d x b => b || mem_fn deBruijn_fn_t2 i d x) (fun d => lift deBruijn_t2 d k) 0 (S d'));simpl;intros.
    rewrite plus_comm;simpl.
    rewrite H0.
    apply (fold2_equal map2_ext fold2_ext);intros.
    rewrite (lift_In_fn deBruijn_fn_t1 d' i H d0 k x0);auto;rewrite plus_comm;simpl;auto.
    rewrite (lift_In_fn deBruijn_fn_t2 d' i H d0 k x0);auto;rewrite plus_comm;simpl;auto.
    (* lift_not_fn *)
    intros;simpl.
    generalize (fold2_map2_simpl map2_ext fold2_ext (fun d x b => b || mem_fn deBruijn_fn_t1 i d x) (fun d => lift deBruijn_t1 d k) (fun d x b => b || mem_fn deBruijn_fn_t2 i d x) (fun d => lift deBruijn_t2 d k) 0 0);simpl;intros.
    rewrite H0.
    apply (fold2_neutral map2_ext fold2_ext);intros;simpl.
    apply lift_not_fn;auto.
    apply lift_not_fn;auto.
    (* swapR_low1_comm *)
    intros;simpl.
    generalize (map2_compose_simpl map2_ext 0 (S d') 0 (fun d => low1 deBruijn_t1 d) (fun d => swapR deBruijn_t1 d k) (fun d x => low1 deBruijn_t1 d (swapR deBruijn_t1 (d+S d') k x)) (fun d => low1 deBruijn_t2 d) (fun d => swapR deBruijn_t2 d k) (fun d x => low1 deBruijn_t2 d (swapR deBruijn_t2 (d+S d') k x)));simpl;intros.
    repeat (rewrite <- plus_n_Sm).
    rewrite plus_comm.
    rewrite H0;try (intros;rewrite <- plus_n_Sm;rewrite plus_comm;reflexivity).
    generalize (map2_compose_simpl map2_ext d' 0 0 (fun d => swapR deBruijn_t1 d k) (fun d => low1 deBruijn_t1 d) (fun d x => swapR deBruijn_t1 (d+d') k (low1 deBruijn_t1 d x)) (fun d => swapR deBruijn_t2 d k) (fun d => low1 deBruijn_t2 d) (fun d x => swapR deBruijn_t2 (d+d') k (low1 deBruijn_t2 d x)));simpl;intros.
    rewrite H1;try (intros;rewrite plus_comm;reflexivity).
    generalize (fold2_orb_eq map2_ext fold2_ext (fun _ => True) (fun _ d x => mem_fn deBruijn_fn_t1 0 d x) (fun d x => low1 deBruijn_t1 d (swapR deBruijn_t1 (d+S d') k x)) (fun d x => swapR deBruijn_t1 (d+d') k (low1 deBruijn_t1 d x)) (fun _ d x => mem_fn deBruijn_fn_t2 0 d x) (fun d x => low1 deBruijn_t2 d (swapR deBruijn_t2 (d+S d') k x)) (fun d x => swapR deBruijn_t2 (d+d') k (low1 deBruijn_t2 d x)) 0 0 0);simpl;intros.
    apply H2;intros;auto.
    eapply swapR_low1_comm;eauto.
    eapply swapR_low1_comm;eauto.
    (* swapR_in_fn *)
    intros;simpl.
    generalize (fold2_map2_simpl map2_ext fold2_ext (fun d x b => b || mem_fn deBruijn_fn_t1 i d x) (fun d => swapR deBruijn_t1 d k) (fun d x b => b || mem_fn deBruijn_fn_t2 i d x) (fun d => swapR deBruijn_t2 d k) 0 (S d'));simpl;intros.
    rewrite plus_comm;simpl.
    rewrite H0.
    apply (fold2_equal map2_ext fold2_ext);intros.
    rewrite (swapR_In_fn deBruijn_fn_t1 d' i H d0 k x0);auto;rewrite plus_comm;simpl;auto.
    rewrite (swapR_In_fn deBruijn_fn_t2 d' i H d0 k x0);auto;rewrite plus_comm;simpl;auto.
    (* swapR_eq_cond *)
    intros;simpl.
    generalize (fold2_orb_eq map2_ext fold2_ext (fun i => i <= k) (fun i d x => mem_fn deBruijn_fn_t1 i d x) (fun d => swapR deBruijn_t1 d k) (fun d x => x) (fun i d x => mem_fn deBruijn_fn_t2 i d x) (fun d => swapR deBruijn_t2 d k) (fun d x => x) 0 0 0);simpl;intros.
    rewrite H0;intros;auto.
    apply (map2_id map2_ext).
    eapply swapR_eq_cond;auto.
    eapply swapR_eq_cond;auto.
    (* swapR_fn *)
    intros;simpl.
    generalize (fold2_map2_simpl map2_ext fold2_ext (fun d x b => b || mem_fn deBruijn_fn_t1 (DeBruijnNat.swapR 0 k i) d x) (fun d => swapR deBruijn_t1 d k) (fun d x b => b || mem_fn deBruijn_fn_t2 (DeBruijnNat.swapR 0 k i) d x) (fun d => swapR deBruijn_t2 d k) 0 0);simpl;intros.
    rewrite H.
    apply (fold2_equal map2_ext fold2_ext);intros.
    rewrite swapR_fn;auto.
    rewrite swapR_fn;auto.
    (* swapR_fn_same *)
    intros;simpl.
    generalize (fold2_map2_simpl map2_ext fold2_ext (fun d x b => b || mem_fn deBruijn_fn_t1 i d x) (fun d => swapR deBruijn_t1 d k) (fun d x b => b || mem_fn deBruijn_fn_t2 i d x) (fun d => swapR deBruijn_t2 d k) (k+S d') 0);simpl;intros.
    rewrite <- plus_assoc;rewrite plus_comm.
    rewrite H.
    generalize (fold2_map2_simpl map2_ext fold2_ext (fun d x b => b || mem_fn deBruijn_fn_t1 i d x) (fun d x => x) (fun d x b => b || mem_fn deBruijn_fn_t2 i d x) (fun d x => x) (k+S d') 0);simpl;intros.
    generalize (H0 d x false).
    rewrite (map2_id map2_ext);intro.
    rewrite H1.
    apply (fold2_equal map2_ext fold2_ext);intros.
    rewrite (plus_comm (k+S d'));rewrite plus_assoc;rewrite swapR_fn_same;auto.
    rewrite (plus_comm (k+S d'));rewrite plus_assoc;rewrite swapR_fn_same;auto.
    (* swapR_low1_comm2 *)
    intros;simpl.
    generalize (map2_compose_simpl map2_ext d' 0 0 (fun d => swapR deBruijn_t1 d k) (fun d => low1 deBruijn_t1 d) (fun d x => swapR deBruijn_t1 (d+d') k (low1 deBruijn_t1 d x)) (fun d => swapR deBruijn_t2 d k) (fun d => low1 deBruijn_t2 d) (fun d x => swapR deBruijn_t2 (d+d') k (low1 deBruijn_t2 d x)));simpl;intros.
    rewrite plus_comm.
    rewrite H0;try (intros;rewrite plus_comm;reflexivity).
    generalize (map2_compose_simpl map2_ext 0 (S d') 0 (fun d => low1 deBruijn_t1 d) (fun d => swapR deBruijn_t1 d k) (fun d x => low1 deBruijn_t1 d (swapR deBruijn_t1 (d+S d') k x)) (fun d => low1 deBruijn_t2 d) (fun d => swapR deBruijn_t2 d k) (fun d x => low1 deBruijn_t2 d (swapR deBruijn_t2 (d+S d') k x)));simpl;intros.
    rewrite <- plus_n_Sm.
    rewrite plus_comm.
    rewrite H1;try (intros;rewrite <- plus_n_Sm;rewrite plus_comm;reflexivity).
    generalize (fold2_orb_eq map2_ext fold2_ext (fun i => True) (fun _ d x => mem_fn deBruijn_fn_t1 0 d x) (fun d x => swapR deBruijn_t1 (d+d') k (low1 deBruijn_t1 d x)) (fun d x => low1 deBruijn_t1 d (swapR deBruijn_t1 (d+S d') k x)) (fun _ d x => mem_fn deBruijn_fn_t2 0 d x) (fun d x => swapR deBruijn_t2 (d+d') k (low1 deBruijn_t2 d x)) (fun d x => low1 deBruijn_t2 d (swapR deBruijn_t2 (d+S d') k x)) 0 0 0);simpl;intros.
    apply H2;intros;auto.
    eapply swapR_low1_comm2;eauto.
    eapply swapR_low1_comm2;eauto.
    (* low1_swapR_comm *)
    intros;simpl.
    generalize (map2_compose_simpl map2_ext (k+S d') 0 0 (fun d => low1 deBruijn_t1 d) (fun d => swapR deBruijn_t1 d k) (fun d x => low1 deBruijn_t1 (d+k+S d') (swapR deBruijn_t1 d k x)) (fun d => low1 deBruijn_t2 d) (fun d => swapR deBruijn_t2 d k) (fun d x => low1 deBruijn_t2 (d+k+S d') (swapR deBruijn_t2 d k x)));simpl;intros.
    rewrite <- plus_assoc;rewrite plus_comm.
    rewrite H0;try (intros;rewrite plus_comm;rewrite plus_assoc;reflexivity).
    generalize (map2_compose_simpl map2_ext 0 (k+S d') 0 (fun d => swapR deBruijn_t1 d k) (fun d => low1 deBruijn_t1 d) (fun d x => swapR deBruijn_t1 d k (low1 deBruijn_t1 (d+k+S d') x)) (fun d => swapR deBruijn_t2 d k) (fun d => low1 deBruijn_t2 d) (fun d x => swapR deBruijn_t2 d k (low1 deBruijn_t2 (d+k+S d') x)));simpl;intros.
    rewrite H1;try (intros;rewrite plus_comm;rewrite plus_assoc;reflexivity).
    generalize (fold2_orb_eq map2_ext fold2_ext (fun i => True) (fun _ d x => mem_fn deBruijn_fn_t1 0 d x) (fun d x => low1 deBruijn_t1 (d+k+S d') (swapR deBruijn_t1 d k x)) (fun d x => swapR deBruijn_t1 d k (low1 deBruijn_t1 (d+k+S d') x)) (fun _ d x => mem_fn deBruijn_fn_t2 0 d x) (fun d x => low1 deBruijn_t2 (d+k+S d') (swapR deBruijn_t2 d k x)) (fun d x => swapR deBruijn_t2 d k (low1 deBruijn_t2 (d+k+S d') x)) (k+S d') 0 0 );simpl;intros.
    apply H2;intros;auto.
    eapply low1_swapR_comm;eauto.
    rewrite <- plus_assoc;rewrite plus_comm;eauto.
    eapply low1_swapR_comm;eauto.
    rewrite <- plus_assoc;rewrite plus_comm;eauto.
    rewrite (plus_comm (k+S d'));rewrite plus_assoc;auto.
    (* lift_low1_fn *)
    intros;simpl.
    generalize (fold2_map2_simpl map2_ext fold2_ext (fun d x b => b || mem_fn deBruijn_fn_t1 i d x) (fun d => low1 deBruijn_t1 d) (fun d x b => b || mem_fn deBruijn_fn_t2 i d x) (fun d => low1 deBruijn_t2 d) 0 0);simpl;intros.
    rewrite H0.
    generalize (fold2_equal_cond map2_ext fold2_ext (fun i => True) (fun i d x => mem_fn deBruijn_fn_t1 0 d x) (fun d x b => b || mem_fn deBruijn_fn_t1 i d (low1 deBruijn_t1 d x)) (fun d x b => b || mem_fn deBruijn_fn_t1 (DeBruijnNat.lift 0 1 i) d x) (fun i d x => mem_fn deBruijn_fn_t2 0 d x) (fun d x b => b || mem_fn deBruijn_fn_t2 i d (low1 deBruijn_t2 d x)) (fun d x b => b || mem_fn deBruijn_fn_t2 (DeBruijnNat.lift 0 1 i) d x) 0 0 0);simpl;intros.
    apply H1;intros;auto.
    rewrite lift_low1_fn;auto.
    rewrite lift_low1_fn;auto.
    (* low1_In_fn *)
    intros;simpl.
    generalize (fold2_map2_simpl map2_ext fold2_ext (fun d x b => b || mem_fn deBruijn_fn_t1 i d x) (fun d => low1 deBruijn_t1 d) (fun d x b => b || mem_fn deBruijn_fn_t2 i d x) (fun d => low1 deBruijn_t2 d) 0 (S d'));simpl;intros.
    rewrite plus_comm;simpl.
    rewrite H1.
    generalize (fold2_equal_cond map2_ext fold2_ext (fun i => True) (fun _ d x => mem_fn deBruijn_fn_t1 0 d x) (fun d x b => b || mem_fn deBruijn_fn_t1 i d x) (fun d x b => b || mem_fn deBruijn_fn_t1 i d (low1 deBruijn_t1 (S (d'+d)) x)) (fun _ d x => mem_fn deBruijn_fn_t2 0 d x) (fun d x b => b || mem_fn deBruijn_fn_t2 i d x) (fun d x b => b || mem_fn deBruijn_fn_t2 i d (low1 deBruijn_t2 (S (d'+d)) x)) (S d') 0 0);simpl;intros.
    apply H2;intros;auto.
    rewrite plus_comm;rewrite plus_n_Sm.
    rewrite <- low1_In_fn;auto.
    rewrite plus_comm in H3;rewrite plus_n_Sm in H3;auto.
    rewrite plus_comm;rewrite plus_n_Sm;auto.
    rewrite <- low1_In_fn;auto.
    rewrite plus_comm in H3;rewrite plus_n_Sm in H3;auto.
    rewrite plus_comm;rewrite plus_n_Sm;auto.
    (* low1_In_fn_cond *)
    intros;simpl.
    generalize (fold2_map2_simpl map2_ext fold2_ext (fun d x b => b || mem_fn deBruijn_fn_t1 i d x) (fun d => low1 deBruijn_t1 d) (fun d x b => b || mem_fn deBruijn_fn_t2 i d x) (fun d => low1 deBruijn_t2 d) d 0);simpl;intros.
    rewrite H0.
    generalize (fold2_exists map2_ext fold2_ext (fun d x => mem_fn deBruijn_fn_t1 (DeBruijnNat.lift 0 1 i) d x) (fun d0 x => mem_fn deBruijn_fn_t1 i (d+d0) (low1 deBruijn_t1 d0 x)) (fun d x => mem_fn deBruijn_fn_t2 (DeBruijnNat.lift 0 1 i) d x) (fun d0 x => mem_fn deBruijn_fn_t2 i (d+d0) (low1 deBruijn_t2 d0 x)) d 0);simpl;intros.
    apply H1;intros;auto.
    apply low1_In_fn_cond;auto.
    apply low1_In_fn_cond;auto.
    (* fn_plus_lift *)
    intros.
    generalize (fold2_map2_simpl map2_ext fold2_ext (fun d x b => b || mem_fn deBruijn_fn_t1 i d x) (fun d x => x) (fun d x b => b || mem_fn deBruijn_fn_t2 i d x) (fun d x => x) d' 0 d x false);simpl;intros.
    rewrite (map2_id map2_ext) in H.
    rewrite plus_comm.
    rewrite H.
    apply (fold2_equal map2_ext fold2_ext);intros;auto.
    rewrite fn_plus_lift;rewrite plus_comm;auto.
    rewrite fn_plus_lift;rewrite plus_comm;auto.
    (* low1_comm *)
    intros;simpl.
    generalize (map2_compose_simpl map2_ext 0 (S d') 0 (fun d => low1 deBruijn_t1 d) (fun d => low1 deBruijn_t1 d) (fun d x => low1 deBruijn_t1 d (low1 deBruijn_t1 (d+S d') x)) (fun d => low1 deBruijn_t2 d) (fun d => low1 deBruijn_t2 d) (fun d x => low1 deBruijn_t2 d (low1 deBruijn_t2 (d+S d') x)));simpl;intros.
    rewrite <- plus_n_Sm.
    rewrite plus_comm.
    rewrite H0;try (intros;rewrite <- plus_n_Sm;rewrite plus_comm;reflexivity).
    generalize (map2_compose_simpl map2_ext d' 0 0 (fun d => low1 deBruijn_t1 d) (fun d => low1 deBruijn_t1 d) (fun d x => low1 deBruijn_t1 (d+d') (low1 deBruijn_t1 d x)) (fun d => low1 deBruijn_t2 d) (fun d => low1 deBruijn_t2 d) (fun d x => low1 deBruijn_t2 (d+d') (low1 deBruijn_t2 d x)));simpl;intros.
    rewrite H1;try (intros;rewrite plus_comm;reflexivity).
    generalize (fold2_orb_eq map2_ext fold2_ext (fun _ => True) (fun _ d x => mem_fn deBruijn_fn_t1 0 d x) (fun d x => low1 deBruijn_t1 d (low1 deBruijn_t1 (d+S d') x)) (fun d x => low1 deBruijn_t1 (d+d') (low1 deBruijn_t1 d x)) (fun _ d x => mem_fn deBruijn_fn_t2 0 d x) (fun d x => low1 deBruijn_t2 d (low1 deBruijn_t2 (d+S d') x)) (fun d x => low1 deBruijn_t2 (d+d') (low1 deBruijn_t2 d x)) (S d') 0 0);simpl;intros.
    apply H2;intros;auto.
    eapply low1_comm.
    rewrite plus_comm;simpl;eauto.
    eapply low1_comm.
    rewrite plus_comm;simpl;eauto.
    rewrite plus_comm;rewrite plus_n_Sm;auto.
    (* low1_swapR_simpl *)
    intros;simpl.
    generalize (map2_compose_simpl map2_ext 0 0 0 (fun d => low1 deBruijn_t1 d) (fun d => swapR deBruijn_t1 d k) (fun d x => low1 deBruijn_t1 d (swapR deBruijn_t1 d k x)) (fun d => low1 deBruijn_t2 d) (fun d => swapR deBruijn_t2 d k) (fun d x => low1 deBruijn_t2 d (swapR deBruijn_t2 d k x)));simpl;intros.
    rewrite H0;try reflexivity.
    generalize (fold2_orb_eq map2_ext fold2_ext (fun _ => True) (fun _ d x => mem_fn deBruijn_fn_t1 0 d x) (fun d x => low1 deBruijn_t1 d (swapR deBruijn_t1 d k x)) (fun d => low1 deBruijn_t1 d) (fun _ d x => mem_fn deBruijn_fn_t2 0 d x) (fun d x => low1 deBruijn_t2 d (swapR deBruijn_t2 d k x)) (fun d => low1 deBruijn_t2 d) k 0 k);simpl;intros.
    rewrite plus_comm;apply H1;intros;auto.
    rewrite plus_comm.
    rewrite (low1_swapR_simpl deBruijn_fn_t1);auto.
    rewrite plus_comm;auto.
    rewrite (low1_swapR_simpl deBruijn_fn_t2);auto.
    rewrite plus_comm;auto.
    rewrite plus_comm;auto.
    rewrite plus_comm;auto.
    (* Nmin_true *)
    intros.
    apply (fold2_neutral map2_ext fold2_ext);intros;apply Nmin_true.
    (* Nmin_le_false *)
    intros.
    generalize (fold2_invariant map2_ext fold2_ext (fun accu:bool*nat => let (b,k) := accu in k <= kmin) (fun d x accu => Nmin deBruijn_fn_t1 d k x accu) (fun d x accu => Nmin deBruijn_fn_t2 d k x accu)).
    intros.
    assert (forall (d:nat)(x:t1)(a:bool*nat),(let (_,k) := a in k <= kmin)->(let (_,k) := Nmin deBruijn_fn_t1 d k x a in k <= kmin)).
    intros.
    destruct a as [ba ka].
    destruct ba.
    rewrite Nmin_true;auto.
    generalize (Nmin_le_false deBruijn_fn_t1 d0 k x0 ka).
    destruct (Nmin deBruijn_fn_t1 d0 k x0 (false,ka));intros.
    generalize (H2 b0 n (refl_equal (b0,n)));intros.
    omega.
    assert (forall (d:nat)(x:t2)(a:bool*nat),(let (_,k) := a in k <= kmin)->(let (_,k) := Nmin deBruijn_fn_t2 d k x a in k <= kmin)).
    intros.
    destruct a as [ba ka].
    destruct ba.
    rewrite Nmin_true;auto.
    generalize (Nmin_le_false deBruijn_fn_t2 d0 k x0 ka).
    destruct (Nmin deBruijn_fn_t2 d0 k x0 (false,ka));intros.
    generalize (H3 b0 n (refl_equal (b0,n)));intros.
    omega.
    generalize (H0 H1 H2 d x (false,kmin)).
    rewrite H;auto.
    (* Nmin_lift *)
    intros.
    simpl.
    generalize (fold2_map2_simpl map2_ext fold2_ext (fun d x accu => Nmin deBruijn_fn_t1 d k x accu) (fun d => lift deBruijn_t1 d k') (fun d x accu => Nmin deBruijn_fn_t2 d k x accu) (fun d => lift deBruijn_t2 d k') 0 (k+S d'));simpl;intros.
    rewrite <- plus_assoc;rewrite plus_comm.
    rewrite H0.
    apply (fold2_equal_invariant map2_ext fold2_ext (fun accu:bool*nat => let (_,km) := accu in km <= k));auto;intros.
    destruct a.
    rewrite (plus_comm (k+S d'));rewrite plus_assoc.
    apply (Nmin_lift deBruijn_fn_t1);auto.
    destruct a.
    rewrite (plus_comm (k+S d'));rewrite plus_assoc.
    apply (Nmin_lift deBruijn_fn_t2);auto.
    destruct a.
    destruct b0.
    rewrite (Nmin_true deBruijn_fn_t1);auto.
    generalize (Nmin_le_false deBruijn_fn_t1 d0 k x0 n).
    destruct (Nmin deBruijn_fn_t1 d0 k x0 (false,n)).
    intros.
    generalize (H2 b0 n0 (refl_equal (b0,n0)));intro.
    omega.
    destruct a.
    destruct b0.
    rewrite (Nmin_true deBruijn_fn_t2);auto.
    generalize (Nmin_le_false deBruijn_fn_t2 d0 k x0 n).
    destruct (Nmin deBruijn_fn_t2 d0 k x0 (false,n)).
    intros.
    generalize (H2 b0 n0 (refl_equal (b0,n0)));intro.
    omega.
    (* Nmin_swapR *)
    intros.
    simpl.
    generalize (fold2_map2_simpl map2_ext fold2_ext (fun d x accu => Nmin deBruijn_fn_t1 d k x accu) (fun d => swapR deBruijn_t1 d k') (fun d x accu => Nmin deBruijn_fn_t2 d k x accu) (fun d => swapR deBruijn_t2 d k') 0 (k+S d'));simpl;intros.
    rewrite <- plus_assoc;rewrite plus_comm.
    rewrite H0.
    apply (fold2_equal_invariant map2_ext fold2_ext (fun accu:bool*nat => let (_,km) := accu in km <= k));auto;intros.
    destruct a.
    rewrite (plus_comm (k+S d'));rewrite plus_assoc.
    apply (Nmin_swapR deBruijn_fn_t1);auto.
    destruct a.
    rewrite (plus_comm (k+S d'));rewrite plus_assoc.
    apply (Nmin_swapR deBruijn_fn_t2);auto.
    destruct a.
    destruct b0.
    rewrite (Nmin_true deBruijn_fn_t1);auto.
    generalize (Nmin_le_false deBruijn_fn_t1 d0 k x0 n).
    destruct (Nmin deBruijn_fn_t1 d0 k x0 (false,n)).
    intros.
    generalize (H2 b0 n0 (refl_equal (b0,n0)));intro.
    omega.
    destruct a.
    destruct b0.
    rewrite (Nmin_true deBruijn_fn_t2);auto.
    generalize (Nmin_le_false deBruijn_fn_t2 d0 k x0 n).
    destruct (Nmin deBruijn_fn_t2 d0 k x0 (false,n)).
    intros.
    generalize (H2 b0 n0 (refl_equal (b0,n0)));intro.
    omega.
    (* Nmin_low1 *)
    intros.
    simpl.
    generalize (fold2_map2_simpl map2_ext fold2_ext (fun d x accu => Nmin deBruijn_fn_t1 d k x accu) (fun d => low1 deBruijn_t1 d) (fun d x accu => Nmin deBruijn_fn_t2 d k x accu) (fun d => low1 deBruijn_t2 d) 0 (k+S d'));simpl;intros.
    rewrite <- plus_assoc;rewrite plus_comm.
    rewrite H1.
    generalize (fold2_equal_cond_invariant map2_ext fold2_ext (fun accu : bool*nat => let (_,km) := accu in km <= k) (fun _ => True) (fun _ d x => mem_fn deBruijn_fn_t1 (k+S d') d x) (fun d x accu => Nmin deBruijn_fn_t1 d k x accu) (fun d x accu => Nmin deBruijn_fn_t1 d k (low1 deBruijn_t1 (k+S d'+d) x) accu) (fun _ d x => mem_fn deBruijn_fn_t2 (k+S d') d x) (fun d x accu => Nmin deBruijn_fn_t2 d k x accu) (fun d x accu => Nmin deBruijn_fn_t2 d k (low1 deBruijn_t2 (k+S d'+d) x) accu) 0 0 0);simpl;intro.
    apply H2;intros;auto.
    rewrite (plus_comm (k+S d'));rewrite plus_assoc.
    destruct a.
    apply (Nmin_low1 deBruijn_fn_t1);auto.
    rewrite (plus_comm (k+S d'));rewrite plus_assoc.
    destruct a.
    apply (Nmin_low1 deBruijn_fn_t2);auto.
    destruct a.
    destruct b0.
    rewrite Nmin_true;auto.
    generalize (Nmin_le_false deBruijn_fn_t1 d0 k x0 n).
    destruct (Nmin deBruijn_fn_t1 d0 k x0 (false,n));intros.
    generalize (H4 b0 n0 (refl_equal (b0,n0)));intro;omega.
    destruct a.
    destruct b0.
    rewrite Nmin_true;auto.
    generalize (Nmin_le_false deBruijn_fn_t2 d0 k x0 n).
    destruct (Nmin deBruijn_fn_t2 d0 k x0 (false,n));intros.
    generalize (H4 b0 n0 (refl_equal (b0,n0)));intro;omega.
    (* Nmin_swapR_gt *)
    intros;simpl.
    generalize (fold2_map2_simpl map2_ext fold2_ext (fun d x accu => Nmin deBruijn_fn_t1 d k x accu) (fun d => swapR deBruijn_t1 d (k2-k1)) (fun d x accu => Nmin deBruijn_fn_t2 d k x accu) (fun d => swapR deBruijn_t2 d (k2-k1)) 0 k1);simpl;intros.
    rewrite plus_comm.
    rewrite H2.
    apply (fold2_equal_invariant map2_ext fold2_ext (fun accu : bool*nat => let (_,km) := accu in km <= k));intros;auto.
    rewrite plus_comm;destruct a.
    rewrite (Nmin_swapR_gt deBruijn_fn_t1 d0 k k1 k2);auto.
    rewrite plus_comm;destruct a.
    rewrite (Nmin_swapR_gt deBruijn_fn_t2 d0 k k1 k2);auto.
    destruct a.
    rewrite (pair_split (Nmin deBruijn_fn_t1 d0 k (swapR deBruijn_t1 (k1+d0) (k2-k1) x0) (b0,n))).
    apply le_trans with n.
    apply Nmin_le.
    auto.
    destruct a.
    rewrite (pair_split (Nmin deBruijn_fn_t2 d0 k (swapR deBruijn_t2 (k1+d0) (k2-k1) x0) (b0,n))).
    apply le_trans with n.
    apply Nmin_le.
    auto.
    (* low1_injective *)
    simpl;intros.
    eapply (fold2_orb_eq_map2_inversion map2_ext fold2_ext (fun x => True) (fun i d x => mem_fn deBruijn_fn_t1 0 d x) (fun d => low1 deBruijn_t1 d) (fun d => low1 deBruijn_t1 d) (fun i d x => mem_fn deBruijn_fn_t2 0 d x) (fun d => low1 deBruijn_t2 d) (fun d => low1 deBruijn_t2 d) 0 0 0);simpl;intros;eauto.
    eapply (low1_injective deBruijn_fn_t1);eauto.
    eapply (low1_injective deBruijn_fn_t2);eauto.
  Defined.

  Record deBruijn_extension_fn (t:Set)(deBruijn_t:deBruijn_base t)(deBruijn_fn_t:deBruijn_fn t deBruijn_t)(t':Set)(deBruijn_t':deBruijn_base t')(deBruijn_fn_t':deBruijn_fn t' deBruijn_t')(deBruijn_ext:deBruijn_extension t deBruijn_t t' deBruijn_t') : Set := {
    j_mem_fn : forall (i d:nat)(x:t),(mem_fn deBruijn_fn_t i d x)=(mem_fn deBruijn_fn_t' i d (j deBruijn_ext x))
  }.

  Implicit Arguments j_mem_fn [t deBruijn_t deBruijn_fn_t t' deBruijn_t' deBruijn_fn_t' deBruijn_ext].

  Record deBruijn_subst_fn (t:Set)(deBruijn_t:deBruijn_base t)(deBruijn_fn_t:deBruijn_fn t deBruijn_t)(to:Set)(deBruijn_to:deBruijn_base to)(deBruijn_fn_to:deBruijn_fn to deBruijn_to)(t_subst_to:deBruijn_subst t deBruijn_t to deBruijn_to) : Set := {
    low_subst_fn_not_supp : forall (i:nat)(d d' k:nat)(es:list to)(x:t),(i <= d')->(deBruijn.mem_fn deBruijn_fn_t i d x = deBruijn.mem_fn deBruijn_fn_t i d (deBruijn.low_subst t_subst_to (d+S d') k es x));
    low_subst_low1_comm : forall (d d' k:nat)(es:list to)(x:t),(mem_fn deBruijn_fn_t 0 d x = false)->(low_subst t_subst_to (d+d') k es (low1 deBruijn_t d x))=(low1 deBruijn_t d (low_subst t_subst_to (d+S d') k es x));
    low_subst_low1_same : forall (d:nat)(e:to)(x:t),(mem_fn deBruijn_fn_t 0 d x = false)->(low_subst t_subst_to d 0 (cons e nil) x = low1 deBruijn_t d x)
  }.

  Implicit Arguments low_subst_fn_not_supp [t deBruijn_t deBruijn_fn_t to deBruijn_to deBruijn_fn_to t_subst_to].
  Implicit Arguments low_subst_low1_comm [t deBruijn_t deBruijn_fn_t to deBruijn_to deBruijn_fn_to t_subst_to].
  Implicit Arguments low_subst_low1_same [t deBruijn_t deBruijn_fn_t to deBruijn_to deBruijn_fn_to t_subst_to].

  Definition deBruijn_s_fn_fold_struct (t:Set)(deBruijn_t:deBruijn_base t)(deBruijn_fn_t:deBruijn_fn t deBruijn_t)(to:Set)(deBruijn_to:deBruijn_base to)(deBruijn_fn_to:deBruijn_fn to deBruijn_to)(t_subst_to:deBruijn_subst t deBruijn_t to deBruijn_to)(t_subst_fn_to:deBruijn_subst_fn t deBruijn_t deBruijn_fn_t to deBruijn_to deBruijn_fn_to t_subst_to)(t':Set)(map_fun:(nat->t->t)->nat->t'->t')(map_ext:map_struct t t' map_fun)(fold_fun:forall (A:Set),(nat->t->A->A)->nat->t'->A->A)(fold_ext:fold_struct t t' map_fun map_ext fold_fun) : (deBruijn_subst_fn t' (deBruijn_map_struct t deBruijn_t t' map_fun map_ext) (deBruijn_fn_map_struct t t' map_fun map_ext fold_fun deBruijn_t deBruijn_fn_t fold_ext) to deBruijn_to deBruijn_fn_to (deBruijn_s_map_struct t deBruijn_t to deBruijn_to t_subst_to t' map_fun map_ext)).
    intros.
    split.
    (* low_subst_fn_not_supp *)
    intros.
    simpl.
    generalize (fold_map_simpl map_ext fold_ext (fun d x b => b || mem_fn deBruijn_fn_t i d x) (fun d => low_subst t_subst_to d k es) 0 (S d') d).
    rewrite plus_0_l.
    rewrite plus_comm.
    intros.
    rewrite H0.
    eapply (fold_equal map_ext fold_ext).
    intros.
    rewrite plus_0_l.
    rewrite plus_comm.
    rewrite <- (low_subst_fn_not_supp t_subst_fn_to);auto.
    (* low_subst_low1_comm *)
    intros.
    simpl.
    generalize (map_compose_simpl map_ext d' 0 0 (fun d => low_subst t_subst_to d k es) (fun d => low1 deBruijn_t d) (fun d x => low_subst t_subst_to (d' + d) k es (low1 deBruijn_t d x)));simpl;intros.
    rewrite plus_comm.
    rewrite H0;auto.
    clear H0.
    generalize (map_compose_simpl map_ext 0 (S d') 0 (fun d => low1 deBruijn_t d) (fun d => low_subst t_subst_to d k es) (fun d x => low1 deBruijn_t d (low_subst t_subst_to (d+S d') k es x)));intros.
    simpl in H0.
    rewrite plus_comm;simpl.
    rewrite H0.
    generalize (fold_orb_eq map_ext fold_ext (fun _ => True) (fun _ d x=> mem_fn deBruijn_fn_t 0 d x) (fun d x => low_subst t_subst_to (d' + d) k es (low1 deBruijn_t d x)) (fun d x => low1 deBruijn_t d (low_subst t_subst_to (d+S d') k es x)) 0 0 0);simpl;intros.
    apply H1;intros;auto.
    rewrite <- (low_subst_low1_comm t_subst_fn_to d0 d' k es x0);auto.
    rewrite plus_comm;auto.
    intros.
    rewrite <- plus_n_Sm;rewrite plus_comm;auto.
    (* low_subst_low1_same *)
    intros.
    simpl.
    generalize (fold_orb_eq map_ext fold_ext (fun _ => True) (fun _ d x=> mem_fn deBruijn_fn_t 0 d x) (fun d => low_subst t_subst_to d 0 (cons e nil)) (fun d => low1 deBruijn_t d) 0 0 0);simpl;intros.
    apply H0;intros.
    apply (low_subst_low1_same t_subst_fn_to).
    auto.
    simpl in H.
    auto.
  Qed.

  Definition deBruijn_s_fn_fold2_struct (t1:Set)(deBruijn_t1:deBruijn_base t1)(deBruijn_fn_t1:deBruijn_fn t1 deBruijn_t1)(t2:Set)(deBruijn_t2:deBruijn_base t2)(deBruijn_fn_t2:deBruijn_fn t2 deBruijn_t2)(to:Set)(deBruijn_to:deBruijn_base to)(deBruijn_fn_to:deBruijn_fn to deBruijn_to)(t1_subst_to:deBruijn_subst t1 deBruijn_t1 to deBruijn_to)(t1_subst_fn_to:deBruijn_subst_fn t1 deBruijn_t1 deBruijn_fn_t1 to deBruijn_to deBruijn_fn_to t1_subst_to)(t2_subst_to:deBruijn_subst t2 deBruijn_t2 to deBruijn_to)(t2_subst_fn_to:deBruijn_subst_fn t2 deBruijn_t2 deBruijn_fn_t2 to deBruijn_to deBruijn_fn_to t2_subst_to)(t:Set)(map2_fun:(nat->t1->t1)->(nat->t2->t2)->nat->t->t)(map2_ext:map2_struct t1 t2 t map2_fun)(fold2_fun:forall (A:Set),(nat->t1->A->A)->(nat->t2->A->A)->nat->t->A->A)(fold2_ext:fold2_struct t1 t2 t map2_fun map2_ext fold2_fun) : (deBruijn_subst_fn t (deBruijn_map2_struct t1 deBruijn_t1 t2 deBruijn_t2 t map2_fun map2_ext) (deBruijn_fn_map2_struct t1 t2 t map2_fun map2_ext fold2_fun deBruijn_t1 deBruijn_fn_t1 deBruijn_t2 deBruijn_fn_t2 fold2_ext) to deBruijn_to deBruijn_fn_to (deBruijn_s_map2_struct t1 deBruijn_t1 t2 deBruijn_t2 to deBruijn_to t1_subst_to t2_subst_to t map2_fun map2_ext)).
    intros.
    split.
    (* low_subst_fn_not_supp *)
    intros.
    simpl.
    generalize (fold2_map2_simpl map2_ext fold2_ext (fun d x b => b || mem_fn deBruijn_fn_t1 i d x) (fun d => low_subst t1_subst_to d k es) (fun d x b => b || mem_fn deBruijn_fn_t2 i d x) (fun d => low_subst t2_subst_to d k es) 0 (S d') d).
    repeat (rewrite plus_0_l).
    rewrite plus_comm.
    intros.
    rewrite H0.
    eapply (fold2_equal map2_ext fold2_ext).
    intros.
    rewrite plus_0_l.
    rewrite plus_comm.
    rewrite <- (low_subst_fn_not_supp t1_subst_fn_to);auto.
    intros.
    rewrite plus_0_l.
    rewrite plus_comm.
    rewrite <- (low_subst_fn_not_supp t2_subst_fn_to);auto.
    (* low_subst_low1_comm *)
    intros.
    simpl.
    generalize (map2_compose_simpl map2_ext d' 0 0 (fun d => low_subst t1_subst_to d k es) (fun d => low1 deBruijn_t1 d) (fun d x => low_subst t1_subst_to (d' + d) k es (low1 deBruijn_t1 d x)) (fun d => low_subst t2_subst_to d k es) (fun d => low1 deBruijn_t2 d) (fun d x => low_subst t2_subst_to (d' + d) k es (low1 deBruijn_t2 d x)));simpl;intros.
    rewrite plus_comm.
    rewrite H0;auto.
    clear H0.
    generalize (map2_compose_simpl map2_ext 0 (S d') 0 (fun d => low1 deBruijn_t1 d) (fun d => low_subst t1_subst_to d k es) (fun d x => low1 deBruijn_t1 d (low_subst t1_subst_to (d+S d') k es x))  (fun d => low1 deBruijn_t2 d) (fun d => low_subst t2_subst_to d k es) (fun d x => low1 deBruijn_t2 d (low_subst t2_subst_to (d+S d') k es x)));intros.
    simpl in H0.
    rewrite plus_comm;simpl.
    rewrite H0.
    generalize (fold2_orb_eq map2_ext fold2_ext (fun _ => True) (fun _ d x=> mem_fn deBruijn_fn_t1 0 d x) (fun d x => low_subst t1_subst_to (d' + d) k es (low1 deBruijn_t1 d x)) (fun d x => low1 deBruijn_t1 d (low_subst t1_subst_to (d+S d') k es x)) (fun _ d x=> mem_fn deBruijn_fn_t2 0 d x) (fun d x => low_subst t2_subst_to (d' + d) k es (low1 deBruijn_t2 d x)) (fun d x => low1 deBruijn_t2 d (low_subst t2_subst_to (d+S d') k es x)) 0 0 0);simpl;intros.
    apply H1;intros;auto.
    rewrite <- (low_subst_low1_comm t1_subst_fn_to);auto.
    rewrite plus_comm;auto.
    rewrite <- (low_subst_low1_comm t2_subst_fn_to);auto.
    rewrite plus_comm;auto.
    intros;rewrite <- plus_n_Sm;rewrite plus_comm;auto.
    intros;rewrite <- plus_n_Sm;rewrite plus_comm;auto.
    (* low_subst_low1_same *)
    intros.
    simpl.
    generalize (fold2_orb_eq map2_ext fold2_ext (fun _ => True) (fun _ d x=> mem_fn deBruijn_fn_t1 0 d x) (fun d => low_subst t1_subst_to d 0 (cons e nil)) (fun d => low1 deBruijn_t1 d) (fun _ d x=> mem_fn deBruijn_fn_t2 0 d x) (fun d => low_subst t2_subst_to d 0 (cons e nil)) (fun d => low1 deBruijn_t2 d) 0 0 0);simpl;intros.
    apply H0;intros.
    apply (low_subst_low1_same t1_subst_fn_to);auto.
    apply (low_subst_low1_same t2_subst_fn_to);auto.
    simpl in H.
    auto.
  Qed.
    
  Record deBruijn_subst_gen_fn (t:Set)(deBruijn_t:deBruijn_base t)(deBruijn_fn_t:deBruijn_fn t deBruijn_t)(to:Set)(deBruijn_to:deBruijn_base to)(deBruijn_fn_to:deBruijn_fn to deBruijn_to)(deBruijn_ext:deBruijn_extension t deBruijn_t to deBruijn_to)(deBruijn_fn_ext:deBruijn_extension_fn t deBruijn_t deBruijn_fn_t to deBruijn_to deBruijn_fn_to deBruijn_ext)(t_subst_to:deBruijn_subst_gen t deBruijn_t to deBruijn_to deBruijn_ext) : Set := {
    low_subst_gen_fn_not_supp : forall (i:nat)(d d' k:nat)(es:list to)(x:t),(i <= d')->(mem_fn deBruijn_fn_t i d x = mem_fn deBruijn_fn_to i d (low_subst_gen t_subst_to (d+S d') k es x));
    low_subst_gen_low1_comm : forall (d d' k:nat)(es:list to)(x:t),(mem_fn deBruijn_fn_t 0 d x = false)->(low_subst_gen t_subst_to (d+d') k es (low1 deBruijn_t d x))=(low1 deBruijn_to d (low_subst_gen t_subst_to (d+S d') k es x));
    low_subst_gen_low1_same : forall (d:nat)(e:to)(x:t),(mem_fn deBruijn_fn_t 0 d x = false)->(low_subst_gen t_subst_to d 0 (cons e nil) x = j deBruijn_ext (low1 deBruijn_t d x))
  }.

  Implicit Arguments low_subst_gen_fn_not_supp [t deBruijn_t deBruijn_fn_t to deBruijn_to deBruijn_fn_to deBruijn_ext deBruijn_fn_ext t_subst_to].
  Implicit Arguments low_subst_gen_low1_comm [t deBruijn_t deBruijn_fn_t to deBruijn_to deBruijn_fn_to deBruijn_ext deBruijn_fn_ext t_subst_to].
  Implicit Arguments low_subst_gen_low1_same [t deBruijn_t deBruijn_fn_t to deBruijn_to deBruijn_fn_to deBruijn_ext deBruijn_fn_ext t_subst_to].

  Record fold_extension (t:Set)(deBruijn_t:deBruijn_base t)(t':Set)(deBruijn_t':deBruijn_base t')(deBruijn_ext:deBruijn_extension t deBruijn_t t' deBruijn_t')(map_fun:(nat->t->t)->nat->t'->t')(map_ext:map_struct t t' map_fun)(map_gen_fun:(nat->t->t')->nat->t'->t')(fold_fun:forall (A:Set),(nat->t->A->A)->nat->t'->A->A)(fold_ext:fold_struct t t' map_fun map_ext fold_fun) : Set := {
    j_fold : forall (A:Set)(f:nat->t->A->A)(d:nat)(x:t)(e:A),fold_fun A f d (j deBruijn_ext x) e = f d x e;  
    fold_map_gen_simpl : forall (A:Set)(f:nat->t->A->A)(g:nat->t->t')(d1 d2:nat),forall (d:nat)(x:t')(a:A),(fold_fun A f (d1+d) (map_gen_fun g (d2+d) x) a)=(fold_fun A (fun d x a => fold_fun A (fun d x a => f (d1+d) x a) d (g (d2+d) x) a) d x a);
    fold_orb_eq_gen : forall (P:nat->Prop)(fi:nat->nat->t->bool)(g h:nat->t->t')(d1 d2 d3:nat),(forall (d:nat)(x:t),(forall (i:nat),(P i)->(fi i (d1+d) x)=false)->g (d2+d) x = h (d3+d) x)->(forall (d:nat)(x:t'),(forall (i:nat),(P i)->(fold_fun bool (fun d x b => b || (fi i d x)) (d1+d) x false = false))->(map_gen_fun g (d2+d) x)=(map_gen_fun h (d3+d) x))
  }.

  Implicit Arguments j_fold [t deBruijn_t t' deBruijn_t' deBruijn_ext map_fun map_ext map_gen_fun fold_fun fold_ext].
  Implicit Arguments fold_map_gen_simpl [t deBruijn_t t' deBruijn_t' deBruijn_ext map_fun map_ext map_gen_fun fold_fun fold_ext].
  Implicit Arguments fold_orb_eq_gen [t deBruijn_t t' deBruijn_t' deBruijn_ext map_fun map_ext map_gen_fun fold_fun fold_ext].

  Definition deBruijn_s_gen_fn_fold_struct 
    (t:Set)(deBruijn_t:deBruijn_base t)(deBruijn_fn_t:deBruijn_fn t deBruijn_t)
    (to:Set)(map_fun:(nat->t->t)->nat->to->to)(map_str:map_struct t to map_fun)
    (deBruijn_ext:deBruijn_extension t deBruijn_t to (deBruijn_map_struct t deBruijn_t to map_fun map_str))
    (deBruijn_s_gen:deBruijn_subst_gen t deBruijn_t to (deBruijn_map_struct t deBruijn_t to map_fun map_str) deBruijn_ext)
    (map_gen_fun:(nat->t->to)->nat->to->to)
    (map_ext:map_extension t deBruijn_t to (deBruijn_map_struct t deBruijn_t to map_fun map_str) deBruijn_ext map_fun map_str map_gen_fun)
    (fold_fun:forall (A:Set),(nat->t->A->A)->nat->to->A->A) 
    (fold_str:fold_struct t to map_fun map_str fold_fun)
    (fold_ext:fold_extension t deBruijn_t to (deBruijn_map_struct t deBruijn_t to map_fun map_str) deBruijn_ext map_fun map_str map_gen_fun fold_fun fold_str)
    (deBruijn_fn_ext:deBruijn_extension_fn t deBruijn_t deBruijn_fn_t to (deBruijn_map_struct t deBruijn_t to map_fun map_str) (deBruijn_fn_map_struct t to map_fun map_str fold_fun deBruijn_t deBruijn_fn_t fold_str) deBruijn_ext) 
    (deBruijn_s_gen_fn:deBruijn_subst_gen_fn t deBruijn_t deBruijn_fn_t to (deBruijn_map_struct t deBruijn_t to map_fun map_str) (deBruijn_fn_map_struct t to map_fun map_str fold_fun deBruijn_t deBruijn_fn_t fold_str) deBruijn_ext deBruijn_fn_ext deBruijn_s_gen)
    :
    deBruijn_subst_fn to (deBruijn_map_struct t deBruijn_t to map_fun map_str) (deBruijn_fn_map_struct t to map_fun map_str fold_fun deBruijn_t deBruijn_fn_t fold_str) to (deBruijn_map_struct t deBruijn_t to map_fun map_str) (deBruijn_fn_map_struct t to map_fun map_str fold_fun deBruijn_t deBruijn_fn_t fold_str) (deBruijn_subst_gen_subst t deBruijn_t to map_fun map_str deBruijn_ext deBruijn_s_gen map_gen_fun map_ext).
    intros.
    split.
    (* low_subst_gen_fn_not_supp *)
    intros.
    simpl.
    generalize (fold_map_gen_simpl fold_ext bool (fun d x b => b || mem_fn deBruijn_fn_t i d x) (fun d => low_subst_gen deBruijn_s_gen d k es) 0 (S d') d x false).
    rewrite plus_0_l.
    rewrite plus_comm.
    intro.
    rewrite H0.
    apply (fold_equal map_str fold_str).
    intros.
    rewrite plus_comm.
    simpl.
    generalize (low_subst_gen_fn_not_supp deBruijn_s_gen_fn i d0 d' k es x0 H).
    intros.
    simpl in H1.
    destruct a;simpl;auto.
    rewrite (fold_neutral map_str fold_str (fun d x a => a || mem_fn deBruijn_fn_t i d x) true);auto.
    (* low_subst_gen_low1_comm *)
    intros.
    simpl.
    rewrite (map_gen_j_map map_ext).
    rewrite (map_gen_j_map map_ext).
    generalize (map_gen_compose map_ext d' 0 (fun d => low_subst_gen deBruijn_s_gen d k es) (fun d x => j deBruijn_ext (low1 deBruijn_t d x)));simpl;intros.
    rewrite plus_comm.
    rewrite H0;auto.
    clear H0.
    generalize (map_gen_compose map_ext 0 (S d') (fun d x => j deBruijn_ext (low1 deBruijn_t d x)) (fun d => low_subst_gen deBruijn_s_gen d k es));intros.
    simpl in H0.
    rewrite plus_comm;simpl.
    rewrite H0.
    clear H0.
    simpl in H.
    generalize (fold_orb_eq_gen fold_ext (fun _ => True) (fun _ d x=> mem_fn deBruijn_fn_t 0 d x) (fun d x => map_gen_fun (fun d => low_subst_gen deBruijn_s_gen d k es) (d'+d) (j deBruijn_ext (low1 deBruijn_t d x))) (fun d x => map_gen_fun (fun d x => j deBruijn_ext (low1 deBruijn_t d x)) d (low_subst_gen deBruijn_s_gen (S (d'+d)) k es x)) 0 0 0).
    simpl;intros.
    apply H0;intros;auto.
    clear H0.
    rewrite (map_gen_j map_ext).
    rewrite <- (map_gen_j_map map_ext (fun d => low1 deBruijn_t d)).
    generalize (low_subst_gen_low1_comm deBruijn_s_gen_fn d0 d' k es x0);simpl;intros.
    rewrite plus_comm.
    rewrite H0;auto.
    rewrite plus_n_Sm.
    auto.
    (* low_subst_gen_low1_same *)
    intros.
    simpl in H.
    simpl.
    rewrite (map_gen_j_map map_ext).
    generalize (fold_orb_eq_gen fold_ext (fun _ => True) (fun _ d x=> mem_fn deBruijn_fn_t 0 d x) (fun d => low_subst_gen deBruijn_s_gen d 0 (cons e nil)) (fun d x => j deBruijn_ext (low1 deBruijn_t d x)) 0 0 0).
    simpl;intros.
    apply H0;intros;auto.
    apply (low_subst_gen_low1_same deBruijn_s_gen_fn).
    auto.
  Qed.

  (* tools *)
  Definition map_list (t:Set)(f:nat->t->t)(d:nat)(ts:list t) : list t := map (f d) ts.

  Lemma map_list_app : forall (t:Set)(f:nat->t->t)(d:nat)(ts us:list t),(map_list t f d (ts ++ us))=(map_list t f d ts)++(map_list t f d us).
    intros.
    unfold map_list.
    generalize us;clear us.
    induction ts;simpl;intros;try reflexivity.
    rewrite IHts.
    auto.
  Qed.

  Lemma map_struct_map_list : forall (t:Set),(map_struct t (list t) (map_list t)).
    split;intros.
    induction x;simpl;try congruence.
    unfold injective.
    induction x;destruct y;simpl;intro H0;try (discriminate H0);trivial.
    injection H0;clear H0;intros.
    rewrite (IHx y);trivial.
    rewrite (H d a t0 H1);trivial.
    induction x;simpl;intros;trivial.
    rewrite IHx.
    rewrite H;trivial.
    induction x;simpl;intros;trivial.
    rewrite IHx.
    rewrite H;trivial.
  Qed.
    
  Definition deBruijn_base_list (t:Set)(deBruijn_t:deBruijn_base t) : deBruijn_base (list t) := deBruijn_map_struct t deBruijn_t (list t) (map_list t) (map_struct_map_list t).

  Definition deBruijn_extension_list (t:Set)(deBruijn_t:deBruijn_base t) : deBruijn_extension t deBruijn_t (list t) (deBruijn_base_list t deBruijn_t).
    intros.
    split with (fun x:t => cons x nil).
    unfold injective.
    intros.
    injection H;auto.
    intros;simpl;trivial.
    intros;simpl;trivial.
    intros;simpl;trivial.
  Defined.

  Definition deBruijn_subst_list (t:Set)(deBruijn_t:deBruijn_base t)(to:Set)(deBruijn_to:deBruijn_base to)(t_subst_to:deBruijn_subst t deBruijn_t to deBruijn_to) : deBruijn_subst (list t) (deBruijn_base_list t deBruijn_t) to deBruijn_to :=  deBruijn_s_map_struct t deBruijn_t to deBruijn_to t_subst_to (list t) (map_list t) (map_struct_map_list t).

  Definition map_pair (t1:Set)(t2:Set)(f1:nat->t1->t1)(f2:nat->t2->t2)(d:nat)(z:t1*t2) := let (x,y) := z in (f1 d x,f2 d y).
  
  Lemma map2_struct_map_pair : forall (t1 t2:Set),(map2_struct t1 t2 (t1*t2) (map_pair t1 t2)).
    split;intros.
    destruct x;reflexivity.
    unfold injective.
    destruct x;destruct y;simpl.
    intros.
    injection H1;clear H1;intros.
    rewrite (H0 d t0 t4 H1).
    rewrite (H d t t3 H2);trivial.
    destruct x;simpl.
    rewrite H.
    rewrite H0;trivial.
    destruct x;simpl.
    rewrite H.
    rewrite H0;trivial.
  Qed.

  Definition deBruijn_base_pair (t1:Set)(deBruijn_t1:deBruijn_base t1)(t2:Set)(deBruijn_t2:deBruijn_base t2) : deBruijn_base (t1*t2) := deBruijn_map2_struct t1 deBruijn_t1 t2 deBruijn_t2 (t1*t2) (map_pair t1 t2) (map2_struct_map_pair t1 t2).

  Definition deBruijn_subst_pair (t1:Set)(deBruijn_t1:deBruijn_base t1)(t2:Set)(deBruijn_t2:deBruijn_base t2)(to:Set)(deBruijn_to:deBruijn_base to)(t1_subst_to:deBruijn_subst t1 deBruijn_t1 to deBruijn_to)(t2_subst_to:deBruijn_subst t2 deBruijn_t2 to deBruijn_to) : deBruijn_subst (t1*t2) (deBruijn_base_pair t1 deBruijn_t1 t2 deBruijn_t2) to deBruijn_to :=  deBruijn_s_map2_struct t1 deBruijn_t1 t2 deBruijn_t2 to deBruijn_to t1_subst_to t2_subst_to (t1*t2) (map_pair t1 t2) (map2_struct_map_pair t1 t2).

  Fixpoint fold_list (t:Set)(A:Set)(f:nat->t->A->A)(d:nat)(ts:list t)(a:A) {struct ts} : A :=
    match ts with
      | nil => a
      | (cons x ts) => fold_list t A f d ts (f d x a)
    end.

  Lemma fold_list_app : forall (t:Set)(A:Set)(f:nat->t->A->A)(d:nat)(xs ys:list t)(a:A),(fold_list t A f d (app xs ys) a)=(fold_list t A f d ys (fold_list t A f d xs a)).
    induction xs;intros;try reflexivity.
    simpl.
    rewrite IHxs.
    auto.
  Qed.

  Lemma fold_list_false_inv : forall (t:Set)(f:nat->t->bool)(d:nat)(x:list t)(b:bool),(fold_list t bool (fun d x b => b || f d x) d x b)=false->b=false.
    induction x;simpl;intros;auto.
    generalize (IHx (b || (f d a)) H);intro.
    elim (orb_false_elim b (f d a) H0);auto.
  Qed.

  Implicit Arguments fold_list_false_inv [t f d x b].

  Lemma fold_list_true : forall (t:Set)(f:nat->t->bool)(d:nat)(x:list t),(fold_list t bool (fun d x b => b || f d x) d x true)=true.
    induction x;simpl;intros;auto.
  Qed.

  Lemma fold_struct_fold_list : forall (t:Set),(fold_struct t (list t) (map_list t) (map_struct_map_list t) (fold_list t)).
    intro;split.
    (* fold_orb_eq *)
    intros.
    induction x;simpl;auto.
    rewrite H.
    rewrite IHx.
    auto.
    intros.
    simpl in H0.
    generalize (H0 i H1).
    intro.
    rewrite (fold_list_false_inv H2) in H2.
    auto.
    intros.
    simpl in H0.
    apply (fold_list_false_inv (H0 i H1)).
    (* fold_exists *)
    induction x;simpl;intros;auto.
    case (Bool_elim_bool (f (d1+d) a)).
    intro.
    rewrite H;auto.
    apply fold_list_true.
    intro.
    rewrite H1 in H0.
    destruct (g (d2+d) a).
    apply fold_list_true.
    auto.
    (* fold_map_simpl *)
    induction x;simpl;intros;auto.
    (* fold_equal_cond_invariant *)
    induction x;simpl;intros;auto.
    rewrite <- H;auto.
    rewrite <- IHx;auto.
    intros.
    generalize (H1 i H3);intro.
    rewrite (fold_list_false_inv H4) in H4;auto.
    intros.
    apply (fold_list_false_inv (H1 i H3)).
    (* fold_invariant *)
    induction x;simpl;intros;auto.
    (* fold_orb_eq_map_inversion *)
    induction x;simpl;intros.
    destruct y;try (discriminate H2);auto.
    destruct y;try (discriminate H2).
    simpl in H2.
    injection H2;intros;auto.
    rewrite (H d a t0);intros;auto.
    rewrite (IHx y);intros;auto.
    generalize (H0 i H5).
    destruct (fi i (d1+d) a);auto.
    rewrite fold_list_true;intros.
    discriminate H6.
    generalize (H1 i H5).
    simpl.
    destruct (fi i (d1+d) t0);auto.
    rewrite fold_list_true;intros.
    discriminate H6.
    generalize (H0 i H5).
    destruct (fi i (d1+d) a);auto.
    rewrite fold_list_true;intros.
    discriminate H6.
    generalize (H1 i H5).
    simpl.
    destruct (fi i (d1+d) t0);auto.
    rewrite fold_list_true;intros.
    discriminate H6.
  Qed.

  Definition deBruijn_fn_base_list (t:Set)(deBruijn_t:deBruijn_base t)(deBruijn_fn_t:deBruijn_fn t deBruijn_t) : deBruijn_fn (list t) (deBruijn_base_list t deBruijn_t) := deBruijn_fn_map_struct t (list t) (map_list t) (map_struct_map_list t) (fold_list t) deBruijn_t deBruijn_fn_t (fold_struct_fold_list t).

  Definition deBruijn_fn_subst_list (t:Set)(deBruijn_t:deBruijn_base t)(deBruijn_fn_t:deBruijn_fn t deBruijn_t)(to:Set)(deBruijn_to:deBruijn_base to)(deBruijn_fn_to:deBruijn_fn to deBruijn_to) (t_subst_to:deBruijn_subst t deBruijn_t to deBruijn_to) (t_fn_subst_to:deBruijn_subst_fn t deBruijn_t deBruijn_fn_t to deBruijn_to deBruijn_fn_to t_subst_to) : deBruijn_subst_fn (list t) (deBruijn_base_list t deBruijn_t) (deBruijn_fn_base_list t deBruijn_t deBruijn_fn_t) to deBruijn_to deBruijn_fn_to (deBruijn_subst_list t deBruijn_t to deBruijn_to t_subst_to) := deBruijn_s_fn_fold_struct t deBruijn_t deBruijn_fn_t to deBruijn_to deBruijn_fn_to t_subst_to t_fn_subst_to (list t) (map_list t) (map_struct_map_list t) (fold_list t) (fold_struct_fold_list t).

  Lemma mem_fn_nil : forall (t:Set)(deBruijn_t:deBruijn_base t)(deBruijn_fn_t:deBruijn_fn t deBruijn_t),forall (n d:nat),(deBruijn.mem_fn (deBruijn_fn_base_list t deBruijn_t deBruijn_fn_t) n d nil)=false.
    intros;reflexivity.
  Qed.

  Lemma mem_fn_cons : forall (t:Set)(deBruijn_t:deBruijn_base t)(deBruijn_fn_t:deBruijn_fn t deBruijn_t),forall (n d:nat)(x:t)(xs:list t),(deBruijn.mem_fn (deBruijn_fn_base_list t deBruijn_t deBruijn_fn_t) n d (cons x xs))=orb (deBruijn.mem_fn deBruijn_fn_t n d x) (deBruijn.mem_fn (deBruijn_fn_base_list t deBruijn_t deBruijn_fn_t) n d xs).
    intros.
    case (Bool_elim_bool (mem_fn deBruijn_fn_t n d x));intro.
    simpl;simpl in H;rewrite H.
    rewrite fold_list_true.
    auto.
    simpl;simpl in H;rewrite H.
    simpl;auto.
  Qed.

  Lemma mem_fn_app : forall (t:Set)(deBruijn_t:deBruijn_base t)(deBruijn_fn_t:deBruijn_fn t deBruijn_t),forall (n d:nat)(xs ys:list t),(deBruijn.mem_fn (deBruijn_fn_base_list t deBruijn_t deBruijn_fn_t) n d (app xs ys))=(orb (deBruijn.mem_fn (deBruijn_fn_base_list t deBruijn_t deBruijn_fn_t) n d xs) (deBruijn.mem_fn (deBruijn_fn_base_list t deBruijn_t deBruijn_fn_t) n d ys)).
    intros.
    simpl.
    rewrite deBruijn.fold_list_app.
    case (Bool_elim_bool (deBruijn.mem_fn (deBruijn_fn_base_list t deBruijn_t deBruijn_fn_t) n d xs));simpl;intros.
    rewrite H.
    simpl.
    generalize (deBruijn.fold_list_true t (fun d0 x => mem_fn deBruijn_fn_t n d0 x));simpl;intro.
    apply H0.
    rewrite H;auto.
  Qed.

  Implicit Arguments mem_fn_app [t deBruijn_t deBruijn_fn_t].

  Implicit Arguments map_list [t].

  Definition fold_pair (t1 t2:Set)(A:Set)(f1:nat->t1->A->A)(f2:nat->t2->A->A)(d:nat)(z:t1*t2)(a:A) : A := 
    let (x,y) := z in (f1 d x (f2 d y a)).

  Lemma fold_pair_false_inv : forall (t1 t2:Set)(f1:nat->t1->bool)(f2:nat->t2->bool)(d:nat)(x:t1*t2)(b:bool),(fold_pair t1 t2 bool (fun d x b => b || f1 d x) (fun d x b => b || f2 d x) d x b)=false->b=false.
    destruct x;simpl;intros;auto.
    destruct b;auto.
  Qed.

  Implicit Arguments fold_pair_false_inv [t1 t2 f1 f2 d x b].

  Lemma fold_pair_true : forall (t1 t2:Set)(f1:nat->t1->bool)(f2:nat->t2->bool)(d:nat)(x:t1*t2),(fold_pair t1 t2 bool (fun d x b => b || f1 d x) (fun d x b => b || f2 d x) d x true)=true.
    destruct x;simpl;intros;auto.
  Qed.

  Lemma fold2_struct_fold_pair : forall (t1 t2:Set),(fold2_struct t1 t2 (t1*t2) (map_pair t1 t2) (map2_struct_map_pair t1 t2) (fold_pair t1 t2)).
    split.
    (* fold_orb_eq *)
    destruct x;simpl;intros;auto.
    rewrite H;auto.
    rewrite H0;auto.
    intros.
    elim (orb_false_elim (f'i i (d1+d) t0) (fi i (d1+d) t) (H1 i H2));auto.
    intros.
    elim (orb_false_elim (f'i i (d1+d) t0) (fi i (d1+d) t) (H1 i H2));auto.
    (* fold_exists *)
    destruct x;simpl;intros;auto.
    case (Bool_elim_bool (f1 (d1+d) t));intro.
    rewrite H;auto.
    rewrite orb_b_true;auto.
    rewrite H2 in H1.
    rewrite orb_b_false in H1.
    rewrite H0;auto.
    (* fold_map_simpl *)
    destruct x;simpl;intros;auto.
    (* fold_equal_cond_invariant *)
    destruct x;simpl;intros;auto.
    rewrite <- H0;auto.
    rewrite <- H;auto.
    intros.
    elim (orb_false_elim (f'i i (d1+d) t0) (fi i (d1+d) t) (H3 i H5));auto.
    intros.
    elim (orb_false_elim (f'i i (d1+d) t0) (fi i (d1+d) t) (H3 i H5));auto.
    (* fold_invariant *)
    destruct x;simpl;intros;auto.
    (* fold_orb_eq_map_inversion *)
    intros.
    destruct x as [x1 x2].
    destruct y as [y1 y2].
    simpl in H3.
    simpl in H2.
    simpl in H1.
    injection H3;clear H3;intros.
    rewrite (H d x1 y1);intros;auto.
    rewrite (H0 d x2 y2);intros;auto.
    generalize (H1 i H5).
    destruct (f'i i (d1+d) x2);intros;auto.
    generalize (H2 i H5).
    destruct (f'i i (d1+d) y2);intros;auto.
    generalize (H1 i H5).
    destruct (fi i (d1+d) x1);intros;auto.
    rewrite orb_b_true in H6;discriminate H6.
    generalize (H2 i H5).
    destruct (fi i (d1+d) y1);intros;auto.
    rewrite orb_b_true in H6;discriminate H6.
  Qed.

  Definition deBruijn_fn_base_pair (t1:Set)(deBruijn_t1:deBruijn_base t1)(deBruijn_fn_t1:deBruijn_fn t1 deBruijn_t1)(t2:Set)(deBruijn_t2:deBruijn_base t2)(deBruijn_fn_t2:deBruijn_fn t2 deBruijn_t2) : deBruijn_fn (t1*t2) (deBruijn_base_pair t1 deBruijn_t1 t2 deBruijn_t2) := deBruijn_fn_map2_struct t1 t2 (t1*t2) (map_pair t1 t2) (map2_struct_map_pair t1 t2) (fold_pair t1 t2) deBruijn_t1 deBruijn_fn_t1 deBruijn_t2 deBruijn_fn_t2 (fold2_struct_fold_pair t1 t2).

  Definition deBruijn_fn_subst_pair (t1:Set)(deBruijn_t1:deBruijn_base t1)(deBruijn_fn_t1:deBruijn_fn t1 deBruijn_t1)(t2:Set)(deBruijn_t2:deBruijn_base t2)(deBruijn_fn_t2:deBruijn_fn t2 deBruijn_t2)(to:Set)(deBruijn_to:deBruijn_base to)(deBruijn_fn_to:deBruijn_fn to deBruijn_to) (t1_subst_to:deBruijn_subst t1 deBruijn_t1 to deBruijn_to) (t1_fn_subst_to:deBruijn_subst_fn t1 deBruijn_t1 deBruijn_fn_t1 to deBruijn_to deBruijn_fn_to t1_subst_to)(t2_subst_to:deBruijn_subst t2 deBruijn_t2 to deBruijn_to) (t2_fn_subst_to:deBruijn_subst_fn t2 deBruijn_t2 deBruijn_fn_t2 to deBruijn_to deBruijn_fn_to t2_subst_to) : deBruijn_subst_fn (t1*t2) (deBruijn_base_pair t1 deBruijn_t1 t2 deBruijn_t2) (deBruijn_fn_base_pair t1 deBruijn_t1 deBruijn_fn_t1 t2 deBruijn_t2 deBruijn_fn_t2) to deBruijn_to deBruijn_fn_to (deBruijn_subst_pair t1 deBruijn_t1 t2 deBruijn_t2 to deBruijn_to t1_subst_to t2_subst_to) := deBruijn_s_fn_fold2_struct t1 deBruijn_t1 deBruijn_fn_t1 t2 deBruijn_t2 deBruijn_fn_t2 to deBruijn_to deBruijn_fn_to t1_subst_to t1_fn_subst_to t2_subst_to t2_fn_subst_to (t1*t2) (map_pair t1 t2) (map2_struct_map_pair t1 t2) (fold_pair t1 t2) (fold2_struct_fold_pair t1 t2).
End deBruijn.








