(* A formalisation of the spi calculus in Coq *)
(* Copyright (C) 2005-2007 Sebastien Briais *)
(* http://lamp.epfl.ch/~sbriais/ *)

(* This library is free software; you can redistribute it and/or modify *)
(* it under the terms of the GNU Lesser General Public License as *)
(* published by the Free Software Foundation; either version 2.1 of the *)
(* License, or (at your option) any later version. *)

(* This library is distributed in the hope that it will be useful, but *)
(* WITHOUT ANY WARRANTY; without even the implied warranty of *)
(* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU *)
(* Lesser General Public License for more details. *)

(* You should have received a copy of the GNU Lesser General Public *)
(* License along with this library; if not, write to the Free Software *)
(* Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA *)
(* 02110-1301 USA *)

Require Import missing.
Require Import ordmissing.
Require Import tactics.
Require Import Relations.
Require Import FSets.
Require Import Arith.
Require Import Max.
Require Import name.
Require Import opSpi.
Require Import messageSpi.
Require Import deBruijnType.
Require Import permutation.

Module Type MsgSet := FSetInterface.S with Module E := MessageSpi.Ord.

Module MessagePair := OrderedPair MessageSpi.Ord MessageSpi.Ord.

Module Type Hedge := FSetInterface.S with Module E := MessagePair.

(* Module HedgeOp (H:Hedge)(M:MsgSet). *)
Module HedgeOp (H:Hedge).
  (** a hedge is a finite set of pairs of messages *)
  Definition hedge := H.t.
  Module HedgeEltProps := OrderedTypeFacts H.E.
  Module HedgeEqProp := FSetEqProperties.EqProperties H.
  Module HedgeProp := FSetProperties.Properties H.
  Module HedgeOtherProp := MissingSetProperties H.
  Definition HedgeSetoid := HedgeOtherProp.setSetoid.

  Require Import Arith.
  Require Import Max.

  (** we define the size of a pair of message *)
  (** this will be used to prove termination of hedge_analysis *)
  Definition size_pair (mn:H.elt) := let (m,n) := mn in max (MessageSpi.size m) (MessageSpi.size n).

  Definition size_pair_ge (x y:H.elt) := (size_pair x)>=(size_pair y).

  Require Import set_order.

  Module HedgeOrder := SetOrder(H).

  Lemma size_pair_ge_refl : forall (x y:H.elt),(H.E.eq x y)->(size_pair_ge x y).
    unfold size_pair_ge.
    destruct x;destruct y.
    simpl;intros;and_elim.
    rewrite H;rewrite H0;auto with arith.
  Qed.

  Lemma size_pair_ge_trans : forall (x y z:H.elt),(size_pair_ge x y)->(size_pair_ge y z)->(size_pair_ge x z).
    unfold size_pair_ge;unfold ge;intros.
    eapply le_trans;eauto.
  Qed.

  Lemma size_pair_ge_total_dec : forall (x y:H.elt),{size_pair_ge x y}+{size_pair_ge y x}.
    unfold size_pair_ge;unfold ge;intros.
    case (le_lt_dec (size_pair y) (size_pair x));intro.
    left;trivial.
    right.
    auto with arith.
  Qed.

  Lemma size_pair_eq : forall (x y:H.elt),(H.E.eq x y)->(size_pair x)=(size_pair y).
    destruct x;destruct y;simpl;intros;and_elim.
    rewrite H;rewrite H0;trivial.
  Qed.

  (** induction principle: it is the set induction principle but we chose that the element added is of maximal size *)
  Lemma set_induction_size_pair : forall (P : H.t -> Prop),(forall (s:H.t),(H.Empty s)->(P s)) -> (forall (s s':H.t),(P s) -> forall (x:H.elt),~(H.In x s) -> (HedgeProp.Add x s s') -> (H.For_all (fun y => (size_pair y) <= (size_pair x)) s) -> (P s')) -> forall (s:H.t),(P s).
    intros.
    apply (HedgeOtherProp.induction_such_that size_pair_ge size_pair_ge_total_dec size_pair_ge_refl size_pair_ge_trans);auto.
  Qed.

  Definition add_single_op (f:H.elt->H.elt)(mn:H.elt)(h:hedge) := H.add (f mn) h.

  Lemma compat_op_add_single_op : forall (f:H.elt->H.elt),compat_op H.E.eq H.Equal (add_single_op f).
    unfold add_single_op.
    unfold compat_op.
    intros.
    generalize HedgeOtherProp.compat_op_add.
    intro.
    destruct x;destruct x';simpl in H;and_elim.
    rewrite H;rewrite H2.
    apply H1;auto.
  Qed.

  Lemma transpose_add_single_op : forall (f:H.elt->H.elt),transpose H.Equal (add_single_op f).
    unfold add_single_op, transpose.
    intros.
    generalize HedgeOtherProp.transpose_add.
    intros.
    apply H.
  Qed.

  Definition hedge_map_single_op (f:H.elt->H.elt)(h:hedge) := H.fold (add_single_op f) h H.empty.

  Definition f_injective (f:H.elt->H.elt) := forall (x y:H.elt),(H.E.eq (f x) (f y))->(H.E.eq x y).

  Lemma hedge_map_single_op_prop : forall (f:H.elt->H.elt),(f_injective f)->forall (h:hedge)(mn:H.elt),(H.In mn h)<->(H.In (f mn) (hedge_map_single_op f h)).
    intros f f_inj h.
    elim h using HedgeProp.set_induction;clear h;intros.
    unfold hedge_map_single_op.
    generalize (HedgeProp.fold_1 (s:=s) HedgeSetoid H.empty (add_single_op f) H);intro.
    split;intros.
    elim (H mn);auto.
    generalize (HedgeOtherProp.Equal_empty_empty (H.fold (add_single_op f) s H.empty) H0);intro.
    elim (H2 (f mn));auto.
    unfold hedge_map_single_op.
    generalize (HedgeProp.fold_2 (s:=s) (s':=s') (x:=x) HedgeSetoid H.empty (compat_op_add_single_op f) (transpose_add_single_op f) H0 H1);intro.
    assert (HedgeProp.Add (f x) (H.fold (add_single_op f) s H.empty) (H.fold (add_single_op f) s' H.empty)).
    generalize (HedgeProp.Add_add (H.fold (add_single_op f) s H.empty) (f x));intro.
    fold (add_single_op f x (H.fold (add_single_op f) s H.empty)) in H3.
    split;intros.
    elim (H3 y);intros.
    elim (H2 y);auto.
    elim (H3 y);intros.
    elim (H2 y);auto.
    split;intro.
    elim (H1 mn);intros.
    case (H5 H4);intros.
    destruct mn;destruct x;simpl in H7;and_elim.
    rewrite <- H7;rewrite <- H8.
    elim (H3 (f (t1,t2)));auto.
    elim (H mn);intros.
    elim (H3 (f mn));auto.
    elim (H3 (f mn));intros.
    case (H5 H4);intros.
    generalize (f_inj mn x (H.E.eq_sym _ _ H7));intro.
    eapply H.In_1.
    apply H.E.eq_sym.
    apply H8.
    elim (H1 x);auto.
    elim (H mn);intros.
    elim (H1 mn);auto.
  Qed.

  Lemma hedge_map_single_op_inv : forall (f:H.elt->H.elt)(h:hedge)(mn:H.elt),(H.In mn (hedge_map_single_op f h))->(exists mn':H.elt,(H.In mn' h)/\(mn=(f mn'))).
    unfold hedge_map_single_op.
    intros f h.
    elim h using HedgeProp.set_induction;clear h;intros.
    generalize (HedgeProp.fold_1 (s:=s) HedgeSetoid H.empty (add_single_op f) H);intro.
    elim (H1 mn);intros.
    generalize (H2 H0);intro.
    elim (H.empty_1 H4).
    generalize (HedgeProp.fold_2 (s:=s) (s':=s') (x:=x) HedgeSetoid H.empty (compat_op_add_single_op f) (transpose_add_single_op f) H0 H1);intro.
    elim (H3 mn);intros.
    generalize (H4 H2);clear H4;clear H5;intro.
    case (HedgeEltProps.eq_dec (f x) mn);intro.
    exists x.
    split.
    elim (H1 x);auto.
    destruct mn;destruct (f x);simpl in e;and_elim.
    rewrite e;rewrite H5;trivial.
    generalize (H.add_3 n H4);intro.
    elim (H mn H5);intro mn';intro.
    and_elim.
    exists mn'.
    elim (H1 mn');split;auto.
  Qed.

  Definition add_two_op (f:H.elt->H.elt->H.elt)(mn1 mn2:H.elt)(h:hedge) := H.add (f mn1 mn2) h.

  Lemma add_two_op_single_op : forall (f:H.elt->H.elt->H.elt)(mn1 mn2:H.elt)(h:hedge),(H.Equal (add_two_op f mn1 mn2 h) (add_single_op (f mn1) mn2 h)).
    intros.
    unfold add_two_op.
    unfold add_single_op.
    apply HedgeProp.equal_refl.
  Qed.

  Lemma compat_op_add_two_op : forall (f:H.elt->H.elt->H.elt)(mn1:H.elt),compat_op H.E.eq H.Equal (add_two_op f mn1).
    unfold compat_op.
    intros.
    eapply HedgeProp.equal_trans.
    apply add_two_op_single_op.
    apply HedgeProp.equal_sym.
    eapply HedgeProp.equal_trans.
    apply add_two_op_single_op.
    apply HedgeProp.equal_sym.
    apply compat_op_add_single_op;auto.
  Qed.

  Lemma transpose_add_two_op : forall (f:H.elt->H.elt->H.elt)(mn1:H.elt),transpose H.Equal (add_two_op f mn1).
    unfold add_two_op, transpose.
    intros.
    generalize HedgeOtherProp.transpose_add.
    intros.
    apply H.
  Qed.

  Lemma add_single_op_add : forall (f:H.elt->H.elt)(h h':hedge)(x:H.elt),(H.Equal (H.fold (add_single_op f) h (H.add x h')) (H.add x (H.fold (add_single_op f) h h'))).
    intros.
    elim h using HedgeProp.set_induction;clear h;intros.
    eapply HedgeProp.equal_trans.
    apply HedgeProp.fold_1;trivial.
    exact HedgeSetoid.
    apply HedgeOtherProp.compat_op_add.
    auto.
    apply HedgeProp.equal_sym.
    apply HedgeProp.fold_1;trivial.
    exact HedgeSetoid.
    eapply HedgeProp.equal_trans.
    apply (HedgeProp.fold_2 (s:=s) (s':=s') (x:=x0) HedgeSetoid);auto.
    apply compat_op_add_single_op.
    apply transpose_add_single_op.
    apply HedgeProp.equal_sym.
    eapply HedgeProp.equal_trans.
    apply HedgeOtherProp.compat_op_add.
    apply H.E.eq_refl.
    apply (HedgeProp.fold_2 (s:=s) (s':=s') (x:=x0) HedgeSetoid);auto.
    apply compat_op_add_single_op.
    apply transpose_add_single_op.
    eapply HedgeProp.equal_trans.
    unfold add_single_op.
    apply HedgeOtherProp.transpose_add.
    unfold add_single_op.
    apply HedgeOtherProp.compat_op_add.
    auto.
    apply HedgeProp.equal_sym;auto.
  Qed.

  Lemma compat_op_add_two_op_hedge : forall (f:H.elt->H.elt->H.elt)(h2:hedge),compat_op H.E.eq H.Equal (fun mn1 => fun h => (H.fold (add_two_op f mn1) h2 h)).
    unfold compat_op.
    intros.
    destruct x;destruct x'.
    simpl in H;and_elim.
    rewrite H;rewrite H1.
    elim h2 using HedgeProp.set_induction;clear h2;intros.
    eapply HedgeProp.equal_trans.
    apply HedgeProp.fold_1.
    exact HedgeSetoid.
    trivial.
    eapply HedgeProp.equal_trans.
    apply H0.
    apply HedgeProp.equal_sym.
    apply HedgeProp.fold_1.
    exact HedgeSetoid.
    trivial.
    eapply HedgeProp.equal_trans.
    apply (HedgeProp.fold_2 (s:=s) (s':=s') (x:=x) HedgeSetoid y (compat_op_add_two_op f (t1,t2)) (transpose_add_two_op f (t1,t2)) H3 H4).
    eapply HedgeProp.equal_trans.
    apply compat_op_add_two_op.
    apply H.E.eq_refl.
    apply H2.
    apply HedgeProp.equal_sym.
    apply HedgeProp.fold_2;trivial.
    exact HedgeSetoid.
    apply compat_op_add_two_op.
    apply transpose_add_two_op.
  Qed.

  Lemma transpose_add_two_op_hedge : forall (f:H.elt->H.elt->H.elt)(h2:hedge),transpose H.Equal (fun mn1 => fun h => (H.fold (add_two_op f mn1) h2 h)).
    unfold transpose.
    intros.
    elim h2 using HedgeProp.set_induction;clear h2;intros.
    eapply HedgeProp.equal_trans.
    apply HedgeProp.fold_1;auto.
    exact HedgeSetoid.
    eapply HedgeProp.equal_trans.
    apply HedgeProp.fold_1;auto.
    exact HedgeSetoid.
    apply HedgeProp.equal_sym.
    eapply HedgeProp.equal_trans.
    apply HedgeProp.fold_1;auto.
    exact HedgeSetoid.
    apply HedgeProp.fold_1;auto.
    exact HedgeSetoid.
    eapply HedgeProp.equal_trans.
    apply (HedgeProp.fold_2 (s:=s) (s':=s') (x:=x0) HedgeSetoid);auto.
    apply compat_op_add_two_op.
    apply transpose_add_two_op.
    apply HedgeProp.equal_sym.
    eapply HedgeProp.equal_trans.
    apply (HedgeProp.fold_2 (s:=s) (s':=s') (x:=x0) HedgeSetoid);auto.
    apply compat_op_add_two_op.
    apply transpose_add_two_op.
    eapply HedgeProp.equal_trans.
    apply compat_op_add_two_op.
    apply H.E.eq_refl.
    apply compat_op_add_two_op_hedge.
    apply H.E.eq_refl.
    apply (HedgeProp.fold_2 (s:=s) (s':=s') (x:=x0) HedgeSetoid);auto.
    apply compat_op_add_two_op.
    apply transpose_add_two_op.
    eapply HedgeProp.equal_sym.
    eapply HedgeProp.equal_trans.
    apply compat_op_add_two_op.
    apply H.E.eq_refl.
    apply compat_op_add_two_op_hedge.
    apply H.E.eq_refl.
    apply (HedgeProp.fold_2 (s:=s) (s':=s') (x:=x0) HedgeSetoid);auto.
    apply compat_op_add_two_op.
    apply transpose_add_two_op.
    eapply HedgeProp.equal_trans.
    apply compat_op_add_two_op.
    apply H.E.eq_refl.
    unfold add_two_op.
    fold (add_single_op (f x)).
    apply add_single_op_add.
    apply HedgeProp.equal_sym.
    eapply HedgeProp.equal_trans.
    apply compat_op_add_two_op.
    apply H.E.eq_refl.
    unfold add_two_op.
    fold (add_single_op (f y)).
    apply add_single_op_add.
    unfold add_two_op.
    eapply HedgeProp.equal_trans.
    apply HedgeOtherProp.transpose_add.
    apply HedgeOtherProp.compat_op_add.
    auto.
    apply HedgeOtherProp.compat_op_add.
    auto.
    apply HedgeProp.equal_sym.
    apply H.
  Qed.

  Definition hedge_map_two_op (f:H.elt->H.elt->H.elt)(h1 h2:hedge) := H.fold (fun mn1 => fun h => H.fold (add_two_op f mn1) h2 h) h1 H.empty.

  Lemma hedge_map_two_op_prop_aux : forall (f:H.elt->H.elt->H.elt)(mn1:H.elt)(h h2:hedge)(mn:H.elt),(H.In mn (H.fold (add_two_op f mn1) h2 h))<->(H.In mn h \/ (exists mn2:H.elt,(H.In mn2 h2)/\(H.E.eq mn (f mn1 mn2)))).
    intros f mn1 h.
    intro h2.
    elim h2 using HedgeProp.set_induction;clear h2;intros.
    generalize (HedgeProp.fold_1 (s:=s) HedgeSetoid h (add_two_op f mn1) H);intro.
    split;intros.
    elim (H0 mn);intros;tauto.
    case H1;intros.
    elim (H0 mn);intros;auto.
    ex_elim H2 mn2.
    elim (H mn2);trivial.
    generalize (HedgeProp.fold_2 (s:=s) (s':=s') (x:=x) HedgeSetoid h (compat_op_add_two_op f mn1) (transpose_add_two_op f mn1) H0 H1);intro.
    split;intros.
    elim (H2 mn);intros.
    generalize (H4 H3);clear H4;clear H5;intro.
    case (HedgeEltProps.eq_dec (f mn1 x) mn);intro.
    right.
    exists x.
    split;auto.
    elim (H1 x);auto.
    generalize (H.add_3 n H4);intro.
    elim (H mn);intros.
    case (H6 H5);clear H6;clear H7;intros.
    left;trivial.
    right.
    ex_elim H6 mn2.
    exists mn2.
    split;trivial.
    elim (H1 mn2);auto.
    case H3;intro.
    elim (H2 mn);intros.
    apply H6;clear H5;clear H6.
    unfold add_two_op.
    apply H.add_2.
    fold (add_two_op f mn1).
    elim (H mn);intros.
    apply H6.
    left;trivial.
    ex_elim H4 mn2;clear H3.
    case (HedgeEltProps.eq_dec x mn2);intro.
    elim (H2 mn);intros.
    apply H6;clear H3;clear H6.
    unfold add_two_op.
    apply H.add_1.
    destruct x;destruct mn2.
    simpl in e;and_elim.
    rewrite e;rewrite H3;auto.
    elim (H2 mn);intros.
    apply H6.
    unfold add_two_op.
    apply H.add_2.
    fold (add_two_op f mn1).
    elim (H1 mn2);intros.
    case (H7 H4);intro.
    elim n.
    auto.
    elim (H mn);intros.
    apply H11.
    right.
    exists mn2.
    tauto.
  Qed.

  Definition f_injective2 (f:H.elt->H.elt->H.elt) := forall (mn1 mn2 mn1' mn2':H.elt),(H.E.eq (f mn1 mn2) (f mn1' mn2'))->(H.E.eq mn1 mn1')/\(H.E.eq mn2 mn2').

  Lemma hedge_map_two_op_prop : forall (f:H.elt->H.elt->H.elt),(f_injective2 f)->forall (h2 h1:hedge)(mn1 mn2:H.elt),(H.In mn1 h1)/\(H.In mn2 h2)<->(H.In (f mn1 mn2) (hedge_map_two_op f h1 h2)).
    unfold hedge_map_two_op.
    intros f f_inj2 h2.
    intro h1.
    elim h1 using HedgeProp.set_induction;clear h1;intros.
    generalize (HedgeProp.fold_1 (s:=s) HedgeSetoid H.empty (fun mn h => H.fold (add_two_op f mn) h2 h) H);intros.
    split;intros;and_elim.
    elim (H mn1);auto.
    elim (H0 (f mn1 mn2));intros.
    generalize (H2 H1);clear H2;clear H3;intro.
    elim (H.empty_1 H2).
    generalize (HedgeProp.fold_2 (s:=s) (s':=s') (x:=x) HedgeSetoid H.empty (compat_op_add_two_op_hedge f h2) (transpose_add_two_op_hedge f h2) H0 H1);intro.
    elim (H2 (f mn1 mn2));intros.
    split;intros.
    and_elim.
    apply H4;clear H3;clear H4.
    elim (hedge_map_two_op_prop_aux f x (H.fold (fun (mn3 : H.E.t) (h : H.t) => H.fold (add_two_op f mn3) h2 h) s H.empty) h2 (f mn1 mn2));intros.
    apply H4.
    elim (H1 mn1);intros.
    case (H7 H5);clear H7;clear H8;intro.
    right.
    exists mn2.
    split;trivial.
    destruct mn1;destruct x;simpl in H7;and_elim.
    rewrite H7;rewrite H8.
    auto.
    left.
    elim (H mn1 mn2);intros.
    apply H8.
    split;trivial.
    generalize (H3 H5);clear H3;clear H4;intros.
    elim (hedge_map_two_op_prop_aux f x (H.fold (fun (mn3 : H.E.t) (h : H.t) => H.fold (add_two_op f mn3) h2 h) s H.empty) h2 (f mn1 mn2));intros.
    case (H4 H3);clear H4;clear H6;intro.
    elim (H mn1 mn2);intros.
    generalize (H7 H4);clear H6;clear H7;intro.
    split;try tauto.
    and_elim.
    elim (H1 mn1);auto.
    ex_elim H4 mn3.
    elim (f_inj2 mn1 mn2 x mn3 H6);intros.
    split.
    elim (H1 x);intros.
    eapply H.In_1.
    apply H.E.eq_sym.
    apply H7.
    auto.
    eapply H.In_1.
    apply H.E.eq_sym.
    apply H8.
    trivial.
  Qed.

  Lemma hedge_map_two_op_inv : forall (f:H.elt->H.elt->H.elt)(h2 h1:hedge)(mn:H.elt),(H.In mn (hedge_map_two_op f h1 h2))->(exists mn1:H.elt,exists mn2:H.elt,(H.In mn1 h1)/\(H.In mn2 h2)/\mn=(f mn1 mn2)).
    intros f h2.
    unfold hedge_map_two_op.
    intro h1.
    elim h1 using HedgeProp.set_induction;clear h1;intros.
    generalize (HedgeProp.fold_1 (s:=s) HedgeSetoid H.empty (fun (mn1 : H.elt) (h : hedge) => H.fold (add_two_op f mn1) h2 h) H);intro.
    elim (H1 mn);intros.
    elim (H.empty_1 (H2 H0)).
    generalize (HedgeProp.fold_2 (s:=s) (s':=s') (x:=x) HedgeSetoid H.empty (compat_op_add_two_op_hedge f h2) (transpose_add_two_op_hedge f h2) H0 H1);intro.
    elim (H3 mn);intros.
    generalize (H4 H2);clear H4;clear H5;intro.
    elim (hedge_map_two_op_prop_aux f x (H.fold (fun (mn1 : H.E.t) (h : H.t) => H.fold (add_two_op f mn1) h2 h) s H.empty) h2 mn);intros.
    case (H5 H4);clear H5;clear H6;intro.
    elim (H mn H5).
    intro mn1;intro.
    ex_elim H6 mn2.
    exists mn1.
    exists mn2.
    split.
    elim (H1 mn1);auto.
    split;trivial.
    ex_elim H5 mn2.
    exists x.
    exists mn2.
    split.
    elim (H1 x);auto.
    split;trivial.
    destruct mn;destruct (f x mn2);simpl in H6;and_elim.
    rewrite H6;rewrite H7;trivial.
  Qed.

  (** we first define the transposition of a hedge *)
  (** it consists of swapping the two messages of the pairs *)
  Definition swap_pair (mn:H.elt) := let (m,n) := mn in (n,m).

  Definition hedge_transpose (h:hedge) : hedge := hedge_map_single_op swap_pair h.

  Lemma hedge_transpose_prop : forall (h:hedge)(m n:MessageSpi.message),(H.In (m,n) h)<->(H.In (n,m) (hedge_transpose h)).
    unfold hedge_transpose.
    intros.
    replace (n,m) with (swap_pair (m,n));try reflexivity.
    apply hedge_map_single_op_prop.
    unfold f_injective.
    intros.
    destruct x;destruct y;simpl in H;and_elim.
    simpl;tauto.
  Qed.

  (** transposition is involutive *)
  Lemma hedge_transpose_involutive : forall (h:hedge),(H.Equal h (hedge_transpose (hedge_transpose h))).
    unfold H.Equal.
    intros;destruct a.
    elim (hedge_transpose_prop (hedge_transpose h) t0 t).
    elim (hedge_transpose_prop h t t0);intros.
    split;auto.
  Qed.

(*   (** projection *) *)
(*   Module MOtherProps := MissingSetProperties M. *)
(*   Definition MSetoid := MOtherProps.setSetoid. *)
(*   Module MProps := OrderedTypeFacts M.E. *)

(*   Definition add_first_proj (mn:H.elt) (ms:M.t) := M.add (let (m,_) := mn in m) ms. *)

(*   Lemma compat_op_add_first_proj : compat_op H.E.eq M.Equal add_first_proj. *)
(*     unfold compat_op. *)
(*     unfold add_first_proj. *)
(*     destruct x;destruct x';intros. *)
(*     simpl in H;and_elim. *)
(*     rewrite H. *)
(*     apply MOtherProps.compat_op_add;auto. *)
(*   Qed. *)

(*   Lemma transpose_add_first_proj : transpose M.Equal add_first_proj. *)
(*     unfold transpose. *)
(*     unfold add_first_proj. *)
(*     intros. *)
(*     apply MOtherProps.transpose_add. *)
(*   Qed. *)

(*   Definition hedge_first (h:hedge) := H.fold add_first_proj h M.empty. *)

(*   Lemma hedge_first_iff : forall (h:hedge)(m:MessageSpi.message),(M.In m (hedge_first h))<->(exists n:MessageSpi.message, H.In (m,n) h). *)
(*     split. *)
(*     unfold hedge_first. *)
(*     elim h using HedgeProp.set_induction;clear h;intros. *)
(*     generalize (HedgeProp.fold_1 (s:=s) MSetoid M.empty add_first_proj H);intro. *)
(*     elim (H1 m);intros. *)
(*     elim (M.empty_1 (H2 H0)). *)
(*     generalize (HedgeProp.fold_2 (s:=s) (s':=s') (x:=x) MSetoid M.empty compat_op_add_first_proj transpose_add_first_proj H0 H1);intro. *)
(*     elim (H3 m);intros. *)
(*     destruct x. *)
(*     generalize (H4 H2). *)
(*     unfold add_first_proj. *)
(*     simpl. *)
(*     fold add_first_proj;clear H4;clear H5;intros. *)
(*     case (MProps.eq_dec t m);intro. *)
(*     exists t0. *)
(*     elim (H1 (t,t0));intros. *)
(*     unfold M.E.eq in e. *)
(*     rewrite <- e. *)
(*     apply H6;auto. *)
(*     generalize (M.add_3 n H4);intro. *)
(*     elim (H H5). *)
(*     intros. *)
(*     exists x. *)
(*     elim (H1 (m,x));auto. *)
(*     intro. *)
(*     ex_elim H n. *)
(*     assert (~H.In (m,n) (H.remove (m,n) h)). *)
(*     auto. *)
(*     assert (HedgeProp.Add (m,n) (H.remove (m,n) h) h). *)
(*     auto. *)
(*     generalize (HedgeProp.fold_2 (s:=(H.remove (m,n) h)) (s':=h) (x:=(m,n)) MSetoid M.empty compat_op_add_first_proj transpose_add_first_proj H0 H1). *)
(*     intro. *)
(*     unfold hedge_first. *)
(*     elim (H2 m);intros. *)
(*     apply H4. *)
(*     unfold add_first_proj;simpl;fold add_first_proj. *)
(*     apply M.add_1. *)
(*     auto. *)
(*   Qed. *)

(*   Definition hedge_second (h:hedge) := hedge_first (hedge_transpose h). *)

(*   Lemma hedge_second_iff : forall (h:hedge)(n:MessageSpi.message),(M.In n (hedge_second h))<->(exists m:MessageSpi.message,(H.In (m,n) h)). *)
(*     intros. *)
(*     unfold hedge_second. *)
(*     generalize (hedge_first_iff (hedge_transpose h) n);intro. *)
(*     elim H;intros. *)
(*     generalize (hedge_transpose_prop h);intro. *)
(*     split;intros. *)
(*     elim (H0 H3);intro m;intro. *)
(*     exists m. *)
(*     elim (H2 m n);auto. *)
(*     apply H1. *)
(*     ex_elim H3 m. *)
(*     exists m. *)
(*     elim (H2 m n);auto. *)
(*   Qed. *)

  (** shortcut *)
  Definition message_relation := relation MessageSpi.message.

  (** the canonical predicate induced by a hedge *)
  Definition relation_of_hedge (h:hedge) (m n:MessageSpi.message) := H.In (m,n) h.

  (** we prove some lemma relating hedge seen as a relation and hedge seen as a finite set *)
  Lemma Subset_impl_inclusion : forall (h h':hedge),(H.Subset h h')->(inclusion MessageSpi.message (relation_of_hedge h) (relation_of_hedge h')).
    unfold H.Subset.
    intros.
    unfold relation_of_hedge.
    unfold inclusion.
    intros.
    apply H.
    trivial.
  Qed.

  Lemma inclusion_impl_Subset : forall (h h':hedge),(inclusion MessageSpi.message (relation_of_hedge h) (relation_of_hedge h'))->(H.Subset h h').
    unfold H.Subset.
    unfold relation_of_hedge.
    unfold inclusion.
    intros.
    destruct a.
    apply H;trivial.
  Qed.

  Lemma same_union_union : forall (h h':hedge),(same_relation MessageSpi.message (relation_of_hedge (H.union h h')) (union MessageSpi.message (relation_of_hedge h) (relation_of_hedge h'))).
    unfold same_relation;unfold inclusion.
    unfold relation_of_hedge.
    split;intros.
    case (H.union_1 H);intro;[left | right];trivial.
    case H;intro;[apply H.union_2 | apply H.union_3];trivial.
  Qed.

  Lemma Equal_impl_same : forall (h h':hedge),(H.Equal h h')->(same_relation MessageSpi.message (relation_of_hedge h) (relation_of_hedge h')).
    unfold same_relation.
    split;intros.
    apply Subset_impl_inclusion.
    apply HedgeProp.subset_equal;trivial.
    apply Subset_impl_inclusion.
    apply HedgeProp.subset_equal;auto.
    apply HedgeProp.equal_sym;auto.
  Qed.

  Lemma same_impl_Equal : forall (h h':hedge),(same_relation MessageSpi.message (relation_of_hedge h) (relation_of_hedge h'))->(H.Equal h h').
    unfold same_relation;intros.
    and_elim.
    apply HedgeProp.subset_antisym;apply inclusion_impl_Subset;auto.
  Qed.

  Lemma same_transpose_transp : forall (h:hedge),(same_relation MessageSpi.message (relation_of_hedge (hedge_transpose h)) (transp MessageSpi.message (relation_of_hedge h))).
    unfold same_relation, inclusion, relation_of_hedge, transp.
    intro h.
    generalize (hedge_transpose_prop h);intro.
    split;intros.
    elim (H y x);auto.
    elim (H y x);auto.
  Qed.

  Lemma hedge_transpose_grow : forall (g h:hedge),(H.Subset g h)->(H.Subset (hedge_transpose g) (hedge_transpose h)).
    intros.
    apply inclusion_impl_Subset.
    generalize (Subset_impl_inclusion g h H);intro.
    elim (same_transpose_transp g);intros.
    elim (same_transpose_transp h);intros.
    unfold inclusion;intros.
    apply H4.
    unfold transp.
    apply H0.
    apply H1.
    trivial.
  Qed.

  Lemma hedge_transpose_morphism : forall (g h:hedge),(H.Equal g h)->(H.Equal (hedge_transpose g) (hedge_transpose h)).
    intros.
    apply HedgeProp.subset_antisym;apply hedge_transpose_grow;apply HedgeProp.subset_equal;trivial.
    apply HedgeProp.equal_sym;trivial.
  Qed.
  
  Lemma hedge_transpose_union : forall (g h:hedge),(H.Equal (hedge_transpose (H.union g h)) (H.union (hedge_transpose g) (hedge_transpose h))).
    intros.
    generalize (hedge_transpose_prop (H.union g h));intro.
    generalize (hedge_transpose_prop g);intro.
    generalize (hedge_transpose_prop h);intro.
    split;intros.
    destruct a as [m n].
    elim (H n m);intros.
    case (H.union_1 (H4 H2));intro.
    apply H.union_2.
    elim (H0 n m);auto.
    apply H.union_3.
    elim (H1 n m);auto.
    destruct a as [m n].
    case (H.union_1 H2);intro.
    elim (H n m);intros.
    apply H4.
    apply H.union_2.
    elim (H0 n m);auto.
    elim (H n m);intros.
    apply H4.
    apply H.union_3.
    elim (H1 n m);auto.
  Qed.

  (** here we give the specification of the synthesis relation *)
  Inductive hedge_synthesis_relation (h:message_relation) : message_relation :=
    | SynInc : forall (m n:MessageSpi.message),(h m n)->(hedge_synthesis_relation h m n)
    | SynEnc_s : forall (m n k l:MessageSpi.message),(hedge_synthesis_relation h m n)->(hedge_synthesis_relation h k l)->(hedge_synthesis_relation h (MessageSpi.MEnc_s m k) (MessageSpi.MEnc_s n l))
    | SynEnc_a : forall (m n k l:MessageSpi.message),(hedge_synthesis_relation h m n)->(hedge_synthesis_relation h k l)->(hedge_synthesis_relation h (MessageSpi.MEnc_a m k) (MessageSpi.MEnc_a n l))
    | SynPair : forall (m1 n1 m2 n2:MessageSpi.message),(hedge_synthesis_relation h m1 n1)->(hedge_synthesis_relation h m2 n2)->(hedge_synthesis_relation h (MessageSpi.MPair m1 m2) (MessageSpi.MPair n1 n2))
    | SynOp : forall (m n:MessageSpi.message)(o:OpSpi.op),(hedge_synthesis_relation h m n)->(hedge_synthesis_relation h (MessageSpi.MOp o m) (MessageSpi.MOp o n)).

  (** we now define the "in_synthesis" function *)
  Fixpoint hedge_in_synthesis (h:hedge) (m:MessageSpi.message) (n:MessageSpi.message) {struct m} : bool :=
    if H.mem (m,n) h then true 
      else 
	match m with
	  | (MessageSpi.MPair m1 m2) =>
	    (match n with
	       | (MessageSpi.MPair n1 n2) => andb (hedge_in_synthesis h m1 n1) (hedge_in_synthesis h m2 n2)
	       | _ => false
	     end)
	  | (MessageSpi.MEnc_s m1 m2) =>
	    (match n with
	       | (MessageSpi.MEnc_s n1 n2) => andb (hedge_in_synthesis h m1 n1) (hedge_in_synthesis h m2 n2)
	       | _ => false
	     end)
	  | (MessageSpi.MEnc_a m1 m2) =>
	    (match n with
	       | (MessageSpi.MEnc_a n1 n2) => andb (hedge_in_synthesis h m1 n1) (hedge_in_synthesis h m2 n2)
	       | _ => false
	     end)
	  | (MessageSpi.MOp o m) => 
	    (match n with
	       | (MessageSpi.MOp o' n) =>
		 (if (OpSpi.op_eq_dec o o') then (hedge_in_synthesis h m n)
		   else false)
	       | _ => false
	     end)
	  | _ => false
	end.

  (** specification of the "in_synthesis" function *)
  Definition hedge_synthesis_spec (h:hedge) := hedge_synthesis_relation (relation_of_hedge h).

  (** we prove that "in_synthesis" is sound ... *)
  Lemma hedge_in_synthesis_impl_hedge_synthesis_spec : forall (h:hedge),forall (m n:MessageSpi.message),(hedge_in_synthesis h m n)=true->(hedge_synthesis_spec h m n).
    unfold hedge_synthesis_spec;unfold relation_of_hedge.
    induction m;simpl;intro.
    elim (Bool_elim_bool (H.mem (MessageSpi.MNam t,n) h));intros.
    apply SynInc;apply H.mem_2;trivial.
    rewrite H in H0.
    discriminate H0.
    elim (Bool_elim_bool (H.mem (MessageSpi.MPair m1 m2,n) h));intros.
    apply SynInc;apply H.mem_2;trivial.
    rewrite H in H0.
    destruct n;try (discriminate H0).
    elim (andb_prop (hedge_in_synthesis h m1 n1) (hedge_in_synthesis h m2 n2) H0);intros.
    apply SynPair;auto.
    elim (Bool_elim_bool (H.mem (MessageSpi.MEnc_s m1 m2,n) h));intros.
    apply SynInc;apply H.mem_2;trivial.
    rewrite H in H0;destruct n;try (discriminate H0).
    elim (andb_prop (hedge_in_synthesis h m1 n1) (hedge_in_synthesis h m2 n2) H0);intros.
    apply SynEnc_s;auto.
    elim (Bool_elim_bool (H.mem (MessageSpi.MEnc_a m1 m2,n) h));intros.
    apply SynInc;apply H.mem_2;trivial.
    rewrite H in H0;destruct n;try (discriminate H0).
    elim (andb_prop (hedge_in_synthesis h m1 n1) (hedge_in_synthesis h m2 n2) H0);intros.
    apply SynEnc_a;auto.
    elim (Bool_elim_bool (H.mem (MessageSpi.MOp o m, n) h));intros.
    apply SynInc;auto.
    rewrite H in H0;destruct n;try (discriminate H0).
    generalize H0.
    case (OpSpi.op_eq_dec o o0);intros.
    subst o0.
    elim (OpSpi.op_eq_dec_refl o);intros.
    rewrite H1 in H0.
    apply SynOp;auto.
    discriminate H1.
  Qed.
  
  (** ... and complete *)
  Lemma hedge_synthesis_spec_impl_hedge_in_synthesis : forall (h:hedge),forall (m n:MessageSpi.message),(hedge_synthesis_spec h m n)->(hedge_in_synthesis h m n)=true.
    unfold hedge_synthesis_spec;unfold relation_of_hedge.
    induction 1.
    generalize (H.mem_1 H);clear H;intro.
    induction m;simpl;try (rewrite H;trivial).
    simpl;rewrite IHhedge_synthesis_relation1;rewrite IHhedge_synthesis_relation2;case H.mem;trivial.
    simpl;rewrite IHhedge_synthesis_relation1;rewrite IHhedge_synthesis_relation2;case H.mem;trivial.
    simpl;rewrite IHhedge_synthesis_relation1;rewrite IHhedge_synthesis_relation2;case H.mem;trivial.
    simpl;rewrite IHhedge_synthesis_relation;case H.mem;trivial.
    elim (OpSpi.op_eq_dec_refl o);intros.
    rewrite H0;trivial.
  Qed.

  (** a simple lemma if a couple (a,n) is in S(h) and a is a name then (a,n) is in h *)
  Theorem name_left_synthesis : forall (h:hedge)(a:Name.t)(n:MessageSpi.message),(hedge_in_synthesis h (MessageSpi.MNam a) n)=(H.mem ((MessageSpi.MNam a),n) h).
    intros;simpl.
    elim (Bool_elim_bool (H.mem ((MessageSpi.MNam a),n) h));intro H;rewrite H;trivial.
  Qed.

  (** the same lemma but for the right side *)
  Theorem name_right_synthesis : forall (h:hedge)(m:MessageSpi.message)(b:Name.t),(hedge_in_synthesis h m (MessageSpi.MNam b))=(H.mem (m,(MessageSpi.MNam b)) h).
    intros;destruct m;simpl.
    elim (Bool_elim_bool (H.mem ((MessageSpi.MNam t),(MessageSpi.MNam b)) h));intro H;rewrite H;trivial.
    elim (Bool_elim_bool (H.mem (MessageSpi.MPair m1 m2, MessageSpi.MNam b) h));intro H;rewrite H;trivial.
    elim (Bool_elim_bool (H.mem (MessageSpi.MEnc_s m1 m2, MessageSpi.MNam b) h));intro H;rewrite H;trivial.
    elim (Bool_elim_bool (H.mem (MessageSpi.MEnc_a m1 m2, MessageSpi.MNam b) h));intro H;rewrite H;trivial.
    elim (Bool_elim_bool (H.mem ((MessageSpi.MOp o m),(MessageSpi.MNam b)) h));intro H;rewrite H;trivial.
  Qed.

  (** h is included in S(h) *)
  Lemma h_incl_syn_h : forall (h:hedge)(m n:MessageSpi.message),(H.In (m,n) h)->(hedge_in_synthesis h m n)=true.
    intros.
    apply hedge_synthesis_spec_impl_hedge_in_synthesis.
    unfold hedge_synthesis_spec.
    apply SynInc;trivial.
  Qed.

  (** g <= h if S(g) is subset of S(h) *)
  Definition le_h (g h:message_relation) : Prop := forall (m n:MessageSpi.message),(hedge_synthesis_relation g m n)->(hedge_synthesis_relation h m n).

  (** g ~ h if S(g) <= S(h) and S(h) <= S(g) *)
  Definition equiv_h (g h:message_relation) : Prop := (le_h g h)/\(le_h h g).

  (** <= is an "order" *)
  Lemma le_h_refl : forall (h:message_relation),(le_h h h).
    unfold le_h;auto.
  Qed.

  Lemma le_h_trans : forall (f g h:message_relation),(le_h f g)->(le_h g h)->(le_h f h).
    unfold le_h.
    auto.
  Qed.

  Lemma le_h_antisym : forall (g h:message_relation),(le_h g h)->(le_h h g)->(equiv_h g h).
    unfold equiv_h;tauto.
  Qed.

  (** ~ is an equivalence relation *)
  Lemma equiv_h_refl : forall (h:message_relation),(equiv_h h h).
    unfold equiv_h.
    split;apply le_h_refl.
  Qed.

  Lemma equiv_h_trans : forall (f g h:message_relation),(equiv_h f g)->(equiv_h g h)->(equiv_h f h).
    unfold equiv_h.
    split;and_elim;eapply le_h_trans;eauto.
  Qed.

  Lemma equiv_h_sym : forall (g h:message_relation),(equiv_h g h)->(equiv_h h g).
    unfold equiv_h;tauto.
  Qed.

  (** if g <= h then g is included in S(h) *)
  Lemma le_h_impl_incl_syn : forall (g h:message_relation),(le_h g h)->(forall (m n:MessageSpi.message),(g m n)->(hedge_synthesis_relation h m n)).
    unfold le_h.
    intros.
    apply H.
    apply SynInc.
    trivial.
  Qed.

  (** if g is included in S(h) then g <= h *)
  Lemma incl_syn_impl_le_h : forall (g h:message_relation),(forall (m n:MessageSpi.message),(g m n)->(hedge_synthesis_relation h m n))->(le_h g h).
    unfold le_h;intros.
    induction H0.
    apply H;trivial.
    apply SynEnc_s;trivial.
    apply SynEnc_a;trivial.
    apply SynPair;trivial.
    apply SynOp;trivial.
  Qed.

  (** g is included in S(h) iff g <= h *)
  Theorem le_h_iff_incl_syn : forall (g h:message_relation),(le_h g h)<->(forall (m n:MessageSpi.message),(g m n)->(hedge_synthesis_relation h m n)).
    split.
    apply le_h_impl_incl_syn.
    apply incl_syn_impl_le_h.
  Qed.

  (** if g is included in h then g <= h *)
  Theorem incl_impl_le_h : forall (g h:message_relation),(inclusion MessageSpi.message g h)->(le_h g h).
    unfold H.Subset.
    intros.
    apply incl_syn_impl_le_h.
    intros.
    apply SynInc.
    apply H;trivial.
  Qed.

  (** if g = h then g ~ h *)
  Lemma same_impl_equiv_h : forall (g h:message_relation),(same_relation MessageSpi.message g h)->(equiv_h g h).
    unfold same_relation.
    unfold equiv_h.
    split;and_elim;apply incl_impl_le_h;trivial.
  Qed.

  (** if f <= h and g <= h then f U g <= h *)
  Theorem le_h_le_h_impl_union_le_h : forall (f g h:message_relation),(le_h f h)->(le_h g h)->(le_h (union MessageSpi.message f g) h).
    intros.
    apply incl_syn_impl_le_h.
    intros.
    case H1;intro;[apply H | apply H0];apply SynInc;trivial.
  Qed.

  (** if g1 <= h1 and g2 <= h2 then g1 U g2 <= h1 U h2 *)
  Theorem le_h_le_h_impl_le_h_union : forall (g1 g2 h1 h2:message_relation),(le_h g1 h1)->(le_h g2 h2)->(le_h (union MessageSpi.message g1 g2) (union MessageSpi.message h1 h2)).
    intros.
    apply le_h_le_h_impl_union_le_h;eapply le_h_trans;[apply H | apply incl_impl_le_h | apply H0 | apply incl_impl_le_h];unfold inclusion;unfold union;tauto.
  Qed.

  (** we show that <= is decidable (using the previous results) *)
  Definition is_le_h (g h:hedge) := (H.for_all (fun mn => let (m,n) := mn in hedge_in_synthesis h m n) g).

  (** here it is! *)
  Lemma is_le_h_iff_le_h : forall (g h:hedge),(is_le_h g h = true)<->(le_h (relation_of_hedge g) (relation_of_hedge h)).
    intros.
    unfold is_le_h.
    assert (FSetInterface.compat_bool H.E.eq (fun mn => let (m,n) := mn in hedge_in_synthesis h m n)).
    unfold FSetInterface.compat_bool.
    intros.
    destruct x;destruct y;simpl in H;and_elim.
    rewrite <- H;rewrite <- H0;trivial.
    split;intros.
    generalize (H.for_all_2 H H0).
    unfold H.For_all.
    intros.
    apply incl_syn_impl_le_h.
    intros.
    apply hedge_in_synthesis_impl_hedge_synthesis_spec.
    apply (H1 (m,n));trivial.
    apply H.for_all_1;trivial.
    unfold H.For_all;intros.
    destruct x.
    apply hedge_synthesis_spec_impl_hedge_in_synthesis.
    unfold hedge_synthesis_spec.
    eapply le_h_impl_incl_syn.
    apply H0.
    trivial.
  Qed.

  (** analysis specification *)
  Definition inverse_keys (k l:MessageSpi.message) := 
    match (MessageSpi.inverse_key k,MessageSpi.inverse_key l) with
      | (Some k',Some l') => Some (k',l')
      | _ => None
    end.

  Inductive hedge_analysis_relation (h:message_relation) : message_relation :=
    | AnaInc : forall (m n:MessageSpi.message),(h m n)->(hedge_analysis_relation h m n)
    | AnaSplitL : forall (m1 m2 n1 n2:MessageSpi.message),(hedge_analysis_relation h (MessageSpi.MPair m1 m2) (MessageSpi.MPair n1 n2))->(hedge_analysis_relation h m1 n1)
    | AnaSplitR : forall (m1 m2 n1 n2:MessageSpi.message),(hedge_analysis_relation h (MessageSpi.MPair m1 m2) (MessageSpi.MPair n1 n2))->(hedge_analysis_relation h m2 n2)
    | AnaDec_s : forall (m k n l:MessageSpi.message),(hedge_analysis_relation h (MessageSpi.MEnc_s m k) (MessageSpi.MEnc_s n l))->(hedge_synthesis_relation h k l)->(hedge_analysis_relation h m n)
    | AnaDec_a : forall (m k n l:MessageSpi.message),(hedge_analysis_relation h (MessageSpi.MEnc_a m k) (MessageSpi.MEnc_a n l))->forall (k' l':MessageSpi.message),(inverse_keys k l = Some (k',l'))->(hedge_synthesis_relation h k' l')->(hedge_analysis_relation h m n).

  Definition hedge_analysis_step_spec (h:hedge) := hedge_analysis_relation (relation_of_hedge h).

  Fixpoint analz_elt (h:hedge) (m:MessageSpi.message)(n:MessageSpi.message) {struct m} : H.t :=
    match (m,n) with
      | ((MessageSpi.MPair m1 m2),(MessageSpi.MPair n1 n2)) => H.add (m1,n1) (H.add (m2,n2) (H.union (analz_elt h m1 n1) (analz_elt h m2 n2)))
      | ((MessageSpi.MEnc_s m1 m2),(MessageSpi.MEnc_s n1 n2)) =>
	if hedge_in_synthesis h m2 n2 
	  then H.add (m1,n1) (analz_elt h m1 n1)
	  else H.empty
      | ((MessageSpi.MEnc_a m1 m2),(MessageSpi.MEnc_a n1 n2)) =>
	match (inverse_keys m2 n2) with
	  | Some (m'2,n'2) => 
	    if hedge_in_synthesis h m'2 n'2 
	      then H.add (m1,n1) (analz_elt h m1 n1)
	      else H.empty
	  | None => H.empty
	end
      | _ => H.empty
    end.

  Definition analz_elt_aux h (mn:H.elt) := let (m,n) := mn in analz_elt h m n.

  Definition analz_aux (h h':hedge) := (HedgeOtherProp.big_union (analz_elt_aux h) h').

  Definition analz (h:hedge) := H.union h (analz_aux h h).

  Lemma analz_elt_aux_compat : forall (h:H.t)(x y:H.elt),(H.E.eq x y)->(H.Equal ((analz_elt_aux h) x) ((analz_elt_aux h) y)).
    intros.
    destruct x;destruct y;simpl in H;and_elim.
    rewrite H.
    rewrite H0.
    auto.
    apply HedgeProp.equal_refl.
  Qed.

  Lemma analz_elt_subset_analysis : forall (h:hedge)(m n:MessageSpi.message),(hedge_analysis_step_spec h m n)->(forall (m' n':MessageSpi.message),(H.In (m',n') (analz_elt h m n))->(hedge_analysis_step_spec h m' n')).
    induction m;intros.
    simpl in H0.
    elim (H.empty_1 H0).
    destruct n;simpl in H0;try (elim (H.empty_1 H0);fail).
    elim (HedgeProp.Add_add (H.add (m2, n2) (H.union (analz_elt h m1 n1) (analz_elt h m2 n2))) (m1,n1) (m',n'));intros.
    case (H1 H0);clear H0;clear H1;clear H2;intro.
    simpl in H0;and_elim;rewrite <- H0;rewrite <- H1.
    unfold hedge_analysis_step_spec.
    eapply AnaSplitL.
    apply H.
    elim (HedgeProp.Add_add (H.union (analz_elt h m1 n1) (analz_elt h m2 n2)) (m2,n2) (m',n'));intros.
    case (H1 H0);clear H0;clear H1;clear H2;intro.
    simpl in H0;and_elim;rewrite <- H0;rewrite <- H1.
    unfold hedge_analysis_step_spec.
    eapply AnaSplitR.
    apply H.
    case (H.union_1 H0);clear H0;intro.
    eapply IHm1;eauto.
    unfold hedge_analysis_step_spec.
    eapply AnaSplitL.
    apply H.
    eapply IHm2;eauto.
    unfold hedge_analysis_step_spec.
    eapply AnaSplitR.
    apply H.
    destruct n;simpl in H0;try (elim (H.empty_1 H0);fail).
    case (Bool_elim_bool (hedge_in_synthesis h m2 n2));intro H1;rewrite H1 in H0;try (elim (H.empty_1 H0);fail).
    elim (HedgeProp.Add_add (analz_elt h m1 n1) (m1,n1) (m',n'));intros.
    case (H2 H0);clear H0;clear H2;clear H3;intro.
    simpl in H0;and_elim;rewrite <- H0;rewrite <- H2.
    unfold hedge_analysis_step_spec.
    eapply AnaDec_s.
    apply H.
    apply hedge_in_synthesis_impl_hedge_synthesis_spec.
    apply H1.
    eapply IHm1;eauto.
    unfold hedge_analysis_step_spec.
    eapply AnaDec_s.
    apply H.
    apply hedge_in_synthesis_impl_hedge_synthesis_spec.
    apply H1.
    destruct n;simpl in H0;try (elim (H.empty_1 H0);fail).
    case (Option_elim_option _ (inverse_keys m2 n2));intro H1;try (rewrite H1 in H0;try (elim (H.empty_1 H0);fail)).
    ex_elim H1 y.
    destruct y as [m'2 n'2].
    rewrite H1 in H0.
    case (Bool_elim_bool (hedge_in_synthesis h m'2 n'2));intro H2;rewrite H2 in H0;try (elim (H.empty_1 H0);fail).
    elim (HedgeProp.Add_add (analz_elt h m1 n1) (m1,n1) (m',n'));intros.
    case (H3 H0);clear H0;clear H3;clear H4;intro.
    simpl in H0;and_elim;rewrite <- H0;rewrite <- H3.
    unfold hedge_analysis_step_spec.
    eapply AnaDec_a.
    apply H.
    apply H1.
    apply hedge_in_synthesis_impl_hedge_synthesis_spec.
    apply H2.
    eapply IHm1;eauto.
    unfold hedge_analysis_step_spec.
    eapply AnaDec_a.
    apply H.
    apply H1.
    apply hedge_in_synthesis_impl_hedge_synthesis_spec.
    apply H2.
    simpl in H0;elim (H.empty_1 H0).
  Qed.

  Lemma analz_subset_analysis : forall (h:H.t)(m n:MessageSpi.message),(H.In (m,n) (analz h))->(hedge_analysis_step_spec h m n).
    unfold analz.
    intros.
    case (H.union_1 H);intro.
    unfold hedge_analysis_step_spec.
    apply AnaInc.
    auto.
    elim (HedgeOtherProp.In_big_union (analz_elt_aux h) (analz_elt_aux_compat h) h (m,n)).
    intros.
    elim (H1 H0);intro x;clear H0;clear H1;clear H2.
    intros;and_elim.
    destruct x.
    eapply analz_elt_subset_analysis.
    unfold hedge_analysis_step_spec.
    apply AnaInc.
    unfold relation_of_hedge.
    apply H0.
    simpl in H1.
    auto.
  Qed.

  Lemma analz_AnaFst : forall (h:hedge)(m n:MessageSpi.message)(m1 n1 m2 n2:MessageSpi.message),(H.In (MessageSpi.MPair m1 m2,MessageSpi.MPair n1 n2) (analz_elt h m n))->(H.In (m1,n1) (analz_elt h m n)).
    induction m;destruct n;intros;simpl in H;try (elim (H.empty_1 H);fail).
    elim (HedgeProp.Add_add (H.add (m2, n2) (H.union (analz_elt h m1 n1) (analz_elt h m2 n2))) (m1,n1) (MessageSpi.MPair m0 m3,MessageSpi.MPair n0 n3));intros.
    case (H0 H);clear H;clear H0;clear H1;intro.
    simpl in H;and_elim.
    rewrite H;rewrite H0.
    simpl.
    apply H.add_2.
    apply H.add_2.
    apply H.union_2.
    apply H.add_1;auto.
    elim (HedgeProp.Add_add (H.union (analz_elt h m1 n1) (analz_elt h m2 n2)) (m2,n2) (MessageSpi.MPair m0 m3,MessageSpi.MPair n0 n3));intros.
    case (H0 H);clear H;clear H0;clear H1;intro.
    simpl in H;and_elim.
    rewrite H;rewrite H0.
    simpl.
    apply H.add_2.
    apply H.add_2.
    apply H.union_3.
    apply H.add_1;auto.
    case (H.union_1 H);clear H;intro.
    simpl.
    apply H.add_2.
    apply H.add_2.
    apply H.union_2.
    eapply IHm1;eauto.
    simpl.
    apply H.add_2.
    apply H.add_2.
    apply H.union_3.
    eapply IHm2;eauto.
    case (Bool_elim_bool (hedge_in_synthesis h m2 n2));intro H1;rewrite H1 in H;try (elim (H.empty_1 H);fail).
    elim (HedgeProp.Add_add (analz_elt h m1 n1) (m1,n1) (MessageSpi.MPair m0 m3,MessageSpi.MPair n0 n3));intros.
    case (H0 H);clear H;clear H0;clear H2;intro.
    simpl in H;and_elim.
    rewrite H;rewrite H0.
    simpl.
    rewrite H1.
    apply H.add_2.
    apply H.add_1;auto.
    simpl.
    rewrite H1.
    apply H.add_2.
    eauto.
    case (Option_elim_option _ (inverse_keys m2 n2));intro H1;try (rewrite H1 in H;try (elim (H.empty_1 H);fail)).
    ex_elim H1 y.
    destruct y as [m'2 n'2].
    rewrite H1 in H.
    case (Bool_elim_bool (hedge_in_synthesis h m'2 n'2));intro H2;rewrite H2 in H;try (elim (H.empty_1 H);fail).
    elim (HedgeProp.Add_add (analz_elt h m1 n1) (m1,n1) (MessageSpi.MPair m0 m3,MessageSpi.MPair n0 n3));intros.
    case (H0 H);clear H;clear H0;clear H3;intro.
    simpl in H;and_elim.
    rewrite H;rewrite H0.
    simpl.
    rewrite H1.
    rewrite H2.
    apply H.add_2.
    apply H.add_1;auto.
    simpl.
    rewrite H1.
    rewrite H2.
    apply H.add_2.
    eauto.
  Qed.

  Lemma analz_AnaSnd : forall (h:hedge)(m n:MessageSpi.message)(m1 n1 m2 n2:MessageSpi.message),(H.In (MessageSpi.MPair m1 m2,MessageSpi.MPair n1 n2) (analz_elt h m n))->(H.In (m2,n2) (analz_elt h m n)).
    induction m;destruct n;intros;simpl in H;try (elim (H.empty_1 H);fail).
    elim (HedgeProp.Add_add (H.add (m2, n2) (H.union (analz_elt h m1 n1) (analz_elt h m2 n2))) (m1,n1) (MessageSpi.MPair m0 m3, MessageSpi.MPair n0 n3));intros.
    case (H0 H);clear H;clear H0;clear H1;intro.
    simpl in H;and_elim.
    rewrite H;rewrite H0.
    simpl.
    apply H.add_2.
    apply H.add_2.
    apply H.union_2.
    apply H.add_2.
    apply H.add_1;auto.
    elim (HedgeProp.Add_add (H.union (analz_elt h m1 n1) (analz_elt h m2 n2)) (m2,n2) (MessageSpi.MPair m0 m3, MessageSpi.MPair n0 n3));intros.
    case (H0 H);clear H;clear H0;clear H1;intro.
    simpl in H;and_elim.
    rewrite H;rewrite H0.
    simpl.
    apply H.add_2.
    apply H.add_2.
    apply H.union_3.
    apply H.add_2.
    apply H.add_1;auto.
    case (H.union_1 H);clear H;intro.
    simpl.
    apply H.add_2.
    apply H.add_2.
    apply H.union_2.
    eauto.
    simpl.
    apply H.add_2.
    apply H.add_2.
    apply H.union_3.
    eauto.
    case (Bool_elim_bool (hedge_in_synthesis h m2 n2));intro H1;rewrite H1 in H;try (elim (H.empty_1 H);fail).
    elim (HedgeProp.Add_add (analz_elt h m1 n1) (m1,n1) (MessageSpi.MPair m0 m3, MessageSpi.MPair n0 n3));intros.
    case (H0 H);clear H;clear H0;clear H2;intro.
    simpl in H;and_elim.
    rewrite H;rewrite H0.
    simpl.
    rewrite H1.
    apply H.add_2.
    apply H.add_2.
    apply H.add_1;auto.
    simpl.
    rewrite H1.
    apply H.add_2.
    eauto.
    case (Option_elim_option _ (inverse_keys m2 n2));intro H1;try (rewrite H1 in H;try (elim (H.empty_1 H);fail)).
    ex_elim H1 y.
    destruct y as [m'2 n'2].
    rewrite H1 in H.
    case (Bool_elim_bool (hedge_in_synthesis h m'2 n'2));intro H2;rewrite H2 in H;try (elim (H.empty_1 H);fail).
    elim (HedgeProp.Add_add (analz_elt h m1 n1) (m1,n1) (MessageSpi.MPair m0 m3, MessageSpi.MPair n0 n3));intros.
    case (H0 H);clear H;clear H0;clear H3;intro.
    simpl in H;and_elim.
    rewrite H;rewrite H0.
    simpl.
    rewrite H1.
    rewrite H2.
    apply H.add_2.
    apply H.add_2.
    apply H.add_1;auto.
    simpl.
    rewrite H1.
    rewrite H2.
    apply H.add_2.
    eauto.
  Qed.

  Lemma analz_AnaDec_s : forall (h:hedge)(m n:MessageSpi.message)(m1 n1 m2 n2:MessageSpi.message),(H.In (MessageSpi.MEnc_s m1 m2, MessageSpi.MEnc_s n1 n2) (analz_elt h m n))->(hedge_in_synthesis h m2 n2)=true->(H.In (m1,n1) (analz_elt h m n)).
    induction m;destruct n;intros;simpl in H;try (elim (H.empty_1 H);fail).
    elim (HedgeProp.Add_add (H.add (m2, n2) (H.union (analz_elt h m1 n1) (analz_elt h m2 n2))) (m1,n1) (MessageSpi.MEnc_s m0 m3, MessageSpi.MEnc_s n0 n3));intros.
    case (H1 H);clear H;clear H1;clear H2;intro.
    simpl in H;and_elim.
    rewrite H;rewrite H1.
    simpl.
    apply H.add_2.
    apply H.add_2.
    apply H.union_2.
    rewrite H0.
    apply H.add_1;auto.
    elim (HedgeProp.Add_add (H.union (analz_elt h m1 n1) (analz_elt h m2 n2)) (m2,n2) (MessageSpi.MEnc_s m0 m3, MessageSpi.MEnc_s n0 n3));intros.
    case (H1 H);clear H;clear H1;clear H2;intro.
    simpl in H;and_elim.
    rewrite H;rewrite H1.
    simpl.
    apply H.add_2.
    apply H.add_2.
    apply H.union_3.
    rewrite H0.
    apply H.add_1;auto.
    case (H.union_1 H);clear H;intro.
    simpl.
    apply H.add_2.
    apply H.add_2.
    apply H.union_2.
    eauto.
    simpl.
    apply H.add_2.
    apply H.add_2.
    apply H.union_3.
    eauto.
    case (Bool_elim_bool (hedge_in_synthesis h m2 n2));intro H1;rewrite H1 in H;try (elim (H.empty_1 H);fail).
    elim (HedgeProp.Add_add (analz_elt h m1 n1) (m1,n1) (MessageSpi.MEnc_s m0 m3, MessageSpi.MEnc_s n0 n3));intros.
    case (H2 H);clear H;clear H2;clear H3;intro.
    simpl in H;and_elim.
    rewrite H;rewrite H2.
    simpl.
    rewrite H1.
    apply H.add_2.
    rewrite H0.
    apply H.add_1;auto.
    simpl.
    rewrite H1.
    apply H.add_2.
    eauto.
    case (Option_elim_option _ (inverse_keys m2 n2));intro H1;try (rewrite H1 in H;try (elim (H.empty_1 H);fail)).
    ex_elim H1 y.
    destruct y as [m'2 n'2].
    rewrite H1 in H.
    case (Bool_elim_bool (hedge_in_synthesis h m'2 n'2));intro H2;rewrite H2 in H;try (elim (H.empty_1 H);fail).
    elim (HedgeProp.Add_add (analz_elt h m1 n1) (m1,n1) (MessageSpi.MEnc_s m0 m3, MessageSpi.MEnc_s n0 n3));intros.
    case (H3 H);clear H;clear H3;clear H4;intro.
    simpl in H;and_elim.
    rewrite H;rewrite H3.
    simpl.
    rewrite H1.
    rewrite H2.
    apply H.add_2.
    rewrite H0.
    apply H.add_1;auto.
    simpl.
    rewrite H1.
    rewrite H2.
    apply H.add_2.
    eauto.
  Qed.

  Lemma analz_AnaDec_a : forall (h:hedge)(m n:MessageSpi.message)(m1 n1 m2 n2:MessageSpi.message),(H.In (MessageSpi.MEnc_a m1 m2, MessageSpi.MEnc_a n1 n2) (analz_elt h m n))->forall (m'2 n'2:MessageSpi.message),(inverse_keys m2 n2 = Some (m'2,n'2))->(hedge_in_synthesis h m'2 n'2)=true->(H.In (m1,n1) (analz_elt h m n)).
    induction m;destruct n;intros;simpl in H;try (elim (H.empty_1 H);fail).
    elim (HedgeProp.Add_add (H.add (m2, n2) (H.union (analz_elt h m1 n1) (analz_elt h m2 n2))) (m1,n1) (MessageSpi.MEnc_a m0 m3, MessageSpi.MEnc_a n0 n3));intros.
    case (H2 H);clear H;clear H2;clear H3;intro.
    simpl in H;and_elim.
    rewrite H;rewrite H2.
    simpl.
    rewrite H0.
    rewrite H1.
    apply H.add_2.
    apply H.add_2.
    apply H.union_2.
    auto.
    elim (HedgeProp.Add_add (H.union (analz_elt h m1 n1) (analz_elt h m2 n2)) (m2,n2) (MessageSpi.MEnc_a m0 m3, MessageSpi.MEnc_a n0 n3));intros.
    case (H2 H);clear H;clear H2;clear H3;intro.
    simpl in H;and_elim.
    rewrite H;rewrite H2.
    simpl.
    rewrite H0.
    rewrite H1.
    apply H.add_2.
    apply H.add_2.
    apply H.union_3.
    auto.
    case (H.union_1 H);clear H;intro.
    simpl.
    apply H.add_2.
    apply H.add_2.
    apply H.union_2.
    eauto.
    simpl.
    apply H.add_2.
    apply H.add_2.
    apply H.union_3.
    eauto.
    case (Bool_elim_bool (hedge_in_synthesis h m2 n2));intro H2;rewrite H2 in H;try (elim (H.empty_1 H);fail).
    elim (HedgeProp.Add_add (analz_elt h m1 n1) (m1,n1) (MessageSpi.MEnc_a m0 m3, MessageSpi.MEnc_a n0 n3));intros.
    case (H3 H);clear H;clear H3;clear H4;intro.
    simpl in H;and_elim.
    rewrite H;rewrite H3.
    simpl.
    rewrite H0.
    rewrite H1.
    rewrite H2.
    apply H.add_2.
    auto.
    simpl.
    rewrite H2.
    apply H.add_2.
    eauto.
    case (Option_elim_option _ (inverse_keys m2 n2));intro H2;try (rewrite H2 in H;try (elim (H.empty_1 H);fail)).
    ex_elim H2 y.
    destruct y as [m' n'].
    rewrite H2 in H.
    case (Bool_elim_bool (hedge_in_synthesis h m' n'));intro H3;rewrite H3 in H;try (elim (H.empty_1 H);fail).
    elim (HedgeProp.Add_add (analz_elt h m1 n1) (m1,n1) (MessageSpi.MEnc_a m0 m3, MessageSpi.MEnc_a n0 n3));intros.
    case (H4 H);clear H;clear H4;clear H5;intro.
    simpl in H;and_elim.
    rewrite H;rewrite H4.
    simpl.
    rewrite H2.
    rewrite H3.
    rewrite H0.
    rewrite H1.
    apply H.add_2.
    auto.
    simpl.
    rewrite H2.
    rewrite H3.
    apply H.add_2.
    eauto.
  Qed.

  Lemma analysis_subset_analz : forall (h:H.t)(m n:MessageSpi.message),(hedge_analysis_step_spec h m n)->(H.In (m,n) (analz h)).
    unfold analz.
    intro h.
    generalize (HedgeOtherProp.In_big_union (analz_elt_aux h) (analz_elt_aux_compat h) h);intro.
    induction 1;intros.
    apply H.union_2;auto.
    case (H.union_1 IHhedge_analysis_relation);intro.
    apply H.union_3.
    elim (H (m1,n1));intros.
    apply H3.
    exists (MessageSpi.MPair m1 m2, MessageSpi.MPair n1 n2).
    split;simpl;auto.
    elim (H (MessageSpi.MPair m1 m2, MessageSpi.MPair n1 n2));intros.
    elim (H2 H1);clear H1;clear H2;clear H3.
    intro x;intros;and_elim.
    destruct x.
    apply H.union_3.
    elim (H (m1,n1));intros.
    apply H4.
    exists (t,t0).
    split;trivial.
    simpl.
    eapply analz_AnaFst.
    apply H2.
    case (H.union_1 IHhedge_analysis_relation);intro.
    apply H.union_3.
    elim (H (m2,n2));intros.
    apply H3.
    exists (MessageSpi.MPair m1 m2, MessageSpi.MPair n1 n2).
    split;simpl;auto.
    elim (H (MessageSpi.MPair m1 m2, MessageSpi.MPair n1 n2));intros.
    elim (H2 H1);clear H1;clear H2;clear H3.
    intro x;intros;and_elim.
    destruct x.
    apply H.union_3.
    elim (H (m2,n2));intros.
    apply H4.
    exists (t,t0).
    split;trivial.
    simpl.
    eapply analz_AnaSnd.
    apply H2.
    case (H.union_1 IHhedge_analysis_relation);intro.
    apply H.union_3.
    elim (H (m,n));intros.
    apply H4.
    exists (MessageSpi.MEnc_s m k, MessageSpi.MEnc_s n l).
    split;simpl;auto.
    rewrite hedge_synthesis_spec_impl_hedge_in_synthesis;auto.
    elim (H (MessageSpi.MEnc_s m k, MessageSpi.MEnc_s n l));intros.
    elim (H3 H2);clear H2;clear H3;clear H4.
    intro x;intros;and_elim.
    destruct x.
    apply H.union_3.
    elim (H (m,n));intros.
    apply H5.
    exists (t,t0).
    split;trivial.
    simpl.
    eapply analz_AnaDec_s.
    apply H3.
    apply hedge_synthesis_spec_impl_hedge_in_synthesis.
    auto.
    case (H.union_1 IHhedge_analysis_relation);intro.
    apply H.union_3.
    elim (H (m,n));intros.
    apply H5.
    exists (MessageSpi.MEnc_a m k, MessageSpi.MEnc_a n l).
    split;simpl;auto.
    rewrite H1.
    rewrite hedge_synthesis_spec_impl_hedge_in_synthesis;auto.
    elim (H (MessageSpi.MEnc_a m k, MessageSpi.MEnc_a n l));intros.
    elim (H4 H3);clear H3;clear H4;clear H5.
    intro x;intros;and_elim.
    destruct x.
    apply H.union_3.
    elim (H (m,n));intros.
    apply H6.
    exists (t,t0).
    split;trivial.
    simpl.
    eapply analz_AnaDec_a.
    apply H4.
    apply H1.
    apply hedge_synthesis_spec_impl_hedge_in_synthesis.
    auto.
  Qed.

  Definition hedge_analysis_step (h:hedge) := analz h.

  (** we show hedge_analysis_step is sound ... *)
  Lemma hedge_analysis_step_impl_hedge_analysis_step_spec : forall (h:hedge)(m n:MessageSpi.message),(H.In (m,n) (hedge_analysis_step h))->(hedge_analysis_step_spec h m n).
    intros.
    apply analz_subset_analysis.
    apply H.
  Qed.

  (** ... and complete *)
  Lemma hedge_analysis_step_spec_impl_hedge_analysis_step : forall (h:hedge)(m n:MessageSpi.message),(hedge_analysis_step_spec h m n)->(H.In (m,n) (hedge_analysis_step h)).
    intros.
    apply analysis_subset_analz.
    apply H.
  Qed.

  (** we now show that it is sufficient to complete hedge_analysis_fold on a subset of h to get the result of hedge_analysis_step h *)
  (** (m,n) is analysable by h if (m,n) are both pairs or both encrypted messages whose inverse keys are synthesisable by h *)
  Definition is_analysable (h:hedge) (m n:MessageSpi.message) := 
    match m with
      | (MessageSpi.MPair m1 m2) => 
	(match n with
	   | (MessageSpi.MPair n1 n2) => true
	   | _ => false
	 end)
      | (MessageSpi.MEnc_s m1 m2) =>
	(match n with
	   | (MessageSpi.MEnc_s n1 n2) => hedge_in_synthesis h m2 n2
	   | _ => false
	 end)
      | (MessageSpi.MEnc_a m1 m2) =>
	(match n with
	   | (MessageSpi.MEnc_a n1 n2) => 
	     (match (inverse_keys m2 n2) with
		| Some (m'2,n'2) => hedge_in_synthesis h m'2 n'2
		| None => false
	      end)
	   | _ => false
	 end)
      | _ => false
    end.

  Lemma is_analysable_compat_bool :  forall (h:hedge),(FSetInterface.compat_bool H.E.eq (fun mn => let (m,n) := mn in is_analysable h m n)).
    unfold FSetInterface.compat_bool.
    destruct x;destruct y;simpl;intros;and_elim.
    rewrite H;rewrite H0;trivial.
  Qed.

  Lemma is_analysable_Subset_true : forall (h h':hedge)(m n:MessageSpi.message),(H.Subset h h')->(is_analysable h m n)=true->(is_analysable h' m n)=true.
    destruct m;destruct n;simpl;auto;intros.
    apply hedge_synthesis_spec_impl_hedge_in_synthesis.
    unfold hedge_synthesis_spec.
    apply (incl_impl_le_h (relation_of_hedge h) (relation_of_hedge h') (Subset_impl_inclusion h h' H)).
    apply hedge_in_synthesis_impl_hedge_synthesis_spec.
    trivial.
    destruct (inverse_keys m2 n2);try (discriminate H0).
    destruct p.
    apply hedge_synthesis_spec_impl_hedge_in_synthesis.
    unfold hedge_synthesis_spec.
    apply (incl_impl_le_h (relation_of_hedge h) (relation_of_hedge h') (Subset_impl_inclusion h h' H)).
    apply hedge_in_synthesis_impl_hedge_synthesis_spec.
    trivial.
  Qed.

  Lemma is_analysable_Subset_false : forall (h h':hedge)(m n:MessageSpi.message),(H.Subset h' h)->(is_analysable h m n)=false->(is_analysable h' m n)=false.
    intros.
    case (Bool_elim_bool (is_analysable h' m n));intro;trivial.
    rewrite (is_analysable_Subset_true h' h m n) in H0;trivial.
    discriminate H0.
  Qed.

  Lemma analz_elt_empty_is_analysable : forall (h:hedge)(m n:MessageSpi.message),(H.Equal (analz_elt h m n) H.empty)<->(is_analysable h m n = false).
    destruct m;simpl;intros;split;auto;destruct n;simpl;auto;intros.
    elim (H (m1,n1));intros.
    assert (H.In (m1,n1) H.empty).
    apply H0;auto.
    elim (H.empty_1 H2).
    discriminate H.
    destruct (hedge_in_synthesis h m2 n2);auto.
    elim (H (m1,n1));intros.
    assert (H.In (m1,n1) H.empty).
    apply H0;auto.
    elim (H.empty_1 H2).
    rewrite H.
    auto.
    destruct (inverse_keys m2 n2);auto.
    destruct p.
    destruct (hedge_in_synthesis h m m0);auto.
    elim (H (m1,n1));intros.
    assert (H.In (m1,n1) H.empty).
    apply H0;auto.
    elim (H.empty_1 H2).
    destruct (inverse_keys m2 n2);auto.
    destruct p.
    rewrite H.
    auto.
  Qed.

  Lemma analz_aux_is_analysable : forall (h h':hedge),(H.Equal (analz_aux h h') (analz_aux h (H.filter (fun mn => let (m,n) := mn in is_analysable h m n) h'))).
    split;intros.
    unfold analz_aux.
    elim (HedgeOtherProp.In_big_union (analz_elt_aux h) (analz_elt_aux_compat h) h' a);intros.
    elim (H0 H).
    intro x;intros;and_elim.
    destruct x as [m n].
    simpl in H3.
    elim (HedgeOtherProp.In_big_union (analz_elt_aux h) (analz_elt_aux_compat h) (H.filter (fun mn => let (m,n) := mn in is_analysable h m n) h') a);intros.
    apply H5.
    exists (m,n).
    simpl.
    split;trivial.
    apply H.filter_3.
    apply is_analysable_compat_bool.
    auto.
    elim (analz_elt_empty_is_analysable h m n);intros.
    destruct (is_analysable h m n);auto.
    elim (H7 (refl_equal false) a);intros.
    elim (H.empty_1 (H8 H3)).
    elim (HedgeOtherProp.In_big_union (analz_elt_aux h) (analz_elt_aux_compat h) (H.filter (fun mn => let (m,n) := mn in is_analysable h m n) h') a);intros.
    elim (H0 H);clear H;clear H0;clear H1;intros;and_elim.
    elim (HedgeOtherProp.In_big_union (analz_elt_aux h) (analz_elt_aux_compat h) h' a);intros.
    apply H2.
    exists x.
    split;auto.
    apply (H.filter_1 (is_analysable_compat_bool h) H).
  Qed.
    
  (** a pair (m,n) is already analysed if (m,n) are both pairs whose both projections are in h *)
  (** of if they are encrypted messages whose both inverse keys are synthesisable by h and contents are in h *)
  Definition is_analysed (h:hedge) (m n:MessageSpi.message) : bool :=
    match m with
      | (MessageSpi.MPair m1 m2) =>
	(match n with
	   | (MessageSpi.MPair n1 n2) => andb (H.mem (m1,n1) h) (H.mem (m2,n2) h)
	   | _ => true
	 end)
      | (MessageSpi.MEnc_s m1 m2) =>
	(match n with
	   | (MessageSpi.MEnc_s n1 n2) => andb (hedge_in_synthesis h m2 n2) (H.mem (m1,n1) h)
	   | _ => true
	 end)
      | (MessageSpi.MEnc_a m1 m2) =>
	(match n with
	   | (MessageSpi.MEnc_a n1 n2) => 
	     (match (inverse_keys m2 n2) with
		| Some (m'2,n'2) => andb (hedge_in_synthesis h m'2 n'2) (H.mem (m1,n1) h)
		| _ => true
	      end)
	   | _ => true
	 end)
      | _ => true
    end.

  Lemma is_analysed_compat_bool : (forall (h:hedge), FSetInterface.compat_bool H.E.eq (fun mn : H.elt => let (m, n) := mn in negb (is_analysed h m n))).
    unfold FSetInterface.compat_bool.
    destruct x;destruct y;simpl;intros;and_elim.
    rewrite H;rewrite H0;trivial.
  Qed.

  Lemma is_analysed_Subset_true : forall (h h':hedge)(m n:MessageSpi.message),(H.Subset h h')->(is_analysed h m n)=true->(is_analysed h' m n)=true.
    destruct m;destruct n;simpl;auto;intros.
    assert (H.In (m1,n1) h).
    apply H.mem_2.
    destruct (H.mem (m1,n1) h);try (discriminate H0);auto.
    assert (H.In (m2,n2) h).
    apply H.mem_2.
    destruct (H.mem (m1,n1) h);try (discriminate H0).
    destruct (H.mem (m2,n2) h);try (discriminate H0);auto.
    repeat (rewrite H.mem_1);try reflexivity.
    apply H;auto.
    apply H;auto.
    assert ((hedge_in_synthesis h m2 n2)=true).
    destruct (hedge_in_synthesis h m2 n2);try (discriminate H0);auto.
    assert (H.In (m1,n1) h).
    apply H.mem_2.
    destruct (hedge_in_synthesis h m2 n2);try (discriminate H0).
    destruct (H.mem (m1,n1) h);try (discriminate H0);auto.
    rewrite H.mem_1.
    rewrite andb_b_true.
    apply hedge_synthesis_spec_impl_hedge_in_synthesis.
    unfold hedge_synthesis_spec.
    apply (incl_impl_le_h (relation_of_hedge h) (relation_of_hedge h') (Subset_impl_inclusion h h' H)).
    apply hedge_in_synthesis_impl_hedge_synthesis_spec.
    trivial.
    auto.
    destruct (inverse_keys m2 n2);auto.
    destruct p;auto.
    assert ((hedge_in_synthesis h m m0)=true).
    destruct (hedge_in_synthesis h m m0);try (discriminate H0);auto.
    rewrite H1 in H0;simpl in H0.
    apply andb_true_intro.
    split.
    apply hedge_synthesis_spec_impl_hedge_in_synthesis.
    unfold hedge_synthesis_spec.
    apply (incl_impl_le_h (relation_of_hedge h) (relation_of_hedge h') (Subset_impl_inclusion h h' H)).
    apply hedge_in_synthesis_impl_hedge_synthesis_spec.
    trivial.
    apply H.mem_1.
    apply H.
    apply H.mem_2;auto.
  Qed.

  Lemma is_analysed_Subset_false : forall (h h':hedge)(m n:MessageSpi.message),(H.Subset h' h)->(is_analysed h m n)=false->(is_analysed h' m n)=false.
    intros.
    case (Bool_elim_bool (is_analysed h' m n));intro;trivial.
    rewrite (is_analysed_Subset_true h' h m n) in H0;trivial.
    discriminate H0.
  Qed.

  Lemma in_analysed : forall (h:hedge)(m n:MessageSpi.message),(is_analysed h m n = true)->forall (a:H.elt),(H.In a (analz_elt h m n))->(H.In a h)\/(exists m':MessageSpi.message,exists n':MessageSpi.message,(H.In (m',n') h)/\(negb (is_analysed h m' n'))=true/\(H.In a (analz_elt h m' n'))).
    induction m;simpl;intros.
    elim (H.empty_1 H0).
    destruct n;try (elim (H.empty_1 H0);fail).
    elim (HedgeProp.Add_add (H.add (m2, n2) (H.union (analz_elt h m1 n1) (analz_elt h m2 n2))) (m1,n1) a);intros.
    case (H1 H0);clear H0;clear H1;clear H2;intros.
    left.
    eapply H.In_1.
    apply H0.
    apply H.mem_2.
    destruct (H.mem (m1,n1) h);try (discriminate H);auto.
    elim (HedgeProp.Add_add (H.union (analz_elt h m1 n1) (analz_elt h m2 n2)) (m2,n2) a);intros.
    case (H1 H0);clear H0;clear H1;clear H2;intros.
    left.
    eapply H.In_1.
    apply H0.
    apply H.mem_2.
    destruct (H.mem (m1,n1) h);try (discriminate H).
    destruct (H.mem (m2,n2) h);try (discriminate H);auto.
    case (H.union_1 H0);clear H0;intro.
    case (Bool_elim_bool (is_analysed h m1 n1));intro.
    eapply IHm1;eauto.
    right.
    exists m1;exists n1.
    rewrite H1;simpl.
    repeat (split;trivial).
    eapply H.mem_2.
    destruct (H.mem (m1,n1) h);try (discriminate H);auto.
    case (Bool_elim_bool (is_analysed h m2 n2));intro.
    eapply IHm2;eauto.
    right.
    exists m2;exists n2.
    rewrite H1;simpl.
    repeat (split;trivial).
    eapply H.mem_2.
    destruct (H.mem (m1,n1) h);try (discriminate H).
    destruct (H.mem (m2,n2) h);try (discriminate H);auto.
    destruct n;try (elim (H.empty_1 H0);fail).
    destruct (hedge_in_synthesis h m2 n2);try (discriminate H).
    elim (HedgeProp.Add_add (analz_elt h m1 n1) (m1,n1) a);intros.
    case (H1 H0);clear H0;clear H1;clear H2;intros.
    left.
    eapply H.In_1.
    apply H0.
    apply H.mem_2.
    destruct (H.mem (m1,n1) h);try (discriminate H);auto.
    case (Bool_elim_bool (is_analysed h m1 n1));intro.
    eapply IHm1;eauto.
    right.
    exists m1;exists n1.
    rewrite H1;simpl.
    repeat (split;trivial).
    eapply H.mem_2.
    destruct (H.mem (m1,n1) h);try (discriminate H);auto.
    destruct n;try (elim (H.empty_1 H0);fail).
    destruct (inverse_keys m2 n2).
    destruct p as [m'2 n'2].
    destruct (hedge_in_synthesis h m'2 n'2).
    simpl in H.
    elim (HedgeProp.Add_add (analz_elt h m1 n1) (m1,n1) a);intros.
    case (H1 H0);clear H0;clear H1;clear H2;intros.
    left.
    eapply H.In_1.
    apply H0.
    apply H.mem_2.
    auto.
    case (Bool_elim_bool (is_analysed h m1 n1));intro.
    eapply IHm1;eauto.
    right.
    exists m1;exists n1.
    rewrite H1;simpl.
    repeat (split;trivial).
    eapply H.mem_2.
    auto.
    elim (H.empty_1 H0).
    elim (H.empty_1 H0).
    elim (H.empty_1 H0).
  Qed.

  Lemma analz_aux_is_analysed : forall (h:hedge),(H.Equal (H.union h (analz_aux h h)) (H.union h (analz_aux h (H.filter (fun mn => let (m,n) := mn in negb (is_analysed h m n)) h)))).
    split;intros.
    case (H.union_1 H);intro.
    apply H.union_2;auto.
    unfold analz_aux.
    elim (HedgeOtherProp.In_big_union (analz_elt_aux h) (analz_elt_aux_compat h) h a);intros.
    elim (H1 H0);intro x;intros;and_elim.
    destruct x as [m n].
    simpl in H4.
    case (Bool_elim_bool (is_analysed h m n));intro.
    case (in_analysed h m n H5 a H4);intro;auto.
    ex_elim H6 m'.
    ex_elim H6 n'.
    elim (HedgeOtherProp.In_big_union (analz_elt_aux h) (analz_elt_aux_compat h) (H.filter (fun mn => let (m,n) := mn in negb (is_analysed h m n)) h) a);intros.
    apply H.union_3.
    apply H10.
    exists (m',n').
    simpl.
    split;auto.
    apply H.filter_3;auto.
    apply is_analysed_compat_bool.
    apply H.union_3.
    elim (HedgeOtherProp.In_big_union (analz_elt_aux h) (analz_elt_aux_compat h) (H.filter (fun mn => let (m,n) := mn in negb (is_analysed h m n)) h) a);intros.
    apply H7.
    exists (m,n).
    simpl.
    split;auto.
    apply H.filter_3;auto.
    apply is_analysed_compat_bool.
    rewrite H5;reflexivity.
    case (H.union_1 H);intro.
    apply H.union_2;auto.
    elim (HedgeOtherProp.In_big_union (analz_elt_aux h) (analz_elt_aux_compat h) (H.filter (fun mn => let (m,n) := mn in negb (is_analysed h m n)) h) a);intros.
    elim (H1 H0);clear H0;clear H1;clear H2.
    intro x;intros.
    destruct x as [m n];and_elim.
    simpl in H1.
    apply H.union_3.
    elim (HedgeOtherProp.In_big_union (analz_elt_aux h) (analz_elt_aux_compat h) h a);intros.
    apply H3.
    exists (m,n).
    simpl.
    split;auto.
    eapply H.filter_1;eauto.
    apply is_analysed_compat_bool.
  Qed.

  Definition not_yet_analysed_elts (h:hedge) := H.filter (fun mn => let (m,n) := mn in (negb (is_analysed h m n))) h.

  Definition analysable_elts (h:hedge) := H.filter (fun mn => let (m,n) := mn in (is_analysable h m n)) (not_yet_analysed_elts h).

  Theorem analz_simpl : forall (h:hedge),H.Equal (analz h) (H.union h (analz_aux h (analysable_elts h))).
    unfold analz.
    intros.
    eapply HedgeProp.equal_trans.
    apply analz_aux_is_analysed.    
    unfold analysable_elts.
    apply HedgeOtherProp.Equal_union_Equal_Equal.
    apply HedgeProp.equal_refl.
    eapply HedgeProp.equal_trans.
    apply analz_aux_is_analysable.
    unfold not_yet_analysed_elts.
    apply HedgeProp.equal_refl.
  Qed.

  Lemma empty_analysable_elts : forall (h:hedge),(H.Empty (analysable_elts h))->(H.Equal (analz h) h).
    intros.
    eapply HedgeProp.equal_trans.
    apply analz_simpl.
    split;intros.
    case (H.union_1 H0);auto. 
    intro.
    elim (HedgeOtherProp.In_big_union (analz_elt_aux h) (analz_elt_aux_compat h) (analysable_elts h) a);intros.
    elim (H2 H1);intro x;intros;and_elim.
    elim (H x H4).
    auto.
  Qed.

  Lemma In_analysable_elts : forall (h:hedge)(m n:MessageSpi.message),(H.In (m,n) (analysable_elts h))<->((H.In (m,n) h)/\(is_analysable h m n = true)/\(is_analysed h m n = false)).
    unfold analysable_elts.
    unfold not_yet_analysed_elts.
    split;intros.
    split.
    eapply H.filter_1.
    apply (is_analysed_compat_bool h).
    eapply H.filter_1.
    apply (is_analysable_compat_bool h).
    auto.
    split.
    generalize (H.filter_2 (is_analysable_compat_bool h) H).
    auto.
    generalize (H.filter_2 (is_analysed_compat_bool h) (H.filter_1 (is_analysable_compat_bool h) H)).
    intro.
    destruct (is_analysed h m n);try (discriminate H0);auto.
    and_elim.
    apply H.filter_3.
    apply is_analysable_compat_bool.
    apply H.filter_3.
    apply is_analysed_compat_bool.
    auto.
    rewrite H1;reflexivity.
    auto.
  Qed.
    
  Lemma is_analysable_is_analysed_analz : forall (h:hedge)(m n:MessageSpi.message),(is_analysable h m n = true)->(H.In (m,n) h)->(is_analysed (analz h) m n = true).
    induction m;simpl;intros;auto.
    destruct n;try (discriminate H).
    apply andb_true_intro.
    unfold analz.
    unfold analz_aux.
    generalize (HedgeOtherProp.In_big_union (analz_elt_aux h) (analz_elt_aux_compat h) h).
    intro.
    split.
    apply H.mem_1.
    apply H.union_3.
    elim (H1 (m1,n1));intros.
    apply H3.
    exists (MessageSpi.MPair m1 m2, MessageSpi.MPair n1 n2).
    split;trivial.
    simpl.
    auto.
    apply H.mem_1.
    apply H.union_3.
    elim (H1 (m2,n2));intros.
    apply H3.
    exists (MessageSpi.MPair m1 m2, MessageSpi.MPair n1 n2).
    split;trivial.
    simpl;auto.
    destruct n;try (discriminate H).
    apply andb_true_intro.
    unfold analz.
    unfold analz_aux.
    split.
    apply hedge_synthesis_spec_impl_hedge_in_synthesis.
    generalize (incl_impl_le_h (relation_of_hedge h) (relation_of_hedge (H.union h (HedgeOtherProp.big_union (analz_elt_aux h) h)))).
    unfold le_h.
    intros.
    unfold hedge_synthesis_spec.
    apply H1.
    apply Subset_impl_inclusion.
    red;intros;auto.
    apply hedge_in_synthesis_impl_hedge_synthesis_spec.
    apply H.
    elim (HedgeOtherProp.In_big_union (analz_elt_aux h) (analz_elt_aux_compat h) h (m1,n1)).
    intros.
    apply H.mem_1.
    apply H.union_3.
    apply H2.
    exists (MessageSpi.MEnc_s m1 m2, MessageSpi.MEnc_s n1 n2).
    split;auto.
    simpl.
    rewrite H.
    auto.
    destruct n;try (discriminate H).
    case (Option_elim_option _ (inverse_keys m2 n2));intro H1;try (rewrite H1);auto.
    ex_elim H1 y.
    destruct y as [m'2 n'2].
    rewrite H1.
    rewrite H1 in H.
    apply andb_true_intro.
    unfold analz.
    unfold analz_aux.
    split.
    apply hedge_synthesis_spec_impl_hedge_in_synthesis.
    generalize (incl_impl_le_h (relation_of_hedge h) (relation_of_hedge (H.union h (HedgeOtherProp.big_union (analz_elt_aux h) h)))).
    unfold le_h.
    intros.
    unfold hedge_synthesis_spec.
    apply H2.
    apply Subset_impl_inclusion.
    red;intros;auto.
    apply hedge_in_synthesis_impl_hedge_synthesis_spec.
    apply H.
    elim (HedgeOtherProp.In_big_union (analz_elt_aux h) (analz_elt_aux_compat h) h (m1,n1)).
    intros.
    apply H.mem_1.
    apply H.union_3.
    apply H3.
    exists (MessageSpi.MEnc_a m1 m2, MessageSpi.MEnc_a n1 n2).
    split;auto.
    simpl.
    rewrite H1.
    rewrite H.
    auto.
  Qed.

  Lemma size_pair_analz_elt : forall (h:hedge)(m n:MessageSpi.message)(m' n':MessageSpi.message),(H.In (m',n') (analz_elt h m n))->(size_pair (m',n') < size_pair (m,n)).
    induction m;intros.
    simpl in H.
    elim (H.empty_1 H).
    destruct n;try (elim (H.empty_1 H);fail).
    simpl in H.
    elim (HedgeProp.Add_add (H.add (m2, n2) (H.union (analz_elt h m1 n1) (analz_elt h m2 n2))) (m1,n1) (m',n'));intros.
    case (H0 H);clear H;clear H0;clear H1;intros.
    simpl in H;and_elim.
    rewrite H;rewrite H0.
    simpl.
    apply le_lt_n_Sm.
    apply max_case2.
    rewrite Max_assoc.
    apply le_max_l.
    eapply le_trans.
    apply le_max_l.
    apply le_max_r.
    elim (HedgeProp.Add_add (H.union (analz_elt h m1 n1) (analz_elt h m2 n2)) (m2,n2) (m',n'));intros.
    case (H0 H);clear H;clear H0;clear H1;intros.
    simpl in H;and_elim.
    rewrite H;rewrite H0.
    simpl.
    apply le_lt_n_Sm.
    apply max_case2.
    eapply le_trans.
    apply le_max_r.
    apply le_max_l.
    eapply le_trans.
    apply le_max_r.
    apply le_max_r.
    case (H.union_1 H);intro.
    generalize (IHm1 n1 m' n' H0).
    intro.
    eapply lt_trans.
    apply H1.
    simpl.
    apply le_lt_n_Sm.
    apply max_case2.
    rewrite Max_assoc.
    apply le_max_l.
    eapply le_trans.
    apply le_max_l.
    apply le_max_r.
    generalize (IHm2 n2 m' n' H0).
    intro.
    eapply lt_trans.
    apply H1.
    simpl.
    apply le_lt_n_Sm.
    apply max_case2.
    eapply le_trans.
    apply le_max_r.
    apply le_max_l.
    eapply le_trans.
    apply le_max_r.
    apply le_max_r.
    simpl in H.
    destruct n;try (elim (H.empty_1 H);fail).
    destruct (hedge_in_synthesis h m2 n2);try (elim (H.empty_1 H);fail).
    elim (HedgeProp.Add_add (analz_elt h m1 n1) (m1,n1) (m',n'));intros.
    case (H0 H);clear H;clear H0;clear H1;intros.
    simpl in H;and_elim.
    rewrite H;rewrite H0.
    simpl.
    apply le_lt_n_Sm.
    apply max_case2.
    rewrite Max_assoc.
    apply le_max_l.
    eapply le_trans.
    apply le_max_l.
    apply le_max_r.
    generalize (IHm1 n1 m' n' H);intro.
    eapply lt_trans.
    apply H0.
    simpl.
    apply le_lt_n_Sm.
    apply max_case2.
    rewrite Max_assoc.
    apply le_max_l.
    eapply le_trans.
    apply le_max_l.
    apply le_max_r.
    simpl in H.
    destruct n;try (elim (H.empty_1 H);fail).
    destruct (inverse_keys m2 n2);try (elim (H.empty_1 H);fail).
    destruct p as [m'2 n'2].
    destruct (hedge_in_synthesis h m'2 n'2);try (elim (H.empty_1 H);fail).
    elim (HedgeProp.Add_add (analz_elt h m1 n1) (m1,n1) (m',n'));intros.
    case (H0 H);clear H;clear H0;clear H1;intros.
    simpl in H;and_elim.
    rewrite H;rewrite H0.
    simpl.
    apply le_lt_n_Sm.
    apply max_case2.
    rewrite Max_assoc.
    apply le_max_l.
    eapply le_trans.
    apply le_max_l.
    apply le_max_r.
    generalize (IHm1 n1 m' n' H);intro.
    eapply lt_trans.
    apply H0.
    simpl.
    apply le_lt_n_Sm.
    apply max_case2.
    rewrite Max_assoc.
    apply le_max_l.
    eapply le_trans.
    apply le_max_l.
    apply le_max_r.
    simpl in H.
    elim (H.empty_1 H).
  Qed.

  Lemma In_not_yet_analysed : forall (h:hedge)(m n:MessageSpi.message),(H.In (m,n) (not_yet_analysed_elts h))<->(H.In (m,n) h)/\(is_analysed h m n = false).
    unfold not_yet_analysed_elts.
    split;intros.
    split.
    eapply H.filter_1.
    apply (is_analysed_compat_bool h).
    auto.
    generalize (H.filter_2 (is_analysed_compat_bool h) H).
    intro.
    destruct (is_analysed h m n);try (discriminate H0);auto.
    and_elim.
    apply H.filter_3;auto.
    apply is_analysed_compat_bool.
    rewrite H0;reflexivity.
  Qed.

  Lemma analysable_elts_decrease : forall (h:hedge)(m n:MessageSpi.message),(H.In (m,n) (not_yet_analysed_elts (analz h)))->((H.In (m,n) (not_yet_analysed_elts h))/\(is_analysable h m n = false))\/(exists m':MessageSpi.message,exists n':MessageSpi.message,(H.In (m',n') (analysable_elts h))/\(size_pair (m,n) < size_pair (m',n'))).
    intros.
    elim (In_not_yet_analysed (analz h) m n);intros.
    elim (H0 H);clear H;clear H0;clear H1;intros;and_elim.
    assert (is_analysed h m n = false).
    eapply is_analysed_Subset_false with (analz h).
    unfold analz.
    red;intros;auto.
    auto.
    elim (analz_simpl h (m,n));intros.
    case (H.union_1 (H2 H));clear H2;clear H3;intro.
    case (Bool_elim_bool (is_analysable h m n));intro.
    rewrite (is_analysable_is_analysed_analz h m n) in H0;auto.
    discriminate H0.
    left.
    repeat (split;trivial).
    elim (In_not_yet_analysed h m n);intros.
    apply H5;tauto.
    unfold analz_aux in H2.
    elim (HedgeOtherProp.In_big_union (analz_elt_aux h) (analz_elt_aux_compat h) (analysable_elts h) (m,n));intros.
    elim (H3 H2);clear H2;clear H3;clear H4;intros;and_elim.
    destruct x as [m' n'].
    right.
    exists m'.
    exists n'.
    unfold analz_elt_aux in H3.
    split;trivial.
    eapply size_pair_analz_elt.    
    apply H3.
  Qed.

  Definition lt_not_yet_analysed (h h':hedge) := (HedgeOrder.lt_set1 size_pair (not_yet_analysed_elts h) (not_yet_analysed_elts h')).

  Lemma non_empty_analysable_elts : forall (h:hedge),(~H.Empty (analysable_elts h))->(lt_not_yet_analysed (analz h) h).
    unfold lt_not_yet_analysed.
    intros.
    apply HedgeOrder.cond_lt_set1.
    exists (H.diff (not_yet_analysed_elts h) (analysable_elts h)).
    assert (forall (m n:MessageSpi.message),(H.In (m,n) (H.diff (not_yet_analysed_elts h) (analysable_elts h)))<->((H.In (m,n) (not_yet_analysed_elts h))/\(is_analysable h m n = false))).
    split;intros.
    generalize (H.diff_1 H0);intro.
    generalize (H.diff_2 H0);intro.
    split;trivial.
    elim (In_analysable_elts h m n);intros.
    destruct (is_analysable h m n);auto.
    elim H2.
    elim (In_not_yet_analysed h m n);intros.
    elim (H5 H1);intros;and_elim.
    apply H4;try tauto.
    and_elim.
    apply H.diff_3.
    trivial.
    red;intro.
    elim (In_analysable_elts h m n);intros.
    elim (H3 H2);intros;and_elim.
    rewrite H6 in H1;discriminate H1.
    assert (H.Equal (H.diff (not_yet_analysed_elts h) (H.diff (not_yet_analysed_elts h) (analysable_elts h))) (analysable_elts h)).
    apply HedgeOtherProp.Subset_diff_diff.
    red;intros.
    destruct a.
    elim (In_analysable_elts h t t0);intros.
    elim (H2 H1);clear H1;clear H2;clear H3;intros;and_elim.
    elim (In_not_yet_analysed h t t0);intros.
    apply H5;try tauto.
    split.
    elim (HedgeOtherProp.strict_Subset_iff_1 (H.diff (not_yet_analysed_elts h) (analysable_elts h)) (not_yet_analysed_elts h));intros.    
    apply H3.
    split.
    red;intros.
    eapply H.diff_1.
    apply H4.
    red;intro.
    apply H.
    red;intros.
    red;intro.
    elim (H4 a).
    elim (H1 a);auto.
    intros.
    destruct x as [m n].
    case (analysable_elts_decrease h m n H2);intro.
    left.
    elim (H0 m n);intros;auto.
    right.
    ex_elim H3 m'.
    ex_elim H3 n'.
    exists (m',n').
    split;trivial.
    elim (H1 (m',n'));auto.
  Qed.

  Require Import Wellfounded.

  Lemma lt_not_yet_analysed_wf : well_founded lt_not_yet_analysed.
    unfold lt_not_yet_analysed.
    apply (wf_inverse_image hedge hedge (HedgeOrder.lt_set1 size_pair) (fun h => not_yet_analysed_elts h)).
    apply HedgeOrder.well_founded_lt_set1.
    apply size_pair_eq.
  Qed.

  (** h is stable by applying hedge_analysis_step *)
  Definition stable_analysis_step (h:hedge) := (H.Subset (hedge_analysis_step h) h).

  Definition stable_analysis_relation (h:message_relation) := (inclusion MessageSpi.message (hedge_analysis_relation h) h).

  Lemma stable_analysis_step_impl_stable_analysis_relation : forall (h:hedge),(stable_analysis_step h)->(stable_analysis_relation (relation_of_hedge h)).
    unfold stable_analysis_step, stable_analysis_relation;intros.
    unfold inclusion;intros.
    fold (hedge_analysis_step_spec h) in H0.
    generalize (hedge_analysis_step_spec_impl_hedge_analysis_step h x y H0);intro.
    generalize (H (x,y) H1);intro.
    trivial.
  Qed.

  Lemma stable_analysis_relation_impl_stable_analysis_step : forall (h:hedge),(stable_analysis_relation (relation_of_hedge h))->(stable_analysis_step h).
    unfold stable_analysis_step, stable_analysis_relation;intros.
    unfold H.Subset;intros.
    destruct a.
    apply H.
    apply (hedge_analysis_step_impl_hedge_analysis_step_spec h t t0 H0).
  Qed.

  (** if h is stable by hedge_analysis_step then h = hedge_analysis_step h *)
  Lemma stable_closed : forall (h:hedge),(stable_analysis_step h)->(H.Equal h (hedge_analysis_step h)).
    unfold stable_analysis_step.
    intros.
    eapply HedgeProp.subset_antisym;trivial.
    intro;intro.
    destruct a.
    apply hedge_analysis_step_spec_impl_hedge_analysis_step.
    unfold hedge_analysis_step_spec;apply AnaInc.
    trivial.
  Qed.

  (** ah can be the analysis of h if h is included in ah and ah is stable by hedge_analysis_step *)
  Definition analysis_cond (h ah:hedge) := (H.Subset h ah)/\(stable_analysis_step ah).

  Definition analysis_cond_relation (h ah:message_relation) := (inclusion MessageSpi.message h ah)/\(stable_analysis_relation ah).
  
  Lemma analysis_cond_impl_analysis_cond_relation : forall (h ah:hedge),(analysis_cond h ah)->(analysis_cond_relation (relation_of_hedge h) (relation_of_hedge ah)).
    unfold analysis_cond, analysis_cond_relation;intros;and_elim.
    split.
    apply (Subset_impl_inclusion h ah H).
    apply stable_analysis_step_impl_stable_analysis_relation;trivial.
  Qed.

  Lemma analysis_cond_relation_impl_analysis_cond : forall (h ah:hedge),(analysis_cond_relation (relation_of_hedge h) (relation_of_hedge ah))->(analysis_cond h ah).
    unfold analysis_cond, analysis_cond_relation;intros;and_elim.
    split.
    apply (inclusion_impl_Subset h ah H).
    apply stable_analysis_relation_impl_stable_analysis_step;trivial.
  Qed.

  (** ah is the analysis of h if ah is the least subset that satisfies analysis_cond *)
  Definition is_analysis (h ah:hedge) := (analysis_cond h ah)/\(forall (ah':hedge),(analysis_cond h ah')->(H.Subset ah ah')).

  Definition is_analysis_relation (h ah:message_relation) := (analysis_cond_relation h ah)/\(forall (ah':message_relation),(analysis_cond_relation h ah')->(inclusion MessageSpi.message ah ah')).

  Lemma hedge_analysis_step_grow : forall (h h':hedge),(H.Subset h h')->(H.Subset (hedge_analysis_step h) (hedge_analysis_step h')).
    intros.
    intro;intro.
    destruct a as (m,n).
    generalize (hedge_analysis_step_impl_hedge_analysis_step_spec h m n H0);clear H0;intro.
    apply hedge_analysis_step_spec_impl_hedge_analysis_step.
    induction H0.
    unfold hedge_analysis_step_spec;apply AnaInc.
    apply (Subset_impl_inclusion h h' H);trivial.
    unfold hedge_analysis_step_spec;eapply AnaSplitL.
    unfold hedge_analysis_step_spec in IHhedge_analysis_relation;eauto.
    unfold hedge_analysis_step_spec;eapply AnaSplitR.
    unfold hedge_analysis_step_spec in IHhedge_analysis_relation;eauto.
    unfold hedge_analysis_step_spec;eapply AnaDec_s.
    unfold hedge_analysis_step_spec in IHhedge_analysis_relation;eauto.
    apply (incl_impl_le_h (relation_of_hedge h) (relation_of_hedge h') (Subset_impl_inclusion h h' H)).
    trivial.
    unfold hedge_analysis_step_spec;eapply AnaDec_a.
    unfold hedge_analysis_step_spec in IHhedge_analysis_relation;eauto.
    apply H1.
    apply (incl_impl_le_h (relation_of_hedge h) (relation_of_hedge h') (Subset_impl_inclusion h h' H)).
    trivial.
  Qed.

  Lemma analysable_empty_analz_is_analysis : forall (h:hedge),(H.Empty (analysable_elts h))->(is_analysis h h).
    intros.
    repeat (split;auto).
    apply HedgeProp.subset_refl.
    unfold stable_analysis_step.
    generalize (empty_analysable_elts h H);intro.
    eapply HedgeProp.subset_equal.
    apply H0.
    intros.
    elim H0;clear H0;intros;and_elim.
    auto.
  Qed.

  Lemma is_analysis_analz_is_analysis : forall (h h':hedge),(is_analysis (analz h) h')->(is_analysis h h').
    intros.
    elim H;clear H;intros;and_elim.
    elim H;clear H;intros;and_elim.
    repeat (split;auto).
    apply HedgeProp.subset_trans with (analz h).
    unfold analz.
    red;intros;auto.
    apply H.
    intros.
    apply H0.
    elim H2;clear H2;intros;and_elim.
    split;auto.
    unfold stable_analysis_step in H3.
    unfold hedge_analysis_step in H3.
    eapply HedgeProp.subset_trans.
    apply hedge_analysis_step_grow.
    apply H2.
    unfold hedge_analysis_step;auto.
  Qed.

  (** here it is: A(h) is decidable *)
  Definition analysis (h:hedge) : {ah:hedge | is_analysis h ah}.
    intro h.
    elim h using (well_founded_induction_type lt_not_yet_analysed_wf);clear h;intros.
    case (HedgeOtherProp.empty_dec (analysable_elts x));intro.
    exists x.
    apply analysable_empty_analz_is_analysis.
    red;intro.
    red;intro.
    elim (e a);intros.
    elim (H.empty_1 (H1 H0)).
    assert (~H.Empty (analysable_elts x)).    
    red;intro.
    elim n.
    apply HedgeProp.subset_antisym;auto.
    generalize (non_empty_analysable_elts x H0);intro.
    elim (H (analz x) H1).
    intro ah;intro.
    exists ah.
    apply is_analysis_analz_is_analysis;auto.
  Defined.
    
  Definition hedge_analysis (h:hedge) := let (ah,_) := analysis h in ah.

  Theorem hedge_analysis_is_analysis : forall (h:hedge),(is_analysis h (hedge_analysis h)).
    unfold hedge_analysis.
    intro.
    elim (analysis h);auto.
  Qed.

  (** A(h) is unique *)
  Lemma analysis_unique : forall (g h h':hedge),(is_analysis g h)->(is_analysis g h')->(H.Equal h h').
    unfold is_analysis;intros;and_elim.
    eapply HedgeProp.subset_antisym.
    apply H2;trivial.
    apply H1;trivial.
  Qed.

  (** we show that A(h) is the iteration of hedge_analysis_step (useful for proving some properties of A(h)) *)
  Fixpoint hedge_analysis_step_iter (h:hedge) (n:nat) {struct n} : hedge :=
    match n with
      | O => h
      | (S n) => (hedge_analysis_step_iter (hedge_analysis_step h) n)
    end.

  Lemma hedge_analysis_hedge_analysis_step : forall (h:hedge),(H.Equal (hedge_analysis h) (hedge_analysis (hedge_analysis_step h))).
    intros.
    eapply analysis_unique with (hedge_analysis_step h).
    generalize (hedge_analysis_is_analysis h).
    unfold is_analysis;unfold analysis_cond;intros;and_elim.
    split.
    split.
    unfold stable_analysis_step in H1.
    eapply HedgeProp.subset_trans.
    apply hedge_analysis_step_grow.
    apply H.
    trivial.
    trivial.
    intros.
    and_elim.
    apply H0.
    split;trivial.
    apply HedgeProp.subset_trans with (hedge_analysis_step h);trivial.
    intro;intro.
    destruct a.
    apply hedge_analysis_step_spec_impl_hedge_analysis_step.
    unfold hedge_analysis_step_spec.
    apply AnaInc.
    trivial.
    apply (hedge_analysis_is_analysis (hedge_analysis_step h)).
  Qed.

  Lemma hedge_analysis_step_iter_Sn : forall (n:nat)(h:hedge),(H.Equal (hedge_analysis_step_iter h (S n)) (hedge_analysis_step (hedge_analysis_step_iter h n))).
    induction n;simpl;intros.
    apply HedgeProp.equal_refl.
    simpl in IHn.
    eapply HedgeProp.equal_trans.
    apply IHn.
    apply HedgeProp.equal_refl.
  Qed.

  Lemma hedge_analysis_step_morphism : forall (g h:hedge),(H.Equal g h)->(H.Equal (hedge_analysis_step g) (hedge_analysis_step h)).
    intros.
    apply HedgeProp.subset_antisym;apply hedge_analysis_step_grow;apply HedgeProp.subset_equal;trivial.
    apply HedgeProp.equal_sym;trivial.
  Qed.

  Lemma hedge_analysis_step_iter_plus : forall (n m:nat)(h:hedge),(H.Equal (hedge_analysis_step_iter h (n+m)) (hedge_analysis_step_iter (hedge_analysis_step_iter h m) n)).
    induction n.
    simpl;intros.
    apply HedgeProp.equal_refl.
    intros.
    replace (S n + m) with (S (n+m));try reflexivity.
    eapply HedgeProp.equal_trans.
    apply hedge_analysis_step_iter_Sn.
    eapply HedgeProp.equal_trans.
    apply hedge_analysis_step_morphism.
    apply IHn.
    apply HedgeProp.equal_sym.
    apply hedge_analysis_step_iter_Sn.
  Qed.

  Lemma hedge_analysis_step_iter_grow_aux : forall (n m:nat)(h:hedge),(H.Subset (hedge_analysis_step_iter h m) (hedge_analysis_step_iter h (n+m))).
    induction n.
    simpl;intros.
    apply HedgeProp.subset_equal.
    apply HedgeProp.equal_refl.
    intros.
    replace (S n+m) with (S (n+m));try reflexivity.
    eapply HedgeProp.subset_trans.
    apply IHn.
    eapply HedgeProp.subset_trans with (hedge_analysis_step (hedge_analysis_step_iter h (n+m))).
    unfold hedge_analysis_step.
    unfold analz.
    red;intros;auto.
    apply HedgeProp.subset_equal.
    apply HedgeProp.equal_sym.
    apply hedge_analysis_step_iter_Sn.
  Qed.

  Lemma hedge_analysis_step_iter_grow : forall (n m:nat)(h:hedge),(n<=m)->(H.Subset (hedge_analysis_step_iter h n) (hedge_analysis_step_iter h m)).
    intros.
    assert (m=n+(m-n));auto with arith.
    rewrite plus_comm in H0.
    rewrite H0.
    apply hedge_analysis_step_iter_grow_aux.
  Qed.

  (** A(h) = hedge_analysis_step^n(h) for some n *)
  Theorem hedge_analysis_is_iter : forall (h:hedge),{n:nat | (H.Equal (hedge_analysis h) (hedge_analysis_step_iter h n))}.
    intro h.
    elim h using (well_founded_induction_type lt_not_yet_analysed_wf);clear h;intros.
    case (HedgeOtherProp.empty_dec (analysable_elts x));intro.
    exists O.
    simpl.
    eapply analysis_unique with x.
    apply hedge_analysis_is_analysis.
    apply analysable_empty_analz_is_analysis.
    red;intro.
    red;intro.
    elim (e a);intros.
    elim (H.empty_1 (H1 H0)).
    assert (~H.Empty (analysable_elts x)).    
    red;intro.
    elim n.
    apply HedgeProp.subset_antisym;auto.
    generalize (non_empty_analysable_elts x H0);intro.
    elim (H (analz x) H1).
    intro p;intros.
    exists (S p).
    simpl.
    unfold hedge_analysis_step.
    eapply HedgeProp.equal_trans.
    apply hedge_analysis_hedge_analysis_step.
    unfold hedge_analysis_step.
    auto.
  Defined.

  Lemma analysis_trans : forall (g h h':hedge),(is_analysis g h)->(is_analysis h h')->(is_analysis g h').
    unfold is_analysis;unfold analysis_cond;intros;and_elim.
    split.
    split.
    intro;intro.
    apply H0.
    apply H;trivial.
    trivial.
    intros.
    and_elim.
    apply H1.
    split;trivial.
    intro;intro.
    apply H3;tauto.
  Qed.

  Lemma analysis_analysed : forall (g h h':hedge),(is_analysis g h)->(is_analysis h h')->(H.Equal h h').
    intros.
    eapply analysis_unique.
    apply H.
    eapply analysis_trans.
    apply H.
    trivial.
  Qed.

  Lemma hedge_analysis_step_iter_analysis : forall (n:nat)(h:hedge),(H.Subset (hedge_analysis_step_iter (hedge_analysis h) n) (hedge_analysis h)).
    induction n.
    simpl;intros.
    apply HedgeProp.subset_equal.
    apply HedgeProp.equal_refl.
    intros.
    eapply HedgeProp.subset_trans.
    apply HedgeProp.subset_equal.
    apply hedge_analysis_step_iter_Sn.
    eapply HedgeProp.subset_trans.
    apply hedge_analysis_step_grow.
    apply IHn.
    generalize (hedge_analysis_is_analysis h).
    unfold is_analysis;intros;and_elim.
    elim H;intros.
    apply H2.
  Qed.

  Lemma hedge_analysis_step_iter_morphism : forall (n:nat)(g h:hedge),(H.Equal g h)->(H.Equal (hedge_analysis_step_iter g n) (hedge_analysis_step_iter h n)).
    induction n.
    simpl;trivial.
    intros.
    eapply HedgeProp.equal_trans.
    apply hedge_analysis_step_iter_Sn.
    eapply HedgeProp.equal_trans.
    apply hedge_analysis_step_morphism.
    apply IHn.
    apply H.
    apply HedgeProp.equal_sym.
    apply hedge_analysis_step_iter_Sn.
  Qed.

  Lemma hedge_analysis_step_iter_Subset_hedge_analysis : forall (n:nat)(h:hedge),(H.Subset (hedge_analysis_step_iter h n) (hedge_analysis h)).
    intros.
    elim (hedge_analysis_is_iter h);intro m;intros.
    case (le_lt_dec n m);intro.
    eapply HedgeProp.subset_trans.
    apply hedge_analysis_step_iter_grow.
    apply l.
    apply HedgeProp.subset_equal.
    apply HedgeProp.equal_sym;trivial.
    assert (n=m+(n-m));auto with arith.
    rewrite plus_comm in H.
    rewrite H.
    eapply HedgeProp.subset_trans.
    apply HedgeProp.subset_equal.
    apply hedge_analysis_step_iter_plus.
    eapply HedgeProp.subset_trans.
    apply HedgeProp.subset_equal.
    apply hedge_analysis_step_iter_morphism.
    apply HedgeProp.equal_sym.
    apply p.
    apply hedge_analysis_step_iter_analysis.
  Qed.

  Lemma hedge_analysis_step_iter_compat_Subset : forall (n:nat)(g h:hedge),(H.Subset g h)->(H.Subset (hedge_analysis_step_iter g n) (hedge_analysis_step_iter h n)).
    induction n;intros;simpl;auto.
    apply IHn.
    apply hedge_analysis_step_grow.
    auto.
  Qed.

  Lemma hedge_analysis_grow : forall (g h:hedge),(H.Subset g h)->(H.Subset (hedge_analysis g) (hedge_analysis h)).
    intros.
    elim (hedge_analysis_is_iter g).
    intro n;intros.
    eapply HedgeProp.subset_trans.
    apply HedgeProp.subset_equal.
    apply p.
    eapply HedgeProp.subset_trans.
    apply hedge_analysis_step_iter_compat_Subset.
    apply H.
    apply hedge_analysis_step_iter_Subset_hedge_analysis.
  Qed.

  Lemma hedge_analysis_morphism : forall (g h:hedge),(H.Equal g h)->(H.Equal (hedge_analysis g) (hedge_analysis h)).
    intros.
    apply HedgeProp.subset_antisym.
    apply hedge_analysis_grow.
    apply HedgeProp.subset_equal.
    auto.
    apply hedge_analysis_grow.
    apply HedgeProp.subset_equal.
    apply HedgeProp.equal_sym;auto.
  Qed.

  (** (m,n) is "strictly" synthesisable *)
  Definition is_synthesisable (h:hedge) (m n:MessageSpi.message) : bool :=
    match (m,n) with
      | (MessageSpi.MEnc_s m k, MessageSpi.MEnc_s n l) => andb (hedge_in_synthesis h m n) (hedge_in_synthesis h k l)
      | (MessageSpi.MEnc_a m k, MessageSpi.MEnc_a n l) => andb (hedge_in_synthesis h m n) (hedge_in_synthesis h k l)
      | (MessageSpi.MPair m1 m2, MessageSpi.MPair n1 n2) => andb (hedge_in_synthesis h m1 n1) (hedge_in_synthesis h m2 n2)
      | (MessageSpi.MOp o1 m, MessageSpi.MOp o2 n) => if (OpSpi.op_eq_dec o1 o2) then (hedge_in_synthesis h m n) else false
      | _ => false
    end.

  (** hedge_reduce h is the "irreducible part" of h *)
  Definition hedge_reduce (h:hedge) := H.filter (fun mn => let (m,n) := mn in negb (is_synthesisable h m n)) h.

  Lemma is_synthesisable_compat_bool :  forall (h:hedge),(FSetInterface.compat_bool H.E.eq (fun mn => let (m,n) := mn in negb (is_synthesisable h m n))).
    unfold FSetInterface.compat_bool.
    destruct x;destruct y;simpl;intros;and_elim.
    rewrite H;rewrite H0;trivial.
  Qed.

  Lemma hedge_reduce_Subset : forall (h:hedge),(H.Subset (hedge_reduce h) h).
    intros;unfold hedge_reduce;intro;intros.
    apply (H.filter_1 (is_synthesisable_compat_bool h) H).
  Qed.

  Lemma is_synthesisable_le_h : forall (g h:hedge),(le_h (relation_of_hedge g) (relation_of_hedge h))->forall (m n:MessageSpi.message),(is_synthesisable g m n = true)->(is_synthesisable h m n = true).
    destruct m;unfold is_synthesisable;intros;auto;destruct n;try (discriminate H0).
    elim (andb_prop (hedge_in_synthesis g m1 n1) (hedge_in_synthesis g m2 n2) H0);intros.
    apply andb_true_intro.
    split;apply hedge_synthesis_spec_impl_hedge_in_synthesis;unfold hedge_synthesis_spec;apply H;apply hedge_in_synthesis_impl_hedge_synthesis_spec;auto.
    elim (andb_prop (hedge_in_synthesis g m1 n1) (hedge_in_synthesis g m2 n2) H0);intros.
    apply andb_true_intro.
    split;apply hedge_synthesis_spec_impl_hedge_in_synthesis;unfold hedge_synthesis_spec;apply H;apply hedge_in_synthesis_impl_hedge_synthesis_spec;auto.
    elim (andb_prop (hedge_in_synthesis g m1 n1) (hedge_in_synthesis g m2 n2) H0);intros.
    apply andb_true_intro.
    split;apply hedge_synthesis_spec_impl_hedge_in_synthesis;unfold hedge_synthesis_spec;apply H;apply hedge_in_synthesis_impl_hedge_synthesis_spec;auto.
    destruct (OpSpi.op_eq_dec o o0);try (discriminate H0).
    apply hedge_synthesis_spec_impl_hedge_in_synthesis;unfold hedge_synthesis_spec;apply H;apply hedge_in_synthesis_impl_hedge_synthesis_spec;auto.
  Qed.

  (** h <= reduce(h) *)
  Lemma h_le_h_reduce : forall (h:hedge),(le_h (relation_of_hedge h) (relation_of_hedge (hedge_reduce h))).
    unfold le_h.
    induction m;intros.
    inversion H.
    apply SynInc.
    unfold hedge_reduce.
    apply (H.filter_3 (is_synthesisable_compat_bool h) H0);trivial.
    unfold hedge_reduce.
    destruct n;inversion H;try (apply SynInc;apply (H.filter_3 (is_synthesisable_compat_bool h) H0);auto;fail).
    case (Bool_elim_bool (is_synthesisable h (MessageSpi.MPair m1 m2) (MessageSpi.MPair n1 n2)));intro.
    unfold is_synthesisable in H0.
    elim (andb_prop (hedge_in_synthesis h m1 n1) (hedge_in_synthesis h m2 n2) H3);intros.
    fold (hedge_reduce h).
    apply SynPair.
    apply IHm1.
    apply hedge_in_synthesis_impl_hedge_synthesis_spec;trivial.
    apply IHm2.
    apply hedge_in_synthesis_impl_hedge_synthesis_spec;trivial.
    apply SynInc.
    apply (H.filter_3 (is_synthesisable_compat_bool h) H0);trivial.
    rewrite H3;trivial.
    fold (hedge_reduce h).
    apply SynPair;auto.
    destruct n;inversion H;try (apply SynInc;apply (H.filter_3 (is_synthesisable_compat_bool h) H0);auto;fail).
    case (Bool_elim_bool (is_synthesisable h (MessageSpi.MEnc_s m1 m2) (MessageSpi.MEnc_s n1 n2)));intro.
    unfold is_synthesisable in H3.
    elim (andb_prop (hedge_in_synthesis h m1 n1) (hedge_in_synthesis h m2 n2) H3);intros.
    apply SynEnc_s.
    apply IHm1.
    apply hedge_in_synthesis_impl_hedge_synthesis_spec;trivial.
    apply IHm2.
    apply hedge_in_synthesis_impl_hedge_synthesis_spec;trivial.
    apply SynInc.
    unfold hedge_reduce.
    apply (H.filter_3 (is_synthesisable_compat_bool h) H0).
    rewrite H3;trivial.
    apply SynEnc_s;auto.
    destruct n;inversion H;try (apply SynInc;apply (H.filter_3 (is_synthesisable_compat_bool h) H0);auto;fail).
    case (Bool_elim_bool (is_synthesisable h (MessageSpi.MEnc_a m1 m2) (MessageSpi.MEnc_a n1 n2)));intro.
    unfold is_synthesisable in H3.
    elim (andb_prop (hedge_in_synthesis h m1 n1) (hedge_in_synthesis h m2 n2) H3);intros.
    apply SynEnc_a.
    apply IHm1.
    apply hedge_in_synthesis_impl_hedge_synthesis_spec;trivial.
    apply IHm2.
    apply hedge_in_synthesis_impl_hedge_synthesis_spec;trivial.
    apply SynInc.
    unfold hedge_reduce.
    apply (H.filter_3 (is_synthesisable_compat_bool h) H0).
    rewrite H3;trivial.
    apply SynEnc_a;auto.
    destruct n;inversion H;try (apply SynInc;apply (H.filter_3 (is_synthesisable_compat_bool h) H0);auto;fail).  
    case (Bool_elim_bool (is_synthesisable h (MessageSpi.MOp o m) (MessageSpi.MOp o0 n)));intro.
    unfold is_synthesisable in H3.
    generalize H3.
    case OpSpi.op_eq_dec;intros.
    subst o0.
    apply SynOp.
    apply IHm.
    apply hedge_in_synthesis_impl_hedge_synthesis_spec;trivial.
    discriminate H4.
    apply SynInc.
    unfold hedge_reduce.
    apply (H.filter_3 (is_synthesisable_compat_bool h) H0).
    rewrite H3;trivial.
    apply SynOp;auto.
  Qed.

  (** h ~ reduce(h) *)
  Lemma h_equiv_h_red : forall (h:hedge),(equiv_h (relation_of_hedge h) (relation_of_hedge (hedge_reduce h))).
    split.
    apply h_le_h_reduce.
    apply incl_impl_le_h.
    apply Subset_impl_inclusion.
    apply hedge_reduce_Subset.
  Qed.

  (** if g is included in h and h <= g then reduce(g) <= reduce(h) *)
  Lemma Subset_le_h_reduce_Subset : forall (g h:hedge),(H.Subset g h)->(le_h (relation_of_hedge h) (relation_of_hedge g))->(H.Subset (hedge_reduce g) (hedge_reduce h)).
    unfold H.Subset, hedge_reduce, le_h;intros.
    destruct a as (m,n).
    generalize (H.filter_1 (is_synthesisable_compat_bool g) H1);intro.
    generalize (H.filter_2 (is_synthesisable_compat_bool g) H1);intro.
    generalize (H (m,n) H2);intro.
    unfold is_synthesisable in H3.
    destruct m;try (discriminate H3).
    apply (H.filter_3 (is_synthesisable_compat_bool h) H4);reflexivity.
    destruct n;apply (H.filter_3 (is_synthesisable_compat_bool h) H4);try reflexivity.
    unfold is_synthesisable.
    rewrite negb_andb in H3.
    rewrite negb_andb.
    apply orb_true_intro.
    case (orb_prop (negb (hedge_in_synthesis g m1 n1)) (negb (hedge_in_synthesis g m2 n2)) H3);intro;[left | right].
    case (Bool_elim_bool (hedge_in_synthesis h m1 n1));intro H6;rewrite H6;try reflexivity.
    generalize (hedge_in_synthesis_impl_hedge_synthesis_spec h m1 n1 H6);intro.
    generalize (H0 m1 n1 H7);intro. 
    rewrite (hedge_synthesis_spec_impl_hedge_in_synthesis g m1 n1 H8) in H5;discriminate H5.
    case (Bool_elim_bool (hedge_in_synthesis h m2 n2));intro H6;rewrite H6;try reflexivity.
    generalize (hedge_in_synthesis_impl_hedge_synthesis_spec h m2 n2 H6);intro.
    generalize (H0 m2 n2 H7);intro. 
    rewrite (hedge_synthesis_spec_impl_hedge_in_synthesis g m2 n2 H8) in H5;discriminate H5.
    destruct n;apply (H.filter_3 (is_synthesisable_compat_bool h) H4);try reflexivity.
    unfold is_synthesisable.
    rewrite negb_andb in H3. 
    rewrite negb_andb.
    apply orb_true_intro.
    case (orb_prop (negb (hedge_in_synthesis g m1 n1)) (negb (hedge_in_synthesis g m2 n2)) H3);intro;[left | right].
    case (Bool_elim_bool (hedge_in_synthesis h m1 n1));intro H6;rewrite H6;try reflexivity.
    generalize (hedge_in_synthesis_impl_hedge_synthesis_spec h m1 n1 H6);intro.
    generalize (H0 m1 n1 H7);intro.
    rewrite (hedge_synthesis_spec_impl_hedge_in_synthesis g m1 n1 H8) in H5;discriminate H5.
    case (Bool_elim_bool (hedge_in_synthesis h m2 n2));intro H6;rewrite H6;try reflexivity.
    generalize (hedge_in_synthesis_impl_hedge_synthesis_spec h m2 n2 H6);intro.
    generalize (H0 m2 n2 H7);intro. 
    rewrite (hedge_synthesis_spec_impl_hedge_in_synthesis g m2 n2 H8) in H5;discriminate H5.
    destruct n;apply (H.filter_3 (is_synthesisable_compat_bool h) H4);try reflexivity.
    unfold is_synthesisable.
    rewrite negb_andb in H3. 
    rewrite negb_andb.
    apply orb_true_intro.
    case (orb_prop (negb (hedge_in_synthesis g m1 n1)) (negb (hedge_in_synthesis g m2 n2)) H3);intro;[left | right].
    case (Bool_elim_bool (hedge_in_synthesis h m1 n1));intro H6;rewrite H6;try reflexivity.
    generalize (hedge_in_synthesis_impl_hedge_synthesis_spec h m1 n1 H6);intro.
    generalize (H0 m1 n1 H7);intro.
    rewrite (hedge_synthesis_spec_impl_hedge_in_synthesis g m1 n1 H8) in H5;discriminate H5.
    case (Bool_elim_bool (hedge_in_synthesis h m2 n2));intro H6;rewrite H6;try reflexivity.
    generalize (hedge_in_synthesis_impl_hedge_synthesis_spec h m2 n2 H6);intro.
    generalize (H0 m2 n2 H7);intro. 
    rewrite (hedge_synthesis_spec_impl_hedge_in_synthesis g m2 n2 H8) in H5;discriminate H5.
    destruct n;apply (H.filter_3 (is_synthesisable_compat_bool h) H4);try reflexivity.
    unfold is_synthesisable.
    generalize H3;case (OpSpi.op_eq_dec o o0);intros;try reflexivity.
    case (Bool_elim_bool (hedge_in_synthesis h m n));intro H6;rewrite H6;try reflexivity.
    generalize (hedge_in_synthesis_impl_hedge_synthesis_spec h m n H6);intro.
    generalize (H0 m n H7);intro.
    rewrite (hedge_synthesis_spec_impl_hedge_in_synthesis g m n H8) in H5;discriminate.
  Qed.

  (** if g = h then reduce(g) = reduce(h) *)
  Lemma Equal_reduce_Equal : forall (g h:hedge),(H.Equal g h)->(H.Equal (hedge_reduce g) (hedge_reduce h)).
    intros.
    apply HedgeProp.subset_antisym.
    apply Subset_le_h_reduce_Subset.
    apply HedgeProp.subset_equal;trivial.
    apply incl_impl_le_h.
    apply Subset_impl_inclusion.
    apply HedgeProp.subset_equal.
    apply HedgeProp.equal_sym;trivial.
    apply Subset_le_h_reduce_Subset.
    apply HedgeProp.subset_equal.
    apply HedgeProp.equal_sym;auto.
    apply incl_impl_le_h.
    apply Subset_impl_inclusion.
    apply HedgeProp.subset_equal.
    trivial.
  Qed.

  (** if reduce(g) is included in h and h <= reduce(g) then reduce(g) is included in reduce(h) *)
  Lemma Subset_le_h_impl_Subset_reduce : forall (g h:hedge),(H.Subset (hedge_reduce g) h)->(le_h (relation_of_hedge h) (relation_of_hedge (hedge_reduce g)))->(H.Subset (hedge_reduce g) (hedge_reduce h)).
    intros g h.
    generalize (incl_impl_le_h (relation_of_hedge (hedge_reduce g)) (relation_of_hedge g) (Subset_impl_inclusion (hedge_reduce g) g (hedge_reduce_Subset g))).
    unfold H.Subset, le_h;intros.
    destruct a as (m,n).
    generalize (H0 (m,n) H2);intro.
    unfold hedge_reduce in H2.
    generalize (H.filter_1 (is_synthesisable_compat_bool g) H2);intro.
    generalize (H.filter_2 (is_synthesisable_compat_bool g) H2);intro.
    unfold is_synthesisable in H5.
    unfold hedge_reduce.
    destruct m.
    apply (H.filter_3 (is_synthesisable_compat_bool h) H3);reflexivity.
    destruct n;apply (H.filter_3 (is_synthesisable_compat_bool h) H3);try reflexivity.
    unfold is_synthesisable.
    rewrite negb_andb in H5.
    rewrite negb_andb.
    apply orb_true_intro.
    case (orb_prop (negb (hedge_in_synthesis g m1 n1)) (negb (hedge_in_synthesis g m2 n2)) H5);intro;[left | right].
    case (Bool_elim_bool (hedge_in_synthesis h m1 n1));intro H7;rewrite H7;try reflexivity.
    generalize (hedge_in_synthesis_impl_hedge_synthesis_spec h m1 n1 H7);intro.
    generalize (H1 m1 n1 H8);intro.
    generalize (H m1 n1 H9);intro.
    rewrite (hedge_synthesis_spec_impl_hedge_in_synthesis g m1 n1) in H6;trivial.
    case (Bool_elim_bool (hedge_in_synthesis h m2 n2));intro H7;rewrite H7;try reflexivity.
    generalize (hedge_in_synthesis_impl_hedge_synthesis_spec h m2 n2 H7);intro.
    generalize (H1 m2 n2 H8);intro.
    generalize (H m2 n2 H9);intro.
    rewrite (hedge_synthesis_spec_impl_hedge_in_synthesis g m2 n2) in H6;trivial.
    destruct n;apply (H.filter_3 (is_synthesisable_compat_bool h) H3);try reflexivity.  
    unfold is_synthesisable.
    rewrite negb_andb in H5.
    rewrite negb_andb.
    apply orb_true_intro.
    case (orb_prop (negb (hedge_in_synthesis g m1 n1)) (negb (hedge_in_synthesis g m2 n2)) H5);intro;[left | right].
    case (Bool_elim_bool (hedge_in_synthesis h m1 n1));intro H7;rewrite H7;try reflexivity.
    generalize (hedge_in_synthesis_impl_hedge_synthesis_spec h m1 n1 H7);intro.
    generalize (H1 m1 n1 H8);intro.
    generalize (H m1 n1 H9);intro.
    rewrite (hedge_synthesis_spec_impl_hedge_in_synthesis g m1 n1) in H6;trivial.
    case (Bool_elim_bool (hedge_in_synthesis h m2 n2));intro H7;rewrite H7;try reflexivity.
    generalize (hedge_in_synthesis_impl_hedge_synthesis_spec h m2 n2 H7);intro.
    generalize (H1 m2 n2 H8);intro.
    generalize (H m2 n2 H9);intro.
    rewrite (hedge_synthesis_spec_impl_hedge_in_synthesis g m2 n2) in H6;trivial.
    destruct n;apply (H.filter_3 (is_synthesisable_compat_bool h) H3);try reflexivity.  
    unfold is_synthesisable.
    rewrite negb_andb in H5.
    rewrite negb_andb.
    apply orb_true_intro.
    case (orb_prop (negb (hedge_in_synthesis g m1 n1)) (negb (hedge_in_synthesis g m2 n2)) H5);intro;[left | right].
    case (Bool_elim_bool (hedge_in_synthesis h m1 n1));intro H7;rewrite H7;try reflexivity.
    generalize (hedge_in_synthesis_impl_hedge_synthesis_spec h m1 n1 H7);intro.
    generalize (H1 m1 n1 H8);intro.
    generalize (H m1 n1 H9);intro.
    rewrite (hedge_synthesis_spec_impl_hedge_in_synthesis g m1 n1) in H6;trivial.
    case (Bool_elim_bool (hedge_in_synthesis h m2 n2));intro H7;rewrite H7;try reflexivity.
    generalize (hedge_in_synthesis_impl_hedge_synthesis_spec h m2 n2 H7);intro.
    generalize (H1 m2 n2 H8);intro.
    generalize (H m2 n2 H9);intro.
    rewrite (hedge_synthesis_spec_impl_hedge_in_synthesis g m2 n2) in H6;trivial.
    destruct n;apply (H.filter_3 (is_synthesisable_compat_bool h) H3);try reflexivity.
    unfold is_synthesisable;generalize H5.
    case OpSpi.op_eq_dec;intros;try reflexivity.
    case (Bool_elim_bool (hedge_in_synthesis h m n));intro H7;rewrite H7;try reflexivity.
    generalize (hedge_in_synthesis_impl_hedge_synthesis_spec h m n H7);intro.
    generalize (H1 m n H8);intro.
    generalize (H m n H9);intro.
    rewrite (hedge_synthesis_spec_impl_hedge_in_synthesis g m n) in H6;trivial.
  Qed.

  (** reduce(reduce(h)) = reduce(h) *)
  Lemma reduce_idempotent : forall (h:hedge),(H.Equal (hedge_reduce h) (hedge_reduce (hedge_reduce h))).
    intros.
    apply HedgeProp.subset_antisym.
    apply Subset_le_h_impl_Subset_reduce.
    unfold H.Subset;auto.
    apply le_h_refl.
    apply hedge_reduce_Subset.
  Qed.

  (** if reduce(g) ~ reduce(h) then reduce(g) is included in h *)
  Lemma reduce_equiv_h_impl_reduce_Subset : forall (g h:hedge),(equiv_h (relation_of_hedge (hedge_reduce g)) (relation_of_hedge h))->(H.Subset (hedge_reduce g) h).
    intros.
    assert (equiv_h (relation_of_hedge g) (relation_of_hedge h)).
    eapply equiv_h_trans.
    apply (h_equiv_h_red g).
    trivial.
    unfold equiv_h in H;and_elim.
    unfold equiv_h in H0;and_elim.
    generalize (le_h_impl_incl_syn (relation_of_hedge (hedge_reduce g)) (relation_of_hedge h) H);intros.
    unfold H.Subset;intros.
    destruct a as (m,n).
    generalize (H3 m n H4);intro.
    unfold hedge_reduce in H4.
    generalize (H.filter_1 (is_synthesisable_compat_bool g) H4);intro.
    generalize (H.filter_2 (is_synthesisable_compat_bool g) H4);intro.
    unfold is_synthesisable in H7.
    destruct m.
    inversion H5;trivial.
    destruct n;inversion H5;trivial.
    rewrite negb_andb in H7.
    case (orb_prop (negb (hedge_in_synthesis g m1 n1)) (negb (hedge_in_synthesis g m2 n2)) H7);intro.
    rewrite (hedge_synthesis_spec_impl_hedge_in_synthesis g m1 n1) in H14;try (discriminate H14).
    unfold hedge_synthesis_spec;apply H2;trivial.
    rewrite (hedge_synthesis_spec_impl_hedge_in_synthesis g m2 n2) in H14;try (discriminate H14).
    unfold hedge_synthesis_spec;apply H2;trivial.
    destruct n;inversion H5;trivial.
    rewrite negb_andb in H7.
    case (orb_prop (negb (hedge_in_synthesis g m1 n1)) (negb (hedge_in_synthesis g m2 n2)) H7);intro.
    rewrite (hedge_synthesis_spec_impl_hedge_in_synthesis g m1 n1) in H14;try (discriminate H14).
    unfold hedge_synthesis_spec;apply H2;trivial.
    rewrite (hedge_synthesis_spec_impl_hedge_in_synthesis g m2 n2) in H14;try (discriminate H14).
    unfold hedge_synthesis_spec;apply H2;trivial.
    destruct n;inversion H5;trivial.
    rewrite negb_andb in H7.
    case (orb_prop (negb (hedge_in_synthesis g m1 n1)) (negb (hedge_in_synthesis g m2 n2)) H7);intro.
    rewrite (hedge_synthesis_spec_impl_hedge_in_synthesis g m1 n1) in H14;try (discriminate H14).
    unfold hedge_synthesis_spec;apply H2;trivial.
    rewrite (hedge_synthesis_spec_impl_hedge_in_synthesis g m2 n2) in H14;try (discriminate H14).
    unfold hedge_synthesis_spec;apply H2;trivial.
    destruct n;inversion H5;trivial.
    subst o0.
    generalize H7.
    case OpSpi.op_eq_dec;intro H20;try (elim H20;trivial;fail);clear H20;intro.
    rewrite (hedge_synthesis_spec_impl_hedge_in_synthesis g m n) in H11;try (discriminate H11).
    unfold hedge_synthesis_spec;apply H2;trivial.
  Qed.

  (** if g <= h then reduce(g) <= reduce(h) *)
  Lemma reduce_stable_le_h : forall (g h:hedge),(le_h (relation_of_hedge g) (relation_of_hedge h))->(le_h (relation_of_hedge (hedge_reduce g)) (relation_of_hedge (hedge_reduce h))).
    intros.
    apply le_h_trans with (relation_of_hedge h).
    apply le_h_trans with (relation_of_hedge g);trivial.
    apply incl_impl_le_h.
    apply Subset_impl_inclusion.
    apply hedge_reduce_Subset.
    apply h_le_h_reduce.
  Qed.

  (** h <= A(h) *)
  Theorem h_le_h_ana_h : forall (h:hedge),(le_h (relation_of_hedge h) (relation_of_hedge (hedge_analysis h))).
    intro.
    apply incl_impl_le_h.
    apply Subset_impl_inclusion.
    generalize (hedge_analysis_is_analysis h);intro.
    unfold is_analysis in H.
    unfold analysis_cond in H;and_elim.
    trivial.
  Qed.

  (** I(h) = reduce(A(h)) *)
  Definition hedge_irreducible (h:hedge) := (hedge_reduce (hedge_analysis h)).

  Lemma irreducible_enc_s : forall (h:hedge)(m n k l:MessageSpi.message),(H.In (MessageSpi.MEnc_s m k, MessageSpi.MEnc_s n l) (hedge_irreducible h))<->((H.In (MessageSpi.MEnc_s m k, MessageSpi.MEnc_s n l) (hedge_analysis h))/\(~(hedge_synthesis_spec (hedge_analysis h) m n) \/ ~(hedge_synthesis_spec (hedge_analysis h) k l))).
    unfold hedge_irreducible.
    unfold hedge_reduce.
    split;intros.
    generalize (H.filter_1 (is_synthesisable_compat_bool (hedge_analysis h)) H);intro.
    generalize (H.filter_2 (is_synthesisable_compat_bool (hedge_analysis h)) H);intro.
    split;trivial.
    unfold is_synthesisable in H1;rewrite negb_andb in H1.
    case (orb_prop (negb (hedge_in_synthesis (hedge_analysis h) m n)) (negb (hedge_in_synthesis (hedge_analysis h) k l)) H1);intro;[left | right];intro;rewrite hedge_synthesis_spec_impl_hedge_in_synthesis in H2;try (discriminate H2);trivial.
    and_elim.
    apply (H.filter_3 (is_synthesisable_compat_bool (hedge_analysis h)) H).
    unfold is_synthesisable.
    rewrite negb_andb.
    apply orb_true_intro.
    case H0;intro;[left | right].
    case (Bool_elim_bool (hedge_in_synthesis (hedge_analysis h) m n));intro H2;rewrite H2;try reflexivity.
    elim H1.
    apply hedge_in_synthesis_impl_hedge_synthesis_spec;trivial.
    case (Bool_elim_bool (hedge_in_synthesis (hedge_analysis h) k l));intro H2;rewrite H2;try reflexivity.
    elim H1.
    apply hedge_in_synthesis_impl_hedge_synthesis_spec;trivial.
  Qed.

  Lemma irreducible_enc_a : forall (h:hedge)(m n k l:MessageSpi.message),(H.In (MessageSpi.MEnc_a m k, MessageSpi.MEnc_a n l) (hedge_irreducible h))<->((H.In (MessageSpi.MEnc_a m k, MessageSpi.MEnc_a n l) (hedge_analysis h))/\(~(hedge_synthesis_spec (hedge_analysis h) m n) \/ ~(hedge_synthesis_spec (hedge_analysis h) k l))).
    unfold hedge_irreducible.
    unfold hedge_reduce.
    split;intros.
    generalize (H.filter_1 (is_synthesisable_compat_bool (hedge_analysis h)) H);intro.
    generalize (H.filter_2 (is_synthesisable_compat_bool (hedge_analysis h)) H);intro.
    split;trivial.
    unfold is_synthesisable in H1;rewrite negb_andb in H1.
    case (orb_prop (negb (hedge_in_synthesis (hedge_analysis h) m n)) (negb (hedge_in_synthesis (hedge_analysis h) k l)) H1);intro;[left | right];intro;rewrite hedge_synthesis_spec_impl_hedge_in_synthesis in H2;try (discriminate H2);trivial.
    and_elim.
    apply (H.filter_3 (is_synthesisable_compat_bool (hedge_analysis h)) H).
    unfold is_synthesisable.
    rewrite negb_andb.
    apply orb_true_intro.
    case H0;intro;[left | right].
    case (Bool_elim_bool (hedge_in_synthesis (hedge_analysis h) m n));intro H2;rewrite H2;try reflexivity.
    elim H1.
    apply hedge_in_synthesis_impl_hedge_synthesis_spec;trivial.
    case (Bool_elim_bool (hedge_in_synthesis (hedge_analysis h) k l));intro H2;rewrite H2;try reflexivity.
    elim H1.
    apply hedge_in_synthesis_impl_hedge_synthesis_spec;trivial.
  Qed.

  Lemma irreducible_op : forall (h:hedge)(oM oN:OpSpi.op)(m n:MessageSpi.message),(H.In (MessageSpi.MOp oM m, MessageSpi.MOp oN n) (hedge_irreducible h))<->((H.In (MessageSpi.MOp oM m, MessageSpi.MOp oN n) (hedge_analysis h))/\(~oM=oN \/ ~(hedge_synthesis_spec (hedge_analysis h) m n))).
    unfold hedge_irreducible.
    unfold hedge_reduce.
    split;intros.
    generalize (H.filter_1 (is_synthesisable_compat_bool (hedge_analysis h)) H);intro.
    generalize (H.filter_2 (is_synthesisable_compat_bool (hedge_analysis h)) H);intro.
    split;trivial.
    unfold is_synthesisable in H1.
    generalize H1.
    case OpSpi.op_eq_dec;intros.
    right;intro.
    rewrite hedge_synthesis_spec_impl_hedge_in_synthesis in H2;try (discriminate H2);trivial.
    left;trivial.
    and_elim.
    apply (H.filter_3 (is_synthesisable_compat_bool (hedge_analysis h)) H).
    unfold is_synthesisable.
    case H0;intro.
    case OpSpi.op_eq_dec;intro.
    elim H1;trivial.
    reflexivity.
    case OpSpi.op_eq_dec;intro;try reflexivity.
    case (Bool_elim_bool (hedge_in_synthesis (hedge_analysis h) m n));intro H2;rewrite H2;try reflexivity.
    elim H1.
    apply hedge_in_synthesis_impl_hedge_synthesis_spec;trivial.
  Qed.

  Lemma irreducible_pair : forall (h:hedge)(m1 m2 n1 n2:MessageSpi.message),~(H.In (MessageSpi.MPair m1 m2, MessageSpi.MPair n1 n2) (hedge_irreducible h)).
    unfold hedge_irreducible.
    intros.
    unfold hedge_reduce.
    intro.
    generalize (H.filter_1 (is_synthesisable_compat_bool (hedge_analysis h)) H);intro.
    generalize (H.filter_2 (is_synthesisable_compat_bool (hedge_analysis h)) H);intro.
    unfold is_synthesisable in H1.
    rewrite negb_andb in H1.
    generalize (hedge_analysis_is_analysis h);intro.
    unfold is_analysis in H2.
    unfold analysis_cond in H2;and_elim.
    generalize (stable_closed (hedge_analysis h) H4);intro.
    unfold stable_analysis_step in H4.
    generalize (HedgeProp.subset_equal H5);intro.
    rewrite hedge_synthesis_spec_impl_hedge_in_synthesis in H1.
    rewrite hedge_synthesis_spec_impl_hedge_in_synthesis in H1.
    simpl in H1.
    discriminate H1.
    unfold hedge_synthesis_spec.
    apply SynInc.
    apply (Subset_impl_inclusion (hedge_analysis_step (hedge_analysis h)) (hedge_analysis h) H4).
    unfold relation_of_hedge.
    apply hedge_analysis_step_spec_impl_hedge_analysis_step.
    unfold hedge_analysis_step_spec;eapply AnaSplitR.
    apply hedge_analysis_step_impl_hedge_analysis_step_spec.
    apply H6.
    apply H0.
    unfold hedge_synthesis_spec.
    apply SynInc.
    apply (Subset_impl_inclusion (hedge_analysis_step (hedge_analysis h)) (hedge_analysis h) H4).
    unfold relation_of_hedge.
    apply hedge_analysis_step_spec_impl_hedge_analysis_step.
    unfold hedge_analysis_step_spec;eapply AnaSplitL.
    apply hedge_analysis_step_impl_hedge_analysis_step_spec.
    apply H6.
    apply H0.
  Qed.

  Theorem ana_h_equiv_h_irr_h : forall (h:hedge),(equiv_h (relation_of_hedge (hedge_analysis h)) (relation_of_hedge (hedge_irreducible h))).
    unfold hedge_irreducible.
    intros.
    apply h_equiv_h_red.
  Qed.

  Theorem h_le_h_irr_h : forall (h:hedge),(le_h (relation_of_hedge h) (relation_of_hedge (hedge_irreducible h))).
    intros.
    eapply le_h_trans with (relation_of_hedge (hedge_analysis h)).
    apply h_le_h_ana_h.
    elim (ana_h_equiv_h_irr_h h);intros.
    trivial.
  Qed.

  Lemma ana_irr_Subset_ana : forall (h:hedge),(H.Subset (hedge_analysis (hedge_irreducible h)) (hedge_analysis h)).
    intro.
    generalize (hedge_analysis_is_analysis (hedge_irreducible h)).
    unfold is_analysis;unfold analysis_cond;intros;and_elim.
    apply H0.
    split.
    unfold hedge_irreducible.
    apply hedge_reduce_Subset.
    generalize (hedge_analysis_is_analysis h);unfold is_analysis;unfold analysis_cond;tauto.
  Qed.

  Lemma irr_irr_Subset_irr : forall (h:hedge),(H.Subset (hedge_irreducible (hedge_irreducible h)) (hedge_irreducible h)).
    intros.
    unfold hedge_irreducible.
    apply Subset_le_h_reduce_Subset;fold (hedge_irreducible h).
    apply ana_irr_Subset_ana.
    elim (ana_h_equiv_h_irr_h h);intros.
    eapply le_h_trans.
    apply H.
    unfold hedge_irreducible.
    apply h_le_h_ana_h.
  Qed.

  Lemma irr_Subset_irr_irr : forall (h:hedge),(H.Subset (hedge_irreducible h) (hedge_irreducible (hedge_irreducible h))).
    intros.
    unfold hedge_irreducible.
    apply Subset_le_h_impl_Subset_reduce.
    fold (hedge_irreducible h).
    generalize (hedge_analysis_is_analysis (hedge_irreducible h)).
    unfold is_analysis;unfold analysis_cond;intros;and_elim.
    apply H.
    eapply le_h_trans with (relation_of_hedge (hedge_analysis h)).
    apply incl_impl_le_h.
    fold (hedge_irreducible h).
    apply Subset_impl_inclusion.
    apply ana_irr_Subset_ana.
    fold (hedge_irreducible h).
    elim (ana_h_equiv_h_irr_h h);auto.
  Qed.

  Definition is_irreducible (h:hedge) := H.Equal h (hedge_irreducible h).

  Theorem hedge_irreducible_is_irreducible : forall (h:hedge),(is_irreducible (hedge_irreducible h)).
    intros.
    unfold is_irreducible.
    apply HedgeProp.subset_antisym.
    apply irr_Subset_irr_irr.
    apply irr_irr_Subset_irr.
  Qed.

  Theorem equiv_h_irr_impl_Subset : forall (g h:hedge),(equiv_h (relation_of_hedge g) (relation_of_hedge h))->(is_irreducible g)->(H.Subset g h).
    unfold is_irreducible.
    unfold hedge_irreducible.
    intros.
    eapply HedgeProp.subset_trans with (hedge_irreducible g).
    apply HedgeProp.subset_equal.
    apply H0.
    unfold hedge_irreducible.
    apply reduce_equiv_h_impl_reduce_Subset.
    eapply equiv_h_trans.
    apply equiv_h_sym.
    apply same_impl_equiv_h.
    apply Equal_impl_same.
    apply H0.
    trivial.
  Qed.

  Theorem equiv_h_irr_irr_impl_Equal : forall (g h:hedge),(equiv_h (relation_of_hedge g) (relation_of_hedge h))->(is_irreducible g)->(is_irreducible h)->(H.Equal g h).
    intros.
    apply HedgeProp.subset_antisym.
    apply equiv_h_irr_impl_Subset;trivial.
    apply equiv_h_irr_impl_Subset;trivial.
    apply equiv_h_sym;trivial.
  Qed.

  Lemma h_le_h_analysis_step_spec : forall (h:message_relation),(le_h h (hedge_analysis_relation h)).
    unfold le_h;intros.
    induction H.
    apply SynInc.
    apply AnaInc;trivial.
    apply SynEnc_s;trivial.
    apply SynEnc_a;trivial.
    apply SynPair;trivial.
    apply SynOp;trivial.
  Qed.

  Lemma le_h_ana_step_spec_le_h_irr : forall (g h:hedge),(le_h (relation_of_hedge g) (relation_of_hedge h))->(le_h (hedge_analysis_step_spec g) (relation_of_hedge (hedge_irreducible h))).
    intros.
    apply incl_syn_impl_le_h.
    intros.
    induction H0.
    apply h_le_h_irr_h.
    apply H.
    apply SynInc;trivial.
    inversion IHhedge_analysis_relation;trivial.
    unfold relation_of_hedge in H1.
    elim (irreducible_pair h m1 m2 n1 n2 H1).
    inversion IHhedge_analysis_relation;trivial.
    unfold relation_of_hedge in H1.
    elim (irreducible_pair h m1 m2 n1 n2 H1).
    inversion IHhedge_analysis_relation;trivial.
    elim (irreducible_enc_s h m n k l);intros.
    unfold relation_of_hedge in H2.
    elim (H5 H2);clear H5;clear H6;intros.
    generalize (H k l H1);intro.
    fold (hedge_synthesis_spec h) in H7.
    assert (H.In (m,n) (hedge_analysis h)).
    generalize (hedge_analysis_is_analysis h);intros.
    unfold is_analysis in H8;and_elim.
    unfold analysis_cond in H8;and_elim.
    unfold stable_analysis_step in H10.
    apply H10.
    apply hedge_analysis_step_spec_impl_hedge_analysis_step.
    unfold hedge_analysis_step_spec.
    eapply AnaDec_s.
    apply AnaInc.
    unfold relation_of_hedge.
    apply H5.
    apply h_le_h_ana_h.
    fold (hedge_synthesis_spec h).
    trivial.
    elim (ana_h_equiv_h_irr_h h);intros.
    apply H9.
    apply SynInc.
    unfold relation_of_hedge.
    trivial.
    inversion IHhedge_analysis_relation;trivial.
    elim (irreducible_enc_a h m n k l);intros.
    unfold relation_of_hedge in H3.
    elim (H6 H3);clear H6;clear H7;intros.
    generalize (H k' l' H2);intro.
    fold (hedge_synthesis_spec h) in H8.
    assert (H.In (m,n) (hedge_analysis h)).
    generalize (hedge_analysis_is_analysis h);intros.
    unfold is_analysis in H9;and_elim.
    unfold analysis_cond in H9;and_elim.
    unfold stable_analysis_step in H11.
    apply H11.
    apply hedge_analysis_step_spec_impl_hedge_analysis_step.
    unfold hedge_analysis_step_spec.
    eapply AnaDec_a.
    apply AnaInc.
    unfold relation_of_hedge.
    apply H6.
    apply H1.
    apply h_le_h_ana_h.
    fold (hedge_synthesis_spec h).
    trivial.
    elim (ana_h_equiv_h_irr_h h);intros.
    apply H10.
    apply SynInc.
    unfold relation_of_hedge.
    trivial.
  Qed.

  Lemma le_h_ana_step_le_h_irr : forall (g h:hedge),(le_h (relation_of_hedge g) (relation_of_hedge h))->(le_h (relation_of_hedge (hedge_analysis_step g)) (relation_of_hedge (hedge_irreducible h))).
    intros.
    eapply le_h_trans.
    elim (same_impl_equiv_h (relation_of_hedge (hedge_analysis_step g)) (hedge_analysis_step_spec g));intros.
    apply H0.
    unfold same_relation;unfold inclusion;split;intros.
    unfold relation_of_hedge in H0.
    apply hedge_analysis_step_impl_hedge_analysis_step_spec;trivial.
    unfold relation_of_hedge.
    apply hedge_analysis_step_spec_impl_hedge_analysis_step;trivial.
    apply le_h_ana_step_spec_le_h_irr;trivial.
  Qed.

  Lemma le_h_ana_step_iter_le_h_irr : forall (n:nat)(g h:hedge),(le_h (relation_of_hedge g) (relation_of_hedge h))->(le_h (relation_of_hedge (hedge_analysis_step_iter g n)) (relation_of_hedge (hedge_irreducible h))).
    induction n;intros.
    simpl.
    eapply le_h_trans.
    apply H.
    apply h_le_h_irr_h.
    generalize (Equal_impl_same (hedge_analysis_step_iter g (S n)) (hedge_analysis_step (hedge_analysis_step_iter g n)) (hedge_analysis_step_iter_Sn n g));intro.
    elim (same_impl_equiv_h (relation_of_hedge (hedge_analysis_step_iter g (S n))) (relation_of_hedge (hedge_analysis_step (hedge_analysis_step_iter g n))) H0);intros.
    eapply le_h_trans.
    apply H1.
    generalize (hedge_irreducible_is_irreducible h);unfold is_irreducible;intros.
    generalize (Equal_impl_same (hedge_irreducible h) (hedge_irreducible (hedge_irreducible h)) H3);intro.
    elim (same_impl_equiv_h (relation_of_hedge (hedge_irreducible h)) (relation_of_hedge (hedge_irreducible (hedge_irreducible h))) H4);intros.
    eapply le_h_trans.
    apply le_h_ana_step_le_h_irr.
    apply IHn.
    apply H.
    trivial.
  Qed.

  Lemma le_h_ana_le_h : forall (g h:hedge),(le_h (relation_of_hedge g) (relation_of_hedge h))->(le_h (relation_of_hedge (hedge_analysis g)) (relation_of_hedge (hedge_analysis h))).
    intros.
    elim (ana_h_equiv_h_irr_h h);intros.
    apply le_h_trans with (relation_of_hedge (hedge_irreducible h));trivial.
    elim (hedge_analysis_is_iter g);intro n;intro H2.
    generalize (Equal_impl_same (hedge_analysis g) (hedge_analysis_step_iter g n) H2);intro.
    elim (same_impl_equiv_h (relation_of_hedge (hedge_analysis g)) (relation_of_hedge (hedge_analysis_step_iter g n)) H3);intros.
    eapply le_h_trans.
    apply H4.
    apply le_h_ana_step_iter_le_h_irr.
    trivial.
  Qed.

  Theorem le_h_irr_le_h : forall (g h:hedge),(le_h (relation_of_hedge g) (relation_of_hedge h))->(le_h (relation_of_hedge (hedge_irreducible g)) (relation_of_hedge (hedge_irreducible h))).
    intros.
    elim (ana_h_equiv_h_irr_h g);intros.
    elim (ana_h_equiv_h_irr_h h);intros.
    eapply le_h_trans.
    apply H1.
    eapply le_h_trans.
    apply le_h_ana_le_h;eauto.
    trivial.
  Qed.

  Theorem equiv_h_impl_irr_unique : forall (g h:hedge),(equiv_h (relation_of_hedge g) (relation_of_hedge h))->(H.Equal (hedge_irreducible g) (hedge_irreducible h)).
    unfold equiv_h;intros;and_elim.
    generalize (hedge_irreducible_is_irreducible g);intro.
    generalize (hedge_irreducible_is_irreducible h);intro.
    assert (equiv_h (relation_of_hedge (hedge_irreducible g)) (relation_of_hedge (hedge_irreducible h))).
    split;apply le_h_irr_le_h;trivial.
    apply HedgeProp.subset_antisym.
    apply equiv_h_irr_impl_Subset;trivial.
    apply equiv_h_irr_impl_Subset;trivial.
    apply equiv_h_sym;trivial.
  Qed.

  Theorem update_irr : forall (g h:hedge),(H.Equal (hedge_irreducible (H.union (hedge_irreducible h) g)) (hedge_irreducible (H.union h g))).
    intros.
    apply equiv_h_irr_irr_impl_Equal;try (apply hedge_irreducible_is_irreducible).
    unfold equiv_h;split.
    apply le_h_trans with (relation_of_hedge (hedge_irreducible (hedge_irreducible (H.union h g)))).
    apply le_h_irr_le_h.
    elim (same_impl_equiv_h (relation_of_hedge (H.union (hedge_irreducible h) g)) (union MessageSpi.message (relation_of_hedge (hedge_irreducible h)) (relation_of_hedge g)) (same_union_union (hedge_irreducible h) g));intros.
    eapply le_h_trans.
    apply H.
    apply le_h_le_h_impl_union_le_h.
    apply le_h_irr_le_h.
    apply incl_impl_le_h.
    unfold relation_of_hedge;unfold inclusion;intros.
    apply H.union_2;trivial.
    eapply le_h_trans with (relation_of_hedge (hedge_irreducible g)).
    apply h_le_h_irr_h.
    apply le_h_irr_le_h.
    apply incl_impl_le_h.
    unfold relation_of_hedge;unfold inclusion;intros.
    apply H.union_3;trivial.
    generalize (hedge_irreducible_is_irreducible (H.union h g));intro.
    unfold is_irreducible in H.
    generalize (Equal_impl_same (hedge_irreducible (H.union h g)) (hedge_irreducible (hedge_irreducible (H.union h g))) H);intro.
    elim (same_impl_equiv_h (relation_of_hedge (hedge_irreducible (H.union h g))) (relation_of_hedge (hedge_irreducible (hedge_irreducible (H.union h g)))) H0);intros.
    eapply le_h_trans.
    apply H2.
    apply le_h_refl.
    apply le_h_irr_le_h.
    elim (same_impl_equiv_h (relation_of_hedge (H.union h g)) (union MessageSpi.message (relation_of_hedge h) (relation_of_hedge g)));intros.
    eapply le_h_trans.
    apply H.
    elim (same_impl_equiv_h (relation_of_hedge (H.union (hedge_irreducible h) g)) (union MessageSpi.message (relation_of_hedge (hedge_irreducible h)) (relation_of_hedge g)));intros.
    apply le_h_trans with (union MessageSpi.message (relation_of_hedge (hedge_irreducible h)) (relation_of_hedge g));trivial.
    apply le_h_le_h_impl_le_h_union.
    apply h_le_h_irr_h.
    apply le_h_refl.
    apply same_union_union.
    apply same_union_union.
  Qed.

  Definition name_left_name (h:message_relation) := forall (m n:MessageSpi.message),(h m n)->forall (a:Name.t),(m=MessageSpi.MNam a)->(exists b:Name.t, n=(MessageSpi.MNam b)).

  Lemma inclusion_impl_name_left_name_impl : forall (g h:message_relation),(inclusion MessageSpi.message g h)->name_left_name h->name_left_name g.
    unfold name_left_name;intros.
    eapply (H0 m n (H m n H1));eauto.
  Qed.

  Definition injective_left (h:message_relation) := forall (m n:MessageSpi.message),(h m n)->forall (m' n':MessageSpi.message),(h m' n')->(m=m')->(n=n').

  Lemma inclusion_impl_injective_left_impl : forall (g h:message_relation),(inclusion MessageSpi.message g h)->(injective_left h)->(injective_left g).
    unfold injective_left;intros.
    eapply H0;eauto.
  Qed.

  Definition op_synthesis_left (h:message_relation) := forall (m n:MessageSpi.message),(h m n)->forall (o:OpSpi.op)(m':MessageSpi.message),(m=MessageSpi.MOp o m')->forall (n':MessageSpi.message),~(hedge_synthesis_relation h m' n').

  Lemma inclusion_impl_op_synthesis_left_impl : forall (g h:message_relation),(inclusion MessageSpi.message g h)->(op_synthesis_left h)->(op_synthesis_left g).
    unfold op_synthesis_left;intros.
    intro.
    generalize (incl_impl_le_h g h H m' n' H3).
    fold (~hedge_synthesis_relation h m' n').
    eapply H0;eauto.
  Qed.
    
  Definition enc_s_synthesis_left (h:message_relation) := forall (m n:MessageSpi.message),(h m n)->forall (m' k:MessageSpi.message),(m=MessageSpi.MEnc_s m' k)->(forall (n':MessageSpi.message),~(hedge_synthesis_relation h m' n'))\/(forall (l:MessageSpi.message),~(hedge_synthesis_relation h k l)).

  Lemma inclusion_impl_enc_s_synthesis_left_impl : forall (g h:message_relation),(inclusion MessageSpi.message g h)->(enc_s_synthesis_left h)->(enc_s_synthesis_left g).
    unfold enc_s_synthesis_left;intros.
    generalize (incl_impl_le_h g h H);intros.
    case (H0 m n (H m n H1) m' k H2);[left | right];intros;intro.
    elim (H4 n');auto.
    elim (H4 l);auto.
  Qed.

  Definition enc_s_analysis_left (h:message_relation) := forall (m n:MessageSpi.message),(h m n)->forall (m' k:MessageSpi.message),m=(MessageSpi.MEnc_s m' k)->forall (l:MessageSpi.message),(hedge_synthesis_relation h k l)->(exists n':MessageSpi.message,n=(MessageSpi.MEnc_s n' l)/\(hedge_synthesis_relation h m' n')).

  Lemma same_impl_enc_s_analysis_left_impl : forall (g h:message_relation),(same_relation MessageSpi.message g h)->(enc_s_analysis_left g)->(enc_s_analysis_left h).
    unfold enc_s_analysis_left;intros.
    elim (same_impl_equiv_h g h H);intros.
    elim H;intros.
    elim (H0 m n (H7 m n H1) m' k H2 l (H5 k l H3)).
    intro n';intro;and_elim.
    exists n'.
    auto.
  Qed.

  Definition enc_s_not_ana_left (h:message_relation) := forall (m n:MessageSpi.message),(h m n)->forall (m' k:MessageSpi.message),m=(MessageSpi.MEnc_s m' k)->forall (l:MessageSpi.message),~(hedge_synthesis_relation h k l).

  Lemma enc_s_syn_ana_iff_enc_s_not_ana : forall (h:message_relation),(enc_s_synthesis_left h)/\(enc_s_analysis_left h)<->(enc_s_not_ana_left h).
    split;intros;and_elim.
    unfold enc_s_not_ana_left.
    intros.
    unfold enc_s_synthesis_left in H.
    unfold enc_s_analysis_left in H0.
    case (H m n H1 m' k H2);intros;auto.
    red;intro.
    elim (H0 m n H1 m' k H2 l H4).
    intro n';intros.
    and_elim.
    elim (H3 n');auto.
    unfold enc_s_not_ana_left in H.
    unfold enc_s_synthesis_left.
    unfold enc_s_analysis_left.
    split;intros.
    right;intros.
    eapply H;eauto.
    elim (H m n H0 m' k H1 l).
    auto.
  Qed.

  Lemma same_impl_enc_s_not_ana_left_impl : forall (g h:message_relation),(same_relation MessageSpi.message g h)->(enc_s_not_ana_left g)->(enc_s_not_ana_left h).
    unfold enc_s_not_ana_left;intros.
    elim (same_impl_equiv_h g h H);intros.
    elim H;intros.
    red;intro.
    apply (H0 m n (H6 m n H1) m' k H2 l).
    apply H4;auto.
  Qed.

  Definition enc_a_synthesis_left (h:message_relation) := forall (m n:MessageSpi.message),(h m n)->forall (m' k:MessageSpi.message),(m=MessageSpi.MEnc_a m' k)->(forall (n':MessageSpi.message),~(hedge_synthesis_relation h m' n'))\/(forall (l:MessageSpi.message),~(hedge_synthesis_relation h k l)).

  Lemma inclusion_impl_enc_a_synthesis_left_impl : forall (g h:message_relation),(inclusion MessageSpi.message g h)->(enc_a_synthesis_left h)->(enc_a_synthesis_left g).
    unfold enc_a_synthesis_left;intros.
    generalize (incl_impl_le_h g h H);intros.
    case (H0 m n (H m n H1) m' k H2);[left | right];intros;intro.
    elim (H4 n');auto.
    elim (H4 l);auto.
  Qed.

  Definition enc_a_analysis_left (h:message_relation) := forall (m n:MessageSpi.message),(h m n)->forall (m' k:MessageSpi.message),m=(MessageSpi.MEnc_a m' k)->forall (k':MessageSpi.message),(MessageSpi.inverse_key k = Some k')->forall (l':MessageSpi.message),(hedge_synthesis_relation h k' l')->(exists l:MessageSpi.message,exists n':MessageSpi.message,MessageSpi.inverse_key l = Some l' /\ n=(MessageSpi.MEnc_a n' l)/\(hedge_synthesis_relation h m' n')).

  Lemma same_impl_enc_a_analysis_left_impl : forall (g h:message_relation),(same_relation MessageSpi.message g h)->(enc_a_analysis_left g)->(enc_a_analysis_left h).
    unfold enc_a_analysis_left;intros.
    elim (same_impl_equiv_h g h H);intros.
    elim H;intros.
    elim (H0 m n (H8 m n H1) m' k H2 k' H3 l' (H6 k' l' H4)).
    intro l;intro.
    ex_elim H9 n'.
    exists l.
    exists n'.
    split;trivial.
    split;trivial.
    auto.
  Qed.

  Definition pair_free_left (h:message_relation) := forall (m1 m2 n:MessageSpi.message),~(h (MessageSpi.MPair m1 m2) n).

  Lemma inclusion_impl_pair_free_left_impl : forall (g h:message_relation),(inclusion MessageSpi.message g h)->(pair_free_left h)->(pair_free_left g).
    unfold pair_free_left;intros.
    intro.
    apply (H0 m1 m2 n);auto.
  Qed.

  Definition injective_inv_left (h:message_relation) := forall (m n:MessageSpi.message),(h m n)->forall (m' n':MessageSpi.message),(h m' n')->((Some m')=MessageSpi.inverse_key m)->((Some n')=MessageSpi.inverse_key n).

  Lemma inclusion_impl_injective_inv_left_impl : forall (g h:message_relation),(inclusion MessageSpi.message g h)->(injective_inv_left h)->(injective_inv_left g).
    unfold injective_inv_left;intros.
    eapply H0;eauto.
  Qed.

  Theorem inj_left_syn : forall (h:message_relation),(injective_left h)->(enc_s_synthesis_left h)->(enc_a_synthesis_left h)->(op_synthesis_left h)->(pair_free_left h)->(injective_left (hedge_synthesis_relation h)).
    unfold injective_left, enc_s_synthesis_left, enc_a_synthesis_left, op_synthesis_left, pair_free_left;intros h H0 H1 H2 H3 H4.
    induction 1.
    induction 1;intros.
    eapply H0;eauto.
    case (H1 m n H m0 k H5);intro H6.
    elim (H6 n0);trivial.
    elim (H6 l);trivial.
    case (H2 m n H m0 k H5);intro H6.
    elim (H6 n0);trivial.
    elim (H6 l);trivial.
    elim (H4 m1 m2 n);subst m;trivial.
    elim (H3 m n H o m0 H6 n0);trivial.
    intros.
    subst m'.
    inversion H6.
    case (H1 (MessageSpi.MEnc_s m k) n' H7 m k);trivial;intro H10.
    elim (H10 n);trivial.
    elim (H10 l);trivial.
    rewrite (IHhedge_synthesis_relation1 m n0);trivial.
    rewrite (IHhedge_synthesis_relation2 k l0);trivial.
    intros.
    subst m'.
    inversion H6.
    case (H2 (MessageSpi.MEnc_a m k) n' H7 m k);trivial;intro H10.
    elim (H10 n);trivial.
    elim (H10 l);trivial.
    rewrite (IHhedge_synthesis_relation1 m n0);trivial.
    rewrite (IHhedge_synthesis_relation2 k l0);trivial.
    intros.
    subst m'.
    inversion H6.
    elim (H4 m1 m2 n');trivial.
    rewrite (IHhedge_synthesis_relation1 m1 n0);trivial.
    rewrite (IHhedge_synthesis_relation2 m2 n3);trivial.
    intros.
    subst m'.
    inversion H5.
    elim (H3 (MessageSpi.MOp o m) n' H6 o m) with n;trivial.
    rewrite (IHhedge_synthesis_relation m n0 H9);trivial.
  Qed.

  Theorem inj_inv_left_syn : forall (h:message_relation),(injective_left h)->(enc_s_synthesis_left h)->(enc_a_synthesis_left h)->(op_synthesis_left h)->(pair_free_left h)->(injective_inv_left h)->(injective_inv_left (hedge_synthesis_relation h)).
    unfold injective_left, enc_s_synthesis_left, enc_a_synthesis_left, op_synthesis_left, pair_free_left, injective_inv_left;intros h H0 H1 H2 H3 H4 H5.
    generalize (inj_left_syn h H0 H1 H2 H3 H4).
    unfold injective_left;intro H6.
    induction 1.
    induction 1;intros.
    eapply H5;eauto.
    destruct m;try (discriminate H7);destruct o;try (discriminate H7).
    destruct m;try (discriminate H7);destruct o;try (discriminate H7).
    destruct m;try (discriminate H7);destruct o;try (discriminate H7).
    destruct m;try (discriminate H8);destruct o0;try (discriminate H8);injection H8;intros;subst m0;subst o.
    elim (H3 (MessageSpi.MOp OpSpi.Pub m) n H OpSpi.Pub m (refl_equal (MessageSpi.MOp OpSpi.Pub m)) n0 H7).
    elim (H3 (MessageSpi.MOp OpSpi.Priv m) n H OpSpi.Priv m (refl_equal (MessageSpi.MOp OpSpi.Priv m)) n0 H7).
    intros.
    discriminate H9.
    intros.
    discriminate H9.
    intros.
    discriminate H9.
    intros.
    destruct o;try (discriminate H8);injection H8;intros.
    subst m'.
    inversion H7.
    elim (H3 (MessageSpi.MOp OpSpi.Priv m) n' H9 OpSpi.Priv m (refl_equal (MessageSpi.MOp OpSpi.Priv m)) n H).
    simpl.
    rewrite <- (H6 m n H m n0 H12 (refl_equal m));trivial.
    subst m'.
    inversion H7.
    elim (H3 (MessageSpi.MOp OpSpi.Pub m) n' H9 OpSpi.Pub m (refl_equal (MessageSpi.MOp OpSpi.Pub m)) n H).
    simpl.
    rewrite <- (H6 m n H m n0 H12 (refl_equal m));trivial.
  Qed.

  Definition is_left_consistent_relation (h:message_relation) := (name_left_name h)/\(injective_left h)/\(op_synthesis_left h)/\(pair_free_left h)/\(enc_s_synthesis_left h)/\(enc_s_analysis_left h)/\(enc_a_synthesis_left h)/\(enc_a_analysis_left h)/\(injective_inv_left h).

  Definition same_impl_is_left_consistent_relation_impl : forall (g h:message_relation),(same_relation MessageSpi.message g h)->(is_left_consistent_relation g)->(is_left_consistent_relation h).
    unfold is_left_consistent_relation;intros;and_elim.
    elim H;intros.
    split.
    eapply inclusion_impl_name_left_name_impl;eauto.
    split.
    eapply inclusion_impl_injective_left_impl;eauto.
    split.
    eapply inclusion_impl_op_synthesis_left_impl;eauto.
    split.
    eapply inclusion_impl_pair_free_left_impl;eauto.
    split.
    eapply inclusion_impl_enc_s_synthesis_left_impl;eauto.
    split.
    eapply same_impl_enc_s_analysis_left_impl;eauto.
    split.
    eapply inclusion_impl_enc_a_synthesis_left_impl;eauto.
    split.
    eapply same_impl_enc_a_analysis_left_impl;eauto.
    eapply inclusion_impl_injective_inv_left_impl;eauto.
  Qed.

  Definition is_left_consistent (h:hedge) := is_left_consistent_relation (relation_of_hedge h).

  Definition is_consistent_relation (h:message_relation) := (is_left_consistent_relation h)/\(is_left_consistent_relation (transp MessageSpi.message h)).

  Lemma same_impl_is_consistent_relation_impl : forall (g h:message_relation),(same_relation MessageSpi.message g h)->(is_consistent_relation g)->(is_consistent_relation h).
    unfold is_consistent_relation;intros;and_elim.
    split.
    eapply same_impl_is_left_consistent_relation_impl;eauto.
    apply same_impl_is_left_consistent_relation_impl with (transp MessageSpi.message g);auto.
    elim H;intros.
    unfold same_relation, transp, inclusion;auto.
  Qed.

  Definition is_consistent (h:hedge) := (is_consistent_relation (relation_of_hedge h)).

  Lemma is_consistent_iff : forall (h:hedge),(is_consistent h)<->((is_left_consistent h)/\(is_left_consistent (hedge_transpose h))).
    unfold is_consistent, is_consistent_relation, is_left_consistent.
    intro h.
    generalize (same_transpose_transp h);intro.
    split;intros;and_elim;split;trivial.
    eapply same_impl_is_left_consistent_relation_impl;eauto.
    unfold same_relation;elim H;auto.
    eapply same_impl_is_left_consistent_relation_impl;eauto.
  Qed.

  Definition no_pairs (h:message_relation) := forall (m1 m2 n1 n2:MessageSpi.message),~(h (MessageSpi.MPair m1 m2) (MessageSpi.MPair n1 n2)).

  Lemma inclusion_impl_no_pairs_impl : forall (g h:message_relation),(inclusion MessageSpi.message g h)->(no_pairs h)->(no_pairs g).
    unfold no_pairs;intros.
    intro.
    elim (H0 m1 m2 n1 n2).
    apply H.
    trivial.
  Qed.

  Definition no_syn_ops (h:message_relation) := forall (o:OpSpi.op)(m n:MessageSpi.message),(h (MessageSpi.MOp o m) (MessageSpi.MOp o n))->~(hedge_synthesis_relation h m n).

  Lemma inclusion_impl_no_syn_ops_impl : forall (g h:message_relation),(inclusion MessageSpi.message g h)->(no_syn_ops h)->(no_syn_ops g).
    unfold no_syn_ops;intros.
    intro.
    elim (H0 o m n).
    apply H;trivial.
    apply (incl_impl_le_h g h);trivial.
  Qed.

  Definition no_syn_cyph_s (h:message_relation) := forall (m n k l:MessageSpi.message),(h (MessageSpi.MEnc_s m k) (MessageSpi.MEnc_s n l))->~(hedge_synthesis_relation h m n)\/~(hedge_synthesis_relation h k l).

  Lemma inclusion_impl_no_syn_cyph_s_impl : forall (g h:message_relation),(inclusion MessageSpi.message g h)->(no_syn_cyph_s h)->(no_syn_cyph_s g).
    unfold no_syn_cyph_s;intros.
    case (H0 m n k l (H (MessageSpi.MEnc_s m k) (MessageSpi.MEnc_s n l) H1));[left | right];intro;elim H2;apply (incl_impl_le_h g h);trivial.
  Qed.

  Definition cyph_s_dec (h:message_relation) := forall (m n k l:MessageSpi.message),(h (MessageSpi.MEnc_s m k) (MessageSpi.MEnc_s n l))->(hedge_synthesis_relation h k l)->(hedge_synthesis_relation h m n).

  Lemma same_impl_cyph_s_dec_impl : forall (g h:message_relation),(same_relation MessageSpi.message g h)->(cyph_s_dec g)->(cyph_s_dec h).
    unfold cyph_s_dec;intros.
    elim (same_impl_equiv_h g h H);intros.
    elim H;intros.
    apply H3.
    eapply H0.
    apply H6.
    apply H1.
    apply H4;trivial.
  Qed.

  Definition no_syn_cyph_a (h:message_relation) := forall (m n k l:MessageSpi.message),(h (MessageSpi.MEnc_a m k) (MessageSpi.MEnc_a n l))->~(hedge_synthesis_relation h m n)\/~(hedge_synthesis_relation h k l).

  Lemma inclusion_impl_no_syn_cyph_a_impl : forall (g h:message_relation),(inclusion MessageSpi.message g h)->(no_syn_cyph_a h)->(no_syn_cyph_a g).
    unfold no_syn_cyph_a;intros.
    case (H0 m n k l (H (MessageSpi.MEnc_a m k) (MessageSpi.MEnc_a n l) H1));[left | right];intro;elim H2;apply (incl_impl_le_h g h);trivial.
  Qed.

  Definition cyph_a_dec (h:message_relation) := forall (m n k l:MessageSpi.message),(h (MessageSpi.MEnc_a m k) (MessageSpi.MEnc_a n l))->forall (k' l':MessageSpi.message),(inverse_keys k l = Some (k',l'))->(hedge_synthesis_relation h k' l')->(hedge_synthesis_relation h m n).

  Lemma same_impl_cyph_a_dec_impl : forall (g h:message_relation),(same_relation MessageSpi.message g h)->(cyph_a_dec g)->(cyph_a_dec h).
    unfold cyph_a_dec;intros.
    elim (same_impl_equiv_h g h H);intros.
    elim H;intros.
    apply H4.
    eapply H0.
    apply H7.
    apply H1.
    apply H2.
    apply H5;trivial.
  Qed.

  Definition cond_irr_relation (h:message_relation) := (no_pairs h)/\(no_syn_ops h)/\(no_syn_cyph_s h)/\(cyph_s_dec h)/\(no_syn_cyph_a h)/\(cyph_a_dec h).

  Lemma same_impl_cond_irr_relation_impl : forall (g h:message_relation),(same_relation MessageSpi.message g h)->(cond_irr_relation g)->(cond_irr_relation h).
    unfold cond_irr_relation;intros;and_elim.
    elim H;intros.
    split.
    eapply inclusion_impl_no_pairs_impl;eauto.
    split.
    eapply inclusion_impl_no_syn_ops_impl;eauto.
    split.
    eapply inclusion_impl_no_syn_cyph_s_impl;eauto.
    split.
    eapply same_impl_cyph_s_dec_impl;eauto.
    split.
    eapply inclusion_impl_no_syn_cyph_a_impl;eauto.
    eapply same_impl_cyph_a_dec_impl;eauto.
  Qed.

  Definition cond_irr (h:hedge) := cond_irr_relation (relation_of_hedge h).

  Lemma is_left_consistent_impl_cond_irr_strong : forall (h:message_relation),(name_left_name h)->(injective_left h)->(op_synthesis_left h)->(pair_free_left h)->(enc_s_synthesis_left h)->(enc_s_analysis_left h)->(enc_a_synthesis_left h)->(enc_a_analysis_left h)->(cond_irr_relation h).
    unfold name_left_name, injective_left, op_synthesis_left, pair_free_left, enc_s_synthesis_left, enc_s_analysis_left, enc_a_synthesis_left, enc_a_analysis_left;intros.
    split.
    unfold no_pairs;intros.
    apply H2.
    split.
    unfold no_syn_ops;intros.
    eapply H1.
    apply H7.
    reflexivity.
    split.
    unfold no_syn_cyph_s;intros.
    case (H3 (MessageSpi.MEnc_s m k) (MessageSpi.MEnc_s n l) H7 m k);auto.
    split.
    unfold cyph_s_dec;intros.
    elim (H4 (MessageSpi.MEnc_s m k) (MessageSpi.MEnc_s n l) H7 m k) with l;trivial.
    intro n0;intros;and_elim.    
    injection H9;intro;subst n0;trivial.
    split.
    unfold no_syn_cyph_a;intros.
    case (H5 (MessageSpi.MEnc_a m k) (MessageSpi.MEnc_a n l) H7 m k);auto.
    unfold cyph_a_dec;intros.
    unfold inverse_keys in H8.
    case (Option_elim_option _ (MessageSpi.inverse_key k));intro Hinvk.
    ex_elim Hinvk k''.
    rewrite Hinvk in H8.
    case (Option_elim_option _ (MessageSpi.inverse_key l));intro Hinvl.
    ex_elim Hinvl l''.
    rewrite Hinvl in H8.
    injection H8;intros;subst k'';subst l''.
    elim (H6 (MessageSpi.MEnc_a m k) (MessageSpi.MEnc_a n l) H7 m k (refl_equal (MessageSpi.MEnc_a m k)) k' Hinvk l' H9).
    intro l'';intro.
    ex_elim H10 n'.
    auto.
    injection H11;intros.
    subst n';auto.
    rewrite Hinvl in H8;discriminate H8.
    rewrite Hinvk in H8;discriminate H8.
  Qed.

  Lemma is_left_consistent_impl_cond_irr : forall (h:message_relation),(is_left_consistent_relation h)->(cond_irr_relation h).
    unfold is_left_consistent_relation;intros;and_elim.
    apply is_left_consistent_impl_cond_irr_strong;auto.
  Qed.

  Lemma is_consistent_impl_cond_irr : forall (h:message_relation),(is_consistent_relation h)->(cond_irr_relation h).
    unfold is_consistent_relation;intros;and_elim.
    apply is_left_consistent_impl_cond_irr;trivial.
  Qed.

  Lemma cond_irr_impl_le_h_ana_step_relation : forall (h:message_relation),(cond_irr_relation h)->(le_h (hedge_analysis_relation h) h).
    unfold cond_irr_relation;intros;and_elim.
    apply incl_syn_impl_le_h;intros.
    induction H5.
    apply SynInc;trivial.
    inversion IHhedge_analysis_relation;trivial.
    elim (H m1 m2 n1 n2);trivial.
    inversion IHhedge_analysis_relation;trivial.
    elim (H m1 m2 n1 n2);trivial.
    inversion IHhedge_analysis_relation;trivial.
    eapply H2;eauto.
    inversion IHhedge_analysis_relation;trivial.
    eapply H4;eauto.
  Qed.

  Lemma same_relation_hedge_analysis_step : forall (h:hedge),(same_relation MessageSpi.message (hedge_analysis_relation (relation_of_hedge h)) (relation_of_hedge (hedge_analysis_step h))).
    intro.
    generalize (hedge_analysis_step_spec_impl_hedge_analysis_step h);intro.
    generalize (hedge_analysis_step_impl_hedge_analysis_step_spec h);intro.    
    split;unfold inclusion;intros.
    unfold relation_of_hedge.
    apply H;trivial.
    apply H0;trivial.
  Qed.
    
  Lemma cond_irr_impl_le_h_ana_step : forall (h:hedge),(cond_irr h)->(le_h (relation_of_hedge (hedge_analysis_step h)) (relation_of_hedge h)).
    unfold cond_irr;intros.
    eapply le_h_trans with (hedge_analysis_relation (relation_of_hedge h)).
    generalize (same_relation_hedge_analysis_step h);intro.
    elim (same_impl_equiv_h (hedge_analysis_relation (relation_of_hedge h)) (relation_of_hedge (hedge_analysis_step h)));intros;trivial.
    apply cond_irr_impl_le_h_ana_step_relation;trivial.
  Qed.
  
  Lemma le_h_ana_step_le_h_relation : forall (g h:message_relation),(le_h g h)->(le_h (hedge_analysis_relation g) (hedge_analysis_relation h)).
    intros.
    apply incl_syn_impl_le_h.
    unfold inclusion;intros.
    induction H0.
    apply h_le_h_analysis_step_spec.
    apply H.
    apply SynInc;trivial.
    inversion IHhedge_analysis_relation;trivial.
    apply SynInc.
    eapply AnaSplitL.
    apply H1.
    inversion IHhedge_analysis_relation;trivial.
    apply SynInc.
    eapply AnaSplitR.
    apply H1.
    inversion IHhedge_analysis_relation;trivial.
    apply SynInc.
    eapply AnaDec_s.
    apply H2.
    apply H.
    trivial.
    inversion IHhedge_analysis_relation;trivial.
    apply SynInc.
    eapply AnaDec_a.
    apply H3.
    apply H1.
    apply H.
    auto.
  Qed.

  Lemma le_h_ana_step_le_h : forall (g h:hedge),(le_h (relation_of_hedge g) (relation_of_hedge h))->(le_h (relation_of_hedge (hedge_analysis_step g)) (relation_of_hedge (hedge_analysis_step h))).
    intros.
    elim (same_impl_equiv_h (hedge_analysis_relation (relation_of_hedge g)) (relation_of_hedge (hedge_analysis_step g)) (same_relation_hedge_analysis_step g));intros.
    elim (same_impl_equiv_h (hedge_analysis_relation (relation_of_hedge h)) (relation_of_hedge (hedge_analysis_step h)) (same_relation_hedge_analysis_step h));intros.
    eapply le_h_trans.
    apply H1.
    eapply le_h_trans.
    apply le_h_ana_step_le_h_relation.
    apply H.
    trivial.
  Qed.

  Lemma cond_irr_impl_le_h_ana_step_iter : forall (n:nat)(h:hedge),(cond_irr h)->(le_h (relation_of_hedge (hedge_analysis_step_iter h n)) (relation_of_hedge h)).
    unfold cond_irr;induction n;intros.
    simpl.
    apply le_h_refl.
    generalize (Equal_impl_same (hedge_analysis_step_iter h (S n)) (hedge_analysis_step (hedge_analysis_step_iter h n)) (hedge_analysis_step_iter_Sn n h));intro.
    elim (same_impl_equiv_h (relation_of_hedge (hedge_analysis_step_iter h (S n))) (relation_of_hedge (hedge_analysis_step (hedge_analysis_step_iter h n))));intros;trivial.
    eapply le_h_trans.
    apply H1.
    generalize (le_h_ana_step_le_h (hedge_analysis_step_iter h n) h (IHn h H));intro.
    eapply le_h_trans.
    apply H3.
    apply cond_irr_impl_le_h_ana_step.
    apply H.
  Qed.

  Lemma cond_irr_impl_le_h_ana : forall (h:hedge),(cond_irr h)->(le_h (relation_of_hedge (hedge_analysis h)) (relation_of_hedge h)).
    intros.
    elim (hedge_analysis_is_iter h);intro n;intro H0.
    elim (same_impl_equiv_h (relation_of_hedge (hedge_analysis h)) (relation_of_hedge (hedge_analysis_step_iter h n)) (Equal_impl_same (hedge_analysis h) (hedge_analysis_step_iter h n) H0));intros.
    eapply le_h_trans.
    apply H1.
    apply cond_irr_impl_le_h_ana_step_iter.
    trivial.
  Qed.

  Lemma cond_irr_impl_equiv_h_irr : forall (h:hedge),(cond_irr h)->(equiv_h (relation_of_hedge h) (relation_of_hedge (hedge_irreducible h))).
    intros.
    apply equiv_h_trans with (relation_of_hedge (hedge_analysis h)).
    split.
    apply h_le_h_ana_h.
    apply cond_irr_impl_le_h_ana;trivial.
    apply ana_h_equiv_h_irr_h.
  Qed.

  Lemma cond_irr_impl_irr_Subset_h : forall (h:hedge),(cond_irr h)->(H.Subset (hedge_irreducible h) h).
    intros.
    apply equiv_h_irr_impl_Subset.
    apply equiv_h_sym.
    apply cond_irr_impl_equiv_h_irr;trivial.
    apply hedge_irreducible_is_irreducible.
  Qed.

  Lemma cond_irr_impl_stable_reduce : forall (h:hedge),(cond_irr h)->(H.Equal h (hedge_reduce h)).
    unfold cond_irr, cond_irr_relation;intros;and_elim.
    apply HedgeProp.subset_antisym;try (apply hedge_reduce_Subset);unfold H.Subset;unfold hedge_reduce;intros.
    destruct a as (m,n).
    apply (H.filter_3 (is_synthesisable_compat_bool h) H5).
    unfold is_synthesisable.
    destruct m;try reflexivity;destruct n;try reflexivity.
    elim (H m1 m2 n1 n2);trivial.
    rewrite negb_andb.
    case (H1 m1 n1 m2 n2);trivial;intro.
    case (Bool_elim_bool (hedge_in_synthesis h m1 n1));intro H7;rewrite H7;try (auto with bool).
    elim H6.
    apply hedge_in_synthesis_impl_hedge_synthesis_spec;trivial.
    case (Bool_elim_bool (hedge_in_synthesis h m2 n2));intro H7;rewrite H7;try (auto with bool).
    elim H6.
    apply hedge_in_synthesis_impl_hedge_synthesis_spec;trivial.
    rewrite negb_andb.
    case (H3 m1 n1 m2 n2);trivial;intro.
    case (Bool_elim_bool (hedge_in_synthesis h m1 n1));intro H7;rewrite H7;try (auto with bool).
    elim H6.
    apply hedge_in_synthesis_impl_hedge_synthesis_spec;trivial.
    case (Bool_elim_bool (hedge_in_synthesis h m2 n2));intro H7;rewrite H7;try (auto with bool).
    elim H6.
    apply hedge_in_synthesis_impl_hedge_synthesis_spec;trivial.
    case OpSpi.op_eq_dec;try reflexivity;intro H6;subst o0.
    generalize (H0 o m n H5);intro.
    case (Bool_elim_bool (hedge_in_synthesis h m n));intro H7;rewrite H7;try reflexivity.
    elim H6.
    apply hedge_in_synthesis_impl_hedge_synthesis_spec;trivial.
  Qed.

  Lemma cond_irr_impl_h_Subset_irr : forall (h:hedge),(cond_irr h)->(H.Subset h (hedge_irreducible h)).
    intros.
    apply HedgeProp.subset_trans with (hedge_reduce h).
    apply HedgeProp.subset_equal.
    apply cond_irr_impl_stable_reduce;trivial.
    apply reduce_equiv_h_impl_reduce_Subset.
    apply equiv_h_trans with (relation_of_hedge h).
    apply equiv_h_sym.
    apply h_equiv_h_red.
    apply cond_irr_impl_equiv_h_irr;trivial.
  Qed.

  Lemma cond_irr_impl_irreducible : forall (h:hedge),(cond_irr h)->(is_irreducible h).
    unfold is_irreducible;intros.
    apply HedgeProp.subset_antisym.
    apply cond_irr_impl_h_Subset_irr;trivial.
    apply cond_irr_impl_irr_Subset_h;trivial.
  Qed.

  Lemma irreducible_impl_cond_irr : forall (h:hedge),(is_irreducible h)->(cond_irr h).
    unfold is_irreducible;intros.
    split.
    unfold no_pairs;intros.
    intro.
    elim (H ((MessageSpi.MPair m1 m2),(MessageSpi.MPair n1 n2)));intros.
    elim (irreducible_pair h m1 m2 n1 n2).
    auto.
    split.
    unfold no_syn_ops;intros.
    elim (H ((MessageSpi.MOp o m),(MessageSpi.MOp o n)));intros.
    intro.
    elim (irreducible_op h o o m n);intros.
    elim (H4 (H1 H0));intros.
    case H7;intro.
    elim H8;trivial.
    elim H8.
    unfold hedge_synthesis_spec.
    apply h_le_h_ana_h;trivial.
    split.
    unfold no_syn_cyph_s;intros.
    elim (H ((MessageSpi.MEnc_s m k),(MessageSpi.MEnc_s n l)));intros.
    elim (irreducible_enc_s h m n k l);intros.
    elim (H3 (H1 H0));intros.
    case H6;intro H7;[left | right];intro H8;elim H7;unfold hedge_synthesis_spec;apply h_le_h_ana_h;trivial.
    split.
    unfold cyph_s_dec;intros.
    elim (H ((MessageSpi.MEnc_s m k),(MessageSpi.MEnc_s n l)));intros.
    generalize (H2 H0);clear H2;clear H3;intro.
    elim (same_impl_equiv_h (relation_of_hedge h) (relation_of_hedge (hedge_irreducible h)) (Equal_impl_same h (hedge_irreducible h) H));intros.
    apply H4.
    elim (ana_h_equiv_h_irr_h h);intros.
    apply H5.
    unfold hedge_irreducible in H2.
    unfold hedge_reduce in H2.
    generalize (H.filter_1 (is_synthesisable_compat_bool (hedge_analysis h)) H2);intro.
    generalize (H.filter_2 (is_synthesisable_compat_bool (hedge_analysis h)) H2);intro.
    generalize (hedge_analysis_is_analysis h);intro.
    unfold is_analysis in H9;and_elim.
    unfold analysis_cond in H9;and_elim.
    unfold stable_analysis_step in H11.
    apply SynInc.
    unfold relation_of_hedge.
    apply H11.
    apply hedge_analysis_step_spec_impl_hedge_analysis_step.
    unfold hedge_analysis_step_spec.
    eapply AnaDec_s.
    apply AnaInc.
    unfold relation_of_hedge.
    apply H7.
    apply H6.
    apply H3.
    trivial.
    split.
    unfold no_syn_cyph_a;intros.
    elim (H ((MessageSpi.MEnc_a m k),(MessageSpi.MEnc_a n l)));intros.
    elim (irreducible_enc_a h m n k l);intros.
    elim (H3 (H1 H0));intros.
    case H6;intro H7;[left | right];intro H8;elim H7;unfold hedge_synthesis_spec;apply h_le_h_ana_h;trivial.
    unfold cyph_a_dec;intros.
    elim (H ((MessageSpi.MEnc_a m k),(MessageSpi.MEnc_a n l)));intros.
    generalize (H3 H0);clear H3;clear H4;intro.
    elim (same_impl_equiv_h (relation_of_hedge h) (relation_of_hedge (hedge_irreducible h)) (Equal_impl_same h (hedge_irreducible h) H));intros.
    apply H5.
    elim (ana_h_equiv_h_irr_h h);intros.
    apply H6.
    unfold hedge_irreducible in H3.
    unfold hedge_reduce in H3.
    generalize (H.filter_1 (is_synthesisable_compat_bool (hedge_analysis h)) H3);intro.
    generalize (H.filter_2 (is_synthesisable_compat_bool (hedge_analysis h)) H3);intro.
    generalize (hedge_analysis_is_analysis h);intro.
    unfold is_analysis in H10;and_elim.
    unfold analysis_cond in H10;and_elim.
    unfold stable_analysis_step in H12.
    apply SynInc.
    unfold relation_of_hedge.
    apply H12.
    apply hedge_analysis_step_spec_impl_hedge_analysis_step.
    unfold hedge_analysis_step_spec.
    eapply AnaDec_a.
    apply AnaInc.
    unfold relation_of_hedge.
    apply H8.
    apply H1.
    apply H7.
    apply H4.
    trivial.
  Qed.

  Theorem irreducible_iff_cond_irr : forall (h:hedge),(cond_irr h)<->(is_irreducible h).
    split;intros.
    apply cond_irr_impl_irreducible;trivial.
    apply irreducible_impl_cond_irr;trivial.
  Qed.

  Theorem is_consistent_impl_irreducible : forall (h:hedge),(is_consistent h)->(is_irreducible h).
    intros.
    apply cond_irr_impl_irreducible.
    unfold cond_irr.
    apply is_consistent_impl_cond_irr.
    unfold is_consistent in H;trivial.
  Qed.

  Theorem equiv_h_consistent_consistent_Equal : forall (g h:hedge),(equiv_h (relation_of_hedge g) (relation_of_hedge h))->(is_consistent g)->(is_consistent h)->(H.Equal g h).
    intros.
    apply equiv_h_irr_irr_impl_Equal;trivial.
    apply is_consistent_impl_irreducible;trivial.
    apply is_consistent_impl_irreducible;trivial.
  Qed.

  Theorem irr_le_h_left_consistent_impl_left_consistent : forall (g h:hedge),(is_irreducible g)->(is_left_consistent h)->(le_h (relation_of_hedge g) (relation_of_hedge h))->(is_left_consistent g).
    unfold is_left_consistent, is_left_consistent_relation, is_irreducible;intros;and_elim.
    split.
    unfold name_left_name.
    intros.
    apply (H0 m n) with a;auto.
    generalize (H1 m n(SynInc (relation_of_hedge g) m n H10));intro.
    generalize (hedge_synthesis_spec_impl_hedge_in_synthesis h m n H12);subst m;rewrite name_left_synthesis;intro.
    unfold relation_of_hedge.
    apply H.mem_2;trivial.
    split.
    unfold injective_left;intros.
    eapply (inj_left_syn (relation_of_hedge h) H2 H5 H7 H3 H4).
    apply H1;apply SynInc;apply H10.
    apply H1;apply SynInc;apply H11.
    trivial.
    split.
    unfold op_synthesis_left;intros.
    intro.
    generalize (H1 m' n' H12);intro.
    generalize (le_h_impl_incl_syn (relation_of_hedge g) (relation_of_hedge h) H1 m n H10);intro.
    rewrite H11 in H14.
    inversion H14.
    elim (H3 (MessageSpi.MOp o m') n H15 o m') with n';trivial.
    subst o0;subst m0.
    rewrite <- (inj_left_syn (relation_of_hedge h) H2 H5 H7 H3 H4 m' n' H13 m' n0 H18) in H16;trivial.
    elim (H (m,n));intros.
    generalize (H15 H10);clear H15;clear H17;intros.
    subst m;subst n.
    elim (irreducible_op g o o m' n');intros.
    elim (H11 H15);clear H11;clear H16;intros.
    case H16;intro.
    elim H17;trivial.
    elim H17.
    unfold hedge_synthesis_spec.
    apply h_le_h_ana_h.
    trivial.
    split.
    unfold pair_free_left;intros.
    intro.
    generalize (le_h_impl_incl_syn (relation_of_hedge g) (relation_of_hedge h) H1 (MessageSpi.MPair m1 m2) n H10);intro.
    inversion H11.
    elim (H4 m1 m2 n H12).
    subst n.
    elim (H (MessageSpi.MPair m1 m2, MessageSpi.MPair n1 n2));intros.
    elim (irreducible_pair g m1 m2 n1 n2);auto.
    split.
    unfold enc_s_synthesis_left;intros.
    generalize (le_h_impl_incl_syn (relation_of_hedge g) (relation_of_hedge h) H1 m n H10);intro.
    rewrite H11 in H12.
    inversion H12.
    case (H5 (MessageSpi.MEnc_s m' k) n H13 m' k);trivial;intros;[left | right];intros;intro.
    elim (H16 n');auto.
    elim (H16 l);auto.
    subst m0;subst k0.
    elim (H (m,n));intros.
    generalize (H13 H10);clear H13;clear H14;intros.
    unfold hedge_irreducible, hedge_reduce in H13.
    generalize (H.filter_1 (is_synthesisable_compat_bool (hedge_analysis g)) H13);intro.
    generalize (H.filter_2 (is_synthesisable_compat_bool (hedge_analysis g)) H13);clear H13;intro.
    rewrite H11 in H13;rewrite <- H16 in H13;unfold is_synthesisable in H13.
    rewrite negb_andb in H13.
    elim (same_impl_equiv_h (relation_of_hedge g) (relation_of_hedge (hedge_irreducible g)) (Equal_impl_same g (hedge_irreducible g) H));intros.
    elim (ana_h_equiv_h_irr_h g);intros.
    case (orb_prop (negb (hedge_in_synthesis (hedge_analysis g) m' n0)) (negb (hedge_in_synthesis (hedge_analysis g) k l)) H13);intro;[left | right];intros;intro.
    rewrite (inj_left_syn (relation_of_hedge h) H2 H5 H7 H3 H4 m' n0 H15 m' n' (H1 m' n' H23)) in H22;trivial.
    case (Bool_elim_bool (hedge_in_synthesis (hedge_analysis g) m' n'));intro.
    rewrite H24 in H22;discriminate H22.
    rewrite hedge_synthesis_spec_impl_hedge_in_synthesis in H24;try (discriminate H24).
    unfold hedge_synthesis_spec;auto.
    rewrite (inj_left_syn (relation_of_hedge h) H2 H5 H7 H3 H4 k l H17 k l0 (H1 k l0 H23)) in H22;trivial.
    case (Bool_elim_bool (hedge_in_synthesis (hedge_analysis g) k l0));intro.
    rewrite H24 in H22;discriminate H22.
    rewrite hedge_synthesis_spec_impl_hedge_in_synthesis in H24;try (discriminate H24).
    unfold hedge_synthesis_spec;auto.
    split.
    unfold enc_s_analysis_left;intros.
    elim (same_impl_equiv_h (relation_of_hedge g) (relation_of_hedge (hedge_irreducible g)) (Equal_impl_same g (hedge_irreducible g) H));intros.
    elim (ana_h_equiv_h_irr_h g);intros.
    generalize (le_h_impl_incl_syn (relation_of_hedge g) (relation_of_hedge h) H1 m n H10);intro.
    rewrite H11 in H17.
    inversion H17.
    generalize (H1 k l H12);intro.
    elim (H6 (MessageSpi.MEnc_s m' k) n H18 m' k) with l;trivial.
    intro n';intro;and_elim.
    exists n'.
    split;trivial.
    apply H14.
    apply H15.
    apply SynInc.
    generalize (hedge_analysis_is_analysis g).
    unfold is_analysis, analysis_cond;intros;and_elim.
    unfold stable_analysis_step in H26.
    unfold relation_of_hedge.
    apply H26.
    apply hedge_analysis_step_spec_impl_hedge_analysis_step.
    unfold hedge_analysis_step_spec.
    rewrite H11 in H10.
    rewrite H22 in H10.
    eapply AnaDec_s.
    apply AnaInc.
    unfold relation_of_hedge.
    apply H24.
    apply H10.
    apply H16.
    apply H13.
    trivial.
    exists n0.
    subst m0;subst k0.
    generalize (inj_left_syn (relation_of_hedge h) H2 H5 H7 H3 H4 k l0 H22 k l (H1 k l H12) (refl_equal k));intro.
    subst l0.
    split;trivial.
    rewrite H11 in H10.
    rewrite <- H21 in H10.
    apply H14.
    apply H15.
    generalize (hedge_analysis_is_analysis g);unfold is_analysis, analysis_cond;intros;and_elim.
    unfold stable_analysis_step in H23.
    apply SynInc.
    unfold relation_of_hedge.
    apply H23.
    apply hedge_analysis_step_spec_impl_hedge_analysis_step.
    unfold hedge_analysis_step_spec.
    eapply AnaDec_s.
    apply AnaInc.
    unfold relation_of_hedge.
    apply H18.
    apply H10.
    apply H16.
    apply H13.
    trivial.
    split.
    unfold enc_a_synthesis_left;intros.
    generalize (le_h_impl_incl_syn (relation_of_hedge g) (relation_of_hedge h) H1 m n H10);intro.
    rewrite H11 in H12.
    inversion H12.
    case (H7 (MessageSpi.MEnc_a m' k) n H13 m' k);trivial;intros;[left | right];intros;intro.
    elim (H16 n');auto.
    elim (H16 l);auto.
    subst m0;subst k0.
    elim (H (m,n));intros.
    generalize (H13 H10);clear H13;clear H14;intros.
    unfold hedge_irreducible, hedge_reduce in H13.
    generalize (H.filter_1 (is_synthesisable_compat_bool (hedge_analysis g)) H13);intro.
    generalize (H.filter_2 (is_synthesisable_compat_bool (hedge_analysis g)) H13);clear H13;intro.
    rewrite H11 in H13;rewrite <- H16 in H13;unfold is_synthesisable in H13.
    rewrite negb_andb in H13.
    elim (same_impl_equiv_h (relation_of_hedge g) (relation_of_hedge (hedge_irreducible g)) (Equal_impl_same g (hedge_irreducible g) H));intros.
    elim (ana_h_equiv_h_irr_h g);intros.
    case (orb_prop (negb (hedge_in_synthesis (hedge_analysis g) m' n0)) (negb (hedge_in_synthesis (hedge_analysis g) k l)) H13);intro;[left | right];intros;intro.
    rewrite (inj_left_syn (relation_of_hedge h) H2 H5 H7 H3 H4 m' n0 H15 m' n' (H1 m' n' H23)) in H22;trivial.
    case (Bool_elim_bool (hedge_in_synthesis (hedge_analysis g) m' n'));intro.
    rewrite H24 in H22;discriminate H22.
    rewrite hedge_synthesis_spec_impl_hedge_in_synthesis in H24;try (discriminate H24).
    unfold hedge_synthesis_spec;auto.
    rewrite (inj_left_syn (relation_of_hedge h) H2 H5 H7 H3 H4 k l H17 k l0 (H1 k l0 H23)) in H22;trivial.
    case (Bool_elim_bool (hedge_in_synthesis (hedge_analysis g) k l0));intro.
    rewrite H24 in H22;discriminate H22.
    rewrite hedge_synthesis_spec_impl_hedge_in_synthesis in H24;try (discriminate H24).
    unfold hedge_synthesis_spec;auto.
    split.
    unfold enc_a_analysis_left;intros.
    elim (same_impl_equiv_h (relation_of_hedge g) (relation_of_hedge (hedge_irreducible g)) (Equal_impl_same g (hedge_irreducible g) H));intros.
    elim (ana_h_equiv_h_irr_h g);intros.
    generalize (le_h_impl_incl_syn (relation_of_hedge g) (relation_of_hedge h) H1 m n H10);intro.
    rewrite H11 in H18.
    inversion H18.
    generalize (H1 k' l' H13);intro.
    elim (H8 (MessageSpi.MEnc_a m' k) n H19 m' k (refl_equal (MessageSpi.MEnc_a m' k)) k' H12 l' H22);intro l;intro.
    ex_elim H23 n'.
    exists l.
    exists n'.
    split;trivial.
    split;trivial.
    apply H15.
    apply H16.
    apply SynInc.
    generalize (hedge_analysis_is_analysis g).
    unfold is_analysis, analysis_cond;intros;and_elim.
    unfold stable_analysis_step in H28.
    unfold relation_of_hedge.
    apply H28.
    apply hedge_analysis_step_spec_impl_hedge_analysis_step.
    unfold hedge_analysis_step_spec.
    rewrite H11 in H10.
    rewrite H24 in H10.
    eapply AnaDec_a.
    apply AnaInc.
    unfold relation_of_hedge.
    apply H26.
    apply H10.
    unfold inverse_keys.
    rewrite H12;rewrite H23.
    reflexivity.
    apply H17.
    apply H14.
    apply H13.
    exists l.
    exists n0.
    subst m0;subst k0.
    generalize (inj_inv_left_syn (relation_of_hedge h) H2 H5 H7 H3 H4 H9 k l H23 k' l' (H1 k' l' H13));intro. 
    assert (MessageSpi.inverse_key l = Some l').
    symmetry.
    apply H19.
    symmetry;auto.
    split;trivial.
    split;trivial.
    rewrite H11 in H10.
    rewrite <- H22 in H10.
    apply H15.
    apply H16.
    generalize (hedge_analysis_is_analysis g);unfold is_analysis, analysis_cond;intros;and_elim.
    unfold stable_analysis_step in H26.
    apply SynInc.
    unfold relation_of_hedge.
    apply H26.
    apply hedge_analysis_step_spec_impl_hedge_analysis_step.
    unfold hedge_analysis_step_spec.
    eapply AnaDec_a.
    apply AnaInc.
    unfold relation_of_hedge.
    apply H24.
    apply H10.
    unfold inverse_keys.
    rewrite H20.
    rewrite H12.
    reflexivity.
    apply H17.
    apply H14.
    auto.
    unfold injective_inv_left;intros.
    generalize (le_h_impl_incl_syn (relation_of_hedge g) (relation_of_hedge h) H1);intro.
    apply (inj_inv_left_syn (relation_of_hedge h) H2 H5 H7 H3 H4 H9 m n (H13 m n H10) m' n' (H13 m' n' H11));trivial.
  Qed.

  Lemma synthesis_transp_transp_synthesis : forall (h:message_relation),(same_relation MessageSpi.message (transp MessageSpi.message (hedge_synthesis_relation h)) (hedge_synthesis_relation (transp MessageSpi.message h))).
    unfold same_relation, inclusion, transp;split;intros.
    induction H.
    apply SynInc;trivial.
    apply SynEnc_s;auto.
    apply SynEnc_a;auto.
    apply SynPair;auto.
    apply SynOp;auto.
    induction H.
    apply SynInc;trivial.
    apply SynEnc_s;auto.
    apply SynEnc_a;auto.
    apply SynPair;auto.
    apply SynOp;auto.
  Qed.

  Lemma synthesis_morphism : forall (g h:message_relation),(same_relation MessageSpi.message g h)->(same_relation MessageSpi.message (hedge_synthesis_relation g) (hedge_synthesis_relation h)).
    intros.
    elim (same_impl_equiv_h g h H).
    unfold le_h.
    unfold same_relation.
    unfold inclusion.
    tauto.
  Qed.

  Lemma inverse_keys_transpose : forall (k l:MessageSpi.message)(k' l':MessageSpi.message),(inverse_keys k l = Some (k',l'))->(inverse_keys l k = Some (l',k')).
    unfold inverse_keys.
    intros.
    destruct k;try (discriminate H);destruct o;try (discriminate H);simpl in H;destruct l;try (discriminate H);destruct o;try (discriminate H);simpl in H;injection H;intros;simpl;congruence.
  Qed.

  Lemma analysis_transp_transp_analysis : forall (h:message_relation),(same_relation MessageSpi.message (transp MessageSpi.message (hedge_analysis_relation h)) (hedge_analysis_relation (transp MessageSpi.message h))).
    unfold same_relation, inclusion;split;intros.
    unfold transp in H;induction H.
    apply AnaInc;trivial.
    eapply AnaSplitL;eauto.
    eapply AnaSplitR;eauto.
    eapply AnaDec_s;eauto.
    elim (synthesis_transp_transp_synthesis h);intros.
    apply H1.
    auto.
    eapply AnaDec_a;eauto.
    apply inverse_keys_transpose.
    apply H0.
    elim (synthesis_transp_transp_synthesis h);intros.
    apply H2.
    apply H1.
    unfold transp;induction H.
    apply AnaInc;trivial.
    eapply AnaSplitL;eauto.
    eapply AnaSplitR;eauto.
    eapply AnaDec_s;eauto.
    elim (synthesis_transp_transp_synthesis h);intros.
    apply H2.
    trivial.
    eapply AnaDec_a;eauto.
    apply inverse_keys_transpose.
    apply H0.
    elim (synthesis_transp_transp_synthesis h);intros.
    apply H3.
    auto.
  Qed.

  Lemma analysis_morphism : forall (g h:message_relation),(same_relation MessageSpi.message g h)->(same_relation MessageSpi.message (hedge_analysis_relation g) (hedge_analysis_relation h)).
    unfold same_relation, inclusion;intros;and_elim;split;intros.
    induction H1.
    apply AnaInc;auto.
    eapply AnaSplitL;eauto.
    eapply AnaSplitR;eauto.
    eapply AnaDec_s;eauto.
    elim (synthesis_morphism g h);intros.
    apply H3;trivial.
    unfold same_relation;tauto.
    eapply AnaDec_a;eauto.
    elim (synthesis_morphism g h);intros.
    apply H4;trivial.
    unfold same_relation;tauto.
    induction H1.
    apply AnaInc;auto.
    eapply AnaSplitL;eauto.
    eapply AnaSplitR;eauto.
    eapply AnaDec_s;eauto.
    elim (synthesis_morphism h g);intros.
    apply H3;trivial.
    unfold same_relation;tauto.
    eapply AnaDec_a;eauto.
    elim (synthesis_morphism g h);intros.
    apply H5;trivial.
    unfold same_relation;tauto.
  Qed.

  Lemma same_relation_trans : forall (A:Set)(f g h:relation A),(same_relation A f g)->(same_relation A g h)->(same_relation A f h).
    unfold same_relation, inclusion;intros;and_elim.
    split;auto.
  Qed.

  Lemma same_relation_sym : forall (A:Set)(g h:relation A),(same_relation A g h)->(same_relation A h g).
    unfold same_relation;intros;tauto.
  Qed.

  Lemma synthesis_transpose_transp_synthesis : forall (h:hedge),(same_relation MessageSpi.message (hedge_synthesis_spec (hedge_transpose h)) (transp MessageSpi.message (hedge_synthesis_spec h))).
    unfold hedge_synthesis_spec.
    intros.
    eapply same_relation_trans.
    apply synthesis_morphism.
    apply same_transpose_transp.
    apply same_relation_sym.
    apply synthesis_transp_transp_synthesis.
  Qed.

  Lemma analysis_step_same : forall (h:hedge),(same_relation MessageSpi.message (relation_of_hedge (hedge_analysis_step h)) (hedge_analysis_step_spec h)).
    split;unfold inclusion;intros.
    apply hedge_analysis_step_impl_hedge_analysis_step_spec.
    apply H.
    unfold relation_of_hedge.
    apply hedge_analysis_step_spec_impl_hedge_analysis_step.
    trivial.
  Qed.

  Lemma same_impl_same_transp : forall (A:Set)(g h:relation A),(same_relation A g h)->(same_relation A (transp A g) (transp A h)).
    unfold transp, same_relation, inclusion;intros;and_elim;split;intros;auto.
  Qed.
    
  Lemma analysis_step_transpose_transpose_analysis_step : forall (h:hedge),(H.Equal (hedge_analysis_step (hedge_transpose h)) (hedge_transpose (hedge_analysis_step h))).
    intros.
    apply same_impl_Equal.
    eapply same_relation_trans.
    apply analysis_step_same.
    eapply same_relation_trans.
    unfold hedge_analysis_step_spec.
    apply analysis_morphism.
    apply same_transpose_transp.
    eapply same_relation_trans.
    apply same_relation_sym.
    apply analysis_transp_transp_analysis.
    apply same_relation_sym.
    eapply same_relation_trans.
    apply same_transpose_transp.
    apply same_impl_same_transp.
    fold (hedge_analysis_step_spec h).
    apply analysis_step_same.
  Qed.

  Lemma in_synthesis_transpose : forall (h:hedge)(m n:MessageSpi.message),(hedge_in_synthesis (hedge_transpose h) m n)=(hedge_in_synthesis h n m).
    intros.
    elim (synthesis_transpose_transp_synthesis h);intros.
    case (Bool_elim_bool (hedge_in_synthesis (hedge_transpose h) m n));intro H1;rewrite H1;symmetry.
    apply hedge_synthesis_spec_impl_hedge_in_synthesis.
    apply H.
    apply hedge_in_synthesis_impl_hedge_synthesis_spec.
    trivial.
    case (Bool_elim_bool (hedge_in_synthesis h n m));intro;trivial.
    generalize (hedge_in_synthesis_impl_hedge_synthesis_spec h n m H2);intro.
    rewrite (hedge_synthesis_spec_impl_hedge_in_synthesis) in H1;try (discriminate H1).
    apply H0.
    apply H3.
  Qed.

  Lemma is_synthesisable_transpose : forall (h:hedge)(m n:MessageSpi.message),(is_synthesisable (hedge_transpose h) m n)=(is_synthesisable h n m).
    unfold is_synthesisable;intros;destruct m;destruct n;try reflexivity;repeat (rewrite in_synthesis_transpose);trivial.
    case OpSpi.op_eq_dec;case OpSpi.op_eq_dec;intros;auto;elim n0;congruence.
  Qed.

  Lemma hedge_reduce_transpose_hedge_transpose_reduce : forall (h:hedge),(H.Equal (hedge_reduce (hedge_transpose h)) (hedge_transpose (hedge_reduce h))).
    unfold H.Equal;split;intros;destruct a as (m,n);elim (hedge_transpose_prop (hedge_reduce h) n m);elim (hedge_transpose_prop h n m);intros.
    generalize (H.filter_1 (is_synthesisable_compat_bool (hedge_transpose h)) H);intro.
    generalize (H.filter_2 (is_synthesisable_compat_bool (hedge_transpose h)) H);intro.
    apply H2.
    apply (H.filter_3 (is_synthesisable_compat_bool h) (H1 H4)).
    rewrite is_synthesisable_transpose in H5;trivial.
    generalize (H3 H);intro.
    generalize (H.filter_1 (is_synthesisable_compat_bool h) H4);intro.
    generalize (H.filter_2 (is_synthesisable_compat_bool h) H4);intro.
    apply (H.filter_3 (is_synthesisable_compat_bool (hedge_transpose h)) (H0 H5)).
    rewrite is_synthesisable_transpose;trivial.
  Qed.

  Lemma analysis_iter_transpose_transpose_analysis_iter : forall (n:nat)(h:hedge),(H.Equal (hedge_analysis_step_iter (hedge_transpose h) n) (hedge_transpose (hedge_analysis_step_iter h n))).
    induction n;intro.
    simpl.
    apply HedgeProp.equal_refl.
    eapply HedgeProp.equal_trans.
    apply hedge_analysis_step_iter_Sn.
    apply HedgeProp.equal_sym.
    eapply HedgeProp.equal_trans.
    apply hedge_transpose_morphism.
    apply hedge_analysis_step_iter_Sn.
    eapply HedgeProp.equal_trans.
    apply HedgeProp.equal_sym.
    apply analysis_step_transpose_transpose_analysis_step.
    apply hedge_analysis_step_morphism.
    apply HedgeProp.equal_sym;trivial.
  Qed.

  Lemma hedge_analysis_transpose_hedge_transpose_analysis : forall (h:hedge),(H.Equal (hedge_analysis (hedge_transpose h)) (hedge_transpose (hedge_analysis h))).
    intro.
    elim (hedge_analysis_is_iter (hedge_transpose h));intro n;intro.
    eapply HedgeProp.equal_trans.
    apply p.
    eapply HedgeProp.equal_trans.
    apply analysis_iter_transpose_transpose_analysis_iter.
    apply hedge_transpose_morphism.
    apply HedgeProp.equal_sym.
    generalize (hedge_analysis_is_analysis h);unfold is_analysis;intros;and_elim.
    apply HedgeProp.subset_antisym.
    apply H0.
    unfold analysis_cond.
    split.
    eapply HedgeProp.subset_trans with (hedge_analysis_step_iter h 0).
    simpl.
    apply HedgeProp.subset_equal.
    apply HedgeProp.equal_refl.
    apply hedge_analysis_step_iter_grow;auto with arith.
    unfold stable_analysis_step.
    generalize (hedge_analysis_is_analysis (hedge_transpose h));unfold is_analysis;intros;and_elim.
    elim H1;intros.
    unfold stable_analysis_step in H4.
    unfold H.Subset;intros.
    destruct a.
    elim (hedge_transpose_prop (hedge_analysis_step_iter h n) t t0);intros.
    apply H7.
    elim (analysis_iter_transpose_transpose_analysis_iter n h (t0,t));intros.
    apply H8.
    elim (p (t0,t));intros.
    apply H10.
    apply H4.
    apply (hedge_analysis_step_grow (hedge_analysis_step_iter (hedge_transpose h) n) (hedge_analysis (hedge_transpose h))).
    apply hedge_analysis_step_iter_Subset_hedge_analysis.
    elim (hedge_analysis_step_iter_Sn n h (t,t0));intros.
    elim (hedge_analysis_step_iter_Sn n (hedge_transpose h) (t0,t));intros.
    apply H14.
    elim (analysis_iter_transpose_transpose_analysis_iter (S n) h (t0,t));intros.
    apply H17.
    elim (hedge_transpose_prop (hedge_analysis_step_iter h (S n)) t t0);intros.
    apply H18.
    auto.
    apply hedge_analysis_step_iter_Subset_hedge_analysis.
  Qed.

  Lemma in_synthesis_Equal : forall (g h:hedge),(H.Equal g h)->(forall (m n:MessageSpi.message),(hedge_in_synthesis g m n)=(hedge_in_synthesis h m n)).
    intros.
    generalize (Equal_impl_same g h H);intros.
    generalize (same_impl_equiv_h (relation_of_hedge g) (relation_of_hedge h) H0);unfold equiv_h;intros;and_elim.
    case (Bool_elim_bool (hedge_in_synthesis g m n));intro H3;rewrite H3;case (Bool_elim_bool (hedge_in_synthesis h m n));intro H4;rewrite H4;trivial.
    generalize (hedge_in_synthesis_impl_hedge_synthesis_spec g m n H3);intro.
    generalize (H1 m n H5);intro.
    fold (hedge_synthesis_spec h) in H6.
    rewrite (hedge_synthesis_spec_impl_hedge_in_synthesis h m n H6) in H4;trivial.
    generalize (hedge_in_synthesis_impl_hedge_synthesis_spec h m n H4);intro.
    generalize (H2 m n H5);intro.
    fold (hedge_synthesis_spec g) in H6.
    rewrite (hedge_synthesis_spec_impl_hedge_in_synthesis g m n H6) in H3.
    discriminate H3.
  Qed.

  Lemma is_synthesisable_Equal : forall (g h:hedge),(H.Equal g h)->(forall (m n:MessageSpi.message),(is_synthesisable g m n)=(is_synthesisable h m n)).
    unfold is_synthesisable.
    intros.
    destruct m;try reflexivity;destruct n;try reflexivity;repeat (rewrite (in_synthesis_Equal g h H));trivial.
  Qed.

  Lemma hedge_reduce_morphism_aux : forall (g h:hedge),(H.Equal g h)->(H.Subset (hedge_reduce g) (hedge_reduce h)).
    intros.
    unfold hedge_reduce.
    unfold H.Subset.
    intros.
    destruct a.
    generalize (H.filter_1 (is_synthesisable_compat_bool g) H0);intro.
    generalize (H.filter_2 (is_synthesisable_compat_bool g) H0);intro.
    apply H.filter_3.
    apply is_synthesisable_compat_bool.
    elim (H (t,t0));intros;auto.
    rewrite (is_synthesisable_Equal h g).
    auto.
    apply HedgeProp.equal_sym;auto.
  Qed.

  Lemma hedge_reduce_morphism : forall (g h:hedge),(H.Equal g h)->(H.Equal (hedge_reduce g) (hedge_reduce h)).
    intros.
    apply HedgeProp.subset_antisym.
    apply hedge_reduce_morphism_aux;trivial.
    apply hedge_reduce_morphism_aux.
    apply HedgeProp.equal_sym;auto.
  Qed.

  Lemma hedge_irreducible_morphism : forall (g h:hedge),(H.Equal g h)->(H.Equal (hedge_irreducible g) (hedge_irreducible h)).
    unfold hedge_irreducible.
    intros.
    apply hedge_reduce_morphism.
    apply hedge_analysis_morphism.
    auto.
  Qed.

  Lemma hedge_irreducible_transpose_hedge_tranpose_irreducible : forall (h:hedge),(H.Equal (hedge_irreducible (hedge_transpose h)) (hedge_transpose (hedge_irreducible h))).
    unfold hedge_irreducible.
    intros.
    eapply HedgeProp.equal_trans.
    apply hedge_reduce_morphism.
    apply hedge_analysis_transpose_hedge_transpose_analysis.
    apply hedge_reduce_transpose_hedge_transpose_reduce.
  Qed.

  Lemma is_irreducible_transpose_is_irreducible : forall (h:hedge),(is_irreducible h)->(is_irreducible (hedge_transpose h)).
    unfold is_irreducible.
    intros.
    eapply HedgeProp.equal_trans.
    apply hedge_transpose_morphism.
    apply H.
    apply HedgeProp.equal_sym.
    apply hedge_irreducible_transpose_hedge_tranpose_irreducible.
  Qed.

  Lemma le_h_transpose_le_h : forall (g h:hedge),(le_h (relation_of_hedge g) (relation_of_hedge h))->(le_h (relation_of_hedge (hedge_transpose g)) (relation_of_hedge (hedge_transpose h))).
    intros.
    unfold le_h.
    intros.
    fold (hedge_synthesis_spec (hedge_transpose g)) in H0.
    fold (hedge_synthesis_spec (hedge_transpose h)).
    generalize (hedge_synthesis_spec_impl_hedge_in_synthesis (hedge_transpose g) m n H0);intro.
    apply hedge_in_synthesis_impl_hedge_synthesis_spec.
    rewrite in_synthesis_transpose.
    rewrite in_synthesis_transpose in H1.
    apply hedge_synthesis_spec_impl_hedge_in_synthesis.
    unfold hedge_synthesis_spec.
    apply H.
    apply hedge_in_synthesis_impl_hedge_synthesis_spec.
    trivial.
  Qed.

  Theorem irr_le_h_consistent_impl_consistent : forall (g h:hedge),(is_irreducible g)->(is_consistent h)->(le_h (relation_of_hedge g) (relation_of_hedge h))->(is_consistent g).
    intros.
    elim (is_consistent_iff h);intros.
    elim (H2 H0);clear H2;clear H3;clear H0;intros.
    elim (is_consistent_iff g);intros.
    apply H4.
    split.
    eapply irr_le_h_left_consistent_impl_left_consistent.
    apply H.
    apply H0.
    trivial.
    eapply irr_le_h_left_consistent_impl_left_consistent.
    apply is_irreducible_transpose_is_irreducible.
    apply H.
    apply H2.
    apply le_h_transpose_le_h.  
    trivial.
  Qed.
  
  (* Consistency definition: *)
  
  (* Definition name_left_name (h:message_relation) := forall (m n:MessageSpi.message),(h m n)->forall (a:name),(m=MNam a)->(exists b:name, n=(MessageSpi.MNam b)). *)
  (* Definition injective_left (h:message_relation) := forall (m n:MessageSpi.message),(h m n)->forall (m' n':MessageSpi.message),(h m' n')->(m=m')->(n=n'). *)
  
  (* Definition op_synthesis_left (h:message_relation) := forall (m n:MessageSpi.message),(h m n)->forall (o:op)(m':MessageSpi.message),(m=MOp o m')->forall (n':MessageSpi.message),~(hedge_synthesis_relation h m' n'). *)

  (* Definition pair_free_left (h:message_relation) := forall (m1 m2 n:MessageSpi.message),~(h (MessageSpi.MPair m1 m2) n). *)
  
  (* Definition enc_s_not_ana_left (h:message_relation) := forall (m n:MessageSpi.message),(h m n)->forall (m' k:MessageSpi.message),m=(MessageSpi.MEnc_s m' k)->forall (l:MessageSpi.message),~(hedge_synthesis_relation h k l). *)
  
  (* Definition enc_a_synthesis_left (h:message_relation) := forall (m n:MessageSpi.message),(h m n)->forall (m' k:MessageSpi.message),(m=MEnc_a m' k)->(forall (n':MessageSpi.message),~(hedge_synthesis_relation h m' n'))\/(forall (l:MessageSpi.message),~(hedge_synthesis_relation h k l)). *)
  (* Definition enc_a_analysis_left (h:message_relation) := forall (m n:MessageSpi.message),(h m n)->forall (m' k:MessageSpi.message),m=(MessageSpi.MEnc_a m' k)->forall (k':MessageSpi.message),(inverse_key k = Some k')->forall (l':MessageSpi.message),(hedge_synthesis_relation h k' l')->(exists l:MessageSpi.message,exists n':MessageSpi.message,inverse_key l = Some l' /\ n=(MessageSpi.MEnc_a n' l)/\(hedge_synthesis_relation h m' n')). *)
  
  (* Definition injective_inv_left (h:message_relation) := forall (m n:MessageSpi.message),(h m n)->forall (m' n':MessageSpi.message),(h m' n')->((Some m')=inverse_key m)->((Some n')=inverse_key n). *)

  (* first, we give a simplified definition of left consistency *)
  Lemma is_left_consistent_iff : forall (h:message_relation),(is_left_consistent_relation h)<->(name_left_name h)/\(injective_left h)/\(op_synthesis_left h)/\(pair_free_left h)/\(enc_s_not_ana_left h)/\(enc_a_synthesis_left h)/\(enc_a_analysis_left h)/\(injective_inv_left h).
    unfold is_left_consistent_relation.
    split;intros;and_elim.
    repeat (split;auto).
    elim (enc_s_syn_ana_iff_enc_s_not_ana h);intros.
    apply H8;try tauto.
    repeat (split;auto).
    elim (enc_s_syn_ana_iff_enc_s_not_ana h);intros.
    elim (H8 H3);auto.
    elim (enc_s_syn_ana_iff_enc_s_not_ana h);intros.
    elim (H8 H3);auto.
  Qed.

  (* Since we are often interested in checking consistency of irreducible hedges *)
  (* we look for a characterisation of consistency of reduce(h) in terms of h (without involving reduce) *)

  Lemma hedge_reduce_iff : forall (h:hedge)(m n:MessageSpi.message),(H.In (m,n) (hedge_reduce h))<->((H.In (m,n) h)/\(is_synthesisable h m n = false)).
    unfold hedge_reduce.
    split;intros.
    split.
    apply (H.filter_1 (s:=h) (x:=(m,n)) (is_synthesisable_compat_bool h));auto.
    generalize (H.filter_2 (s:=h) (x:=(m,n)) (is_synthesisable_compat_bool h) H).
    intro.
    destruct (is_synthesisable h m n);try (discriminate H0);auto.
    and_elim.
    apply H.filter_3;auto.
    apply is_synthesisable_compat_bool.
    rewrite H0;reflexivity.
  Qed.

  (* if g < h then inj_left (S h) -> inj_left (S g) *)
  Lemma le_h_inj_left_syn : forall (g h:message_relation),(le_h g h)->(injective_left (hedge_synthesis_relation h))->(injective_left (hedge_synthesis_relation g)).
    unfold injective_left;intros.
    eapply H0;eauto.
  Qed.

  Lemma le_h_inj_inv_left_syn : forall (g h:message_relation),(le_h g h)->(injective_inv_left (hedge_synthesis_relation h))->(injective_inv_left (hedge_synthesis_relation g)).
    unfold injective_inv_left;intros.
    eapply H0;eauto.
  Qed.

  Lemma inclusion_inj_left : forall (g h:message_relation),(inclusion MessageSpi.message g h)->(injective_left h)->(injective_left g).
    intros.
    unfold injective_left.
    intros.
    apply (H0 m n (H m n H1) m' n' (H m' n' H2) H3).
  Qed.

  Lemma inclusion_inj_inv_left : forall (g h:message_relation),(inclusion MessageSpi.message g h)->(injective_inv_left h)->(injective_inv_left g).
    intros.
    unfold injective_inv_left.
    intros.
    apply (H0 m n (H m n H1) m' n' (H m' n' H2) H3).
  Qed.

  Lemma le_h_inj_left : forall (g h:message_relation),(le_h g h)->(injective_left (hedge_synthesis_relation h))->(injective_left g).
    intros.
    eapply inclusion_inj_left with (hedge_synthesis_relation g).
    unfold inclusion;intros.
    apply SynInc;auto.
    eapply le_h_inj_left_syn;eauto.
  Qed.

  Lemma le_h_inj_inv_left : forall (g h:message_relation),(le_h g h)->(injective_inv_left (hedge_synthesis_relation h))->(injective_inv_left g).
    intros.
    eapply inclusion_inj_inv_left with (hedge_synthesis_relation g).
    unfold inclusion;intros.
    apply SynInc;auto.
    eapply le_h_inj_inv_left_syn;eauto.
  Qed.

  (* corollaries *)
  Lemma inj_left_syn_reduce_iff : forall (h:hedge),(injective_left (hedge_synthesis_relation (relation_of_hedge h)))<->(injective_left (hedge_synthesis_relation (relation_of_hedge (hedge_reduce h)))).
    split;intros;elim (h_equiv_h_red h);intros;eapply le_h_inj_left_syn;eauto.
  Qed.

  Lemma inj_inv_left_syn_reduce_iff : forall (h:hedge),(injective_inv_left (hedge_synthesis_relation (relation_of_hedge h)))<->(injective_inv_left (hedge_synthesis_relation (relation_of_hedge (hedge_reduce h)))).
    split;intros;elim (h_equiv_h_red h);intros;eapply le_h_inj_inv_left_syn;eauto.
  Qed.


  (* name_left_name *)
  Lemma name_left_name_reduce_iff : forall (h:hedge),(name_left_name (relation_of_hedge (hedge_reduce h)))<->(name_left_name (relation_of_hedge h)).
    intro h.
    unfold name_left_name;split;intro.    
    intros.
    elim (H m n) with a;auto.
    intro b;intro.
    exists b;auto.
    unfold relation_of_hedge.
    elim (hedge_reduce_iff h m n);intros.
    apply H3.
    split;auto.
    rewrite H1;reflexivity.
    intros.
    elim (H m n) with a;auto.
    intro b;intro.
    exists b;auto.
    unfold relation_of_hedge.
    unfold relation_of_hedge in H0.
    elim (hedge_reduce_iff h m n);intros.
    elim (H2 H0);intros;auto.
  Qed.

  (* op_synthesis_left *)
  Definition op_synthesis_left_aux (h:message_relation) := forall (m n:MessageSpi.message),(h m n)->forall (o:OpSpi.op)(m':MessageSpi.message),(m=MessageSpi.MOp o m')->forall (n':MessageSpi.message),(hedge_synthesis_relation h m' n')->(n=MessageSpi.MOp o n').

  Lemma op_synthesis_left_impl_reduce : forall (h:hedge),(op_synthesis_left_aux (relation_of_hedge h))->(op_synthesis_left (relation_of_hedge (hedge_reduce h))).
    intro h.
    elim (h_equiv_h_red h);intros.
    unfold op_synthesis_left, op_synthesis_left_aux;intros.
    red;intro.
    generalize (H0 m' n' H4);intro.
    elim (hedge_reduce_iff h m n);intros.
    elim (H6 H2);intros.
    generalize (H1 m n H8 o m' H3 n' H5);intro.
    subst m;subst n.
    unfold is_synthesisable in H9.
    generalize H9;clear H9.
    case OpSpi.op_eq_dec;intro.
    rewrite hedge_synthesis_spec_impl_hedge_in_synthesis.
    intro.
    discriminate H9.
    apply H5.
    elim n;auto.
  Qed.

  Lemma op_synthesis_left_reduce_impl : forall (h:hedge),(op_synthesis_left (relation_of_hedge (hedge_reduce h)))->(injective_left (hedge_synthesis_relation (relation_of_hedge (hedge_reduce h))))->(op_synthesis_left_aux (relation_of_hedge h)).
    intro h.
    elim (h_equiv_h_red h);intros.
    unfold op_synthesis_left_aux;intros.
    assert (hedge_synthesis_relation (relation_of_hedge h) (MessageSpi.MOp o m') (MessageSpi.MOp o n')).
    apply SynOp;auto.
    assert (hedge_synthesis_relation (relation_of_hedge h) m n).
    apply SynInc;auto.
    elim (inj_left_syn_reduce_iff h);intros.
    generalize (H9 H2);clear H8;clear H9;intro.
    apply (H8 m n H7 (MessageSpi.MOp o m') (MessageSpi.MOp o n') H6 H4).
  Qed.

  (* pair free left *)
  Definition pair_free_left_aux (h:message_relation) := forall (m n:MessageSpi.message),(h m n)->forall (m1 m2:MessageSpi.message),m=MessageSpi.MPair m1 m2->exists n1:MessageSpi.message,exists n2:MessageSpi.message,(hedge_synthesis_relation h m1 n1)/\(hedge_synthesis_relation h m2 n2)/\n=MessageSpi.MPair n1 n2.

  Lemma pair_free_left_reduce_iff : forall (h:hedge),(pair_free_left_aux (relation_of_hedge h))<->(pair_free_left (relation_of_hedge (hedge_reduce h))).
    intro h.
    split.
    unfold pair_free_left;intros.
    red;intro.
    elim (hedge_reduce_iff h (MessageSpi.MPair m1 m2) n);intros.
    elim (H1 H0);intros.
    elim (H (MessageSpi.MPair m1 m2) n H3 m1 m2 (refl_equal (MessageSpi.MPair m1 m2))).
    intro n1;intro.
    ex_elim H5 n2.
    rewrite H7 in H4.
    unfold is_synthesisable in H4.
    rewrite hedge_synthesis_spec_impl_hedge_in_synthesis in H4.
    rewrite hedge_synthesis_spec_impl_hedge_in_synthesis in H4.
    discriminate H4.
    apply H6.
    apply H5.
    unfold pair_free_left_aux;intros.
    generalize (H m1 m2 n);intro.
    rewrite <- H1 in H2.
    assert (is_synthesisable h m n = true).
    elim (hedge_reduce_iff h m n);intros.
    destruct (is_synthesisable h m n);auto.
    elim H2.
    apply H4.
    split;auto.
    rewrite H1 in H3.
    unfold is_synthesisable in H3.
    destruct n;try (discriminate H3).
    exists n1.
    exists n2.
    elim (andb_prop (hedge_in_synthesis h m1 n1) (hedge_in_synthesis h m2 n2) H3);intros.
    repeat (split;auto).
    apply hedge_in_synthesis_impl_hedge_synthesis_spec;auto.
    apply hedge_in_synthesis_impl_hedge_synthesis_spec;auto.
  Qed.

  (* enc_s_not_ana_left *)
  Definition enc_s_not_ana_left_aux (h:message_relation) := forall (m n:MessageSpi.message),(h m n)->forall (m' k:MessageSpi.message),m=(MessageSpi.MEnc_s m' k)->forall (l:MessageSpi.message),(hedge_synthesis_relation h k l)->exists n':MessageSpi.message,(hedge_synthesis_relation h m' n')/\n=MessageSpi.MEnc_s n' l.

  Lemma enc_s_not_ana_left_impl_reduce : forall (h:hedge),(enc_s_not_ana_left_aux (relation_of_hedge h))->(enc_s_not_ana_left (relation_of_hedge (hedge_reduce h))).
    unfold enc_s_not_ana_left;intros.
    red;intro.
    elim (h_equiv_h_red h);intros.
    generalize (H4 k l H2);intro.
    elim (hedge_reduce_iff h m n);intros.
    elim (H6 H0);intros.
    elim (H m n H8 m' k H1 l H5);intro n';intros;and_elim.
    rewrite H1 in H9.
    rewrite H11 in H9.
    unfold is_synthesisable in H9.
    rewrite hedge_synthesis_spec_impl_hedge_in_synthesis in H9;auto.
    rewrite hedge_synthesis_spec_impl_hedge_in_synthesis in H9;auto.
    discriminate H9.
  Qed.

  Lemma enc_s_not_ana_left_reduce_impl : forall (h:hedge),(enc_s_not_ana_left (relation_of_hedge (hedge_reduce h)))->(injective_left (hedge_synthesis_relation (relation_of_hedge (hedge_reduce h))))->(enc_s_not_ana_left_aux (relation_of_hedge h)).
    unfold enc_s_not_ana_left_aux;intros.
    elim (h_equiv_h_red h);intros.
    generalize (H4 k l H3);intro.
    case (Bool_elim_bool (is_synthesisable h m n)).
    rewrite H2.
    unfold is_synthesisable.
    intro.
    destruct n;try (discriminate H7).
    elim (andb_prop (hedge_in_synthesis h m' n1) (hedge_in_synthesis h k n2) H7);intros.
    assert (hedge_synthesis_relation (relation_of_hedge (hedge_reduce h)) k n2).
    apply H4.
    apply hedge_in_synthesis_impl_hedge_synthesis_spec;auto.
    generalize (H0 k l H6 k n2 H10 (refl_equal k));intro.
    subst n2.
    exists n1.
    split;auto.
    apply hedge_in_synthesis_impl_hedge_synthesis_spec;auto.
    intro.
    elim (hedge_reduce_iff h m n);intros.
    assert (relation_of_hedge (hedge_reduce h) m n).
    apply H9.
    split;auto.
    elim (H m n H10 m' k H2 l);auto.
  Qed.

  (* we now give a function to compute the "left synthesis" of a hedge *)
  Definition make_pair (mn1 mn2:H.elt) := let (m1,n1) := mn1 in let (m2,n2) := mn2 in (MessageSpi.MPair m1 m2, MessageSpi.MPair n1 n2).

  Definition make_all_pairs (h1 h2:hedge) := hedge_map_two_op make_pair h1 h2.

  Lemma make_all_pairs_prop : forall (h1 h2:hedge)(m1 n1 m2 n2:MessageSpi.message),(H.In (m1,n1) h1)/\(H.In (m2,n2) h2)<->(H.In (MessageSpi.MPair m1 m2, MessageSpi.MPair n1 n2) (make_all_pairs h1 h2)).
    unfold make_all_pairs.
    intros.
    replace (MessageSpi.MPair m1 m2, MessageSpi.MPair n1 n2) with (make_pair (m1,n1) (m2,n2));try reflexivity.
    apply hedge_map_two_op_prop.
    unfold f_injective2.
    intros.
    destruct mn1;destruct mn2;destruct mn1';destruct mn2'.
    simpl in H;and_elim.
    simpl.
    injection H;injection H0;intros.
    rewrite H1;rewrite H2;rewrite H3;rewrite H4.
    auto.
  Qed.

  Lemma make_all_pairs_inv : forall (h1 h2:hedge)(m n:MessageSpi.message),(H.In (m,n) (make_all_pairs h1 h2))->(exists m1:MessageSpi.message,exists n1:MessageSpi.message,exists m2:MessageSpi.message,exists n2:MessageSpi.message,m=(MessageSpi.MPair m1 m2)/\n=(MessageSpi.MPair n1 n2)/\(H.In (m1,n1) h1)/\(H.In (m2,n2) h2)).
    intros.
    unfold make_all_pairs in H.
    elim (hedge_map_two_op_inv make_pair h2 h1 (m,n) H).
    intro mn1;intro.
    ex_elim H0 mn2.
    destruct mn1;destruct mn2;simpl in H2.
    injection H2;intros.
    exists t.
    exists t0.
    exists t1.
    exists t2.
    auto.
  Qed.

  Definition make_cyph_s (mn1 mn2:H.elt) := let (m1,n1) := mn1 in let (m2,n2) := mn2 in (MessageSpi.MEnc_s m1 m2, MessageSpi.MEnc_s n1 n2).

  Definition make_all_cyphs_s (h1 h2:hedge) := hedge_map_two_op make_cyph_s h1 h2.

  Lemma make_all_cyphs_s_prop : forall (h1 h2:hedge)(m1 n1 m2 n2:MessageSpi.message),(H.In (m1,n1) h1)/\(H.In (m2,n2) h2)<->(H.In (MessageSpi.MEnc_s m1 m2, MessageSpi.MEnc_s n1 n2) (make_all_cyphs_s h1 h2)).
    unfold make_all_cyphs_s.
    intros.
    replace (MessageSpi.MEnc_s m1 m2, MessageSpi.MEnc_s n1 n2) with (make_cyph_s (m1,n1) (m2,n2));try reflexivity.
    apply hedge_map_two_op_prop.
    unfold f_injective2.
    intros.
    destruct mn1;destruct mn2;destruct mn1';destruct mn2'.
    simpl in H;and_elim.
    simpl.
    injection H;injection H0;intros.
    rewrite H1;rewrite H2;rewrite H3;rewrite H4.
    auto.
  Qed.

  Lemma make_all_cyphs_s_inv : forall (h1 h2:hedge)(m n:MessageSpi.message),(H.In (m,n) (make_all_cyphs_s h1 h2))->(exists m1:MessageSpi.message,exists n1:MessageSpi.message,exists m2:MessageSpi.message,exists n2:MessageSpi.message,m=(MessageSpi.MEnc_s m1 m2)/\n=(MessageSpi.MEnc_s n1 n2)/\(H.In (m1,n1) h1)/\(H.In (m2,n2) h2)).
    intros.
    unfold make_all_cyphs_s in H.
    elim (hedge_map_two_op_inv make_cyph_s h2 h1 (m,n) H).
    intro mn1;intro.
    ex_elim H0 mn2.
    destruct mn1;destruct mn2;simpl in H2.
    injection H2;intros.
    exists t.
    exists t0.
    exists t1.
    exists t2.
    auto.
  Qed.

  Definition make_cyph_a (mn1 mn2:H.elt) := let (m1,n1) := mn1 in let (m2,n2) := mn2 in (MessageSpi.MEnc_a m1 m2, MessageSpi.MEnc_a n1 n2).

  Definition make_all_cyphs_a (h1 h2:hedge) := hedge_map_two_op make_cyph_a h1 h2.

  Lemma make_all_cyphs_a_prop : forall (h1 h2:hedge)(m1 n1 m2 n2:MessageSpi.message),(H.In (m1,n1) h1)/\(H.In (m2,n2) h2)<->(H.In (MessageSpi.MEnc_a m1 m2, MessageSpi.MEnc_a n1 n2) (make_all_cyphs_a h1 h2)).
    unfold make_all_cyphs_a.
    intros.
    replace (MessageSpi.MEnc_a m1 m2, MessageSpi.MEnc_a n1 n2) with (make_cyph_a (m1,n1) (m2,n2));try reflexivity.
    apply hedge_map_two_op_prop.
    unfold f_injective2.
    intros.
    destruct mn1;destruct mn2;destruct mn1';destruct mn2'.
    simpl in H;and_elim.
    simpl.
    injection H;injection H0;intros.
    rewrite H1;rewrite H2;rewrite H3;rewrite H4.
    auto.
  Qed.

  Lemma make_all_cyphs_a_inv : forall (h1 h2:hedge)(m n:MessageSpi.message),(H.In (m,n) (make_all_cyphs_a h1 h2))->(exists m1:MessageSpi.message,exists n1:MessageSpi.message,exists m2:MessageSpi.message,exists n2:MessageSpi.message,m=(MessageSpi.MEnc_a m1 m2)/\n=(MessageSpi.MEnc_a n1 n2)/\(H.In (m1,n1) h1)/\(H.In (m2,n2) h2)).
    intros.
    unfold make_all_cyphs_a in H.
    elim (hedge_map_two_op_inv make_cyph_a h2 h1 (m,n) H).
    intro mn1;intro.
    ex_elim H0 mn2.
    destruct mn1;destruct mn2;simpl in H2.
    injection H2;intros.
    exists t.
    exists t0.
    exists t1.
    exists t2.
    auto.
  Qed.

  Definition make_op (o:OpSpi.op) (mn:H.elt) := let (m,n) := mn in (MessageSpi.MOp o m, MessageSpi.MOp o n).

  Definition make_all_ops (o:OpSpi.op) (h:hedge) := hedge_map_single_op (make_op o) h.

  Lemma make_all_ops_prop : forall (o:OpSpi.op)(h:hedge)(m n:MessageSpi.message),(H.In (m,n) h)<->(H.In (MessageSpi.MOp o m, MessageSpi.MOp o n) (make_all_ops o h)).
    intros.
    replace (MessageSpi.MOp o m, MessageSpi.MOp o n) with (make_op o (m,n));try reflexivity.
    unfold make_all_ops.
    apply hedge_map_single_op_prop.
    unfold f_injective.
    intros.
    destruct x;destruct y;simpl in H;and_elim.
    injection H;injection H0;intros.
    rewrite H1;rewrite H2;auto.
  Qed.

  Lemma make_all_ops_inv : forall (o:OpSpi.op)(h:hedge)(m n:MessageSpi.message),(H.In (m,n) (make_all_ops o h))->(exists m':MessageSpi.message,exists n':MessageSpi.message,m=(MessageSpi.MOp o m')/\n=(MessageSpi.MOp o n')/\(H.In (m',n') h)).
    unfold make_all_ops.
    intros.
    elim (hedge_map_single_op_inv (make_op o) h (m,n) H);intro mn'.
    intros.
    and_elim.
    destruct mn'.
    simpl in H1.
    injection H1;intros.
    exists t.
    exists t0.
    auto.
  Qed.

  Definition filter_left (m:MessageSpi.message) (mn:H.elt) := let (m',n') := mn in (if MessageSpi.eq_dec m m' then true else false).

  Lemma compat_bool_filter_left : forall (m:MessageSpi.message),FSetInterface.compat_bool H.E.eq (filter_left m).
    unfold FSetInterface.compat_bool.
    intros.
    destruct x;destruct y;simpl in H;and_elim.
    rewrite H;rewrite H0;trivial.
  Qed.

  Lemma filter_left_ok : forall (h:hedge)(m m' n:MessageSpi.message),(H.In (m',n) (H.filter (filter_left m) h))<->m=m'/\(H.In (m,n) h).
    split;intros.
    generalize (H.filter_1 (compat_bool_filter_left m) H);intros.
    generalize (H.filter_2 (compat_bool_filter_left m) H);intros.
    simpl in H1.
    generalize H1.
    case MessageSpi.eq_dec.
    intros.
    subst m'.
    split;trivial.
    intros.
    discriminate H2.
    and_elim.
    subst m'.
    apply (H.filter_3 (compat_bool_filter_left m) H0).
    simpl.
    case MessageSpi.eq_dec;trivial.
    intros.
    elim n0;trivial.
  Qed.
    
  Fixpoint hedge_synthesis_left (h:hedge) (m:MessageSpi.message) {struct m} : hedge :=
    let h_base := H.filter (filter_left m) h in
      H.union h_base
      (match m with
	 | (MessageSpi.MPair m1 m2) =>
	   let h_m1 := hedge_synthesis_left h m1 in
	     let h_m2 := hedge_synthesis_left h m2 in
	       make_all_pairs h_m1 h_m2
	 | (MessageSpi.MEnc_s m1 m2) =>
	   let h_m1 := hedge_synthesis_left h m1 in
	     let h_m2 := hedge_synthesis_left h m2 in
	       make_all_cyphs_s h_m1 h_m2
	 | (MessageSpi.MEnc_a m1 m2) =>
	   let h_m1 := hedge_synthesis_left h m1 in
	     let h_m2 := hedge_synthesis_left h m2 in
	       make_all_cyphs_a h_m1 h_m2
	 | (MessageSpi.MOp o m) =>
	   let h_m := hedge_synthesis_left h m in
	     make_all_ops o h_m
	 | _ => H.empty
       end).

  Lemma hedge_synthesis_left_ok : forall (h:hedge)(m:MessageSpi.message)(m' n:MessageSpi.message),(H.In (m',n) (hedge_synthesis_left h m))<->m=m'/\(hedge_synthesis_spec h m n).
    induction m;split;simpl;intros;try (case (H.union_1 H));intros;and_elim.
    elim (filter_left_ok h (MessageSpi.MNam t) m' n);intros.
    elim (H1 H0);clear H1;clear H2;intros.
    split;trivial.
    unfold hedge_synthesis_spec.
    apply SynInc.
    apply H2.
    elim (H.empty_1 H0).
    elim (filter_left_ok h (MessageSpi.MNam t) m' n);intros.
    apply H.union_2.
    apply H2.
    split;trivial.
    apply H.mem_2.
    rewrite <- name_left_synthesis.
    apply hedge_synthesis_spec_impl_hedge_in_synthesis.
    trivial.
    elim (filter_left_ok h (MessageSpi.MPair m1 m2) m' n);intros.
    elim (H1 H0);intros.
    split;trivial.
    unfold hedge_synthesis_spec.
    apply SynInc.
    apply H4.
    elim (make_all_pairs_inv (hedge_synthesis_left h m1) (hedge_synthesis_left h m2) m' n H0).
    intro m1';intro.
    ex_elim H1 n1'.
    ex_elim H1 m2'.
    ex_elim H1 n2'.
    elim (IHm1 m1' n1');intros.
    elim (IHm2 m2' n2');intros.
    elim (H5 H3);clear H5;clear H6;intros.
    elim (H7 H4);clear H7;clear H8;intros.
    subst m1';subst m2'.
    subst m'.
    split;trivial.
    unfold hedge_synthesis_spec.
    subst n.
    apply SynPair.
    apply H6.
    apply H8.
    inversion H0.
    unfold relation_of_hedge in H1.
    apply H.union_2.
    elim (filter_left_ok h (MessageSpi.MPair m1 m2) m' n);intros.
    auto.
    fold (hedge_synthesis_spec h) in H3.
    fold (hedge_synthesis_spec h) in H5.
    apply H.union_3.
    subst m'.
    elim (make_all_pairs_prop (hedge_synthesis_left h m1) (hedge_synthesis_left h m2) m1 n1 m2 n2);intros.
    apply H.
    elim (IHm1 m1 n1);intros.
    elim (IHm2 m2 n2);intros.
    split;auto.
    elim (filter_left_ok h (MessageSpi.MEnc_s m1 m2) m' n);intros.
    elim (H1 H0);intros.
    split;trivial.
    unfold hedge_synthesis_spec.
    apply SynInc.
    apply H4.
    elim (make_all_cyphs_s_inv (hedge_synthesis_left h m1) (hedge_synthesis_left h m2) m' n H0).
    intro m1';intro.
    ex_elim H1 n1'.
    ex_elim H1 m2'.
    ex_elim H1 n2'.
    elim (IHm1 m1' n1');intros.
    elim (IHm2 m2' n2');intros.
    elim (H5 H3);intros.
    elim (H7 H4);intros.
    subst m1';subst m2'.
    split;auto.
    subst n.
    unfold hedge_synthesis_spec.
    apply SynEnc_s.
    apply H10.
    apply H12.
    inversion H0.
    apply H.union_2.
    elim (filter_left_ok h (MessageSpi.MEnc_s m1 m2) m' n);intros.
    apply H5.
    split;trivial.
    subst m'.
    apply H.union_3.
    elim (make_all_cyphs_s_prop (hedge_synthesis_left h m1) (hedge_synthesis_left h m2) m1 n0 m2 l);intros.
    apply H.
    elim (IHm1 m1 n0);intros.
    elim (IHm2 m2 l);intros.
    split.
    apply H8;auto.
    apply H10;auto.
    elim (filter_left_ok h (MessageSpi.MEnc_a m1 m2) m' n);intros.
    elim (H1 H0);intros.
    split;trivial.
    unfold hedge_synthesis_spec.
    apply SynInc.
    apply H4.
    elim (make_all_cyphs_a_inv (hedge_synthesis_left h m1) (hedge_synthesis_left h m2) m' n H0).
    intro m1';intro.
    ex_elim H1 n1'.
    ex_elim H1 m2'.
    ex_elim H1 n2'.
    elim (IHm1 m1' n1');intros.
    elim (IHm2 m2' n2');intros.
    elim (H5 H3);intros.
    elim (H7 H4);intros.
    subst m1';subst m2'.
    split;auto.
    subst n.
    unfold hedge_synthesis_spec.
    apply SynEnc_a.
    apply H10.
    apply H12.
    inversion H0.
    apply H.union_2.
    elim (filter_left_ok h (MessageSpi.MEnc_a m1 m2) m' n);intros.
    apply H5.
    split;trivial.
    subst m'.
    apply H.union_3.
    elim (make_all_cyphs_a_prop (hedge_synthesis_left h m1) (hedge_synthesis_left h m2) m1 n0 m2 l);intros.
    apply H.
    elim (IHm1 m1 n0);intros.
    elim (IHm2 m2 l);intros.
    split.
    apply H8;auto.
    apply H10;auto.
    elim (filter_left_ok h (MessageSpi.MOp o m) m' n);intros.
    elim (H1 H0);intros.
    split;trivial.
    unfold hedge_synthesis_spec.
    apply SynInc.
    apply H4.
    elim (make_all_ops_inv o (hedge_synthesis_left h m) m' n H0).
    intro m1;intro.
    ex_elim H1 n1.
    elim (IHm m1 n1);intros.
    elim (H4 H3);intros.
    subst m1.
    split;auto.
    subst n.
    unfold hedge_synthesis_spec.
    apply SynOp.
    auto.
    inversion H0.
    apply H.union_2.
    elim (filter_left_ok h (MessageSpi.MOp o m) m' n);intros.
    apply H5.
    auto.
    subst n;subst o0.
    apply H.union_3.
    subst m'.
    elim (make_all_ops_prop o (hedge_synthesis_left h m) m n0);intros.
    apply H.
    elim (IHm m n0);intros.
    apply H5.
    auto.
  Qed.

  Lemma hedge_synthesis_left_empty : forall (h:hedge)(m:MessageSpi.message),(H.Empty (hedge_synthesis_left h m))<->(forall (n:MessageSpi.message),~(hedge_synthesis_spec h m n)).
    split;intros.
    intro.
    elim (H (m,n)).
    elim (hedge_synthesis_left_ok h m m n);intros;auto.
    case (HedgeOtherProp.empty_dec (hedge_synthesis_left h m));intro.
    apply HedgeOtherProp.Equal_empty_empty;trivial.
    assert (~H.Empty (hedge_synthesis_left h m)).
    intro.
    elim n.
    apply HedgeOtherProp.empty_is_empty;trivial.
    case (Option_elim_option _ (H.choose (hedge_synthesis_left h m)));intro.
    ex_elim s mn.
    generalize (H.choose_1 s);intro.
    destruct mn as (m',n').
    elim (hedge_synthesis_left_ok h m m' n');intros.
    elim (H2 H1);intros.
    elim (H n').
    trivial.
    eapply H.choose_2.
    trivial.
  Qed.

  Lemma hedge_in_left_synthesis_dec : forall (h:hedge)(m:MessageSpi.message),{n:MessageSpi.message|(hedge_synthesis_spec h m n)}+{forall (n:MessageSpi.message),~(hedge_synthesis_spec h m n)}.
    intros.
    case (Bool_eq_dec (H.is_empty (hedge_synthesis_left h m)));intro.
    right.
    elim (hedge_synthesis_left_empty h m);intros.
    apply H.
    auto.
    left.
    case (Option_elim_option _ (H.choose (hedge_synthesis_left h m)));intro.
    ex_elim s mn.
    destruct mn as [m' n].
    exists n.
    elim (hedge_synthesis_left_ok h m m' n);intros.
    elim (H (H.choose_1 s));intros.
    auto.
    generalize (H.choose_2 e0);intro.
    rewrite H.is_empty_1 in e.
    discriminate e.
    auto.
  Qed.

  Lemma hedge_in_left_synthesis_or : forall (h:hedge)(m:MessageSpi.message),(exists n:MessageSpi.message,(hedge_synthesis_spec h m n))\/(forall (n:MessageSpi.message),~(hedge_synthesis_spec h m n)).
    intros.
    case (hedge_in_left_synthesis_dec h m);intro.
    ex_elim s n.
    left;exists n;auto.
    right;auto.
  Qed.

  (* Definition enc_a_synthesis_left (h:message_relation) := forall (m n:MessageSpi.message),(h m n)->forall (m' k:MessageSpi.message),(m=MEnc_a m' k)->(forall (n':MessageSpi.message),~(hedge_synthesis_relation h m' n'))\/(forall (l:MessageSpi.message),~(hedge_synthesis_relation h k l)). *)

  Definition enc_a_synthesis_left_aux (h:message_relation) := forall (m n:MessageSpi.message),(h m n)->forall (m' k:MessageSpi.message),(m=MessageSpi.MEnc_a m' k)->forall (n':MessageSpi.message),(hedge_synthesis_relation h m' n')->forall (l:MessageSpi.message),(hedge_synthesis_relation h k l)->n=MessageSpi.MEnc_a n' l.

  Lemma enc_a_synthesis_left_impl_reduce : forall (h:hedge),(enc_a_synthesis_left_aux (relation_of_hedge h))->(enc_a_synthesis_left (relation_of_hedge (hedge_reduce h))).
    intro h.
    elim (h_equiv_h_red h);intros.
    unfold enc_a_synthesis_left;intros.
    elim (hedge_reduce_iff h m n);intros.
    elim (H4 H2);intros.
    rewrite H3 in H7.
    case (hedge_in_left_synthesis_or h m');case (hedge_in_left_synthesis_or h k);intros.
    ex_elim H8 l.
    ex_elim H9 n'.
    generalize (H1 m n H6 m' k H3 n' H9 l H8);intros.
    rewrite H10 in H7.
    unfold is_synthesisable in H7.
    rewrite hedge_synthesis_spec_impl_hedge_in_synthesis in H7;auto.    
    rewrite hedge_synthesis_spec_impl_hedge_in_synthesis in H7;auto.
    discriminate H7.
    right;intros.
    red;intro.
    apply (H8 l).
    apply (H0 k l H10);auto.
    left;intros.
    red;intro.
    apply (H9 n').
    apply (H0 m' n' H10);auto.
    left;intros.
    red;intro.
    apply (H9 n').
    apply (H0 m' n' H10);auto.
  Qed.

  Lemma enc_a_synthesis_left_reduce_impl : forall (h:hedge),(enc_a_synthesis_left (relation_of_hedge (hedge_reduce h)))->(injective_left (hedge_synthesis_relation (relation_of_hedge (hedge_reduce h))))->(enc_a_synthesis_left_aux (relation_of_hedge h)).
    intro h.
    elim (h_equiv_h_red h);intros.
    unfold enc_a_synthesis_left_aux;intros.
    assert (hedge_synthesis_relation (relation_of_hedge (hedge_reduce h)) m' n').
    apply H;auto.
    assert (hedge_synthesis_relation (relation_of_hedge (hedge_reduce h)) k l).
    apply H;auto.
    case (Bool_elim_bool (is_synthesisable h m n)).
    rewrite H4;unfold is_synthesisable;intro.
    destruct n;try (discriminate H9).
    elim (andb_prop (hedge_in_synthesis h m' n1) (hedge_in_synthesis h k n2) H9);intros.
    assert (hedge_synthesis_relation (relation_of_hedge (hedge_reduce h)) m' n1).
    apply H.
    apply hedge_in_synthesis_impl_hedge_synthesis_spec;auto.
    assert (hedge_synthesis_relation (relation_of_hedge (hedge_reduce h)) k n2).
    apply H.
    apply hedge_in_synthesis_impl_hedge_synthesis_spec;auto.
    rewrite (H2 m' n' H7 m' n1 H12 (refl_equal m')).
    rewrite (H2 k l H8 k n2 H13 (refl_equal k));auto.
    intro.
    elim (hedge_reduce_iff h m n);intros.
    assert (relation_of_hedge (hedge_reduce h) m n).
    apply H11.
    split;auto.
    case (H1 m n H12 m' k H4);intro.
    elim (H13 n');auto.
    elim (H13 l);auto.
  Qed.

  (* Definition enc_a_analysis_left (h:message_relation) := forall (m n:MessageSpi.message),(h m n)->forall (m' k:MessageSpi.message),m=(MessageSpi.MEnc_a m' k)->forall (k':MessageSpi.message),(inverse_key k = Some k')->forall (l':MessageSpi.message),(hedge_synthesis_relation h k' l')->(exists l:MessageSpi.message,exists n':MessageSpi.message,inverse_key l = Some l' /\ n=(MessageSpi.MEnc_a n' l)/\(hedge_synthesis_relation h m' n')). *)

  Lemma enc_a_analysis_left_impl_reduce : forall (h:hedge),(enc_a_analysis_left (relation_of_hedge h))->(enc_a_analysis_left (relation_of_hedge (hedge_reduce h))).
    intro h.
    elim (h_equiv_h_red h);intros.
    unfold enc_a_analysis_left;intros.
    generalize (H0 k' l' H5);intro.
    elim (hedge_reduce_iff h m n);intros.
    elim (H7 H2);intros.
    elim (H1 m n H9 m' k H3 k' H4 l' H6);intro l;intros.
    ex_elim H11 n'.
    exists l.
    exists n'.
    split;trivial.
    split;trivial.
    apply H;auto.
  Qed.

  Lemma enc_a_analysis_left_reduce_impl : forall (h:hedge),(enc_a_analysis_left (relation_of_hedge (hedge_reduce h)))->(injective_inv_left (hedge_synthesis_relation (relation_of_hedge h)))->(enc_a_analysis_left (relation_of_hedge h)).
    intro h.
    elim (h_equiv_h_red h);intros.
    unfold enc_a_analysis_left;intros.
    case (Bool_elim_bool (is_synthesisable h m n));intro.
    rewrite H4 in H7;unfold is_synthesisable in H7;destruct n;try (discriminate H7).
    elim (andb_prop (hedge_in_synthesis h m' n1) (hedge_in_synthesis h k n2) H7);intros.
    assert (hedge_synthesis_relation (relation_of_hedge h) m' n1).
    apply hedge_in_synthesis_impl_hedge_synthesis_spec;auto.
    assert (hedge_synthesis_relation (relation_of_hedge h) k n2).
    apply hedge_in_synthesis_impl_hedge_synthesis_spec;auto.
    symmetry in H5.
    generalize (H2 k n2 H11 k' l' H6 H5);intro.
    symmetry in H12.
    exists n2.
    exists n1.
    split;trivial.
    split;trivial.
    elim (hedge_reduce_iff h m n);intros.
    assert (relation_of_hedge (hedge_reduce h) m n).
    apply H9.
    split;auto.
    elim (H1 m n H10 m' k H4 k' H5 l' (H k' l' H6));intro l;intro.
    ex_elim H11 n'.
    exists l.
    exists n'.
    split;trivial.
    split;trivial.
    apply H0;auto.
  Qed.

  Definition is_left_consistent_relation_aux (h:message_relation) := (name_left_name h)/\(injective_left h)/\(op_synthesis_left_aux h)/\(pair_free_left_aux h)/\(enc_s_not_ana_left_aux h)/\(enc_a_synthesis_left_aux h)/\(enc_a_analysis_left h)/\(injective_inv_left h).

  Lemma is_left_consistent_reduce_impl : forall (h:hedge),(is_left_consistent_relation (relation_of_hedge (hedge_reduce h)))->(is_left_consistent_relation_aux (relation_of_hedge h)).
    unfold is_left_consistent_relation_aux, is_left_consistent_relation;intros;and_elim.
    repeat (split;idtac).
    elim (name_left_name_reduce_iff h);auto.
    eapply le_h_inj_left.
    apply h_le_h_reduce.
    apply inj_left_syn;auto.
    apply op_synthesis_left_reduce_impl;auto.
    apply inj_left_syn;auto.
    elim (pair_free_left_reduce_iff h);auto.
    apply enc_s_not_ana_left_reduce_impl;auto.
    elim (enc_s_syn_ana_iff_enc_s_not_ana (relation_of_hedge (hedge_reduce h)));intros;try tauto.
    apply inj_left_syn;auto.
    apply enc_a_synthesis_left_reduce_impl;auto.
    apply inj_left_syn;auto.
    apply enc_a_analysis_left_reduce_impl;auto.
    eapply le_h_inj_inv_left_syn.
    apply h_le_h_reduce.
    apply inj_inv_left_syn;auto.
    eapply le_h_inj_inv_left.
    apply h_le_h_reduce.
    apply inj_inv_left_syn;auto.
  Qed.

  Lemma is_left_consistent_impl_reduce : forall (h:hedge),(is_left_consistent_relation_aux (relation_of_hedge h))->(is_left_consistent_relation (relation_of_hedge (hedge_reduce h))).
    unfold is_left_consistent_relation_aux, is_left_consistent_relation;intros;and_elim.
    repeat (split;idtac).
    elim (name_left_name_reduce_iff h);auto.
    eapply inclusion_inj_left with (relation_of_hedge h).
    unfold inclusion;intros.
    elim (hedge_reduce_iff h x y);intros.
    elim (H8 H7);auto.
    trivial.
    apply op_synthesis_left_impl_reduce;auto.
    elim (pair_free_left_reduce_iff h);auto.
    elim (enc_s_syn_ana_iff_enc_s_not_ana (relation_of_hedge (hedge_reduce h)));intros;try tauto.
    elim (H8 (enc_s_not_ana_left_impl_reduce h H3));auto.
    elim (enc_s_syn_ana_iff_enc_s_not_ana (relation_of_hedge (hedge_reduce h)));intros;try tauto.
    elim (H8 (enc_s_not_ana_left_impl_reduce h H3));auto.
    apply enc_a_synthesis_left_impl_reduce;auto.
    apply enc_a_analysis_left_impl_reduce;auto.
    eapply inclusion_inj_inv_left with (relation_of_hedge h).
    unfold inclusion;intros.
    elim (hedge_reduce_iff h x y);intros.
    elim (H8 H7);auto.
    trivial.
  Qed.

  (* properties of the synthesis of a left consistent hedge *)
  Lemma enc_s_syn_left_consistent : forall (h:message_relation),(is_left_consistent_relation h)->forall (m n:MessageSpi.message),(hedge_synthesis_relation h m n)->forall (m1 m2:MessageSpi.message),(m=MessageSpi.MEnc_s m1 m2)->forall (n2:MessageSpi.message),(hedge_synthesis_relation h m2 n2)->exists n1:MessageSpi.message,((n=MessageSpi.MEnc_s n1 n2)/\(hedge_synthesis_relation h m1 n1)).
    intros h H.
    unfold is_left_consistent_relation in H.
    and_elim.
    induction 1;intros.
    apply (H4 m n H8 m1 m2 H9 n2 H10).
    injection H8;clear H8;intros.
    subst k;subst m.
    generalize (inj_left_syn h H0 H3 H5 H1 H2);intro.
    generalize (H8 m2 n2 H9 m2 l H8_0 (refl_equal m2));intro.
    subst l.
    exists n.
    split;auto.
    discriminate H8.
    discriminate H8.
    discriminate H9.
  Qed.

  Lemma enc_a_syn_left_consistent : forall (h:message_relation),(is_left_consistent_relation h)->forall (m n:MessageSpi.message),(hedge_synthesis_relation h m n)->forall (m1 m2:MessageSpi.message),(m=MessageSpi.MEnc_a m1 m2)->forall (m'2:MessageSpi.message),(MessageSpi.inverse_key m2 = Some m'2)->forall (n'2:MessageSpi.message),(hedge_synthesis_relation h m'2 n'2)->exists n2:MessageSpi.message,exists n1:MessageSpi.message,((MessageSpi.inverse_key n2 = Some n'2)/\(n=MessageSpi.MEnc_a n1 n2)/\(hedge_synthesis_relation h m1 n1)).
    intros h H.
    unfold is_left_consistent_relation in H.
    and_elim.
    induction 1;intros.
    apply (H6 m n H8 m1 m2 H9 m'2 H10 n'2 H11).
    discriminate H8.
    injection H8;clear H8;intros.
    subst k;subst m.
    generalize (inj_inv_left_syn h H0 H3 H5 H1 H2 H7);intro.
    symmetry in H9.
    generalize (H8 m2 l H8_0 m'2 n'2 H10 H9);intro.
    symmetry in H9;symmetry in H11.
    exists l.
    exists n.
    auto.
    discriminate H8.
    discriminate H9.
  Qed.

  Lemma pair_syn_left_consistent : forall (h:message_relation),(is_left_consistent_relation h)->forall (m n:MessageSpi.message),(hedge_synthesis_relation h m n)->forall (m1 m2:MessageSpi.message),(m=MessageSpi.MPair m1 m2)->exists n1:MessageSpi.message,exists n2:MessageSpi.message,((n = MessageSpi.MPair n1 n2)/\(hedge_synthesis_relation h m1 n1)/\(hedge_synthesis_relation h m2 n2)).
    intros h H.
    unfold is_left_consistent_relation in H.
    and_elim.
    induction 1;intros.
    elim (H2 m1 m2 n).
    rewrite <- H9;auto.
    discriminate H8.
    discriminate H8.
    injection H8;intros;subst m0;subst m3.
    exists n1.
    exists n2.
    auto.
    discriminate H9.
  Qed.

  (* reflexive hedges *)
  Definition is_reflexive (h:message_relation) := forall (m n:MessageSpi.message),(h m n)->m=n.

  Lemma inclusion_is_reflexive : forall (g h:message_relation),(inclusion _ g h)->(is_reflexive h)->(is_reflexive g).
    intros.
    red;intro.
    intros.
    eapply H0.
    apply H.
    auto.
  Qed.

  Definition id_left (mn:H.elt) := let (m,n) := mn in (m,m).

  Definition hedge_id_left (h:hedge) := hedge_map_single_op id_left h.

  Lemma hedge_id_left_iff : forall (h:hedge)(m n:MessageSpi.message),(H.In (m,n) (hedge_id_left h))<->((m=n) /\ (exists n':MessageSpi.message,H.In (m,n') h)).
    split.
    intros.
    elim (hedge_map_single_op_inv id_left h (m,n) H).
    intros.
    and_elim.
    destruct x;simpl in H1.
    injection H1;intros.
    subst t.
    subst n.
    split;trivial.
    exists t0.
    auto.
    intro.
    and_elim.
    ex_elim H0 n'.    
    assert (HedgeProp.Add (m,n') (H.remove (m,n') h) h);auto.
    assert (~H.In (m,n') (H.remove (m,n') h));auto.
    unfold hedge_id_left.
    unfold hedge_map_single_op.
    elim (HedgeProp.fold_2 HedgeSetoid (s:=(H.remove (m,n') h)) (s':= h) (x:=(m,n')) H.empty (compat_op_add_single_op id_left) (transpose_add_single_op id_left) H2 H1 (m,n)).
    intros.
    apply H4.
    unfold add_single_op.
    unfold id_left.
    apply H.add_1.
    rewrite <- H;auto.
  Qed.

  Lemma is_reflexive_id_left : forall (h:hedge),(is_reflexive (relation_of_hedge (hedge_id_left h))).
    unfold is_reflexive;intros.
    elim (hedge_id_left_iff h m n);intros.
    elim (H0 H);intros;auto.
  Qed.

  Lemma is_reflexive_synthesis : forall (h:hedge),(is_reflexive (relation_of_hedge h))->(is_reflexive (hedge_synthesis_relation (relation_of_hedge h))).
    unfold is_reflexive.
    intros.
    induction H0;auto;try congruence.
  Qed.

  Lemma is_reflexive_analysis : forall (h:hedge),(is_reflexive (relation_of_hedge h))->(is_reflexive (hedge_analysis_relation (relation_of_hedge h))).
    unfold is_reflexive;intros.
    induction H0;auto;try congruence.
  Qed.

  Lemma is_reflexive_analysis_step : forall (h:hedge),(is_reflexive (relation_of_hedge h))->(is_reflexive (relation_of_hedge (hedge_analysis_step h))).
    intros.
    unfold is_reflexive;intros.
    generalize (hedge_analysis_step_impl_hedge_analysis_step_spec h m n H0);intro.
    apply (is_reflexive_analysis h H).
    auto.
  Qed.
  
  Lemma is_reflexive_analysis_step_iter : forall (n:nat)(h:hedge),(is_reflexive (relation_of_hedge h))->(is_reflexive (relation_of_hedge (hedge_analysis_step_iter h n))).
    induction n;intros.
    simpl;auto.
    simpl.
    apply IHn.
    apply is_reflexive_analysis_step;auto.
  Qed.

  Lemma is_reflexive_ana : forall (h:hedge),(is_reflexive (relation_of_hedge h))->(is_reflexive (relation_of_hedge (hedge_analysis h))).
    intros.
    unfold is_reflexive.
    intros.
    elim (hedge_analysis_is_iter h).
    intro n0;intros.
    elim (p (m,n));intros.
    apply (is_reflexive_analysis_step_iter n0 h);auto.
    apply H1.
    apply H0.
  Qed.

  Lemma is_reflexive_reduce : forall (h:hedge),(is_reflexive (relation_of_hedge h))->(is_reflexive (relation_of_hedge (hedge_reduce h))).
    intros.
    unfold is_reflexive;intros.
    elim (hedge_reduce_iff h m n);intros.
    elim (H1 H0);intros.
    apply H;auto.
  Qed.

  Lemma is_reflexive_irreducible : forall (h:hedge),(is_reflexive (relation_of_hedge h))->(is_reflexive (relation_of_hedge (hedge_irreducible h))).
    unfold hedge_irreducible.
    intros.
    apply is_reflexive_reduce.
    apply is_reflexive_ana;auto.
  Qed.

  Lemma is_reflexive_left_consistent : forall (h:hedge),(is_reflexive (relation_of_hedge h))->(is_left_consistent (hedge_irreducible h)).
    unfold is_left_consistent.
    intros.
    generalize (is_reflexive_irreducible h H);intro.
    generalize (is_reflexive_synthesis (hedge_irreducible h) H0);intro.
    unfold is_left_consistent_relation.
    repeat split.
    unfold name_left_name;intros.
    exists a.
    generalize (H0 m n H2);try congruence.
    unfold injective_left;intros.
    generalize (H0 m n H2);intros.
    generalize (H0 m' n' H3);intros.
    congruence.
    unfold op_synthesis_left;intros.
    red;intro.
    elim (hedge_reduce_iff (hedge_analysis h) m n);intros.
    elim (H5 H2);intros.
    rewrite <- (H0 m n H2) in H8.
    rewrite H3 in H8.
    unfold is_synthesisable in H8.
    generalize H8;clear H8.
    case OpSpi.op_eq_dec;intros.
    rewrite <- (H1 m' n' H4) in H4.
    rewrite hedge_synthesis_spec_impl_hedge_in_synthesis in H8.
    discriminate H8.
    elim (ana_h_equiv_h_irr_h h);intros.
    unfold le_h in H10.
    unfold hedge_synthesis_spec.
    apply H10.
    apply H4.
    elim n0;auto.
    unfold pair_free_left;intros.
    red;intro.
    generalize (H0 (MessageSpi.MPair m1 m2) n H2);intro.
    subst n.
    elim (hedge_reduce_iff (hedge_analysis h) (MessageSpi.MPair m1 m2) (MessageSpi.MPair m1 m2));intros.
    elim (H3 H2);intros.
    unfold is_synthesisable in H6.
    generalize (hedge_analysis_is_analysis h).
    unfold is_analysis.
    unfold analysis_cond.
    intros.
    and_elim.
    unfold stable_analysis_step in H9.
    rewrite hedge_synthesis_spec_impl_hedge_in_synthesis in H6.
    rewrite hedge_synthesis_spec_impl_hedge_in_synthesis in H6.
    discriminate H6.
    unfold hedge_synthesis_spec.
    apply SynInc.
    apply (H9 (m2,m2)).
    apply hedge_analysis_step_spec_impl_hedge_analysis_step.
    unfold hedge_analysis_step_spec.
    eapply AnaSplitR.
    apply AnaInc.
    unfold relation_of_hedge.
    apply H5.
    unfold hedge_synthesis_spec.
    apply SynInc.
    apply (H9 (m1,m1)).
    apply hedge_analysis_step_spec_impl_hedge_analysis_step.
    unfold hedge_analysis_step_spec.
    eapply AnaSplitL.
    apply AnaInc.
    unfold relation_of_hedge.
    apply H5.
    unfold enc_s_synthesis_left;intros.
    generalize (H0 m n H2);intro.
    subst n.
    elim (hedge_reduce_iff (hedge_analysis h) m m);intros.
    elim (H4 H2);intros.
    rewrite H3 in H7.
    unfold is_synthesisable in H7.
    case (Bool_elim_bool (hedge_in_synthesis (hedge_analysis h) m' m'));intro.
    rewrite H8 in H7;simpl in H7.
    right.
    intro;red;intro.
    rewrite <- (H1 k l H9) in H9.
    elim (ana_h_equiv_h_irr_h h);intros.
    generalize (H11 k k H9);intro.
    rewrite hedge_synthesis_spec_impl_hedge_in_synthesis in H7;auto;discriminate H7.
    left.
    intro;red;intro.
    rewrite <- (H1 m' n' H9) in H9.
    elim (ana_h_equiv_h_irr_h h);intros.
    generalize (H11 m' m' H9);intro.
    rewrite hedge_synthesis_spec_impl_hedge_in_synthesis in H8;auto;discriminate H8.
    unfold enc_s_analysis_left;intros.
    generalize (H1 k l H4);intro.
    subst l.
    exists m'.
    generalize (H0 m n H2);intro;subst n.
    split;trivial.
    rewrite H3 in H2.
    elim (hedge_reduce_iff (hedge_analysis h) (MessageSpi.MEnc_s m' k) (MessageSpi.MEnc_s m' k));intros.
    elim (H5 H2);intros.
    unfold is_synthesisable in H8.
    elim (ana_h_equiv_h_irr_h h);intros.
    apply H9.
    apply SynInc.
    generalize (hedge_analysis_is_analysis h).
    unfold is_analysis, analysis_cond;intros;and_elim.
    unfold stable_analysis_step in H13.
    unfold relation_of_hedge.
    apply H13.
    apply hedge_analysis_step_spec_impl_hedge_analysis_step.
    unfold hedge_analysis_step_spec.
    eapply AnaDec_s.
    eapply AnaInc.
    unfold relation_of_hedge.
    apply H7.
    apply H10.
    auto.
    unfold enc_a_synthesis_left;intros.
    generalize (H0 m n H2);intro.
    subst n.
    elim (hedge_reduce_iff (hedge_analysis h) m m);intros.
    elim (H4 H2);intros.
    rewrite H3 in H7.
    unfold is_synthesisable in H7.
    case (Bool_elim_bool (hedge_in_synthesis (hedge_analysis h) m' m'));intro.
    rewrite H8 in H7;simpl in H7.
    right.
    intro;red;intro.
    rewrite <- (H1 k l H9) in H9.
    elim (ana_h_equiv_h_irr_h h);intros.
    generalize (H11 k k H9);intro.
    rewrite hedge_synthesis_spec_impl_hedge_in_synthesis in H7;auto;discriminate H7.
    left.
    intro;red;intro.
    rewrite <- (H1 m' n' H9) in H9.
    elim (ana_h_equiv_h_irr_h h);intros.
    generalize (H11 m' m' H9);intro.
    rewrite hedge_synthesis_spec_impl_hedge_in_synthesis in H8;auto;discriminate H8.
    unfold enc_a_analysis_left;intros.
    generalize (H1 k' l' H5);intro.
    subst l'.
    exists k.
    exists m'.
    generalize (H0 m n H2);intro;subst n.
    split;trivial.
    split;trivial.
    rewrite H3 in H2.
    elim (hedge_reduce_iff (hedge_analysis h) (MessageSpi.MEnc_a m' k) (MessageSpi.MEnc_a m' k));intros.
    elim (H6 H2);intros.
    unfold is_synthesisable in H9.
    elim (ana_h_equiv_h_irr_h h);intros.
    apply H10.
    apply SynInc.
    generalize (hedge_analysis_is_analysis h).
    unfold is_analysis, analysis_cond;intros;and_elim.
    unfold stable_analysis_step in H4.
    unfold relation_of_hedge.
    apply H14.
    apply hedge_analysis_step_spec_impl_hedge_analysis_step.
    unfold hedge_analysis_step_spec.
    eapply AnaDec_a.
    eapply AnaInc.
    unfold relation_of_hedge.
    apply H8.
    unfold inverse_keys.
    rewrite H4;reflexivity.
    apply H11.
    auto.
    unfold injective_inv_left;intros.
    generalize (H0 m n H2);intro.
    subst n.
    generalize (H0 m' n' H3);intro.
    subst n';auto.
  Qed.

  Lemma is_reflexive_transp : forall (h:message_relation),(is_reflexive h)->(same_relation _ h (transp _ h)).
    intros.
    split;unfold inclusion, transp;intros.
    generalize (H _ _ H0).
    intros.
    subst y.
    auto.
    generalize (H _ _ H0).
    intros.
    subst y.
    auto.
  Qed.

  Definition inclusion_left (g h:message_relation) := forall (m n:MessageSpi.message),(g m n)->(exists n':MessageSpi.message,(h m n')).

  Lemma inclusion_left_id_left : forall (h:hedge),(inclusion_left (relation_of_hedge h) (relation_of_hedge (hedge_id_left h))).
    unfold inclusion_left;intros.
    exists m.
    elim (hedge_id_left_iff h m m);intros.
    apply H1;split;auto.
    exists n;auto.
  Qed.
  
  Lemma inclusion_left_syn : forall (g h:message_relation),(inclusion_left g h)->(inclusion_left (hedge_synthesis_relation g) (hedge_synthesis_relation h)).
    unfold inclusion_left;intros.
    induction H0.
    elim (H m n H0);intro n';intros.
    exists n'.
    apply SynInc;auto.
    elim IHhedge_synthesis_relation1;intro m';intro.
    elim IHhedge_synthesis_relation2;intro k';intro.
    exists (MessageSpi.MEnc_s m' k').
    apply SynEnc_s;auto.
    elim IHhedge_synthesis_relation1;intro m';intro.
    elim IHhedge_synthesis_relation2;intro k';intro.
    exists (MessageSpi.MEnc_a m' k').
    apply SynEnc_a;auto.
    elim IHhedge_synthesis_relation1;intro m1';intro.
    elim IHhedge_synthesis_relation2;intro m2';intro.
    exists (MessageSpi.MPair m1' m2').
    apply SynPair;auto.
    elim IHhedge_synthesis_relation;intro m';intro.
    exists (MessageSpi.MOp o m').
    apply SynOp;auto.
  Qed.

  Lemma inclusion_left_ana_step : forall (n:nat)(g h:hedge),(inclusion_left (relation_of_hedge g) (relation_of_hedge h))->(is_reflexive (relation_of_hedge h))->(inclusion_left (relation_of_hedge (hedge_analysis_step_iter g n)) (relation_of_hedge (hedge_analysis_step_iter h n))).
    induction n;simpl;intros;auto.
    apply IHn.
    unfold inclusion_left.
    intros.
    generalize (hedge_analysis_step_impl_hedge_analysis_step_spec g m n0 H1);intro.
    cut (exists n':MessageSpi.message,hedge_analysis_step_spec h m n').
    intro.
    ex_elim H3 n'.
    exists n'.
    unfold relation_of_hedge.
    apply hedge_analysis_step_spec_impl_hedge_analysis_step;auto.
    clear H1.
    generalize (is_reflexive_analysis_step h H0);intro.
    unfold hedge_analysis_step_spec.
    induction H2.
    elim (H m n0 H2);intro n1;intro.
    exists n1.
    apply AnaInc;auto.
    elim IHhedge_analysis_relation;intro n';intro.
    generalize (H1 (MessageSpi.MPair m1 m2) n' (hedge_analysis_step_spec_impl_hedge_analysis_step h (MessageSpi.MPair m1 m2) n' H3));intro.
    subst n'.
    exists m1.
    eapply AnaSplitL;eauto.
    elim IHhedge_analysis_relation;intro n';intro.
    generalize (H1 (MessageSpi.MPair m1 m2) n' (hedge_analysis_step_spec_impl_hedge_analysis_step h (MessageSpi.MPair m1 m2) n' H3));intro.
    subst n'.
    exists m2.
    eapply AnaSplitR;eauto.
    elim IHhedge_analysis_relation;intro n';intro.
    generalize (H1 (MessageSpi.MEnc_s m k) n' (hedge_analysis_step_spec_impl_hedge_analysis_step h (MessageSpi.MEnc_s m k) n' H4));intro.
    subst n'.
    exists m.
    eapply AnaDec_s;eauto.
    elim (inclusion_left_syn (relation_of_hedge g) (relation_of_hedge h) H k l H3);intro k';intro.
    generalize (is_reflexive_synthesis h H0 k k' H5);intro;subst k';auto.
    elim IHhedge_analysis_relation;intro n';intro.
    generalize (H1 (MessageSpi.MEnc_a m k) n' (hedge_analysis_step_spec_impl_hedge_analysis_step h (MessageSpi.MEnc_a m k) n' H5));intro.
    subst n'.
    exists m.
    eapply AnaDec_a;eauto.
    unfold inverse_keys in H3.
    unfold inverse_keys.
    destruct (MessageSpi.inverse_key k);try (discriminate H3).
    destruct (MessageSpi.inverse_key l);try (discriminate H3).
    injection H3;intros;subst m0;subst m1.
    reflexivity.
    elim (inclusion_left_syn (relation_of_hedge g) (relation_of_hedge h) H k' l' H4);intro k'';intro.
    generalize (is_reflexive_synthesis h H0 k' k'' H6);intro;subst k'';auto.
    apply is_reflexive_analysis_step;auto.
  Qed.

  Lemma inclusion_left_analysis : forall (g h:hedge),(inclusion_left (relation_of_hedge g) (relation_of_hedge h))->(is_reflexive (relation_of_hedge h))->(inclusion_left (relation_of_hedge (hedge_analysis g)) (relation_of_hedge (hedge_analysis h))).
    intros.
    elim (hedge_analysis_is_iter g);intro n0;intros.
    generalize (inclusion_left_ana_step n0 g h H H0);intro. 
    unfold inclusion_left;intros.
    elim (p (m,n));intros.
    elim (H1 m n (H3 H2));intro n';intro.
    exists n'.   
    unfold relation_of_hedge.
    apply (hedge_analysis_step_iter_Subset_hedge_analysis n0 h).
    auto.
  Qed.

  Lemma inclusion_left_trans : forall (x y z:message_relation),(inclusion_left x y)->(inclusion_left y z)->(inclusion_left x z).
    unfold inclusion_left;intros.
    elim (H m n H1);intro n';intros.
    elim (H0 m n' H2);intro n'';intros.
    exists n'';auto.
  Qed.

  Lemma inclusion_impl_inclusion_left : forall (x y:message_relation),(inclusion MessageSpi.message x y)->(inclusion_left x y).
    unfold inclusion_left, inclusion;intros.
    exists n.
    apply H;auto.
  Qed.

  Lemma inclusion_left_analysis_id_left : forall (h:hedge),(inclusion_left (relation_of_hedge h) (relation_of_hedge (hedge_analysis (hedge_id_left h)))).
    intros.
    apply inclusion_left_trans with (relation_of_hedge (hedge_analysis h)).
    apply inclusion_impl_inclusion_left.
    generalize (hedge_analysis_is_analysis h).
    unfold is_analysis, analysis_cond;intros;and_elim.
    unfold inclusion;intros.
    unfold relation_of_hedge.
    apply H.
    apply H2.
    apply inclusion_left_analysis.
    apply inclusion_left_id_left.
    apply is_reflexive_id_left.
  Qed.

  Lemma inclusion_id_left_left : forall (h:hedge),(inclusion_left (relation_of_hedge (hedge_id_left h)) (relation_of_hedge h)).
    unfold inclusion_left;intros.
    elim (hedge_id_left_iff h m n);intros.
    elim (H0 H);intros.
    auto.
  Qed.

  Lemma inclusion_left_id_left_analysis_impl : forall (h:hedge),(injective_left (hedge_synthesis_relation (relation_of_hedge h)))->(inclusion_left (relation_of_hedge (hedge_analysis (hedge_id_left h))) (relation_of_hedge h))->(enc_s_not_ana_left_aux (relation_of_hedge h)).
    intro h.
    intro Hinj.
    intros.
    generalize (inclusion_left_analysis_id_left h);intro.
    generalize (is_reflexive_id_left h);intro.
    generalize (is_reflexive_ana (hedge_id_left h) H1);intro.
    generalize (is_reflexive_synthesis (hedge_id_left h) H1);intro.
    generalize (inclusion_left_syn (relation_of_hedge h) (relation_of_hedge (hedge_id_left h)) (inclusion_left_id_left h));intro.
    generalize (inclusion_left_syn (relation_of_hedge (hedge_id_left h)) (relation_of_hedge h) (inclusion_id_left_left h));intro.
    unfold enc_s_not_ana_left_aux;intros.
    elim (H0 m n H6);intro n';intro.
    generalize (H2 m n' H9);intro.
    subst n'.
    elim (H4 k l H8);intro k';intro.
    generalize (H3 k k' H10);intro.
    subst k'.
    generalize (hedge_analysis_is_analysis (hedge_id_left h)).
    unfold is_analysis, analysis_cond;intros;and_elim.
    elim (H m' m').
    intro n';intro.
    exists n'.
    split.
    apply SynInc;auto.
    apply (Hinj m n (SynInc (relation_of_hedge h) m n H6) (MessageSpi.MEnc_s m' k) (MessageSpi.MEnc_s n' l) (SynEnc_s (relation_of_hedge h) m' n' k l (SynInc (relation_of_hedge h) m' n' H14) H8) H7).
    unfold stable_analysis_step in H13.
    unfold relation_of_hedge.
    apply H13.
    apply hedge_analysis_step_spec_impl_hedge_analysis_step.
    unfold hedge_analysis_step_spec.
    eapply AnaDec_s.
    apply AnaInc.
    rewrite H7 in H9.
    apply H9.
    generalize (h_le_h_ana_h (hedge_id_left h));intro.
    apply H14;auto.
  Qed.

  Lemma inclusion_left_syn_impl_inclusion_left : forall (g h:message_relation),(inclusion_left g (hedge_synthesis_relation h))->(inclusion_left (hedge_synthesis_relation g) (hedge_synthesis_relation h)).
    unfold inclusion_left;intros.
    induction H0.
    eapply H;eauto.
    elim IHhedge_synthesis_relation1;intro m';intro.
    elim IHhedge_synthesis_relation2;intro k';intro.
    exists (MessageSpi.MEnc_s m' k').
    apply SynEnc_s;auto.
    elim IHhedge_synthesis_relation1;intro m';intro.
    elim IHhedge_synthesis_relation2;intro k';intro.
    exists (MessageSpi.MEnc_a m' k').
    apply SynEnc_a;auto.
    elim IHhedge_synthesis_relation1;intro m'1;intro.
    elim IHhedge_synthesis_relation2;intro m'2;intro.
    exists (MessageSpi.MPair m'1 m'2).
    apply SynPair;auto.
    elim IHhedge_synthesis_relation;intro m';intro.
    exists (MessageSpi.MOp o m').
    apply SynOp;auto.
  Qed.
  
  Lemma left_consistent_impl_inclusion_left_ana_step : forall (h:hedge),(is_left_consistent h)->forall (n:nat),(inclusion_left (relation_of_hedge (hedge_analysis_step_iter (hedge_id_left h) n)) (hedge_synthesis_relation (relation_of_hedge h))).
    unfold is_left_consistent, is_left_consistent_relation;intros;and_elim.
    induction n.
    simpl.
    unfold inclusion_left;intros.
    elim (hedge_id_left_iff h m n);intros.
    elim (H9 H8);intros.
    ex_elim H12 m'.
    exists m'.
    apply SynInc;auto.
    unfold inclusion_left;intros.
    elim (hedge_analysis_step_iter_Sn n (hedge_id_left h) (m,n0));intros.
    generalize (H9 H8);intro.
    generalize (hedge_analysis_step_impl_hedge_analysis_step_spec (hedge_analysis_step_iter (hedge_id_left h) n) m n0 H11);intro.
    clear H8;clear H9;clear H10;clear H11.
    generalize (inclusion_left_syn_impl_inclusion_left (relation_of_hedge (hedge_analysis_step_iter (hedge_id_left h) n)) (relation_of_hedge h) IHn);intro.
    induction H12.
    eapply IHn;eauto.
    elim IHhedge_analysis_relation;intro n';intro.
    inversion H9.
    elim (H2 m1 m2 n' H10).
    exists n0;auto.
    elim IHhedge_analysis_relation;intro n';intro.
    inversion H9.
    elim (H2 m1 m2 n' H10).
    exists n3;auto.
    elim IHhedge_analysis_relation;intro n';intro.
    inversion H10.
    elim (H8 k l H9);intro l';intro.
    elim (H4 (MessageSpi.MEnc_s m k) n' H11 m k (refl_equal (MessageSpi.MEnc_s m k)) l' H15);intro m';intro;and_elim.
    exists m';auto.
    exists n1;auto.
    elim IHhedge_analysis_relation;intro n';intro.
    inversion H11.
    elim (H8 k' l' H10);intro l'';intro.
    assert (MessageSpi.inverse_key k = Some k').
    unfold inverse_keys in H9.
    destruct (MessageSpi.inverse_key k);try (discriminate H9).
    destruct (MessageSpi.inverse_key l);try (discriminate H9).
    injection H9;intros;try congruence.
    elim (H6 (MessageSpi.MEnc_a m k) n' H13 m k (refl_equal (MessageSpi.MEnc_a m k)) k' H17 l'' H16);intro k'';intro.
    ex_elim H18 m'.
    exists m';auto.
    exists n1;auto.
  Qed.

  Lemma left_consistent_impl_inclusion_left_analysis : forall (h:hedge),(is_left_consistent h)->(inclusion_left (relation_of_hedge (hedge_analysis (hedge_id_left h))) (hedge_synthesis_relation (relation_of_hedge h))).
    intros.
    elim (hedge_analysis_is_iter (hedge_id_left h));intro n0;intro.
    unfold inclusion_left;intros.
    elim (p (m,n));intros.
    generalize (H1 H0);intro.
    apply (left_consistent_impl_inclusion_left_ana_step h H n0 m n H3).
  Qed.

  (** Decidability of consistency *)
  Definition name_left_name_pred (mn:H.elt) : bool :=
    let (m,n) := mn in
      match m with
	| (MessageSpi.MNam a) =>
	  (match n with
	     | (MessageSpi.MNam b) => true
	     | _ => false
	   end)
	| _ => true
      end.

  Lemma name_left_name_pred_compat_bool : FSetInterface.compat_bool H.E.eq name_left_name_pred.
    unfold FSetInterface.compat_bool.
    intros.
    destruct x;destruct y;simpl in H.
    and_elim.
    simpl.
    rewrite H;rewrite H0.
    trivial.
  Qed.

  Definition name_left_name_proc (h:hedge) := H.for_all name_left_name_pred h.

  Lemma name_left_name_ok : forall (h:hedge),(name_left_name (relation_of_hedge h))<->(name_left_name_proc h=true).
    split;intros.
    unfold name_left_name in H.
    unfold name_left_name_proc.
    apply H.for_all_1.
    apply name_left_name_pred_compat_bool.
    unfold H.For_all.
    intros.
    destruct x.
    destruct t;try reflexivity.
    elim (H (MessageSpi.MNam t) t0 H0 t);trivial.
    intros.
    rewrite H1.
    reflexivity.
    unfold name_left_name_proc in H.
    generalize (H.for_all_2 name_left_name_pred_compat_bool H).
    unfold H.For_all.
    intros.
    unfold name_left_name.
    intros.
    subst m.
    generalize (H0 ((MessageSpi.MNam a),n) H1).
    simpl.
    destruct n;intro H2;try (discriminate H2).
    exists t;trivial.
  Qed.

  Definition injective_left_pred (mn:H.elt) (mn':H.elt) :=
    let (m,n) := mn in
      let (m',n') := mn' in
	if MessageSpi.eq_dec m m' then
	  if MessageSpi.eq_dec n n' then true else false
	  else true.

  Lemma injective_left_pred_compat_bool : forall (mn:H.elt),(FSetInterface.compat_bool H.E.eq (injective_left_pred mn)).
    intros.
    unfold FSetInterface.compat_bool.
    intros.
    destruct x;destruct y;simpl in H;and_elim.
    rewrite H;rewrite H0.
    trivial.
  Qed.

  Lemma injective_left_pred_hedge_compat_bool : forall (h:hedge),(FSetInterface.compat_bool H.E.eq (fun mn => H.for_all (injective_left_pred mn) h)).
    unfold FSetInterface.compat_bool.
    intros.
    destruct x;destruct y;simpl in H;and_elim.
    rewrite H;rewrite H0;trivial.
  Qed.

  Definition injective_left_proc (h:hedge) := H.for_all (fun mn => H.for_all (injective_left_pred mn) h) h.

  Lemma injective_left_ok : forall (h:hedge),(injective_left (relation_of_hedge h))<->(injective_left_proc h=true).
    unfold injective_left_proc, injective_left.
    split;intros.
    apply H.for_all_1.
    apply injective_left_pred_hedge_compat_bool.
    unfold H.For_all.
    intros.
    apply H.for_all_1.
    apply injective_left_pred_compat_bool.
    unfold H.For_all.
    intros.
    destruct x;destruct x0.
    simpl.
    case (MessageSpi.eq_dec t t1);intro;trivial.
    case (MessageSpi.eq_dec t0 t2);intro;trivial.
    elim n.
    eapply H;unfold relation_of_hedge.
    apply H0.
    apply H1.
    trivial.
    generalize (H.for_all_2 (injective_left_pred_hedge_compat_bool h) H).
    unfold H.For_all;intros.
    generalize (H3 (m,n) H0);intro.
    generalize (H.for_all_2 (injective_left_pred_compat_bool (m,n)) H4).
    unfold H.For_all;intros.
    generalize (H5 (m',n') H1);intro.
    subst m'.
    simpl in H6.
    generalize H6.
    case (MessageSpi.eq_dec m m);intro.
    case (MessageSpi.eq_dec n n');intros;auto.
    discriminate H2.
    elim n0;trivial.
  Qed.

  Definition pair_free_left_pred (mn:H.elt) :=
    let (m,n) := mn in
      match m with
	| (MessageSpi.MPair m1 m2) => false
	| _ => true
      end.

  Lemma pair_free_left_pred_compat_bool : FSetInterface.compat_bool H.E.eq pair_free_left_pred.
    unfold FSetInterface.compat_bool.
    intros.
    destruct x;destruct y;simpl in H;and_elim.
    rewrite H;rewrite H0;trivial.
  Qed.

  Definition pair_free_left_proc (h:hedge) := H.for_all pair_free_left_pred h.

  Lemma pair_free_left_ok : forall (h:hedge),(pair_free_left (relation_of_hedge h))<->(pair_free_left_proc h=true).
    unfold pair_free_left_proc, pair_free_left.
    split;intros.
    apply H.for_all_1.
    apply pair_free_left_pred_compat_bool.
    unfold H.For_all.
    intros.
    destruct x.
    destruct t;simpl;trivial.
    elim (H t1 t2 t0).
    apply H0.
    intro.
    generalize (H.for_all_2 pair_free_left_pred_compat_bool H).
    unfold H.For_all;intros.
    generalize (H1 (MessageSpi.MPair m1 m2,n) H0).
    intro.
    discriminate H2.
  Qed.

  Definition injective_inv_left_pred (mn:H.elt) (mn':H.elt) :=
    let (m,n) := mn in
      let (m',n') := mn' in
        match MessageSpi.inverse_key m with
          | Some minv =>
            if MessageSpi.eq_dec minv m' then
              match MessageSpi.inverse_key n with
                | Some ninv => 
                  if MessageSpi.eq_dec ninv n' then true else false
                | None => false
              end
            else true
          | None => true
        end.

  Lemma injective_inv_left_pred_compat_bool : forall (mn:H.elt),(FSetInterface.compat_bool H.E.eq (injective_inv_left_pred mn)).
    intros.
    unfold FSetInterface.compat_bool.
    intros.
    destruct x;destruct y;simpl in H;and_elim.
    rewrite H;rewrite H0.
    trivial.
  Qed.

  Lemma injective_inv_left_pred_hedge_compat_bool : forall (h:hedge),(FSetInterface.compat_bool H.E.eq (fun mn => H.for_all (injective_inv_left_pred mn) h)).
    unfold FSetInterface.compat_bool.
    intros.
    destruct x;destruct y;simpl in H;and_elim.
    rewrite H;rewrite H0;trivial.
  Qed.

  Definition injective_inv_left_proc (h:hedge) := H.for_all (fun mn => H.for_all (injective_inv_left_pred mn) h) h.

  Lemma injective_inv_left_ok : forall (h:hedge),(injective_inv_left (relation_of_hedge h))<->(injective_inv_left_proc h=true).
    unfold injective_inv_left_proc, injective_inv_left.
    split;intros.
    apply H.for_all_1.
    apply injective_inv_left_pred_hedge_compat_bool.
    unfold H.For_all.
    intros.
    apply H.for_all_1.
    apply injective_inv_left_pred_compat_bool.
    unfold H.For_all.
    intros.
    destruct x;destruct x0.
    simpl.
    generalize (H t t0 H0 t1 t2 H1).
    destruct (MessageSpi.inverse_key t);destruct (MessageSpi.inverse_key t0);intros;auto.
    case (MessageSpi.eq_dec m t1);intro;trivial.
    case (MessageSpi.eq_dec m0 t2);intro;trivial.
    elim n.
    rewrite e in H2.
    generalize (H2 (refl_equal (Some t1))).
    intros.
    injection H3;auto.
    case (MessageSpi.eq_dec m t1);intro;trivial.
    rewrite e in H2.
    generalize (H2 (refl_equal (Some t1))).
    intro.
    discriminate H3.
    generalize (H.for_all_2 (injective_inv_left_pred_hedge_compat_bool h) H).
    unfold H.For_all;intros.
    generalize (H3 (m,n) H0);intro.
    generalize (H.for_all_2 (injective_inv_left_pred_compat_bool (m,n)) H4).
    unfold H.For_all;intros.
    generalize (H5 (m',n') H1).
    unfold injective_inv_left_pred.
    rewrite <- H2.
    case (MessageSpi.eq_dec m' m');intro.
    destruct (MessageSpi.inverse_key n).
    case (MessageSpi.eq_dec m0 n');intro.
    subst m0;auto.
    intro.
    discriminate H6.
    intro.
    discriminate H6.
    elim n0.
    auto.
  Qed.

  Definition op_synthesis_left_pred (h:hedge)(mn:H.elt) :=
    let (m,n) := mn in
      match m with
	| (MessageSpi.MOp o m) => H.is_empty (hedge_synthesis_left h m)
	| _ => true
      end.

  Lemma op_synthesis_left_pred_compat_bool : forall (h:hedge),(FSetInterface.compat_bool H.E.eq (op_synthesis_left_pred h)).
    unfold FSetInterface.compat_bool.
    intros.
    destruct x;destruct y;simpl in H;and_elim.
    rewrite H;rewrite H0;trivial.
  Qed.

  Definition op_synthesis_left_proc (h:hedge) := H.for_all (op_synthesis_left_pred h) h.

  Lemma op_synthesis_left_ok : forall (h:hedge),(op_synthesis_left (relation_of_hedge h))<->(op_synthesis_left_proc h = true).
    unfold op_synthesis_left, op_synthesis_left_proc.
    split;intros.
    apply H.for_all_1.
    apply op_synthesis_left_pred_compat_bool.
    unfold H.For_all;intros.
    destruct x as (m,n).
    destruct m;try reflexivity.
    simpl.
    apply H.is_empty_1.
    elim (hedge_synthesis_left_empty h m);intros.
    apply H2;intro.
    unfold hedge_synthesis_spec.
    eapply H.
    unfold relation_of_hedge.
    apply H0.
    reflexivity.
    generalize (H.for_all_2 (op_synthesis_left_pred_compat_bool h) H).
    unfold H.For_all;intros.
    generalize (H2 (m,n) H0);intro.
    rewrite H1 in H3;simpl in H3.
    generalize (H.is_empty_2 H3);intro.
    elim (hedge_synthesis_left_empty h m');intros.
    fold (hedge_synthesis_spec h);auto.
  Qed.

  Definition enc_s_synthesis_left_pred (h:hedge)(mn:H.elt) :=
    let (m,n) := mn in
      match m with
	| (MessageSpi.MEnc_s m k) => orb (H.is_empty (hedge_synthesis_left h m)) (H.is_empty (hedge_synthesis_left h k))
	| _ => true
      end.

  Lemma enc_s_synthesis_left_pred_compat_bool : forall (h:hedge),(FSetInterface.compat_bool H.E.eq (enc_s_synthesis_left_pred h)).
    unfold FSetInterface.compat_bool.
    intros.
    destruct x;destruct y;simpl in H;and_elim.
    rewrite H;rewrite H0;trivial.
  Qed.

  Definition enc_s_synthesis_left_proc (h:hedge) := H.for_all (enc_s_synthesis_left_pred h) h.

  Lemma enc_s_synthesis_left_ok : forall (h:hedge),(enc_s_synthesis_left (relation_of_hedge h))<->(enc_s_synthesis_left_proc h = true).
    unfold enc_s_synthesis_left, enc_s_synthesis_left_proc.
    split;intros.
    apply H.for_all_1.
    apply enc_s_synthesis_left_pred_compat_bool.
    unfold H.For_all;intros.
    destruct x as (m,n).
    destruct m;try reflexivity.
    simpl.
    apply orb_true_intro.
    generalize (H (MessageSpi.MEnc_s m1 m2) n H0 m1 m2 (refl_equal (MessageSpi.MEnc_s m1 m2)));intro.
    case H1;clear H1;intro;[left | right];apply H.is_empty_1.
    elim (hedge_synthesis_left_empty h m1);auto.
    elim (hedge_synthesis_left_empty h m2);auto.
    generalize (H.for_all_2 (enc_s_synthesis_left_pred_compat_bool h) H);unfold H.For_all;intros.
    generalize (H2 (m,n) H0);subst m;simpl;intros.
    case (orb_prop (H.is_empty (hedge_synthesis_left h m')) (H.is_empty (hedge_synthesis_left h k)) H1);intro;generalize (H.is_empty_2 H3);intro;[left | right];intros;fold (hedge_synthesis_spec h).
    elim (hedge_synthesis_left_empty h m');auto.
    elim (hedge_synthesis_left_empty h k);auto.
  Qed.

  Definition enc_s_analysis_left_pred_aux (h:hedge)(m n:MessageSpi.message)(ikl:H.elt) :=
    let (ik,il) := ikl in
      match n with
	| (MessageSpi.MEnc_s n l) =>
	  if MessageSpi.eq_dec l il then hedge_in_synthesis h m n
	    else false
	| _ => false
      end.

  Lemma enc_s_analysis_left_pred_aux_compat_bool : forall (h:hedge)(m n:MessageSpi.message),FSetInterface.compat_bool H.E.eq (enc_s_analysis_left_pred_aux h m n).
    unfold FSetInterface.compat_bool;intros.
    destruct x;destruct y;simpl in H;and_elim.
    rewrite H;rewrite H0;trivial.
  Qed.

  Definition enc_s_analysis_left_pred (h:hedge)(mn:H.elt) :=
    let (m,n) := mn in
      match m with
	| (MessageSpi.MEnc_s m k) =>
	  H.for_all (enc_s_analysis_left_pred_aux h m n) (hedge_synthesis_left h k)
	| _ => true
      end.

  Lemma enc_s_analysis_left_pred_compat_bool : forall (h:hedge),FSetInterface.compat_bool H.E.eq (enc_s_analysis_left_pred h).
    unfold FSetInterface.compat_bool;intros.
    destruct x;destruct y;simpl in H;and_elim.
    rewrite H;rewrite H0;trivial.
  Qed.

  Definition enc_s_analysis_left_proc (h:hedge) := H.for_all (enc_s_analysis_left_pred h) h.

  Lemma enc_s_analysis_left_ok : forall (h:hedge),(enc_s_analysis_left (relation_of_hedge h))<->(enc_s_analysis_left_proc h = true).
    unfold enc_s_analysis_left, enc_s_analysis_left_proc;intros.
    split;intros.
    apply H.for_all_1.
    apply enc_s_analysis_left_pred_compat_bool.
    unfold H.For_all;intros.
    destruct x as (m,n).
    destruct m;try reflexivity.
    simpl.
    apply H.for_all_1.
    apply enc_s_analysis_left_pred_aux_compat_bool.
    unfold H.For_all;intros.
    destruct x as (ik,il).
    elim (hedge_synthesis_left_ok h m2 ik il);intros.
    elim (H2 H1);intros.
    fold (hedge_synthesis_spec h) in H.
    elim (H (MessageSpi.MEnc_s m1 m2) n H0 m1 m2) with il;trivial.
    intro n1;intro;and_elim.
    rewrite H6;simpl.
    case MessageSpi.eq_dec;intro.
    apply hedge_synthesis_spec_impl_hedge_in_synthesis;trivial.
    elim n0;trivial.
    generalize (H.for_all_2 (enc_s_analysis_left_pred_compat_bool h) H);unfold H.For_all;intros.
    generalize (H3 (m,n) H0);intro.
    rewrite H1 in H4;simpl in H4.
    generalize (H.for_all_2 (enc_s_analysis_left_pred_aux_compat_bool h m' n) H4);unfold H.For_all;intros.
    fold (hedge_synthesis_spec h) in H2.
    elim (hedge_synthesis_left_ok h k k l);intros.
    assert (H.In (k,l) (hedge_synthesis_left h k)).
    apply H7.
    split;trivial.
    generalize (H5 (k,l) H8).
    destruct n;simpl;intros;try (discriminate H9).
    generalize H9.
    case MessageSpi.eq_dec;intros.
    exists n1.
    rewrite <- e.
    split;trivial.
    apply hedge_in_synthesis_impl_hedge_synthesis_spec;trivial.
    discriminate H10.
  Qed.

  Definition enc_a_synthesis_left_pred (h:hedge)(mn:H.elt) :=
    let (m,n) := mn in
      match m with
	| (MessageSpi.MEnc_a m k) => orb (H.is_empty (hedge_synthesis_left h m)) (H.is_empty (hedge_synthesis_left h k))
	| _ => true
      end.

  Lemma enc_a_synthesis_left_pred_compat_bool : forall (h:hedge),(FSetInterface.compat_bool H.E.eq (enc_a_synthesis_left_pred h)).
    unfold FSetInterface.compat_bool.
    intros.
    destruct x;destruct y;simpl in H;and_elim.
    rewrite H;rewrite H0;trivial.
  Qed.

  Definition enc_a_synthesis_left_proc (h:hedge) := H.for_all (enc_a_synthesis_left_pred h) h.

  Lemma enc_a_synthesis_left_ok : forall (h:hedge),(enc_a_synthesis_left (relation_of_hedge h))<->(enc_a_synthesis_left_proc h = true).
    unfold enc_a_synthesis_left, enc_a_synthesis_left_proc.
    split;intros.
    apply H.for_all_1.
    apply enc_a_synthesis_left_pred_compat_bool.
    unfold H.For_all;intros.
    destruct x as (m,n).
    destruct m;try reflexivity.
    simpl.
    apply orb_true_intro.
    generalize (H (MessageSpi.MEnc_a m1 m2) n H0 m1 m2 (refl_equal (MessageSpi.MEnc_a m1 m2)));intro.
    case H1;clear H1;intro;[left | right];apply H.is_empty_1.
    elim (hedge_synthesis_left_empty h m1);auto.
    elim (hedge_synthesis_left_empty h m2);auto.
    generalize (H.for_all_2 (enc_a_synthesis_left_pred_compat_bool h) H);unfold H.For_all;intros.
    generalize (H2 (m,n) H0);subst m;simpl;intros.
    case (orb_prop (H.is_empty (hedge_synthesis_left h m')) (H.is_empty (hedge_synthesis_left h k)) H1);intro;generalize (H.is_empty_2 H3);intro;[left | right];intros;fold (hedge_synthesis_spec h).
    elim (hedge_synthesis_left_empty h m');auto.
    elim (hedge_synthesis_left_empty h k);auto.
  Qed.

  Definition enc_a_analysis_left_pred_aux (h:hedge)(m n:MessageSpi.message)(ikl:H.elt) :=
    let (ik,il) := ikl in
      match n with
	| (MessageSpi.MEnc_a n l) =>
          match MessageSpi.inverse_key l with
            | Some(linv) => if MessageSpi.eq_dec linv il then hedge_in_synthesis h m n else false
            | None => false
          end
	| _ => false
      end.

  Lemma enc_a_analysis_left_pred_aux_compat_bool : forall (h:hedge)(m n:MessageSpi.message),FSetInterface.compat_bool H.E.eq (enc_a_analysis_left_pred_aux h m n).
    unfold FSetInterface.compat_bool;intros.
    destruct x;destruct y;simpl in H;and_elim.
    rewrite H;rewrite H0;trivial.
  Qed.

  Definition enc_a_analysis_left_pred (h:hedge)(mn:H.elt) :=
    let (m,n) := mn in
      match m with
	| (MessageSpi.MEnc_a m k) =>
          match MessageSpi.inverse_key k with
            | Some(kinv) => H.for_all (enc_a_analysis_left_pred_aux h m n) (hedge_synthesis_left h kinv)
            | None => true
          end
	| _ => true
      end.

  Lemma enc_a_analysis_left_pred_compat_bool : forall (h:hedge),FSetInterface.compat_bool H.E.eq (enc_a_analysis_left_pred h).
    unfold FSetInterface.compat_bool;intros.
    destruct x;destruct y;simpl in H;and_elim.
    rewrite H;rewrite H0;trivial.
  Qed.

  Definition enc_a_analysis_left_proc (h:hedge) := H.for_all (enc_a_analysis_left_pred h) h.

  Lemma enc_a_analysis_left_ok : forall (h:hedge),(enc_a_analysis_left (relation_of_hedge h))<->(enc_a_analysis_left_proc h = true).
    unfold enc_a_analysis_left, enc_a_analysis_left_proc;intros.
    split;intros.
    apply H.for_all_1.
    apply enc_a_analysis_left_pred_compat_bool.
    unfold H.For_all;intros.
    destruct x as (m,n).
    destruct m;try reflexivity.
    simpl.
    generalize (H _ _ H0 _ _ (refl_equal _)).
    destruct (MessageSpi.inverse_key m2);intros;auto.
    apply H.for_all_1.
    apply enc_a_analysis_left_pred_aux_compat_bool.
    unfold H.For_all;intros.
    destruct x as (ik,il).
    elim (hedge_synthesis_left_ok h m ik il);intros.
    elim (H3 H2);intros.
    fold (hedge_synthesis_spec h) in H1.
    elim (H1 _ (refl_equal _) _ H6).
    intro n1;intro.
    ex_elim H7 l.
    rewrite H8;simpl.
    rewrite H7.
    case MessageSpi.eq_dec;intro.
    apply hedge_synthesis_spec_impl_hedge_in_synthesis;trivial.
    elim n0;trivial.
    generalize (H.for_all_2 (enc_a_analysis_left_pred_compat_bool h) H);unfold H.For_all;intros.
    generalize (H4 (m,n) H0);intro.
    rewrite H1 in H5;simpl in H5.
    rewrite H2 in H5.
    generalize (H.for_all_2 (enc_a_analysis_left_pred_aux_compat_bool h m' n) H5);unfold H.For_all;intros.
    fold (hedge_synthesis_spec h) in H3.
    elim (hedge_synthesis_left_ok h k' k' l');intros.
    assert (H.In (k',l') (hedge_synthesis_left h k')).
    apply H8.
    split;trivial.
    generalize (H6 (k',l') H9).
    destruct n;simpl;intros;try (discriminate H10).
    exists n2.
    exists n1.
    destruct (MessageSpi.inverse_key n2);try (discriminate H10).
    generalize H10;clear H10.
    case MessageSpi.eq_dec;intros;try (discriminate H10).
    rewrite e.
    split;trivial.
    split;trivial.
    apply hedge_in_synthesis_impl_hedge_synthesis_spec;trivial.
  Qed.

  Definition hedge_is_left_consistent (h:hedge) := andb (name_left_name_proc h) (andb (injective_left_proc h) (andb (op_synthesis_left_proc h) (andb (pair_free_left_proc h) (andb (enc_s_synthesis_left_proc h) (andb (enc_s_analysis_left_proc h) (andb (enc_a_synthesis_left_proc h) (andb (enc_a_analysis_left_proc h) (injective_inv_left_proc h)))))))).

  Definition hedge_is_left_consistent_ok : forall (h:hedge),(is_left_consistent h)<->(hedge_is_left_consistent h = true).
    unfold is_left_consistent.
    unfold is_left_consistent_relation.
    unfold hedge_is_left_consistent.
    intro h.
    elim (name_left_name_ok h);intros.
    elim (injective_left_ok h);intros.
    elim (op_synthesis_left_ok h);intros.
    elim (pair_free_left_ok h);intros.
    elim (enc_s_synthesis_left_ok h);intros.
    elim (enc_s_analysis_left_ok h);intros.
    elim (enc_a_synthesis_left_ok h);intros.
    elim (enc_a_analysis_left_ok h);intros.
    elim (injective_inv_left_ok h);intros.
    split;intros;and_elim.
    repeat (apply andb_true_intro;split;auto).
    elim (andb_prop _ _ H17);clear H17;intros.
    elim (andb_prop _ _ H18);clear H18;intros.
    elim (andb_prop _ _ H19);clear H19;intros.
    elim (andb_prop _ _ H20);clear H20;intros.
    elim (andb_prop _ _ H21);clear H21;intros.
    elim (andb_prop _ _ H22);clear H22;intros.
    elim (andb_prop _ _ H23);clear H23;intros.
    elim (andb_prop _ _ H24);clear H24;intros.
    repeat (split;auto).
  Qed.

  Theorem hedge_is_left_consistent_dec : forall (h:hedge),{is_left_consistent h}+{~(is_left_consistent h)}.
    intro h.
    case (Bool_eq_dec (hedge_is_left_consistent h));[left | right];intros.
    elim (hedge_is_left_consistent_ok h);auto.
    elim (hedge_is_left_consistent_ok h);intros.
    intro.
    rewrite H in e;trivial.
    discriminate e.
  Defined.

  Definition hedge_is_consistent (h:hedge) := andb (hedge_is_left_consistent h) (hedge_is_left_consistent (hedge_transpose h)).

  Definition hedge_is_consistent_ok : forall (h:hedge),(is_consistent h)<->(hedge_is_consistent h = true).
    intro.
    elim (is_consistent_iff h);intros.
    elim (hedge_is_left_consistent_ok h);intros.
    elim (hedge_is_left_consistent_ok (hedge_transpose h));intros.
    unfold hedge_is_consistent.
    split;intros.
    elim (H H5);intros.
    repeat (apply andb_true_intro;split;auto).
    apply H0.
    elim (andb_prop _ _ H5);intros.
    split;auto.
  Qed.

  (** hedges containing only names *)
  Definition hedge_subset_NxN (h:hedge) := forall (m n:MessageSpi.message),(H.In (m,n) h)->(exists a:Name.t,exists b:Name.t,m=(MessageSpi.MNam a) /\ n=(MessageSpi.MNam b)).

  Lemma hedge_subset_NxN_Subset : forall (h h':hedge),(H.Subset h h')->(hedge_subset_NxN h')->(hedge_subset_NxN h).
    unfold hedge_subset_NxN;intros.
    elim (H0 m n (H (m,n) H1)).
    intro a;intros.
    ex_elim H2 b.
    exists a;exists b.
    split;auto.
  Qed.

  Definition is_in_NxN (mn:H.elt) :=
    let (m,n) := mn in
      match (m,n) with
	| (MessageSpi.MNam _,MessageSpi.MNam _) => true
	| _ => false
      end.

  Lemma compat_bool_is_in_NxN : FSetInterface.compat_bool H.E.eq is_in_NxN.
    unfold FSetInterface.compat_bool.
    destruct x;destruct y;simpl;intros.
    and_elim.
    rewrite H;rewrite H0.
    trivial.
  Qed.

  Definition is_hedge_subset_NxN (h:hedge) := H.for_all is_in_NxN h.

  Lemma hedge_subset_NxN_iff : forall (h:hedge),(hedge_subset_NxN h)<->(is_hedge_subset_NxN h = true).
    split;intros.
    unfold is_hedge_subset_NxN.
    apply H.for_all_1.
    apply compat_bool_is_in_NxN.
    unfold H.For_all.
    destruct x;intros.
    elim (H t t0 H0).
    intro a;intro.
    ex_elim H1 b.
    subst t;subst t0.
    reflexivity.
    unfold hedge_subset_NxN.
    intros.
    unfold is_hedge_subset_NxN in H.
    generalize (H.for_all_2 compat_bool_is_in_NxN H).
    unfold H.For_all;intros.
    generalize (H1 (m,n) H0).
    unfold is_in_NxN.
    destruct m;intros H2;try (discriminate H2).
    destruct n;try (discriminate H2).
    exists t.
    exists t0.
    tauto.
  Qed.

  Lemma hedge_is_left_consistent_NxN : forall (h:hedge),(hedge_subset_NxN h)->((is_left_consistent h)<->(injective_left (relation_of_hedge h))).
    intros.
    unfold is_left_consistent;unfold is_left_consistent_relation.
    split;intro;try tauto.
    split.
    unfold name_left_name.
    intros.
    unfold relation_of_hedge in H1.
    elim (H m n H1).
    intro a';intro.
    ex_elim H3 b;intros.
    exists b;trivial.
    split;trivial.
    split.
    unfold op_synthesis_left.
    intros.
    unfold relation_of_hedge in H1.
    elim (H m n H1).
    intro a;intro.
    ex_elim H3 b;intros.
    subst m.
    discriminate H3.
    split.
    unfold pair_free_left.
    intros.
    intro.
    unfold relation_of_hedge in H1.
    elim (H (MessageSpi.MPair m1 m2) n H1);intro a;intro.
    ex_elim H2 b;intros.
    discriminate H2.
    split.
    unfold enc_s_synthesis_left.
    intros.
    unfold relation_of_hedge in H1.
    elim (H m n H1);intro a;intros.
    ex_elim H3 b.
    subst m.
    discriminate H3.
    split.
    unfold enc_s_analysis_left.
    intros.
    unfold relation_of_hedge in H1.
    elim (H m n H1).
    intro a;intro.
    ex_elim H4 b;intros.
    subst m.
    discriminate H4.
    split.
    unfold enc_a_synthesis_left.
    intros.
    unfold relation_of_hedge in H1.
    elim (H m n H1);intro a;intros.
    ex_elim H3 b.
    subst m.
    discriminate H3.
    split.
    unfold enc_a_analysis_left.
    intros.
    unfold relation_of_hedge in H1.
    elim (H m n H1).
    intro a;intro.
    ex_elim H5 b;intros.
    subst m.
    discriminate H5.
    unfold injective_inv_left;intros.
    elim (H m n H1);intro a;intros.
    ex_elim H4 b;intros.
    subst m.
    simpl in H3.
    discriminate H3.
  Qed.

  Lemma hedge_transpose_NxN (h:hedge) : (hedge_subset_NxN h)->(hedge_subset_NxN (hedge_transpose h)).
    unfold hedge_subset_NxN.
    intros.
    elim (hedge_transpose_prop h n m);intros.
    generalize (H2 H0);clear H1;clear H2.
    intro.
    elim (H n m H1);intro a;intros.
    ex_elim H2 b.
    exists b.
    exists a.
    auto.
  Qed.

  Definition injective_both (h:message_relation) := forall (m n:MessageSpi.message),(h m n)->forall (m' n':MessageSpi.message),(h m' n')->((m=m')<->(n=n')).

  Lemma injective_both_inclusion : forall (h h':message_relation),(inclusion _ h h')->(injective_both h')->(injective_both h).
    unfold injective_both;intros.
    apply H0;auto.
  Qed.

  Lemma hedge_is_consistent_NxN : forall (h:hedge),(hedge_subset_NxN h)->((is_consistent h)<->(injective_both (relation_of_hedge h))).
    intros.
    generalize (hedge_transpose_NxN h H);intro.
    elim (is_consistent_iff h);intros.
    split;intros.
    elim (H1 H3);clear H1;intros.
    elim (hedge_is_left_consistent_NxN h H);intros.
    elim (hedge_is_left_consistent_NxN (hedge_transpose h) H0);intros.
    unfold injective_both.
    intros.
    split;intro.
    apply (H5 H1 m n H9 m' n');auto.
    elim (hedge_transpose_prop h m n);intros.
    elim (hedge_transpose_prop h m' n');intros.
    apply (H7 H4 n m (H12 H9) n' m');auto.
    apply H14.
    auto.
    apply H2.
    split.
    elim (hedge_is_left_consistent_NxN h H);intros.
    apply H5.
    unfold injective_left;intros.
    unfold injective_both in H3.
    elim (H3 m n H6 m' n' H7);auto.
    elim (hedge_is_left_consistent_NxN (hedge_transpose h) H0);intros.
    apply H5.
    unfold injective_left.
    intros.
    elim (hedge_transpose_prop h n m);intros.
    elim (hedge_transpose_prop h n' m');intros.
    unfold injective_both in H3.
    elim (H3 n m (H10 H6) n' m' (H12 H7));auto.
  Qed.

  (** names and hedges *)
  Definition is_left_name (x:nat)(h:message_relation) :=
    exists m:MessageSpi.message,exists n:MessageSpi.message,(h m n)/\(deBruijn.mem_fn MessageSpi.deBruijn_fn_base x 0 m = true).

  Lemma inclusion_is_left_name : forall (h h':message_relation),(inclusion _ h h')->forall (x:nat),(is_left_name x h)->(is_left_name x h').
    unfold is_left_name;intros.
    ex_elim H0 m.
    ex_elim H0 n.
    exists m;exists n.
    split;auto.
  Qed.

  Lemma same_is_left_name : forall (h h':message_relation),(same_relation _ h h')->forall (x:nat),(is_left_name x h)<->(is_left_name x h').
    intros.
    split;intros.
    apply inclusion_is_left_name with h;auto.
    elim H;auto.
    apply inclusion_is_left_name with h';auto.
    elim H;auto.
  Qed.

  Definition is_left_name_pred (x:nat)(mn:H.elt) :=
    let (m,n) := mn in
      deBruijn.mem_fn MessageSpi.deBruijn_fn_base x 0 m.

  Lemma is_left_name_pred_compat_bool : forall (x:nat),FSetInterface.compat_bool H.E.eq (is_left_name_pred x).
    intros.
    unfold FSetInterface.compat_bool.
    intros.
    destruct x0;destruct y.
    simpl in H.
    and_elim.
    rewrite H;rewrite H0;auto.
  Qed.

  Definition hedge_is_left_name (x:nat)(h:H.t) :=
    (H.exists_ (is_left_name_pred x) h).

  Definition is_left_name_spec (x:nat)(h:H.t) := is_left_name x (relation_of_hedge h).

  Lemma hedge_is_left_name_ok : forall (x:nat)(h:H.t),(is_left_name_spec x h)<->(hedge_is_left_name x h = true).
    unfold hedge_is_left_name, is_left_name_spec, is_left_name.
    split;intros.
    ex_elim H m.
    ex_elim H n.
    apply H.exists_1.
    apply is_left_name_pred_compat_bool.
    unfold H.Exists.
    exists (m,n).
    split;auto.
    elim (H.exists_2 (is_left_name_pred_compat_bool x) H).
    intros.
    and_elim.
    destruct x0.
    exists t.
    exists t0.
    split;auto.
  Qed.

  Lemma mem_fn_left_synthesis : forall (h:message_relation),forall (x:nat),(is_left_name x (hedge_synthesis_relation h))->(is_left_name x h).
    unfold is_left_name.
    intros.
    ex_elim H m.
    ex_elim H n.
    generalize H0;clear H0.
    induction H;intros.
    exists m.
    exists n.
    split;auto.
    rewrite MessageSpi.mem_fn_MEnc_s in H1.
    case (orb_prop _ _ H1);intro;auto.
    rewrite MessageSpi.mem_fn_MEnc_a in H1.
    case (orb_prop _ _ H1);intro;auto.
    rewrite MessageSpi.mem_fn_MPair in H1.
    case (orb_prop _ _ H1);intro;auto.
    rewrite MessageSpi.mem_fn_MOp in H0.
    auto.
  Qed.

  Lemma mem_fn_le_h : forall (g h:message_relation),(le_h g h)->forall (x:nat),(is_left_name x g)->(is_left_name x h).
    intros.
    generalize (le_h_impl_incl_syn _ _ H);intro.    
    apply mem_fn_left_synthesis.
    eapply inclusion_is_left_name.
    unfold inclusion;intros.
    apply H1.
    apply H2.
    auto.
  Qed.

  Definition is_fresh_left (x:nat)(h:message_relation) := ~(is_left_name x h).

  Lemma inclusion_is_fresh_left : forall (h h':message_relation),(inclusion _ h h')->forall (x:nat),(is_fresh_left x h')->(is_fresh_left x h).
    unfold is_fresh_left;intros.
    red;intro.
    apply H0.
    eapply inclusion_is_left_name;eauto.
  Qed.

  Lemma same_is_fresh_left : forall (h h':message_relation),(same_relation _ h h')->forall (x:nat),(is_fresh_left x h)<->(is_fresh_left x h').
    unfold is_fresh_left;intros.
    elim (same_is_left_name _ _ H x);intros.
    split;red;intros;auto.
  Qed.

  Definition hedge_is_fresh_left (x:nat)(h:H.t) := negb (hedge_is_left_name x h).

  Definition is_fresh_left_spec (x:nat)(h:H.t) := is_fresh_left x (relation_of_hedge h).

  Lemma hedge_is_fresh_left_ok : forall (x:nat)(h:H.t),(is_fresh_left_spec x h)<->(hedge_is_fresh_left x h = true).
    unfold hedge_is_fresh_left, is_fresh_left_spec, is_fresh_left.
    intros.
    elim (hedge_is_left_name_ok x h);intros.
    split;intros.
    destruct (hedge_is_left_name x h);try reflexivity.
    elim H1;auto.
    apply H0.
    auto.
    red;intro.
    rewrite (H H2) in H1.
    discriminate H1.
  Qed.

  Lemma not_in_hedge_relation_cond : forall (h:message_relation)(m:MessageSpi.message)(x:nat),(deBruijn.mem_fn MessageSpi.deBruijn_fn_base x 0 m = true)->(is_fresh_left x h)->forall (n:MessageSpi.message),~(h m n).
    intros.
    red;intro.
    apply H0.
    exists m.
    exists n.
    split;auto.
  Qed.

  Lemma mem_fn_left_synthesis_union : forall (g h:message_relation),forall (m:MessageSpi.message),(forall (m' n':MessageSpi.message),g m' n'->exists x:nat,(deBruijn.mem_fn MessageSpi.deBruijn_fn_base x 0 m' = true)/\(deBruijn.mem_fn MessageSpi.deBruijn_fn_base x 0 m = false))->forall (n:MessageSpi.message),(hedge_synthesis_relation (union _ g h) m n)->(hedge_synthesis_relation h m n).
    intros.
    induction H0.
    case H0;intro.
    elim (H _ _ H1);intros.
    and_elim.
    rewrite H2 in H3;discriminate H3.
    apply SynInc;auto.
    apply SynEnc_s.
    apply IHhedge_synthesis_relation1.
    intros.
    elim (H _ _ H0);intro x;intros;and_elim.
    exists x.
    split;auto.
    rewrite MessageSpi.mem_fn_MEnc_s in H2.
    elim (orb_false_elim _ _ H2);intros;auto.
    apply IHhedge_synthesis_relation2.
    intros.
    elim (H _ _ H0);intro x;intros;and_elim.
    exists x.
    split;auto.
    rewrite MessageSpi.mem_fn_MEnc_s in H2.
    elim (orb_false_elim _ _ H2);intros;auto.
    apply SynEnc_a.
    apply IHhedge_synthesis_relation1.
    intros.
    elim (H _ _ H0);intro x;intros;and_elim.
    exists x.
    split;auto.
    rewrite MessageSpi.mem_fn_MEnc_a in H2.
    elim (orb_false_elim _ _ H2);intros;auto.
    apply IHhedge_synthesis_relation2.
    intros.
    elim (H _ _ H0);intro x;intros;and_elim.
    exists x.
    split;auto.
    rewrite MessageSpi.mem_fn_MEnc_a in H2.
    elim (orb_false_elim _ _ H2);intros;auto.
    apply SynPair.
    apply IHhedge_synthesis_relation1.
    intros.
    elim (H _ _ H0);intro x;intros;and_elim.
    exists x.
    split;auto.
    rewrite MessageSpi.mem_fn_MPair in H2.
    elim (orb_false_elim _ _ H2);intros;auto.
    apply IHhedge_synthesis_relation2.
    intros.
    elim (H _ _ H0);intro x;intros;and_elim.
    exists x.
    split;auto.
    rewrite MessageSpi.mem_fn_MPair in H2.
    elim (orb_false_elim _ _ H2);intros;auto.
    apply SynOp.
    auto.
  Qed.

  Definition is_right_name (x:nat)(h:message_relation) := is_left_name x (transp _ h).

  Definition is_right_name_spec (x:nat)(h:H.t) := is_right_name x (relation_of_hedge h).

  Lemma is_right_name_spec_iff_is_left_name_transpose : forall (x:nat)(h:H.t),(is_right_name_spec x h)<->(is_left_name_spec x (hedge_transpose h)).
    unfold is_right_name_spec, is_left_name_spec, is_right_name.
    split;intros.
    elim H.
    intro m;intros.
    ex_elim H0 n.
    exists m.
    exists n.
    split;auto.
    elim (same_transpose_transp h).
    intros.
    apply H3.
    auto.
    elim H.
    intro m;intros.
    ex_elim H0 n.
    exists m.
    exists n.
    split;auto.
    elim (same_transpose_transp h).
    intros.
    apply H2.
    auto.
  Qed.

  Lemma inclusion_is_right_name : forall (h h':message_relation),(inclusion _ h h')->forall (x:nat),(is_right_name x h)->(is_right_name x h').
    unfold is_right_name;intros.
    eapply inclusion_is_left_name with (transp _ h).
    red;intros.
    unfold transp.
    apply H.
    unfold transp in H1;auto.
    auto.
  Qed.

  Lemma same_is_right_name : forall (h h':message_relation),(same_relation _ h h')->forall (x:nat),(is_right_name x h)<->(is_right_name x h').
    intros.
    elim H;intros.
    split;intro;eapply inclusion_is_right_name;eauto.
  Qed.

  Definition is_fresh_right (x:nat)(h:message_relation) := ~(is_right_name x h).

  Definition is_fresh_right_spec (x:nat)(h:H.t) := is_fresh_right x (relation_of_hedge h).

  Lemma is_fresh_right_spec_iff_is_fresh_left_transpose : forall (x:nat)(h:H.t),(is_fresh_right_spec x h)<->(is_fresh_left_spec x (hedge_transpose h)).
    unfold is_fresh_right_spec, is_fresh_left_spec.
    intros.
    unfold is_fresh_right, is_fresh_left.
    elim (is_right_name_spec_iff_is_left_name_transpose x h).
    unfold is_right_name_spec, is_left_name_spec.
    intros.
    split;intros;red;intro.
    apply H1;auto.
    apply H1;auto.
  Qed.

  Lemma inclusion_is_fresh_right : forall (h h':message_relation),(inclusion _ h h')->forall (x:nat),(is_fresh_right x h')->(is_fresh_right x h).
    unfold is_fresh_right;intros.
    red;intro.
    apply H0.
    eapply inclusion_is_right_name;eauto.
  Qed.

  Lemma same_is_fresh_right : forall (h h':message_relation),(same_relation _ h h')->forall (x:nat),(is_fresh_right x h)<->(is_fresh_right x h').
    unfold is_fresh_right.
    intros.
    elim (same_is_right_name _ _ H x);intros.
    split;red;intros;auto.
  Qed.

  Lemma is_reflexive_left_name : forall (h:message_relation),(is_reflexive h)->forall (x:nat),(is_left_name x h)<->(is_right_name x h).
    intros.
    split;intros.
    elim H0.
    intro m;intros.
    ex_elim H1 n.
    unfold is_right_name.
    generalize (H _ _ H1).
    intros.
    subst n.
    exists m.
    exists m.
    split;auto.
    elim H0.
    intro m;intros.
    ex_elim H1 n.
    generalize (H _ _ H1).
    intros.
    subst n.
    exists m.
    exists m.
    split;auto.
  Qed.

  Lemma is_reflexive_fresh_left : forall (h:message_relation),(is_reflexive h)->forall (x:nat),(is_fresh_left x h)<->(is_fresh_right x h).
    unfold is_fresh_right.
    unfold is_fresh_left.
    intros.
    split;red;intros.
    apply H0.
    elim (is_reflexive_left_name h H x);auto.
    apply H0.
    elim (is_reflexive_left_name h H x);auto.
  Qed.

  Definition are_fresh_names (h:H.t)(B:H.t) := 
    (hedge_subset_NxN B) /\ 
    (injective_both (relation_of_hedge B)) /\ 
    (forall (b1 b2:nat),(H.In (MessageSpi.MNam (Name.Ref b1),MessageSpi.MNam (Name.Ref b2)) B)->
      forall (m n:MessageSpi.message),(H.In (m,n) h)->
        (deBruijn.mem_fn MessageSpi.deBruijn_fn_base b1 0 m)=false/\
        (deBruijn.mem_fn MessageSpi.deBruijn_fn_base b2 0 n)=false).

  Lemma are_fresh_names_Subset : forall (h h' B B':H.t),(H.Subset h h')->(H.Subset B B')->(are_fresh_names h' B')->(are_fresh_names h B).
    unfold are_fresh_names;intros;and_elim.
    split.
    eapply hedge_subset_NxN_Subset;eauto.
    split.
    eapply injective_both_inclusion.
    apply Subset_impl_inclusion.
    apply H0.
    apply H2.
    intros.
    apply H3;auto.
  Qed.

  Lemma are_fresh_names_synthesis : 
    forall (h:H.t)(B:H.t),(are_fresh_names h B)->
      forall (m n:MessageSpi.t),(hedge_synthesis_spec (H.union h B) m n)->
        (forall (b1 b2:nat),(H.In (MessageSpi.MNam (Name.Ref b1),MessageSpi.MNam (Name.Ref b2)) B)->
          (deBruijn.mem_fn MessageSpi.deBruijn_fn_base b1 0 m)=false)->
        (hedge_synthesis_spec h m n).
    unfold hedge_synthesis_spec, relation_of_hedge.
    intros.
    induction H0.
    case (H.union_1 H0);intro.
    apply SynInc.
    auto.
    elim H;intros;and_elim.
    elim (H3 _ _ H2).
    intro a;intros.
    ex_elim H6 b.
    destruct a as [b1].
    destruct b as [b2].
    subst m;subst n.
    generalize (H1 _ _ H2).
    intros.
    simpl in H6.
    generalize (Name.fn_0_name b1).
    intro.
    simpl in H7.
    rewrite H7 in H6;discriminate H6.
    apply SynEnc_s.
    apply IHhedge_synthesis_relation1.
    intros.
    generalize (H1 _ _ H0).
    rewrite MessageSpi.mem_fn_MEnc_s.
    intro.
    elim (orb_false_elim _ _ H2);auto.
    apply IHhedge_synthesis_relation2.
    intros.
    generalize (H1 _ _ H0).
    rewrite MessageSpi.mem_fn_MEnc_s.
    intro.
    elim (orb_false_elim _ _ H2);auto.
    apply SynEnc_a.
    apply IHhedge_synthesis_relation1.
    intros.
    generalize (H1 _ _ H0).
    rewrite MessageSpi.mem_fn_MEnc_a.
    intro.
    elim (orb_false_elim _ _ H2);auto.
    apply IHhedge_synthesis_relation2.
    intros.
    generalize (H1 _ _ H0).
    rewrite MessageSpi.mem_fn_MEnc_a.
    intro.
    elim (orb_false_elim _ _ H2);auto.
    apply SynPair.
    apply IHhedge_synthesis_relation1.
    intros.
    generalize (H1 _ _ H0).
    rewrite MessageSpi.mem_fn_MPair.
    intro.
    elim (orb_false_elim _ _ H2);auto.
    apply IHhedge_synthesis_relation2.
    intros.
    generalize (H1 _ _ H0).
    rewrite MessageSpi.mem_fn_MPair.
    intro.
    elim (orb_false_elim _ _ H2);auto.
    apply SynOp.
    apply IHhedge_synthesis_relation.
    intros.
    generalize (H1 _ _ H2).
    rewrite MessageSpi.mem_fn_MOp.
    auto.
  Qed.

  Lemma are_fresh_names_left_consistent : forall (h:H.t),(is_left_consistent h)->forall (B:H.t),(are_fresh_names h B)->(is_left_consistent (H.union h B)).
    intros.
    elim H;clear H;intros;and_elim.
    elim H0;intros;and_elim.
    elim (hedge_is_consistent_NxN _ H9);intros.
    generalize (H13 H10);clear H12;clear H13;intros.
    elim H12;clear H12;intros.
    clear H13.
    elim H12;clear H12;intros;and_elim.
    split;unfold relation_of_hedge.
    unfold name_left_name;intros.
    case (H.union_1 H21);intros.
    apply (H _ _ H23 _ H22).
    apply (H12 _ _ H23 _ H22).
    split.
    unfold injective_left;intros.
    case (H.union_1 H21);case (H.union_1 H22);intros.
    apply (H1 _ _ H25 _ _ H24 H23).
    elim (H9 _ _ H24);intro a;intros.
    ex_elim H26 b.
    subst m'.
    subst m;subst n'.
    destruct a as [b1].
    destruct b as [b2].
    elim (H11 _ _ H24 _ _ H25).
    intros.
    simpl in H23.
    generalize (Name.fn_0_name b1).
    simpl.
    intros.
    rewrite H27 in H23;discriminate H23.
    elim (H9 _ _ H25);intro a;intros.
    ex_elim H26 b.
    subst m.
    subst m';subst n.
    destruct a as [b1].
    destruct b as [b2].
    elim (H11 _ _ H25 _ _ H24).
    intros.
    simpl in H23.
    generalize (Name.fn_0_name b1).
    simpl.
    intros.
    rewrite H27 in H23;discriminate H23.
    apply (H13 _ _ H25 _ _ H24 H23).
    split.
    unfold op_synthesis_left;intros.
    red;intro.
    case (H.union_1 H21);intro.
    apply (H2 _ _ H24 _ _ H22 n').
    apply (are_fresh_names_synthesis _ _ H0 _ _ H23).
    intros.
    elim (H11 _ _ H25 _ _ H24).
    intros.
    subst m.
    rewrite MessageSpi.mem_fn_MOp in H26;auto.
    elim (H9 _ _ H24).
    intro a;intros.
    ex_elim H25 b.
    subst m;discriminate H25.
    split.
    unfold pair_free_left;intros.
    red;intro.
    case (H.union_1 H21);intro.
    apply (H3 _ _ _ H22).
    apply (H15 _ _ _ H22).
    split.
    unfold enc_s_synthesis_left;intros.
    case (H.union_1 H21);intro.
    case (H4 _ _ H23 _ _ H22);intro.
    left.
    intros.
    red;intro.
    apply (H24 n').
    apply (are_fresh_names_synthesis _ _ H0 _ _ H25).
    intros.
    elim (H11 _ _ H26 _ _ H23).
    subst m;intros.
    rewrite MessageSpi.mem_fn_MEnc_s in H22.
    elim (orb_false_elim _ _ H22);auto.
    right.
    intro.
    red;intro.
    apply (H24 l).
    apply (are_fresh_names_synthesis _ _ H0 _ _ H25).
    intros.
    elim (H11 _ _ H26 _ _ H23).
    subst m;intros.
    rewrite MessageSpi.mem_fn_MEnc_s in H22.
    elim (orb_false_elim _ _ H22);auto.
    elim (H9 _ _ H23);intro a;intros.
    ex_elim H24 b.
    subst m;discriminate H24.
    split.
    unfold enc_s_analysis_left;intros.
    case (H.union_1 H21);intro.
    assert (hedge_synthesis_spec h k l).
    apply (are_fresh_names_synthesis _ _ H0 _ _ H23).
    intros.
    elim (H11 _ _ H25 _ _ H24).
    subst m;intros.
    rewrite MessageSpi.mem_fn_MEnc_s in H22.
    elim (orb_false_elim _ _ H22);auto.
    elim (H5 _ _ H24 _ _ H22 _ H25).
    intro n';intros;and_elim.
    exists n'.
    split;auto.
    generalize incl_impl_le_h.
    unfold le_h.
    intros.
    eapply H28;eauto.
    red;intros.
    apply H.union_2.
    auto.
    elim (H9 _ _ H24);intro a;intros.
    ex_elim H25 b.
    subst m;discriminate H25.
    split.
    unfold enc_a_synthesis_left;intros.
    case (H.union_1 H21);intro.
    case (H6 _ _ H23 _ _ H22);intro.
    left.
    intros.
    red;intro.
    apply (H24 n').
    apply (are_fresh_names_synthesis _ _ H0 _ _ H25).
    intros.
    elim (H11 _ _ H26 _ _ H23).
    subst m;intros.
    rewrite MessageSpi.mem_fn_MEnc_a in H22.
    elim (orb_false_elim _ _ H22);auto.
    right.
    intro.
    red;intro.
    apply (H24 l).
    apply (are_fresh_names_synthesis _ _ H0 _ _ H25).
    intros.
    elim (H11 _ _ H26 _ _ H23).
    subst m;intros.
    rewrite MessageSpi.mem_fn_MEnc_a in H22.
    elim (orb_false_elim _ _ H22);auto.
    elim (H9 _ _ H23);intro a;intros.
    ex_elim H24 b.
    subst m;discriminate H24.
    split.
    unfold enc_a_analysis_left;intros.
    case (H.union_1 H21);intro.
    assert (hedge_synthesis_spec h k' l').
    apply (are_fresh_names_synthesis _ _ H0 _ _ H24).
    intros.
    elim (H11 _ _ H26 _ _ H25).
    subst m;intros.
    rewrite MessageSpi.mem_fn_MEnc_a in H22.
    elim (orb_false_elim _ _ H22);intros.
    rewrite (MessageSpi.mem_fn_inverse_key 0 b1 _ _ H23) in H29;auto.
    generalize (H7 _ _ H25 _ _ H22 _ H23 _ H26).
    intros.
    ex_elim H27 l.
    ex_elim H27 n'.
    exists l.
    exists n'.
    split;auto.
    split;auto.
    generalize incl_impl_le_h.
    unfold le_h.
    intros.
    eapply H30;eauto.
    red;intros.
    apply H.union_2.
    auto.
    elim (H9 _ _ H25);intro a;intros.
    ex_elim H26 b.
    subst m;discriminate H26.
    unfold injective_inv_left;intros.
    case (H.union_1 H21);case (H.union_1 H22);intros.
    apply (H8 _ _ H25 _ _ H24 H23).
    elim (H9 _ _ H24);intro a;intros.
    ex_elim H26 b.
    subst m';subst n'.
    destruct a as [b1].
    destruct b as [b2].
    elim (H11 _ _ H24 _ _ H25).
    intros.
    symmetry in H23.
    rewrite (MessageSpi.mem_fn_inverse_key 0 b1 _ _ H23) in H26.
    simpl in H26.
    generalize (Name.fn_0_name b1).
    simpl.
    intros.
    rewrite H28 in H26;discriminate H26.
    elim (H9 _ _ H25);intro a;intros.
    ex_elim H26 b.
    subst m;subst n.
    destruct a as [b1].
    destruct b as [b2].
    elim (H11 _ _ H25 _ _ H24).
    intros.
    symmetry in H23.
    rewrite <- (MessageSpi.mem_fn_inverse_key 0 b1 _ _ H23) in H26.
    simpl in H26.
    generalize (Name.fn_0_name b1).
    simpl.
    intros.
    rewrite H28 in H26;discriminate H26.
    apply (H20 _ _ H25 _ _ H24 H23).
  Qed.

  Lemma are_fresh_names_transpose : forall (h:H.t)(B:H.t),are_fresh_names h B -> are_fresh_names (hedge_transpose h) (hedge_transpose B).
    unfold are_fresh_names.
    intros.
    and_elim.
    split.
    apply hedge_transpose_NxN;auto.
    split.
    unfold relation_of_hedge.
    unfold injective_both;intros.
    generalize (hedge_transpose_prop B).
    intros.
    elim (H4 n m);intros.
    elim (H4 n' m');intros.
    generalize (H0 _ _ (H6 H2) _ _ (H8 H3)).
    tauto.
    intros.
    generalize (hedge_transpose_prop B);intros.
    generalize (hedge_transpose_prop h);intros.
    elim (H4 (MessageSpi.MNam (Name.Ref b2)) (MessageSpi.MNam (Name.Ref b1))).
    intros.
    elim (H5 n m);intros.
    generalize (H1 _ _ (H7 H2) _ _ (H9 H3)).
    tauto.
  Qed.

(*   Inductive pairwise_distinct : list Name.t -> Prop := *)
(*   | pwdNil : pairwise_distinct nil *)
(*   | pwdCons : forall (l:list Name.t),pairwise_distinct l -> forall (n:Name.t),~(In n l)->(pairwise_distinct (cons n l)). *)

  Definition are_fresh_left (h:H.t)(l:list Name.t) :=
    forall (x:nat),(In (Name.Ref x) l)->(is_fresh_left_spec x h).

  Lemma are_fresh_left_Subset : forall (h h':H.t)(l:list Name.t),(H.Subset h h')->(are_fresh_left h' l)->(are_fresh_left h l).
    unfold are_fresh_left;intros.
    unfold is_fresh_left_spec.
    eapply inclusion_is_fresh_left.
    apply Subset_impl_inclusion.
    apply H.
    apply H0;auto.
  Qed.

  Fixpoint are_fresh_left_proc (h:H.t)(l:list Name.t) {struct l} : bool :=
    match l with
      | nil => true
      | cons (Name.Ref x) l => 
        andb (hedge_is_fresh_left x h) (are_fresh_left_proc h l)
    end.

  Lemma are_fresh_left_proc_ok : forall (h:H.t)(l:list Name.t),(are_fresh_left h l)<->(are_fresh_left_proc h l = true).
    unfold are_fresh_left.
    split;intros.
    induction l.
    reflexivity.
    destruct a.
    simpl.
    rewrite IHl.
    elim (hedge_is_fresh_left_ok n h).
    intros.
    rewrite H0.
    reflexivity.
    apply H.
    simpl.
    left;auto.
    intros.
    apply H.
    simpl.
    right;auto.
    induction l.
    inversion H0.
    destruct a.
    simpl in H.
    elim (andb_prop _ _ H);clear H;intros.
    simpl in H0.
    case H0.
    intros.
    injection H2;intros;subst x.
    elim (hedge_is_fresh_left_ok n h);auto.
    auto.
  Qed.

  Definition are_fresh_right (h:H.t)(l:list Name.t) :=
    forall (x:nat),(In (Name.Ref x) l)->(is_fresh_right_spec x h).

  Lemma are_fresh_right_Subset : forall (h h':H.t)(l:list Name.t),(H.Subset h h')->(are_fresh_right h' l)->(are_fresh_right h l).
    unfold are_fresh_right;intros.
    unfold is_fresh_right_spec.
    eapply inclusion_is_fresh_right.
    apply Subset_impl_inclusion.
    apply H.
    apply H0;auto.
  Qed.

  Lemma are_fresh_right_iff_are_fresh_left_transpose : forall (h:H.t)(l:list Name.t),(are_fresh_right h l)<->(are_fresh_left (hedge_transpose h) l).
    unfold are_fresh_right.
    unfold are_fresh_left.
    split;intros.
    elim (is_fresh_right_spec_iff_is_fresh_left_transpose x h).
    intros.
    apply H1.
    apply H.
    auto.
    elim (is_fresh_right_spec_iff_is_fresh_left_transpose x h).
    intros.
    apply H2.
    apply H.
    auto.
  Qed.

(*   Definition n_pwd (n:nat)(l:list Name.t) := (length l = n) /\ (pairwise_distinct l). *)

  Definition n_pwd (n:nat)(l:list Name.t) := (length l = n) /\ (is_set Name.t l).

  Lemma n_pwd_nil : n_pwd 0 nil.
    split.
    reflexivity.
    apply nil_is_set.
(*     apply pwdNil. *)
  Qed.

  Lemma n_pwd_single : forall (x:Name.t),(n_pwd 1 (cons x nil)).
    intros.
    split.
    reflexivity.
    apply cons_is_set.
    apply nil_is_set.
(*     apply pwdCons. *)
(*     apply pwdNil. *)
    red;intro.
    inversion H.
  Qed.

  Lemma are_fresh_left_nil : forall (h:H.t),(are_fresh_left h nil).
    unfold are_fresh_left.
    intros.
    inversion H.
  Qed.

  Lemma are_fresh_left_single : forall (h:H.t)(k:nat),(are_fresh_left h (cons (Name.Ref k) nil))<->(is_fresh_left_spec k h).
    unfold are_fresh_left.
    split;intros.
    apply H.
    simpl.
    left;auto.
    inversion H0.
    injection H1;intros.
    subst x;auto.
    inversion H1.
  Qed.

  Definition n_fresh_left (h:H.t)(n:nat)(l:list Name.t) := (n_pwd n l) /\ (are_fresh_left h l).

  Lemma n_fresh_left_single : forall (h:H.t)(l:list Name.t),(n_fresh_left h 1 l)->exists k:nat,l=(cons (Name.Ref k) nil) /\ (is_fresh_left_spec k h).
    unfold n_fresh_left.
    intros.
    and_elim.
    unfold n_pwd in H.
    and_elim.
    destruct l;try (discriminate H).
    destruct l;try (discriminate H).
    destruct t.
    exists n.
    split;auto.
    unfold are_fresh_left in H0.
    apply H0.
    simpl.
    left;auto.
  Qed.

  Definition n_fresh_right (h:H.t)(n:nat)(l:list Name.t) := (n_pwd n l) /\ (are_fresh_right h l).

  Lemma is_reflexive_n_fresh_left : forall (h:H.t),(is_reflexive (relation_of_hedge h))->forall (n:nat)(l:list Name.t),(n_fresh_left h n l)->(n_fresh_right h n l).
    intros.
    elim H0.
    intros.
    split;auto.
    unfold are_fresh_right.
    intros.
    unfold are_fresh_left in H2.
    unfold is_fresh_right_spec.
    unfold is_fresh_right.
    red;intro.
    elim (H2 _ H3).
    elim H4.
    intro a;intros.
    ex_elim H5 b.
    unfold transp in H5.
    generalize (H _ _ H5).
    intro.
    subst b.
    exists a.
    exists a.
    split;auto.
  Qed.

  Lemma n_fresh_left_Subset : forall (h h':H.t)(n:nat)(l:list Name.t),(H.Subset h h')->(n_fresh_left h' n l)->(n_fresh_left h n l).
    unfold n_fresh_left;intros.
    and_elim.
    split;auto.
    eapply are_fresh_left_Subset.
    apply H.
    auto.
  Qed.

  Lemma n_fresh_right_Subset : forall (h h':H.t)(n:nat)(l:list Name.t),(H.Subset h h')->(n_fresh_right h' n l)->(n_fresh_right h n l).
    unfold n_fresh_right;intros.
    and_elim.
    split;auto.
    eapply are_fresh_right_Subset.
    apply H.
    auto.
  Qed.

  Lemma n_pwd_permutation : forall (n:nat)(l l':list Name.t),(n_pwd n l)->(is_permutation _ l l')->(n_pwd n l').
    unfold n_pwd;intros.
    and_elim.
    split.
    rewrite <- (permutation_length _ _ _ H0).
    auto.
    eapply is_permutation_set;eauto.
  Qed.

  Lemma n_fresh_left_permutation : forall (h:H.t)(n:nat)(l:list Name.t),(n_fresh_left h n l)->forall (l':list Name.t),(is_permutation _ l l')->(n_fresh_left h n l').
    intros.
    elim H;clear H;intros.
    split.
    eapply n_pwd_permutation;eauto.
    unfold are_fresh_left.
    intros.
    elim (permutation_in _ _ _ H0 (Name.Ref x)).
    intros.
    apply H1.
    auto.
  Qed.

  Lemma n_fresh_right_permutation : forall (h:H.t)(n:nat)(l:list Name.t),(n_fresh_right h n l)->forall (l':list Name.t),(is_permutation _ l l')->(n_fresh_right h n l').
    intros.
    elim H;clear H;intros.
    split.
    eapply n_pwd_permutation;eauto.
    unfold are_fresh_right.
    intros.
    elim (permutation_in _ _ _ H0 (Name.Ref x)).
    intros.
    apply H1.
    auto.
  Qed.
End HedgeOp.






